#
# Copyright (C) 1996 by the Board of Trustees
#    of Leland Stanford Junior University.
# 
# This file is part of the SimOS distribution. 
# See LICENSE file for terms of the license. 
#

## 
## kernel_stats.tcl
##

FileSourced osf/kernel_stats.tcl

timing create kstats
log [timing fields]

###
### statistics logging
###

annotation set simos exit {
    timing exit kstats
    log [timing dump kstats]
    console "Dumped KSTATS\n"
}

annotation set simos sigusr {
    log [timing dump kstats]
}

###
### thread naming
###

if { $PARAM(MACHINE.Count) == 1 } {
    proc ktname {} {
        global PID CPU
		return $PID($CPU)
    }
    proc pid2ktname {pid} {
		return $pid
    }
} else {
    proc ktname {} {
        global M PID CPU
       	return $M($CPU)_$PID($CPU)
    }
    proc pid2ktname {pid} {
        global M CPU
       	return $M($CPU)_$pid
    }
}


###
### thread control
###

proc kstatsSwitch args {
    global PROCESS CYCLES kstatsThread M CPU PARAM
    global CPU 

#     if {$CYCLES > 57522000} {
#         console "KSTATSSWITCH [ktname] with $args on $CPU\n"
#     }
    if { $PARAM(MACHINE.Count) == 1 } {
        eval timing switch kstats [ktname] $args
    } else {
        eval timing switch kstats [ktname] machine$M($CPU) $args
    }

    if ![info exists kstatsThread([ktname])] {
        set kstatsThread([ktname]) "runq"
    } else {
        timing end kstats $kstatsThread([ktname])
    }
}

annotation set simos enter {
    if [inUserMode] {
        kstatsSwitch 
    } elseif {$PROCESS($CPU) == "boot"} {
        kstatsSwitch bootproc
    } elseif {$PROCESS($CPU) == "kernel idle"} {
        kstatsSwitch idleproc idle
    } else {
        kstatsSwitch startup
    }
}

annotation set osEvent switchOut {
    # has thread still active (i.e. has not exited)
    if [info exists kstatsThread([ktname])] {
        if {[timing current kstats] == "ROOT"} { 
            console "KSTATS: CPU $CPU CYCLE $CYCLES: bad DESCHED pid $PID($CPU)\n"
        }
        
        timing start kstats $kstatsThread([ktname])
    }
}


proc kernelThreadState {proc} {
    global kernelThreadNames

   if [info exists kernelThreadNames($proc)] {
        return $proc
    } else {
        return fork
    }
}

annotation set osEvent switchIn {
    if {$PROCESS($CPU) == "kernel idle"} {
        kstatsSwitch idleproc
    } elseif {$PROCESS($CPU) == "runq"} {
        kstatsSwitch runqproc
    } else {
        kstatsSwitch [kernelThreadState $PROCESS($CPU)]
    }
}

annotation set osEvent procexit {
    timing terminate kstats [ktname]
    unset kstatsThread([ktname])
}



###
### kernel services
###

annotation set osEvent startPal {
    timing start? kstats PAL
}

annotation set pc $PALPC(TRAP_NDTB_MISS) {
    timing vector kstats dtlb
}
annotation set pc $PALPC(TRAP_PDTB_MISS) {
    timing vector kstats 2xtlb
}
annotation set pc $PALPC(TRAP_ITB_MISS) {
    timing vector kstats itlb
}

annotation set osEvent endPal {
    timing end kstats [timing current kstats]
}

annotation set osEvent startKernel {
    #ugly hack because gamma/kappa mess up on the pc
    if {$PARAM(CPU.Model) == "KAPPA"} {
        set t_pc [hex [expr $pc - 4]]
    } elseif {$PARAM(CPU.Model) == "GAMMA"} {
        set t_pc [hex [expr $pc - 4]]
    } else {
        set t_pc $pc
    }

    if {$KENTRY($t_pc) == "SysCall"} {
        set syscall_dec [expr $v0 + 0]
        if [catch {set syscall $osf1_syscall($syscall_dec)} msg] {
            set syscall "SysCall"
        }
        timing start kstats $syscall
    } elseif {$KENTRY($t_pc) == "Intr"} {
        set intr_dec [expr $a0 + 0]
        if [catch {set itype $osf1_intr($intr_dec)} msg] {
            set itype "Intr"
        }
        timing start kstats $itype
    } elseif {$KENTRY($t_pc) == "MM_FAULT"} {
        set mm_dec [expr $a1 + 0]
        if [catch {set mmtype $osf1_MM($mm_dec)} msg] {
            set mmtype "MM_FAULT"
        }
        timing start kstats $mmtype
    } else {
        timing start kstats $KENTRY($t_pc)
    }
}

annotation set osEvent endKernel {
    #return from callsys or interrupt
    set current [timing current kstats]
    if {($current != "ROOT") && ($current != "idle") && ($current != "sync")} {
        timing end kstats $current
    } else {
        console "KSTATS: CPU $CPU CYCLE $CYCLES: bad rfe pid $PID($CPU) $PROCESS($CPU) $current\n"
        timing dumpstack kstats
    }
}

annotation set osEvent startIdle {
    timing start kstats idle
}

annotation set osEvent endIdle {
    timing end kstats idle
}

annotation set osEvent startSync {
    timing start kstats sync
}

annotation set osEvent endSync {
    timing end kstats sync
}
# If lock is hard to get (too many tries)
# then it goes to a more involved routine
# simple_lock_miss
# Track this routine as a regular kstats routine
annotation set pc vmunix::simple_lock_miss:START {
    timing start kstats simple_lock_miss
}
# kernel dependent (ret in the middle of a routine)
annotation set pc $simple_lock_miss_end {
    timing end kstats simple_lock_miss
}

# Desched for disk and other reasons
annotation set pc vmunix::simos_go:START {
    set kstatsThread([ktname]) diskIO
}

annotation set pc vmunix::assert_wait_mesg_head:START {
    if {$a0 != 0} {
        if {$kstatsThread([ktname]) == "runq"} {
            set kstatsThread([ktname]) semawait
        }
    }
}
annotation set pc vmunix::assert_wait_mesg:START {
    if {$a0 != 0} {
        if {$kstatsThread([ktname]) == "runq"} {
            set kstatsThread([ktname]) semawait
        }
    }
}
annotation set pc vmunix::assert_wait_mesg_timo:START {
    if {$a0 != 0} {
        if {$kstatsThread([ktname]) == "runq"} {
            set kstatsThread([ktname]) semawait
        }
    }
}
