/*
 * Copyright (C) 1996 by the Board of Trustees
 *    of Leland Stanford Junior University.
 * 
 * This file is part of the SimOS distribution. 
 * See LICENSE file for terms of the license. 
 *
 */

/* 
 * ethersim.c --
 *
 *	 Simuation of a local ethernet with a gateway interface.
 *	 This program simulates a local ethernet with a gateway machine
 *	 attached to it.  The simulated ethernet accepts ethernet packets
 *	 that arrive on the special UDP port.  It is assumed that these
 *	 packets are send by the network module of simos services running
 *	 on top of Unix.  This program decodes the ethernet header and 
 *	 uses it to determine which UDP port to output the packet on.
 *	 This is done by keeping track of the ethernet address of the machine
 *	 sending packets to the ethernet port.
 *
 *	 When this program receives an ethernet packet of type IP to an IP
 *	 destination address outside the subnet of the local ethernet, it
 *	 outputs the packet to the local (phsyical) ethernet attached to 
 *	 the machine.  By doing this the program acts like a gateway machine.
 *
 *	 The program also opens the network interface of the machines and 
 *	 grabs any packets with an IP destination of the simulated subnet.  
 *	 This is done using Sun's NIT protocol and packet filter.  Packets
 *	 addressed to a known simulated host will be forward on to that 
 *	 host's UDP port.  
 *
 *	 NOTES:
 *	    1) Because this program uses NIT to get packets from the outside
 *	       packets arriving at network interfaces other than the one
 *	       specified are not seen.  In particular, packets sent thru
 *	       the loopback (lo0) interface are not seen by this program. 
 *	       This means that processes on the local machine can't talk IP
 *	       to the simulated machines.  
 *	    2) This program is a big security hole because it accepts packets
 *	       from an non-privledged UDP port and forwards them on to the
 *	       local ethernet.  I think that outputPacket() does enough 
 *	       checking to stop this from being a problem.
 *
 * DEC Alpha port: Artur Klauser, University of Colorado at Boulder
 *                 klauser@cs.colorado.edu
 *
 * $Id: ethersim.c,v 1.1 1998/09/08 23:48:27 verghese Exp $
 * $Log: ethersim.c,v $
 * Revision 1.1  1998/09/08 23:48:27  verghese
 * This is the initial checkin of ethersim for digital UNIX.
 * The cleanup and initial port of this version was done by Artur
 * Clauser at Colorado, based on the code for the Sun and SGI versions.
 *
 * I have modified and debugged this code to support the proxy
 * form of ethersim, as this can be run without configuration
 * changes to a router.
 *
 * Revision 1.1  1998/02/19 07:10:45  klauser
 * SimOS ethersim DEC Alpha port - initial revision
 *
 */


#include <sys/types.h>
#include <sys/param.h>
#include <sys/stat.h>
#include <sys/ioctl.h>
#include <sys/socket.h>
#include <sys/file.h>
#include <sys/wait.h>
#include <sys/time.h>

#include <net/if.h>
#include <net/pfilt.h>
#include <netinet/in_systm.h>
#include <netinet/in.h>
#include <netinet/ip.h>
#include <netinet/udp.h>
#include <netinet/tcp.h>
#include <netinet/ip_icmp.h>
#include <netinet/if_ether.h>

#include <arpa/inet.h>

#include <stdio.h>
#include <signal.h>
#include <pwd.h>
#include <setjmp.h>
#include <netdb.h>
#include <errno.h>
#include <strings.h>
#include <varargs.h>
#include <stropts.h>
#include <poll.h>
#include <search.h>
#include <memory.h>
#include <stdlib.h>
#include <unistd.h>

#include "hash.h"
#include "ethersim.h"

/*
 * Program options setable from the command line.
 */

in_addr_t subnetaddr = 0x808af1c0; /* Subnet address (i.e. 128.138.241.192). */

int    simportnumber = ETHER_SIM_PORT;  /* UDP port of simulated ethernet. */
int    noipforwarding = 0;		/* Boolean - Don't forward ip packet. */
char   interfacename[32] = "tu0";	/* Name of interface to forward IP 
					 * packets on. */
int    verbose = 0;                     /* verbose: tell user about pkts */

in_addr_t   routeraddr = (in_addr_t)-1;
/*
 * Buffer to hold the current packet being transmitted on the ethernet. 
 * This must be big enought to hold an ethernet MTU and its header.  We
 * add 128 bytes for good measure.
 */
char curPacket[ETHERMTU + sizeof(struct ether_header) + 128];

int	simfd;		/* Open file descriptor of simulated ethernet port. */
int	netfd;          /* Open file descriptor of network interface tap. */

/*
 * The ethernet address and IP network address of the interface being used.
 */

struct ether_addr myetheraddr;
in_addr_t mynetaddr;

/*
 * The internet address and netmask of the interface being used.
 */
in_addr_t myinetAddr;
in_addr_t mynetmask;
in_addr_t hostnetmask = 0;

/*
 * HostInfoRecord - Info kept per simulated host.
 */

typedef struct HostInfoRecord {
    struct ether_addr etheraddr;/* Ethernet address of host. */
    in_addr_t inetAddr;		/* Internet address of host. */
    struct sockaddr fromaddr;	/* Socket the of the simulated machines 
				 * ethernet interface. */
    int	 fromlen;		/* Length in bytes of fromaddr. */

} HostInfoRecord;

/*
 * Hash tables that map 
 *	etheraddrTable - Ethernet addresses to HostInfoRecord.
 *	ipaddrTable    - IP addresses to HostInfoRecord. 
 */

Hash_Table etheraddrTable;
Hash_Table ipaddrTable;

#define MAX_ADDRS 16
u_long nproxys = 0;
in_addr_t proxyaddr[MAX_ADDRS];


/*
 * offsetof() return the offset of a field into the structure.
 */
#ifndef offsetof
#define	offsetof(_t, _f)  ((int) &(((_t *) 0)->_f))
#endif

#define D(x) (hdrPtr->ether_dhost[x])
#define S(x) (hdrPtr->ether_shost[x])

/*
 * Forward routine declartions.
 */
void dispEtherPacket _ARGS_((int n, struct ether_header* hdrPtr, int tab));
int serviceSimPort _ARGS_((void));
int serviceNetPort _ARGS_((void));
int initNITdevice  _ARGS_((char *device, in_addr_t subNetAddr, in_addr_t netmask));
int initNITdeviceFP  _ARGS_((char *device, in_addr_t *proxyaddr, u_long nproxy, in_addr_t netmask));
int outputPacket   _ARGS_((char *packet, int size));
int lookupInterface _ARGS_((char *interfacename, in_addr_t *inetAddrPtr, 
				in_addr_t *netmaskPtr));
int validPacket _ARGS_((char *packet, int size, struct sockaddr *fromAddr, 
				int fromlen));
int ArpRequest _ARGS_((void *pin, int pinlen, void *pout, int *poutlen));

/* Digital Unix missing proto */
int pfopen _ARGS_((char *ifname, int flags));


/*
 *----------------------------------------------------------------------
 *
 * main -- 
 *
 *       Main routine and loop for ethersim.
 *
 * Results:
 *      Only returns if an error occurs.
 *
 * Side effects:
 *
 *----------------------------------------------------------------------
 */
int
main(argc, argv)
int argc;
char *argv[];
{
    int	c, errflg;
    struct sockaddr_in addr;
    int	   nfds;
    struct pollfd  fds[3];
    int	   (*serviceRoutine[3])();

    simfd = netfd = -1;
    errflg = 0;

#define OPTS "vfg:p:r:i:x:n:"
    while ((c = getopt(argc, argv,  OPTS)) != -1) {
       switch (c) {
        case 'v':
	     verbose = 1;
	     break;
	case 'f':
	     noipforwarding = 1;
	     break;
	case 'i':
	     strncpy(interfacename, optarg, sizeof(optarg)-1);
	     break;
	case 'g':
	     subnetaddr = inet_network(optarg);
	     if (subnetaddr == (in_addr_t) -1) {
		 fprintf(stderr, "%s: Malformed gateway address %s\n", argv[0],
					optarg);
		 errflg++;
	     }
	     break;
        case 'x':
           proxyaddr[nproxys] = inet_network(optarg);
           if (proxyaddr[nproxys] == (u_long) -1) {
              fprintf(stderr, "%s: Malformed proxy address %s\n", argv[0],
                      optarg);
              errflg++;
           }
           nproxys++;
           break;
        case 'n':
           hostnetmask = inet_network(optarg);
           if (hostnetmask == (u_long) -1) {
              fprintf(stderr, "%s: Malformed netmask %s\n", argv[0],
                      optarg);
              errflg++;
           }
           break;
	case 'r':
	     routeraddr = inet_network(optarg);
	     if (routeraddr == (in_addr_t) -1) {
		 fprintf(stderr, "%s: Malformed router address %s\n", argv[0],
					optarg);
		 errflg++;
	     }
	     break;
	case 'p':
	     simportnumber = atoi(optarg);
	     if (simportnumber < 0) {
		 fprintf(stderr, "%s: Illegal port number %s\n", argv[0], 
						optarg);
		 errflg++;
	     }
	     break;
	case '?':
	     errflg++;
	}
    }
    if (errflg) {
	(void)fprintf(stderr, 
		"usage: %s [-v] [-g netaddr] [-p portnumber] [-n netmask]\n", 
				argv[0]);
	exit (2);
    }


    /*
     * Create the socket that transmits on the simulated ethernet will be
     * sent.  It is a UDP port number ETHER_SIM_PORT.  The mostly likely
     * cause of failure is another ethersim running causing the bind
     * to fail with address in use.
     */

    simfd = socket(AF_INET, SOCK_DGRAM, 0);
    if (simfd < 0) {
        perror("main: socket");
        exit(1);
    }


    bzero((char *)&addr, sizeof(addr));
    addr.sin_port = htons(simportnumber);
    addr.sin_family = AF_INET;
    if (bind(simfd, (struct sockaddr *) &addr, sizeof(addr)) < 0) {
        perror("main: bind");
        exit(2);
    }

    if (!noipforwarding && (geteuid() != 0)) {
	fprintf(stderr, "%s: not running as root - IP forwarding disabled\n",
			argv[0]);
	noipforwarding = 1;
    }

    if (!noipforwarding) {
	if (lookupInterface(interfacename, &myinetAddr, &mynetmask) < 0) {
	    fprintf(stderr, "Couldn't find interface %s\n", interfacename);
	    exit(3);
	}
	{
		struct in_addr tmp;
		tmp.s_addr = myinetAddr;
		mynetaddr = inet_netof(tmp);
	}

        if (hostnetmask == 0) {
           hostnetmask = mynetmask;
        }

        if(nproxys > 0) {
           netfd = initNITdeviceFP(interfacename, proxyaddr, nproxys, hostnetmask);
        } else {
           netfd = initNITdevice(interfacename, subnetaddr, hostnetmask);
        }
	if (netfd < 0) {
	    fprintf(stderr, "Couldn't initialize network tap\n");
	    exit(3);
	}
    } 
    /*
     * Initialize mapping hash tables.
     */
    Hash_InitTable(&etheraddrTable, 0, 2);
    Hash_InitTable(&ipaddrTable, 0, HASH_ONE_WORD_KEYS);

    /*
     * Loop waiting for packets to arrive.
     */
    nfds = 0;
    fds[nfds].fd = simfd;
    fds[nfds].events = POLLIN;
    serviceRoutine[nfds] = serviceSimPort;
    nfds++;
    if (netfd >= 0) {
	fds[nfds].fd = netfd;
	fds[nfds].events = POLLIN;
    	serviceRoutine[nfds] = serviceNetPort;
	nfds++;
    }

    if (verbose) {
      fprintf (stderr, "start system poll loop\n");
    }
    while (1) {
        int     n;
	int	i;

	fds[0].revents = fds[1].revents = fds[2].revents = 0;
	n = poll(fds, nfds, -1);
	if (n < 0) {
	    if (errno == EINTR) continue;
	    perror("poll");
	    exit(1);
	}
	for (i = 0; i < nfds; i++) { 
	  if (fds[i].revents & POLLIN) {
	    (void) serviceRoutine[i]();
	  }
	}
    }
}


/*
 * Assumes that eaddr points to a 6 byte ethernet address
 * and that buffer points to a 32 byte character buffer
 * returns a pointer to the filled buffer
 */

u_char *
ether_ntoa(u_char *eaddr, u_char *buffer)
{
   u_char *ptr = buffer;
   int i;

   for (i=0; i<6; i++) {
      ptr += sprintf(ptr, "%x:", eaddr[i]);
   }

   *ptr = (u_char)0;

   return buffer;
}



/*
 *----------------------------------------------------------------------
 *
 * dispEtherPacket -- 
 *
 *       display packet for verbose mode
 *
 *----------------------------------------------------------------------
 */
void
dispEtherPacket(int n, struct ether_header* hdrPtr, int tab)
{
    struct ip *ipPtr;
    struct udphdr *udpPtr;
    struct tcphdr *tcpPtr;
    struct icmp *icmpPtr;
    struct ether_arp *arpPtr;
    char *dataPtr;
    struct in_addr ipdst, ipsrc;
    char blankhead[40];
    u_char ebuf[32];

    if (tab == 0)
       sprintf(blankhead,"%s"," ");
    else if (tab == 1)
       sprintf(blankhead,"%s","\t\t");
    else if (tab == 2)
       sprintf(blankhead,"%s","\t\t\t\t");


    if (0) {
      u_char *p = (u_char*) hdrPtr;
      int i;

      for (i = 0; i < n; i++) {
        if ((i == 14) || (i == 34)) {
          printf("\n");
        }
        printf("%02x ", p[i]);
      }
      printf("\n");
    }

    printf("%s<D>%x:%x:%x:%x:%x:%x\n",blankhead,D(0),D(1),D(2),D(3),D(4),D(5));
    printf("%s<S>%x:%x:%x:%x:%x:%x\n",blankhead,S(0),S(1),S(2),S(3),S(4),S(5));
    if (ntohs(hdrPtr->ether_type) == ETHERTYPE_IP){
       printf("%sether_type = IP\n",blankhead);
    } else if (ntohs(hdrPtr->ether_type)  == ETHERTYPE_ARP) {
       printf("%sether_type = ARP\n",blankhead);
       arpPtr = (struct ether_arp *) ((char*)hdrPtr + sizeof(struct ether_header));
       bcopy((u_char*)& arpPtr->arp_spa, (u_char*)& ipsrc.s_addr ,sizeof(struct in_addr));
       bcopy((u_char*)& arpPtr->arp_tpa, (u_char*)& ipdst.s_addr ,sizeof(struct in_addr));
       printf("%s<S>%s\n",blankhead,
               ether_ntoa(arpPtr->arp_sha, ebuf));
       printf("%s<S>0x%x(%s)\n",blankhead,ntohl(ipsrc.s_addr),
              inet_ntoa(ipsrc));
       printf("%s<D>%s\n",blankhead,
               ether_ntoa(arpPtr->arp_tha, ebuf));
       printf("%s<D>0x%x(%s)\n",blankhead,ntohl(ipdst.s_addr),
              inet_ntoa(ipdst));
       return;
       
    } else {
       printf("%sether_type = %u\n",blankhead,(hdrPtr->ether_type));
    }

    ipPtr = (struct ip *) ((char*)hdrPtr + sizeof(struct ether_header));

    bcopy((u_char*)& ipPtr->ip_src, (u_char*)& ipsrc.s_addr ,sizeof(struct in_addr));
    bcopy((u_char*)& ipPtr->ip_dst, (u_char*)& ipdst.s_addr ,sizeof(struct in_addr));

    printf("%s<S>0x%x(%s)\n",blankhead,ntohl(ipsrc.s_addr),
            inet_ntoa(ipsrc));
    printf("%s<D>0x%x(%s)\n",blankhead,ntohl(ipdst.s_addr),
            inet_ntoa(ipdst));
    printf("%sip_len:%d\n",blankhead,ntohs(ipPtr->ip_len));

    if ( ipPtr->ip_p == IPPROTO_TCP) {
         printf("%sTCP:\n",blankhead);
         tcpPtr = (struct tcphdr *) ((char*) ipPtr + sizeof(struct ip));
         dataPtr = (char*) ((char*)tcpPtr + sizeof(struct tcphdr));
    } else if (ipPtr->ip_p == IPPROTO_UDP) {
         printf("%sUDP:\n",blankhead);
         udpPtr = (struct udphdr*) ((char*) ipPtr + sizeof(struct ip));
         dataPtr = (char*) ((char*)udpPtr + sizeof(struct udphdr));
    } else  if (ipPtr->ip_p == IPPROTO_ICMP){
         printf("%sICMP: ",blankhead);
         icmpPtr = (struct icmp*) ((char*) ipPtr + sizeof(struct ip));
         if (icmpPtr->icmp_type == ICMP_ECHO) {
           printf("ECHO\n");
         } else if (icmpPtr->icmp_type == ICMP_ECHOREPLY) {
           printf("ECHOREPLY\n");
         } else {
           printf("type: %d  code: %d\n",
                  icmpPtr->icmp_type, icmpPtr->icmp_code);
         }
    } else  if (ipPtr->ip_p == IPPROTO_IGMP){
         printf("%sIGMP:\n",blankhead);
    } else  if (ipPtr->ip_p == IPPROTO_GGP){
         printf("%sGGP:\n",blankhead);
    } else  if (ipPtr->ip_p == IPPROTO_RAW){
         printf("%sRAW:\n",blankhead);
    } else {
         printf("%sprotocol:%d\n",blankhead,ipPtr->ip_p);
    }
}


/*
 *----------------------------------------------------------------------
 *
 * serviceSimPort -- 
 *
 *       Service a request that arrives at the simulated ethernet port.
 *
 * Results:
 *      -1 if problem occur with
 *
 * Side effects:
 *
 *----------------------------------------------------------------------
 */
int
serviceSimPort()
{
    struct  ether_header  *hdrPtr;
    struct sockaddr from;
    int	    fromlen, n;
    int	    etherAddrKey[2];
    int xfersize;
    Hash_Entry *entryPtr;
    HostInfoRecord *srcHostInfo;
    HostInfoRecord *dstHostInfo;

    if (verbose) {
      fprintf(stderr, "serviceSimPort()\n");
    }

    fromlen = sizeof (from);
    n = recvfrom(simfd, curPacket, sizeof(curPacket), 0, &from, &fromlen);
    if (n < 0) {
	perror("recvfrom");
	return -1;
    }
    if (!validPacket(curPacket, n, &from, fromlen)) {
	return 0;
    }
    if ((n < ETHERMIN + sizeof(struct ether_header)) || 
        (n > ETHERMTU + sizeof(struct ether_header))) {
	fprintf(stderr, "Bad packet size %d\n", n);
	return 0;
    }

    hdrPtr = (struct ether_header *) curPacket;
    if (verbose) {
        printf("Packet received from UDP host 0x%x (%s)\n",
               ntohl(((struct sockaddr_in *)&from)->sin_addr.s_addr),
               inet_ntoa(((struct sockaddr_in *)&from)->sin_addr));
        dispEtherPacket(n, hdrPtr, 0);
    }


    /*
     * Ensure that the mapping exists for this ethernet address.
     */
    etherAddrKey[1] = 0;
    bcopy(hdrPtr->ether_shost, (char *) etherAddrKey, 6);
    entryPtr = Hash_CreateEntry(&etheraddrTable, (Address) etherAddrKey, NULL);
    srcHostInfo = (HostInfoRecord *) Hash_GetValue(entryPtr);
    if (srcHostInfo == NULL) {
      srcHostInfo = (HostInfoRecord *) calloc(1, sizeof(HostInfoRecord));
      if (verbose) {
	fprintf(stderr, "setting srcHostInfo %lx\n", (long) srcHostInfo);
      }
      bcopy(hdrPtr->ether_shost, 
	    (caddr_t)&srcHostInfo->etheraddr, 6);
      Hash_SetValue(entryPtr, srcHostInfo);
    }
    bcopy((caddr_t)&from, (caddr_t)&srcHostInfo->fromaddr, fromlen);
    srcHostInfo->fromlen = fromlen;
    if (ntohs(hdrPtr->ether_type) == ETHERTYPE_IP) {
      in_addr_t	inetAddr;

      if (verbose) {
      struct in_addr	sAddr, dAddr;
	fprintf(stderr, "ETHERTYPE_IP packet %#x\n", ntohs(hdrPtr->ether_type));
        bcopy(curPacket + sizeof(struct ether_header) + offsetof(struct ip, ip_src),
              (char *) &sAddr, sizeof(sAddr));
        bcopy(curPacket + sizeof(struct ether_header) + offsetof(struct ip, ip_dst),
              (char *) &dAddr, sizeof(dAddr));
	fprintf(stderr, "IP: Src %s ", inet_ntoa(sAddr));
 	fprintf(stderr, "Dst %s\n",  inet_ntoa(dAddr));
      }

      bcopy(curPacket + sizeof(struct ether_header) + offsetof(struct ip, ip_src),
	    (char *) &inetAddr, sizeof(inetAddr));
      if (srcHostInfo->inetAddr != inetAddr) {
	if (verbose) {
	  fprintf(stderr, "srcHostInfo->inetAddr != inetAddr\n");
	}
	srcHostInfo->inetAddr = inetAddr;
	entryPtr = Hash_CreateEntry(&ipaddrTable, (Address)(u_long)inetAddr,NULL);
	Hash_SetValue(entryPtr, srcHostInfo);
      }
     }


    if ((D(0) == 0xff) && (D(1) == 0xff) && (D(2) == 0xff) &&
	(D(3) == 0xff) && (D(4) == 0xff) && (D(5) == 0xff)) {
	Hash_Search search;
	/*
	 * Broadcast. 
	 */
	if (verbose) {
	  fprintf(stderr, "broadcast packet\n");
	}

        if (ArpRequest(curPacket, n, curPacket, &n)) {
    		if (n < ETHERMIN + sizeof(struct ether_header)) {
			n =  ETHERMIN + sizeof(struct ether_header);
		}
                if (verbose) {
                   fprintf(stderr, "Sending ARP reply packet size %d\n", n);
                }
		goto reply;
        }

	for (entryPtr = Hash_EnumFirst(&etheraddrTable, &search); 
	     entryPtr != NULL; entryPtr = Hash_EnumNext(&search)) {
	    dstHostInfo = (HostInfoRecord *) Hash_GetValue(entryPtr);
	    if (dstHostInfo != srcHostInfo) { 
		xfersize = sendto(simfd, curPacket, n, 0x0, 
				    (struct sockaddr *) &dstHostInfo->fromaddr,
				    dstHostInfo->fromlen);
		if (xfersize < 0) {
		    perror("sendto");
		}
	    }
	}
	return 0;
    } 
  reply:
    etherAddrKey[1] = 0;
    bcopy(hdrPtr->ether_dhost, (char *) etherAddrKey, 6);

    entryPtr = Hash_FindEntry(&etheraddrTable, (Address)etherAddrKey);
    if (entryPtr != NULL) {
	dstHostInfo = (HostInfoRecord *) Hash_GetValue(entryPtr);

	if (verbose) {
	  fprintf(stderr, "sending packet\n");
	}

	xfersize = sendto(simfd, curPacket, n, 0x0, 
				(struct sockaddr *) &dstHostInfo->fromaddr,
				dstHostInfo->fromlen);
	if (xfersize < 0) {
	    perror("sendto");
	}
        return 0;

    } 

    if ((ntohs(hdrPtr->ether_type) == ETHERTYPE_IP) && (netfd > 0)) {
	in_addr_t	destAddr;

	if (verbose) {
	  fprintf(stderr, "outputPacket\n");
	}

	bcopy(curPacket + sizeof(struct ether_header) + offsetof(struct ip, ip_dst),
		(char *) &destAddr, sizeof(destAddr));
	destAddr = ntohl(destAddr);
	if ((destAddr & mynetmask) != subnetaddr) {
	    outputPacket(curPacket, n);
	}
        return 0;
    }


    fprintf(stderr, "Unknown dest address %x:%x:%x:%x:%x:%x from %x:%x:%x:%x:%x:%x\n", D(0), D(1), D(2), D(3), D(4), D(5), 
	    S(0), S(1), S(2), S(3), S(4), S(5));

     return 0;

}

/*
 *----------------------------------------------------------------------
 *
 * serviceNetPort -- 
 *
 *       Service a request that arrives at the network interface tap port.
 *
 * Results:
 *      -1 if problem occur with
 *
 * Side effects:
 *
 *----------------------------------------------------------------------
 */
int
serviceNetPort()
{
    struct  ether_header  *hdrPtr;
    int	     n;
    in_addr_t	inetAddr;
    Hash_Entry  *entryPtr;
    HostInfoRecord *hostInfoRecord;
    int xfersize;

    if (verbose) {
      fprintf(stderr, "serviceNetPort()\n");
    }

    n = read(netfd, curPacket, sizeof(curPacket));
    if (n < 0) {
	perror("read of serviceNetPort");
	return -1;
    }
    if ((n < ETHERMIN + sizeof(struct ether_header)) || 
	(n > ETHERMTU + sizeof(struct ether_header))) {
	fprintf(stderr, "Bad packet size %d\n", n);
	return -1;
    }

    hdrPtr = (struct ether_header *) curPacket;
    if ( verbose ) {
      struct in_addr ipadr;
      bcopy (curPacket + sizeof(struct ether_header) + 
             offsetof(struct ip, ip_src), (u_char*)&ipadr, 4);
      printf("Packet received from ETHERNET host 0x%x (%s)\n",
             ntohl(ipadr.s_addr), inet_ntoa(ipadr));
      bcopy (curPacket + sizeof(struct ether_header) + 
             offsetof(struct ip, ip_dst), (u_char*)&ipadr, 4);
      printf("  to SIM host 0x%x (%s)\n",
             ntohl(ipadr.s_addr), inet_ntoa(ipadr));
      dispEtherPacket (n, hdrPtr, 0);
    }

    if (ntohs(hdrPtr->ether_type) != ETHERTYPE_IP) {
      if (verbose) {
	fprintf(stderr, "ntohs(hdrPtr->ether_type) != ETHERTYPE_IP\n");
      }
      return 0;
    }

    bcopy(curPacket + sizeof(struct ether_header) + offsetof(struct ip, ip_dst),
		(char *) &inetAddr, sizeof(inetAddr));
    entryPtr = Hash_FindEntry(&ipaddrTable, (Address)(u_long)inetAddr);
    if (entryPtr == NULL) {
      if (verbose) {
	fprintf(stderr, "entryPtr == NULL\n");
      }
      return 0;
    }
    hostInfoRecord = (HostInfoRecord *) Hash_GetValue(entryPtr);
    if (verbose) {
       struct in_addr	sAddr, dAddr;
       bcopy(curPacket + sizeof(struct ether_header) + offsetof(struct ip, ip_src),
             (char *) &sAddr, sizeof(sAddr));
       bcopy(curPacket + sizeof(struct ether_header) + offsetof(struct ip, ip_dst),
             (char *) &dAddr, sizeof(dAddr));
       fprintf(stderr, "IP: Src %s ", inet_ntoa(sAddr));
       fprintf(stderr, "Dst %s\n",  inet_ntoa(dAddr));
       fprintf(stderr, "sending...\n");
    }
    xfersize = sendto(simfd, curPacket, n, 0x0, 
				(struct sockaddr *) &hostInfoRecord->fromaddr,
				hostInfoRecord->fromlen);

    if (xfersize < 0) {
	perror("sendto");
    }
    return 0;
}

/*
 *----------------------------------------------------------------------
 *
 * validPacket -- 
 *
 *       Check to see if we should accept this packet.
 *
 * Results:
 *      1 if packet ok.
 *      0 if we should request it.
 *
 * Side effects:
 *
 *----------------------------------------------------------------------
 */

int
validPacket(packet, size, fromAddr, fromlen)
    char *packet;	/* The contents of the packet. */
    int	 size;		/* The size of the packet. */
    struct sockaddr *fromAddr; /* Where the packet was from. */
    int	  fromlen;	       /* The length of fromAddr. */
{
    struct  ether_header  *hdrPtr;
    struct  sockaddr_in   *inaddrPtr;
    in_addr_t  srcAddr;

    hdrPtr = (struct ether_header *) packet;
    if ((size < ETHERMIN + sizeof(struct ether_header)) || 
        (size > ETHERMTU + sizeof(struct ether_header))) {
	fprintf(stderr, "Bad packet size %d\n", size);
	return 0;
    }
    /* 
     * XXX - Do security checking here.  Currently we reject all packets 
     * from hosts outside our network.  We only if we are forwarding
     * packets.
     */
    if (!noipforwarding) { 
	if ((fromAddr->sa_family != AF_INET) || 
	    (fromlen < offsetof(struct sockaddr_in, sin_zero))) {
	    fprintf(stderr, "Packet from bad fromaddr\n");
	    return 0;
	}
	inaddrPtr = (struct sockaddr_in *) fromAddr;
	srcAddr = inet_netof(inaddrPtr->sin_addr);
	if ((srcAddr != mynetaddr) && (srcAddr != IN_LOOPBACKNET)) {
	    fprintf(stderr, "Security violation: Packet from host %s\n",
			    inet_ntoa(inaddrPtr->sin_addr));
	    return 0;
	}
    }
    return 1;
}



/*
 *----------------------------------------------------------------------
 *
 * initNITdevice -- 
 *
 *       Initialize the network interface tap (NIT) file descriptor so
 *	 we can catch IP packets for this subnet.
 *
 * Results:
 *      None.
 *
 * Side effects:
 *
 *----------------------------------------------------------------------
 */

int 
initNITdevice(device, subNetAddr, netmask)
    char	    *device;    /* Network interface name. */
    in_addr_t 	   subNetAddr;  /* IP address of the subnet. */
    in_addr_t	   netmask;	/* Network address mask of the interface. */
{
      struct ifreq ifr;
      struct enfilter pf;
      int   err, cmd;
      int fd;
      int allow_copyall;
      u_short bits;
      struct endevp param;

      fd = pfopen(device, O_RDWR);
      if (fd < 0) {
	  perror("pfopen");
	  return -1;
      }

      /* make sure we can see packets to/from ethersim host too */
      allow_copyall = 1;
      if (ioctl(fd, EIOCALLOWCOPYALL, &allow_copyall)) {
        perror("ioctl: EIOCALLOWCOPYALL");
      }
        
      bits = (ENTSTAMP | ENBATCH | ENHOLDSIG | ENNONEXCL);
      if (ioctl(fd, EIOCMBIC, &bits)) {
          perror("ioctl: EIOCMBIC");
          return -1;
      }

      bits = (ENPROMISC | ENCOPYALL);
      if ( ioctl(fd, EIOCMBIS, &bits)) {
          perror("ioctl: EIOCMBIS");
          return -1;
      }

      /*
       * check if we got the right interface
       */
      if (ioctl(fd, EIOCIFNAME, &ifr)) {
          perror("ioctl: EIOCIFNAME");
          return -1;
      }
      if (strcmp (device, ifr.ifr_name) != 0) {
        fprintf (stderr, "wrong network interface %s (requested %s)\n",
          ifr.ifr_name, device);
      }
      if (verbose) {
        fprintf (stderr, "connected to network interface %s\n",
          ifr.ifr_name);
      }

      /*
       * Read out the ethernet address so we can fill them into packets
       * being sent.
       */
      if (ioctl(fd, EIOCDEVP, &param)) {
          perror("ioctl: EIOCDEVP");
          return -1;
      }
      if (verbose) {
        if (param.end_dev_type == ENDT_10MB) {
          printf ("device type: %s\n", "10MB ethernet");
        } else {
          printf ("device type: %d\n", param.end_dev_type);
        }
        printf ("address len: %d\n", param.end_addr_len);
        printf ("header len:  %d\n", param.end_hdr_len);
        printf ("MTU:         %d\n", param.end_MTU);
        printf ("hardware address:  %x:%x:%x:%x:%x:%x\n",
          param.end_addr[0], param.end_addr[1], param.end_addr[2],
          param.end_addr[3], param.end_addr[4], param.end_addr[5]);
        printf ("broadcast address: %x:%x:%x:%x:%x:%x\n",
          param.end_broadaddr[0], param.end_broadaddr[1], 
          param.end_broadaddr[2], param.end_broadaddr[3], 
          param.end_broadaddr[4], param.end_broadaddr[5]);
      }

      bcopy(param.end_addr, (char *) &myetheraddr, sizeof(myetheraddr));

      /*
       * Add the packet filter to only return IP packets to the specified
       * network. 
       */

      pf.enf_Priority = 37;       /* anything > 2 should work */

      cmd = 0;
      pf.enf_Filter[cmd++] = ENF_PUSHWORD + (u_short)
		 offsetof(struct ether_header, ether_type)/sizeof(u_short);
      pf.enf_Filter[cmd++] = ENF_PUSHLIT | ENF_CAND;
      pf.enf_Filter[cmd++] = htons(ETHERTYPE_IP);

      /* handles all propper sub-netmasks */
      /* compare low-order address bits first */
      /* since they are more likely to disagree */
      if ((netmask & 0xffff) != 0x0000) {
        u_short lowval, highval;
        lowval = (((in_addr_t)subNetAddr) & netmask) & 0xffff;
        highval = lowval | (~(netmask & 0xffff));
        lowval = htons (lowval);
        highval = htons (highval);
        pf.enf_Filter[cmd++] = ENF_PUSHWORD + (u_short) 
                    sizeof(struct ether_header)/sizeof(u_short) +
                     offsetof(struct ip, ip_dst)/sizeof(u_short) + 1;
        pf.enf_Filter[cmd++] = ENF_PUSHLIT | ENF_GE;
        pf.enf_Filter[cmd++] = lowval;
        pf.enf_Filter[cmd++] = ENF_PUSHWORD + (u_short) 
                    sizeof(struct ether_header)/sizeof(u_short) +
                     offsetof(struct ip, ip_dst)/sizeof(u_short) + 1;
        pf.enf_Filter[cmd++] = ENF_PUSHLIT | ENF_LE;
        pf.enf_Filter[cmd++] = highval;
        pf.enf_Filter[cmd++] = ENF_CAND;
      }
      if ((netmask >> 16) == 0xffff) {
        /* compare high-order address bits now */
        pf.enf_Filter[cmd++] = ENF_PUSHWORD + (u_short) 
                    sizeof(struct ether_header)/sizeof(u_short) +
                     offsetof(struct ip, ip_dst)/sizeof(u_short);
        pf.enf_Filter[cmd++] = ENF_PUSHLIT | ENF_EQ;
        pf.enf_Filter[cmd++] = htons((((in_addr_t)subNetAddr)&netmask)>>16);
      } else {
        u_short lowval, highval;
        lowval = (((in_addr_t)subNetAddr) & netmask) >> 16;
        highval = lowval | (~(netmask >> 16));
        lowval = htons (lowval);
        highval = htons (highval);
        pf.enf_Filter[cmd++] = ENF_PUSHWORD + (u_short) 
                    sizeof(struct ether_header)/sizeof(u_short) +
                     offsetof(struct ip, ip_dst)/sizeof(u_short);
        pf.enf_Filter[cmd++] = ENF_PUSHLIT | ENF_GE;
        pf.enf_Filter[cmd++] = lowval;
        pf.enf_Filter[cmd++] = ENF_PUSHWORD + (u_short) 
                    sizeof(struct ether_header)/sizeof(u_short) +
                     offsetof(struct ip, ip_dst)/sizeof(u_short);
        pf.enf_Filter[cmd++] = ENF_PUSHLIT | ENF_LE;
        pf.enf_Filter[cmd++] = highval;
        pf.enf_Filter[cmd++] = ENF_CAND;
      }
      pf.enf_FilterLen = cmd;

      if (ioctl(fd, EIOCSETF, &pf)) {
          perror("ioctl: EIOCSETF");
	  goto bad;
      }

      /* Flush the read queue, to get rid of anything that accumulated
	 before the device reached its final configuration. */
      err = ioctl(fd, EIOCFLUSH, 0);
      if (err < 0) {
	  perror("EIOCFLUSH");
	  goto bad;
      }
      return fd;
  bad:
      (void) close(fd);
      return -1;
}




/*
 *----------------------------------------------------------------------
 *
 * initNITdevice -- 
 *
 *       Initialize the network interface tap (NIT) file descriptor so
 *	 we can catch IP packets for this subnet. (proxy version)
 *
 * Results:
 *      None.
 *
 * Side effects:
 *
 *----------------------------------------------------------------------
 */

int 
initNITdeviceFP(device, proxyaddr, nproxys, netmask)
    char	   *device;    /* Network interface name. */
    in_addr_t      *proxyaddr;  /* list of proxy addresses */
    u_long         nproxys;     /* number of proxy addresses */
    in_addr_t	   netmask;	/* Network address mask of the interface. */
{
      struct ifreq ifr;
      struct enfilter pf;
      int   err, cmd, i;
      int fd;
      int allow_copyall;
      u_short bits;
      struct endevp param;

      fd = pfopen(device, O_RDWR);
      if (fd < 0) {
	  perror("pfopen");
	  return -1;
      }

      /* make sure we can see packets to/from ethersim host too */
      allow_copyall = 1;
      if (ioctl(fd, EIOCALLOWCOPYALL, &allow_copyall)) {
        perror("ioctl: EIOCALLOWCOPYALL");
      }
        
      bits = (ENTSTAMP | ENBATCH | ENHOLDSIG | ENNONEXCL);
      if (ioctl(fd, EIOCMBIC, &bits)) {
          perror("ioctl: EIOCMBIC");
          return -1;
      }

      bits = (ENPROMISC | ENCOPYALL);
      if ( ioctl(fd, EIOCMBIS, &bits)) {
          perror("ioctl: EIOCMBIS");
          return -1;
      }

      /*
       * check if we got the right interface
       */
      if (ioctl(fd, EIOCIFNAME, &ifr)) {
          perror("ioctl: EIOCIFNAME");
          return -1;
      }
      if (strcmp (device, ifr.ifr_name) != 0) {
        fprintf (stderr, "wrong network interface %s (requested %s)\n",
          ifr.ifr_name, device);
      }
      if (verbose) {
        fprintf (stderr, "connected to network interface %s\n",
          ifr.ifr_name);
      }

      /*
       * Read out the ethernet address so we can fill them into packets
       * being sent.
       */
      if (ioctl(fd, EIOCDEVP, &param)) {
          perror("ioctl: EIOCDEVP");
          return -1;
      }
      if (verbose) {
        if (param.end_dev_type == ENDT_10MB) {
          printf ("device type: %s\n", "10MB ethernet");
        } else {
          printf ("device type: %d\n", param.end_dev_type);
        }
        printf ("address len: %d\n", param.end_addr_len);
        printf ("header len:  %d\n", param.end_hdr_len);
        printf ("MTU:         %d\n", param.end_MTU);
        printf ("hardware address:  %x:%x:%x:%x:%x:%x\n",
          param.end_addr[0], param.end_addr[1], param.end_addr[2],
          param.end_addr[3], param.end_addr[4], param.end_addr[5]);
        printf ("broadcast address: %x:%x:%x:%x:%x:%x\n",
          param.end_broadaddr[0], param.end_broadaddr[1], 
          param.end_broadaddr[2], param.end_broadaddr[3], 
          param.end_broadaddr[4], param.end_broadaddr[5]);
      }

      bcopy(param.end_addr, (char *) &myetheraddr, sizeof(myetheraddr));

      /*
       * Add the packet filter to only return IP packets to the specified
       * network. 
       */

      pf.enf_Priority = 37;       /* anything > 2 should work */

      cmd = 0;
      pf.enf_Filter[cmd++] = ENF_PUSHWORD + (u_short)
		 offsetof(struct ether_header, ether_type)/sizeof(u_short);
      pf.enf_Filter[cmd++] = ENF_PUSHLIT | ENF_CAND;
      pf.enf_Filter[cmd++] = htons(ETHERTYPE_IP);

      for(i=0; i<nproxys; i++) {
         u_short lowval, highval;
         lowval = (proxyaddr[i] & 0xffff);
         highval = ((proxyaddr[i]>>16) & 0xffff);
         lowval = htons (lowval);
         highval = htons (highval);
         
         pf.enf_Filter[cmd++] = ENF_PUSHWORD + (u_short) 
            sizeof(struct ether_header)/sizeof(u_short) +
            offsetof(struct ip, ip_dst)/sizeof(u_short) + 1;
         pf.enf_Filter[cmd++] = ENF_PUSHLIT | ENF_EQ;
         pf.enf_Filter[cmd++] = lowval;
         
         pf.enf_Filter[cmd++] = ENF_PUSHWORD + (u_short) 
            sizeof(struct ether_header)/sizeof(u_short) +
            offsetof(struct ip, ip_dst)/sizeof(u_short);
         pf.enf_Filter[cmd++] = ENF_PUSHLIT | ENF_EQ;
         pf.enf_Filter[cmd++] = highval;

         pf.enf_Filter[cmd++] = ENF_AND;
/*
         pf.enf_Filter[cmd++] = ENF_PUSHZERO;
         pf.enf_Filter[cmd++] = ENF_CNAND;
         */
         printf("lowval %x highval %x\n", lowval, highval);
      }
      pf.enf_FilterLen = cmd;

      if (ioctl(fd, EIOCSETF, &pf)) {
          perror("ioctl: EIOCSETF");
	  goto bad;
      }

      /* Flush the read queue, to get rid of anything that accumulated
	 before the device reached its final configuration. */
      err = ioctl(fd, EIOCFLUSH, 0);
      if (err < 0) {
	  perror("EIOCFLUSH");
	  goto bad;
      }
      return fd;
  bad:
      (void) close(fd);
      return -1;
}

/*
 *----------------------------------------------------------------------
 *
 * outputPacket -- 
 *
 *
 * Results:
 *      None.
 *
 * Side effects:
 *
 *----------------------------------------------------------------------
 */


int
outputPacket(packet, size)
    char *packet;  /* The contents of the packet. */
    int	 size;     /* The size of the packet. */
{
    struct ether_header	*hdrPtr;
    in_addr_t	srcaddr;

    /*
     * Ethernet source address is set by kernel
     */
    hdrPtr = (struct ether_header *) packet;
    bzero(hdrPtr->ether_shost, 6);

    bcopy(packet + sizeof(struct ether_header) + offsetof(struct ip, ip_src),
	  (char *) &srcaddr, sizeof(srcaddr));
    srcaddr = ntohl(srcaddr);

    /*
     * Security: 
     *	1) Don't broadcast or send to all zeros.
     *  2) Must be of type IP.
     *  3) Must have a ip_src address on the simulated subnet.
     */

    if (((D(0) == 0xff) && (D(1) == 0xff) && (D(2) == 0xff) &&
	(D(3) == 0xff) && (D(4) == 0xff) && (D(5) == 0xff)) ||
        ((D(0) == 0) && (D(1) == 0) && (D(2) == 0) &&
	(D(3) == 0) && (D(4) == 0) && (D(5) == 0)) ||
	(ntohs(hdrPtr->ether_type) != ETHERTYPE_IP) ||
	(((srcaddr & mynetmask) != subnetaddr) && (nproxys == 0))) {

	fprintf(stderr, 
	    "outputPacket: Tossing packet sent to %x:%x:%x:%x:%x:%x\n",
		D(0), D(1), D(2), D(3), D(4), D(5));
	return 0;
    }

    /*
     * send the data
     */
    if (write(netfd, packet, size) != size) {
	    perror("outputPacket: write");
	    return (-1);
    }
#ifdef notdef
    fprintf(stderr, "Not sending packet to %x:%x:%x:%x:%x:%x from %x:%x:%x:%x:%x:%x\n", D(0), D(1), D(2), D(3), D(4), D(5), 
	    S(0), S(1), S(2), S(3), S(4), S(5));
#endif	

    return 0;
}
/*
 *----------------------------------------------------------------------
 *
 * lookupInterface -- 
 *
 *       Find the specified interface and read its address and netmask.
 *	
 *
 * Results:
 *     0 if we found things, -1 otherwise.
 *
 * Side effects:
 *
 *----------------------------------------------------------------------
 */

int
lookupInterface(interfacename, inetAddrPtr, netmaskPtr)
    char *interfacename;
    in_addr_t *inetAddrPtr;
    in_addr_t *netmaskPtr;
{
    char cmdBuf[256];
    char inetString[32], netmaskString[32];
    int ch;

    FILE *pfile;

    /* 
     * We cheat and use ifconfig to lookup the info. This should probably
     * be changed to do all kinds of nasty I/O controls to lookup the info.
     */
    sprintf(cmdBuf, "/usr/sbin/ifconfig %s", interfacename);
    pfile = popen(cmdBuf, "r");
    if (pfile == (FILE *) NULL) {
	fprintf(stderr, "Can't execute %s\n", cmdBuf);
	return -1;
    }
    /*
     * Skip the first line which should look something like:
     * le0: flags=63<UP,BROADCAST,NOTRAILERS,RUNNING>
     */
    for (ch = fgetc(pfile); (ch != '\n') && (ch != EOF); ch = fgetc(pfile)) 
	;
    if (fscanf(pfile, "     inet %s netmask %s ", inetString, netmaskString) != 2){
	(void) pclose(pfile);
	return -1;
    }
    *inetAddrPtr = inet_addr(inetString);
    *netmaskPtr = strtol(netmaskString, (char **)NULL, 16);
    if (routeraddr == (in_addr_t) -1) {
	routeraddr = ((*inetAddrPtr) & (*netmaskPtr)) | 1; 
		/* Default to host 1 on our network */
    }
    (void) pclose(pfile);
    return 0;

}

/*
 * FindGateway is used to resolve ARP requests
 * myinetaddr is returned in network form
 * mynetmask is host form (UGH!!!)
 */
in_addr_t
FindGateway(dstaddr)
   in_addr_t dstaddr;
{
   in_addr_t raddr = htonl(routeraddr);
   in_addr_t nmask = htonl(mynetmask);

   if (verbose) {
      fprintf (stderr, "find gateway dst(%8x) me(%8x) subnet(%8x) router
(%8x) mask (%8x)\n",
               dstaddr, myinetAddr, subnetaddr, raddr, nmask);
   }

   if ((dstaddr & nmask) == (myinetAddr & nmask)) 
      return dstaddr;
   if ((dstaddr & nmask) == htonl(subnetaddr)) 
      return -1;
   return (raddr);
}
	



