/*
 * Copyright (C) 1996-1998 by the Board of Trustees
 *    of Leland Stanford Junior University.
 * 
 * This file is part of the SimOS distribution. 
 * See LICENSE file for terms of the license. 
 *
 */

/*****************************************************************
 * scache.c
 * Secondary cache model
 * 
 * $Author: stets $
 * $Date: 2001/07/31 23:16:08 $
 *****************************************************************/
#include <assert.h>
#include <malloc.h>
#include "syslimits.h"
#include "simutil.h"
#include "sim_error.h"
#include "scache.h"
#include "memsys.h"
#include "cpu_interface.h"
#include "hw_events.h"
#include "cpu_stats.h"

#ifdef HWBCOPY
#  include "hw_bcopy.h"
#endif
#include "arch_specifics.h"

/* #define CC_CHECKER */
#ifdef CC_CHECKER
static int  CacheChecker(PA paddr, int cpuNum, int isExcl);
#endif

#if defined(SIM_MIPS32) || defined(SIM_MIPS64)
extern int sim_magic_OSPC_access(int cpuNum, uint VA, byte *data);
extern void sim_magic_OSPC_stalled(int cpuNum, uint VA, int stalled);
/*#define DEBUG_OSPC_STALL*/
#endif

/*
 * Possible return status from the miss handling table.
 * SMHTSUCCESS - Entry allocated everything looks good.
 * SMHTMERGE   - This request merged with another. Returned other.
 * SMHTFULL    - Entry not allocated because table was full.
 * SMHTCONFLICT - Entry not allocated because it conflicted with one already
 *               allocated.
 */
typedef enum { SMHTSUCCESS = 0, SMHTFULL, SMHTMERGE, SMHTCONFLICT} SMHTStatus;

/* The second level caches for all of the processors */
SCache *SCACHE;

/* Local Functions */
static void SCacheHitEvent(int cpuNum,EventCallbackHdr *event, void *arg);
static SMHTStatus AllocSMHT(int cpuNum, MCMD mcmd, PA pAddr, int lru, int *smhtind);
static void FreeSMHT(int cpuNum, int entryNum);

static void SCacheUpdate(int cpuNum, SMHT *smht, int mode, int status,
			 int result, byte *data);
static FlushStatus SCacheFlush(int cpuNum, int writeback, int retain,
			       PA pAddr, int size, byte *data);

static SCResult SCacheUpgradeMiss(int cpuNum, VA vAddr, PA pAddr, SCacheCmd cmd, 
				  int mhtind, int way, struct SCacheSet *set,
				  SimCounter *missCntPtr);
static SCResult SCacheMiss(int cpuNum, VA vAddr, PA pAddr, SCacheCmd cmd, int mhtind,
			   struct SCacheSet* set, SimCounter *missCntPtr);

static void SCacheLRUInit(uint *lruword);
static void SCacheLRUMake(uint *lruword, int set);
static int  SCacheLRU(uint *lruword);
static void SCacheLRUTouch(uint *lruword,int set);
#ifdef DATA_HANDLING
static void DumpCacheLine(byte *data, int length);
#endif

/* Some shared functions for cache manipulation */
int SameCacheLine(PA addr1, PA addr2, unsigned cacheLineSize);
PA QuadWordAlign(PA offset);

#define CURRENT_PC(_cpu) (CPUVec.CurrentPC(_cpu))

/*Hack for supporting non-power-of-two cache sizes */
int scacheTagShift;

/*****************************************************************
 * InitSCaches
 *****************************************************************/
void
InitSCaches(void)
{
   int i,j,k;
#ifdef DATA_HANDLING
   char *pool;
#endif

   scacheTagShift = GetLog2(SCACHE_SIZE/SCACHE_ASSOC);
   if (scacheTagShift == -1) {
      CPUError("Warning: TagShift is not power of 2, %d/%d\n", SCACHE_SIZE, SCACHE_ASSOC);
   }

   SCACHE = (SCache *)calloc(TOTAL_CPUS,sizeof(SCache)); 

   CPUPrint("SCACHE: Assoc = %d Size = %#x Line = %d HitTime %d\n", 
            SCACHE_ASSOC, SCACHE_SIZE, SCACHE_LINE_SIZE, SCACHE_HIT_TIME);
   
   for (j=0; j < TOTAL_CPUS; j++) {
      SCACHE[j].set = (struct SCacheSet *)
	 calloc(SCACHE_INDEX, sizeof(struct SCacheSet));
      if (SCACHE[j].set == NULL) {
	 CPUError("calloc failed on scache allocation (%d bytes)\n",
                  SCACHE_INDEX*sizeof(struct SCacheSet));
      }
#ifdef DATA_HANDLING
      pool = (char *) calloc(SCACHE_LINE_SIZE * SCACHE_INDEX 
                             * SCACHE_ASSOC, sizeof(char));
#endif 
      for (i=0; i < SCACHE_INDEX; i++) {
	 for (k=0; k < SCACHE_ASSOC; k++) {
	    SCACHE[j].set[i].tags[k] = INVALID_TAG;
#ifdef DATA_HANDLING 
            {
               if (pool == NULL) {
                  CPUError("calloc failed on scache pool allocation\n");
               }
               /* Allocate the actual memory for the cache line's data */
               SCACHE[j].set[i].data[k] = pool;
               pool += SCACHE_LINE_SIZE;
            }
#endif
         }
	 SCacheLRUInit(&SCACHE[j].set[i].LRU[0]);
      }
      for (i=0; i < SMHT_SIZE; i++) {
	 SCACHE[j].SMHT[i].inuse = FALSE;
      }
      SCACHE[j].SMHTnumInuse = 0;
   }

}

/*****************************************************************
 * SCacheFetch
 *
 * Process a first level cache miss.
 *****************************************************************/
SCResult
SCacheFetch(int cpuNum, VA vAddr, PA pAddr, SCacheCmd cmd, int mhtind)
{
   struct SCacheSet*  set;
   SimCounter *missCntPtr;
   int way;
   int cacheNum  = GET_CACHE_NUM(cpuNum);
   int scacheNum = GET_SCACHE_NUM(cpuNum);
   uint ind      = SCACHE_INDEXOF(pAddr); 
   PA tag        = SCACHE_TAG(pAddr);

   if (interest(pAddr)) {
      LogEntry("fetch",cpuNum,"pA=0%08x state=%d tag=0x%x foo=0x%x\n",pAddr, 
               MSCacheState(cpuNum, pAddr),tag,SCACHE[scacheNum].set[ind].tags[0]);
   }
   switch (cmd) {
   case SC_IGET:
      SCACHE[scacheNum].stats.Igets++;
      missCntPtr = &SCACHE[scacheNum].stats.IgetMisses;
      break;
   case SC_DGET:
   case SC_DLLGET:
      SCACHE[scacheNum].stats.Dgets++;
      missCntPtr = &SCACHE[scacheNum].stats.DgetMisses;
#ifdef HWBCOPY
      if ((SimConfigGetBool("Hwbcopy.Streaming")) &&
          (SCACHE[scacheNum].get.last_index + 1) == ind) {
        int run = ++SCACHE[scacheNum].get.consecutive;
        int range = SimConfigGetInt("Hwbcopy.StreamRange");
        int depth = SimConfigGetInt("Hwbcopy.StreamDepth");
        int i;
        for (i = 1; i < depth; i++) {
           if ((run >= range*i) && (run <= range*(i+1))) {
              PA pfAddr = pAddr + (i * SCACHE_LINE_SIZE);
              VA vfAddr = vAddr + (i * SCACHE_LINE_SIZE);
              if (IS_VALID_PA(M_FROM_CPU(cpuNum), pfAddr)) {
                 SCACHE[scacheNum].stats.StreamGets[i]++;
                 SCachePrefetch(cpuNum, vfAddr, pfAddr, MEMSYS_GET);
              }
           }
        }
        if (run >= range*depth) {
           PA pfAddr = pAddr + (depth * SCACHE_LINE_SIZE);
           VA vfAddr = vAddr + (depth * SCACHE_LINE_SIZE);
           if (IS_VALID_PA(M_FROM_CPU(cpuNum), pfAddr)) {
              SCACHE[scacheNum].stats.StreamGets[depth]++;
              SCachePrefetch(cpuNum, vfAddr, pfAddr, MEMSYS_GET);
           }
        }
      }
      else if ((SCACHE[scacheNum].get.last_index) != ind)
        SCACHE[scacheNum].get.consecutive = 0;

      SCACHE[scacheNum].get.last_index = ind;
#endif
      break;
   case SC_DGETX:
   case SC_DLLGETX:
      SCACHE[scacheNum].stats.DgetXs++;
      missCntPtr = &SCACHE[scacheNum].stats.DgetXMisses;
#ifdef HWBCOPY
      if ((SimConfigGetBool("Hwbcopy.Streaming")) &&
          (SCACHE[scacheNum].getx.last_index + 1) == ind) {
        int run = ++SCACHE[scacheNum].getx.consecutive;
        int range = SimConfigGetInt("Hwbcopy.StreamRange");
        int depth = SimConfigGetInt("Hwbcopy.StreamDepth");
        int i;
        for (i = 1; i < depth; i++) {
           if ((run >= range*i) && (run <= range*(i+1))) {
              PA pfAddr = pAddr + (i * SCACHE_LINE_SIZE);
              VA vfAddr = vAddr + (i * SCACHE_LINE_SIZE);
              if (IS_VALID_PA(M_FROM_CPU(cpuNum), pfAddr)) {
                 SCACHE[scacheNum].stats.StreamGetXs[i]++;
                 SCachePrefetch(cpuNum, vfAddr, pfAddr, MEMSYS_GETX);
              }
           }
        }
        if (run >= range*depth) {
           PA pfAddr = pAddr + (depth * SCACHE_LINE_SIZE);
           VA vfAddr = vAddr + (depth * SCACHE_LINE_SIZE);
           if (IS_VALID_PA(M_FROM_CPU(cpuNum), pfAddr)) {
              SCACHE[scacheNum].stats.StreamGetXs[depth]++;
              SCachePrefetch(cpuNum, vfAddr, pfAddr, MEMSYS_GETX);
           }
        }
      }
      else if ((SCACHE[scacheNum].getx.last_index) != ind)
        SCACHE[scacheNum].getx.consecutive = 0;

      SCACHE[scacheNum].getx.last_index = ind;
#endif
      break;
   case SC_DUGETX:
   case SC_DSCUGETX:
      SCACHE[scacheNum].stats.Dupgrades++;
      missCntPtr = &SCACHE[scacheNum].stats.DupgradeMisses;
      break;
   default:
      CPUError("Unknown SC command 0x%x\n", cmd);
      missCntPtr = 0;
      break;
   }

/*   pAddr = pAddr & ~(PA)(SCACHE_LINE_SIZE-1);*/

  set = & SCACHE[scacheNum].set[ind];
  for( way=0;way<SCACHE_ASSOC;++way ) {
     if((set->tags[way] & ~EXCLUSIVE_TAG) == tag) {
        break;
     }
  }
  if (way==SCACHE_ASSOC) {
     /* A miss */
     return SCacheMiss(cpuNum, vAddr, pAddr, cmd, mhtind, set, missCntPtr);
  }

  if ((cmd & EXCLUSIVE_CMD) &&
      !(set->tags[way] & EXCLUSIVE_TAG)) {
     if ((cmd == SC_DGETX) || (cmd == SC_DLLGETX)) {
        /* A first level getx which requires an upgrade in the scache --  *
         * change the statistics to record this as an upgrade, not a getx */
        SCACHE[scacheNum].stats.DgetXs--;
        SCACHE[scacheNum].stats.Dupgrades++;
        missCntPtr = &SCACHE[scacheNum].stats.DupgradeMisses;
     }
     return SCacheUpgradeMiss(cpuNum, vAddr, pAddr, cmd, mhtind, way,
			      set, missCntPtr);
  }
 
   if (VERBOSE_DEBUG) {
      CPUPrint("%d: SCacheFetch: hit for %8.8lx\n",cpuNum,pAddr);
   }

   /* A hit in the second level, update LRU and PUT the line in the first
    * level cache.
    */ 
   SCacheLRUTouch(&SCACHE[scacheNum].set[ind].LRU[0], way);
   
   /*
    * Return the line to the first level after the specified delay.
    */
   CACHE[cacheNum].MHT[mhtind].smhtEntry = SECOND_LEVEL_HIT_ENTRY;
   if (set->tags[way] & EXCLUSIVE_TAG) { 
      CACHE[cacheNum].MHT[mhtind].mode = MEMSYS_EXCLUSIVE;
    } else {
       CACHE[cacheNum].MHT[mhtind].mode = MEMSYS_SHARED;
    }

   if (memsysVec.NoMemoryDelay || SCACHE_HIT_TIME == 0 || IN_MAGIC_SECTION(cpuNum)) {
      MHTReqDone(cpuNum, CACHE[cacheNum].MHT + mhtind, 
                 CACHE[cacheNum].MHT[mhtind].mode,
		 MEMSYS_STATUS_SUCCESS);
      return SCSUCCESS;
   } else { 
      EventDoCallback(cpuNum, SCacheHitEvent,
		      (EventCallbackHdr *)(CACHE[cacheNum].MHT + mhtind),
		      (void *)NULL, SCACHE_HIT_TIME);
      return SCSTALL;
   }
}

	 
/*****************************************************************
 * SCacheMiss
 * 2nd level miss, forward to memory system. Start by allocating
 * space in the SMHT. 
 *****************************************************************/

SCResult
SCacheMiss(int cpuNum, VA vAddr, PA pAddr, SCacheCmd cmd, int mhtind, 
           struct SCacheSet* set, SimCounter *missCntPtr)
{
   int cacheNum = GET_CACHE_NUM(cpuNum);
   int scacheNum = GET_SCACHE_NUM(cpuNum);
   PA replacePaddr;
   int writeback;
   MCMD mcmd;
   int way;
   int replace;
   int waiter;
   SMHTStatus ret;
   Result mret;
   int smhtind;
   /* Quad-word align the miss */
   pAddr = QuadWordAlign(pAddr);
#ifndef SOLO
   if (!IS_VALID_PA(M_FROM_CPU(cpuNum), pAddr)) {
      return SCBUSERROR;
   }
#endif

   /* Check if there are any lines marked INVALID and replace them
    * rather than the LRU line 
    */ 
   for (replace=0; replace<SCACHE_ASSOC; ++replace) {
      if (set->tags[replace] & INVALID_TAG ) {
         break;
      }
   }
   if (replace == SCACHE_ASSOC) {
      way = SCacheLRU(&set->LRU[0]);
   } else {
      way = replace;
   }

   mcmd = (cmd & EXCLUSIVE_CMD) ? MEMSYS_GETX : MEMSYS_GET;

   if ((cmd == SC_DLLGET) || (cmd == SC_DLLGETX)) {
      mcmd |= MEMSYS_LLFLAVOR;
   }

   if (cmd == SC_IGET) {
      mcmd |= MEMSYS_IFFLAVOR;
   }

   ret = AllocSMHT(cpuNum, mcmd, pAddr, way, &smhtind);
   if ((ret == SMHTFULL) || (ret == SMHTCONFLICT)) {
      /*We've overrun or conflicted in the SMHT. Return 
       * a retry so we will try again. 
       */
      if (VERBOSE_DEBUG) {
         CPUPrint("%d: SCacheMiss:  VA: %8.8lx  PA: %8.8lx RETRY\n",
                  cpuNum, vAddr, pAddr);
      }
      /* NOTE: I've already counted this reference as a L2 Ref. To make
         the stats work, I need a count of these gets that have to
         retry since the retry doesn't show up as an L1 Miss. */

      switch (cmd) {
      case SC_IGET:
        SCACHE[scacheNum].stats.IGetMHTRetries++;
        break;
      case SC_DGET:
      case SC_DLLGET:
        SCACHE[scacheNum].stats.DGetMHTRetries++;
        break;
      case SC_DGETX:
      case SC_DLLGETX:
        SCACHE[scacheNum].stats.DGetXMHTRetries++;
        break;
      case SC_DUGETX:
      case SC_DSCUGETX:
        SCACHE[scacheNum].stats.DUGetXMHTRetries++;
        break;
      default:
        CPUWarning("Bad cmd found at SMHT conflict\n");
        ASSERT(0);
      }
      return SCRETRY;
   }
   ASSERT((ret == SMHTSUCCESS) || (ret == SMHTMERGE));
	 
   /*
    * Add the MHT entry to the lists of entries waiting
    * for the line to be filled or upgraded.
    */
   ASSERT(SCACHE[scacheNum].SMHT[smhtind].numMHTwait < MHT_SIZE);
   waiter = SCACHE[scacheNum].SMHT[smhtind].numMHTwait++;
   SCACHE[scacheNum].SMHT[smhtind].mhtInd[waiter] = mhtind;
   SCACHE[scacheNum].SMHT[smhtind].vAddr = vAddr;
   CACHE[cacheNum].MHT[mhtind].smhtEntry = smhtind;
   if (ret == SMHTMERGE) {
      /*
       * This line already has an outstanding miss of the right
       * type.  Merge into it.
       */
      if (VERBOSE_DEBUG) {
         CPUPrint("%d: SCacheMiss:  VA: %8.8lx  PA: %8.8lx MERGE\n",
                  cpuNum, vAddr, pAddr);
      }
      SCACHE[scacheNum].stats.MergeMisses++;
      return SCSTALL;
   }
   /* Now we need to allocate clear out space in the cache
    * for it. If the entry we choose is valid we flush
    * it from the primary cache(s). We pass the replaced
    * address to the memory system so it can simulated writebacks
    * and replacement hints, etc. 
    */
 
   if (VERBOSE_DEBUG) {
      CPUPrint("%d: SCacheMiss:  VA: %8.8lx  PA: %8.8lx\n",cpuNum, vAddr, pAddr);
   }

   if (!(set->tags[way] & INVALID_TAG)) {
      replacePaddr = STAG_TO_PA(set->tags[way],set-SCACHE[scacheNum].set);
      writeback = ((set->tags[way] & EXCLUSIVE_TAG) != 0);
#ifndef SOLO
      ASSERT (IS_VALID_PA(M_FROM_CPU(cpuNum), replacePaddr));
#endif
      
      if (VERBOSE_DEBUG) {
	 CPUPrint("%d: SCacheMiss:  replacing PA: %8.8lx EX:%d\n",cpuNum,
		    replacePaddr, writeback);
      }

      /* Writeback if dirty... retain == 0 =>remove the line */
      PrimaryFlush(cpuNum, writeback, 0, replacePaddr, SCACHE_LINE_SIZE);

      { 
         uint type;
         if (writeback) {
            type = E_L2 | E_WRITEBACK;
         } else {
            type = E_L2 | E_FLUSH_CLEAN;
         }
         L2_LINE_TRANS_EVENT(cpuNum, replacePaddr, type, vAddr, way, 
                             IS_KUSEG(CURRENT_PC(cpuNum)));
      }
  		      
      set->tags[way] = INVALID_TAG;
      if (writeback) {
	 SCACHE[scacheNum].stats.Writebacks++;
      } else {
	 SCACHE[scacheNum].stats.Replacements++;
      }
   } else {
      replacePaddr = MEMSYS_NOADDR;
      writeback = 0;
   }
	
/* if not a NAK'ed reference */
   (*missCntPtr)++;

#ifdef DATA_HANDLING
   /* Store the cache line pointer in the SMHT */
   /* for use later when the reply arrives */
   SCACHE[scacheNum].SMHT[smhtind].lineData = set->data[way];
   mret = memsysVec.MemsysCmd(cpuNum, SCACHE[scacheNum].SMHT[smhtind].memsyscmd, 
		    pAddr, smhtind, replacePaddr, writeback, set->data[way]);
#else
   mret = memsysVec.MemsysCmd(cpuNum, SCACHE[scacheNum].SMHT[smhtind].memsyscmd, 
		    pAddr, smhtind, replacePaddr, writeback, NULL);
#endif

   if (mret == STALL) {
      return SCSTALL;
   }
   
   /* magic memory system */
   ASSERT(mret == SUCCESS);
#ifndef BUFFER_WRITES
   ASSERT(!SCACHE[scacheNum].SMHT[smhtind].inuse);
   SCacheLRUTouch(&set->LRU[0], way);
   return SCSUCCESS;
#endif
}


/*****************************************************************
 *
 *****************************************************************/
PFResult
SCachePrefetch(int cpuNum, VA vAddr, PA pAddr, MCMD mcmd)
{
   int scacheNum = GET_SCACHE_NUM(cpuNum);
   PA replacePaddr;
   int writeback;
   int way;
   int replace;
   SMHTStatus ret;
   Result mret;
   int smhtind;
   struct SCacheSet *set = & SCACHE[scacheNum].set[SCACHE_INDEXOF(pAddr)];
	 
   if (interest(pAddr)) {
      LogEntry("l2-pf-issue",cpuNum,"pAddr=0x%x vAddr=0x%x foo=0x%x\n",
               pAddr,vAddr,set->tags[0]);
   }
   /* Quad-word align the address */
   pAddr = QuadWordAlign(pAddr);
#ifndef SOLO
   if (!IS_VALID_PA(M_FROM_CPU(cpuNum), pAddr)) {
      CPUWarning("%d: SCachePrefetch:  VA: %8.8llx  PA: %8.8lx FAIL out of range!\n",
                  cpuNum, (Reg64)vAddr, pAddr);

      return PF_SUCCESS; /* Makes everyone above happy */
   }
#endif   

   /* Check to see if line is already resident */
   way = SCACHE_ASSOC;
   for(replace=0; replace<SCACHE_ASSOC; ++replace) {
      if ((set->tags[replace] & ~EXCLUSIVE_TAG) == SCACHE_TAG(pAddr)) {
         /* Line is resident in cache -- see if we need an upgrade */
         if (((mcmd == MEMSYS_GETX) && (set->tags[replace] & EXCLUSIVE_TAG)) ||
             (mcmd == MEMSYS_GET)) {
            return PF_RESIDENT;
         } else {
            mcmd = MEMSYS_UPGRADE; /* Need to upgrade here, a shared
                                      copy is present */
            way = replace;
            break;
         }
      } else if (set->tags[replace] & INVALID_TAG) {
         way = replace;
      }
   }
   
   mcmd |= MEMSYS_PREFETCH;
            
   /* Allocate space in the SMHT for the prefetch */
   if (way == SCACHE_ASSOC) {
      way = SCacheLRU(&set->LRU[0]);
   }

   ret = AllocSMHT(cpuNum, mcmd, pAddr, way, &smhtind);
   /* If SMHT is full, punt the prefetch */
   if ((ret == SMHTCONFLICT) || (ret == SMHTFULL)) {
      if (VERBOSE_DEBUG) {
         CPUPrint("%d: SCachePrefetch:  VA: %8.8lx  PA: %8.8lx FAIL\n",
                  cpuNum, vAddr, pAddr);
      }
      return PF_FAILURE;
   }

   if (ret == SMHTMERGE) {
     /* This line already has an outstanding miss of the right type -- punt */
      if (VERBOSE_DEBUG) {
         CPUPrint("%d: SCachePrefetch:  VA: %8.8lx  PA: %8.8lx MERGE\n",
                  cpuNum, vAddr, pAddr);
      }
     return PF_MERGE;
   }

   ASSERT(ret == SMHTSUCCESS);

   SCACHE[scacheNum].SMHT[smhtind].pfvAddr = vAddr;

   /* Now we need to allocate clear out space in the cache
    * for it. If the entry we choose is valid we flush
    * it from the primary cache(s). We pass the replaced
    * address to the memory system so it can simulated writebacks
    * and replacement hints, etc. 
    */

   if (VERBOSE_DEBUG) {
      CPUPrint("%d: SCachePrefetch:  VA: %8.8lx  PA: %8.8lx\n",
               cpuNum, vAddr, pAddr);
   }

   if (!((mcmd & MEMSYS_CMDMASK) == MEMSYS_UPGRADE) && 
       !(set->tags[way] & INVALID_TAG)) {
      replacePaddr = STAG_TO_PA(set->tags[way],set-SCACHE[scacheNum].set);
      writeback = (set->tags[way] & EXCLUSIVE_TAG) != 0;
      
#ifndef SOLO
      ASSERT (IS_VALID_PA(M_FROM_CPU(cpuNum), replacePaddr));
#endif
      if (VERBOSE_DEBUG) {
	 CPUPrint("%d: SCachePrefetch:  replacing PA: %8.8lx EX:%d\n",cpuNum,
		    replacePaddr, writeback);
      }

      PrimaryFlush(cpuNum, writeback, 0, replacePaddr, SCACHE_LINE_SIZE);

      { 
         uint type;
         if (writeback) {
            type = E_L2 | E_WRITEBACK;
         } else {
            type = E_L2 | E_FLUSH_CLEAN;
         }
         L2_LINE_TRANS_EVENT(cpuNum, replacePaddr, type, vAddr, 
                             way, IS_KUSEG(CURRENT_PC(cpuNum)));
      }

      set->tags[way] = INVALID_TAG;
      if (writeback) {
	 SCACHE[scacheNum].stats.Writebacks++;
      } else {
	 SCACHE[scacheNum].stats.Replacements++;
      }
   } else {
      replacePaddr = MEMSYS_NOADDR;
      writeback = 0;
   }

   SCACHE[scacheNum].stats.Prefetches++;

   PREFETCH_EVENT(CPUVec.CycleCount(cpuNum),cpuNum,CURRENT_PC(cpuNum),vAddr);

#ifdef DATA_HANDLING
				/* Store the cache line pointer in the SMHT */
				/* for use later when the reply arrives */
   SCACHE[scacheNum].SMHT[smhtind].lineData = set->data[way];
   mret = memsysVec.MemsysCmd(cpuNum, SCACHE[scacheNum].SMHT[smhtind].memsyscmd, 
		    pAddr, smhtind, replacePaddr, writeback, set->data[way]);
#else
   mret = memsysVec.MemsysCmd(cpuNum, SCACHE[scacheNum].SMHT[smhtind].memsyscmd, 
		    pAddr, smhtind, replacePaddr, writeback, NULL);
#endif

   if (mret == STALL) return (((mcmd & MEMSYS_CMDMASK) == MEMSYS_UPGRADE) ? 
                              PF_UPGRADE : PF_STALL);

   ASSERT(mret == SUCCESS);
   ASSERT(!SCACHE[scacheNum].SMHT[smhtind].inuse);
   SCacheLRUTouch(&set->LRU[0], way);
   return PF_SUCCESS;
}


SCResult
SCacheUpgradeMiss(int cpuNum, VA vAddr, PA pAddr, SCacheCmd cmd, int mhtind,
		  int way, struct SCacheSet *set, SimCounter *missCntPtr)
{
   int cacheNum = GET_CACHE_NUM(cpuNum);
   int scacheNum = GET_SCACHE_NUM(cpuNum);
   MCMD mcmd = MEMSYS_UPGRADE;
   SMHTStatus ret;
   Result mret;
   int smhtind;

   if (cmd == SC_DSCUGETX) {
      mcmd |= MEMSYS_SCFLAVOR;
   }

   /* Quad-word align the miss */
   pAddr = QuadWordAlign(pAddr);
#ifndef SOLO   
   if (!IS_VALID_PA(M_FROM_CPU(cpuNum), pAddr)) {
      return SCBUSERROR;
   }
#endif
   ret = AllocSMHT(cpuNum, mcmd, pAddr, way, &smhtind);
   if ((ret == SMHTFULL) || (ret == SMHTCONFLICT)) {
      /* We've overrun or conflicted in the SMHT. Return 
       * a retry so we will try again. 
       */
      SCACHE[scacheNum].stats.DUGetXMHTRetries++;
      if (VERBOSE_DEBUG) {
         CPUPrint("%d: SCacheUpgradeMiss:  VA: %8.8lx  PA: %8.8lx RETRY\n",cpuNum, vAddr, pAddr);
      }
      return SCRETRY;
   }

   ASSERT((ret == SMHTSUCCESS) || (ret == SMHTMERGE));
   ASSERT(SCACHE[scacheNum].SMHT[smhtind].numMHTwait < MHT_SIZE);
   SCACHE[scacheNum].SMHT[smhtind].vAddr = vAddr;
   SCACHE[scacheNum].SMHT[smhtind].mhtInd[SCACHE[scacheNum].SMHT[smhtind].numMHTwait++] = mhtind;
   CACHE[cacheNum].MHT[mhtind].smhtEntry = smhtind;
   if (ret == SMHTMERGE) {
      /* This line already has an outstanding miss of the right
       * type.  Merge into it.
       */
      if (VERBOSE_DEBUG) {
         CPUPrint("%d: SCacheUpgradeMiss:  VA: %8.8lx  PA: %8.8lx MERGE\n",cpuNum, vAddr, pAddr);
      }
      return SCSTALL;
   }

   (*missCntPtr)++;

   if (VERBOSE_DEBUG) {
      CPUPrint("%d: SCacheUpgradeMiss:  VA: %8.8lx  PA: %8.8lx\n",cpuNum, vAddr, pAddr);
   }
   
#ifdef DATA_HANDLING
				/* Store the cache line pointer in the SMHT */
				/* for use later IN CASE DATA ARRIVES WITH */
				/* the message */
   SCACHE[scacheNum].SMHT[smhtind].lineData = set->data[way];
#endif

   mret = memsysVec.MemsysCmd(cpuNum, SCACHE[scacheNum].SMHT[smhtind].memsyscmd, pAddr, 
                           smhtind, MEMSYS_NOADDR, 0, (byte *) 0);

   if (mret == SUCCESS) {
      /* magic memory system */
#ifndef BUFFER_WRITES
/* WARNING: --- this might be wrong: --- */
   ASSERT(!SCACHE[scacheNum].SMHT[smhtind].inuse);
#endif   
      return SCSUCCESS;
   } else {
      return SCSTALL;
   }
}

#ifdef DATA_HANDLING
/*****************************************************************
 * DumpCacheLine
 *****************************************************************/
void
DumpCacheLine(byte *data,int length)
{
   unsigned long *ld = (unsigned long *)data;
   int i;

   for (i=0; i< (length/sizeof(long)); i+=4) {
      CPUPrint("%8.8lx %8.8lx %8.8lx %8.8lx ",ld[i],ld[i+1], ld[i+2],ld[i+3]);
      if ((i+4)%8 == 0) {
	 CPUPrint("\n");
      }
   }
}
#endif

/***************************************************************
 * IsInSCache
 *
 * The data parameter is used to return a POINTER to the caller
 * indicating the line in the scache which is relevant.
 *
 * If the line is not in the cache, 0 is returned and the
 * data parameter is unchanged.
 *
 *
 ****************************************************************/
int
IsInSCache(int cpuNum, PA pAddr, int mode, char **data, int *way)
{
   int scacheNum = GET_SCACHE_NUM(cpuNum);
   struct SCacheSet* set;
   PA tag;
   bool foundit = FALSE;
   uint w;

   if (skipCaches) {
      return foundit;
   }

   set = & SCACHE[scacheNum].set[SCACHE_INDEXOF(pAddr)];
   if (mode == MEMSYS_EXCLUSIVE) { 
      tag = SCACHE_TAG_EX(pAddr);
      for(w=0;w<SCACHE_ASSOC; ++w) {
	 if (set->tags[w] == tag) {
	    foundit=1;
#ifdef DATA_HANDLING
	    *data = set->data[w];
				/* Return a pointer to the scache line */
#endif
	    break;		/* if we found it, why not bail out!!!! */
	 }
      }
   } else {
       tag = SCACHE_TAG(pAddr);
       for(w=0;w<SCACHE_ASSOC;++w) {
	  if ((set->tags[w] & ~EXCLUSIVE_TAG) == tag) {
	     foundit=1;
#ifdef DATA_HANDLING
	     *data = set->data[w];
#endif
				/* Return a pointer to the scache line */
	     break;		/* if we found it, why not bail out!!!! */
	  }
       }
   }
   *way = w;
   
   return foundit;
}

/*****************************************************************
 * SCacheHitEvent
 *
 * Event callback for a first level miss that hits in the second level.  
 * Basically, stuff the line into the cache if it is still in the second 
 * level cache.
 *****************************************************************/
static void
SCacheHitEvent(int cpuNum,EventCallbackHdr *event, void *arg)
{
   MHT *mht = (MHT *) event;
   int result;
   int status;
   char *bogus;
   int way=0;
   uint type;

   result = IsInSCache(cpuNum, mht->pAddr, mht->mode, &bogus, &way);
   if (result) {
      status = MEMSYS_STATUS_SUCCESS;
   } else { 
      status = MEMSYS_STATUS_NAK;
   }
   /*
    * Modify MHT and unstall processor with first data word as soon
    * as the bus is free (if bus utilization is on).
    */
   
   if (mht->cmd & SC_IGET) {
      /* Sending to ICache */
      if (status != MEMSYS_STATUS_NAK) {
         mht->status = status;
      }
      mht->stall += 
         CPUVec.FillCacheLine(cpuNum, mht->PC, mht->vAddr, mht->pAddr, E_I);
      type = MHTReqDone(cpuNum, mht, mht->mode, status);
   } else {
      /* Sending to DCache */
      if ((status != MEMSYS_STATUS_NAK) && (mht->cmd != SC_DSCUGETX) &&
          (mht->cmd != SC_DUGETX)) {
         mht->status = status;
      }
      mht->stall += 
         CPUVec.FillCacheLine(cpuNum, mht->PC, mht->vAddr, mht->pAddr, E_D);
      type = MHTReqDone(cpuNum, mht, mht->mode, status);
   }
}

/*****************************************************************
 * SCachePUT
 *****************************************************************/
void
SCachePUT(int cpuNum, SMHT *smht, int mode, byte *fillData)
{
   int scacheNum = GET_SCACHE_NUM(cpuNum);
   PA pAddr = smht->pAddr;
   uint ind = (pAddr/SCACHE_LINE_SIZE) % SCACHE_INDEX;
   PA tag   = SCACHE_TAG(pAddr);
   int way  = smht->scacheLRU;
   struct SCacheSet *set =  &SCACHE[scacheNum].set[ind];
	
   if (interest(pAddr)) {
      LogEntry("l2-PUT",cpuNum,"pAddr=0x%x tag=0x%x foo=0x%x\n",
               pAddr,tag,set->tags[0]);
   }

   QUICK_ICACHE_CLEAR(cpuNum);
#ifndef SOLO
   ASSERT (IS_VALID_PA(M_FROM_CPU(cpuNum), pAddr));
/*   ASSERT( tag < SCACHE_TAG( MEM_SIZE(M_FROM_CPU(cpuNum))));*/
#endif

#ifdef DATA_HANDLING
   ASSERT( SCACHE[scacheNum].set[ind].tags[way] == tag ||
	   fillData);		/* We either have the data already, and */
				/* this is an upgrade, or fillData is set */
				/* so we can get the correct data*/
#endif

   if (mode & MEMSYS_EXCLUSIVE) {
      tag |= EXCLUSIVE_TAG;
   }

   SCACHE[scacheNum].set[ind].tags[way] = tag;
   SCacheLRUTouch(&SCACHE[scacheNum].set[ind].LRU[0], way);
   
   /* WARNING: I think this was too restrictive:  ASSERT(fillData != 0); */
   /* This parameter should never be zero when I get here */

#ifdef DATA_HANDLING
   if (fillData != 0) {
      ASSERT(smht->lineData != (char *) 0xffffffff);
      bcopy(fillData, smht->lineData, SCACHE_LINE_SIZE);

      if (VERBOSE_DEBUG) {
	 CPUPrint("%d: SCacheMiss: filling %8.8lx (from flashlite) ",cpuNum,
		    smht->pAddr);
	 CPUPrint("%d: bcopy %x -> %x (%d)  Data: ",cpuNum,
		    fillData, smht->lineData, SCACHE_LINE_SIZE);      
	 DumpCacheLine(smht->lineData, SCACHE_LINE_SIZE);		 
      }
      smht->lineData = (char *) 0xffffffff;
      /* Set lineData to be invalid to prevent erroneous reuse of this */
      /* pointer, such as occured in our heinous bug of 3/28/95 */
   }
#endif
   
#ifdef CC_CHECKER
   CacheChecker(pAddr, cpuNum, (tag & EXCLUSIVE_TAG)!=0); 
#endif
}




/****************************************************************
 * SCacheFlush
 *
 * See DCacheFlush, the semantics is identical
 ****************************************************************/
static FlushStatus
SCacheFlush(int cpuNum, int writeback, int retain, PA paddr, int size,
	    byte *data)
{
   int scacheNum = GET_SCACHE_NUM(cpuNum);
   PA a;  
   uint ind;
   struct SCacheSet* set;
   int way;
   FlushStatus s = FLUSH_NOTFOUND;

   if (interest(paddr)) {
      LogEntry("flush",cpuNum,"%x state=%d\n",paddr, MSCacheState(cpuNum, paddr));
   }
   ASSERT(writeback || !retain);  /* One of these two should be set */
   
   if (writeback && retain) {
      SCACHE[scacheNum].stats.Downgrade++;
   }

   for (a = (paddr & ~(PA)(SCACHE_LINE_SIZE-1)); 
	a < (paddr & ~(PA)(SCACHE_LINE_SIZE-1)) + size; 
	a += SCACHE_LINE_SIZE) {

      ind = SCACHE_INDEXOF(a);
      set = & SCACHE[scacheNum].set[ind];

      /*
       * Direct-mapped caches. We do not rely on presence in the
       * cache to determine the success of the SC, but rather
       * on the LLbit.  
       */

      if (SCACHE_ASSOC == 1  && CPUVec.GetLockAddr(cpuNum) == a) { 
         if (!retain) { 
            /* invalidated from the cache */
	   CPUVec.ClearLockFlag(cpuNum);
         } 
      }
  
      for (way=0; way<SCACHE_ASSOC; ++way) {
	 if (set->tags[way] == SCACHE_TAG_EX(a)) {
	    goto ScFlush_Exclusive;
	 }
	 if ((set->tags[way] & ~EXCLUSIVE_TAG) == SCACHE_TAG(a))  {
	    goto ScFlush_Shared;
	 }
      }

      if (way==SCACHE_ASSOC) {
         uint type;
	 /* not found */
         if (writeback & retain) { 
            type = E_L2 | E_WRITEBACK | E_FAKE_EXTERNAL;
         } else {
            type = E_L2 | E_FLUSH_CLEAN | E_FAKE_EXTERNAL;
         }
         L2_LINE_TRANS_EVENT(cpuNum, a, type, 0, 0,IS_KUSEG(CURRENT_PC(cpuNum)));
         continue;
      }
      
      ASSERT(0);		/* Control should never reach this point */
      return s;

ScFlush_Exclusive:		/* Found an exclusive line in the flush */
				/* region...handle it based on the flags*/
      s = FLUSH_EXCLUSIVE;

      PrimaryFlush(cpuNum, writeback, retain, a, SCACHE_LINE_SIZE);
         
      SCACHE[scacheNum].stats.LinesWriteback++;

      /* WARNING: I don't require the user to writeback the line if it's */
      /* dirty since I want to support invalidations of dirty lines for */
      /* memory copy, where I squash a line regardless of its state */
      
      if (retain) {
         /* Remove exclusive ownership */
	 set->tags[way] &= SCACHE_TAG(a);
#ifndef SOLO
         ASSERT(IS_VALID_PA(M_FROM_CPU(cpuNum), STAG_TO_PA(set->tags[way],0)));
/*         ASSERT( (set->tags[way]&~EXCLUSIVE_TAG) < 
                 SCACHE_TAG(MEM_SIZE(M_FROM_CPU(cpuNum))));*/
#endif
         L2_LINE_TRANS_EVENT(cpuNum, a, E_L2 | E_EXTERNAL | E_DOWNGRADE, 0, way,
                             CURRENT_PC(cpuNum));
      } else {
	 set->tags[way] = INVALID_TAG;
	 L2_LINE_TRANS_EVENT(cpuNum, a, E_L2 | E_EXTERNAL | E_WRITEBACK, 0, way,
                             IS_KUSEG(CURRENT_PC(cpuNum)));
      }


#ifdef DATA_HANDLING
      if (writeback) {
	 ASSERT(data != NULL);
	 if (data != NULL) {	/* If I want a data reply */
            bcopy(set->data[way], data, SCACHE_LINE_SIZE);
	 }
         /* SAH: Why isn't there a writeback here? */
	 if (VERBOSE_DEBUG) {
	    CPUPrint("%d: bcopy %x -> %x (%d) (flashlite) Data:\n", cpuNum,
		       set->data[way],(char *)data,SCACHE_LINE_SIZE);
	    DumpCacheLine(set->data[way],SCACHE_LINE_SIZE);
	 }
      }
#endif /* DATA_HANDLING */
      continue;

ScFlush_Shared:
      if (!retain) {
	 if (VERBOSE_DEBUG) {
	    CPUPrint("%d: SCacheMiss:  spilling (shared) PA: %8.8lx \n",cpuNum,
		       paddr);
	 }
	 PrimaryFlush(cpuNum, writeback, retain, a, SCACHE_LINE_SIZE);
	 SCACHE[scacheNum].stats.LinesInval++;

	 s = FLUSH_SHARED;
	 set->tags[way] = INVALID_TAG;

	 SCacheLRUMake(&set->LRU[0], way);

         L2_LINE_TRANS_EVENT(cpuNum, a, (E_L2 | E_EXTERNAL | E_FLUSH_CLEAN), 
                             0, way,IS_KUSEG(CURRENT_PC(cpuNum)));
      }
      continue;
   }

   return s;
}


/****************************************************************
 * SCacheFlushIndex
 *
 * Similar to SCacheFlush, but flushes one of the cached lines at
 * an index matching the one of the given paddr. Used to implement
 * the CACH_SD option to the CACHE instruction.
 ****************************************************************/
static FlushStatus
SCacheFlushIndex(int cpuNum, int writeback, int retain,
		 PA paddr, PA *real_paddr, int size, byte *data)
{
   int scacheNum = GET_SCACHE_NUM(cpuNum);
   PA a, aa;  
   uint ind;
   struct SCacheSet* set;
   int way;
   FlushStatus s = FLUSH_NOTFOUND;

   if( interest(paddr))
      LogEntry("flush index",cpuNum,"%x \n",paddr);

   ASSERT(writeback || !retain);  /* One of these two should be set */
   
   if (writeback && retain)
      SCACHE[scacheNum].stats.Downgrade++;

   for (a = (paddr & ~(PA)(SCACHE_LINE_SIZE-1)); 
	a < (paddr & ~(PA)(SCACHE_LINE_SIZE-1)) + size; 
	a += SCACHE_LINE_SIZE) {

#ifndef SOLO
      ASSERT (IS_VALID_PA(M_FROM_CPU(cpuNum), paddr));
#endif
      ind = SCACHE_INDEXOF(a);
      set = & SCACHE[scacheNum].set[ind];

      /*
       * Direct-mapped caches. We do not rely on presence in the
       * cache to determine the success of the SC, but rather
       * on the LLbit.  
       */

      if (SCACHE_ASSOC == 1  && CPUVec.GetLockAddr(cpuNum) == a) { 
         if (!retain) { 
            /* invalidated from the cache */
	   CPUVec.ClearLockFlag(cpuNum);
         } 
      }
  
      for (way=0; way<SCACHE_ASSOC; ++way)
	if (set->tags[way] != INVALID_TAG) {
	  if ((set->tags[way] & EXCLUSIVE_TAG) != 0) {
	    s = FLUSH_EXCLUSIVE;
	    aa = STAG_TO_PA(set->tags[way], ind);
	    if (real_paddr) *real_paddr = aa;
	    
#ifndef SOLO
            ASSERT (IS_VALID_PA(M_FROM_CPU(cpuNum), aa));
#endif
	    PrimaryFlush(cpuNum, writeback, retain, aa, SCACHE_LINE_SIZE);
	    
	    SCACHE[scacheNum].stats.LinesWriteback++;
	    
	    /* WARNING: I don't require the user to writeback the line if it's */
	    /* dirty since I want to support invalidations of dirty lines for */
	    /* memory copy, where I squash a line regardless of its state */
	    
	    if (retain) {
	      /* Remove exclusive ownership */
	      set->tags[way] &= ~EXCLUSIVE_TAG;
	      L2_LINE_TRANS_EVENT(cpuNum, aa, E_L2|E_EXTERNAL|E_DOWNGRADE, 0, way,
                                  IS_KUSEG(CURRENT_PC(cpuNum)));
	    } else {
	      set->tags[way] = INVALID_TAG;
	      L2_LINE_TRANS_EVENT(cpuNum, aa, E_L2|E_EXTERNAL|E_WRITEBACK, 0, way,
                                  IS_KUSEG(CURRENT_PC(cpuNum)));
	    }

#ifdef DATA_HANDLING
	    if (writeback) {
	      ASSERT(data != NULL);
	      if (data != NULL) {	/* If I want a data reply */
		bcopy(set->data[way],data, SCACHE_LINE_SIZE);
	      }
	      if (VERBOSE_DEBUG) {
		CPUPrint("%d: bcopy %x -> %x (%d) (flashlite) Data:\n", cpuNum,
			   set->data[way],(char *)data,SCACHE_LINE_SIZE);
		DumpCacheLine(set->data[way],SCACHE_LINE_SIZE);
	      }
	    }
#endif /* DATA_HANDLING */
	    return s; /* invalidate only one line */
	  } else {
	    aa = STAG_TO_PA(set->tags[way], ind);
	    if (real_paddr) *real_paddr = aa;
	    if (!retain) {
	      if (VERBOSE_DEBUG) {
		CPUPrint("%d: SCacheMiss: spilling (shared) PA: %8.8lx \n",
                         cpuNum, paddr);
	      }
	      PrimaryFlush(cpuNum, writeback, retain, aa, SCACHE_LINE_SIZE);
	      SCACHE[scacheNum].stats.LinesInval++;
	      
	      s = FLUSH_SHARED;
	      set->tags[way] = INVALID_TAG;
	      
	      SCacheLRUMake(&set->LRU[0], way);
	      
	      L2_LINE_TRANS_EVENT(cpuNum, aa, (E_L2 | E_EXTERNAL | E_FLUSH_CLEAN), 
				  0, way,IS_KUSEG(CURRENT_PC(cpuNum)));
	    }
	    return s; /* invalidate only one line */
	  }
	}
   }

   return s;
}


/*****************************************************************
 * CacheFakeInvalidate 
 * Called when a line has been invalidated by a cpu, and is known not
 * to be in the cache of cpuNum. This allows Memstat to maintain its
 * state machine. This is never called in perfectmem.
 *
 *****************************************************************/
void
CacheFakeInvalidate(int cpuNum,PA paddr, int size)
{
   int scacheNum = GET_SCACHE_NUM(cpuNum);
   PA a;
   uint ind;
   struct SCacheSet* set;
   int way;

   for (a = paddr; a < paddr + size; a += SCACHE_LINE_SIZE) {
      ind = SCACHE_INDEXOF(a);  /* (a/SCACHE_LINE_SIZE) % SCACHE_INDEX; */
      set = & SCACHE[scacheNum].set[ind];
      
      if( SCACHE_ASSOC == 1  && 
          CPUVec.GetLockAddr(cpuNum) == (a &  ~(SCACHE_LINE_SIZE-1))  ) { 
          /* Direct-mapped cache. Invalidate it */
	CPUVec.ClearLockFlag(cpuNum);
      }

      for(way=0;way<SCACHE_ASSOC;++way  ) {
         if ((set->tags[way] & ~EXCLUSIVE_TAG) == SCACHE_TAG(a)) 
            break;
      }
      if (way==SCACHE_ASSOC) {
         /* not in the cache */
	 L2_LINE_TRANS_EVENT(cpuNum, a, (E_L2 | E_FAKE_EXTERNAL), 0, 0,
                             IS_KUSEG(CURRENT_PC(cpuNum)));
         continue;
      }
      ASSERT( 0 );
   }
}

/***************************************************************
 * Cacheinvalidate
 *
 * This routine destructively invalidates a line from the cache.
 * In the case of shared data, the line is removed, and this
 * operation is equivalent to CacheExtract.  
 *
 * In the case of dirty data, this routine PERMANENTLY DISCARDS
 * the dirty copy and removes the line from the cache.  While
 * this violates most cache coherence protocols, for block
 * transfer this is a useful performance optimization.
 ***************************************************************/
int
CacheInvalidate(int cpuNum, PA paddr, int size, bool writeback, 
                int *wasDirty) 
{
   FlushStatus s;
   
   if (VERBOSE_DEBUG) {
      CPUPrint("%d: CacheInvalidate for paddr %8.8lx\n",
		 cpuNum, paddr);
   }

   if (writeback) {
      /* writeback == TRUE  retain  == FALSE */      
      s = SCacheFlush(cpuNum, TRUE, FALSE, paddr, size, NULL);
   } else {
      /* writeback == FALSE  retain  == FALSE */
      s = SCacheFlush(cpuNum, FALSE, FALSE, paddr, size, NULL);
   }

   *wasDirty = (s == FLUSH_EXCLUSIVE);
   if (s == FLUSH_NOTFOUND) {
      return 0;
   }
   return 1;
}

/***************************************************************
 * CacheinvalidateIndex
 *
 * Invoked through the "cache" op. Stalling the processor 
 * is a little messy here... I assume that if it was a writeback and
 * it was found, then I'll need to stall the processor..   
 ***************************************************************/
int
CacheInvalidateIndex(int cpuNum, PA paddr, int size, bool writeback, 
		     int *wasDirty) 
{
   FlushStatus s;
   
   if (VERBOSE_DEBUG) {
      CPUPrint("%d: CacheInvalidateIndex for paddr %8.8lx\n",
		 cpuNum, paddr);
   }

   if (writeback) {
      /* writeback == TRUE  retain  == FALSE */      
      s = SCacheFlushIndex(cpuNum, TRUE, FALSE, paddr, NULL, size, NULL);
   } else {
      /* writeback == FALSE  retain  == FALSE */
      s = SCacheFlushIndex(cpuNum, FALSE, FALSE, paddr, NULL, size, NULL);
   }

   *wasDirty = (s == FLUSH_EXCLUSIVE);
   if (s == FLUSH_NOTFOUND) {
      return 0;
   }
   return 1;
}

/****************************************************************
 * CacheWriteback
 *
 * This routine forces a line in the cache to be written back,
 * if it is dirty.  The line loses its ownership but remains in
 * the cache. (This is a GET operation)
 ****************************************************************/
int
CacheWriteback(int cpuNum, PA paddr, int size, byte *data)
{
   FlushStatus s;

   if (VERBOSE_DEBUG) {
      CPUPrint("%d: CacheWriteback for paddr %8.8lx\n",
		 cpuNum, paddr);
   }

   s = SCacheFlush(cpuNum, TRUE, TRUE, paddr, size, data);
				/* writeback == TRUE */
				/* retain  == TRUE */
   if (s == FLUSH_NOTFOUND) {
      return 0;
   }
   return 1;
} 

/****************************************************************
 * CacheFlush
 *
 * This routine forces a line in the cache to be written back,
 * if it is dirty.  The line is flushed out of the cache.
 ****************************************************************/
int
CacheFlush(int cpuNum, PA paddr, int size, byte *data)
{
   FlushStatus s;

   if (VERBOSE_DEBUG) {
      CPUPrint("%d: CacheFlush for paddr %8.8lx\n",
		 cpuNum, paddr);
   }

   s = SCacheFlush(cpuNum, TRUE, FALSE, paddr, size, data);
				/* writeback == TRUE */
				/* retain  == FALSE */
   if (s == FLUSH_NOTFOUND) {
      return 0;
   }
   return 1;
} 

/***************************************************************
 * CacheExtract
 *
 * This operation requests a line from the cache, and removes it
 * from the cache.  The line is written back, if dirty.
 * (This is a GETX operation)
 *
 * After this operation, the line is no longer present in the
 * cache.
 ***************************************************************/
int
CacheExtract(int cpuNum, PA paddr, int size, int *writeback, byte *data)
{
   FlushStatus s;


   if (VERBOSE_DEBUG) {
      CPUPrint("%d: CacheExtract for paddr %8.8lx\n",
		 cpuNum, paddr);
   }

   s = SCacheFlush(cpuNum, TRUE, FALSE, paddr, size, data);
				/* writeback == TRUE */
				/* retain  == FALSE */
   *writeback = (s == FLUSH_EXCLUSIVE);
   if (s == FLUSH_NOTFOUND) {
      return 0;
   }
   return 1;
} 


/***************************************************************
 * CacheExtractIndex
 *
 * This operation requests a line from the cache, and removes it
 * from the cache.  The line is written back, if dirty.
 * (This is a GETX operation)
 *
 * After this operation, the line is no longer present in the
 * cache.
 ***************************************************************/
int
CacheExtractIndex(int cpuNum, PA paddr, int size,
		  int *writeback, PA *real_paddr,
		  byte *data) 
{
   FlushStatus s;


   if (VERBOSE_DEBUG) {
      CPUPrint("%d: CacheExtract for paddr %8.8lx\n",
		 cpuNum, paddr);
   }

   s = SCacheFlushIndex(cpuNum, TRUE, FALSE, paddr, real_paddr, size, data);
				/* writeback == TRUE */
				/* retain  == FALSE */
   *writeback = (s == FLUSH_EXCLUSIVE);
   if (s == FLUSH_NOTFOUND) {
      return 0;
   }
   return 1;
} 


/*****************************************************************
 * CacheCmdDone
 *****************************************************************/

struct {

  int stalled;          /* stalled OSPC access on this CPU? */

  int transId;          /* data for completing stalled access */
  int mode;
  int status;
  int result;

  EventCallbackHdr hdr; /* callback header for t/o */

} ospcs[SIM_MAXCPUS];


/* called in the event of an OSPC timeout.
 * Must complete the access and conclude the transaction.
 */

#if defined(SIM_MIPS32) || defined(SIM_MIPS64)
static void
OSPCCallback(int cpuNum, EventCallbackHdr *hdr, void *arg)
{
  int          scacheNum = GET_SCACHE_NUM(cpuNum);
  SMHT*        smht      = SCACHE[scacheNum].SMHT + ospcs[cpuNum].transId;
  static byte* data      = 0;
  int          delay;

  ASSERT(ospcs[cpuNum].stalled);

  if (!data) {
     data = (byte *)malloc(SCACHE_LINE_SIZE);
     bzero((char*)data,SCACHE_LINE_SIZE);
     ASSERT(data);
  }

#ifndef SOLO
  /* let MAGIC know the access has timed out, so it should
   * generate an int when the data finally makes it here.
   */
  sim_magic_OSPC_stalled(cpuNum, smht->pAddr+K0BASE, 0);

  /* get the real OSPC data from MAGIC */
  delay = sim_magic_OSPC_access(cpuNum, smht->pAddr+K0BASE, data);
  ASSERT(delay == 0);

#endif

  /* complete the pending transaction */
  SCacheUpdate(cpuNum, smht,
	       ospcs[cpuNum].mode, ospcs[cpuNum].status, ospcs[cpuNum].result,
	       data);

#ifdef DEBUG_OSPC_STALL
  LogEntry("OSPC access",cpuNum,"pA=0%08x TIMED OUT\n",smht->pAddr);
#endif

  /* mark that all done */
  ospcs[cpuNum].stalled = 0;
}
#endif 

void
CacheCmdDone(int cpuNum, int transId, int mode, int status, int result,
	     byte *data)
{
   int   scacheNum = GET_SCACHE_NUM(cpuNum);
   SMHT *smht      = SCACHE[scacheNum].SMHT + transId;

#if !defined(SOLO) && !defined(SIM_ALPHA) && !defined(SIM_X86)
   int   delay;
#endif

#if !defined(SOLO) &&  !defined(SIM_X86)
   if (transId < 0) {
      DMAStatus res;
      if (status == MEMSYS_STATUS_SUCCESS) {
         res = DMA_OK;
      } else if (status == MEMSYS_STATUS_NAK) {
         res = DMA_NAK;
      } else {
         res = DMA_BAD;
      }
      DMACmdDone(transId, res);
      return;
   }
#else
   ASSERT (transId >= 0);
#endif

   if ((status == MEMSYS_STATUS_NAK) ||
       ((smht->memsyscmd & MEMSYS_CMDMASK) == MEMSYS_UPGRADE)) {
      SCacheUpdate(cpuNum, smht, mode, status, result, data);
      return;
   }

#if !defined(SOLO) && !defined(SIM_ALPHA) && !defined(SIM_X86)
   /* Support for OSPC's.
    *
    * For references to the OSPC area (enabled only when the remap is),
    * we let the reference go as usual through the memory system. When
    * the ref completes, we just substitute the OSPC data for the data
    * provided by the memory system.
    *
    * This is a nasty hack and should eventually be replaced by having
    * all memory accesses go through MAGIC (simmagic.c).
    */
   if ( IS_REMAPPED_PADDR(0, cpuNum)                                  &&
        (smht->memsyscmd & MEMSYS_CMDMASK) == MEMSYS_GET              &&
	status == MEMSYS_STATUS_SUCCESS                               &&
	(delay = sim_magic_OSPC_access(cpuNum,smht->pAddr+K0BASE,data)) != 0 ) {

#ifndef DATA_HANDLING
     ASSERT(0);
#endif

     /* must block on this access */
     ASSERT(!ospcs[cpuNum].stalled);
     ospcs[cpuNum].stalled = 1;
     ospcs[cpuNum].transId = transId;
     ospcs[cpuNum].mode    = mode;
     ospcs[cpuNum].status  = status;
     ospcs[cpuNum].result  = result;

#ifndef SOLO
     /* let MAGIC know the access has stalled, so it shouldn't
      * generate an int when the SIPS arrives.
      */
     sim_magic_OSPC_stalled(cpuNum, smht->pAddr+K0BASE, 1);
#endif

     /* set timeout */
     EventDoCallback(cpuNum, OSPCCallback, &ospcs[cpuNum].hdr, NULL, delay);

#ifdef DEBUG_OSPC_STALL
     LogEntry("OSPC access",cpuNum,"pA=0%08x STALLED\n",smht->pAddr);
#endif

     /* will get a call to CacheCmdUnstall when SIPS arrives, or will
      * time out, call OSPCCallback, and complete the access.
      */
     return;
   }
#endif
   SCacheUpdate(cpuNum, smht, mode, status, result, data);
}

void
CacheCmdUnstall(int cpuNum, VA vAddr)
{
#if defined(SIM_MIPS32) || defined(SIM_MIPS64)
  int          scacheNum = GET_SCACHE_NUM(cpuNum);
  SMHT*        smht      = SCACHE[scacheNum].SMHT + ospcs[cpuNum].transId;
  static byte* data      = 0;
  int          delay;
  

  if (ospcs[cpuNum].stalled && smht->pAddr+K0BASE == vAddr) {
    if (!data) {
       data = (byte*)malloc(SCACHE_LINE_SIZE);
       bzero(data,SCACHE_LINE_SIZE);
       ASSERT(data);
    }
#ifndef SOLO
    sim_magic_OSPC_stalled(cpuNum, smht->pAddr+K0BASE, 0);
    /* get the real OSPC data from MAGIC */
    delay = sim_magic_OSPC_access(cpuNum, vAddr, data);
    ASSERT(delay == 0);
#endif
    
    /* complete the pending transaction */
    SCacheUpdate(cpuNum, smht,
		 ospcs[cpuNum].mode, ospcs[cpuNum].status, ospcs[cpuNum].result,
		 data);
    /* mark that all done and kill timeout */
    ospcs[cpuNum].stalled = 0;
    EventCallbackRemove(&ospcs[cpuNum].hdr);
#ifdef DEBUG_OSPC_STALL
    LogEntry("OSPC access",cpuNum,"pA=0%08x GOT DATA\n",smht->pAddr);
#endif
  } /* else: must be other OSPC priority */
  
#endif
  
}  



VA   CacheGetVaddr(int cpuNum, int transId)
{
   int   scacheNum = GET_SCACHE_NUM(cpuNum);
   SMHT *smht      = SCACHE[scacheNum].SMHT + transId;
   ASSERT(transId >=0);
   return smht->vAddr;
}



  
/*****************************************************************
 * SCacheUpdate
 *****************************************************************/
static void
SCacheUpdate(int cpuNum, SMHT *smht, int mode, int status, int result,
	     byte *data)
{
   int cacheNum = GET_CACHE_NUM(cpuNum);
   int scacheNum = GET_SCACHE_NUM(cpuNum);
   int i;
   int reported = 0;
   uint type = 0;

   QUICK_ICACHE_CLEAR(cpuNum);

   if (status == MEMSYS_STATUS_ERROR) {
      for (i = 0; i < smht->numMHTwait; i++) {
         bool isIRef = FALSE;
         int mhtind  = smht->mhtInd[i];
         MHT *mht    = &(CACHE[cacheNum].MHT[mhtind]);   

         ASSERT (mhtind < MHT_SIZE);
         CPUPrint("MEMSYS ERROR: CPU %d at PC %#x Addr %#x\n", 
                  cpuNum, smht->PC, smht->vAddr);
         if (mht->cmd & SC_IGET) {
            isIRef = TRUE;
         }
         CPUVec.TakeBusError(cpuNum, isIRef, smht->vAddr);
         break;
      }
      FreeSMHT(cpuNum, smht - SCACHE[scacheNum].SMHT);
      return;
   }

   if (status == MEMSYS_STATUS_NAK) {
      if (smht->numMHTwait == 0) {
        SCACHE[scacheNum].stats.Prefetches--;
      }
      for (i = 0; i < smht->numMHTwait; i++) {
         int mhtind = smht->mhtInd[i];
         MHT *mht = &(CACHE[cacheNum].MHT[mhtind]);   
         ASSERT( mhtind < MHT_SIZE );
         if (mht->cmd & SC_IGET) {
            mht->stall +=
               CPUVec.FillCacheLine(cpuNum, mht->PC, mht->vAddr, mht->pAddr, E_I);
         } else {
            mht->stall +=
               CPUVec.FillCacheLine(cpuNum, mht->PC, mht->vAddr, mht->pAddr, E_D);
         }

         type |= MHTReqDone(cpuNum, mht, 0, MEMSYS_STATUS_NAK);
         /* We don't want nak'ed requests to count towards the
            number of misses... the number of read misses should just
            equal the number of read replies. */
         /* RPB: When there are multiple MHT entries waiting, only one of 
            them is counted as a miss by the stats; the rest are merges.
            In order to make the stats come out right, only the counter 
            for the first miss should be decremented. */
         /* If there is a prefetch in this entry, then the other requests 
            must have merged with it (since we would just punt the prefetch
            otherwise), so don't need to decrement the counter in that case
            either. */
         if ((i > 0) || (smht->pfvAddr > 0)) continue;
         if (mht->cmd & SC_IGET) {
            SCACHE[scacheNum].stats.IgetMisses--;
         } else { 
            switch (mht->cmd & (SC_DGET|SC_DGETX|SC_DUGETX|SC_DSCUGETX|SC_DLLGET|SC_DLLGETX)) {
            case SC_DGET:
            case SC_DLLGET:
               SCACHE[scacheNum].stats.DgetMisses--;
               break;
            case SC_DGETX:
            case SC_DLLGETX:
               SCACHE[scacheNum].stats.DgetXMisses--;
               break;
            case SC_DUGETX:
            case SC_DSCUGETX:
               SCACHE[scacheNum].stats.DupgradeMisses--;
               break;
            default:
               ASSERT(0);
            }
         }
      }
      SCACHE[scacheNum].stats.NAKs++;
      FreeSMHT(cpuNum, smht - SCACHE[scacheNum].SMHT);
      /* --- WARNING: This was put in because SYNCs mean that a request ---
       * --- arrival does not necessarily wake up the processor --- */
#ifdef rpb_notdef
      CPUVec.Unstall(cpuNum);
#endif
      return;
   }
   ASSERT(status == MEMSYS_STATUS_SUCCESS);
   SCachePUT(cpuNum, smht, mode, data);
   
   if (smht->numMHTwait == 0) {
      int stall = 0;
      type |= E_L2 | E_D;

      if (mode == MEMSYS_EXCLUSIVE) { 
         if (result & MEMSYS_RESULT_INVALIDATE) {
            type |= E_WRITE | E_CAUSED_INVAL;
         } else {
            type |= E_WRITE;
         }
      } else {
         type |= E_READ;
      }
      
      if ((smht->memsyscmd & MEMSYS_CMDMASK) == MEMSYS_UPGRADE) {
         type |= E_UPGRADE;
      }
      
      ASSERT((result & MEMSYS_RESULT_CACHE) 
             || (result & MEMSYS_RESULT_MEMORY));
      ASSERT(!((result & MEMSYS_RESULT_CACHE) 
               && (result &MEMSYS_RESULT_MEMORY)));
      
      if (result & MEMSYS_RESULT_CACHE) {
         type |= E_FOUND_IN_CACHE;
      }
            
      /*
       * Record the prefetch that did not incur a miss
       */
      if (interest(smht->pAddr)) {
         LogEntry("pfupdate",cpuNum,"vAddr=0x%x pAddr=0x%x type=%x\n",
                  smht->pfvAddr,smht->pAddr, type);
      }
      if (CPUVec.CPUstatus(cpuNum) != cpu_running) { 
         stall = CPUVec.CycleCount(cpuNum) - STATS_VALUE(cpuNum, stallStart);
      }

      if (type & E_UPGRADE) {
         PREFETCH_NOUPGRADE_EVENT(cpuNum,CURRENT_PC(cpuNum),smht->pfvAddr,
                                  smht->pAddr,smht->scacheLRU, stall);
         
      } else { 
         PREFETCH_NOMISS_EVENT(cpuNum,CURRENT_PC(cpuNum),smht->pfvAddr,
                               smht->pAddr,smht->scacheLRU, stall);      
      }         
      reported = 1;
   } else {
      for (i = 0; i < smht->numMHTwait; i++) {
         int mhtind = smht->mhtInd[i];
         MHT *mht = &(CACHE[cacheNum].MHT[mhtind]);   
         ASSERT (mhtind < MHT_SIZE);

         if (mht->cmd & SC_IGET) {
            /* Sending to Icache */

            /* Report to processor and MemStat (only once) */
            if (!reported) {
               type |=  E_L2 | E_I | E_READ;

               mht->stall += CPUVec.FillCacheLine(cpuNum, smht->PC, 
                                                  smht->vAddr, smht->pAddr, E_I);

               if (result & MEMSYS_RESULT_CACHE) {
                  type |= E_FOUND_IN_CACHE;
               }
               if (result & MEMSYS_RESULT_REMOTE_HOME) {
                  type |= E_REMOTE;
               }
               
               L2_IMISS_EVENT(CPUVec.CycleCount(cpuNum), cpuNum, mht->vAddr, 
                              smht->pAddr, mht->stall, type);
               reported = 1;
            }
            mht->status = status;
            /* WARNING:--- This does not work properly with SYNCs --- */
            type |= MHTReqDone(cpuNum, mht, mode, status);
         } else {
            /* Sending to Dcache */

            /* Report to processor and MemStat (only once) */
            if (!reported) {
               type |= E_L2 | E_D;
               mht->stall += CPUVec.FillCacheLine(cpuNum, smht->PC, 
                                                  smht->vAddr, smht->pAddr, E_D);

               if (mode == MEMSYS_EXCLUSIVE) { 
                  if (result & MEMSYS_RESULT_INVALIDATE) {
                     type |= E_WRITE | E_CAUSED_INVAL;
                  } else {
                     type |= E_WRITE;
                  }
               } else {
                  type |= E_READ;
               }
               
               if ((smht->memsyscmd & MEMSYS_CMDMASK) == MEMSYS_UPGRADE) {
                  type |= E_UPGRADE;
               }
               
               if (mht->cmd == SC_DSCUGETX) {
                  type |= E_SC_UPGRADE;
               }
               
               ASSERT((result & MEMSYS_RESULT_CACHE) 
                      || (result & MEMSYS_RESULT_MEMORY));
               ASSERT(!((result & MEMSYS_RESULT_CACHE) 
                        && (result &MEMSYS_RESULT_MEMORY)));

               if (result & MEMSYS_RESULT_CACHE) {
                  type |= E_FOUND_IN_CACHE;
               }
               if (result & MEMSYS_RESULT_REMOTE_HOME) {
                  type |= E_REMOTE;
               }

               L2_DMISS_EVENT(CPUVec.CycleCount(cpuNum), cpuNum, smht->PC, 
                              mht->vAddr, smht->pAddr, mht->stall, 
                              type, smht->scacheLRU);
               reported = 1;
            }
            if ((smht->memsyscmd & MEMSYS_CMDMASK) == MEMSYS_UPGRADE) {
               type |= MHTReqDone(cpuNum, mht, mode, status);
            } else {
               mht->status = status;
               type |= MHTReqDone(cpuNum, mht, mode, status);
            }

         }
      }
   }
   if (!reported) {
      CPUWarning("No mht entries for 2cache miss, not reported to memstat\n");
   }
   FreeSMHT(cpuNum, smht - SCACHE[scacheNum].SMHT);
   /* --- WARNING: This was put in because SYNCs mean that a request ---
    * --- arrival does not necessarily wake up the processor --- */
   CPUVec.Unstall(cpuNum);
}
 

/*****************************************************************
 * AllocSMHT 
 *
 * Allocate an entry in the miss handling table for the
 * specified second level miss. This routine is respondsible for 
 * checking for conflict and merge miss as well.
 *****************************************************************/
static SMHTStatus
AllocSMHT(int cpuNum, MCMD cmd, PA pAddr, int lru, int *mhtind)
{
   int scacheNum = GET_SCACHE_NUM(cpuNum);
   int entry;
   uint ind = SCACHE_INDEXOF(pAddr); /* /SCACHE_LINE_SIZE) % SCACHE_INDEX; */

   if (SCACHE[scacheNum].SMHTnumInuse == 0) { 
      entry = 0;  /* Special case for speed: allocate when SMHT is empty */
   } else { 
      int i;
      
      entry = -1;
      for (i = 0; i < SMHT_SIZE; i++) {
         if (!SCACHE[scacheNum].SMHT[i].inuse) {
	    if (entry == -1) 
	       entry = i;
	 } else {
	    if (SCACHE[scacheNum].SMHT[i].ind == ind) { 
	       *mhtind = i;
	       if (!SameCacheLine(SCACHE[scacheNum].SMHT[i].pAddr, pAddr,
				  SCACHE_LINE_SIZE)) {
		  /* Currently we only support one miss per cache index */
		  return SMHTCONFLICT;
	       }
               if ((SCACHE[scacheNum].SMHT[i].memsyscmd & MEMSYS_IFFLAVOR) != 
                   (cmd & MEMSYS_IFFLAVOR)) {
                   /* return conflict if both accesses are not of the
                    * same i/d flavor. In hardware, its not worth
                    * setting up a path to perform the fill on both
                    * caches. */
		  return SMHTCONFLICT;
	       }
	       /* Except we call them merges if two misses happen to the same
		* line */
               if (((SCACHE[scacheNum].SMHT[i].memsyscmd & MEMSYS_CMDMASK) ==
                    MEMSYS_GET) && ((cmd & MEMSYS_CMDMASK) == MEMSYS_GETX)) {
		  /* Call it a conflict if we need a GETX and only a
		   * GET is outstanding. 
		   */

		  /* --- THIS IS NOT RIGHT (MAH), if its the same cache --- 
		   * --- line it should MERGE and issue as an UPGRADE when ---
		   * --- the SHARED line comes back ---
                   * Currently, we only buffer writes when the request outstanding
                   * can satisfy the write when it returns.
		   */
		  return SMHTCONFLICT;
	       } 
	       return SMHTMERGE;
	    }
	 }
      }
      if (entry == -1) {
         if (i == SMHT_SIZE) { 
	    /* SMHT is full */
	    ASSERT(SCACHE[scacheNum].SMHTnumInuse == SMHT_SIZE); 
	    return SMHTFULL;
         } else {
            entry = i;
         }
      }
   }

   SCACHE[scacheNum].SMHTnumInuse++;
   SCACHE[scacheNum].SMHT[entry].inuse = 1;
   SCACHE[scacheNum].SMHT[entry].memsyscmd = cmd;
   SCACHE[scacheNum].SMHT[entry].pAddr = pAddr;
   /* RPB -- technically, this is not quite right.  Consider the case
      where the write buffer is active and a request is NAKed */
   SCACHE[scacheNum].SMHT[entry].PC = CURRENT_PC(cpuNum);
   SCACHE[scacheNum].SMHT[entry].ind = ind;
   SCACHE[scacheNum].SMHT[entry].scacheLRU = lru;
   SCACHE[scacheNum].SMHT[entry].numMHTwait = 0;
   SCACHE[scacheNum].SMHT[entry].pfvAddr = 0;
   {
      PA tag = SCACHE[scacheNum].set[ind].tags[lru];
      int formerState;
      if (tag & INVALID_TAG) {
         formerState = SCACHE_INVALID;
      } else if (tag & EXCLUSIVE_TAG) {
         formerState = SCACHE_DIRTY_EXCLUSIVE;
      } else {
         formerState = SCACHE_SHARED;
      }
      SCACHE[scacheNum].SMHT[entry].formerState = formerState;
   }
   *mhtind = entry;

   return SMHTSUCCESS;
}

/*****************************************************************
 * FreeSMHT - Delete a MHT entry
 *
 *****************************************************************/
static void
FreeSMHT(int cpuNum, int entryNum)
{
   int scacheNum = GET_SCACHE_NUM(cpuNum);

   ASSERT(SCACHE[scacheNum].SMHT[entryNum].inuse);
   SCACHE[scacheNum].SMHT[entryNum].inuse = 0;
   SCACHE[scacheNum].SMHTnumInuse--;
   ASSERT(SCACHE[scacheNum].SMHTnumInuse >= 0);
}


/*****************************************************************
 * EmptySMHT - Check whether SMHT is empty.
 *
 *****************************************************************/
int
EmptySMHT(int cpuNum)
{
  int scacheNum = GET_SCACHE_NUM(cpuNum);
  int entryNum;

  for (entryNum = 0; entryNum < SMHT_SIZE; entryNum++) {
    if (SCACHE[scacheNum].SMHT[entryNum].inuse) return FALSE;
  }
  return TRUE;
}


#ifdef CC_CHECKER
/* extern void MustBeInDirectory(int cpu, uint); */
static int
CacheChecker(PA paddr, int inCpu, int isExclusive)
{
   int cpu;
   int ind;
   bool foundit=0;

   if (paddr != (PA) -1) {
      PA tag = isExclusive ? SCACHE_TAG(paddr):SCACHE_TAG_EX(paddr);
      ind = SCACHE_INDEXOF(paddr) /* /SCACHE_LINE_SIZE) % SCACHE_INDEX; */
#ifdef DIR_CHECKER
      MustBeInDirectory(inCpu, paddr);
#endif
      for (cpu = 0; cpu < TOTAL_CPUS; cpu++) {
	 if (cpu == inCpu) 
	    continue;
     foundit=0;
	 if (isExclusive) {
        for(way=0;way<SCACHE_ASSOC;++way) {
           foundit |= ((SCACHE[cpu].set[ind].tags[way] & ~EXCLUSIVE_TAG) == tag);
        }
        if (!foundit) 
           CPUError("Cache check failed at paddr 0x%x\n", paddr);
	 } else {
            for(way=0;way<SCACHE_ASSOC;++way ) {
               foundit |= ((SCACHE[cpu].set[ind].tags[way]) == tag);
            }
	    if (!foundit)  
               CPUError("Cache check failed at paddr 0x%x\n", paddr);
	 }
      }
   }		    
   return 0;
}
#endif 


static void 
SCacheLRUTouch(uint *lruword, int set)
{
   switch( SCACHE_ASSOC ) {
   case 1: /* Direct-mapped */
      break;
   case 2:  /* 2-way */
      *lruword = set;
      break;
   case 4: { /* 4-way */
      unsigned char *lru;
      lru = (unsigned char *)lruword;
      if (lru[0] == set) {
         /* Hit on most recently used - do nothing */
      } else if (lru[1] == set) { 
         lru[1] = lru[0];
         lru[0] = set;
      } else if (lru[2] == set) {
         lru[2] = lru[1];
         lru[1] = lru[0];
         lru[0] = set;
      } else if (lru[3] == set) {
         lru[3] = lru[2];
         lru[2] = lru[1];
         lru[1] = lru[0];
         lru[0] = set;
      } else {
         CPUWarning("LRU screw-up in SCacheLRUTouch\n");
         ASSERT(0);
      }
      break;
   }
   case 5: 
   case 6: 
   case 7: 
   case 8: {
      unsigned char *lru;
      int t1, t2, i;

      lru = (unsigned char *)lruword;
      t2 = set;
      for (i=0; i<SCACHE_ASSOC; i++) {
         if (lru[i] == set) {
            break;
         } else {
            t1 = lru[i];
            lru[i] = t2;
            t2 = t1;
         }
      }
      ASSERT(lru[i] == set);
      lru[i] = t2;
      break;
   }
   default: 
      ASSERT(0);
   }
}


static int  
SCacheLRU(uint *lruword)
{
   switch( SCACHE_ASSOC ) {
   case 1: 
      return 0;
   case 2:
      return (!(*lruword));
   case 4: 
      return ((char *)lruword)[3];
   case 5: 
   case 6: 
   case 7: 
   case 8: 
      return ((char *)lruword)[SCACHE_ASSOC - 1];
   default:
      ASSERT(0);
   }
   return( 0 ); /*compiler happy */
}


static void 
SCacheLRUInit(uint *lruword)
{
   unsigned char* lru;
   int i;

   if (SCACHE_ASSOC == 4) {
      *lruword = 0x00010203;
      /* Need to have all sets here or LRU algorithm crashes */
   } else if (SCACHE_ASSOC > 4) {
      lru = (unsigned char*) lruword; 
      for (i=0; i<SCACHE_ASSOC; i++) {
         lru[i] = i;
      }
   } else {
      *lruword = 0;
   }
}

static void 
SCacheLRUMake(uint *lruword, int set)
{
   unsigned char* lru; 
   unsigned char t1,t2;
   int i;

   switch( SCACHE_ASSOC ) {
   case  1: 
      break;
   case 2: 
      *lruword = !set;
      break;
   case 4: 
      lru = (unsigned char*) lruword; 
      if (lru[3] != set) {	
         t1 = lru[2];				
         lru[2] = lru[3];			
         lru[3] = set;			   
         if (t1 != set) {			
            t2 = lru[1];				
            lru[1] = t1;				
            if (t2 != set)  lru[0] = t2;	
         }						
      }						
      break;
   case 5:
   case 6:
   case 7:
   case 8:
      lru = (unsigned char*) lruword; 
      t2 = set;
      for (i=SCACHE_ASSOC-1; i>=0; i--) {
         if (lru[i] == set) {
            break;
         } else {
            t1 = lru[i];
            lru[i] = t2;
            t2 = t1;
         }
      }
      ASSERT(lru[i] == set);
      lru[i] = t2;
      break;
      
   default:
      ASSERT( 0 );
   }
}

/* ======================================================================
   Utility routines
 
   - These routines need to be here, even though they may be specific
     to one memory model or the other

   ======================================================================*/

int
SameCacheLine(PA addr1, PA addr2, unsigned cacheLineSize)
{
   /* Now, match only if SAME cache line */
   return((addr1 & ~(PA)(cacheLineSize-1)) ==
      (addr2 & ~(PA)(cacheLineSize-1)));
}

PA
QuadWordAlign(PA offset)
{
   if (sizeof(PA) == 8) {
      return (offset & ~(0xFLL));
   } else {
      return (offset & ~(0xF));
   }
}

#ifdef DATA_HANDLING
/* write all dirty cache lines back to memory */
void FlushEntireCache(int cpuNum, int leaveShared)
{
#ifndef SOLO
   int scacheNum = GET_SCACHE_NUM(cpuNum);
   int numSets = SCACHE_SIZE/SCACHE_LINE_SIZE/SCACHE_ASSOC;
   int setNum, w;
   struct SCacheSet* set;
   int nDirty = 0;

   for (setNum = 0; setNum < numSets; setNum++) {
      set = &SCACHE[scacheNum].set[setNum];
      for(w=0;w<SCACHE_ASSOC; ++w) {
         if (set->tags[w] & EXCLUSIVE_TAG) {
            PA pAddr = STAG_TO_PA(set->tags[w], setNum);
            byte *memAddr = PHYS_TO_MEMADDR(M_FROM_CPU(cpuNum), pAddr);

            nDirty++;
            SCacheFlush(cpuNum, TRUE, leaveShared, pAddr, SCACHE_LINE_SIZE, memAddr);
         } else if (!(set->tags[w] & INVALID_TAG) && !leaveShared) {
            PA pAddr = STAG_TO_PA(set->tags[w], setNum);
#ifndef SOLO
            ASSERT (IS_VALID_PA(M_FROM_CPU(cpuNum), pAddr));
#endif
            SCacheFlush(cpuNum, FALSE, FALSE, pAddr, SCACHE_LINE_SIZE, NULL);
         }
      }
   }
   CPUPrint("FlushEntireCache: %d dirtly lines found in cache on cpu %d.\n",
              nDirty, cpuNum);
#endif
}
#endif



