/*
 * Copyright (C) 1996-1998 by the Board of Trustees
 *    of Leland Stanford Junior University.
 * 
 * This file is part of the SimOS distribution. 
 * See LICENSE file for terms of the license. 
 *
 */

/****************************************************************
 * eventcallback.c
 * 
 * Generic callback mechanism. 
 * 
 * $Author: stets $
 * $Date: 2000/02/19 00:02:09 $
 *****************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include "simutil.h"
#include "simtypes.h"
#include "sim_error.h"
#include "eventcallback.h"
#include "cpu_interface.h"
#include "machine_params.h"

/* Keep the asserts happy */

#define DEBUGLINE "%10u %-20s: rout %#x elm %#x arg %#x\n"


#ifdef notdef
#define EVENT_LOCK(_x)   { if( !CPUVec.singleEventQueue ) Simcp0Lock(  &(_x->lock)); }
#define EVENT_UNLOCK(_x) { if( !CPUVec.singleEventQueue ) Simcp0Unlock(&(_x->lock)); } 
#endif
#ifndef NO_EMBRA
#define EVENT_LOCK(_x) EventLock(_x)
#define EVENT_UNLOCK(_x) EventUnlock(_x)
extern void EventLock(int cpuNum);
extern void EventUnlock(int cpuNum);
#else
#define EVENT_LOCK(_x) 
#define EVENT_UNLOCK(_x)
#endif


EventQueue *eventQueues = 0;

/*****************************************************************
 * EventDoCallback
 *****************************************************************/
void EventDoCallback(int cpuNum, void (*rout)(int,EventCallbackHdr *,void*),
		     EventCallbackHdr *elm, void *arg, SimTime timeInFuture) 
{ 
   SimTime when, cpuTime; 
   EventCallbackHdr *next;  
   EventQueue *queue;

   EVENT_LOCK(cpuNum);

   ASSERT( elm && CPUVec.CycleCount );

   if (elm->active) {
     /* CPUWarning("Callback bounced on CPU %d\n", cpuNum); */
     EVENT_UNLOCK(cpuNum);
     return;
   }
     
   /* ASSERT( !elm->active ); */
   ASSERT( (int64)timeInFuture >= 0);

   /* On startup, GAMMA sets CycleCount to -1 and so
    * we need to ignore the CycleCount on the first cycle 
    * of execution.
    */
   cpuTime = CPUVec.CycleCount(cpuNum);
   if ((SimTime)-1 != cpuTime)
      when = cpuTime + timeInFuture;
   else
      when = timeInFuture;

   elm->rout = rout;
   elm->arg = arg;
   elm->active = TRUE;
   elm->cpuNum = cpuNum;
   if (CPUVec.singleEventQueue) { 
      queue = eventQueues;
   } else { 
      queue = eventQueues + cpuNum;
   }
   /*
    * This can happen in Embra MPinUP if the cycle counts
    * are off my more that the timeInFuture
    */
   if (when <= queue->lastCall && timeInFuture > 0) {
      when = queue->lastCall +1;
   }
   if (when < queue->lastCall && timeInFuture == 0) {
      when = queue->lastCall;
   }
   
   elm->when = when;
   ASSERT(when < SIM_MAX_TIME);
   next = &queue->sentinel;
   while (1) {
      ASSERT( next->next );
      if (when < next->next->when) {
         elm->next = next->next;
         next->next = elm;
         break;
      } else { 
         next = next->next;
      }
   }
   queue->calltime = queue->sentinel.next->when;
   EVENT_UNLOCK(cpuNum);
}

/*****************************************************************
 * EventQueueInsert
 *****************************************************************/
void EventQueueInsert(EventCallbackHdr *elm, int cpuNum)
{
   EventQueue *queue = eventQueues + cpuNum;
   EventCallbackHdr *next;  
   SimTime when = elm->when;
   
   EVENT_LOCK(cpuNum);

   next = &queue->sentinel;
   while (1) {
      ASSERT( next->next );
      if (when < next->next->when) {
         elm->next = next->next;
         next->next = elm;
         break;
      } else { 
         next = next->next;
      }
   }
   queue->calltime = queue->sentinel.next->when;
   EVENT_UNLOCK(cpuNum);
}

/*****************************************************************
 * EventProcess
 * cpuNum is ignored if all events are in the same queue
 *****************************************************************/
void
EventProcess(int cpuNum, SimTime now) 
{
    EventQueue *queue;
    EventCallbackHdr *h;
    if (CPUVec.singleEventQueue ) {
       queue = eventQueues;
       ASSERT( cpuNum < 0 );
    } else {
       queue = eventQueues + cpuNum;
       ASSERT( cpuNum >= 0 && cpuNum < TOTAL_CPUS );
    }
    h = &queue->sentinel;
    EVENT_LOCK(cpuNum);
    ASSERT(h);
    if (queue->lastCall >=  now) { 
#ifdef WANT_TO_FILL_LOG
       CPUPrint("EventCallback: processing events at %lld. Prev time: %lld\n",
                (uint64)now, (uint64)queue->lastCall);
#endif
    }
    while (h->next->when <= now) {
       EventCallbackHdr *elm = h->next;
       /*
        * Make sure that the time is also up for the
        * particular CPU. If not, bypass the callback
        */
       if (elm->when  <= CPUVec.CycleCount(elm->cpuNum)) {
          h->next = h->next->next; 
          elm->active = FALSE;
	  EVENT_UNLOCK(cpuNum);
          elm->rout(elm->cpuNum, elm, elm->arg);
          EVENT_LOCK(cpuNum);
          /* 
           * Restart from the top of the queue. Event might have been 
           * inserted during the processing
           */
          h = &queue->sentinel;
          /*
           * moved this here, so that it is updated only if
           * an event for the correct CPU is found 
           */ 
          queue->lastCall = now;
       } else {
          h=h->next;
       }
    }
    queue->calltime = queue->sentinel.next->when;
    EVENT_UNLOCK(cpuNum);
}

/*****************************************************************
 * EventCallbackWaiting
 *****************************************************************/
bool
EventCallbackWaiting(EventCallbackHdr * event)
{
   return event->active;
}

/*****************************************************************
 * EventCallbackUpdate
 * 
 *****************************************************************/
void
EventCallbackUpdate(EventCallbackHdr * event, SimTime newtime)
{
   int cpuNum = event->cpuNum;
   ASSERT(event->active);
   ASSERT(newtime < event->when);
   
   EventCallbackRemove(event);
   EventDoCallback(cpuNum,event->rout,event, event->arg, newtime - 
                   CPUVec.CycleCount(cpuNum));
}


/*****************************************************************
 * EventCallbackRemove
 *
 *****************************************************************/
void
EventCallbackRemove(EventCallbackHdr * event)
{
   EventQueue *queue;
   int cpuNum = event->cpuNum;
   EventCallbackHdr *next;
   
   if (CPUVec.singleEventQueue ) {
      queue = eventQueues;
   } else { 
      queue = eventQueues+cpuNum;
   }
   EVENT_LOCK(cpuNum);
   next = &queue->sentinel;
   while (next->next) {
      if (next->next ==  &queue->sentinel ) break;
      if (next->next == event) {
         next->next = event->next;
         event->active = FALSE;
         queue->calltime = queue->sentinel.next->when;
         EVENT_UNLOCK(cpuNum);
         return;
      }
      next = next->next;
   }
   {
      CPUWarning("EventCallbackRemove called on event %x not on list %d\n",
	      event, queue - eventQueues);
      CPUWarning("Attempting to remove from individual queue\n...");
      if (CPUVec.singleEventQueue == 1) {
	 CPUVec.singleEventQueue = 0;
	 EventCallbackRemove(event);
	 CPUVec.singleEventQueue = 1;
      }
   }

}


void EventCallbackInit(int numCPUs)
{
   int i;
   if (!eventQueues) {
      eventQueues = ZMALLOC(sizeof(EventQueue)*numCPUs,"EventCallback");
      ASSERT( eventQueues );
   }
   for(i=0;i<numCPUs;i++) {
      eventQueues[i].sentinel.next =  & eventQueues[i].sentinel;
      eventQueues[i].sentinel.when = SIM_MAX_TIME;
   }

}

/* Remove callbacks that we don't want to happen!!
 * Embra has a CheckInterruptCallback which we don't
 * want (or need) in other, non-parallel CPU simulators.
 */

void EventQueueRemove(int numCPUs, void (*rout)(int, struct EventCallbackHdr *, void *))
{
   int i;
   EventCallbackHdr *next;
   for (i = 0; i < numCPUs; i++) {
      next = eventQueues[i].sentinel.next;
      while (next->when < SIM_MAX_TIME) {
	 if (next->rout == rout)
	    EventCallbackRemove(next);
	 next = next->next;
      }
   }
}

/* XXX Return maximum time for event queue... */
SimTime EventQueueMaxTime(int cpuNum)
{
   SimTime result = 0;
   EventCallbackHdr *next;
   for (next = eventQueues[cpuNum].sentinel.next;
	next != &eventQueues[cpuNum].sentinel; next = next->next) {
      if (next->when > result)
	 result = next->when;
   }
   return result;
}

/* XXX Debugging: Take a look at the event queues when we
 * switch between simulators
 */

void EventQueueDump(int numCPUs)
{
   int i;
   EventCallbackHdr *next;
   for (i = 0; i < numCPUs; i++) {
      for (next = eventQueues[i].sentinel.next;
	   next != &eventQueues[i].sentinel; next = next->next) {
	 CPUWarning("Queue %d has event %x cpu %d at %lld with routine %x\n",
		    i, next, next->cpuNum, next->when, next->rout);
      }
   }
}


/*
 * EventQueueSplit: split into separate queues
 */

void EventQueueSplit(void)
{
   EventQueue *queue = eventQueues;
   EventCallbackHdr *next, *elm;
   EVENT_LOCK(0);
   for (next = &queue->sentinel;
	   next->next != &queue->sentinel; next = next->next) {
      /*
	 CPUWarning("EventQueueSplit: event for cpu %d\n", next->next->cpuNum);
      */
      
      while (next->next->cpuNum != 0) {
	 elm = next->next;
	 /* Delete from eventQueue[0] */
	 next->next = next->next->next;
	 /* Insert in individual queue */
	 EventQueueInsert(elm, elm->cpuNum);
      }
      
   }
   /* XXX Sanity check to make sure split worked!!  */
   for (next = &queue->sentinel;
	   next->next != &queue->sentinel; next = next->next) {
      if (next->next->cpuNum != 0) {
	 EventQueueDump(1);
	 CPUError("EventQueueSplit: split didn't work!!!\n");
      }
   }
   EVENT_UNLOCK(0);
}


/*
 * EventQueueMerge: merge into single event queue
 */

void EventQueueMerge(int numCPUs)
{
   int i;
   EventCallbackHdr *next, *elm;
   for (i = 1; i < numCPUs; i++) {
      EVENT_LOCK(i);
      next = &eventQueues[i].sentinel;
      while (next->next != &eventQueues[i].sentinel) {
	    elm = next->next;
	    /* Delete from eventQueue[i] and update next->next */
	    next->next = next->next->next;
	    /* Insert in eventQueue[0] */
	    EventQueueInsert(elm, 0);
      }
      EVENT_UNLOCK(i);
   }
}






