/*
 * Copyright (C) 1996-1998 by the Board of Trustees
 *    of Leland Stanford Junior University.
 * 
 * This file is part of the SimOS distribution. 
 * See LICENSE file for terms of the license. 
 *
 */

/****************************************************************
 * memstat.c
 *
 * Keep track of memory cache lines and the transitions that they go
 * through.  
 * 
 * Last Modified by: $Author: bosch $
 * $Date: 2000/05/12 19:31:01 $
 ****************************************************************/
 
/* #define DEBUGGING */ 
#include <stdio.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/mman.h>
#include <fcntl.h>
#include <unistd.h>
#include <string.h>
#include <errno.h>
#include <stddef.h>
#include <stdarg.h>
#include "sim_error.h"
#include "simutil.h" 
#include "simrecord.h" 
#include "cpu_interface.h"
#include "memstat.h"
#include "annotations.h"
#include "machine_params.h"
#include "statrecord.h"
#include "false_sharing.h"
#include "hw_events.h"
#include "tcl_init.h"

/* Get rid of this */
#include "../../memsystems/memsys.h"

/*******************************************************************
 * The #define NO_RACE_CONDITION option can be used to debug the 
 * caches by verifying that only certain state transitions actually
 * take place. NO_RACE_CONDITION can only be used with UMA,NUMA and
 * PerfectMem, not with Flash-lite 
 *******************************************************************/
#define NO_RACE_CONDITION ( memsysVec.type != FLASHLITE )

/********************************************************************
 * MemStat state machine. 
 * This simple FSM consists of the states CACHE_COLD ... CACHE_WARMED
 * and is managed for every physical line and processor of the machine
 * The state is encoded in 16 bits, the top 12 being reserved to store
 * the PID of the process that evicted the line. 
 ********************************************************************/

#define CACHE_COLD        0  
#define CACHE_REPLACED_K  1
#define CACHE_REPLACED_U  2
#define CACHE_INVALIDATED 3
#define CACHE_REPL_INV    4
#define CACHE_INCACHE     5
#define CACHE_WARMED      6

char *categName[] =  {"ice","cold",
                      "inv","capinv",
                      "cap","Kcap","SelfCap","OtherCap"};

#define MS_MAXPID         4095  /* 2**12 -1 */
#define PIDOFFSET         4

#ifdef SIM_ORIGIN
#define MAGIC_NODE_SIZE_BITS            32
#define MAGIC_NODE_OFFSET_MASK          ((1LL << MAGIC_NODE_SIZE_BITS) - 1)

#define MEM_NODE(_pA)        ((_pA) >> MAGIC_NODE_SIZE_BITS)
#define MEM_ADDR(_pA)        ((_pA) & MAGIC_NODE_OFFSET_MASK)

#define MEM_LINE(_m,_pA)     \
    ((PA)(MEM_NODE(_pA)*(MEM_SIZE(_m)/NUM_MEMORIES(_m)) + MEM_ADDR(_pA)) >> log2SCACHE_LINE_SIZE)
#define MEM_L1D_LINE(_m,_pA) \
    ((PA)(MEM_NODE(_pA)*(MEM_SIZE(_m)/NUM_MEMORIES(_m)) + MEM_ADDR(_pA)) >> log2DCACHE_LINE_SIZE)
#define MEM_L1I_LINE(_m,_pA) \
    ((PA)(MEM_NODE(_pA)*(MEM_SIZE(_m)/NUM_MEMORIES(_m)) + MEM_ADDR(_pA)) >> log2ICACHE_LINE_SIZE)

#define NUM_NODES(_m) ((NUM_CPUS(_m) + 1)/ 2)
#define MEM_LINES(_m)        ((NUM_NODES(_m)*(MEM_SIZE(_m)/NUM_MEMORIES(_m))) >>log2SCACHE_LINE_SIZE)
#define MEM_L1D_LINES(_m)    ((NUM_NODES(_m)*(MEM_SIZE(_m)/NUM_MEMORIES(_m)))>>log2DCACHE_LINE_SIZE)
#define MEM_L1I_LINES(_m)    ((NUM_NODES(_m)*(MEM_SIZE(_m)/NUM_MEMORIES(_m)))>>log2ICACHE_LINE_SIZE)
#else
#define MEM_LINE(_m, _pA)     ((_pA)>>log2SCACHE_LINE_SIZE)
#define MEM_L1D_LINE(_m, _pA) ((_pA)>>log2DCACHE_LINE_SIZE)
#define MEM_L1I_LINE(_m, _pA) ((_pA)>>log2ICACHE_LINE_SIZE)
#define MEM_LINES(_m)        (MEM_SIZE(_m)>>log2SCACHE_LINE_SIZE)
#define MEM_L1D_LINES(_m)    (MEM_SIZE(_m)>>log2DCACHE_LINE_SIZE)
#define MEM_L1I_LINES(_m)    (MEM_SIZE(_m)>>log2ICACHE_LINE_SIZE)

#endif


#define MS_pid(_cpu) 0

#define MemStatNone 0
#define MemStatGlobal 1
#define MemStatGlobalPID 2

int memstatOption = MemStatGlobal;

static struct MemStatState {
   /* State machine turned on or not! */
   int init;  
   int trackPID;
   int trackL1;

   uint cacheMask;
   int bitsPerState;
   /* L2 State machines */
   uint lastAddr[MAX_MACHINES];
   uint trNumCPUs[MAX_MACHINES];
   uint trWarnings[MAX_MACHINES];
   uint *trPtr[MAX_MACHINES];

   /* L1 State machines (cap/compulsory) */
   uint *l1DtrPtr[MAX_MACHINES];
   uint *l1ItrPtr[MAX_MACHINES];

   /*
    * PID mappings. 
    * Memstat pid table. Hive uses pids sparsly. We therefore 
    * use an indirection table to index the pids. This saves us
    * a couple of bits (exactly 4) under the safe assumption that
    * no more than 4K processes will be spawned during the run.
    */
   int pidCount[MAX_MACHINES];
   int pidTable[MAX_MACHINES][32*1024];
} memStat;

struct MemStatBucket {
   /* Instruction & Data */
   StatRecordFieldDesc dStall;
   StatRecordFieldDesc dReadStall;
   StatRecordFieldDesc dWriteStall;
   StatRecordFieldDesc dL2Stall;
   StatRecordFieldDesc dUpgradeStall;
   StatRecordFieldDesc dL2DirtyStall;
   StatRecordFieldDesc dL2ReadStall;
   StatRecordFieldDesc dL2WriteStall;

   StatRecordFieldDesc prefMHTStall;
   StatRecordFieldDesc readMHTStall;
   StatRecordFieldDesc writeMHTStall;
   StatRecordFieldDesc loadBypassStall;
   StatRecordFieldDesc SMHTStall;

   StatRecordFieldDesc dL1Miss;
   StatRecordFieldDesc dL2Miss;
   StatRecordFieldDesc dL2DirtyMiss;
   StatRecordFieldDesc dL1CapMiss;
   StatRecordFieldDesc dUpgrade;                 
   StatRecordFieldDesc dUpgradeNoInval;
   StatRecordFieldDesc dUpgradeSC;
   StatRecordFieldDesc dUpgradeRemote;
   
   StatRecordFieldDesc read[NUM_CATEGS];
   StatRecordFieldDesc write[NUM_CATEGS];   
   
   /* Instruction only */
   StatRecordFieldDesc instr[NUM_CATEGS];
   StatRecordFieldDesc iL1Miss;
   StatRecordFieldDesc iL2Miss;
   StatRecordFieldDesc iStall;
   StatRecordFieldDesc iL2Stall;

   StatRecordFieldDesc iL2DirtyMiss;
   StatRecordFieldDesc iL1CapMiss;

   StatRecordFieldDesc dL2RemoteMiss;
   StatRecordFieldDesc dL2RemoteStall;

   StatRecordFieldDesc dL2RemoteUpgrade;
   StatRecordFieldDesc dL2RemoteUpgradeStall;

   StatRecordFieldDesc iL2RemoteMiss;
   StatRecordFieldDesc iL2RemoteStall;

   /* Pipeline stall fields -- Sam only */

   StatRecordFieldDesc pipeStall;
   StatRecordFieldDesc pipeICacheStall;
   StatRecordFieldDesc pipeMispredictStall;
   StatRecordFieldDesc pipeTrapStall;
   StatRecordFieldDesc pipeDrainStall;
   StatRecordFieldDesc pipeIntStall;
   StatRecordFieldDesc pipeFPStall;
   StatRecordFieldDesc pipeLdStStall;
   StatRecordFieldDesc pipeOtherStall;
} memStatBucket;


/***********************************************************************
 * Statics and #defines
 **********************************************************************/
#define STAT_RECORD(_f,_inc) \
               StatRecordEntry(cpuNum,pc,vAddr,memStatBucket._f,_inc);
#define STAT_RECORD_I(_f,_inc) \
               StatRecordEntry(cpuNum,pc,0,memStatBucket._f,_inc);
#define STAT_RECORD_D(_f,_inc) \
               StatRecordEntry(cpuNum,0,vAddr,memStatBucket._f,_inc);

static int  CacheTrMiss(int cpu, PA pAddr);

int interest(PA addr);
#define ITRACE(s)  CPUWarning("%10lld (%s,%x,%d)\n", \
      (uint64)CPUVec.CycleCount(cpuNum),s,pAddr & (SCACHE_LINE_SIZE-1), cpuNum);

/*****************************************************************
 * MemStatInitCategoryFields
 *
 * Go through the cache miss categories and define fields for them.
 * This can be used by anyone needing the cache fields.
 *****************************************************************/
void 
MemStatInitCategoryFields(StatRecordFieldDesc *desc, char *tag, int type)
{
   int i;
   char str[128];
   ASSERT (memStat.init);
#ifndef USE_FLASHLITE
   for (i=0; i<NUM_CATEGS; i++) {
      int f = 1;
      switch (i) { 
      case CATEG_cap: 
         if( memStat.trackPID ) f = 0;
         break;
      case CATEG_Kcap: 
      case CATEG_OtherCap: 
      case CATEG_SelfCap: 
          if( !memStat.trackPID ) f=0;
          break;
      }
      if (f) {
         sprintf(str,"%s.%s", tag, categName[i]);
         desc[i] = StatRecordDefineField(str, type);
      }
   }
#endif
}

/****************************************************************************
 * MemStatInit
 ****************************************************************************/
#define DEF_FIELD(_field,_fl) (memStatBucket._field = StatRecordDefineField(# _field,_fl))

extern int isSolo;

void 
MemStatEarlyInit(void)
{
   int machine, i;
   for(machine=0; machine<NUM_MACHINES; machine++) {
       for(i=0;i<32*1024;i++) {
           memStat.pidTable[machine][i] = -1;
       }
   }

   /* Data-cache related counters */
   DEF_FIELD(dL1Miss,     STATRECORD_INSTRUCTION|STATRECORD_DATA);               
   DEF_FIELD(dL2Miss,     STATRECORD_INSTRUCTION|STATRECORD_DATA);
   DEF_FIELD(dL2DirtyMiss, STATRECORD_INSTRUCTION|STATRECORD_DATA);
   DEF_FIELD(dUpgrade,        STATRECORD_INSTRUCTION|STATRECORD_DATA);
   DEF_FIELD(dUpgradeNoInval, STATRECORD_INSTRUCTION|STATRECORD_DATA);
   DEF_FIELD(dUpgradeSC,      STATRECORD_INSTRUCTION|STATRECORD_DATA); 

   /* data stall time */
   DEF_FIELD(dStall,      STATRECORD_INSTRUCTION|STATRECORD_DATA);
   DEF_FIELD(dReadStall,      STATRECORD_INSTRUCTION|STATRECORD_DATA);
   DEF_FIELD(dWriteStall,      STATRECORD_INSTRUCTION|STATRECORD_DATA);
   DEF_FIELD(dL2Stall,    STATRECORD_INSTRUCTION|STATRECORD_DATA);
   DEF_FIELD(dUpgradeStall,   STATRECORD_INSTRUCTION|STATRECORD_DATA);
   DEF_FIELD(dL2DirtyStall, STATRECORD_INSTRUCTION|STATRECORD_DATA);
   DEF_FIELD(dL2ReadStall, STATRECORD_INSTRUCTION|STATRECORD_DATA);
   DEF_FIELD(dL2WriteStall, STATRECORD_INSTRUCTION|STATRECORD_DATA);

   DEF_FIELD(prefMHTStall, STATRECORD_INSTRUCTION|STATRECORD_DATA);
   DEF_FIELD(readMHTStall, STATRECORD_INSTRUCTION|STATRECORD_DATA);
   DEF_FIELD(writeMHTStall, STATRECORD_INSTRUCTION|STATRECORD_DATA);
   DEF_FIELD(loadBypassStall, STATRECORD_INSTRUCTION|STATRECORD_DATA);
   DEF_FIELD(SMHTStall, STATRECORD_INSTRUCTION|STATRECORD_DATA);

   /* Instruction-cache related counters */
   DEF_FIELD(iL1Miss,     STATRECORD_INSTRUCTION); 
   DEF_FIELD(iL2Miss,     STATRECORD_INSTRUCTION );
   DEF_FIELD(iStall,      STATRECORD_INSTRUCTION );
   DEF_FIELD(iL2Stall,    STATRECORD_INSTRUCTION );
   DEF_FIELD(iL2DirtyMiss, STATRECORD_INSTRUCTION);

   /* Local/Remote memory home related counters */
   DEF_FIELD(iL2RemoteMiss,     STATRECORD_INSTRUCTION); 
   DEF_FIELD(iL2RemoteStall,     STATRECORD_INSTRUCTION); 
   DEF_FIELD(dL2RemoteMiss,     STATRECORD_INSTRUCTION|STATRECORD_DATA); 
   DEF_FIELD(dL2RemoteStall,     STATRECORD_INSTRUCTION|STATRECORD_DATA); 
   DEF_FIELD(dL2RemoteUpgrade,     STATRECORD_INSTRUCTION|STATRECORD_DATA); 
   DEF_FIELD(dL2RemoteUpgradeStall,     STATRECORD_INSTRUCTION|STATRECORD_DATA); 

   /* Pipeline stall fields -- Sam only */
   DEF_FIELD(pipeStall, STATRECORD_INSTRUCTION);
   DEF_FIELD(pipeICacheStall, STATRECORD_INSTRUCTION);
   DEF_FIELD(pipeMispredictStall, STATRECORD_INSTRUCTION);
   DEF_FIELD(pipeTrapStall, STATRECORD_INSTRUCTION);
   DEF_FIELD(pipeDrainStall, STATRECORD_INSTRUCTION);
   DEF_FIELD(pipeIntStall, STATRECORD_INSTRUCTION);
   DEF_FIELD(pipeFPStall, STATRECORD_INSTRUCTION);
   DEF_FIELD(pipeLdStStall, STATRECORD_INSTRUCTION);
   DEF_FIELD(pipeOtherStall, STATRECORD_INSTRUCTION);

   switch (memstatOption) {
   case MemStatNone:
      memStat.init = 0;
      return;
   case MemStatGlobal: 
      memStat.init = 1;
      /*  memStat.trackL1 = 1; */
      memStat.trackPID = 0;
      memStat.bitsPerState = 4;
      memStat.cacheMask = 0xf;     
      break;
   case MemStatGlobalPID:
      memStat.init = 1;
      /*   memStat.trackL1 = 1; */
      memStat.trackPID = 1;
      memStat.bitsPerState = 4+12;
      memStat.cacheMask = 0xffff;
      break;
   default: 
      ASSERT(0);
   }
   memStat.init = 1;

   /*
    * Additional counters, enabled only when the state machine is running
    */
   if (memStat.trackL1) {
      DEF_FIELD(dL1CapMiss,      STATRECORD_INSTRUCTION|STATRECORD_DATA);
   }
   MemStatInitCategoryFields(memStatBucket.read, "read",
                             STATRECORD_INSTRUCTION|STATRECORD_DATA);
   MemStatInitCategoryFields(memStatBucket.write, "write",
                             STATRECORD_INSTRUCTION|STATRECORD_DATA);
   if (memStat.trackL1) { 
      DEF_FIELD(iL1CapMiss,  STATRECORD_INSTRUCTION );
   }
   MemStatInitCategoryFields(memStatBucket.instr, "instr", 
                             STATRECORD_INSTRUCTION);   
}

void 
MemStatLateInit(void)
{
    int machine;

    if( !memStat.init) return;
#ifdef USE_FLASHLITE
    return;
#endif
    for (machine = 0; machine < NUM_MACHINES; machine++) {
        for (memStat.trNumCPUs[machine]=1;
             memStat.trNumCPUs[machine] < NUM_CPUS(machine); 
             memStat.trNumCPUs[machine] *=2) {
            continue;
        }
        memStat.lastAddr[machine] = memStat.bitsPerState *
            memStat.trNumCPUs[machine] * (MEM_LINES(machine) / 8);
        ASSERT (memStat.lastAddr[machine]);
        CPUPrint("MemStat: init cache state transitions for mach %d, %i proc,len=%#x\n",
                 machine, memStat.trNumCPUs[machine],
                 memStat.lastAddr[machine]);
        memStat.trPtr[machine] = (unsigned int *) 
            ZALLOC_PERM(memStat.lastAddr[machine],"MemStat");
        if (memStat.trackL1) {
            ASSERT (log2DCACHE_LINE_SIZE);
            ASSERT (log2ICACHE_LINE_SIZE);
            ASSERT (log2ICACHE_SIZE);
            ASSERT (log2DCACHE_SIZE);
            ASSERT (NUM_CPUS(machine));
            memStat.l1DtrPtr[machine] = (unsigned int *) 
                ZALLOC_PERM(NUM_CPUS(machine) *
                            MEM_L1D_LINES(machine) / 8,
                            "MemStat::L1D"); 
            memStat.l1ItrPtr[machine] = (unsigned int *) 
                ZALLOC_PERM(NUM_CPUS(machine) *
                            MEM_L1I_LINES(machine) / 8,
                            "MemStat::L1I"); 
        }
    }
}

void
MemStatImiss(MemStatTime cycle, int cpuNum, VA pc, PA pAddr,
             MemStatTime stallTime, int missType) 
{ 
   int categ = -1;
   int machine;
   int mcpu;

   if (!memStat.init) return;

   machine = M_FROM_CPU(cpuNum);
   mcpu = MCPU_FROM_CPU(cpuNum);
   ASSERT(!(missType & E_D));

#ifdef DEBUGGING
   if( interest(pAddr) ) {
      CPUWarning("%10lld IMISS cpu=%i pAddr=%#x type=%i\n",
                 CPUVec.CycleCount(cpuNum), cpuNum, pAddr,missType);
   }
#endif

   ASSERT( (missType & E_L1) || (missType & E_L2));
#ifndef USE_FLASHLITE
   if (memStat.init) {
      if (!(missType & E_L1)) {
         categ = CacheTrMiss(cpuNum,pAddr);
         ASSERT(categ >= 0);
      }
   } else {
      categ = -1;
   }
#else
   categ = -1;
#endif

   /*
    * L1 cache miss
    */
   if (missType & E_L1) {
      uint l1Line  = mcpu * MEM_L1I_LINES(machine) +
          MEM_L1I_LINE(machine, pAddr); 
      uint l1Index = l1Line /32;
      uint l1Pos   = l1Line %32; 
      
      STAT_RECORD_I(iL1Miss,1);
      STAT_RECORD_I(iStall,stallTime);

      /* L1 capacity */
      if (memStat.trackL1) {
         if (memStat.l1ItrPtr[machine][l1Index] & (1<<l1Pos)) { 
            STAT_RECORD_I(iL1CapMiss,1);
         }
         memStat.l1ItrPtr[machine][l1Index] |= (1<<l1Pos);
      }
      return;
   }
   /*
    * L2 cache miss
    */
   if (missType & E_L2) {
      int i;
      STAT_RECORD_I(iL2Miss,1);
      STAT_RECORD_I(iL2Stall, stallTime);
      if (missType & E_FOUND_IN_CACHE) {
         STAT_RECORD_I(iL2DirtyMiss,1);
      }
      if (missType & E_REMOTE) {
         STAT_RECORD_I(iL2RemoteMiss,1);
         STAT_RECORD_I(iL2RemoteStall,stallTime);
      }

      if (memStat.trackL1) {
         for(i=0; i < SCACHE_LINE_SIZE; i += ICACHE_LINE_SIZE) {
            /* clear the capacity bits in the L1 cache */
            PA  L1pA    = (pAddr & ~(SCACHE_LINE_SIZE-1)) + i;
            uint l1Line  = mcpu * MEM_L1I_LINES(machine) +
                MEM_L1I_LINE(machine, L1pA); 
            uint l1Index = l1Line /32;
            uint l1Pos   = l1Line %32; 
            memStat.l1ItrPtr[machine][l1Index] &= ~(1<<l1Pos);
         }
      }
#ifndef USE_FLASHLITE
      if (memStat.init) {
         ASSERT(categ >= 0 && categ < NUM_CATEGS);
         StatRecordEntry(cpuNum,pc,0,memStatBucket.instr[categ],1);
      }
#endif
   }
}


int
MemStatDmiss(MemStatTime cycle, int cpuNum, VA pc, VA vAddr, 
             PA pAddr, MemStatTime stallTime, int missType, int way)
{
   int categ=-1,i;
   int machine;
   int mcpu;

   if (!memStat.init) return 0;

   machine = M_FROM_CPU(cpuNum);
   mcpu = MCPU_FROM_CPU(cpuNum);

#ifdef DEBUGGING
   if (interest(pAddr)) { 
      CPUWarning("%10lld DMISS cpu=%i pAddr=%#x type=%#x state=%#x upgrade=%i\n",
                 CPUVec.CycleCount(cpuNum), cpuNum, pAddr, missType, 
                 MSCacheState(cpuNum, pAddr),
                 ((missType&E_UPGRADE)!=0));
   }
#endif
   
   if ((missType & E_L1) || ((missType & E_L2) && (missType & E_UPGRADE))) {
      /* no cache transition */
   } else {
      /* transition */
      ASSERT(missType & E_L2);
#ifndef USE_FLASHLITE
      if (memStat.init) {
         categ = CacheTrMiss(cpuNum,pAddr);
         if (missType & E_D) { 
            FalseSharingDefineOffset(cpuNum,pAddr,way,categ);
         }
      } else {
         categ = -1;
      }
#else
      categ = -1;
#endif
   }
   ASSERT((missType & (E_L2 | E_L1)));

   if (missType & E_L1) {
      uint l1Line  = mcpu * MEM_L1D_LINES(machine) +
          MEM_L1D_LINE(machine, pAddr); 
      uint l1Index = l1Line /32;
      uint l1Pos   = l1Line %32; 
      STAT_RECORD(dL1Miss,1);
      STAT_RECORD(dStall,stallTime);
      /* 
       * Classification of miss as read/write is complicated by write buffer:
       * 1. If it's a merge, it must be a read that stalled the CPU (because 
       * a merging write would not have stalled the CPU)
       * 2. If it's a conflict, it must be a write that stalled the CPU
       * (the original miss was a write or else the CPU would have stalled)
       * 3. If it's neither (always true if there's no write buffering),
       * then go by E_READ or E_WRITE
       */
      if (missType & E_MERGE) {
         STAT_RECORD(dReadStall,stallTime);
      } else if (missType & E_CONFLICT) {
         STAT_RECORD(dWriteStall,stallTime);
      } else if (missType & E_READ) {
         STAT_RECORD(dReadStall,stallTime);
      } else if (missType & E_WRITE) {
         STAT_RECORD(dWriteStall,stallTime);
      } else {
         ASSERT(0);
      }

      if (memStat.trackL1) {
         /* L1 capacity */
         if (memStat.l1DtrPtr[machine][l1Index] & (1<<l1Pos)) { 
            STAT_RECORD(dL1CapMiss,1);
         }
         memStat.l1DtrPtr[machine][l1Index] |=  (1<<l1Pos);
#ifdef DEBUGGING
         if (interest(pAddr)) { 
             CPUWarning("%10lld L1 Ref cpu=%i pAddr=%#x type=%#x state=%#x\n",
                       CPUVec.CycleCount(cpuNum), cpuNum, pAddr, missType, 
                       MSCacheState(cpuNum, pAddr));
         }
#endif
      }
      return -1;
   } 

   if ((missType & E_L2) && (missType & E_UPGRADE)) {
#ifndef USE_FLASHLITE
      if (memStat.init && 
          NO_RACE_CONDITION && 
          MSCacheState(cpuNum,pAddr) != CACHE_INCACHE) {
         CPUError("MemStat: %10lld  cpu=%i pAddr=%#x UPGRADE NOT IN CACHE \n",
                  CPUVec.CycleCount(cpuNum), cpuNum, pAddr);
      }
#endif
      STAT_RECORD(dUpgrade,1);
      /* 
       * Upgrade classification is complicated by the write buffer -- 
       * see the explanation above for details
       */
      if (missType & E_MERGE) {
         STAT_RECORD(dL2ReadStall,stallTime);
         STAT_RECORD(dL2Stall, stallTime);
      } else if (missType & E_CONFLICT) {
         STAT_RECORD(dL2WriteStall,stallTime);
         STAT_RECORD(dL2Stall, stallTime);
      } else {
         STAT_RECORD(dUpgradeStall, stallTime);
      }
      if (!(missType & E_CAUSED_INVAL)) {
         STAT_RECORD(dUpgradeNoInval,1);
      }
      if (missType & E_SC_UPGRADE) { 
         STAT_RECORD(dUpgradeSC, 1);
      }
      if (missType & E_REMOTE) {
         STAT_RECORD_I(dL2RemoteUpgrade,1);
         STAT_RECORD_I(dL2RemoteUpgradeStall,stallTime);
      }

#ifdef DEBUGGING
      if (interest(pAddr)) { 
         CPUWarning("%10lld L2-upgrade cpu=%i pAddr=%#x type=%#x state=%#x\n",
                    CPUVec.CycleCount(cpuNum), cpuNum, pAddr, missType, 
                    MSCacheState(cpuNum, pAddr));
      }
#endif
      return -1;
   } 
    
   /* L2 cache Miss */
   ASSERT(missType & E_L2);
   STAT_RECORD(dL2Miss, 1);
   STAT_RECORD(dL2Stall, stallTime);
   if (missType & E_FOUND_IN_CACHE) {
      STAT_RECORD(dL2DirtyMiss,1);
      STAT_RECORD(dL2DirtyStall,stallTime);
      /* Dirty stall is always remote */
      missType |= E_REMOTE;
   }
   if (missType & E_REMOTE) {
      STAT_RECORD_I(dL2RemoteMiss,1);
      STAT_RECORD_I(dL2RemoteStall,stallTime);
   }

   if (memStat.trackL1) {
      /*
       * clear the capacity bits in the L1 cache 
       */
      for(i=0; i<SCACHE_LINE_SIZE; i+= DCACHE_LINE_SIZE) {
         PA  L1pA   = (pAddr & ~(SCACHE_LINE_SIZE-1)) + i;
         uint l1Line = mcpu * MEM_L1D_LINES(machine) +
             MEM_L1D_LINE(machine, L1pA); 
         uint l1Index = l1Line /32;
         uint l1Pos   = l1Line %32; 
         memStat.l1DtrPtr[machine][l1Index] &=  ~(1<<l1Pos);
      }
   }

   /* 
    * Miss classification is complicated by the write buffer -- 
    * see the explanation above for details
    */
   if (missType & E_MERGE) {
      STAT_RECORD(dL2ReadStall,stallTime);
   } else if (missType & E_CONFLICT) {
      STAT_RECORD(dL2WriteStall,stallTime);
   } else if (missType & E_READ) {
      STAT_RECORD(dL2ReadStall,stallTime);
   } else if (missType & E_WRITE) {
      STAT_RECORD(dL2WriteStall,stallTime);
   } else {
      ASSERT(0);
   }

   if (!memStat.init) return -1;
#ifdef USE_FLASHLITE
   return -1;
#endif
   ASSERT(categ >= 0);

   if (interest(pAddr)) {
      CPUWarning("%10lld CacheTrmiss : %i\n",CPUVec.CycleCount(cpuNum),categ);
   }
   if (missType & E_WRITE) {
      StatRecordEntry(cpuNum, pc, vAddr, memStatBucket.write[categ], 1);
   } else  { 
      StatRecordEntry(cpuNum, pc, vAddr, memStatBucket.read[categ], 1);
   }
   return missType;
}

void
MemStatPipelineStall(int cpuNum, VA pc, VA vAddr,
                     MemStatTime stallTime, int stallType)
{
   STAT_RECORD_I(pipeStall, stallTime);
   switch (stallType) {
   case E_PIPE_ICACHE:
      STAT_RECORD_I(pipeICacheStall, stallTime);
      break;
   case E_PIPE_MISPREDICT:
      STAT_RECORD_I(pipeMispredictStall, stallTime);
      break;
   case E_PIPE_TRAP:
      STAT_RECORD_I(pipeTrapStall, stallTime);
      break;
   case E_PIPE_DRAIN:
      STAT_RECORD_I(pipeDrainStall, stallTime);
      break;
   case E_PIPE_INT:
      STAT_RECORD_I(pipeIntStall, stallTime);
      break;
   case E_PIPE_FP:
      STAT_RECORD_I(pipeFPStall, stallTime);
      break;
   case E_PIPE_LDST:
      STAT_RECORD_I(pipeLdStStall, stallTime);
      break;
   case E_PIPE_OTHER:
      STAT_RECORD_I(pipeOtherStall, stallTime);
      break;
   }
}

void  
MemStatCacheTrans(int cpuNum, PA pAddr, int type, VA evictorAddr, int isUserMode)
{
   unsigned int index,pos,state,off;
   int machine;
   int mcpu;

   if (!memStat.init) return;
#ifdef USE_FLASHLITE
   return;
#endif

   machine = M_FROM_CPU(cpuNum);
   mcpu = MCPU_FROM_CPU(cpuNum);

   /* debug */
   if (interest(pAddr)) {
      char c[5];
      c[0] = 't';
      if ((type & E_L1) && (type & E_I)) { 
         c[1] = 'i';
      }
      else if ((type & E_L1) && (type & E_D)) { 
         c[1] = 'd';
      }
      else if (type & E_L2) {
         c[1] = '2';
      }
      else 
         ASSERT(0);
      
      if ((!(type & E_EXTERNAL)) && (!(type & E_FAKE_EXTERNAL))) {
         /* Line was replaced */
         c[2] = 'r';
      }
      else if (type & E_EXTERNAL) { 
         c[2] = 'X';
      }
      else if (type & E_FAKE_EXTERNAL) { 
         c[2] = 'F';
      }
      else 
         ASSERT(0);
      
      if (type & E_WRITEBACK) { 
         c[3] = 'W';
      }
      else if (type & E_FLUSH_CLEAN) { 
         c[3] = 'C';
      }
      else if (type & E_DOWNGRADE) { 
         c[3] = 'D';
      }
      else 
         c[3]='X';
      c[4] = '\0';
      ITRACE(c); 
   } 
  
   /* for now we only measure transitions in the L2 cache  */
   
   if (type & E_L1) return;
   ASSERT (type & E_L2);
   
   pAddr &= ~(SCACHE_LINE_SIZE-1);
   ASSERT(MEM_LINE(machine, pAddr) < MEM_LINES(machine));
   
   pos = memStat.bitsPerState *( memStat.trNumCPUs[machine] *
                                 MEM_LINE(machine, pAddr) + mcpu);
   off = pos % 32;
   index = pos / 32;
   ASSERT(index*4<memStat.lastAddr[machine]); /* range check */

   /* get current state */
   state = ( memStat.trPtr[machine][index] >> off ) & memStat.cacheMask;

   /*
    * REPLACEMENT:
    * The line was in the cache and should be in the cache and is replaced. 
    */
   if ((!(type & E_EXTERNAL)) && (!(type & E_FAKE_EXTERNAL))) {  
      /* evict your own line */
      if (state != CACHE_INCACHE) {
         /* you better be in the cache! */
         if (NO_RACE_CONDITION) {
            CPUError("MS %#x %10lld MemStat: EVICT, cpu=%i state=%i type=%i\n",
                     pAddr,CPUVec.CycleCount(cpuNum),cpuNum,state,type);
         }
         ++(memStat.trWarnings[machine]);
      }

      if (isUserMode) {
         type |= E_USER_REPLACED;
      } else {
         type |= E_KERN_REPLACED;
      }

      ASSERT(type & (E_KERN_REPLACED|E_USER_REPLACED));
      if (type & E_USER_REPLACED) {
         uint newState;
         if (memStat.pidTable[machine][MS_pid(cpuNum)] == -1) { 
            memStat.pidTable[machine][MS_pid(cpuNum)]
                = memStat.pidCount[machine]++;
            ASSERT(memStat.pidCount[machine] < MS_MAXPID);
         } 
         if( memStat.trackPID ) { 
            newState = (memStat.pidTable[machine]
                        [MS_pid(cpuNum)]<<PIDOFFSET)|CACHE_REPLACED_U;
         } else { 
            newState = CACHE_REPLACED_U;
         }
         if (interest(pAddr)) {
            LogEntry("newstate",cpuNum,"pAddr=%#x newstate=%i\n",
                      pAddr,newState);
         }
         memStat.trPtr[machine][index] &= ~(memStat.cacheMask<<off);
         memStat.trPtr[machine][index] |=(newState<<off); 
      } else {
         if (interest(pAddr)) {
             LogEntry("newstate",cpuNum,"pAddr=%#x newstate=%i\n",
                      pAddr,CACHE_REPLACED_K);
         }
         memStat.trPtr[machine][index] &= ~(memStat.cacheMask<<off);
         memStat.trPtr[machine][index] |= CACHE_REPLACED_K<<off;
      }
      return ;
   }

   if (type & E_DOWNGRADE) {
      return;    /* no transition on a writeback */
   }

   if (type & E_EXTERNAL) {
      ASSERT(type & (E_WRITEBACK | E_FLUSH_CLEAN));
      /* you better be in the cache! */
      if (state != CACHE_INCACHE) {
         if (NO_RACE_CONDITION) {
            CPUError("MemStat %#x %10lld EXTERNAL cpuNum=%i state=%i type=%i\n",
                     pAddr,(uint64)CPUVec.CycleCount(cpuNum),cpuNum,state,type);
         }
         ++(memStat.trWarnings[machine]);
      }
      state = CACHE_INVALIDATED;
   }
   if (type & E_FAKE_EXTERNAL) {
      switch(state) {
      case CACHE_REPLACED_K:
         state = CACHE_REPL_INV;
         break;
      case CACHE_INCACHE:
         if (NO_RACE_CONDITION) {
            CPUError("MemStat %#x %10lld FAKE_EXTERNAL, cpuNum=%i\n",
                     pAddr,CPUVec.CycleCount(cpuNum), cpuNum);
         }
         ++(memStat.trWarnings[machine]);
         break;
      case CACHE_COLD:
         break;
      default:
         if ((state&0xf)==CACHE_REPLACED_U)
            state=CACHE_REPL_INV;
           
      }
   }
   if (interest(pAddr)) {
      LogEntry("newstate",cpuNum,"pAddr=%#x newstate=%i\n",
               pAddr,state);
   }

   /* record new state */
   memStat.trPtr[machine][index] &= ~(memStat.cacheMask<<off);
   memStat.trPtr[machine][index] |= (state <<off);
}



void MemStatPrefetchNoMiss (int cpuNum, PA pAddr, int way) {
   if (interest(pAddr)) {
      LogEntry("prefetch",cpuNum,"pAddr=0x%x, way=%i \n",
               pAddr,way);
   }
#ifndef USE_FLASHLITE
   if (memStat.init) {
      int categ = CacheTrMiss(cpuNum,pAddr);
      FalseSharingDefineOffset(cpuNum,pAddr,way,categ);
   }
#endif
} 

/*****************************************************************
 * CacheTrMiss
 * 
 * This is called when a line goes back into the cache
 * to maintain the correct state machine.
 *****************************************************************/

static int
CacheTrMiss(int cpuNum, PA pAddr)
{
   int state,i;
   unsigned int pos,index,off;
   int retval = -1;
   int machine;
   int mcpu;

   if (!memStat.init) return 0;
#ifdef USE_FLASHLITE
   return 0;
#endif

   machine = M_FROM_CPU(cpuNum);
   mcpu = MCPU_FROM_CPU(cpuNum);

#ifdef DEBUGGING
   if (interest(pAddr)) { 
      CPUWarning("%10lld trMiss\n",CPUVec.CycleCount(cpuNum));
   }
#endif

   pAddr &=~(SCACHE_LINE_SIZE-1);
   ASSERT( memStat.trPtr[machine] );
   if (MEM_LINE(machine, pAddr) >= MEM_LINES(machine)) {
      CPUError("MemStat::CacheTrMiss. pAddr=0x%llx, MEM_SIZE=0x%llx\n",
               (uint64)pAddr, (uint64)MEM_SIZE(machine));
   }
   pos = memStat.bitsPerState * (memStat.trNumCPUs[machine] *
                                 MEM_LINE(machine, pAddr) + mcpu);
   off = pos% 32;
   index = pos / 32;
   ASSERT( index*4< memStat.lastAddr[machine] );
   /* get old state */
   state = (memStat.trPtr[machine][index] >> off ) & memStat.cacheMask;
   switch( state ) {
   case CACHE_COLD:  
      retval= CATEG_ice;
      break;
   case CACHE_REPLACED_K:
      if (memStat.trackPID) {
         retval = CATEG_Kcap;
      } else { 
         retval = CATEG_cap;
      }
      break;
   case CACHE_INVALIDATED:
      retval = CATEG_inv;
      break;
   case CACHE_REPL_INV:
      retval = CATEG_capinv;
      break;
   case CACHE_WARMED: 
      retval = CATEG_cold;
      break;
   case CACHE_INCACHE:
      if( NO_RACE_CONDITION ) {
         CPUError("%10lld MemStat CacheTrMiss, state=%i pAddr=%#x\n",
                  CPUVec.CycleCount(cpuNum), state, pAddr);
      }
      ++(memStat.trWarnings[machine]);
      break;
   default: 
      ASSERT( (state&0xf)== CACHE_REPLACED_U );
      if( memStat.trackPID) {
         if (memStat.pidTable[machine][MS_pid(cpuNum)] == -1) { 
            memStat.pidTable[machine][MS_pid(cpuNum)]
                = memStat.pidCount[machine]++;
            ASSERT(memStat.pidCount[machine] < MS_MAXPID);
         }   
         if (state>>PIDOFFSET == memStat.pidTable[machine][MS_pid(cpuNum)])
            retval = CATEG_SelfCap;
         else
            retval = CATEG_OtherCap;  
      } else {
         retval = CATEG_cap;
      }
   }
   memStat.trPtr[machine][index] &= ~(memStat.cacheMask <<off);
   memStat.trPtr[machine][index] |= CACHE_INCACHE<<off; /* it is back in the cache */
   
   if (interest(pAddr)) {
      LogEntry("newstate",cpuNum,"pAddr=%#x newstate=IN_CACHE\n",
               pAddr);
   }
   if (state == CACHE_COLD) {
      pos = memStat.bitsPerState * (memStat.trNumCPUs[machine] *
                                    MEM_LINE(machine, pAddr));
      off = pos % 32;
      index = pos / 32;
      for(i=0;i<memStat.trNumCPUs[machine];++i) {
         if( i != mcpu ) {
            memStat.trPtr[machine][index] &= ~(memStat.cacheMask <<off);
            memStat.trPtr[machine][index] |= (CACHE_WARMED<<off);
         }
         off += memStat.bitsPerState;
         if( off >= 32 ) {
            off -=32;
            index ++;
         }  
      }
   }
   if (interest(pAddr)) {
      ITRACE("Cache");
   }
   return retval;
}


int  
MemStatInitialized(void)
{
   return( memStat.init );
}

/*****************************************************************************
 * Support routines for MemStat
 ****************************************************************************/
void 
MemStatLog(FILE *f,char *tag, SimTime time, char *desc, char *text,...)
{
   va_list ap;
   if (f) { 
      if (time) {
         fprintf(f,"%-6s ",tag);
         va_start(ap,text);
         vfprintf(f,text,ap);
         va_end(ap);
         fprintf(f,"\n");
      } else {
         fprintf(f,"DEF %-6s %s\n",tag,desc);
      }
   } 
}


/*****************************************************************
 * Very useful debugging procedures
 *****************************************************************/
int
interest( PA pAddr ) {
#ifdef DEBUGGING
/*#define INTERESTING_LINE  0x67460d80*/
#define INTERESTING_LINE  0x768040
#define SET_PROBLEM 1

   if( (pAddr  & ~(SCACHE_LINE_SIZE-1))== 
       ( INTERESTING_LINE  &~(SCACHE_LINE_SIZE-1))  ) {
      CPU_nop();
      return 1;
   } 
   if( SET_PROBLEM ) { 
      if( ((pAddr >> log2SCACHE_LINE_SIZE) & (SCACHE_INDEX-1))
          == ((INTERESTING_LINE>>log2SCACHE_LINE_SIZE) & (SCACHE_INDEX-1)) ) {
         CPU_nop();
         return 1;
      }
   }
#endif
   return 0;
}

 int
MSCacheState(int cpu, PA pAddr) 
{
   unsigned int pos, off, index;
   int machine;
   int mcpu;

   if (!memStat.init) return -1;
#ifdef USE_FLASHLITE
   return -1;
#endif

   machine = M_FROM_CPU(cpu);
   mcpu = MCPU_FROM_CPU(cpu);

   pos = memStat.bitsPerState *( memStat.trNumCPUs[machine] *
                                 MEM_LINE(machine, pAddr) + mcpu);
   off = pos % 32;
   index = pos / 32;
   ASSERT(index*4<memStat.lastAddr[machine]); /* range check */

   /* get current state */
   return ( memStat.trPtr[machine][index] >> off ) & memStat.cacheMask;
}
