/*
 * Copyright (C) 1996-1998 by the Board of Trustees
 *    of Leland Stanford Junior University.
 * 
 * This file is part of the SimOS distribution. 
 * See LICENSE file for terms of the license. 
 *
 */

/*****************************************************************
 * annotations.c
 *
 *
 * Author: $Author: stets $
 * Date:   $Date: 2000/03/21 06:07:22 $
 ****************************************************************/


#include <stdio.h>
#include <unistd.h>
#include <sys/types.h>
#include <string.h>
#include <stdlib.h>
#include <fcntl.h>
#include <math.h>
#include <setjmp.h>
#include <ctype.h>
#include <assert.h>

#include "sim_error.h"
#include "simutil.h"
#include "memstat.h"
#include "annotations.h"
#include "cpu_interface.h"
#include "symbols.h"
#include "tcl_init.h"
#include "hw_events.h"
#include "visual.h"

RemapVector *remapVec = 0;  /* does not belong there, but cannot be 
                             * defined in the simos directory (-G0 )
                             * has to be defined somewhere
                             */
struct CPUVector CPUVec;    /* has to be defined somewhere */

/*****************************************************************
 *   Annotation Private Declarations
 ****************************************************************/

#if SIM_MAXCPUS > 64
#define BV_SIZE ((SIM_MAXCPUS + 63) / 64)
#define BV_INDEX(_cpu) (_cpu / 64)
#define BV_SHIFT(_cpu) (_cpu % 64)

typedef struct {
   uint64 bv[BV_SIZE];
} CPUBitVec;

#define ENABLE_ALL(_vec) {                               \
   int _i;                                               \
   for (_i = 0; _i < BV_SIZE; _i++) {                    \
      _vec.bv[_i] = ~((uint64) 0);                       \
   }                                                     \
}
#define DISABLE_ALL(_vec) {                              \
   int _i;                                               \
   for (_i = 0; _i < BV_SIZE; _i++) {                    \
      _vec.bv[_i] = (uint64) 0;                          \
   }                                                     \
}
#define ENABLE_CPU(_vec, _cpu)     {(_vec).bv[BV_INDEX(_cpu)] |= (1 << (BV_SHIFT(_cpu)));} 
#define DISABLE_CPU(_vec, _cpu)    {(_vec).bv[BV_INDEX(_cpu)] &= ~(1 << (BV_SHIFT(_cpu)));} 
#define IS_CPU_ENABLED(_vec, _cpu) (((_vec).bv[BV_INDEX(_cpu)]) & (1 << (BV_SHIFT(_cpu))))
#else
typedef uint64 CPUBitVec;
#define ENABLE_ALL(_vec)           {_vec = ~((CPUBitVec) 0);}
#define ENABLE_CPU(_vec, _cpu)     {_vec |= (((CPUBitVec) 1) << (_cpu));} 
#define DISABLE_ALL(_vec)          {_vec = (CPUBitVec) 0;}
#define DISABLE_CPU(_vec, _cpu)    {_vec &= ~(((CPUBitVec) 1) << (_cpu));} 
#define IS_CPU_ENABLED(_vec, _cpu) ((_vec) & (((CPUBitVec) 1) << (_cpu)))
#endif

typedef struct AnnType AnnType;
typedef struct AnnCmd AnnCmd;
typedef struct AnnRec AnnRec;
typedef struct AnnTag AnnTag;

typedef enum {
   ANN_ARGTYPE_NUM = 0,
   ANN_ARGTYPE_SYMBOL,
   ANN_ARGTYPE_ENUM,
   ANN_ARGTYPE_NONE,
   ANN_NUM_ARGTYPES      /* must be last */
} AnnArgType;

struct AnnType {
   char       *name;
   AnnArgType argType;
   AnnRec     *recs;
   char       **enumList;
   int        enumCount;
};

struct AnnRec {
   AnnType    *type;
   uint64     arg;
   AnnCmd     *cmds;
   AnnRec     *next;
   AnnRec     *FMHashLink;
   int        FMHashType;
};

struct AnnCmd {
   Tcl_Interp *interp;
#ifdef TCL8
   Tcl_Obj    *compiledscript;   
#else
   char       *script;
#endif
   int        context;
   AnnRec     *rec;
   AnnCmd     *nextByType;
   AnnTag     *tag;
   AnnCmd     *nextByTag;
};

struct AnnTag {
   CPUBitVec  enabled;
   char       *name;
   AnnCmd     *cmds;
};
   
static Tcl_HashTable types;
static Tcl_HashTable tags;
static int staticAnnsInstalled = 0;
static bool prePCAnnsExist = FALSE;

static void SetAnn(Tcl_Interp *interp, char *script,
                   AnnType *type, uint64 arg, AnnTag *tag);
static int TranslateArg(Tcl_Interp *interp, AnnType *type,
                        char *argStr, uint64 *retval);

static void AnnFMInfo(Tcl_Interp *interp);
static void AnnAddDynamic(AnnPtr ptr, int newrec);

static void FlushAnnCache(void);

/*****************************************************************
 *   Annotation TCL Declarations
 ****************************************************************/

static int cmdSet(Tcl_Interp *interp, int argc, char *argv[]);
static int cmdDisable(Tcl_Interp *interp, int argc, char *argv[]);
static int cmdEnable(Tcl_Interp *interp, int argc, char *argv[]);
static int cmdType(Tcl_Interp *interp, int argc, char *argv[]);
static int cmdExec(Tcl_Interp *interp, int argc, char *argv[]);
static int cmdInfo(Tcl_Interp *interp, int argc, char *argv[]);
static int cmdTrace(Tcl_Interp *interp, int argc, char *argv[]);

static tclcmd annCmds[] = {
{   "set",         -1, cmdSet,         " set type ?arg? ?-tag tagname? script"},
{   "disable",     -1, cmdDisable,     " disable tag ?cpu ...?"},
{   "enable",      -1, cmdEnable,      " enable tag ?cpu ...?"},
{   "type",        -1, cmdType,        " type typeName argtype ?enumlist?"},
{   "exec",        -1, cmdExec,        " exec type ?arg?"},
{   "info",        -1, cmdInfo,        " info ?list|tags|types|fm?"},
{   "trace",       -1, cmdTrace,       " trace on|off"},
{   NULL,           0, NULL,           NULL}
};

static char *argTypeNames[] = { "num", "symbol", "enum", "none" };

/* Boolean indicating whether any scache annotations are defined --
 * moved up here since AnnAddDynamic needs is */
uint scacheAnnotation = 0;
void ScacheAnnLateInit(Tcl_Interp *interp);

/*****************************************************************
 *   Annotation tracing support
 ****************************************************************/

static struct annTraceStruct {
   int        enabled;
#ifdef TCL8
   Tcl_Obj    *script;
#else
   char       *script;
#endif
} annTrace;

/*****************************************************************
 *   Annotation Initialization
 ****************************************************************/

void AnnotationInit(Tcl_Interp *interp)
{
   ASSERT((sizeof(argTypeNames)/sizeof(char*)) == ANN_NUM_ARGTYPES);
   
   Tcl_InitHashTable(&types, TCL_STRING_KEYS);
   Tcl_InitHashTable(&tags, TCL_STRING_KEYS);
   
   Tcl_CreateCommand(interp, "annotation", DispatchCmd, (ClientData)annCmds, NULL);
}


/*****************************************************************
 *   Annotation - TCL Interface
 ****************************************************************/

int cmdSet(Tcl_Interp *interp, int argc, char *argv[])
{
   Tcl_HashEntry *entry;
   AnnTag *tag;
   AnnType *type;
   uint64 arg = 0;
   int i, new;

   if ((argc < 4) || (argc > 7)) {
      Tcl_AppendResult(interp, "wrong # args: should be \"",
                       argv[0], annCmds[0].usage, "\"", NULL);
      return TCL_ERROR;
   }

   entry = Tcl_FindHashEntry(&types, argv[2]);   
   if (!entry) {
      Tcl_AppendResult(interp, "no type named \"", argv[2], "\"", NULL);
      return TCL_ERROR;
   }

   type = (AnnType *) Tcl_GetHashValue(entry);
   ASSERT(type);

   if (argc > 5) {
      /* we must have a tag */
      if (strcmp(argv[argc-3], "-tag")) {
         Tcl_AppendResult(interp, "wrong # args: should be \"",
                          argv[0], annCmds[0].usage, "\"", NULL);
         return TCL_ERROR;
      }

      entry = Tcl_CreateHashEntry(&tags, argv[argc-2], &new);
   
      if (new) {
         tag = (AnnTag*)ZMALLOC(sizeof(AnnTag),"AnnTag");
         ENABLE_ALL(tag->enabled);
         tag->name = Tcl_GetHashKey(&tags, entry);
         tag->cmds = NULL;
         Tcl_SetHashValue(entry, tag);

      } else {
         tag =  (AnnTag*) Tcl_GetHashValue(entry);
         ASSERT(tag);
      }

   } else {
      tag = NULL;
   }
   
   if (type->argType == ANN_ARGTYPE_NONE) {
      if ((argc != 4) && (argc != 6)) {
         Tcl_AppendResult(interp, "wrong # args: should be \"",
                          argv[0], annCmds[0].usage, "\"", NULL);
         return TCL_ERROR;
      }
      
   } else {
      if (((argc == 4) || (argc == 6)) && (type->argType == ANN_ARGTYPE_ENUM)) {
         char *str = SaveString(argv[argc-1]);
         for (i=0; i<type->enumCount; i++) {
            if (type->enumList[i][0]) {
               SetAnn(interp, str, type, i, tag);
            }
         }
         return TCL_OK;

      }
      
      if ((argc != 5) && (argc != 7)) {
         Tcl_AppendResult(interp, "wrong # args: should be \"",
                          argv[0], annCmds[0].usage, "\"", NULL);
         return TCL_ERROR;
      }

      if (TranslateArg(interp, type, argv[3], &arg) != TCL_OK) {
         return TCL_ERROR;
      }
   }
   SetAnn(interp, SaveString(argv[argc-1]), type, arg, tag);
   return TCL_OK;
}

void SetAnn(Tcl_Interp *interp, char *script, AnnType *type, uint64 arg, AnnTag *tag)
{
   AnnRec *rec;
   AnnCmd *cmd, **last;
   int newrec = 0;

   cmd = (AnnCmd*)ZMALLOC(sizeof(AnnCmd),"AnnCmd");

   cmd->interp = interp;
#ifdef TCL8
   cmd->compiledscript = Tcl_NewStringObj(script, -1);
   /* script may now need to be deallocated because newstringobj allocates
    * an internal buffer
    */
#else
   cmd->script = script;
#endif
   cmd->context = VisualGetContext();
   
   for (rec = type->recs; rec && (rec->arg != arg); rec = rec->next) ;
   
   if (!rec) {
      rec = (AnnRec*)ZMALLOC(sizeof(AnnRec),"AnnRec");
      rec->type = type;
      rec->arg = arg;
      rec->cmds = NULL;
      rec->next = type->recs;
      type->recs = rec;
      newrec = 1;
   }

   cmd->rec = rec;
   cmd->nextByType = NULL;
   for (last = &(rec->cmds); *last; last = &((*last)->nextByType)) ;
   *last = cmd;

   if (tag) {
      cmd->tag = tag;
      cmd->nextByTag = tag->cmds;
      tag->cmds = cmd;
   } else {
      cmd->tag = NULL;
      cmd->nextByTag = NULL;
   }

   if (staticAnnsInstalled) {
      AnnAddDynamic((AnnPtr)rec, newrec);
   }
}

int cmdDisable(Tcl_Interp *interp, int argc, char *argv[])
{
   Tcl_HashEntry *entry;
   AnnTag *tag;
   char *tagName;
   int i, cpunum = 0;
   
   if (argc < 3) {
      Tcl_AppendResult(interp, "wrong # args: should be \"",
                       argv[0], annCmds[1].usage, "\"", NULL);
      return TCL_ERROR;
   }

   tagName = argv[2];
   
   entry = Tcl_FindHashEntry(&tags, tagName);
   if (!entry) {
      return TCL_OK;
   }

   tag =  (AnnTag*) Tcl_GetHashValue(entry);
   ASSERT(tag);

   if (argc == 3) {
      DISABLE_ALL(tag->enabled);
   } else {
      for (i=3; i<argc; i++) {
         if (Tcl_GetInt(interp, argv[i], &cpunum) != TCL_OK) {
            Tcl_AppendResult(interp, "bad cpunum \"", argv[i], "\"", NULL);
            return TCL_ERROR;
         } else {
            DISABLE_CPU(tag->enabled, cpunum);
         }
      }
   }

   return TCL_OK;
}

static int cmdEnable(Tcl_Interp *interp, int argc, char *argv[])
{
   Tcl_HashEntry *entry;
   AnnTag *tag;
   char *tagName;
   int i, cpunum = 0;
   
   if (argc < 3) {
      Tcl_AppendResult(interp, "wrong # args: should be \"",
                       argv[0], annCmds[2].usage, "\"", NULL);
      return TCL_ERROR;
   }

   tagName = argv[2];
   
   entry = Tcl_FindHashEntry(&tags, tagName);
   if (!entry) {
      return TCL_OK;
   }

   tag = (AnnTag*) Tcl_GetHashValue(entry);
   ASSERT(tag);

   if (argc == 3) {
      ENABLE_ALL(tag->enabled);
   } else {
      for (i=3; i<argc; i++) {
         if (Tcl_GetInt(interp, argv[i], &cpunum) != TCL_OK) {
            Tcl_AppendResult(interp, "bad cpunum \"", argv[i], "\"", NULL);
            return TCL_ERROR;
         } else {
            ENABLE_CPU(tag->enabled, cpunum);
         }
      }
   }

   return TCL_OK;
}

int cmdType(Tcl_Interp *interp, int argc, char *argv[])
{
   AnnType *type;
   AnnArgType argType;
   char *typeName;
   char *argTypeStr;
   char **enumList = NULL;
   int enumCount = 0;
   int new = 0;
   Tcl_HashEntry *entry;
   
   if ((argc != 4) && (argc != 5)) {
      Tcl_AppendResult(interp, "wrong # args: should be \"",
                       argv[0], annCmds[3].usage, "\"", NULL);
      return TCL_ERROR;
   }

   typeName = argv[2];
   argTypeStr = argv[3];

   for (argType=ANN_ARGTYPE_NUM; argType<ANN_NUM_ARGTYPES; argType++) {
      if ((argTypeNames[argType][0] == argTypeStr[0])
          && (strcmp(argTypeNames[argType], argTypeStr) == 0)) {
         break;
      }
   }
   
   if (argType == ANN_NUM_ARGTYPES) {
      Tcl_AppendResult(interp, "no such argtype \"", argTypeStr, "\"", NULL);
      return TCL_ERROR;
   }

   if (argType == ANN_ARGTYPE_ENUM) {
      if (argc != 5) {
         Tcl_AppendResult(interp, "wrong # args: should be \"",
                          argv[0], annCmds[3].usage, "\"", NULL);
         return TCL_ERROR;
      }

      Tcl_SplitList(interp, argv[4], &enumCount, &enumList);
   }

   entry = Tcl_CreateHashEntry(&types, typeName, &new);
   
   if (!new) {
      Tcl_AppendResult(interp, "type name already taken \"",
                       typeName, "\"", NULL);
      return TCL_ERROR;      
   }
   
   type = (AnnType*)ZMALLOC(sizeof(AnnType),"AnnType");

   type->name = Tcl_GetHashKey(&types, entry);
   type->argType = argType;
   type->recs = NULL;
   type->enumCount = enumCount;
   type->enumList = enumList;
   Tcl_SetHashValue(entry, type);

   return TCL_OK;
}

int cmdExec(Tcl_Interp *interp, int argc, char *argv[])
{
   Tcl_HashEntry *entry;
   AnnType *type;
   AnnRec *rec;
   char *typeName = 0;
   char *argStr = 0;
   uint64 arg = 0;

   if ((argc != 3) && (argc != 4)) {
      Tcl_AppendResult(interp, "wrong # args: should be \"",
                       argv[0], annCmds[4].usage, "\"", NULL);
      return TCL_ERROR;
   }
   
   typeName = argv[2];

   entry = Tcl_FindHashEntry(&types, typeName);   
   if (!entry) {
      Tcl_AppendResult(interp, "no type named \"", typeName, "\"", NULL);
      return TCL_ERROR;
   }

   type =  (AnnType *) Tcl_GetHashValue(entry);
   ASSERT(type);

   if (type->argType == ANN_ARGTYPE_NONE) {
      if (argc != 3) {
         Tcl_AppendResult(interp, "wrong # args: should be \"",
                          argv[0], annCmds[4].usage, "\"", NULL);
         return TCL_ERROR;
      }
      
   } else {
      if (argc != 4) {
         Tcl_AppendResult(interp, "wrong # args: should be \"",
                          argv[0], annCmds[4].usage, "\"", NULL);
         return TCL_ERROR;
      }

      argStr = argv[3];

      if (TranslateArg(interp, type, argStr, &arg) != TCL_OK) {
         return TCL_ERROR;
      } 
   }

   for (rec = type->recs; rec; rec = rec->next) {
      if (rec->arg == arg) {
         AnnExec(rec);
      }
   }
   
   return TCL_OK;
}

int cmdInfo(Tcl_Interp *interp, int argc, char *argv[])
{
   AnnTag *tag;
   AnnRec *rec;
   AnnCmd *cmd;
   AnnType *type;
   Tcl_HashSearch search;
   Tcl_HashEntry *entry;

   char buf[64];
   int numst, numcmds;
   int num;
   int found = 0;
   
   if ((argc != 2) && (argc != 3)) {
      Tcl_AppendResult(interp, "wrong # args: should be \"",
                       argv[0], annCmds[5].usage, "\"", NULL);
      return TCL_ERROR;
   }
   
   if ((argc == 2) || (!strcmp(argv[2], "list"))) {
      Tcl_AppendResult(interp, "Annotation List:\n", NULL);
      
      entry = Tcl_FirstHashEntry(&types, &search);
      
      while (entry) {
         type =  (AnnType *) Tcl_GetHashValue(entry);
         ASSERT(type);

         for (rec = type->recs; rec; rec = rec->next) {
            for (cmd = rec->cmds; cmd; cmd = cmd->nextByType) {
               Tcl_AppendResult(interp, "annotation set ", type->name, " ", NULL);
               
               switch (type->argType) {
               case ANN_ARGTYPE_NUM:
               case ANN_ARGTYPE_SYMBOL:
		  PrintLLX(buf,rec->arg);
		  strcat(buf," ");
                  Tcl_AppendResult(interp, buf, NULL);
                  break;
               case ANN_ARGTYPE_ENUM:
                  if ((rec->arg < type->enumCount) && (type->enumList[rec->arg][0])) {
                     Tcl_AppendResult(interp, type->enumList[rec->arg], " ", NULL);
                  } else {
		     PrintLLD(buf,rec->arg);
		     strcat(buf," ");
                     Tcl_AppendResult(interp, buf, NULL);
                  }
                  break;
               case ANN_ARGTYPE_NONE:
                  break;
               default:
                  ASSERT(0);
                  break;
               }

               if (cmd->tag) {
                  Tcl_AppendResult(interp, "-tag ", cmd->tag->name, " ", NULL);
               }
#ifdef TCL8
               Tcl_AppendResult(interp, "{", 
                                Tcl_GetStringFromObj(cmd->compiledscript, NULL),
                                "}\n\n", NULL);
#else               
               Tcl_AppendResult(interp, "{", (char*)cmd->script, "}\n\n", NULL);
#endif
            }
         }

         entry = Tcl_NextHashEntry(&search);
      }
      
      found = 1;
   }

   if ((argc == 2) || (!strcmp(argv[2], "tags"))) {
      Tcl_AppendResult(interp, "\nAnnotation Tags:\n", NULL);
      
      entry = Tcl_FirstHashEntry(&tags, &search);
      
      while (entry) {
         tag =  (AnnTag*) Tcl_GetHashValue(entry);
         ASSERT(tag);
         Tcl_AppendResult(interp, "\"", tag->name, "\" ", NULL);
         for (num=0, cmd=tag->cmds; cmd; num++, cmd=cmd->nextByTag);
         sprintf(buf, "enabled mask %#llx num %d", tag->enabled, num);
         Tcl_AppendResult(interp, buf, "\n", NULL);
         entry = Tcl_NextHashEntry(&search);
      }
      
      Tcl_AppendResult(interp, "\n", NULL);
      found = 1;
   }

   if ((argc == 2) || (!strcmp(argv[2], "types"))) {
      Tcl_AppendResult(interp, "Annotation Types:\n", NULL);
      
      entry = Tcl_FirstHashEntry(&types, &search);
      
      while (entry) {
         type = (AnnType *) Tcl_GetHashValue(entry);
         ASSERT(type);
         
         Tcl_AppendResult(interp, "\"", type->name, "\" ",
                          argTypeNames[type->argType], NULL);
         
         if (type->argType == ANN_ARGTYPE_ENUM) {
            Tcl_AppendResult(interp, " {",
                             Tcl_Merge(type->enumCount, type->enumList),
                             "}", NULL);
         }
         
         numst=0;
         numcmds=0;
         for (rec=type->recs; rec; numst++, rec=rec->next) {
         for (cmd=rec->cmds; cmd; numcmds++, cmd=cmd->nextByType);
         }
         
         sprintf(buf, "subtypes %d commands %d", numst, numcmds);
         Tcl_AppendResult(interp, " ", buf, "\n", NULL);
         entry = Tcl_NextHashEntry(&search);
      }
      
      Tcl_AppendResult(interp, "\n", NULL);
      found = 1;
   }

   if ((argc == 2) || (!strcmp(argv[2], "fm"))) {
      AnnFMInfo(interp);
      found = 1;
   }

   if (!found) {
      Tcl_AppendResult(interp, "arg not recognized \"", argv[2], "\"", NULL);
      return TCL_ERROR;
   }
   
   return TCL_OK;
}

int cmdTrace(Tcl_Interp *interp, int argc, char *argv[])
{
   char *action = argv[2];

   if (argc != 3) {
      Tcl_AppendResult(interp, "wrong # args: should be \"",
                       argv[0], annCmds[6].usage, "\"", NULL);
      return TCL_ERROR;
   }

   if (!strcmp(action, "off")) {
      annTrace.enabled = 0;
   } else if (!strcmp(action, "on")) {
      annTrace.enabled = 1;
   } else {
      annTrace.enabled = 2;
#ifdef TCL8
      annTrace.script = Tcl_NewStringObj(SaveString(action),-1); 
#else
      annTrace.script = SaveString(action);
#endif
   }
   return TCL_OK;
}


/*****************************************************************
 *   Annotation - SimOS Interface
 ****************************************************************/

AnnPtr AnnFind(char *typeName, char *argStr)
{
   Tcl_HashEntry *entry;
   AnnType *type;
   AnnRec *rec;
   uint64 arg;
   int code;

   entry = Tcl_FindHashEntry(&types, typeName);   
   ASSERT(entry);

   type =  (AnnType*) Tcl_GetHashValue(entry);
   ASSERT(type);
   
   code = TranslateArg(NULL, type, argStr, &arg);
   ASSERT(code == TCL_OK);

   for (rec = type->recs; rec; rec = rec->next) {
      if (rec->arg == arg) {
         return rec;
      }
   }

   return NULL;
}

AnnPtr AnnFindInt(char *typeName, uint64 arg)
{
   Tcl_HashEntry *entry;
   AnnType *type;
   AnnRec *rec;

   entry = Tcl_FindHashEntry(&types, typeName);   
   ASSERT(entry);

   type =  (AnnType*) Tcl_GetHashValue(entry);
   ASSERT(type);
   
   for (rec = type->recs; rec; rec = rec->next) {
      if (rec->arg == arg) {
         return rec;
      }
   }

   return NULL;
}

uint64 AnnFirst(char *typeName, AnnPtr *iter)
{
   Tcl_HashEntry *entry;
   AnnType *type;

   entry = Tcl_FindHashEntry(&types, typeName);   
   ASSERT(entry);

   type =  (AnnType *) Tcl_GetHashValue(entry);
   ASSERT(type);
   
   *iter = type->recs;
   if (*iter) {
      return (*iter)->arg;
   }
   
   *iter = NULL;
   return 0;
}

uint64 AnnNext(AnnPtr *iter)
{
   if (!*iter) return 0;
   *iter = (*iter)->next;

   if (*iter) {
      return (*iter)->arg;
   }

   return 0;
}

int AnnNumArg(char *typeName)
{
   Tcl_HashEntry *entry;
   AnnType *type;

   entry = Tcl_FindHashEntry(&types, typeName);
   ASSERT(entry);
   
   type =  (AnnType *) Tcl_GetHashValue(entry);
   ASSERT(type);
      
   switch (type->argType) {
   case ANN_ARGTYPE_NUM:
   case ANN_ARGTYPE_SYMBOL:
      return -1;
   case ANN_ARGTYPE_ENUM:
      return type->enumCount;
   case ANN_ARGTYPE_NONE:
      return 1;
   default:
      ASSERT(0);
      break;
   }
   
   return 0;
}

static void traceAnnotation(AnnCmd* cmd)
{
   AnnPtr   rec  = cmd->rec;
   uint64   arg  = rec->arg;
   AnnType* type = rec->type;
   int      cpu  = CPUVec.CurrentCpuNum();
   char     cpuName[32];
   char     *pid, *proc;
   uint64   cyclecount;

   
   if (CPUVec.CycleCount) {
      cyclecount = CPUVec.CycleCount(cpu);
   } else {
      cyclecount = 0;
   }

   sprintf(cpuName,"%d",cpu);    
   pid = Tcl_GetVar2(TCLInterp,"PID",cpuName,0);
   proc = Tcl_GetVar2(TCLInterp,"PROCESS",cpuName,0);
   if (!pid)  { pid = "-1";}
   if (!proc) { proc = "NONAME";}

   CPUPrint("%lld ANN cpu %d %s-%s %s ",
            cyclecount, cpu, pid, proc, type->name);

   switch (type->argType) {
   case ANN_ARGTYPE_NUM:
   case ANN_ARGTYPE_SYMBOL:
      CPUPrint("0x%llx ", arg);
      break;
   case ANN_ARGTYPE_ENUM:
      if ((arg < type->enumCount) && (type->enumList[arg][0])) {
         CPUPrint("%s ", type->enumList[arg]);
      } else {
         CPUPrint("%lld ", arg);
      }
      break;
   case ANN_ARGTYPE_NONE:
      break;
   default:
      ASSERT(0);
      break;
   }


#ifdef TCL8
   CPUPrint("{%s}\n", Tcl_GetStringFromObj(cmd->compiledscript, NULL));
   if (annTrace.enabled == 2) {
      extern int Tcl_GlobalEvalObj(Tcl_Interp*, Tcl_Obj*);
      Tcl_GlobalEvalObj(TCLInterp, annTrace.script);
   }
#else               
   CPUPrint("{%s}\n", (char*) annTrace.script);
   if (annTrace.enabled == 2) {
      Tcl_GlobalEval(TCLInterp, annTrace.script);
   }
#endif
}


void AnnExecCPU(AnnPtr ptr, int cpuNum)
{
   AnnCmd *cmd;
   int code;
   int oldContext;

   if (!ptr) {
      return;
   }

   for (cmd = ptr->cmds; cmd; cmd = cmd->nextByType) {
      if (!cmd->tag || (IS_CPU_ENABLED(cmd->tag->enabled, cpuNum))) {
         if (!VisualSetContext(cmd->context, &oldContext)) 
            continue;
         if (annTrace.enabled) {
            traceAnnotation(cmd);
         }

#ifdef TCL8
         code = Tcl_GlobalEvalObj(cmd->interp, cmd->compiledscript);
#else
         code = Tcl_GlobalEval(cmd->interp, cmd->script);
#endif

         VisualSetContext(oldContext, NULL);
         if (code != TCL_OK) {

#ifdef TCL8
               HandleError(cmd->interp, 
                           Tcl_GetStringFromObj(cmd->compiledscript, NULL));
#else
               HandleError(cmd->interp, cmd->script);
#endif
         }
      }
   }
}

void
AnnExec(AnnPtr ann)
{
   AnnExecCPU(ann, CPUVec.CurrentCpuNum());
}

int 
TranslateArg(Tcl_Interp *interp, AnnType *type, char *argStr, uint64 *retval)
{
   int i, code;
   uint64 val = 0;

   
   switch (type->argType) {
   case ANN_ARGTYPE_NUM:
      break;
   case ANN_ARGTYPE_SYMBOL:
      code = SymInt(interp, argStr, &val);
      *retval = val;
      return code;
      /* break; */
   case ANN_ARGTYPE_ENUM:
      for (i=0; i<type->enumCount; i++) {
	if (type->enumList[i][0] && (strcmp(type->enumList[i], argStr) == 0)) {
	   *retval = i;
            return TCL_OK;
         }
      }
      break;
   case ANN_ARGTYPE_NONE:
      *retval = 0;
      return TCL_OK;
      /* break; */
   default:
      ASSERT(0);
      return TCL_ERROR;
      /* break; */
   }

#ifdef __alpha
   if (sscanf(argStr, "%li", retval) != 1) {
#else
   if (sscanf(argStr, "%lli", retval) != 1) {
#endif
      ASSERT(interp);
      Tcl_AppendResult(interp, "bad arg \"", argStr, "\"", NULL);
      switch (type->argType) {
      case ANN_ARGTYPE_NUM:
         Tcl_AppendResult(interp, ": expected a number", NULL);
         break;
      case ANN_ARGTYPE_SYMBOL:
         Tcl_AppendResult(interp, ": expected a symbol", NULL);
         break;
      case ANN_ARGTYPE_ENUM:
         Tcl_AppendResult(interp, ": expected one of ", NULL);
         for (i=0; i<type->enumCount; i++) {
            if (i > 0)
               Tcl_AppendResult(interp, ", ", NULL);
            Tcl_AppendResult(interp, type->enumList[i], NULL);
         }
         break;
      default:
	break;
      }
      return TCL_ERROR;
   }
   
   return TCL_OK;
}



/*****************************************************************
 *   Fast Memory (FM) Annotation Lookup
 *
 *       This module provides a mechanism for cpu simulators
 *       to quickly determine if there is a memory annotation
 *       on a range of addresses.
 *
 ****************************************************************/

#define NUM_HIST 11
#define HASH_TABLE_SIZE 2048
#define SECOND_HASH_SHIFT 16
#define HASH_FUNC(_va) ((((_va) >> log2RangeSize) \
                         ^ ((_va) >> SECOND_HASH_SHIFT)) \
                        % HASH_TABLE_SIZE)


typedef struct hashBucket {
   AnnRec *chain;
   int typesInChain;
} hashBucket;


static hashBucket hashTable[HASH_TABLE_SIZE];

static uint rangeSize = -1;
static uint log2RangeSize = 0;
static uint existingAnns = 0;   /* Keep track of what's been inserted */

void AnnFMInit(uint size)
{
   AnnPtr rec = NULL;
   VA addr;
   int bucket, i;

   if (size == rangeSize) {
      return;
   }

   rangeSize = size;
   log2RangeSize = GetLog2(rangeSize);
   
   for (i=0; i<HASH_TABLE_SIZE; i++) {
      hashTable[i].chain = NULL;
      hashTable[i].typesInChain = 0;
   }

   for (addr = AnnFirst("pc", &rec); rec; addr = AnnNext(&rec)) {
      existingAnns |= ANNFM_PC_TYPE;
      bucket = HASH_FUNC(addr);
      rec->FMHashType = ANNFM_PC_TYPE;
      rec->FMHashLink = hashTable[bucket].chain;
      hashTable[bucket].chain = rec;
      hashTable[bucket].typesInChain |= ANNFM_PC_TYPE;
   }

   for (addr = AnnFirst("pre-pc", &rec); rec; addr = AnnNext(&rec)) {
      existingAnns |= ANNFM_PRE_PC_TYPE;
      bucket = HASH_FUNC(addr);
      rec->FMHashType = ANNFM_PRE_PC_TYPE;
      rec->FMHashLink = hashTable[bucket].chain;
      hashTable[bucket].chain = rec;
      hashTable[bucket].typesInChain |= ANNFM_PRE_PC_TYPE;
      prePCAnnsExist = TRUE;
   }

   for (addr = AnnFirst("load", &rec); rec; addr = AnnNext(&rec)) {
      existingAnns |= ANNFM_LD_TYPE;
      bucket = HASH_FUNC(addr);
      rec->FMHashType = ANNFM_LD_TYPE;
      rec->FMHashLink = hashTable[bucket].chain;
      hashTable[bucket].chain = rec;
      hashTable[bucket].typesInChain |= ANNFM_LD_TYPE;
   }

   for (addr = AnnFirst("store", &rec); rec; addr = AnnNext(&rec)) {
      existingAnns |= ANNFM_ST_TYPE;
      bucket = HASH_FUNC(addr);
      rec->FMHashType = ANNFM_ST_TYPE;
      rec->FMHashLink = hashTable[bucket].chain;
      hashTable[bucket].chain = rec;
      hashTable[bucket].typesInChain |= ANNFM_ST_TYPE;
   }
}

static void AnnFMAddDynamic(AnnPtr rec, int type)
{
   VA addr = rec->arg;
   int bucket = HASH_FUNC(addr);

   /* Check to see if this rec is already in the hash table */
   AnnPtr hrec;
   for (hrec = hashTable[bucket].chain; hrec; hrec = hrec->FMHashLink) {
      if (hrec == rec) {
         break;
      }
   }
   /* If not found, add it to the hash table */
   if (hrec != rec) {
      existingAnns |= type;
      rec->FMHashType = type;
      rec->FMHashLink = hashTable[bucket].chain;
      hashTable[bucket].chain = rec;
      hashTable[bucket].typesInChain |= type;
   }
   /* Flush the annotation cache so the new annotation will be found */
   FlushAnnCache();
}


int AnnFMTypesExist(int types)
{
   return types & existingAnns;
}

int AnnFMRangeCheck(VA addr, int types)
{
   uint block;
   int bucket = HASH_FUNC(addr);
   AnnRec *rec;
   
   if (hashTable[bucket].typesInChain & types) {
      for (rec = hashTable[bucket].chain; rec; rec = rec->FMHashLink) {
         block = addr >> log2RangeSize;
         if ((rec->FMHashType & types) && (block == (rec->arg >> log2RangeSize))) {
            return 1;
         }
      }
   }

   return 0;
}

uint AnnFMRangeVector(VA addr, int types)
{
   uint retval = 0;
   VA block = addr >> log2RangeSize;
   int bucket = HASH_FUNC(addr);
   int word;
   AnnRec *rec;

   if (hashTable[bucket].typesInChain & types) {
      for (rec = hashTable[bucket].chain; rec; rec = rec->FMHashLink) {
         if ((rec->FMHashType & types) 
             && (block == (rec->arg >> log2RangeSize))) {
            word = (rec->arg & (rangeSize - 1)) >> 2;
            retval |= (1 << word);
         }
      }
   }
   return retval;   
}

AnnPtr AnnFMLookup(VA addr, int type)
{
   int bucket = HASH_FUNC(addr);
   AnnRec *rec;

   if (hashTable[bucket].typesInChain & type) {
      for (rec = hashTable[bucket].chain; rec; rec = rec->FMHashLink) {
         if ((rec->FMHashType & type) && (addr == rec->arg)) {
            return rec;
         }
      }
   }

   return NULL;
}

void AnnFMExec(VA addr, int type)
{
   int bucket = HASH_FUNC(addr);
   AnnRec *rec;

   if (hashTable[bucket].typesInChain & type) {
      for (rec = hashTable[bucket].chain; rec; rec = rec->FMHashLink) {
         if ((rec->FMHashType & type) && (addr == rec->arg)) {
            AnnExec(rec);
            return;
         }
      }
   }
}

void AnnFMInfo(Tcl_Interp *interp)
{
   AnnRec *rec;
   char buf[64];
   int hist[NUM_HIST];
   int i, b, count;
   
   Tcl_AppendResult(interp, "\nAnnotation Fast Memory Lookup:\n", NULL);
   
   sprintf(buf, "%d", HASH_TABLE_SIZE);
   Tcl_AppendResult(interp, "Num Buckets: ", buf, "\n", NULL);

   for (i=0; i<NUM_HIST; i++) {
      hist[i] = 0;
   }
   
   for (b=0; b<HASH_TABLE_SIZE; b++) {
      count = 0;
      for (rec = hashTable[b].chain; rec; rec = rec->FMHashLink) {
         count++;
      }
      hist[(count < NUM_HIST) ? count : NUM_HIST-1]++;
   }

   for (i=0; i<NUM_HIST; i++) {
      if (i<NUM_HIST-1) {
         sprintf(buf, "%d elems = %d", i, hist[i]);
      } else {
         sprintf(buf, "%d+ elems = %d", i, hist[i]);
      }
      Tcl_AppendResult(interp, "number of buckets with ", buf, "\n", NULL);
   }

   Tcl_AppendResult(interp, "\n", NULL);
}

/*****************************************************************
 * Commonly used annotation shortcuts
 *****************************************************************/
#define MAX_CYCLE_ANNS 2000
static struct CycleAnnHdr {
   EventCallbackHdr hdr;
   AnnPtr ptr;
} cycleAnnEvent[MAX_CYCLE_ANNS];

static int lastCycleAnn = 0;

static void
CycleAnnCallback(int cpuNum, EventCallbackHdr *hdr, void *arg) {
   struct CycleAnnHdr *cycleAnn = (struct CycleAnnHdr *) hdr;

   AnnExec(cycleAnn->ptr);
}

static void
InstallCycleAnnotation(AnnPtr ptr, uint64 count) {
   SimTime now  = CPUVec.CycleCount(0);
   SimTime when = count;
   SimTime diff = when - now;

   cycleAnnEvent[lastCycleAnn].ptr = ptr;
   EventDoCallback(0, CycleAnnCallback,
                   (EventCallbackHdr *) &cycleAnnEvent[lastCycleAnn],
                   0, diff);
   lastCycleAnn++;
   ASSERT(lastCycleAnn < MAX_CYCLE_ANNS);
}

#define MAX_PERIODIC_ANNS 2000
static struct PeriodicAnnHdr {
   EventCallbackHdr hdr;
   SimTime period;
   AnnPtr ptr;
} periodicAnnEvent[MAX_PERIODIC_ANNS];

static int lastPeriodicAnn = 0;

static void
PeriodicAnnCallback(int cpuNum, EventCallbackHdr *hdr, void *arg) {
   struct PeriodicAnnHdr *periodicAnn = (struct PeriodicAnnHdr *) hdr;
   AnnExec(periodicAnn->ptr);
   EventDoCallback(0, PeriodicAnnCallback, hdr, 0, periodicAnn->period);
}

static void
InstallPeriodicAnnotation(AnnPtr ptr, uint64 period) 
{
   periodicAnnEvent[lastPeriodicAnn].ptr = ptr;
   periodicAnnEvent[lastPeriodicAnn].period = period;

   EventDoCallback(0, PeriodicAnnCallback,
                   (EventCallbackHdr *) &periodicAnnEvent[lastPeriodicAnn],
                   0, period);
   lastPeriodicAnn++;
   ASSERT(lastPeriodicAnn < MAX_PERIODIC_ANNS);
}

AnnPtr *annExcTable;
AnnPtr annRFE;
AnnPtr annUTLB;
#ifdef JUMP_ANNS
AnnPtr annJR;
AnnPtr annJAL;
#endif

int  annExcNum, annInstNum;
bool annLoads  = FALSE;
bool annStores = FALSE;
bool annPCs    = FALSE;
bool annWatchpoints = FALSE;

void 
AnnCommonSetup(void) 
{
   AnnPtr iter = NULL;
   AnnPtr load, store;
   uint64 annCycle, annPeriodic;
   VA annExc, annMem;
   int i;

   annPCs = AnnFMTypesExist(ANNFM_PC_TYPE);

   AnnFirst("load", &load);
   AnnFirst("store", &store);

   /* Do this to keep things fast when there are no watchpoints at all */
   if (load || store) {
      annWatchpoints = TRUE;
      if (load)  { annLoads = TRUE; }
      if (store) { annStores = TRUE; }
      
      for (annMem = AnnFirst("load", &iter);iter;annMem = AnnNext(&iter)) {
         CPUVec.InstallMemAnnotation(annMem);
      }
      for (annMem = AnnFirst("store",&iter);iter;annMem = AnnNext(&iter)) {
         CPUVec.InstallMemAnnotation(annMem);
      }
   }
   
   for (annCycle = AnnFirst("cycle", &iter); iter; annCycle = AnnNext(&iter)) {
      InstallCycleAnnotation(iter, annCycle);
   }

   for (annPeriodic = AnnFirst("periodic", &iter); iter; 
        annPeriodic = AnnNext(&iter)) {
      InstallPeriodicAnnotation(iter, annPeriodic);
   }

   annRFE  = AnnFind("inst", "rfe");
   annUTLB = AnnFind("utlb","");
#ifdef JUMP_ANNS
   annJAL  = AnnFind("inst", "jal");
   annJR   = AnnFind("inst", "jr");
#endif

   annExcNum   = AnnNumArg("exc");
   annExcTable = (AnnPtr*) ZMALLOC (annExcNum * sizeof(AnnPtr),"AnnExcTable");

   for(i=0; i < annExcNum; i++) {
      annExcTable[i] = NULL;
   }

   for (annExc = AnnFirst("exc", &iter); iter; annExc = AnnNext(&iter)) {
      if (annExc < annExcNum) {
         annExcTable[annExc] = iter;
      }
   }

   ScacheAnnLateInit(TCLInterp);

   staticAnnsInstalled = 1;
}


/*****************************************************************
 * Common route for dynamically added annotations. Each annotation
 * goes through this procedure individually. 
 * We check to see if the is an existing AnnRec:
 * If so, there's no need to add it again
 *****************************************************************/
static void
AnnAddDynamic(AnnPtr ptr, int newrec)
{
   AnnRec *rec = ptr;

   if (!strcmp(rec->type->name, "cycle")) {
      if (newrec) {
         InstallCycleAnnotation(rec, rec->arg);
      }
   } else if (!strcmp(rec->type->name, "periodic")) {
      if (newrec) {
         InstallPeriodicAnnotation(rec, rec->arg);
      }
   } else if (!strcmp(rec->type->name, "pc")) {
      if (newrec) {
         AnnFMAddDynamic(rec, ANNFM_PC_TYPE);
      } else {
         FlushAnnCache();
      }
   } else if (!strcmp(rec->type->name, "pre-pc")) {
      if (newrec) {
         AnnFMAddDynamic(rec, ANNFM_PRE_PC_TYPE);
      } else {
         FlushAnnCache();
      }
   } else if (!strcmp(rec->type->name, "load")) {
      if (newrec) {
         AnnFMAddDynamic(rec, ANNFM_LD_TYPE);
      } else {
         FlushAnnCache();
      }
   } else if (!strcmp(rec->type->name, "store")) {
      if (newrec) {
         AnnFMAddDynamic(rec, ANNFM_ST_TYPE);
      } else {
         FlushAnnCache();
      }
   } else if (!strcmp(rec->type->name, "commit")) {
     /* nothing to do here */
   } else if (!strcmp(rec->type->name, "scache")) {
      ScacheAnnLateInit(TCLInterp);
   } else if (!strcmp(rec->type->name, "inst")) {
      annRFE  = AnnFind("inst", "rfe");
#ifdef JUMP_ANNS
      annJAL  = AnnFind("inst", "jal");
      annJR   = AnnFind("inst", "jr");
#endif
   } else if (!strcmp(rec->type->name, "utlb")) {
      annUTLB = AnnFind("utlb","");
   } else if (!strcmp(rec->type->name, "exc")) {
      AnnPtr iter = NULL;
      VA annExc;
      int i;

      annExcNum   = AnnNumArg("exc");
      free(annExcTable);
      annExcTable = (AnnPtr*) ZMALLOC (annExcNum * sizeof(AnnPtr),"AnnExcTable");
      
      for(i=0; i < annExcNum; i++) {
         annExcTable[i] = NULL;
      }
      
      for (annExc = AnnFirst("exc", &iter); iter; annExc = AnnNext(&iter)) {
         if (annExc < annExcNum) {
            annExcTable[annExc] = iter;
         }
      }
   } else {
     /* nothing to do here */
   }      
}

/******************************************************************
 * Scache annotations common to mipsy and embra.
 * Placed here for want of a better place.
 * ***************************************************************/

/*****************************************************************
 * Scache annotation related defines. They are dependent on the 
 * defines of cache events in hw_events.h.
 *****************************************************************/

#define SA_INSTR 0
#define SA_READ 1
#define SA_WRITE 2
#define SA_UPGRADE 6
#define SA_SC_UPGRADE 14

#define SA_DATAMISSSIZE 16
#define SA_DATAMISSSHIFT 4
#define SA_DATAMISSMASK 0x0f

static VA annScacheVaddr = 0;    /* information for a Scache miss */
static PA annScachePaddr = 0;
static uint annScacheType = 0;
static uint annScacheCPU = 0;
static uint annScacheStall = 0;

/* Declaration of scacheAnnotation moved up so AnnAddDynamic can set it */
static AnnPtr scacheInstrAnn = 0;
static AnnPtr scacheDataAnn[SA_DATAMISSSIZE];


/* 
 * Init routine for the scache miss annotation.
 * We set up a table with the appropriate annotation routines
 * as returned by tcl. In late init also set a variable, if any of the scache
 * annotations are set. All of this is to reduce the overhead of
 * of supporting a scache miss annotation
 */
static char *
ScacheMissAccess(ClientData clientData, Tcl_Interp *interp,
                 char *name1, char *name2, int flags)
{
   char buf[32];
   char *pRes;

   if (flags & TCL_TRACE_READS) {
      /* set its value */
      if (!strcmp(name2, "Vaddr")) {
         sprintf(buf, "%#lx", (uint64) annScacheVaddr);
      } else if (!strcmp(name2, "Paddr")) {
         sprintf(buf, "%#lx", (uint64) annScachePaddr);
      } else if (!strcmp(name2, "Type")) {
         sprintf(buf, "%d", annScacheType);
      } else if (!strcmp(name2, "CPU")) {
         sprintf(buf, "%d", annScacheCPU);
      } else if (!strcmp(name2, "Stall")) {
         sprintf(buf, "%d", annScacheStall);
      } else {
         return "Unknown ScacheMiss field";
      }
      pRes = Tcl_SetVar2(interp, name1, name2, buf, 0|TCL_LEAVE_ERR_MSG);
      if (!pRes)
         return pRes;
   } else if (flags & TCL_TRACE_WRITES) {
      return "Writes to ScacheMiss ignored";
   }
   return NULL;
}

void
ScacheAnnInit(Tcl_Interp *interp)
{
   int i;

   for (i=0;i<(SA_DATAMISSSIZE);i++) {
      scacheDataAnn[i] = (AnnPtr) 0;
   }

   /* Variables to track the Scache miss 
    * Need to use an array to support 64-bit since we can't link 64-bit values
    * Must define 1 element of the array so TCL thinks its defined
    */
   Tcl_SetVar2(interp, "ScacheMiss", "CPU", "0", TCL_GLOBAL_ONLY);
   Tcl_TraceVar(interp, "ScacheMiss",
                TCL_TRACE_READS | TCL_TRACE_WRITES | TCL_GLOBAL_ONLY,
                ScacheMissAccess, (ClientData) NULL);
}

void
ScacheAnnLateInit(Tcl_Interp *interp)
{
   AnnPtr ap;

   if ((ap = AnnFind("scache", "instr"))) {
      scacheInstrAnn = ap;
      scacheAnnotation = 1;
   }

   if ((ap = AnnFind("scache", "read"))) {
      scacheDataAnn[SA_READ] = ap;
      scacheAnnotation = 1;
   }

   if ((ap = AnnFind("scache", "write"))) {
      scacheDataAnn[SA_WRITE] = ap;
      scacheAnnotation = 1;
   }

   if ((ap = AnnFind("scache", "upg"))) {
      scacheDataAnn[SA_UPGRADE] = ap;
      scacheAnnotation = 1;
   }

   if ((ap = AnnFind("scache", "sc_upg"))) {
      scacheDataAnn[SA_SC_UPGRADE] = ap;
      scacheAnnotation = 1;
   }
}

/*
 * Routine called from hw_events.h to execute the Scache annotation
 */
void
ScacheAnnotation(VA vAddr, PA pAddr, uint type, uint cpu, uint stallTime)
{
   annScacheVaddr = vAddr;
   annScachePaddr = pAddr;
   annScacheType = type;
   annScacheCPU = cpu;
   annScacheStall = stallTime;

   if (type & E_I) {
      AnnExecCPU(scacheInstrAnn, cpu);
   } else {
      AnnExecCPU(scacheDataAnn[(type >> SA_DATAMISSSHIFT) & SA_DATAMISSMASK], cpu);
   }
}


/*****************************************************************
 * Complete support for pre- and post- pc annotations. 
 *
 * This is EXTREMELY optimized and is thus a little confusing. It 
 * works as follows. 
 * - Each time you enter a new range of PC values, you check a tag. 
 * - If you match the tag, then there are no annotations on this
 *   range (and you are looking at an up-to-date range). 
 * - If you don't match the tag, this is considered "Interesting".
 *   It means that either this is a PC value  in a range not in the
 *   cache or there is an annotation somewhere in its range. 
 * - The determination of which one of these that it is comes from the
 *   highest bit in the tag... if it is set, then there is an
 *   annotation on it. 
 *
 ****************************************************************/

#define ANN_PRESENT_MASK     VA_MASK_HIGH_BIT
#define ANN_SET_PRESENT_MASK VA_HIGH_BIT

#define ANN_RANGE_INTEREST(_tag, _ind) \
   (_tag != (annCache[_ind].postAnnTag & ANN_PRESENT_MASK))

#define ANN_SET_PRESENT(_ind) \
   (annCache[_ind].postAnnTag |= ANN_SET_PRESENT_MASK)

#define ANN_PC_HAS_ANN(_pc, _ind) \
   (annCache[_ind].postAnnBitVec & (1 << ((_pc >> 2) & 0x1f)))

#define PRE_ANN_RANGE_INTEREST(_tag, _ind) \
   (_tag != (annCache[_ind].preAnnTag & ANN_PRESENT_MASK))

#define PRE_ANN_SET_PRESENT(_ind) \
   (annCache[_ind].preAnnTag |= ANN_SET_PRESENT_MASK)

#define PRE_ANN_PC_HAS_ANN(_pc, _ind) \
   (annCache[_ind].preAnnBitVec & (1 << ((_pc >> 2) & 0x1f)))

AnnCache annCache[ANN_CACHE_SIZE];

extern bool tookCheckpoint;

static void 
FlushAnnCache(void)
{
   int i;
   for (i = 0; i < ANN_CACHE_SIZE; i++) {
      annCache[i].postAnnTag = 0;
      annCache[i].preAnnTag = 0;
   }
}

bool
RunPCAnnotations(VA pc, VA nPC) {
   VA   tag   = ANN_TAG(pc);
   uint index = ANN_INDEX(tag);

   if (tag != annCache[index].postAnnTag) {
      if (ANN_RANGE_INTEREST(tag, index)) {
         /* Set up for this new range of PC values for this index. */
         annCache[index].postAnnTag    = tag;
         annCache[index].postAnnBitVec = AnnFMRangeVector(pc, ANNFM_PC_TYPE);
         if (annCache[index].postAnnBitVec) {
            /* Note that there is an annotation somewhere in this range */
            ANN_SET_PRESENT(index);
         }
      }
      if (ANN_PC_HAS_ANN(pc, index)) {
         AnnExec(AnnFMLookup(pc, ANNFM_PC_TYPE)); 

         /* UGLY: If this was a cpt, we've already moved the PC and nPC forward. */ 
         if (tookCheckpoint) { 
            tookCheckpoint = FALSE; 
            return TRUE;
        } 
     }
  } 

  if (prePCAnnsExist) {
     tag = ANN_TAG(nPC);
     index = ANN_INDEX(tag);

     if (tag != annCache[index].preAnnTag) {
        if (PRE_ANN_RANGE_INTEREST(tag, index)) {
           annCache[index].preAnnTag    = tag;
           annCache[index].preAnnBitVec = AnnFMRangeVector(nPC, ANNFM_PRE_PC_TYPE);
           if (annCache[index].preAnnBitVec) {
              PRE_ANN_SET_PRESENT(index);
           }
        }
        if (PRE_ANN_PC_HAS_ANN(nPC, index)) {
           AnnExec(AnnFMLookup(nPC, ANNFM_PRE_PC_TYPE)); 
        }
     } 
  }
  return FALSE;
}

