/*
 * Copyright (C) 1998 by the Board of Trustees
 *    of Leland Stanford Junior University.
 * Copyright (C) 1998 Digital Equipment Corporation
 *
 * This file is part of the SimOS distribution.
 * See LICENSE file for terms of the license.
 *
 */

/*
 *
 */

/* Epsilon CPU Scheduler Interface definition */

#ifndef EPSILON_H
#define EPSILON_H

#undef EPSILON_USE_MEMREF

#ifndef EPSILON_SCHEDULER_H
#include "epsilon-scheduler.h"
#endif

typedef int Instid;
typedef long Inum;
#define N_INFLIGHT_INSTRS 2048
#define inumInstid(inum) ((inum)&(N_INFLIGHT_INSTRS-1))
#define NEXT_INSTID(n) (((n)+1)&(N_INFLIGHT_INSTRS-1))
#define PREV_INSTID(n) (((n)-1)&(N_INFLIGHT_INSTRS-1))

#define N_LOGICAL_REGISTERS 80 /* 32 integer, 32 float, 16 shadow */
#define N_PHYSICAL_REGS 512
#define ZERO_REGISTER_DEST_PHYSREG (ZERO_REGISTER+32)
typedef short regname_t;


typedef struct InflightInstructionState_s {
  VA PC;
  PA ppc;
  union alpha_instruction instr;
  unsigned isMemop : 1;
  Reg64 fpcr;
  Inum inum;       /* fetchnum, unique for a given CPU */
  /* for tracking instruction state */
  unsigned inflight : 1;
  unsigned fetched : 1;
  unsigned fetched2 : 1;
  unsigned mapped : 1;
  unsigned issued : 1;
  unsigned committed : 1;

  unsigned trapfree : 1; /* set this to one when this instruction can no longer cause a trap */
  unsigned untrappable : 1; /* set this to one when we're not allowed to trap an instruction 
                             * (e.g. after we issue an uncached memop request) */

  unsigned trapped : 1;
  unsigned memopRequested : 1;
  unsigned memopReadyToComplete : 1;
  unsigned memopCompleted : 1;
  unsigned done : 1;                         /* ((committed && (!isMemop || memopCompleted)) || trapped) */
  unsigned readyToIssue : 1;
  unsigned readyToRetire : 1;
  unsigned needsReissue : 1;                 /* instruction was issued before dependences satisfied */
  unsigned interrupted : 1;                  /* set if instruction is to be trapped because of an interrupt */

  Inum inorderNext;
  Inum inorderPrev;

  ulong runstate;
  short  immuStatus;
  short  dmmuStatus;
  short  memrefStatus; /* result from cache request */
   /* for keeping track of how to unwind stallFetchUntilInum */
  short stallFetchUntilCommit;

  /* for recovering instruction map on traps */
  regname_t pra; /* physical ra number */
  regname_t prb; /* physical ra number */
  regname_t prc; /* physical ra number */
  regname_t prdest; /* physical destreg number */
  regname_t oldprdest; /* old physical destreg number */
  regname_t lrdest; /* logical destreg number */
  /* for loads and stores: store-buffer */
  VA     eaddr;
  PA     paddr;
  MA     maddr;
  struct StoreBuffer {
    Reg64  vmask;
  } storebuffer;
  Reg64  value; /* for stores only */
  Inum lastStoreInum;
  /* for tracking control flow */
  VA     nextpc;

  /* for speculative exception handling (virtualizing IPR_EXC_ADDR) */
  AlphaTrapState *trapState;
  
  /* per-inflight-instruction pointer used by microarchitecture scheduler */
  void *schedulerInstrDesc;

} InflightInstructionState;

/* 
 * EpsilonState:
 *   
 *   Initial regmap maps each logical register to the physical register of
 *   the same number, except that both ZERO_REGISTER and FP_ZERO_REGISTER
 *   map to ZERO_REGISTER.  Physical register number FP_ZERO_REGISTER is
 *   used to catch the discarded value from instructions whose logical
 *   destination register is either ZERO_REGISTER or FP_ZERO_REGISTER.
 *   
 *   The physical register free list is linked through the register file,
 *   with the last register containing -1.
 *   
 */

typedef struct {
   char *name;
   int needsNewTrapState;
   int delta; /* which inum to kill: 0 for trapCauseInum, 1 for inorderNext, 2 for trapPointInum */
} EpsilonTTInfo;

#define EpsilonTTName(tt_) epsilonTTInfo[tt_].name
extern EpsilonTTInfo epsilonTTInfo[];

typedef struct EpsilonState {
   VA nextpc;
   Inum stallFetchCount; /* if nonzero, stall fetch */
   Inum lastInumFetched;
   Inum lastInumMapped;
   Inum lastInumCommitted;
   Inum lastInumDone;
   Inum lastInumUntrapped;
   Inum nextInum;
   Inum lastStoreInum;
   Inum trapPointInum; /* points to lowest inum that may still be trapped */
   long fetchCount;
   regname_t regmap[N_LOGICAL_REGISTERS];			/* fetch-time regmap */
   regname_t committedRegmap[N_LOGICAL_REGISTERS];		/* committed regmap */
   regname_t freereg;
   int useShadowRegisters;
   int myNum;
   Reg regs[N_PHYSICAL_REGS];
   int regValid[(N_PHYSICAL_REGS+31)/32];
   int regAllocated[(N_PHYSICAL_REGS+31)/32];
   InflightInstructionState iqueue[N_INFLIGHT_INSTRS];
   /* committed trap state */
   AlphaTrapState *oldestTrapState;
   /* speculative trap state */
   AlphaTrapState *newestTrapState;

   /* trap handling */
   int trapPending;
   VA trapPC; /* PC of trapped instruction */
   EpsilonTrapType trapType;
   Inum trapCauseInum; /* inum of cause of trap */
   Inum trapAbortInum; /* first inum to kill */

   struct {
      int pending;  /* there is an ExceptionStateUpdate pending */
      int trapType; /* the type of the trap */
      InflightInstructionState *tciis;
      Inum trapCauseInum;    /* the trapCauseInum */
      Inum trapAbortInum;    /* the trapAbortInum */
   } exceptionStateUpdate;

   /* interrupt handling */
   long pending_irq;

   /* ldst */
   int pendingMemops;
   Inum stalledMemop;

} EpsilonState;

extern EpsilonState *epeArray;
extern EpsilonState **EPE;

#define EpsilonInstructionIsMemop(cpu,instid) EPE[cpu]->iqueue[instid].isMemop
#ifdef op_stb
#define ALPHA_STORE_OPCODES_MASK \
   ((1L<<op_stw)|(1L<<op_stb)|(1L<<op_stq_u)|(1L<<op_stf)|(1L<<op_stg)|(1L<<op_sts)|(1L<<op_stt)|(1L<<op_stl)|(1L<<op_stq)|(1L<<op_stl_c)|(1L<<op_stq_c)|(1L<<PRIV_OP_HW_ST))
#else
#define ALPHA_STORE_OPCODES_MASK \
   ((1L<<op_stq_u)|(1L<<op_stf)|(1L<<op_stg)|(1L<<op_sts)|(1L<<op_stt)|(1L<<op_stl)|(1L<<op_stq)|(1L<<op_stl_c)|(1L<<op_stq_c)|(1L<<PRIV_OP_HW_ST))
#endif
#define ALPHA_LOAD_OPCODES_MASK \
   ((1L<<op_ldwu)|(1L<<op_ldbu)|(1L<<op_ldq_u)|(1L<<op_ldf)|(1L<<op_ldg)|(1L<<op_lds)|(1L<<op_ldt)|(1L<<op_ldl)|(1L<<op_ldq)|(1L<<op_ldl_l)|(1L<<op_ldq_l)|(1L<<PRIV_OP_HW_LD))
#define ALPHA_CONTROL_FLOW_OPCODES_MASK \
   ((1L<<op_br) \
    |(1L<<op_fbeq) \
    |(1L<<op_fblt) \
    |(1L<<op_fble) \
    |(1L<<op_bsr) \
    |(1L<<op_fbne) \
    |(1L<<op_fbge) \
    |(1L<<op_fbgt) \
    |(1L<<op_blbc) \
    |(1L<<op_beq) \
    |(1L<<op_blt) \
    |(1L<<op_ble) \
    |(1L<<op_blbs) \
    |(1L<<op_bne) \
    |(1L<<op_bge) \
    |(1L<<op_bgt) \
    |(1L<<op_jsr) \
    |(1L<<op_call_pal) \
    |(1L<<op_pal1e))

#define EpsilonInstructionIsStore(cpu,instid) \
   ((ALPHA_STORE_OPCODES_MASK >> EPE[cpu]->iqueue[instid].instr.common.opcode)&1)
#define EpsilonInstructionIsLoad(cpu,instid) \
   ((ALPHA_LOAD_OPCODES_MASK >> EPE[cpu]->iqueue[instid].instr.common.opcode)&1)
#define EpsilonInstructionIsControlFlow(cpu,instid) \
   ((ALPHA_CONTROL_FLOW_OPCODES_MASK >> EPE[cpu]->iqueue[instid].instr.common.opcode)&1)

#define EpsilonInstructionNextPC(cpu,instid) EPE[cpu]->iqueue[instid].nextpc

#define EpsilonPhysicalRegisterIsValid(epe,physreg) \
    ((epe->regValid[physreg>>5]>>(physreg&0x1f))&1)
#define EpsilonPhysicalRegisterSetValid(epe,physreg) \
     (epe->regValid[physreg>>5] |= (1 << (physreg&0x1f)))
#define EpsilonPhysicalRegisterClearValid(epe,physreg) \
     (epe->regValid[physreg>>5] &= ~(long)(1L << (physreg&0x1f)))

#define EpsilonPhysicalRegisterIsAllocated(epe,physreg) \
    ((epe->regAllocated[physreg>>5]>>(physreg&0x1f))&1)
#define EpsilonPhysicalRegisterSetAllocated(epe,physreg) \
     (epe->regAllocated[physreg>>5] |= (1 << (physreg&0x1f)))
#define EpsilonPhysicalRegisterClearAllocated(epe,physreg) \
     (epe->regAllocated[physreg>>5] &= ~(long)(1L << (physreg&0x1f)))

typedef struct EpsilonParams {
   int padding;
   int debugScheduler;
   int debugPreCycle;
   int debugOutOfOrderMemop;
   int debugUncachedMemops;
   int debugTrapMask;
   int debugSchedulerDuringInterrupt;
   int usePmintCPUCycle;
   int checkRegmap;
   int trapClearRegisterLocks;
   int trapNoStall;
   int trapNotifyInterrupts;
   int trapNotifyAtFetchTime;
   int trapPalAtFetch2;
   int trapSpeculativeUncachedMemops;
   int interruptTrapUncommitted;
   int width;
   int pipelined;
   int interruptFrequencyMask;
   int newTrapStates;
   int takeInterruptAtRei;
   int useMemRef;
   int fetchAheadLimit;
   int deadlockCycles;
   const char *schedulerLibrary;
   const char *schedulerInitProcedureName;
} EpsilonParams;

extern EpsilonParams epsilonParams;



/* interface functions provided for EPSILON */

/*** FETCH ***/

/* returns instid */
Inum EpsilonFetchInstruction(int cpu, VA predicted_pc);
/* Causes GAMMA to fetch the next instruction from thread 'tid' at address
 * predicted_pc. */

/* 
 * This is called by the scheduler when it's ready to be notified of
 * front-end traps (ITB miss, illegal instruction, etc.)
 */
void EpsilonFetchInstruction2(int cpu, Inum inum);

/* Actually does nothing except set the mapped bit, since we map at fetch time. */
void EpsilonMapInstruction (int cpu, Inum inum);

/*** ISSUE and EXECUTE ***/

void EpsilonIssueInstruction(int cpu, Inum inum); 
/* Computes and returns the virtual address being accessed.
 *
 */

void EpsilonRequestMemop(int cpu, Inum inum); 
/* for a load instruction, reads store-buffer or memory and delivers data to target register,
 * for a write instruction, flushes value from store-buffer to memory 
 */

void EpsilonCompleteMemop(int cpu, Inum inum); 
/* for a load instruction, reads store-buffer or memory and delivers data to target register,
 * for a write instruction, flushes value from store-buffer to memory 
 */

/*** COMMIT ***/

void EpsilonCommitInstruction(int cpu, Inum inum);
/* Notifies EPSILON that instr 'i' has been commited.
 * Requires: instruction 'i' must be an instruction which has
 * already been fetched, but not yet committed or killed.
 */

extern int epsilonTrapNotificationInProgress;
void EpsilonProcessPendingTrap(AlphaState *pe, EpsilonState *epe);
void EpsilonNotifyTrap(int tpuNum, Inum trapInum, int trapType);
void EpsilonUpdateExceptionState(AlphaState *pe, EpsilonState *epe, int completingTrap);

void EpsilonTrapInstruction(int cpu, Inum inum);

void EpsilonCPUVectorInit(void);

void PrintAllInflightInstructions(EpsilonState *epe);



#ifndef PMINT_CPU_CYCLE
#define PMINT_CPU_CYCLE() NOTREACHED()
#endif

#ifndef PMINT_COPYIN
#define PMINT_COPYIN()
#endif
#ifndef PMINT_COPYOUT
#define PMINT_COPYOUT()
#endif

#endif /* EPSILON_H */
