/*
 * branch.c -- Sam pipeline, branch prediction unit
 */

#include "sam.h"
#include "ev5.h"
#include "print.h"
#include "hw_events.h"

/* #define BRANCH_DEBUG  */

/* Definitions for branch prediction algorithm */

#define BranchPredictTaken 0x04
#define BranchPredictMax   0x07

/* SamBPIndex: Mapping from PC to branch prediction table index */

static INLINE int SamBPIndex(VA pc)
{
   return ((pc >> 2) & ((SAM_BP_TABLE_SIZE) - 1));
}

/* SamThreadActivate: Create and activate a new thread */

static INLINE void
SamThreadActivate(SamState *sp, SamThreadState *th, SamThreadState *newth)
{
   int i;
   newth->pc = th->pc;
   for (i = 0; i < ALPHA_NUM_GPRS; i++) {
      SamRegState *gp = th->gprName[i];
      newth->gprName[i] = gp;
      SamAcquireReg(gp);
   }
   for (i = 0; i < ALPHA_NUM_FPRS; i++) {
      SamRegState *fp = th->fprName[i];
      newth->fprName[i] = fp;
      SamAcquireReg(fp);
   }
   newth->cmov2Fetch = 0;
   newth->oldPrediction = 0;
   newth->branchSP = -1;
   newth->status = ThSpec;
   List_InsertDebug(ThreadToList(newth), LIST_ATREAR(sp->activeThreads));
   newth->stallFetch = StallNone;
}

/* SamThreadDeactivate: Deactivate a pruned thread */

void
SamThreadDeactivate(SamState *sp, SamThreadState *th)
{
   /* Release registers */
   int i;
   for (i = 0; i < ALPHA_NUM_GPRS; i++) {
      SamRegState *gp = th->gprName[i];
      SamReleaseGPR(gp, sp);
   }
   for (i = 0; i < ALPHA_NUM_FPRS; i++) {
      SamRegState *fp = th->fprName[i];
      SamReleaseFPR(fp, sp);
   }

   /* Remove thread from active list */
   th->status = ThFree;
   List_RemoveDebug(ThreadToList(th));
   List_InsertDebug(ThreadToList(th), LIST_ATREAR(sp->freeThreads));
}

/* SamBranch{Call,Return,Restore}: Management of target prediction stack */

static INLINE void
SamBranchCall(SamState *sp, SamThreadState *th, VA pc)
{
   th->branchSP = sp->branchSP;
   th->oldPrediction = sp->branchStack[sp->branchSP];
#ifdef BRANCH_DEBUG
   CPUPrint("BP: CPU %d push %d: 0x%llx\n", sp->PE->myNum, sp->branchSP, pc);
#endif   
   sp->branchStack[sp->branchSP] = pc;
   if (++sp->branchSP >= SAM_BP_RETURN_STACK) sp->branchSP = 0;
}

static INLINE VA
SamBranchReturn(SamState *sp, SamThreadState *th)
{
   if (--sp->branchSP < 0) sp->branchSP = SAM_BP_RETURN_STACK - 1;
   th->branchSP = -1;
#ifdef BRANCH_DEBUG
   CPUPrint("BP: CPU %d pop %d: 0x%llx\n", sp->PE->myNum, sp->branchSP, 
            sp->branchStack[sp->branchSP]);
#endif   
   return sp->branchStack[sp->branchSP];
}

static INLINE void
SamBranchRestore(SamState *sp, SamThreadState *th)
{
   if ((th->branchSP >= 0) && (th->oldPrediction != 0)) {
#ifdef BRANCH_DEBUG
      CPUPrint("BP: CPU %d restore %d: 0x%llx\n", sp->PE->myNum, th->branchSP,
               th->oldPrediction);
#endif
      sp->branchStack[th->branchSP] = th->oldPrediction;
      sp->branchSP = th->branchSP;
   }
}

/* SamBranchPrune: Handle result of branch computation */

static void
SamBranchPrune(SamState *sp, SamThreadState *th)
{
   /* Unstall instruction fetch if waiting for a thread */
   SamThreadState *fth;
   SamGetFetchThread(fth, sp);
   fth->stallFetch &= ~StallThread;

   if (th == ListToThread(List_Last(sp->activeThreads))) {
      /* This thread has no child -- either:
       * 1. Unconditional immediate branch
       * 2. Indirect branch without prediction
       * Nothing to do for 1, need to unstall fetch for 2. */
      if (th->prediction == NotPredicted) {
         th->stallFetch &= ~StallBranch;
      }
      return;
   }

   ASSERT(th->prediction != NotPredicted);
   if (th->prediction == th->resolution) {
      /* Prediction was correct: merge this thread with its child */
      SamThreadState *cth = ListToThread(List_Next(ThreadToList(th)));

      List_Links *ilp;

      cth->status = th->status;

      /* Merge parent instructions into child thread */
      LIST_FORALL(sp->reorderBuffer, ilp) {
         SamIwinEntry *ip = ReorderListToIwin(ilp);
         if (ip->thread == th) {
            ip->thread = cth;
         } else if ((ip->thread == cth) && (cth->status != ThSpec)) {
            ip->iwinFlags &= ~IwinSpec;
            SamPrintStage(sp, ip, "NSQ");
            SamCheckInstAvail(ip, sp);
         }
      }
      
      SamThreadDeactivate(sp, th);
   } else {
      /* Prediction was incorrect: destroy all descendants of this thread */
      List_Links *lp = List_Last(sp->activeThreads);
      while (ListToThread(lp) != th) {
         SamThreadState *dth = ListToThread(lp);
         List_Links *ilp;

         /* Squash all instructions owned by this thread */
         LIST_FORALL(sp->reorderBuffer, ilp) {
            SamIwinEntry *ip = ReorderListToIwin(ilp);
            if (ip->thread == dth) {
               SamSquashInst(sp, ip);
               SamPrintStage(sp, ip, "SQ");
            }
         }

         /* Restore branch stack if this was a call */
         if (dth->call) {
            SamBranchRestore(sp, dth);
         }

         lp = List_Prev(lp);
         SamThreadDeactivate(sp, dth);
      }

      /* Account for branch misprediction latency */
      th->stallFetch |= StallMispredict;
      th->mispredictRestartCycle = samCurrentTime + SAM_BP_MISPREDICT_LATENCY;
   }
}

/* SamBranch: Main branch routine */

void
SamBranch(SamState *sp, SamIwinEntry *ip)
{
   bool isIndirect = ip->instFlags & InstIndirect;
   bool isReturn = ip->instFlags & InstReturn;
   bool isConditional = ip->instFlags & InstConditional;
   bool isCall = ip->instFlags & InstCall;

   SamBPEntry *bpEntry;
   SamThreadState *th, *newth;

   SamGetFetchThread(th, sp);
   th->call = isCall;
   th->prediction = NotPredicted;
   th->predictPC = 0x0;

   /* rpb: need to implement coroutine prediction */
   ASSERT(!(ip->instFlags & InstCoroutine));

   if (!isConditional && !isIndirect && !isCall) {
      /* Unconditional immediate branch: just do the branch */
      ip->iwinFlags |= IwinAvail | IwinIssued | IwinExecuted;
      SamPrintStage(sp, ip, "AV");
      SamPrintStage(sp, ip, "IS");
      SamPrintStage(sp, ip, "EX BR");

      th->prediction = th->resolution = Taken;
      th->pc = ip->target;

      /* Alpha: this must be a BR, which saves the updated PC in Ra */
      ASSERT(ip->instr.common.opcode == op_br);
      ip->ra.reg->contents.gpr = ip->pc + 4;
      SamRegWriteback(sp, ip);

      return;
   }

   /* Create a new thread for the branch */
   SamGetFreeThread(newth, sp);
   if (newth == NULL) {
      th->stallFetch |= StallThread;
      return;
   }

   bpEntry = &sp->branchPredictionTable[SamBPIndex(ip->pc)];

   if (!isConditional && !isIndirect && isCall) {
      /* Unconditional immediate call: no prediction is needed */
      SamThreadActivate(sp, th, newth);
      th->prediction = th->resolution = Taken;
      newth->pc = ip->target;
      SamBranchCall(sp, th, th->pc);
   } else if (isConditional) {
      /* Conditional: do branch prediction */
      SamThreadActivate(sp, th, newth);
      th->prediction = (bpEntry->bits & BranchPredictTaken) ? Taken : NotTaken;
      if (th->prediction == Taken) {
         newth->pc = ip->target;
         if (isCall) {
            SamBranchCall(sp, th, th->pc);
         }
      } else {
         th->pc = ip->target;
      }
   } else {
      /* Indirect: do target prediction */
      VA target;
      ASSERT(isIndirect);
      if (isCall) {
         SamBranchCall(sp, th, th->pc);
      }
      if (isReturn) {
         target = SamBranchReturn(sp, th);
      } else {
#ifdef BRANCH_DEBUG
         CPUPrint("BP: CPU %d get entry %d: 0x%llx\n", sp->PE->myNum,
                  SamBPIndex(ip->pc), bpEntry->target);
#endif
         target = bpEntry->target;
      }
      if (target != 0) {
         th->prediction = Taken;
         SamThreadActivate(sp, th, newth);
         th->predictPC = target;
         newth->pc = target;
      } else {
         /* No predicted target: stall until branch is processed */
         List_InsertDebug(ThreadToList(newth), LIST_ATFRONT(sp->freeThreads));
         th->stallFetch |= StallBranch;
      }
   }
}

void
SamBranchResolve(SamState *sp, SamIwinEntry *ip, bool resolution)
{
   SamBPEntry *bpEntry = &sp->branchPredictionTable[SamBPIndex(ip->pc)];
   SamBPStats *bpStats;

   if (resolution) {
      if (ip->instFlags & InstCall) {
         bpStats = &sp->branchStatistics[CallTaken];
      } else {
         bpStats = &sp->branchStatistics[BranchTaken];
      }
      if (bpEntry->bits < BranchPredictMax) bpEntry->bits++;
   } else {
      if (ip->instFlags & InstCall) {
         bpStats = &sp->branchStatistics[CallNotTaken];
      } else {
         bpStats = &sp->branchStatistics[BranchNotTaken];
      }
      if (bpEntry->bits > 0) bpEntry->bits--;
   }
   if (ip->thread != NULL) {
      ip->thread->resolution = resolution ? Taken : NotTaken;

      switch (ip->thread->prediction) {
      case NotPredicted:
         bpStats->unpredicted++;
         if (ip->thread->resolution == Taken) {
            ip->thread->pc = ip->target;
         }
         break;
      case NotTaken:
         if (ip->thread->resolution == Taken) {
            bpStats->incorrect++;
         } else {
            bpStats->correct++;
         }
         break;
      case Taken:
         if (ip->thread->resolution == Taken) {
            bpStats->correct++;
         } else {
            bpStats->incorrect++;
         }
         break;
      }
      SamBranchPrune(sp, ip->thread);
   }
}

void
SamBranchResolveJump(SamState *sp, SamIwinEntry *ip, VA target)
{
   SamBPEntry *bpEntry = &sp->branchPredictionTable[SamBPIndex(ip->pc)];
   SamBPStats *bpStats;

   if (ip->instFlags & InstCall) {
      bpStats = &sp->branchStatistics[CallIndirect];
   } else if (ip->instFlags & InstReturn) {
      bpStats = &sp->branchStatistics[ReturnIndirect];
   } else {
      bpStats = &sp->branchStatistics[JumpIndirect];
   }

   ip->thread->pc = target;
   if (!(ip->instFlags & InstReturn)) {
#ifdef BRANCH_DEBUG
         CPUPrint("BP: CPU %d set entry %d: 0x%llx\n", sp->PE->myNum,
                  SamBPIndex(ip->pc), target);
#endif
      bpEntry->target = target;
   }
   if (ip->thread != NULL) {
#ifdef BRANCH_DEBUG
      CPUPrint("BP: CPU %d predict 0x%llx actual 0x%llx\n",
               sp->PE->myNum, ip->thread->predictPC, target);
#endif

      if (ip->thread->prediction == NotPredicted) {
         bpStats->unpredicted++;
         ip->thread->resolution = NotTaken;
      } else if (ip->thread->predictPC == target) {
         bpStats->correct++;
         ip->thread->resolution = Taken;
      } else {
         bpStats->incorrect++;
         ip->thread->resolution = NotTaken;
      }

      SamBranchPrune(sp, ip->thread);
   }
}
