/*
 * Copyright (C) 1998 by the Board of Trustees
 *    of Leland Stanford Junior University.
 * Copyright (C) 1998 Digital Equipment Corporation
 *
 * This file is part of the SimOS distribution.
 * See LICENSE file for terms of the license.
 *
 */

/*****************************************************************
 * cpu_interface.c
 *****************************************************************/

#include <sys/signal.h>
#include <setjmp.h>

#include "syslimits.h"
#include "simtypes.h"
#include "cpu_interface.h"
#include "cpu_stats.h"
#include "checkpoint.h"

#include "alpha.h"
#include "sam.h"
#include "tcl_init.h" 
#include "machine_params.h"
#include "sim_error.h"
#include "gdb_interface.h"
#include "alpha_gdb.h"
#include "ev5.h"
#include "ev5_ipr.h"
#include "addr_layout.h"

static CPUStatus SamCPUStatus(int cpuNum);
static void   SamHandleDebugSignal(int cpuid, int sigusr); /* Defined in debug.c */
static VA     SamCurrentPC(int);
extern void   SamExitSimulator( CPUType toSim );
static int    SamCurrentCpuNum(void);
static Inst   SamCurrentInstruction(int);
static Result SamGetReg(int cpuNum, int regnum, Reg *val);
static void   SamClearLockFlag(int cpuNum);
static bool   SamGetLockFlag(int cpuNum);
static PA     SamGetLockAddr(int cpu);
static Result SamGetMem(int cpuNum, VA vAddr, uint nbytes, char *buf);
static Result SamTranslateVirtualNoSE(int cpuNum, VA vAddr, PA *paddr);
static void   SamInstallMemAnnotation(VA);
static void   SamDMAInval(int machine, PA* k0list);
static void   SamLaunchSlaveCPU(int cpuNum);
static int    SamDoCheckpoint(int cpuNum);
static void   SamException(int cpuNum, int trap);
static int    SamFillCacheLine(int cpuNum, VA pc, VA vAddr, PA pAddr, int type);
static void   SamFlushCacheLine(int cpuNum, PA pA, int size, int retain);

extern jmp_buf     jmpEnv;                    /* Used to exit the main loop */

int sam_debug_mode = 0;
int sam_break_nexti = 0;
int sam_sigusr = 0;

char *
SamPAToHostAddr(int cpu, PA pAddr, VA vAddr) 
{
#ifndef SOLO
   return (PHYS_TO_MEMADDR(M_FROM_CPU(cpu),pAddr) +
           (IS_REMAPPED_PADDR(pAddr,cpu) ? remapVec->NodeAddr[cpu] : 0));
#endif
#ifdef SOLO
   return (char *)(vAddr);
#endif 
}

extern void SamFinishMemRequests(void);

/*****************************************************************
 * CPUVectorInit
 *****************************************************************/
void 
SamCPUVectorInit(void)
{
   CPUVec.Handle_Debug_Signal  = SamHandleDebugSignal;
   CPUVec.GetRegister          = SamGetReg;
   CPUVec.PutRegister          = SamPutReg;
   CPUVec.ClearLockFlag        = SamClearLockFlag;
   CPUVec.GetLockFlag          = SamGetLockFlag;
   CPUVec.GetLockAddr          = SamGetLockAddr;
   CPUVec.GetMemory            = SamGetMem;
   CPUVec.PutMemory            = SamPutMem;
   CPUVec.TranslateVirtualNoSE = SamTranslateVirtualNoSE;
   CPUVec.InstallMemAnnotation = SamInstallMemAnnotation;
   CPUVec.Unstall              = SamUnstall;
   CPUVec.FlushCacheLine       = SamFlushCacheLine;
   CPUVec.FillCacheLine        = SamFillCacheLine;
   CPUVec.CPUstatus            = SamCPUStatus;
   CPUVec.OldCycleCount        = CPUVec.CycleCount;
   CPUVec.CycleCount           = SamCycleCount;
   CPUVec.DMAInval             = SamDMAInval;
   CPUVec.ExitSimulator        = SamExitSimulator;
   CPUVec.CurrentCpuNum        = SamCurrentCpuNum;
   CPUVec.CurrentPC            = SamCurrentPC;
   CPUVec.CurrentInstruction   = SamCurrentInstruction;
   CPUVec.InstrCount           = SamInstructionCount;
   CPUVec.DoCheckpoint         = SamDoCheckpoint;
   CPUVec.Exception            = SamException;
#if 0
   CPUVec.TakeBusError         = SamTakeBusError;
#endif
   CPUVec.useMemRef            = FALSE;
   CPUVec.singleEventQueue     = TRUE;
   CPUVec.PAToHostAddr         = SamPAToHostAddr;
   CPUVec.LaunchSlaveCPU       = SamLaunchSlaveCPU;
}

void
SamExitSimulator(CPUType toSim)
{
   int cpuNum;

   for (cpuNum = 0; cpuNum < TOTAL_CPUS; cpuNum++) {
      SamState *sp = SPE[cpuNum];
      sp->cpuSwitch = SwitchRequested;
   }
   exitToSimulator = toSim;
   readyToSwitch = 0;
}

int
SamCurrentCpuNum(void)
{
   return curSP->PE->myNum;
}

VA
SamCurrentPC(int cpu)
{
   /* Assume AlphaState is updated before this call is made */
   return PE[cpu]->PC;
}

Inst
SamCurrentInstruction(int cpu)
{
   /* Used indirectly in DTranslateVirtual */
   /* Assume AlphaState is updated before this call is made */
   return PE[cpu]->instr.word;
}

Result
SamGetReg(int cpu, int regnum, Reg *val)
{
   AlphaState *P = PE[cpu];

   /* AlphaState is updated at graduation, so it holds the correct values */
   if (regnum >=0 && regnum < 32) {
      *val = P->reg[regnum];
      return SUCCESS;
   }
   if (regnum >= 32 && regnum < 64) {
      *val = *(Reg *)&P->fp[regnum-32];
      return SUCCESS;
   }
   /* shadow registers */
   if (regnum >= SHADOW_REGNUM && regnum < (SHADOW_REGNUM+8)) {
      int shadowset = 8;
      Reg *regP;
      if (EV5_ShadowRegistersUsed(P, P->PC)) {
         /* return the saved register value */
         shadowset = 0;
      } else {
         /* return the shadow register value */
         shadowset = 8;
      }
      regP = (Reg *)&P->shadowRegs[regnum-67+shadowset];
      *val = *regP;
      return SUCCESS;
   }

   switch (regnum) {
   case INSTR_REGNUM: *val = P->instr.word; break;
   case PC_REGNUM: *val = P->PC; break;
   case FPCR_REGNUM: *val = P->fpcr; break;
   default: 
      CPUError("SamGetReg: gdb register %d not defined \n",regnum);
   }
   return SUCCESS;
}

Result
SamPutReg(int cpu, int regnum, Reg val)
{
   SamState *sp = SPE[cpu];
   SamThreadState *th;
   SamGetGradThread(th, sp);

   CPUWarning("SamPutReg %d %d: this might not work....", cpu, regnum);

   if (regnum >=0 && regnum < 32) {
      th->gprName[regnum]->contents.gpr = val;
      return SUCCESS;
   }
   if (regnum >= 32 && regnum < 64) {
      th->fprName[regnum-32]->contents.fpr = val;
      return SUCCESS;
   }

   switch (regnum) {
   case PC_REGNUM: ASSERT(0); break;
   case FPCR_REGNUM: ASSERT(0); break;
   default: 
      CPUError("SamGetReg: gdb register %d not defined \n",regnum);
   }
   return SUCCESS;
}


static void
SamClearLockFlag(int cpu)
{
   AlphaState *P = PE[cpu];
   P->lock_flag = 0;
}

static bool
SamGetLockFlag(int cpu)
{
   AlphaState *P = PE[cpu];
   return P->lock_flag;
}

static PA
SamGetLockAddr(int cpu)
{
   AlphaState *P = PE[cpu];
   return P->locked_addr;
}

Result
SamGetMem(int cpuNum, VA vAddr, uint nbytes, char *buf)
{
   int i;
   PA pA=0;
   SamState *sp = SPE[cpuNum];
   buf[0] = 0;
   for (i=0;i<nbytes;i++) {
      if (EV5_DTranslateVirtual(PE[cpuNum],vAddr+i,0,0,&pA)==MMU_SUCCESS) {
         bool ret = SamLdstGetByte(sp, pA, &buf[i]);
         if (!ret) {
            buf[i] = *(char *)PHYS_TO_MEMADDR(M_FROM_CPU(cpuNum),pA);
         }
      } else {
	 return FAILURE;
      }
   }
   return SUCCESS;
}

Result
SamPutMem(int cpuNum, VA vAddr, uint nbytes, char *buf)
{

   int i;
   PA pA=0;
   CPUWarning("Put mem size %d addr 0x%lx buf 0x%x (flush cache) \n",nbytes,vAddr,*(uint32*)buf);
   for (i=0;i<nbytes;i++) {
      if (EV5_DTranslateVirtual(PE[cpuNum],vAddr+i,0,0,&pA)==MMU_SUCCESS) {
	 *(char*)PHYS_TO_MEMADDR(M_FROM_CPU(cpuNum),pA) =  buf[i];
      } else {
	 return FAILURE;
      }
   }
   return SUCCESS;
}

Result
SamTranslateVirtualNoSE(int cpuNum, VA vAddr, PA *pAddr)
{
   if (EV5_DTranslateVirtual(PE[cpuNum],vAddr,0,0,pAddr)==MMU_SUCCESS) {
      return SUCCESS;
   } else { 
      return FAILURE;
   }
}

SimTime
SamCycleCount(int cpuNum)
{
   return samCurrentTime;
}

SimTime
SamInstructionCount(int cpuNum)
{
   return STATS_VALUE(cpuNum, numInstructions);
}

/*****************************************************************
 * SamDebug
 *
 * Main entry point into the debugger from normal execution. 
 *
 * NOTE: the second argument is_break_instruction is set when
 * we call in due to a single-step trap (from cpu.c) even though
 * there is no break instruction at the PC.  We need to send
 * the SIGTRAP to gdb in this case.  If we send a SIGUSR2 it
 * stops single-stepping immediately even though we may not
 * have yet reached the next line boundary (for 'n' or 's').
 *
 * Return value 0 means nothing has changed, go through normal
 * instruction completion (pc = npc, npc += 4, annotations, etc).
 *
 * Return value 1 means PC has changed or instruction at PC has
 * been overwritten.  If you entered the debugger between instructions
 * abort anything you have read about the next instruction and restart.
 * If you entered the debugger from a breakpoint, reexecute the
 * instruction at that PC.
 *
 *****************************************************************/
int 
SamDebug(int cpuNum, int is_break_instruction)
{
   Simdebug_result res;

   res = Simdebug_run(is_break_instruction ? SIGTRAP : SIGUSR2, cpuNum);
   switch (res) {
   case SD_CONTINUE:
      sam_break_nexti = SAM_NOBREAK;
      sam_debug_mode = 0;
      break;
   case SD_NEXTI_ANYCPU:
      sam_break_nexti = SAM_BREAKANYCPU;
      sam_debug_mode = 1;
      break;
   default:
      sam_break_nexti = res;
      sam_debug_mode = 1;
      break;
   }
   
   return 1;
}

/*****************************************************************
 * HandleDebugSignal
 *****************************************************************/
static void 
SamHandleDebugSignal(int cpuid, int sigusr)
{
   sam_debug_mode = 1;
   if (sigusr) {
      sam_sigusr = 1;
   } else { 
      if (cpuid == -1) {
         sam_break_nexti = SAM_BREAKANYCPU;
      } else {
         sam_break_nexti = cpuid;
      }
   }
}

static CPUStatus
SamCPUStatus(int cpuNum)
{
  return cpu_running;
}

static void
SamInstallMemAnnotation(VA vA)
{
}


void SamDMAInval(int machine, PA* k0list)
{
   ;
}

/*****************************************************************
 * HandleDebugSignal
 *****************************************************************/

void SamLaunchSlaveCPU(int cpuNum)
{
   CPUWarning("Sam: launching cpu %d \n",cpuNum);
   PE[cpuNum]->cpuState = cpu_running;
}


static int SamDoCheckpoint(int cpuNum)
{
   int i, result=0;
   VA oldPC[SIM_MAXCPUS];
   
   for(i=0;i<TOTAL_CPUS;i++) {
      oldPC[i] = PE[i]->PC;
   }
   if (AnnFMLookup(PE[cpuNum]->PC,ANNFM_PC_TYPE)) {
      /* most likely a post-pc annotation --> increment unless branch */
      uint32 inst =  CPUVec.CurrentInstruction(cpuNum);
      union alpha_instruction ainst;

      ainst.word = inst;
      switch (ainst.common.opcode) {
      case op_call_pal:
      case op_br:
      case op_bsr:
      case op_blbc:
      case op_beq:
      case op_blt:
      case op_ble:
      case op_blbs:
      case op_bne:
      case op_bge:
      case op_bgt:
      case op_fbeq:
      case op_fblt:
      case op_fble:
      case op_fbne:
      case op_fbge:
      case op_fbgt:
      case op_jsr:
         break;
      default:
         PE[cpuNum]->PC += 4;
      }
   }
   result = Simcpt_Checkpoint(CPT_SAVE,"");
   for(i=0;i<TOTAL_CPUS;i++) {
      PE[i]->PC = oldPC[i];
   }
   return result;
}

void
SamUnstall(int cpuNum)
{
   /* Nothing to do here -- processor doesn't stall */
}

static void
SamException(int cpuNum, int trap)
{
   SamState *sp = SPE[cpuNum];
   SamTrapRecord(sp, trap);
}

static int
SamFillCacheLine(int cpuNum, VA pc, VA vAddr, PA pAddr, int type)
{
   SamState *sp = SPE[cpuNum];
   return SamLdstFillLine(sp, pAddr, type);
}

static void
SamFlushCacheLine(int cpuNum, PA pA, int size, int retain)
{
   SamState *sp = SPE[cpuNum];
   SamLdstFlushLine(sp, pA, size, retain);
}
