/*
 * decode.c -- Sam pipeline, instruction decode unit
 */

#include "sam.h"
#include "print.h"
#include "instruction.h"

#define DECODE(TYPE)                                                    \
   void SamDecode ## TYPE (SamState *sp, SamIwinEntry *ip)

DECODE(PAL)
{
   switch (ip->instr.common.opcode) {
   case op_call_pal:
      ip->instFlags = InstCallPAL;
      ip->rdest = NULL;
      break;
   case PRIV_OP_MTPR:
      ip->instFlags = InstIPR;
      ip->ra.type = RegInt; ip->ra.areg = ip->instr.m_format.ra;
      ip->rdest = NULL;
      break;
   case PRIV_OP_MFPR:
      ip->instFlags = InstIPR;
      ip->ra.type = RegInt; ip->ra.areg = ip->instr.m_format.ra;
      ip->rdest = &ip->ra;
      break;
   case PRIV_OP_HW_LD:
      ip->instFlags = InstHWMem | InstLoad;
      ip->ra.type = RegInt; ip->ra.areg = ip->instr.m_format.ra;
      ip->rb.type = RegInt; ip->rb.areg = ip->instr.m_format.rb;
      ip->rdest = &ip->ra;
      break;
   case PRIV_OP_HW_ST:
      ip->instFlags = InstHWMem | InstStore;
      ip->ra.type = RegInt; ip->ra.areg = ip->instr.m_format.ra;
      ip->rb.type = RegInt; ip->rb.areg = ip->instr.m_format.rb;
      ip->rdest = NULL;
      break;
   case PRIV_OP_REI:
      ip->instFlags = InstREI;
      ip->rdest = NULL;
      break;
   }

   SamPrintStage(sp, ip, "ID PAL");
}

DECODE(LoadAddress)
{
   ip->ra.type = RegInt; ip->ra.areg = ip->instr.m_format.ra;
   ip->rb.type = RegInt; ip->rb.areg = ip->instr.m_format.rb;

   ip->rdest = &ip->ra;

   SamPrintStage(sp, ip, "ID LDA");
}

DECODE(MemLoad)
{
   int opcode = ip->instr.common.opcode;
   switch (opcode) {
   case op_ldf: ip->ra.type = RegFFlt; break;
   case op_ldg: ip->ra.type = RegGFlt; break;
   case op_lds: ip->ra.type = RegSFlt; break;
   case op_ldt: ip->ra.type = RegTFlt; break;
   default: ip->ra.type = RegInt; break;
   }
   ip->ra.areg = ip->instr.m_format.ra;
   ip->rb.type = RegInt; 
   ip->rb.areg = ip->instr.m_format.rb;

   ip->rdest = &ip->ra;

   /* Check for write to R31 */
   if (ip->rdest->areg == ZERO_REGISTER) {
      /* From EV6 manual section 2.6, "Load Instructions to R31 and F31":
       * ldq_u: no operation, skip issue and execute stages
       * lds: prefetch with modify intent
       * ldq: prefetch, evict next
       * other: prefetch
       * For now, handle ldq as a simple prefetch. */
      if (opcode == op_ldq_u) {
         ip->instFlags |= InstWriteZero;
      } else if (opcode == op_lds) {
         ip->instFlags |= InstPrefetchMod;
      } else {
         ip->instFlags |= InstPrefetch;
      }
   } else {
      ip->instFlags |= InstLoad;
   }

   SamPrintStage(sp, ip, "ID LD");
}

DECODE(MemStore)
{
   int opcode = ip->instr.common.opcode;
   ip->instFlags = InstStore;
   switch (opcode) {
   case op_stf: ip->ra.type = RegFFlt; break;
   case op_stg: ip->ra.type = RegGFlt; break;
   case op_sts: ip->ra.type = RegSFlt; break;
   case op_stt: ip->ra.type = RegTFlt; break;
   default: ip->ra.type = RegInt; break;
   }
   ip->ra.areg = ip->instr.m_format.ra;
   ip->rb.type = RegInt; 
   ip->rb.areg = ip->instr.m_format.rb;
   if ((opcode == op_stl_c) || (opcode == op_stq_c)) {
      ip->instFlags |= InstSC;
      // Hack: for SC, use RC to hold output version of ra
      ip->rc.type = RegInt; ip->rc.areg = ip->ra.areg;
   }

   if (ip->instFlags & InstSC) {
      ip->rdest = &ip->rc;
   } else {
      ip->rdest = NULL;
   }
   SamPrintStage(sp, ip, "ID ST");
}

DECODE(Branch)
{
   Reg_s disp = ip->instr.b_format.branch_displacement;
   ip->target = ip->pc + (disp << 2) + 4;
   ip->instFlags = InstBranch;

   switch (ip->instr.common.opcode) {
   case op_br:  break;
   case op_bsr: ip->instFlags |= InstCall; break;
   default:     ip->instFlags |= InstConditional; break;
   }
   switch (ip->instr.common.opcode) {
   case op_fbeq: case op_fblt: case op_fble:
   case op_fbne: case op_fbge: case op_fbgt:
      ip->ra.type = RegFlt; ip->ra.areg = ip->instr.b_format.ra;
      break;
   default:
      ip->ra.type = RegInt; ip->ra.areg = ip->instr.b_format.ra;
      break;
   }

   if (ip->instFlags & InstConditional) {
      ip->rdest = NULL;
   } else {
      ip->rdest = &ip->ra;
   }
   SamPrintStage(sp, ip, "ID BR");
}

DECODE(Jump)
{
   char *pszPrint;

   /* Indirect: don't know the target yet */
   ip->target = 0;

   ip->instFlags = InstBranch | InstIndirect;
   switch (ip->instr.j_format.function) {
   case jsr_jmp:           
      pszPrint = "ID JSR jmp";
      break;
   case jsr_jsr:           
      ip->instFlags |= InstCall; 
      pszPrint = "ID JSR jsr";
      break;
   case jsr_ret:           
      ip->instFlags |= InstReturn; 
      pszPrint = "ID JSR ret";
      break;
#ifdef rpb_notdef
   case jsr_jsr_coroutine: 
      ip->instFlags |= InstCoroutine; 
      pszPrint = "ID JSR cor";
      break;
#endif
   default:                
      SamDecodeIllegal(sp, ip); 
      ip->rdest = NULL; 
      return;
   }
   ip->ra.type = RegInt; ip->ra.areg = ip->instr.j_format.ra;
   ip->rb.type = RegInt; ip->rb.areg = ip->instr.j_format.rb;

   ip->rdest = &ip->ra;
   SamPrintStage(sp, ip, pszPrint);
}

DECODE(Operate)
{
   /* Check for conditional move instruction */
   if (ip->instr.common.opcode == op_intl) {
      int func = ip->instr.o_format.function;
      if ((func == intl_cmovlbs) || (func == intl_cmovlbc) ||
          (func == intl_cmoveq) || (func == intl_cmovne) ||
          (func == intl_cmovlt) || (func == intl_cmovge) ||
          (func == intl_cmovle) || (func == intl_cmovgt)) {
         ip->instFlags |= InstCMOV;
      }
   }

   ip->ra.type = RegInt; ip->ra.areg = ip->instr.o_format.ra;
   if (!(ip->instr.o_format.form)) {
      ip->rb.type = RegInt; ip->rb.areg = ip->instr.o_format.rb;
   }
   ip->rc.type = RegInt; ip->rc.areg = ip->instr.o_format.rc;

   ip->rdest = &ip->rc;

   /* Check for write to R31 */
   if (ip->rdest->areg == ZERO_REGISTER) {
      ip->instFlags |= InstWriteZero;
   }

   SamPrintStage(sp, ip, "ID INT");
}

/* SamDecodeCMOV[12]: called after SamDecodeOperate to handle CMOV */

DECODE(CMOV1)
{
   /* CMOV stage 1: copy oldRc to newRc1 -- put oldRc in Rb */
   if (ip->instFlags & InstFloating) {
      ip->rb.type = RegFlt; ip->rb.areg = ip->instr.f_format.fc;
   } else {
      ip->rb.type = RegInt; ip->rb.areg = ip->instr.o_format.rc;
   }
   ip->instFlags |= InstCMOV1;
}
DECODE(CMOV2)
{
   /* CMOV stage 2: copy Rb or newRc1 to newRc2 -- put newRc1 in Ra */
   if (ip->instFlags & InstFloating) {
      ip->ra.type = RegFlt; ip->ra.areg = ip->instr.f_format.fc;
   } else {
      ip->ra.type = RegInt; ip->ra.areg = ip->instr.o_format.rc;
   }
   ip->instFlags |= InstCMOV2;
}

DECODE(Floating)
{
   int func = ip->instr.f_format.function;
   ip->instFlags = InstFloating;

   /* Check for conditional move instruction */
   if (ip->instr.common.opcode == op_fltl) {
      if ((func == fltl_fcmoveq) || (func == fltl_fcmovne) ||
          (func == fltl_fcmovlt) || (func == fltl_fcmovge) ||
          (func == fltl_fcmovle) || (func == fltl_fcmovgt)) {
         ip->instFlags |= InstCMOV;
      }
   }

   if (ip->instr.common.opcode == op_fltm) {
      ip->ra.type = RegInt; 
   } else {
      ip->ra.type = RegFlt;
   }
   ip->ra.areg = ip->instr.f_format.fa;
   ip->rb.type = RegFlt; ip->rb.areg = ip->instr.f_format.fb;
   ip->rc.type = RegFlt; ip->rc.areg = ip->instr.f_format.fc;

   ip->rdest = &ip->rc;

   /* Check for write to R31 */
   if (ip->rdest->areg == ZERO_REGISTER) {
      if (func != fltl_mt_fpcr) {
         ip->instFlags |= InstWriteZero;
      }
   }

   SamPrintStage(sp, ip, "ID FP");
}

DECODE(Misc)
{
   /* Don't want sign extension */
   uint func = ip->instr.m_format.memory_displacement & 0xffff;
   switch (func) {
   case misc_fetch: 
      ip->instFlags = InstPrefetch;
      ip->rb.type = RegInt; ip->rb.areg = ip->instr.m_format.rb;
      ip->rdest = NULL;
      break;
   case misc_fetch_m: 
      ip->instFlags = InstPrefetchMod;
      ip->rb.type = RegInt; ip->rb.areg = ip->instr.m_format.rb;
      ip->rdest = NULL;
      break;
   case misc_trapb:
      ip->instFlags = InstTRAPB;
      ip->rdest = NULL;
      break;
   case misc_excb:
      ip->instFlags = InstEXCB;
      ip->rdest = NULL;
      break;
   case misc_mb: 
      ip->instFlags = InstMB;
      ip->rdest = NULL;
      break;
   case misc_wmb: 
      ip->instFlags = InstWMB;
      ip->rdest = NULL;
      break;
#ifdef misc_wh64
   case misc_wh64: 
      /* rpb: treat as a no-op for now */
      ip->rdest = NULL;
      break;
#endif
   case misc_rc:
   case misc_rs:
   case misc_rpcc: 
      ip->ra.type = RegInt; ip->ra.areg = ip->instr.m_format.ra;
      ip->rdest = &ip->ra;
      break;
   default:
      SamDecodeIllegal(sp, ip);
      ip->rdest = NULL;
      return;
   }
   
   SamPrintStage(sp, ip, "ID MISC");
}

DECODE(Illegal)
{
   ip->instFlags |= InstIllegal;

   ip->rdest = NULL;
   SamPrintStage(sp, ip, "ID ILL");
}

void 
SamDecode(SamState *sp, SamIwinEntry *ip)
{
   int opcode;
   
   /* Initialize instruction state */
   ip->ra.type = ip->rb.type = ip->rc.type = RegNone;
   ip->ra.reg = ip->rb.reg = ip->rc.reg = NULL;
   ip->instFlags = 0;
   ip->iwinFlags = 0;

   /* Decode the instruction */
   opcode = ip->instr.common.opcode;
   (samInstructionTable[opcode].decode)(sp, ip);
}
