/*
 * execute.c -- Sam pipeline, execution stage
 */

#include "sam.h"
#include "print.h"
#include "funcunit.h"
#include "instruction.h"
#include "ev5.h"
#include "ev5_ipr.h"

void
SamExecute(SamState *sp)
{
   List_Links *lp;
   LIST_FORALL(sp->executeQueue, lp) {
      int opcode;
      SamIwinEntry *ip = ListToIwin(lp);
      if (ip->completeTime > samCurrentTime) {
         return;
      }

#ifdef ZERO_DEBUG
      ASSERT(ip->thread->gprName[ZERO_REGISTER]->contents.gpr == 0);
#endif
      
      ip->fpcr = sp->PE->fpcr;
      sp->PE->PC = ip->pc;
      sp->PE->instr = ip->instr;
      sp->executing = ip;
      opcode = ip->instr.common.opcode;
      if (ip->instFlags & InstIllegal) {
         SamExecuteIllegal(sp, ip);
      } else {
         (samInstructionTable[opcode].execute)(sp, ip);
      }
      ip->iwinFlags |= IwinExecuted;
      
      if (SamTrapOccurred(sp)) {
         SamTrapCopy(sp->lastTrap, ip->trap);
         ip->iwinFlags |= IwinFault;
         ip->thread->stallFetch |= StallTrap;
      }
      
      /* If this was not a ldst fault and not an MFPR 
       * then write back registers.
       * Note: MFPR has side-effects; needs to be executed
       *       at grad time, at which point reg writeback 
       *       can be done.
       */
      if (!(ip->iwinFlags & (IwinFault | IwinLdst)) 
          && (PRIV_OP_MFPR != ip->instr.common.opcode)) 
      {
         SamRegWriteback(sp, ip);
      }

      /* Note: the LIST_FORALL still works because 
       * lp->nextPtr is unchanged by the List_Remove operation. */
      List_RemoveDebug(lp);
   }
}

void
SamExecuteAddressCalculation(SamState *sp, SamIwinEntry *ip)
{
   Reg base;
   int disp;
   MMUStatus status;
   /* EV5_DTranslateVirtual expects 0 or 1, not just false or true */
   bool writing = (ip->instFlags & InstStore) ? 1 : 0;

   /* Prefetches on EV6 can generate exceptions */
   bool trapNow = TRUE;
      
   base = ip->rb.reg->contents.gpr;
   disp = ip->instr.m_format.memory_displacement;
   ip->vAddr = base + disp;
   
   sp->PE->PC = ip->pc;
   sp->PE->instr = ip->instr;
   status = EV5_DTranslateVirtual(sp->PE, ip->vAddr, writing, 
                                  trapNow, &ip->pAddr);
   
   if (status == MMU_EXCEPTION) {
      ip->pAddr = (PA) -1;
      if (trapNow) {
         ip->iwinFlags |= IwinTBFault;
         SamTrapCopy(sp->lastTrap, ip->trap);
      }
   } else if (status == MMU_UNCACHED) {
      ip->iwinFlags |= IwinUncached;
   }
}

void
SamExecuteHWAddressCalculation(SamState *sp, SamIwinEntry *ip)
{
   Reg base;
   int disp;
   MMUStatus status;
   /* EV5_DTranslateVirtual expects 0 or 1, not just false or true */
   bool writing = (ip->instFlags & InstStore) ? 1 : 0;

   int size;
   ev5_hwldst_format hwldst;
   hwldst.word = ip->instr.word;
   
   base = ip->rb.reg->contents.gpr;
   disp = hwldst.hwmem_format.disp;
   ip->vAddr = base + disp;
   
   sp->PE->PC = ip->pc;
   sp->PE->instr = ip->instr;

   /* Check if we've incorrectly speculated */
   if (!IS_PAL(sp->PE)) {
      ip->instFlags |= InstIllegal;
      ip->pAddr = (PA) -1;
      return;
   }

   if (writing) {
      status = EV5_HW_Store(sp->PE, ip->instr.word, base, &ip->pAddr, &size);
   } else {
      status = EV5_HW_Load(sp->PE, ip->instr.word, base, &ip->pAddr, &size);
   }
   if (hwldst.hwmem_format.lock_cond) {
      CPUWarning("HW_LD LoadLock: cycle=%ld pc=%lx vAddr=%lx pAddr=%lx",
                 samCurrentTime, ip->pc, ip->vAddr, ip->pAddr);
   }
   
   switch (status) {
   case MMU_SUCCESS:
      ASSERT(ip->pAddr < MEM_SIZE(M_FROM_CPU(sp->PE->myNum)));
      break;
   case MMU_UNCACHED:
      ip->vAddr = (ip->pAddr & PA_MASK) | K1BASE;
      ip->iwinFlags |= IwinUncached;
      break;
   case MMU_EXCEPTION:
      ip->iwinFlags |= IwinTBFault;
      SamTrapCopy(sp->lastTrap, ip->trap);
      /* Mimicking code in EV5_HW_Load */
      if (writing) {
         ASSERT(hwldst.hwmem_format.vpte == 0);
      } else {
         ASSERT(hwldst.hwmem_format.vpte);
         ASSERT(ip->trap.trap == TRAP_NDTB_MISS);
         ip->trap.trap = TRAP_PDTB_MISS;
      }
      break;
   case MMU_LL:
   case MMU_SC:
      ASSERT(0);
      break;
   }
}
