/*
 * fetch.c -- Sam pipeline, instruction fetch stage
 */

#include "sam.h"
#include "print.h"
#include "qc.h"
#include "ev5.h"
#include "cpu_stats.h"
#include "addr_layout.h"
#include "hw_events.h"

static int SamRename(SamState *sp, SamIwinEntry *ip);

void
SamFetch(SamState *sp)
{
   int count = 0;
   PA pAddr = 0;
   MA mAddr = 0;
   int ret;
   Result result;
   int bubble = 0;

   int cpuNum = sp->PE->myNum;

   if (!SamTrapPending(sp) && (sp->cpuSwitch == NoSwitch)) {
      while (count < FETCH_WIDTH && (!SAM_IN_ORDER || (sp->issueSlots > 0))) {
         MMUStatus status;
         SamIwinEntry *ip;
         SamThreadState *th;
         SamGetFetchThread(th, sp);
         
         /* Check if we're ready to restart after a branch misprediction */
         if ((th->stallFetch & StallMispredict) && 
             (samCurrentTime == th->mispredictRestartCycle)) {
            th->stallFetch &= ~StallMispredict;
         }

         /* Check if instruction fetch is stalled */
         if (th->stallFetch != StallNone) {
            if (th->stallFetch & StallICache) {
               bubble = E_PIPE_ICACHE;
            } else if ((th->stallFetch & StallTrap)       ||
                       (th->stallFetch & StallITB)) {
               bubble = E_PIPE_TRAP;
            } else if ((th->stallFetch & StallREI)        ||
                       (th->stallFetch & StallBarrier)    ||
                       (th->stallFetch & StallSC)) {
               bubble = E_PIPE_DRAIN;
            } else if ((th->stallFetch & StallMispredict) ||
                       (th->stallFetch & StallBranch)     ||
                       (th->stallFetch & StallThread)) {
               bubble = E_PIPE_MISPREDICT;
            }
            break;
         }
         
         SamGetFreeIwinEntry(ip, sp);
         /* Check if instruction window is full */
         if (ip == NULL) {
            /* No bubble -- pipeline is full */
            break;
         }
         
         /* PC address translation */
         
         sp->PE->PC = ip->pc = th->pc;
         sp->PE->instr = ip->instr;
         
         if (IS_PAL(sp->PE)) {
            /* PALCode: fetched from physical memory */
            pAddr = th->pc & ~0x3;
            mAddr = PHYS_TO_MEMADDR(cpuNum, pAddr);
         } else if ((mAddr = QCLookup(sp->PE->curIQC, th->pc, 0)) != NULL) {
            pAddr = MEMADDR_TO_PHYS(M_FROM_CPU(cpuNum), mAddr);
         } else {
            status = EV5_ITranslateVirtual(sp->PE, &pAddr, TRUE);
            /* Check if address translation failed */
            if (status != MMU_SUCCESS) {
               ASSERT(status == MMU_EXCEPTION);
               th->stallFetch |= StallITB;
               SamTrapCopy(sp->lastTrap, th->trap);
               bubble = E_PIPE_TRAP;
               goto abort;
            }
            mAddr = PHYS_TO_MEMADDR(M_FROM_CPU(cpuNum), pAddr);
         }         
         ASSERT(mAddr != 0);
         
         /* Fetch instruction from the memory system */
         result = SamRefICache(sp, th->pc, pAddr, &ip->instr.word);
         
         /* Check if instruction was found in the I$ */
         switch (result) {
         case SUCCESS:
            STATS_INC(cpuNum, iReads, 1);
            break;
         case STALL:
            th->stallFetch |= StallICache;
            bubble = E_PIPE_ICACHE;
            goto abort;
         case FAILURE:
            /* Cache wouldn't accept request - try again next cycle. */
            bubble = E_PIPE_ICACHE;
            goto abort;
         case BUSERROR:
         default:
            /* rpb: what's the correct trap to raise here? */
            CPUVec.Exception(sp->PE->myNum, TRAP_UNALIGN);
            SamTrapCopy(sp->lastTrap, th->trap);
            th->stallFetch |= StallTrap;
            bubble = E_PIPE_TRAP;
            goto abort;
         }
         
         SamDecode(sp, ip);

         /* If this is a branch, make sure there's a thread available */
         if (ip->instFlags & InstBranch) {
            if ((NUM_THREADS > 1)  && !SamThreadIsAvailable(sp)) {
               th->stallFetch |= StallThread;
               bubble = E_PIPE_OTHER;
               goto abort;
            }
         }
         
         /* Special case for conditional move (3-operand) instructions: 
          * they must be decomposed into two instructions */
         if (ip->instFlags & InstCMOV) {
            if (th->cmov2Fetch) {
               SamDecodeCMOV2(sp, ip);
            } else {
               SamDecodeCMOV1(sp, ip);
            }
         }

         ret = SamRename(sp, ip);
         /* Check if rename succeeded */
         if (ret < 0) {
            bubble = E_PIPE_OTHER;
            goto abort;
         }
         
         /* Rename succeeded: update thread and machine state */
         
         th->cmov2Fetch = ((ip->instFlags & InstCMOV) && !th->cmov2Fetch);
         if (!th->cmov2Fetch) {
            th->pc += 4;
         }

         /* Set instruction window flags */
         if (ip->instFlags & InstMemory) {
            ip->iwinFlags |= IwinLdst;
         }
         if (ip->instFlags & (InstIPR|InstMemory)) {
            ip->iwinFlags |= IwinDepScbd;
         }
         if (ip->instFlags & InstREI) {
            ip->iwinFlags |= IwinFlush;
         }

         ip->ls = NULL;
         ip->depCount = 0;
         ip->bubble = 0;

         ip->thread = th;
         ip->iwinFlags |= IwinBusy;
         SamPrintFetch(sp, ip);
         SamPrintIwinEntry(sp, ip);
         SamPrintStage(sp, ip, "IF");
         
         if (th->status == ThSpec) {
            ip->iwinFlags |= IwinSpec;
         } else {
            SamPrintStage(sp, ip, "NSQ");
         }
         
         if (ip->instFlags & InstBranch) {
            SamBranch(sp, ip);
         } else if (ip->instFlags & InstSC) {
            th->stallFetch |= StallSC;
         } else if (ip->instFlags & (InstMB | InstWMB)) {
            ip->vAddr = ip->pAddr = 0;
         } else if (ip->instFlags & (InstEXCB | InstTRAPB)) {
            /* rpb: conservatively stall pipeline on EXCB/TRAPB for now */
            th->stallFetch |= StallBarrier;
         } else if (ip->instFlags & InstREI) {
            th->stallFetch |= StallREI;
         } else if (ip->instFlags & InstIllegal) {
            th->stallFetch |= StallTrap;
         }

         /* Short-circuit path for instructions that write R31/F31
          * without side-effects: skip issue/execute stages completely */
         if (ip->instFlags & InstWriteZero) {
            ip->iwinFlags |= IwinAvail | IwinIssued | IwinExecuted;
            SamPrintStage(sp, ip, "AV");
            SamPrintStage(sp, ip, "IS");
            SamPrintStage(sp, ip, "EX W31");
            SamRegWriteback(sp, ip);
         } else {
            SamCheckInstAvail(ip, sp);
         }

         /* Add instruction to issue/reorder/ldst queues */
         List_InsertDebug(IwinToReorderList(ip), LIST_ATREAR(sp->reorderBuffer));
         /* Check for short-circuit execution:
          * write to R31 or unconditional immediate branch */
         if (!(ip->iwinFlags & IwinIssued)) {
            SamIssueEnqueue(ip, sp);
         }
         if (ip->instFlags & InstMemory) {
            List_InsertDebug(IwinToLdstList(ip), LIST_ATREAR(sp->ldstQueue));
         }

         count++;
         continue;
         
   abort:
         /* Instruction fetch/rename failed: return iwin entry to freelist */
         List_InsertDebug(IwinToList(ip), LIST_ATFRONT(sp->freeIwinEntries));
         break;
      }
   }
   HistogramAdd(sp->fetchHistogram, count);

   if (bubble != 0) {
      if (List_IsEmpty(sp->reorderBuffer)) {
         sp->bubble = bubble;
      } else {
         SamIwinEntry *ip = ReorderListToIwin(List_Last(sp->reorderBuffer));
         ip->bubble = bubble;
      }
   }
}

static int
SamRename(SamState *sp, SamIwinEntry *ip)
{
   SamRegState *ra, *rb, *rp;
   SamThreadState *th;
   SamGetFetchThread(th, sp);

   /* Initialize the register information */

   ip->ra.reg = ip->rb.reg = ip->rc.reg = NULL;
   ip->ra.oldreg = ip->rb.oldreg = ip->rc.oldreg = NULL;
   ra = rb = rp = NULL;

   /* Get active names of source registers */

   if ((ip->ra.type != RegNone) && (ip->rdest != &ip->ra)) {
      switch (ip->ra.type) {
      case RegInt:
         ra = th->gprName[ip->ra.areg];
         break;
      case RegFFlt: case RegGFlt: case RegSFlt: case RegTFlt: case RegFlt:
         ra = th->fprName[ip->ra.areg];
         break;
      default: 
         ASSERT(0);
         break;
      }
      ASSERT(ra != NULL);
   }

   if ((ip->rb.type != RegNone) && (ip->rdest != &ip->rb)) {
      switch (ip->rb.type) {
      case RegInt:
         rb = th->gprName[ip->rb.areg];
         break;
      case RegFFlt: case RegGFlt: case RegSFlt: case RegTFlt: case RegFlt:
         rb = th->fprName[ip->rb.areg];
         break;
      default: 
         ASSERT(0);
         break;
      }
      ASSERT(rb != NULL);
   }

   /* Acquire a free register for the target register */

   if (ip->rdest != NULL) {
      if (ip->rdest->areg == ZERO_REGISTER) {
         /* Only need one copy of the zero register -- treat like source */
         if (ip->rdest->type == RegInt) {
            rp = th->gprName[ZERO_REGISTER];
         } else {
            rp = th->fprName[ZERO_REGISTER];
         }
         ASSERT(rp->status & (RegBusy | RegMapped));
         ip->rdest->oldreg = NULL;
      } else {
         if (ip->rdest->type == RegInt) {
            SamGetFreeGPR(rp, sp);
         } else {
            SamGetFreeFPR(rp, sp);
         }

         /* Check if there was a register available */
#ifdef RENAME_DEBUG
         {
            static SimTime renameStallTime[SIM_MAXCPUS];
            if (rp == NULL) {
               if (++renameStallTime[sp->PE->myNum] > 10000) {
                  CPUError("Out of rename registers on CPU %d\n", sp->PE->myNum);
               }
            } else {
               renameStallTime[sp->PE->myNum] = 0;
            }
         }
#endif
         if (rp == NULL) {
            return -1;
         }

         if (ip->rdest->type == RegInt) {
            ip->rdest->oldreg = th->gprName[ip->rdest->areg];
            th->gprName[ip->rdest->areg] = rp;
         } else {
            ip->rdest->oldreg = th->fprName[ip->rdest->areg];
            th->fprName[ip->rdest->areg] = rp;
         }
         
         SamAcquireReg(rp);
         rp->owner = ip;
         rp->status |= RegBusy | RegInWin;
         ip->iwinFlags |= IwinDefine;
      }
      ip->rdest->reg = rp;
   }

   /* Rename succeeded: acquire source registers, release old copy of dest */

   if (ra != NULL) {
      ASSERT(ra->status & (RegBusy | RegMapped));
      ra->ref++;
      if ((ip->ra.areg != ZERO_REGISTER) && (ra->status & RegInWin)) {
         /* Add instruction to dependency list of register owner */
         SamIwinEntry *owner = ra->owner;
         if (owner != ip) {
            int i = owner->depCount++;
            owner->depList[i] = ip;
            ip->iwinFlags |= IwinDepA;
            ip->ra.dep = owner;
         }
      }
      ip->ra.reg = ra;
   }

   if (rb != NULL) {
      ASSERT(rb->status & (RegBusy | RegMapped));
      rb->ref++;
      if ((ip->rb.areg != ZERO_REGISTER) && (rb->status & RegInWin)) {
         /* Add instruction to dependency list of register owner */
         SamIwinEntry *owner = rb->owner;
         if (owner != ip) {
            int i = owner->depCount++;
            owner->depList[i] = ip;
            ip->iwinFlags |= IwinDepB;
            ip->rb.dep = owner;
         }
      }
      ip->rb.reg = rb;
   }

   if ((ip->rdest != NULL) && (ip->rdest->oldreg != NULL)) {
      SamReleaseReg(ip->rdest->oldreg, ip->rdest->type, sp);
   }
   
   return 0;
}
