/*
 * grad.c -- Sam pipeline, graduation stage
 */

#include "sam.h"
#include "print.h"

#include "cpu_stats.h"
#include "hw_events.h"
#include "memstat.h"
#include "print_insts.h"
#include "ev5.h"
#include "ev5_ipr.h"

/* Sanity check -- die if graduation stalls for too long */
static SimTime lastGraduated[SIM_MAXCPUS];
const SimTime SanityCheckCycles = 100000;

static void
SamCompleteInst(SamState *sp, SamIwinEntry *ip, bool trap)
{
   bool squash = ip->iwinFlags & IwinSquash;

   ASSERT(ip->ls == NULL);

   if (ip->iwinFlags & IwinDefine) {
      SamRegState *rp;

      /* Update AlphaState register contents (for GetReg) */
      if (!trap && !squash) {
         int areg = ip->rdest->areg;
         if (areg != ZERO_REGISTER) {
            if (ip->rdest->type == RegInt) {
               sp->PE->reg[areg] = ip->rdest->reg->contents.gpr;
            } else {
               sp->PE->fp[areg] = ip->rdest->reg->contents.fpr;
            }
         }
      }

      /* Undo register rename if we're taking an exception */
      if (trap && !squash) {
         SamAcquireReg(ip->rdest->oldreg);
         SamReleaseReg(ip->rdest->reg, ip->rdest->type, sp);
         if (ip->rdest->type == RegInt) {
            ASSERT(ip->thread->gprName[ip->rdest->areg] == ip->rdest->reg);
            ip->thread->gprName[ip->rdest->areg] = ip->rdest->oldreg;
         } else {
            ASSERT(ip->thread->fprName[ip->rdest->areg] == ip->rdest->reg);
            ip->thread->fprName[ip->rdest->areg] = ip->rdest->oldreg;
         }
      }

      if (trap || squash) {
         /* Free the current name for this register */
         rp = ip->rdest->reg;
      } else {
         /* Free the previous name for this register */
         rp = ip->rdest->oldreg;
      }
      ASSERT(rp != NULL);
      SamFreeReg(rp, ip->rdest->type, sp);
   }

   /* Back out the thread's PC if we're taking an exception */
   if (trap && !squash) {
      ip->thread->pc = ip->pc;
   }

   if (!trap && !squash) {
      /* Scoreboard bits [7:4] are cleared when instruction is retired */
      if (ip->iwinFlags & IwinScbdHi) {
         byte scbd = ~(ip->instr.word & 0xf0);
         ASSERT(ip->instFlags & InstIPR);
         sp->scoreboard &= ~scbd;
         ip->iwinFlags &= ~IwinScbdHi;
      }
      ASSERT(!(ip->iwinFlags & IwinScbdLo));
   }

   /* Update pipeline stall statistics */
   if (sp->stallSlots > 0) {
      int type = SamStallType(ip, sp);
      MemStatPipelineStall(sp->PE->myNum, ip->pc, 0, sp->stallSlots, type);
      sp->stallSlots = 0;
      if (!squash) {
         sp->bubble = ip->bubble;
      }
   }

   /* Remove entry from reorder buffer and return it to free list */
   List_RemoveDebug(IwinToReorderList(ip));
   List_InsertDebug(IwinToList(ip), LIST_ATREAR(sp->freeIwinEntries));

   lastGraduated[sp->PE->myNum] = 0;
}

void
SamGraduate(SamState *sp)
{
   int count = 0;
   int squashCount = 0;
   int cpuNum = sp->PE->myNum;
   SamIwinEntry *ip = NULL;

   ++lastGraduated[cpuNum];
   VASSERT((lastGraduated[cpuNum] < SanityCheckCycles),
           ("ERROR: lastGraduated[%d]=%d >= SanityCheckCycles=%d [%d]\n",
            cpuNum, lastGraduated[cpuNum], SanityCheckCycles, samCurrentTime));

   while (count < GRAD_WIDTH) {
      /* Check for and handle any pending traps */
      SamTrap(sp);

      /* Bail if there are no instructions in the pipeline */
      if (List_IsEmpty(sp->reorderBuffer)) {
         ip = NULL;
         break;
      }

      if (SamTrapPending(sp)) {
         ip = ReorderListToIwin(List_Last(sp->reorderBuffer));
         if (!(ip->iwinFlags & IwinSquash)) {
            SamFlushInst(sp, ip);
         }
         SamPrintStage(sp, ip, "EXC");
         SamPrintIwinEntry(sp, ip);
         SamCompleteInst(sp, ip, TRUE);
         /* Don't count trap-flushed instruction towards graduation slot count */
      } else {
         bool freed, squash, trap;
         ip = ReorderListToIwin(List_First(sp->reorderBuffer));

         squash = ip->iwinFlags & IwinSquash;
         trap = (ip->iwinFlags & IwinFault) && !(ip->iwinFlags & IwinSpec);

         /* Execute MFPR instruction. (It has side-effects and
          * so should not performed early or speculatively.) 
          */
         if (!squash && 
             ((ip->iwinFlags & IwinExecuted) && 
              (ip->instr.common.opcode == PRIV_OP_MFPR))) {
            int priv;
            ASSERT(!trap); /* This instruction should not trap. */
            priv = ip->instr.m_format.memory_displacement;
            ip->ra.reg->contents.gpr = EV5_MoveFromPriv(sp->PE, priv);
            SamRegWriteback(sp, ip);
            SamPrintStage(sp,ip,"MFPR GR");
         }

         freed = ip->iwinFlags & IwinFreed;
         
         if (squash && (freed || trap)) {
            ASSERT(!(ip->iwinFlags & IwinLdst) || !(ip->iwinFlags & IwinLdstBuf));
            SamPrintStage(sp, ip, "GR squashed");
            SamPrintIwinEntry(sp, ip);
            SamCompleteInst(sp, ip, FALSE);
            count++;
            squashCount++;
         } else if (trap) {
            /* Instruction took a trap */
            ASSERT(!SamTrapPending(sp));
            SamTrapCopy(ip->trap, sp->currentTrap);
            /* Instruction is used in EV5_Trap for arithmetic traps */
            sp->PE->instr = ip->instr;
            /* Update the FPCR if necessary to reflect the trap */
            if (ip->trap.trap == TRAP_ARITH) {
#ifdef FPCR_DEBUG
               CPUWarning("Arithmetic trap at 0x%lx: FPCR 0x%lx -> 0x%lx\n",
                          ip->pc, sp->PE->fpcr, ip->fpcr);
#endif
               sp->PE->fpcr = ip->fpcr;
            }
            /* Print this instruction if it's a call_pal */
            if (ip->instr.common.opcode == op_call_pal) {
               STATS_INC(cpuNum, numInstructions, 1);
               PRINT_INSTRUCTION(sp->PE, ip->pc, ip->instr.word);
            }
            /* Leave instruction in reorder buffer to be flushed */
            continue;
         } else if (freed) {
            if (ip->iwinFlags & IwinLdst) {
               if (ip->iwinFlags & IwinLdstBuf) {
                  /* Check if load/store instruction can be retired */
                  int ret = SamLdstBufferRetire(sp, ip->ls);
                  if (ret > 0) {
                     /* Either request hasn't issued to memory yet or 
                      * we're stalled on a cache miss; retry next cycle. */
                     break;
                  } else if (ret < 0) {
                     /* Line was nuked from the cache -- backup and retry */
                     SamTrapRecord(sp, TrapReplay);
                     SamTrapCopy(sp->lastTrap, ip->trap);
                     ip->iwinFlags |= IwinFault;
                     /* Next time through loop, instruction will trap */
                     continue;
                  } else {
                     /* Memory request completed: unstall memory issue */
                     sp->replayStatus = Issue;
                  }
               }
            }
            
            /* Restart instruction fetch if necessary */
            if (ip->instFlags & InstSC) {
               ip->thread->stallFetch &= ~StallSC;
            } else if (ip->instFlags & (InstEXCB | InstTRAPB)) {
               ip->thread->stallFetch &= ~StallBarrier;
            } else if (ip->instFlags & InstREI) {
               ip->thread->stallFetch &= ~StallREI;
            }

            /* Instruction is ready to graduate */
            SamPrintStage(sp, ip, "GR");
            SamPrintIwinEntry(sp, ip);
            
            SamCompleteInst(sp, ip, FALSE);
            count++;

            /* Update FPCR if necessary */
            if (ip->fpcr != sp->PE->fpcr) {
#ifdef FPCR_DEBUG
               CPUWarning("New FPCR value at 0x%lx: FPCR 0x%lx -> 0x%lx\n",
                          ip->pc, sp->PE->fpcr, ip->fpcr);
#endif
               sp->PE->fpcr = ip->fpcr;
            }

            /* Commit IPR change for MTPR instruction */
            if (ip->instr.common.opcode == PRIV_OP_MTPR) {
               int priv = ip->instr.m_format.memory_displacement;
               sp->PE->PC = ip->pc;
               sp->PE->instr = ip->instr;
               EV5_MoveToPriv(sp->PE, priv, ip->ra.reg->contents.gpr);
               ASSERT(!SamTrapPending(sp));
               freeTrapState(sp->currentTrap.state);
               sp->currentTrap.state = allocTrapState(sp->PE->trapState);
            }

            /* Don't consider CMOV1 to be a completed instruction */
            if (!(ip->instFlags & InstCMOV1)) {
               sp->PE->instr = ip->instr;
               sp->PE->PC = ip->pc;
               
               /* Update instruction statistics */
               STATS_INC(cpuNum, numInstructions, 1);
               if (STATS_VALUE(cpuNum, numInstructions) > 
                   STATS_VALUE(cpuNum, nextInstrSample)) { 
                  INST_SAMPLE_EVENT(samCurrentTime, cpuNum, ip->pc);
                  STATS_INC(cpuNum, nextInstrSample, MS_SAMPLE_INSTR_INTERVAL); 
               }

#ifdef PRINT_INSTRUCTIONS
               if (printInstMask[cpuNum]) {
#else
               if (0) {
#endif
                  /* rpb hack: tweak register contents for correct printout of mtpr */
                  SamRegState *ra = NULL;
                  SamRegState *rb = NULL;
                  if (ip->ra.reg != NULL) {
                     if (ip->ra.type == RegInt) {
                        ra = ip->thread->gprName[ip->ra.areg];
                        ip->thread->gprName[ip->ra.areg] = ip->ra.reg;
                     } else {
                        ra = ip->thread->fprName[ip->ra.areg];
                        ip->thread->fprName[ip->ra.areg] = ip->ra.reg;
                     }
                  }
                  if ((ip->rb.reg != NULL) && (ip->rb.areg != ip->ra.areg)) {
                     if (ip->rb.type == RegInt) {
                        rb = ip->thread->gprName[ip->rb.areg];
                        ip->thread->gprName[ip->rb.areg] = ip->rb.reg;
                     } else {
                        rb = ip->thread->fprName[ip->rb.areg];
                        ip->thread->fprName[ip->rb.areg] = ip->rb.reg;
                     }
                  }
                  PRINT_INSTRUCTION(sp->PE, ip->pc, ip->instr.word);
                  if (ip->ra.reg != NULL) {
                     if (ip->ra.type == RegInt) {
                        ip->thread->gprName[ip->ra.areg] = ra;
                     } else {
                        ip->thread->fprName[ip->ra.areg] = ra;
                     }
                  }
                  if ((ip->rb.reg != NULL) && (ip->rb.areg != ip->ra.areg)) {
                     if (ip->rb.type == RegInt) {
                        ip->thread->gprName[ip->rb.areg] = rb;
                     } else {
                        ip->thread->fprName[ip->rb.areg] = rb;
                     }
                  }
               }

               /* Check for PC annotations */
               RunPCAnnotations(ip->pc, 0);

               /* Check for load/store annotations */
               if ((ip->instFlags & InstLoad) && annLoads) {
                  AnnPtr ptr;
                  if ((ptr = AnnFMLookup(ip->vAddr, ANNFM_LD_TYPE)) != NULL) {
                     AnnExec(ptr);
                  }
                  if ((ptr = AnnFMLookup(ip->pAddr, ANNFM_LD_TYPE)) != NULL) {
                     AnnExec(ptr);
                  }
               }
               if ((ip->instFlags & InstStore) && annStores) {
                  /* Don't trigger store annotation on a failed SC */
                  if (!(ip->instFlags & InstSC) || (ip->rdest->reg->contents.gpr == 1)) {
                     AnnPtr ptr;
                     if ((ptr = AnnFMLookup(ip->vAddr, ANNFM_ST_TYPE)) != NULL) {
                        AnnExec(ptr);
                     }
                     if ((ptr = AnnFMLookup(ip->pAddr, ANNFM_ST_TYPE)) != NULL) {
                        AnnExec(ptr);
                     }
                  }
               }
            }
#ifdef PRINT_INSTRUCTIONS
            if (printInstMask[cpuNum]) {
#else
            if (0) {
#endif
#ifdef MEM_DEBUG
               if (ip->instFlags & InstLoad) {
                  CPUPrint("MEM LD: pc=0x%lx va=0x%lx val=0x%lx\n",
                           ip->pc, ip->vAddr, ip->rdest->reg->contents);
               } else if (ip->instFlags & InstStore) {
                  CPUPrint("MEM ST: pc=0x%lx va=0x%lx val=0x%lx\n",
                           ip->pc, ip->vAddr, ip->ra.reg->contents);
               }
#endif
#ifdef RDEST_DEBUG
               if (ip->rdest != NULL) {
                  CPUPrint("RDEST: R%d = 0x%llx\n", ip->rdest->areg, ip->rdest->reg->contents.gpr);
               }
#endif
            }
         } else {
            /* Instruction not eligible to graduate: done for this cycle */
            break;
         }
      }
   }

   /* Currently not recording per-PC squash information */
   if (squashCount > 0) {
      MemStatPipelineStall(cpuNum, 0, 0, squashCount, E_PIPE_MISPREDICT);
   }

   /* Don't count squashed instructions in the graduation histogram */
   HistogramAdd(sp->gradHistogram, count - squashCount);

   /* Stall time bookkeeping */
   sp->stallSlots += GRAD_WIDTH - count;
}

