/*
 * issue.c -- Sam pipeline, issue stage
 */

#include "sam.h"
#include "print.h"

static int issueCount[SIM_MAXCPUS];

void
SamIssue(SamState *sp)
{
   int u;
   int count = 0;
   int cpuNum = sp->PE->myNum;
   List_Links *lp = List_First(sp->issueQueue);
   SamIwinEntry *ip = NULL;

   /* In-order: only consider instructions in the current instruction block */
   int issueWindow = ISSUE_WIDTH - issueCount[cpuNum];

   if (!SamTrapPending(sp)) {
      while ((count < ISSUE_WIDTH) && !List_IsAtEnd(sp->issueQueue, lp)) {
         List_Links *nlp = List_Next(lp);
         ip = ListToIwin(lp);
         ASSERT(ip->iwinFlags & IwinBusy);

         if (SAM_IN_ORDER && (issueWindow-- == 0)) break;
         
         /* Check load/store instructions against scoreboard */

         if ((ip->iwinFlags & IwinLdst) && (ip->iwinFlags & IwinDepScbd)) {
            if ((sp->scoreboard & 0x0f) == 0x00) {
               ip->iwinFlags &= ~IwinDepScbd;
               SamCheckInstAvail(ip, sp);
            }
         }
         
         /* Check MTPR/MFPR instructions against scoreboard */

         if (ip->instFlags & InstIPR) {
            byte scbd = ~(ip->instr.word & 0xff);
            if (ip->iwinFlags & IwinDepScbd) {
               if ((sp->scoreboard & scbd) == 0x00) {
                  ip->iwinFlags &= ~IwinDepScbd;
                  SamCheckInstAvail(ip, sp);
               }
            }
            /* Halt issue if this instruction isn't ready to issue */
            if (!(ip->iwinFlags & IwinAvail)) {
               goto haltIssue;
            } else {
               ip->iwinFlags |= IwinScbdHi | IwinScbdLo;
               sp->scoreboard |= scbd;
            }
         } 

         /* Check instructions that require pipeline flush */

         if (ip->iwinFlags & IwinFlush) {
            if (List_First(sp->reorderBuffer) == IwinToReorderList(ip)) {
               ip->iwinFlags &= ~IwinFlush;
               SamCheckInstAvail(ip, sp);
            } else {
               goto haltIssue;
            }
         }

         /* Skip instruction if it isn't available to issue */
             
         if (!(ip->iwinFlags & IwinAvail)) goto noIssue;
         
         /* Check if instruction should be squashed */
         
         if (ip->iwinFlags & IwinSquash) {
            if (ip->iwinFlags & IwinLdst) {
               List_RemoveDebug(IwinToLdstList(ip));
            }
            SamFreeInst(sp, ip, TRUE);
            SamIssueDequeue(ip, sp);
            goto issue;
         }

         /* Check if memory operation is free to issue */

         if (ip->iwinFlags & IwinLdst) {
            if (sp->replayStatus == Wait) {
               goto noIssue;
            } else if (List_First(sp->ldstQueue) != IwinToLdstList(ip)) {
               goto noIssue;
            } else if (ip->instFlags & (InstMB | InstWMB)) {
               if (!List_IsEmpty(sp->activeLdstEntries)) {
                  goto noIssue;
               }
            }
         }

         /* Check if there is a functional unit available */

         for (u = 0; u < sp->numFuncUnits; u++) {
            int latency;
            int repeatRate;
            SamFuncUnit *unit = &sp->funcUnit[u];
            bool willing = (unit->CanExecute)(ip, &latency, &repeatRate);
            bool able = (unit->availTime <= samCurrentTime);
            if (willing && able) {
               List_Links *elp;
               
               /* Reserve a load/store buffer slot if necessary */
               
               if (ip->iwinFlags & IwinLdst) {
                  if (!SamLdstBufferReserve(sp, ip)) {
                     /* No room in load/store buffer -- can't issue */
                     goto noIssue;
                  } else {
                     List_RemoveDebug(IwinToLdstList(ip));
                     /* If this is a replay, hold off subsequent memory ops */
                     if (sp->replayStatus == Replay) {
                        sp->replayStatus = Wait;
                     }
                  }
               }
               
               /* Execute instruction and remove it from issue queue */

               ip->completeTime = samCurrentTime + latency;
               unit->availTime = samCurrentTime + repeatRate;
               unit->numExecuted++;
               
               /* Remove instruction from issue queue */

               ip->iwinFlags |= IwinIssued;
               SamPrintStage(sp, ip, "IS");
               SamIssueDequeue(ip, sp);
               
               /* Add to execute pqueue, sorted by completeTime */

               LIST_FORALL(sp->executeQueue, elp) {
                  SamIwinEntry *eip = ListToIwin(elp);
                  if (ip->completeTime < eip->completeTime) break;
               }
               List_InsertDebug(lp, LIST_BEFORE(elp));
               goto issue;
            }
         }
         /* No functional unit available -- fall through to noIssue */
         
   noIssue:
         lp = nlp;
         continue;

   haltIssue:
         break;

   issue:
         count++;
         issueCount[cpuNum]++;
         lp = nlp;
         if (ip->instFlags & InstIPR) {
            /* Scoreboard bits [3:0] are cleared when instruction is issued */
            byte scbd = ~(ip->instr.word & 0x0f);
            sp->scoreboard &= ~scbd;
            ip->iwinFlags &= ~IwinScbdLo;
            /* rpb hack: need to halt issue here, but I dunno why.... */
            break;
         }
      }
   }
   HistogramAdd(sp->issueHistogram, count);
   if (issueCount[cpuNum] == ISSUE_WIDTH) {
      issueCount[cpuNum] = 0;
   }
}

