/*
 * iwin.h -- the Sam instruction window structure and macros
 */

#ifndef SAM_IWIN_H
#define SAM_IWIN_H

#include <alpha/inst.h>
#include "list.h"
#include "simtypes.h"
#include "register.h"
#include "trap.h"
#include "sam_params.h"

struct SamLdstEntry_s;

/* The SamIwinEntry uses three List_Links:
 * 1. The free list and functional unit queues
 * 2. The reorder buffer
 * 3. The load/store queue
 * The first is updated as the instruction progresses in the pipeline;
 * the second is enqueued at fetch time and dequeued at graduation;
 * the third is enqueued at fetch time and dequeued when the instruction issues.
 */

typedef struct SamIwinEntry_s {
   List_Links links;        /* Freelist management -- must come first */
   List_Links reorderLinks; /* Reorder buffer management -- must come second */
   List_Links ldstLinks;    /* Load/store queue management -- must come third */

   union alpha_instruction instr;
   int iwinFlags;
   int instFlags;

   VA pc;

   /* Descriptions of the three registers for this instruction:
    * rdest indicates which register (if any) is the destination.
    * NOTE: Rc can never act as a source, only as a destination */
   struct {
      SamRegType type;              /* Integer or floating point D/F/G/S/T */
      int areg;                     /* Architectural register number */
      SamRegState *reg;             /* Pointer to the physical register */
      SamRegState *oldreg;          /* Pointer to the physical register
                                     * that used to represent 
                                     * this architectural register */
      struct SamIwinEntry_s *dep;   /* Instruction that writes this register */
   } ra, rb, rc, *rdest;

   /* Load/store address information */
   VA vAddr;
   PA pAddr;
   struct SamLdstEntry_s *ls;

   /* Branch target address information */
   VA target;

   /* List of iwin entries that are dependent on this one */
   int depCount;
   struct SamIwinEntry_s *depList[SAM_MAX_IWIN_ENTRIES];

   /* Thread that owns this entry */
   SamThreadState *thread;

   /* Execution information: cycle instruction will complete */
   SimTime completeTime;

   /* Exception caused by this instruction */
   SamTrapState trap;

   /* FPCR contents after execution of this instruction */
   Reg fpcr;

   /* Type of pipeline bubble occurring behind this instruction */
   int bubble;
} SamIwinEntry;

/* iwinFlags: Information about an active iwin entry */

#define IwinBusy        0x000001
#define IwinSpec        0x000002
#define IwinFlush       0x000004
#define IwinSquash      0x000008

#define IwinFault       0x000010
#define IwinTBFault     0x000020

#define IwinScbdHi      0x000040
#define IwinScbdLo      0x000080

#define IwinDefine      0x000100
#define IwinDepA        0x000200
#define IwinDepB        0x000400
#define IwinDepScbd     0x000800

#define IwinAvail       0x001000
#define IwinIssued      0x002000
#define IwinExecuted    0x004000
#define IwinFreed       0x008000

#define IwinLdst        0x010000
#define IwinLdstBuf     0x020000

#define IwinUncached    0x100000

/* instFlags: Information about the instruction in the iwin entry */

#define InstLoad          0x0000001
#define InstStore         0x0000002
#define InstPrefetch      0x0000004
#define InstPrefetchMod   0x0000008
#define InstSC            0x0000010

#define InstBranch        0x0000020
#define InstCall          0x0000040
#define InstReturn        0x0000080
#define InstCoroutine     0x0000100
#define InstIndirect      0x0000200
#define InstConditional   0x0000400

#define InstFloating      0x0000800

#define InstMB            0x0001000
#define InstWMB           0x0002000
#define InstTRAPB         0x0004000
#define InstEXCB          0x0008000

#define InstREI           0x0010000
#define InstIPR           0x0020000
#define InstHWMem         0x0040000
#define InstCallPAL       0x0080000

#define InstFPCR          0x0100000

#define InstCMOV          0x0200000
#define InstCMOV1         0x0400000
#define InstCMOV2         0x0800000

#define InstIllegal       0x1000000
#define InstWriteZero     0x2000000

#define InstMemory        0x000300f

/* Convert between iwin and free list pointers */

#define IwinToList(IP) ((List_Links*)(IP))
#define ListToIwin(LP) ((SamIwinEntry*)(LP))

/* Convert between iwin and reorder list pointers */
/* Need to do pointer arithmetic to get back to head of struct */

#define IwinToReorderList(IP) (&(IP)->reorderLinks)
#define ReorderListToIwin(LP) ((SamIwinEntry*)(LP-1))

/* Convert between iwin and load/store list pointers */
/* Need to do pointer arithmetic to get back to head of struct */

#define IwinToLdstList(IP) (&(IP)->ldstLinks)
#define LdstListToIwin(LP) ((SamIwinEntry*)(LP-2))

/* Claim an iwin entry from the free list */

#define SamGetFreeIwinEntry(IP, SP)                     \
{                                                       \
   List_Links *_lp = List_First((SP)->freeIwinEntries); \
   if (_lp == (SP)->freeIwinEntries) {                  \
      IP = NULL;                                        \
   } else {                                             \
      IP = ListToIwin(_lp);                             \
      List_RemoveDebug(_lp);                            \
   }                                                    \
}

/* Check if an instruction is available to issue */

#define SamCheckInstAvail(IP, SP)                                           \
   ASSERT((IP)->iwinFlags & IwinBusy);                                      \
   if (!((IP)->iwinFlags &                                                  \
       (IwinDepA|IwinDepB|IwinFlush|IwinDepScbd|IwinAvail))) {              \
      (IP)->iwinFlags |= IwinAvail;                                         \
      SamPrintStage(SP, IP, "AV");                                          \
   }

/* Manage the issue queue */

#define SamIssueEnqueue(IP, SP)                                         \
   List_InsertDebug(IwinToList(IP), LIST_ATREAR((SP)->issueQueue));     \
   (SP)->issueSlots--;

#define SamIssueDequeue(IP, SP)                 \
   List_RemoveDebug(IwinToList(IP));            \
   (SP)->issueSlots++;

/* Instruction stall classification */
#define SamStallType(IP, SP)                            \
(((SP)->bubble != 0) ? (SP)->bubble :                   \
 (((IP)->instFlags & InstFloating) ? E_PIPE_FP :        \
  (((IP)->instFlags & InstMemory) ? E_PIPE_LDST :       \
   E_PIPE_INT)))

#endif /* SAM_IWIN_H */
