/*
 * ldst.h -- the Sam load/store buffer structure and macros
 */

#ifndef SAM_LDST_H
#define SAM_LDST_H

#include "list.h"
#include "simtypes.h"
#include "register.h"
#include "iwin.h"

#define _SAM_DBG_LDST   0

typedef struct SamLdstEntry_s {
   List_Links links;            /* Freelist management -- must come first */

   int status;
   int type;
   VA  vAddr;
   PA  pAddr;
   PA  pc;
   SamRegState *reg;
   SamRegState *screg;          /* Destination register for SC */
   SamIwinEntry *ip;
   
   byte *dataPtr;
   char data[8];
   char valid[8];
   int size;
   int invalidated;
   struct SamLdstEntry_s *loadhit;

   /* Load-to-use latency modeling */
   Reg loadValue;
   SimTime loadCompleteCycle;

#if _SAM_DBG_LDST
  SimTime       start;
#endif
} SamLdstEntry;

/* status: load/store status bits */

#define LdstPending     0x0001          /* Waiting for data to be ready    */
#define LdstCached      0x0002          /* If already sent to cache        */
#define LdstConflict    0x0004          /* If conflicts with earlier entry */
#define LdstDone        0x0008          /* Operation has been completed    */
#define LdstFailed      0x0010          /* Cache rejected operation        */
#define LdstStall       0x0020          /* Cache will call back            */
#define LdstLdDelay     0x0040          /* Load-to-use latency             */
#define LdstBarrier     0x0080          /* Wait for memory barrier         */
#define LdstFilled      0x0100          /* Line filled in the cache        */

#define LdstUncached    0x0200          /* If is an uncached operation     */

/* type: size/type information about memory operation */

/* specific load/store/prefetch operations */
#define LdstLdLongword                  0x000001
#define LdstLdQuadword                  0x000002
#define LdstLdLongwordLocked            0x000004
#define LdstLdQuadwordLocked            0x000008
#define LdstLdQuadwordUnaligned         0x000010
#define LdstLdWordUnsigned              0x000020
#define LdstLdByteUnsigned              0x000040
#define LdstStLongword                  0x000100
#define LdstStQuadword                  0x000200
#define LdstStLongwordConditional       0x000400
#define LdstStQuadwordConditional       0x000800
#define LdstStQuadwordUnaligned         0x001000
#define LdstStWord                      0x002000
#define LdstStByte                      0x004000
#define LdstPrefetchShared              0x010000
#define LdstPrefetchExclusive           0x020000
#define LdstMB                          0x100000
#define LdstWMB                         0x200000

/* masks for various load/store classes */
#define LdstLoad                        0x0000ff
#define LdstStore                       0x00ff00
#define LdstLocked                      0x00000c
#define LdstConditional                 0x000c00
#define LdstUnaligned                   0x001010
#define LdstLongword                    0x000505
#define LdstQuadword                    0x001a1a
#define LdstWord                        0x002020
#define LdstByte                        0x004040
#define LdstPrefetch                    0x030000

/* Convert between ldst and list pointers */

#define LdstToList(LS) ((List_Links*)(LS))
#define ListToLdst(LP) ((SamLdstEntry*)(LP))

/* Initialize register state */

#define SamLdstInit(LS)                         \
{                                               \
   int _i;                                      \
   for (_i = 0; _i < 8; _i++) {                 \
      (LS)->valid[_i] = 0;                      \
   }                                            \
   (LS)->dataPtr = NULL;                        \
   (LS)->size = 0;                              \
   (LS)->invalidated = 0;                       \
   (LS)->loadhit = NULL;                        \
   (LS)->loadCompleteCycle = 0;                 \
}

/* Claim a load/store entry from the free list */

#define SamGetFreeLdstEntry(LS, SP)                     \
{                                                       \
   List_Links *_lp = List_First((SP)->freeLdstEntries); \
   if (_lp == (SP)->freeLdstEntries) {                  \
      LS = NULL;                                        \
   } else {                                             \
      LS = ListToLdst(_lp);                             \
      List_RemoveDebug(_lp);                            \
   }                                                    \
}

/* Claim a load/store entry from the reserved list */

#define SamGetReservedLdstEntry(LS, SP)                         \
{                                                               \
   List_Links *_lp = List_First((SP)->reservedLdstEntries);     \
   if (_lp == (SP)->reservedLdstEntries) {                      \
      LS = NULL;                                                \
   } else {                                                     \
      LS = ListToLdst(_lp);                                     \
      List_RemoveDebug(_lp);                                    \
   }                                                            \
}

/* Return an address mask corresponding to the load/store size */

#define SamGetLdstMask(LS)                      \
   (((LS)->type & LdstQuadword) ? ~((PA)7) :    \
    (((LS)->type & LdstLongword) ? ~((PA)3) :   \
     (((LS)->type & LdstWord) ? ~((PA)1) :      \
      ~((PA)0))))


#endif /* SAM_LDST_H */
