/*
 * memory.c -- Sam pipeline, memory system interface
 */

#include <string.h>
#include "sam.h"
#include "print.h"
#include "registry.h"
#include "memref.h"
#include "gamma.h"
#include "trap.h"

#include "memstat.h"
#include "hw_events.h"

/* rpb: shouldn't need this stuff here */
#include "pcache.h"
#include "scache.h"
#define dIndexMask  (DCACHE_INDEX - 1)
#define dTagShift   (log2DCACHE_SIZE - 1)

/* SamRefICache: Send instruction requests to the memory system */

Result 
SamRefICache(SamState *sp, VA vAddr, PA pAddr, Inst *inst)
{
   int index = SCACHE_INDEXOF(pAddr);
   List_Links *lp;

   /* Check if there's a write pending on this line */
   LIST_FORALL(sp->activeLdstEntries, lp) {
      SamLdstEntry *ls = ListToLdst(lp);
      if (index == SCACHE_INDEXOF(ls->pAddr)) {
         return FAILURE;
      }
   }

   return MemRefReadInst(sp->PE->myNum, vAddr, pAddr, inst);
}

/* SamRefDCache/SamRefUncached: Send data requests to the memory system */

static Result
SamRefDCache(int cpuNum, SamLdstEntry *ls, byte **dataPtr)
{
   RefSize size = DOUBLE_SZ;
   RefFlavor flavor = NO_FLAVOR;

   /* Check for an out-of-bounds physical address */
   if (ls->pAddr >= MEM_SIZE(M_FROM_CPU(cpuNum))) {
      /* rpb: what's the correct thing to do here? */
      return BUSERROR;
   }

   if (ls->type & LdstConditional) {
      flavor = SC_FLAVOR;
   } else if (ls->type & LdstLocked) {
      flavor = LL_FLAVOR;
   } else if (ls->type & (LdstStore | LdstPrefetchExclusive)) {
      flavor = GETX_FLAVOR;
   } else {
      ASSERT(ls->type & (LdstLoad | LdstPrefetchShared));
      flavor = NO_FLAVOR;
   }
   
   if (ls->type & LdstQuadword) {
      size = DOUBLE_SZ;
   } else if (ls->type & LdstLongword) {
      size = WORD_SZ;
   } else if (ls->type & LdstWord) {
      size = HALF_SZ;
   } else if (ls->type & LdstByte) {
      size = BYTE_SZ;
   } else {
      ASSERT(ls->type & LdstPrefetch);
      size = BYTE_SZ; /* irrelevant for prefetches */
   }

   /* Update AlphaState PC for MemStat bookkeeping */
   PE[cpuNum]->PC = ls->pc;

   if (ls->type & (LdstLoad | LdstPrefetchShared)) {
      return MemRefFetchShared(cpuNum, ls->vAddr, ls->pAddr, 
                               dataPtr, size, flavor);
   } else {
      return MemRefFetchExclusive(cpuNum, ls->vAddr, ls->pAddr, 
                                  dataPtr, size, flavor);
   }
}

typedef int (*MagicFunction)(int   cpuNum, /* CPU making this access */
                             VA    va,     /* accessed virtual address */
                             int   type,   /* access type */
                             void* buff);  /* buff width acc to type */

static int
SamRefUncached(int cpuNum, SamLdstEntry *ls, void *dataPtr)
{
   MagicFunction func;
   unsigned int flag;
   Result result;
   int type;

   if (!RegistryIsInRange(ls->vAddr, (void**)&func, &flag)) {
      CPUError("UNCACHED OPERATION not-in-range at vAddr=0x%lx pc=0x%lx val=0x%lx\n",
               ls->vAddr, ls->pc, ls->reg->contents);
   }
   if (!(flag & REG_FUNC)) {
      CPUError("UNCACHED OPERATION non-func at vAddr=0x%lx pc=0x%lx val=0x%lx\n",
               ls->vAddr, ls->pc, ls->reg->contents);
   }

   switch (ls->type) {
   case LdstLdQuadword:
   case LdstLdQuadwordUnaligned:
      type = BDOOR_LOAD_DOUBLE;
      break;
   case LdstLdLongword:
      type = BDOOR_LOAD_WORD;
      break;
   case LdstLdWordUnsigned:
      type = BDOOR_LOAD_HALF;
      break;
   case LdstLdByteUnsigned:
      type = BDOOR_LOAD_BYTE;
      break;

   case LdstStQuadword:
   case LdstStQuadwordUnaligned:
      type = BDOOR_STORE_DOUBLE;
      break;
   case LdstStLongword:
      type = BDOOR_STORE_WORD;
      break;
   case LdstStWord:
      type = BDOOR_STORE_HALF;
      break;
   case LdstStByte:
      type = BDOOR_STORE_BYTE;
      break;

   default:
      NOTREACHED();
      return 1;
   }

   result = (Result)(func)(cpuNum, ls->vAddr, type, dataPtr); 
   return (result == MMU_SUCCESS) ? 0 : 1;
}

/* SamLdstCheckAlignment: Report if desired access is correctly aligned */

bool
SamLdstCheckAlignment(PA pAddr, int type)
{
   if (((type & LdstQuadword) && (pAddr & 0x7)) ||
       ((type & LdstLongword) && (pAddr & 0x3)) ||
       ((type & LdstWord)     && (pAddr & 0x1))) {
      return FALSE;
   } else {
      return TRUE;
   }
}

/* SamLdstBufferReserve: Reserve a load/store buffer entry */

bool
SamLdstBufferReserve(SamState *sp, SamIwinEntry *ip)
{
   SamLdstEntry *ls;
   SamGetFreeLdstEntry(ls, sp);
   if (ls != NULL) {
      ip->ls = ls;
      sp->activeLdstBufCount++;
      return TRUE;
   } else {
      return FALSE;
   }
}

/* SamLdstBufferRelease: Release a reserved load/store buffer entry */

void
SamLdstBufferRelease(SamState *sp, SamIwinEntry *ip)
{
   SamLdstEntry *ls = ip->ls;
   ASSERT(ls != NULL);
   ip->ls = NULL;
   sp->activeLdstBufCount--;
   List_InsertDebug(LdstToList(ls), LIST_ATREAR(sp->freeLdstEntries));
}

/* SamLdstBufferConflict: Check for a conflict with all preceding entries.
 *    Algorithm: Construct a new set of cache tags for the cache set
 *    targetted by the new access. If the tag set is not large enough to
 *    accomodate the new access and all previous pending accesses or if
 *    the new access directly conflicts with a previous pending access
 *    then signal a conflict.
 *    We need to avoid conflict misses between outstanding accesses
 *    otherwise we could produce livelock.
*/

static int
SamLdstBufferConflict(SamState *sp, SamLdstEntry *ls)
{
   int cacheTags[DCACHE_ASSOC];
   int oldTag, nextTag = 0;
   int i, assoc;
   List_Links *lp;

   /* When checking for a conflict, we need to consider the cache with
    * the smallest associativity in the hierarchy. 
    *    Note: Even though we consider the cache with the smallest 
    *          associativity, we only compare pcache indices. We assume
    *          that the  scache is always larger than the pcache, so
    *          the use of pcache indices is actually conservative.
    */
   assoc = (DCACHE_ASSOC < machines.SCacheAssoc) ? DCACHE_ASSOC : machines.SCacheAssoc;

   /* Add new Ldst access to temporary cache tags */
   cacheTags[nextTag++] = DCACHE_TAG(ls->pAddr);

   for (lp = List_First(sp->activeLdstEntries);
        lp != LdstToList(ls);
        lp = List_Next(lp)) {
      SamLdstEntry *pls = ListToLdst(lp);
      if (pls->type & (LdstMB | LdstWMB)) {
         /* Barrier: no conflict */
         continue;
      }
      if (pls->status & LdstUncached) {
         /* Uncached operation: no conflict */
         continue;
      } 
      if (DCACHE_INDEXOF(pls->pAddr) != DCACHE_INDEXOF(ls->pAddr)) {
         /* Different cache index: no conflict */
         continue;
      }
      /* Potential conflict miss */

      oldTag = DCACHE_TAG(pls->pAddr);
      for (i = 0; i < nextTag; i++) {
         if (oldTag == cacheTags[i]) break;
      }
      
      /* If the old tag is not found
       * then add it to the temporary tags 
       * else verify the new tag does not conflict directly 
       *      with this (pls->pAddr) previous access */       
      if (i == nextTag) {
         /* If tags are full then the new access is a conflict */
         if (nextTag == assoc) {
            return TRUE;
         }
         cacheTags[nextTag++] = oldTag;
      } else {
         if (oldTag == DCACHE_TAG(ls->pAddr)) {
            PA mask = SamGetLdstMask(ls);
            PA pmask = SamGetLdstMask(pls);
            PA cmask = mask & pmask;
            if ((ls->pAddr & cmask) == (pls->pAddr & cmask)) {
               return TRUE;
            }
         }
      }
   }
   return FALSE;
}

/* SamLdstBufferAlloc: Allocate a load/store buffer entry */

static SamLdstEntry *
SamLdstBufferAlloc(SamState *sp, SamIwinEntry *ip, int type)
{
   SamLdstEntry *store1 = NULL;
   SamLdstEntry *store2 = NULL;

   SamLdstEntry *ls = ip->ls;
   ASSERT(ls != NULL);
   
   SamLdstInit(ls);
   ls->pAddr = ip->pAddr;
   List_InsertDebug(LdstToList(ls), LIST_ATREAR(sp->activeLdstEntries));

   if (type & LdstLoad) {
      /* Check if we can get value from a previous store */
      int offset = ls->pAddr & 7;
      List_Links *lp;
      int size;

      if (type & LdstQuadword) {
         size = 8;
      } else if (type & LdstLongword) {
         size = 4;
      } else if (type & LdstWord) {
         size = 2;
      } else if (type & LdstByte) {
         size = 1;
      } else {
         NOTREACHED();
      }

      for (lp = List_First(sp->activeLdstEntries);
           lp != LdstToList(ls);
           lp = List_Next(lp)) {
         SamLdstEntry *pls = ListToLdst(lp);
         if (pls->type & LdstStore) {
            PA maskedPA = ls->pAddr & ~((PA)7);
            PA maskedPPA = pls->pAddr & ~((PA)7);
            if (maskedPA == maskedPPA) {
               if (store1 == NULL) {
                  store1 = pls;
               } else {
                  store2 = pls;
                  break;
               }
            }
         }
      }
      if ((size + offset) > 8) {
         size = 8 - offset;
      }
      if (store1 != NULL) {
         bool anyhit = FALSE;
         bool hitflag = TRUE;
         int b;
         for (b = 0; b < size; b++) {
            if (store1->valid[b + offset]) {
               anyhit = TRUE;
            } else {
               hitflag = FALSE;
            }
         }
         if (hitflag) ls->loadhit = store1;
         if ((!anyhit) && (store2 != NULL)) {
            hitflag = TRUE;
            for (b = 0; b < size; b++) {
               if (!store2->valid[b + offset]) {
                  hitflag = FALSE;
                  break;
               }
            }
            if (hitflag) ls->loadhit = store2;
         }
      }
   }
   return ls;
}

/* SamLdstBufferWrite: Write data from given entry to write buffer */

static void
SamLdstBufferWrite(SamState *sp, SamLdstEntry *ls, Reg *data, int size)
{
   int opcode = ls->ip->instr.common.opcode;
   int lineoffset = ls->pAddr & 7;
   ls->size = size;
   if ((size + lineoffset) > 8) {
      size = 8 - lineoffset;
   }

   switch (size) {
   case 8:
      ASSERT((lineoffset & 0x07) == 0);
      *(uint64 *)(ls->valid + lineoffset) = 0x0101010101010101ULL;
      break;
   case 4:
      ASSERT((lineoffset & 0x03) == 0);
      *(uint32 *)(ls->valid + lineoffset) = 0x01010101;
      break;
   case 2:
      ASSERT((lineoffset & 0x01) == 0);
      *(uint16 *)(ls->valid + lineoffset) = 0x0101;
      break;
   case 1:
      *(uint8 *)(ls->valid + lineoffset) = 0x01;
      break;
   default:
      fprintf(stderr, "Memory op invalid size\r\n");
      break;
   }

   /* Need to use gammaMemoryTable to handle floating point;
    * don't use it if undefined (HW_LD and HW_ST instructions) */
   if (gammaMemoryTable[opcode] != NULL) {
      Reg tdest;
      Reg *dst = (Reg *)(ls->data + lineoffset);
      Reg *src = (Reg *)data;
      gammaMemoryTable[opcode](&tdest, src, dst, ls->pAddr, ls->vAddr);
      if (ls->type & LdstConditional) {
         ASSERT(tdest == 1);
      }
   } else if (size == 8) {
      uint64 qval = *data;
      *(uint64 *)(ls->data + lineoffset) = qval;
   } else if (size == 4) {
      uint32 lval = *data;
      *(uint32 *)(ls->data + lineoffset) = lval;
   } else if (size == 2) {
      uint16 lval = *data;
      *(uint16 *)(ls->data + lineoffset) = lval;
   } else if (size == 1) {
      uint8 lval = *data;
      *(uint8 *)(ls->data + lineoffset) = lval;
   } else {
      NOTREACHED();
   }

   return;
}

/* SamLdstBufferUpdateRead: Read data from given entry, if hit in store buffer */

static bool
SamLdstBufferUpdateRead(SamState *sp, SamLdstEntry *ls, byte *data, int size)
{
   int i;
   int lineoffset;
   char valid[8] = { 0, 0, 0, 0, 0, 0, 0, 0 };
   bool success = TRUE;

   if (ls->loadhit == NULL) return FALSE;

   lineoffset = ls->pAddr & 7;
   if ((size + lineoffset) > 8) {
      size = 8 - lineoffset;
   }

   for (i = 0; i < size; i++) {
      byte v = *(byte *)(ls->loadhit->valid + (lineoffset + i));
      if (v) {
         data[i] = *(byte *)(ls->loadhit->data + (lineoffset + i));
         valid[i + lineoffset] = 1;
      } else {
         success = FALSE;
      }
   }

   return success;
}

/* SamLdstBuffer{Free,Squash,Retire}: release an active load/store buffer entry */

static void
SamLdstBufferFree(SamState *sp, SamLdstEntry *ls)
{
   List_Links *lp;

   /* Update any dependent loads */
   if (ls->type & LdstStore) {
      LIST_FORALL(sp->activeLdstEntries, lp) {
         SamLdstEntry *ils = ListToLdst(lp);
         if (ils->loadhit == ls) {
            ASSERT(ils->type & LdstLoad);
            ils->loadhit = NULL;
            ils->status &= ~(LdstPending | LdstCached);
         }
      }
   }

   /* Free the load/store buffer entry */
   lp = LdstToList(ls);
   List_RemoveDebug(lp);
   sp->activeLdstBufCount--;
   List_InsertDebug(lp, LIST_ATREAR(sp->freeLdstEntries));
}

void
SamLdstBufferSquash(SamState *sp, SamIwinEntry *ip)
{
   SamLdstEntry *ls = ip->ls;
   ASSERT(ip == ls->ip);
   ip->iwinFlags &= ~IwinLdstBuf;
   SamLdstBufferFree(sp, ls);
   ip->ls = NULL;
}

int
SamLdstBufferRetire(SamState *sp, SamLdstEntry *ls)
{
   SamIwinEntry *ip = ls->ip;
   ASSERT((ip == NULL) || !(ip->iwinFlags & IwinSquash));
   if (!(ls->status & LdstUncached) &&
       !(ls->type & (LdstPrefetch | LdstMB | LdstWMB))) {
      if (ls->invalidated) {
         if (ls->type & (LdstLoad | LdstConditional)) {
            /* Fail loads and SCs */
            return -1;
         } else if (ls->dataPtr != NULL) {
            byte *dataPtr;
            int ret = SamRefDCache(sp->PE->myNum, ls, &dataPtr);
            switch (ret) {
            case SUCCESS:
               ls->dataPtr = dataPtr;
               break;
            case STALL:
               ls->status |= LdstStall;
               ls->dataPtr = NULL;
               ls->invalidated = FALSE;
               break;
            case FAILURE:
               return 1;
            default:
               NOTREACHED();
            }
         }
      }
      if ((ls->dataPtr == NULL) && !(ls->type & LdstConditional)) {
         if (SAM_RETIRE_STORES_EARLY && (ip != NULL)) {
            /* Retire the store, call again when line comes back to cache */
            ip->iwinFlags &= ~IwinLdstBuf;
            ip->ls = NULL;
            ls->ip = NULL;
            return 0;
         } else {
            return 1;
         }
      }
      if (ls->type & LdstStore) {
         int offset = (ls->pAddr & 7);
         byte *dst = (byte *)((size_t)ls->dataPtr & ~7) + offset;
         byte *src = ls->data + offset;
         switch (ls->size) {
         case 8:
            ((int64 *)dst)[0] = ((int64 *)src)[0];
            break;
         case 4:
            ((int32 *)dst)[0] = ((int32 *)src)[0];
            break;
         case 2:
            ((int16 *)dst)[0] = ((int16 *)src)[0];
            break;
         case 1:
            ((int8 *)dst)[0] = ((int8 *)src)[0];
            break;
         case 0:
            /* Failed SC case: no data to be copied */
            ASSERT(ls->dataPtr == NULL);
            ASSERT(ls->type & LdstConditional);
            break;
         default:
            NOTREACHED();
            break;
         }
      }
   }
   if (ip != NULL) {
      ip->iwinFlags &= ~IwinLdstBuf;
      ip->ls = NULL;
   }
   SamLdstBufferFree(sp, ls);
   return 0;
}

/* SamLdstCompleteLoad: Write back the results of a load operation */

static void
SamLdstCompleteLoad(SamState *sp, SamLdstEntry *ls)
{
   ls->status |= LdstDone;
   ls->reg->contents.gpr = ls->loadValue;
   SamRegWriteback(sp, ls->ip);
   if (ls->type & LdstLocked) {
      PA align = ls->pAddr & ~(PA)(DCACHE_LINE_SIZE - 1);
      sp->PE->lock_flag = 1;
      sp->PE->locked_addr = align;
   }
}

/* SamLdstOp: Issue store to write buffer, check load against write buffer */

static int
SamLdstOp(SamState *sp, SamLdstEntry *ls, bool updateInst)
{
   static double nullquadword = 0;
   byte *dataPtr = (byte *) &nullquadword;
   Reg_s val;

   if (ls->dataPtr != NULL) {
      dataPtr = ls->dataPtr;
   }

   switch (ls->type) {
   case LdstLdQuadword: 
   case LdstLdQuadwordLocked:
   case LdstLdQuadwordUnaligned:
   {
      int64 qval = *(int64 *)dataPtr;
      SamLdstBufferUpdateRead(sp, ls, (byte *)&qval, 8);
      val = (Reg_s)qval;
      break;
   }
   case LdstLdLongword: 
   case LdstLdLongwordLocked:
   {
      int32 lval = *(int32 *)dataPtr;
      SamLdstBufferUpdateRead(sp, ls, (byte *)&lval, 4);
      val = (Reg_s)lval;
      break;
   }
   case LdstLdWordUnsigned:
   {
      uint16 wval = *(uint16 *)dataPtr;
      SamLdstBufferUpdateRead(sp, ls, (byte *)&wval, 2);
      val = (Reg)wval;
      break;
   }
   case LdstLdByteUnsigned:
   {
      uint8 bval = *(uint8 *)dataPtr;
      SamLdstBufferUpdateRead(sp, ls, (byte *)&bval, 1);
      val = (Reg)bval;
      break;
   }

   case LdstStQuadword:
   case LdstStQuadwordConditional:
   case LdstStQuadwordUnaligned:
      SamLdstBufferWrite(sp, ls, &ls->reg->contents.gpr, 8);
      break;
   case LdstStLongword:
   case LdstStLongwordConditional:
      SamLdstBufferWrite(sp, ls, &ls->reg->contents.gpr, 4);
      break;
   case LdstStWord:
      SamLdstBufferWrite(sp, ls, &ls->reg->contents.gpr, 2);
      break;
   case LdstStByte:
      SamLdstBufferWrite(sp, ls, &ls->reg->contents.gpr, 1);
      break;

   default:
      NOTREACHED();
      break;
   }

   if (ls->type & LdstLoad) {
      int opcode = ls->ip->instr.common.opcode;
      /* Need to use gammaMemoryTable to handle floating point;
       * don't use it if undefined (HW_LD and HW_ST instructions) */
      if (gammaMemoryTable[opcode] != NULL) {
         Reg *dst = &ls->loadValue;
         Reg *src = &val;
         gammaMemoryTable[opcode](dst, NULL, src, ls->pAddr, ls->vAddr);
      } else {
         ls->loadValue = val;
      }
      if (SAM_LOAD_TO_USE_LATENCY > 0) {
         ls->status |= LdstLdDelay;
         ls->loadCompleteCycle = samCurrentTime + SAM_LOAD_TO_USE_LATENCY;
      } else {
         SamLdstCompleteLoad(sp, ls);
      }
   } else {
      if (ls->type & LdstConditional) {
         ls->status |= LdstDone;
         ls->screg->contents.gpr = 1;
         SamRegWriteback(sp, ls->ip);
      } else if (updateInst) {
         ls->status |= LdstDone;
         SamFreeInst(sp, ls->ip, FALSE);
      }
   }

   return 0;
}
   
static int
SamLdstOpUncached(SamState *sp, SamLdstEntry *ls)
{
   int err;
   int64 qval;
   Reg_s val = 0;
   int cpuNum = sp->PE->myNum;

   switch (ls->type) {
   case LdstLdQuadword:
   case LdstLdQuadwordUnaligned:
      err = SamRefUncached(cpuNum, ls, &qval);
      val = (Reg_s)qval;
      break;
   /* LdstLdLongword/LdstLdWordUnsigned/LdstLdByteUnsigned:
    * need to supply quadword-aligned pointer, since uncached expects Reg* */
   case LdstLdLongword:
      err = SamRefUncached(cpuNum, ls, &qval);
      val = (Reg_s)(int32)qval;
      break;
   case LdstLdWordUnsigned:
      err = SamRefUncached(cpuNum, ls, &qval);
      val = (Reg)(uint16)qval;
      break;
   case LdstLdByteUnsigned:
      err = SamRefUncached(cpuNum, ls, &qval);
      val = (Reg)(uint8)qval;
      break;

   case LdstStQuadword:
   case LdstStQuadwordUnaligned:
   case LdstStLongword:
   case LdstStWord:
   case LdstStByte:
      err = SamRefUncached(cpuNum, ls, &(ls->reg->contents.gpr));
      break;
   default:
      NOTREACHED();
      return 1;
   }

   if (err == 0) {
      ls->status |= LdstDone;
      if (ls->type & LdstLoad) {
         ls->reg->contents.gpr = val;
         SamRegWriteback(sp, ls->ip);
      } else {
         SamFreeInst(sp, ls->ip, FALSE);
      }
   }

   return err;
}

/* SamLdstCache: Perform the actual read/write operation */

static int
SamLdstCache(SamState *sp, SamLdstEntry *ls)
{
   int ret;
   int cpuNum = sp->PE->myNum;
   byte *dataPtr = NULL;

   bool store = ls->type & LdstStore;
   bool sc = ls->type & LdstConditional;

   if (sc) {
      PA align = ls->pAddr & ~(PA)(DCACHE_LINE_SIZE - 1);
      if (!sp->PE->lock_flag || 
          (sp->PE->locked_addr != align) ||
          (ls->status & LdstFailed)) {
         goto scfail;
      }
   }

   /* Update AlphaState for miss bookkeeping */
   sp->PE->PC = ls->pc;
   
   if (ls->status & LdstStall) {
      ret = STALL;
   } else {
      ls->status |= LdstFailed;
      ret = SamRefDCache(cpuNum, ls, &dataPtr);
      ls->status &= ~LdstFailed;
   }

   if (ls->type & LdstPrefetch) {
      ls->status |= LdstDone;
      SamFreeInst(sp, ls->ip, FALSE);
      return 0;
   }

   switch (ret) {
   case SUCCESS:
      ls->status |= LdstCached;
      if (!(ls->status & LdstPending)) {
         /* When the operation is done, clear the invalidated
          * flag.  If the flag gets set between here and when
          * this operation is retired, then we have to retry it. */
         ls->dataPtr = dataPtr;
         ls->invalidated = FALSE;
         SamLdstOp(sp, ls, TRUE);
      }
      return 0;
   case FAILURE:
      ls->status |= LdstFailed;
      return -1;
   case STALL:
      ls->status |= LdstStall | LdstCached;
      if (!(ls->status & LdstPending)) {
         if (ls->loadhit || (store && !sc)) {
            /* Loads that hit in the write buffer, and stores
             * other than SC instructions, can be processed */
            SamLdstOp(sp, ls, TRUE);
         }
      }
      return 0;
   case SCFAILURE:
scfail:
      ls->screg->contents.gpr = 0;
      SamRegWriteback(sp, ls->ip);
      ls->size = 0;
      ls->status |= LdstDone;
      return 0;
   case BUSERROR:
      /* rpb: what's the correct trap to raise here? */
      CPUVec.Exception(sp->PE->myNum, TRAP_UNALIGN);
      SamTrapCopy(sp->lastTrap, ls->ip->trap);
      SamRegWriteback(sp, ls->ip);
      ls->ip->iwinFlags |= IwinFault;
      ls->ip->thread->stallFetch |= StallTrap;
      ls->status |= LdstDone;
      ls->size = 0;
      return 0;
   default:
      NOTREACHED();
      return 0;
   }
}

/* SamLdstQueue: Add a request to the load/store buffer */

void
SamLdstQueue(SamState *sp, SamIwinEntry *ip, int type,
             SamRegState *reg, SamRegState *screg)
{
   SamLdstEntry *ls = ip->ls;
   ASSERT(ls != NULL);
   SamLdstBufferAlloc(sp, ip, type);

   ip->iwinFlags |= IwinLdstBuf;

   ls->status = 0;
   ls->type = type;
   ls->pc = ip->pc;
   ls->vAddr = ip->vAddr;
   ls->reg = reg;
   ls->screg = screg;
   ls->ip = ip;
   ls->dataPtr = NULL;
#if _SAM_DBG_LDST
   ls->start = samCurrentTime;
#endif

   if (type & (LdstMB | LdstWMB)) {
      /* Nothing to do for MB/WMB operations: 
         they don't issue until all preceding load/store ops have completed */
      ls->status |= LdstDone;
      SamFreeInst(sp, ip, FALSE);
   } else if (ip->iwinFlags & IwinUncached) {
      ls->status |= LdstUncached;
   } else if (type & LdstLoad) {
      /* Check for conflicts with other cached operations */
      if (SamLdstBufferConflict(sp, ls)) {
         ls->status |= LdstConflict;
      }
      if ((ls->loadhit) && (ls->loadhit->status & LdstPending)) {
         ls->status |= LdstPending;
      }
   } else if (type & LdstStore) {
      /* Check for conflicts with other cached operations */
      if (SamLdstBufferConflict(sp, ls)) {
         ls->status |= LdstConflict;
      }
      /* Do the operation once to set valid bytes information;
       * then check if the value was really available */
      if (!(type & LdstConditional)) {
         SamLdstOp(sp, ls, FALSE);
      }
      if (reg->status & RegInWin) {
         ls->status |= LdstPending;
      }
   } else {
      ASSERT(ls->type & LdstPrefetch);
   }
}

/* SamMemory: Perform memory operations */

void
SamMemory(SamState *sp)
{
   int count = 0;
   bool checkLoads = FALSE;
   List_Links *lp;

   LIST_FORALL(sp->activeLdstEntries, lp) {
      SamLdstEntry *ls = ListToLdst(lp);

      /* Retry any references that have been filled in the cache */

      if (ls->status & LdstFilled) {
         byte *dataPtr;
         int ret = SamRefDCache(sp->PE->myNum, ls, &dataPtr);
         switch (ret) {
         case SUCCESS:
            ls->status &= ~LdstFilled;
            ls->dataPtr = dataPtr;
            break;
         case STALL:
            ls->status &= ~LdstFilled;
            ls->status |= LdstStall;
            break;
         case FAILURE:
            /* retry again next cycle */
            break;
         default:
            NOTREACHED();
            break;
         }
      }

      /* Update the pending flags for store operations */

      if ((ls->type & LdstStore) && (ls->status & LdstPending)) {
         if (!(ls->reg->status & RegInWin)) {
            /* Turn off LdstCached so it will be retried in cache */
            ls->status &= ~(LdstPending | LdstCached);
            /* Write the correct value in the load/store buffer */
            if (!(ls->type & LdstConditional)) {
               SamLdstOp(sp, ls, FALSE);
            }
            checkLoads = TRUE;
         }
      }

      /* Check for completion of load-to-use delay */

      if ((ls->status & LdstLdDelay) &&
          (ls->loadCompleteCycle == samCurrentTime)) {
         SamLdstCompleteLoad(sp, ls);
         ls->status &= ~LdstLdDelay;
      }
   }

   /* Retry any writes for retired instructions */

   if (SAM_RETIRE_STORES_EARLY) {
      /* Can't use LIST_FORALL since SamLdstBufferRetire may invalidate lp */
      lp = List_First(sp->activeLdstEntries);
      while (!List_IsAtEnd(sp->activeLdstEntries, lp)) {
         List_Links *nlp = lp->nextPtr;
         SamLdstEntry *ls = ListToLdst(lp);
         if (ls->ip == NULL) {
            ASSERT(ls->type & LdstStore);
            ASSERT(!(ls->type & LdstConditional));
            ASSERT(!(ls->status & LdstUncached));
            SamLdstBufferRetire(sp, ls);
         }
         lp = nlp;
      }
   }

   /* If a store became available, check for waiting loads */

   if (checkLoads) {
      LIST_FORALL(sp->activeLdstEntries, lp) {
         SamLdstEntry *ls = ListToLdst(lp);
         if ((ls->type & LdstLoad) && (ls->status & LdstPending)) {
            if (!(ls->loadhit->status & LdstPending)) {
               /* Turn off LdstCached so it will be retried in cache */
               ls->status &= ~(LdstPending | LdstCached);
            }
         }
      }
   }

   /* Update the conflict status of entries */

   LIST_FORALL(sp->activeLdstEntries, lp) {
      SamLdstEntry *ls = ListToLdst(lp);
      if (ls->status & LdstConflict) {
         if (!SamLdstBufferConflict(sp, ls)) {
            ls->status &= ~LdstConflict;
         }
      }
   }

   /* Try to send requests to the cache */
   
   LIST_FORALL(sp->activeLdstEntries, lp) {
      SamLdstEntry *ls = ListToLdst(lp);

      /* Don't need to do anything if operation is done */

      if (ls->status & (LdstDone | LdstLdDelay)) continue;

      /* Uncached operations are issued at graduation time */

      if (ls->status & LdstUncached) {
         int ret;
         SamIwinEntry *hp = ReorderListToIwin(List_First(sp->reorderBuffer));
         if ((ls->ip != hp) || (ls->status & LdstPending)) {
            /* Not ready to issue uncached operation: skip it */
            continue;
         }

         ls->dataPtr = (byte *) ls->pAddr;
         ret = SamLdstOpUncached(sp, ls);

         /* Only perform one uncached op per cycle */
         break;
      }

      /* Issue operation to the memory system */

      if (!(ls->status & (LdstCached | LdstConflict))) {
         int ret = SamLdstCache(sp, ls);
         if (ret >= 0) {
            count++;
         }
      }

      /* Bail if the cache couldn't accept the request */

      if (ls->status & LdstFailed) break;           
   }

   /* Track nuber of active load/store buffer entries */
   HistogramAdd(sp->ldstBufHistogram, sp->activeLdstBufCount);

   /* rpb hack -- track number of active misses */
   {
      int cpuNum = sp->PE->myNum;
      int rdmiss = CACHE[cpuNum].readDMisses;
      int rmiss = CACHE[cpuNum].readMisses;
      int wmiss = CACHE[cpuNum].writeMisses;
      int miss = rmiss + wmiss;
      HistogramAdd(sp->readDMissHistogram, rdmiss);
      HistogramAdd(sp->readMissHistogram, rmiss);
      HistogramAdd(sp->writeMissHistogram, wmiss);
      HistogramAdd(sp->cacheMissHistogram, miss);
   }
}

int
SamLdstFillLine(SamState *sp, PA pAddr, int type)
{
   int stall = 0;
   PA pAddrAlign = pAddr & ~(PA)(SCACHE_LINE_SIZE - 1);
   List_Links *lp;
   if (type & E_I) {
      /* Attribute as I$ stall if pipe is empty and fetch is stalled on I$ */
      SamThreadState *th;
      SamGetFetchThread(th, sp);
      if (th->stallFetch & StallICache) {
         th->stallFetch &= ~StallICache;
         if (List_IsEmpty(sp->reorderBuffer)) {
            stall = sp->stallSlots;
            sp->stallSlots = 0;
         }
      }
   } 

   /* Check stalled requests, see if any of them just came back to the cache */
   LIST_FORALL(sp->activeLdstEntries, lp) {
      SamLdstEntry *ls = ListToLdst(lp);
      if (ls->status & LdstStall) {
         PA lsPAddrAlign = ls->pAddr & ~(PA)(SCACHE_LINE_SIZE - 1);
         if (pAddrAlign == lsPAddrAlign) {
            /* Attribute as D$ stall if the instruction at the head
             * of the graduation queue is waiting for this cache line */
            SamIwinEntry *hp = ReorderListToIwin(List_First(sp->reorderBuffer));
            if ((type & E_D) && (ls->ip == hp)) {
               stall = sp->stallSlots;
               sp->stallSlots = 0;
            }      
            ls->status &= ~LdstStall;
            if (ls->status & LdstDone) {
               ls->status |= LdstFilled;
            } else {
               ls->status &= ~LdstCached;
            }
         }
      }
   }
   return stall;
}

void 
SamLdstFlushLine(SamState *sp, PA pAddr, int size, int retain)
{
   PA pAddrEnd = pAddr + size;
   List_Links *lp;

   /* Check stalled requests, see if any of them just got flushed */
   LIST_FORALL(sp->activeLdstEntries, lp) {
      SamLdstEntry *ls = ListToLdst(lp);

      /* Check if we need to invalidate request:    If.... */
      if ((ls->dataPtr != NULL) &&               /* 1. Already have data */
          !(retain && (ls->type & LdstLoad)) &&  /* 2. Not a load downgrade */
          (ls->pAddr >= pAddr) &&                /* 3. In the flush region */
          (ls->pAddr < pAddrEnd)) {

         ls->invalidated = TRUE;                 /* Then we invalidate */
      }
   }
}

/* SamLdstGetByte: Extract value from write buffer */

bool
SamLdstGetByte(SamState *sp, PA pA, byte *buf)
{
   if (SAM_RETIRE_STORES_EARLY) {
      List_Links *lp;
      LIST_FORALL(sp->activeLdstEntries, lp) {
         SamLdstEntry *ls = ListToLdst(lp);
         if (ls->ip == NULL) {
            /* Store instruction already retired: check address */
            PA mask = SamGetLdstMask(ls);
            if ((ls->pAddr & mask) == (pA & mask)) {
               /* Extract value from write buffer and return it */
               int offset = pA & 7;
               *buf = ls->data[offset];
               return TRUE;
            }
         }
      }
   }
   return FALSE;
}
