/*
 * opcodes.c -- Instruction execution emulation
 */

#include "sam.h"
#include "print.h"
#include "instruction.h"
#include "ev5.h"
#include "ev5_ipr.h"
#include "cpu_stats.h"

/* Use gamma opcode implementations for now */
#include "gamma.h"

#define EXECUTE(TYPE)                                           \
   void SamExecute ## TYPE (SamState *sp, SamIwinEntry *ip)

/* Utility routines for handling CMOV, FP disabled, unaligned reference */
static EXECUTE(CMOV2);
static EXECUTE(FPDisabled);
static EXECUTE(Unaligned);

EXECUTE(PAL)
{
   switch (ip->instr.common.opcode) {
   case op_call_pal:
      if ((ip->instr.pal_format.function == 0) ||
          (ip->instr.pal_format.function == 0x80)) {
         /* rpb: treat halt and bpt as an illegal instruction */
         SamExecuteIllegal(sp, ip);
         return;
      }
      EV5_PALOpcode(sp->PE, ip->instr.pal_format.function);
      break;
   case PRIV_OP_MTPR:
   case PRIV_OP_MFPR:
      /* Nothing to do here -- MTPR/MFPR execute at retire time. */
      break;
   case PRIV_OP_HW_LD:
      SamExecuteHWAddressCalculation(sp, ip);
      if (ip->iwinFlags & IwinTBFault) {
         ip->iwinFlags |= IwinFault;
      }
      if (ip->iwinFlags & IwinFault) {
         ip->thread->stallFetch |= StallTrap;
      } else {
         int type;
         ev5_hwldst_format hwldst;
         hwldst.word = ip->instr.word;
         if (hwldst.hwmem_format.quad) {
            type = LdstLdQuadword;
         } else {
            type = LdstLdLongword;
         }
         if (SamLdstCheckAlignment(ip->pAddr, type)) {
            SamLdstQueue(sp, ip, type, ip->ra.reg, NULL);
         } else {
            SamExecuteUnaligned(sp, ip);
         }
      }
      break;
   case PRIV_OP_HW_ST:
      SamExecuteHWAddressCalculation(sp, ip);
      if (ip->iwinFlags & IwinTBFault) {
         ip->iwinFlags |= IwinFault;
      }
      if (ip->iwinFlags & IwinFault) {
         ip->thread->stallFetch |= StallTrap;
      } else {
         ev5_hwldst_format hwldst;
         hwldst.word = ip->instr.word;
         if (hwldst.hwmem_format.quad) {
            SamLdstQueue(sp, ip, LdstStQuadword, ip->ra.reg, NULL);
         } else {
            SamLdstQueue(sp, ip, LdstStLongword, ip->ra.reg, NULL);
         }
      }
      break;
   case PRIV_OP_REI:
      SamExportRegs(sp, ip->thread);
      EV5_HwRei(sp->PE);
      SamImportRegs(sp, ip->thread);
      ip->thread->pc = sp->PE->PC;
      break;
   }
   ASSERT(!SamTrapPending(sp));
   freeTrapState(sp->currentTrap.state);
   sp->currentTrap.state = allocTrapState(sp->PE->trapState);
   SamPrintStage(sp, ip, "EX PAL");
}

EXECUTE(LoadAddress)
{
   VA vAddr;
   Reg base = ip->rb.reg->contents.gpr;
   int disp = ip->instr.m_format.memory_displacement;
   if (ip->instr.common.opcode == op_ldah) {
      vAddr = base + (disp << 16);
   } else {
      vAddr = base + disp;
   }
   ip->ra.reg->contents.gpr = vAddr;
   SamPrintStage(sp, ip, "EX LDA");
}

EXECUTE(MemLoad)
{
   /* For FP loads, check if floating point operations are enabled */
   if ((ip->ra.type != RegInt) && !FPENABLED(sp->PE)) {
      SamExecuteFPDisabled(sp, ip);
      return;
   }

   SamExecuteAddressCalculation(sp, ip);
   if (ip->iwinFlags & IwinTBFault) {
      ip->iwinFlags |= IwinFault;
   }

   if (ip->iwinFlags & IwinFault) {
      ip->thread->stallFetch |= StallTrap;
   } else {
      int type;
      SamRegState *reg = ip->ra.reg;

      if (ip->instFlags & InstPrefetch) {
         type = LdstPrefetchShared;
      } else if (ip->instFlags & InstPrefetchMod) {
         type = LdstPrefetchExclusive;
      } else {
         switch (ip->instr.common.opcode) {
#ifdef op_ldbu
         case op_ldbu:
            type = LdstLdByteUnsigned;
            break;
#endif
#ifdef op_ldwu
         case op_ldwu:
            type = LdstLdWordUnsigned;
            break;
#endif
         case op_ldl: case op_ldf: case op_lds:
            type = LdstLdLongword;
            break;
         case op_ldq: case op_ldg: case op_ldt:
            type = LdstLdQuadword;
            break;
         case op_ldl_l: 
            type = LdstLdLongwordLocked;
            break;
         case op_ldq_l: 
            type = LdstLdQuadwordLocked;
            break;
         case op_ldq_u: 
            type = LdstLdQuadwordUnaligned;
            break;
         default:       
            NOTREACHED();
            return;
         }
      }
      if (type & LdstUnaligned) {
         ip->vAddr &= ~(VA)7;
         ip->pAddr &= ~(PA)7;
      }

      /* rpb hack: stolen from kappa.c */
      if (ip->iwinFlags & IwinUncached) {
         type = LdstLdLongword;
      }

      if (SamLdstCheckAlignment(ip->pAddr, type)) {
         SamLdstQueue(sp, ip, type, reg, NULL);
         STATS_INC(sp->PE->myNum, dReads, 1);
      } else {
         SamExecuteUnaligned(sp, ip);
      }
   }
   SamPrintStage(sp, ip, "EX LD");
}

EXECUTE(MemStore)
{
   SamExecuteAddressCalculation(sp, ip);

   if (ip->iwinFlags & IwinTBFault) {
      ip->iwinFlags |= IwinFault;
   }

   /* For FP stores, check if floating point operations are enabled */
   if ((ip->ra.type != RegInt) && !FPENABLED(sp->PE)) {
      SamExecuteFPDisabled(sp, ip);
      return;
   }

   if (ip->iwinFlags & IwinFault) {
      ip->thread->stallFetch |= StallTrap;
   } else {
      int type;
      SamRegState *reg = ip->ra.reg;
      SamRegState *screg = NULL;

      switch (ip->instr.common.opcode) {
#ifdef op_stb
      case op_stb:
         type = LdstStByte;
         break;
#endif
#ifdef op_stw
      case op_stw:
         type = LdstStWord;
         break;
#endif
      case op_stl: case op_stf: case op_sts:
         type = LdstStLongword; 
         break;
      case op_stq: case op_stg: case op_stt:
         type = LdstStQuadword; 
         break;
      case op_stl_c: 
         type = LdstStLongwordConditional; 
         break;
      case op_stq_c: 
         type = LdstStQuadwordConditional; 
         break;
      case op_stq_u: 
         type = LdstStQuadwordUnaligned; 
         break;
      default:
         NOTREACHED(); 
         return;
      }

      if (type & LdstUnaligned) {
         ip->vAddr &= ~(VA)7;
         ip->pAddr &= ~(PA)7;
      }

      /* rpb hack: stolen from kappa.c */
      if (ip->iwinFlags & IwinUncached) {
         type = LdstStLongword;
      }

      if (type & LdstConditional) {
         // Hack: for SC, use Rc to hold output version of Ra
         screg = ip->rc.reg;
      }

      if (SamLdstCheckAlignment(ip->pAddr, type)) {
         SamLdstQueue(sp, ip, type, reg, screg);
         STATS_INC(sp->PE->myNum, dWrites, 1);
      } else {
         SamExecuteUnaligned(sp, ip);
      }
   }
   SamPrintStage(sp, ip, "EX ST");
}

EXECUTE(Branch)
{
   Reg *raPtr = &ip->ra.reg->contents.gpr;
   bool cond;
   int opcode = ip->instr.common.opcode;

   /* Unconditional: Store updated PC in RA */
   /* Conditional: Compute branch condition */
   if (!(ip->instFlags & InstConditional)) {
      cond = 1;
      ASSERT((opcode == op_br) || (opcode == op_bsr));
      *raPtr = ip->pc + 4;
   } else if (ip->ra.type == RegInt) {
      Reg_s rav = ip->ra.reg->contents.gpr;
      switch (opcode) {
      case op_blbc: cond = !(rav & 1); break;
      case op_beq:  cond = (rav == 0); break;
      case op_blt:  cond = (rav < 0);  break;
      case op_ble:  cond = (rav <= 0); break;
      case op_blbs: cond = (rav & 1);  break;
      case op_bne:  cond = (rav != 0); break;
      case op_bge:  cond = (rav >= 0); break;
      case op_bgt:  cond = (rav > 0);  break;
      default:
         CPUWarning("unknown branch opcode: 0x%x \n", opcode);
         SamExecuteIllegal(sp, ip);
         return;
      }
   } else {
      double fav = ip->ra.reg->contents.fpr;
      /* Check if floating point operations are enabled */
      if (!FPENABLED(sp->PE)) {
         SamExecuteFPDisabled(sp, ip);
         return;
      }
      switch (opcode) {
      case op_fbeq: cond = (fav == 0); break;
      case op_fblt: cond = (fav < 0);  break;
      case op_fble: cond = (fav <= 0); break;
      case op_fbne: cond = (fav != 0); break;
      case op_fbge: cond = (fav >= 0); break;
      case op_fbgt: cond = (fav > 0);  break;
      default: 
         CPUWarning("unknown branch opcode: 0x%x \n", opcode);
         SamExecuteIllegal(sp, ip);
         return;
      }
   }
   SamBranchResolve(sp, ip, cond);
   SamPrintStage(sp, ip, "EX BR");
}

EXECUTE(Jump)
{
   Reg *raPtr = &ip->ra.reg->contents.gpr;
   Reg rbv = ip->rb.reg->contents.gpr;

   /* Ignore low two bits of Rb */
   rbv &= ~(Reg)0x3;
   if (IS_PAL(sp->PE)) {
      rbv |= 0x1;
   }

   /* Store updated PC in RA */
   *raPtr = ip->pc + 4;

   SamBranchResolveJump(sp, ip, rbv);
   SamPrintStage(sp, ip, "EX JSR");
}

EXECUTE(Operate)
{
   Reg rav, rbv, *rcPtr;
   int gfunc;

   if (ip->instFlags & InstCMOV2) {
      SamExecuteCMOV2(sp, ip);
      return;
   }

   rav = ip->ra.reg->contents.gpr;
   rcPtr = &ip->rc.reg->contents.gpr;

   if (ip->instFlags & InstCMOV1) {
      /* Use gammaOperateTable to compute move predicate */
      rbv = 1;
      *rcPtr = 0;
   } else if (ip->instr.o_format.form) {
      rbv = ip->instr.l_format.literal;
   } else {
      rbv = ip->rb.reg->contents.gpr;
   }

   gfunc = (ip->instr.common.opcode * (1<<7)) + ip->instr.o_format.function;
   if (!gammaOperateTable[gfunc]) {
      CPUWarning("\nOperateOpcode, not def "
                 "pc=%x code_func=%#x Ra=%lu Rb=%lu Rc=%lu instr=0x%lx\n",
		 ip->pc, gfunc, ip->ra.areg, ip->rb.areg, ip->rc.areg, 
                 ip->instr.word);
      /* Check for zero register */
      /* rpb: Includes hack from gamma to enable booting 
       * until the palcode is fixed; is this still needed? */
      if ((ip->rc.areg != ZERO_REGISTER) &&
          !((ip->pc == 0x8f15) && (ip->instr.word == 0x47e014b0))) {
         SamExecuteIllegal(sp, ip);
         return;
      }
   } else {
      sp->PE->operating_instr = ip->instr;
      sp->PE->operating_flags = gammaOperateFlags[gfunc];
      gammaOperateTable[gfunc](rav, rbv, rcPtr, rcPtr);
      sp->PE->operating_instr.word = 0;
   }

   if (ip->instFlags & InstCMOV1) {
      /* Set predicate according to conditional move result */
      ip->rc.reg->predicate = ip->rc.reg->contents.gpr;
      /* Copy value from oldRc to newRc1 */
      ip->rc.reg->contents.gpr = ip->rb.reg->contents.gpr;
   }

   if (ip->instFlags & InstFloating) {
      SamPrintStage(sp, ip, "EX FP");
   } else {
      SamPrintStage(sp, ip, "EX INT");
   }
}

EXECUTE(CMOV2)
{
   if (ip->ra.reg->predicate) {
      /* Copy value from Rb to newRc2 */
      if (ip->instFlags & InstFloating) {
         ip->rc.reg->contents.fpr = ip->rb.reg->contents.fpr;
      } else if (ip->instr.o_format.form) {
         ip->rc.reg->contents.gpr = ip->instr.l_format.literal;
      } else {
         ip->rc.reg->contents.gpr = ip->rb.reg->contents.gpr;
      }
   } else {
      /* Copy value from newRc1 to newRc2 */
      if (ip->instFlags & InstFloating) {
         ip->rc.reg->contents.fpr = ip->ra.reg->contents.fpr;
      } else {
         ip->rc.reg->contents.gpr = ip->ra.reg->contents.gpr;
      }
   }

   if (ip->instFlags & InstFloating) {
      SamPrintStage(sp, ip, "EX FP");
   } else {
      SamPrintStage(sp, ip, "EX INT");
   }
}

EXECUTE(Floating)
{
   double *raPtr, *rbPtr, *rcPtr;
   double rbCMOV = 1;
   int gfunc;

   /* Check if floating point operations are enabled */
   if (!FPENABLED(sp->PE)) {
      SamExecuteFPDisabled(sp, ip);
      return;
   }

   if (ip->instFlags & InstCMOV2) {
      SamExecuteCMOV2(sp, ip);
      return;
   }

   raPtr = &ip->ra.reg->contents.fpr;
   rbPtr = &ip->rb.reg->contents.fpr;
   rcPtr = &ip->rc.reg->contents.fpr;

   if (ip->instFlags & InstCMOV1) {
      rbPtr = &rbCMOV;
      *rcPtr = 0;
   }

   gfunc = (ip->instr.common.opcode * (1<<11)) + ip->instr.f_format.function;
   if (!gammaFPTable[gfunc]) {
      CPUWarning("FPOpcode, not def pc=%lx opcode=%x func=%x\n",
                 ip->pc, ip->instr.common.opcode, ip->instr.f_format.function);
      /* Check for zero register */
      if (ip->rc.areg != ZERO_REGISTER) {
         SamExecuteIllegal(sp, ip);
         return;
      }
   } else { 
      sp->PE->operating_instr = ip->instr;
      sp->PE->operating_flags = gammaFPFlags[gfunc];
      gammaFPTable[gfunc](raPtr, rbPtr, rcPtr, rcPtr, &ip->fpcr);
      sp->PE->operating_instr.word = 0;
   }

   if (ip->instFlags & InstCMOV1) {
      /* Set predicate according to conditional move result */
      ip->rc.reg->predicate = ip->rc.reg->contents.fpr;
      /* Copy value from oldRc to newRc1 */
      ip->rc.reg->contents.fpr = ip->rb.reg->contents.fpr;
   }

   SamPrintStage(sp, ip, "EX FP");
}

EXECUTE(Misc)
{
   /* Don't want sign extension */
   uint func = ip->instr.m_format.memory_displacement & 0xffff;
   switch (func) {
   case misc_mb:
      SamLdstQueue(sp, ip, LdstMB, NULL, NULL);
      break;
   case misc_wmb:
      SamLdstQueue(sp, ip, LdstWMB, NULL, NULL);
      break;
   case misc_trapb:
   case misc_excb:
      /* Nothing to do */
      break;
   case misc_rc:
   case misc_rs:
      EV5_MiscReadClearSet(sp->PE, &ip->rdest->reg->contents.gpr, func == misc_rs);
      break;
   case misc_rpcc:
      ip->rdest->reg->contents.gpr = CPUVec.CycleCount(sp->PE->myNum);
      break;
   case misc_wh64:
   case misc_ecb:
      /* ECB: evict cache block
       * WH64: write hint 64
       * ignore these performance hints. ok for our purposes
       */
#if 0
      CPUPrint("Performance hint (%s) ignored.\n", (func == misc_wh64) ?
               "wh64" : "ecb");
#endif
      break;
   default:
      CPUPrint("Unimplemented: opcode=0x%lx func=0x%lx\n", 
               ip->instr.common.opcode, func);
      UNIMPLEMENTED();
   }
   SamPrintStage(sp, ip, "EX MISC");
}

EXECUTE(Illegal)
{
   CPUVec.Exception(sp->PE->myNum, TRAP_OPDEC);
   SamPrintStage(sp, ip, "EX ILL");
}

static 
EXECUTE(FPDisabled)
{
   CPUVec.Exception(sp->PE->myNum, TRAP_FEN);
   SamPrintStage(sp, ip, "EX FEN");
}

static
EXECUTE(Unaligned)
{
   CPUVec.Exception(sp->PE->myNum, TRAP_UNALIGN);
   SamPrintStage(sp, ip, "EX UNALIGNED");
}
