/*
 * register.h -- the Sam register structure and macros
 */

#ifndef SAM_REGISTER_H
#define SAM_REGISTER_H

#include "list.h"
#include "simtypes.h"

struct SamIwinEntry_s;

/* Enumeration used by iwin to track register types */

/* RegFlt is used when exact type is unknown or irrelevant */
typedef enum {
   RegNone, RegInt, RegFFlt, RegGFlt, RegSFlt, RegTFlt, RegFlt
} SamRegType;

typedef struct {
   List_Links links;   /* Freelist management -- must come first */
   union {
      Reg gpr;
      double fpr;
   } contents;
   bool predicate;     /* Used by conditional move instructions */
   struct SamIwinEntry_s *owner;         
   int status;
   int ref;
   int nmap;
} SamRegState;

/* status: Register status information */

#define RegBusy   0x01
#define RegInWin  0x02
#define RegMapped 0x04
#define RegError  0x08   /* Indicates a fault */
#define RegFreed  0x10   /* Register should be freed */

/* Convert between register and list pointers */

#define RegToList(RP) ((List_Links*)(RP))
#define ListToReg(LP) ((SamRegState*)(LP))

/* Initialize register state */

#define SamRegInit(RP)                          \
   (RP)->contents.gpr = 0;                      \
   (RP)->owner = NULL;                          \
   (RP)->status = 0;                            \
   (RP)->ref = 0;                               \
   (RP)->nmap = 0;

/* Claim a register from the free list */

#define SamGetFreeReg(RP, LIST)                 \
{                                               \
   List_Links *_lp = List_First(LIST);          \
   if (_lp == (LIST)) {                         \
      RP = NULL;                                \
   } else {                                     \
      RP = ListToReg(_lp);                      \
      List_RemoveDebug(_lp);                    \
   }                                            \
}
#define SamGetFreeGPR(GP, SP) SamGetFreeReg(GP, (SP)->gprFreeList)
#define SamGetFreeFPR(FP, SP) SamGetFreeReg(FP, (SP)->fprFreeList)

/* Check if register can be returned to the free list */

#define SamCheckRegFree(RP, TYPE, SP)                                        \
   if ((RP)->status & RegFreed) {                                            \
      if (!((RP)->status & RegMapped)) {                                     \
         (RP)->status = 0;                                                   \
         if ((TYPE) == RegInt) {                                             \
            List_InsertDebug(RegToList(RP), LIST_ATREAR((SP)->gprFreeList)); \
         } else {                                                            \
            List_InsertDebug(RegToList(RP), LIST_ATREAR((SP)->fprFreeList)); \
         }                                                                   \
      }                                                                      \
   }

/* Acquire/release a register */

#define SamAcquireReg(RP)                       \
   (RP)->nmap++;                                \
   (RP)->status |= RegMapped;

#define SamReleaseReg(RP, TYPE, SP)             \
   (RP)->nmap--;                                \
   if ((RP)->nmap == 0) {                       \
      (RP)->status &= ~RegMapped;               \
      SamCheckRegFree(RP, TYPE, SP);            \
   }
#define SamReleaseGPR(GP, SP) SamReleaseReg(GP, RegInt, SP)
#define SamReleaseFPR(GP, SP) SamReleaseReg(GP, RegFlt, SP)

#define SamFreeReg(RP, TYPE, SP)                \
   (RP)->status |= RegFreed;                    \
   SamCheckRegFree(RP, TYPE, SP);

#endif /* SAM_REGISTER_H */
