#include <stdlib.h>
#include <setjmp.h>
#include <signal.h>
#include <alpha/fpu.h>
#include "devices.h"
#include "hw_events.h"
#include "stats.h"
#include "sam.h"
#include "cpu_interface.h"
#include "instruction.h"
#include "print.h"
#include "gamma.h"
#include "ev5.h"
#include "ev5_ipr.h"
#include "cpu_stats.h"
#include "memref.h"

/* rpb hack: needed for cache histograms */
#include "pcache.h"

SamState *SPE[SIM_MAXCPUS];
SamState *curSP;
SimTime samCurrentTime;

int samPrintLevel[SIM_MAXCPUS];

int exitToSimulator = NO_CPU;
int readyToSwitch;

static void SamFPInit(void);

/* Forward declaration of functional unit specifications */

static SamFuncUnitInfo SamFixedUnit; // all insts, fixed latency & repeat rate
static SamFuncUnitInfo SamUnifiedUnit; // all insts, EV6 latency & repeat rate
static SamFuncUnitInfo SamEV6ALU;
static SamFuncUnitInfo SamEV6FPU;
static SamFuncUnitInfo SamEV6LSU;

void
SamInit(void) 
{
   int cpuNum;
   SamState  *sp;

   ASSERT(TOTAL_CPUS <= SIM_MAXCPUS);

   bzero((char*)SPE, sizeof(SPE));

   for (cpuNum = 0; cpuNum < TOTAL_CPUS; cpuNum++) {
      int i;

      /* takeInterrupt seems to be have added originally for mipsy support. 
       * The parameter controls whether EV5_SetExtInterrupt calls directly
       * into EV5_CheckForInterrupt. (The latter notices an interrupt and then
       * passes it to the CPU model.) In mipsy, the parameter was used to
       * control whether interrupts were processed immediately once the process
       * is unstalled from a cache miss.
       *
       * We set takeInterrupt to false here so that SetExtInterrupt does not
       * call CheckForInterrupt. If takeInterrupt is enabled and SetExtInterrupt
       * is called twice in a cycle (e.g. for a clock interrupt and disk I/O 
       * completion), the CheckForInterrupt could be called twice, thereby 
       * incorrectly signalling two interrupts on a single cycle. We have 
       * changed the main loop of SamCPURun to explicitly call CheckForInterrupts
       * once per cycle for each processor. RJS 11/19/01
       */
      PE[cpuNum]->takeInterrupt = 0;

      /* SAM only buffers one pending interrupt. A software interrupt
       * concurrent with a hardware interrupt that both occur during
       * a interrupt handling routine can cause SAM's interrupt 
       * structure to overflow unless carefully handled. The parameter
       * takeInterruptAtRei allows the shared ev5 code to launch
       * interrupt handling on a REI instruction. We disable this
       * parameter to force allow interrupt handling to be launched
       * from within SAM, thereby allowing us to avoid overflow
       * of SAM's internal structures.
       */
      PE[cpuNum]->takeInterruptAtRei = 0;

      /* Create the Sam CPU state */

      sp = (SamState *)malloc(sizeof(SamState));
      ASSERT(sp != NULL);
      SPE[cpuNum] = sp;
      sp->PE = PE[cpuNum];

      /* Create the general-purpose physical register state */
      
      sp->gprFreeList = (List_Links *)malloc(sizeof(List_Links));
      List_Init(sp->gprFreeList);
      ASSERT(NUM_GPRS <= SAM_MAX_GPRS);
      for (i = 0; i < NUM_GPRS; i++) {
         SamRegState *gp = &(sp->gpr[i]);
         List_InitElement(RegToList(gp));
      }

      /* Create the floating-point physical register state */

      sp->fprFreeList = (List_Links *)malloc(sizeof(List_Links));
      List_Init(sp->fprFreeList);
      ASSERT(NUM_FPRS <= SAM_MAX_FPRS);
      for (i = 0; i < NUM_FPRS; i++) {
         SamRegState *fp = &(sp->fpr[i]);
         List_InitElement(RegToList(fp));
      }

      /* Create the lists of active/free threads */

      sp->activeThreads = (List_Links *)malloc(sizeof(List_Links));
      sp->freeThreads = (List_Links *)malloc(sizeof(List_Links));
      List_Init(sp->activeThreads);
      List_Init(sp->freeThreads);
      ASSERT(NUM_THREADS <= SAM_MAX_THREADS);
      for (i = 0; i < NUM_THREADS; i++) {
         SamThreadState *th = &(sp->thread[i]);
         List_InitElement(ThreadToList(th));
      }

      /* Create the list of free iwin entries */
      
      sp->freeIwinEntries = (List_Links *)malloc(sizeof(List_Links));
      sp->reorderBuffer = (List_Links *)malloc(sizeof(List_Links));
      sp->ldstQueue = (List_Links *)malloc(sizeof(List_Links));
      sp->issueQueue = (List_Links *)malloc(sizeof(List_Links));
      sp->executeQueue = (List_Links *)malloc(sizeof(List_Links));
      List_Init(sp->freeIwinEntries);
      List_Init(sp->reorderBuffer);
      List_Init(sp->ldstQueue);
      List_Init(sp->issueQueue);
      List_Init(sp->executeQueue);
      ASSERT(NUM_IWIN_ENTRIES <= SAM_MAX_IWIN_ENTRIES);
      for (i = 0; i < NUM_IWIN_ENTRIES; i++) {
         SamIwinEntry *ip = &(sp->iwin[i]);
         List_InitElement(IwinToList(ip));
         List_InitElement(IwinToReorderList(ip));
      }

      /* Create the lists of active/free load/store buffer entries */

      sp->activeLdstEntries = (List_Links *)malloc(sizeof(List_Links));
      sp->freeLdstEntries = (List_Links *)malloc(sizeof(List_Links));
      List_Init(sp->activeLdstEntries);
      List_Init(sp->freeLdstEntries);
      ASSERT(NUM_LDST_ENTRIES <= SAM_MAX_LDST_ENTRIES);
      for (i = 0; i < NUM_LDST_ENTRIES; i++) {
         SamLdstEntry *ls = &(sp->ldstbuf[i]);
         List_InitElement(LdstToList(ls));
      }

      /* Declare the functional units */

      sp->numFuncUnits = 0;
      for (i = 0; i < NUM_FIXED_UNITS; i++) {
         sp->funcUnit[sp->numFuncUnits].typeString = "Fixed";
         sp->funcUnit[sp->numFuncUnits++].CanExecute = SamFixedUnit;
      }
      for (i = 0; i < NUM_UNIFIED_UNITS; i++) {
         sp->funcUnit[sp->numFuncUnits].typeString = "Unified";
         sp->funcUnit[sp->numFuncUnits++].CanExecute = SamUnifiedUnit;
      }
      for (i = 0; i < NUM_ALUS; i++) {
         sp->funcUnit[sp->numFuncUnits].typeString = "ALU";
         sp->funcUnit[sp->numFuncUnits++].CanExecute = SamEV6ALU;
      }
      for (i = 0; i < NUM_FPUS; i++) {
         sp->funcUnit[sp->numFuncUnits].typeString = "FPU";
         sp->funcUnit[sp->numFuncUnits++].CanExecute = SamEV6FPU;
      }
      for (i = 0; i < NUM_LSUS; i++) {
         sp->funcUnit[sp->numFuncUnits].typeString = "LSU";
         sp->funcUnit[sp->numFuncUnits++].CanExecute = SamEV6LSU;
      }

      /* Initialize the functional units */

      ASSERT(sp->numFuncUnits <= SAM_MAX_FUNC_UNITS);
      for (i = 0; i < sp->numFuncUnits; i++) {
         sp->funcUnit[i].availTime = 0;
         sp->funcUnit[i].numExecuted = 0;
      }

      /* Set up pipeline output level */

      samPrintLevel[cpuNum] = 0;

      /* Create statistics histograms */

      sp->activeLdstBufCount = 0;
      HistogramInit(sp->fetchHistogram, 0, FETCH_WIDTH, 1);
      HistogramInit(sp->issueHistogram, 0, ISSUE_WIDTH, 1);
      HistogramInit(sp->gradHistogram, 0, GRAD_WIDTH, 1);
      HistogramInit(sp->ldstBufHistogram, 0, NUM_LDST_ENTRIES, 1);

      HistogramInit(sp->readMissHistogram, 0, MHT_SIZE, 1);
      HistogramInit(sp->readDMissHistogram, 0, MHT_SIZE, 1);
      HistogramInit(sp->writeMissHistogram, 0, MHT_SIZE, 1);
      HistogramInit(sp->cacheMissHistogram, 0, MHT_SIZE, 1);
   }
}

void
SamEnter(void)
{
   int cpuNum;
   int numCPUs = TOTAL_CPUS;

   /* Create Sam data structures if necessary */

   static int initialized = 0;
   if (!initialized) {
      SamInit();
      initialized = 1;
   }

   /* Initialize Sam CPU state from PE array */

   for (cpuNum = 0; cpuNum < numCPUs; cpuNum++) {
      int i;
      SamState *sp = SPE[cpuNum];
      SamTrapClear(sp->lastTrap);
      SamTrapClear(sp->currentTrap);
      SamTrapClear(sp->interrupt);
      sp->currentTrap.state = allocTrapState(sp->PE->trapState);

      sp->scoreboard = 0;
      sp->cpuSwitch = NoSwitch;
      sp->replayStatus = Issue;
      sp->issueSlots = SAM_ISSUE_SLOTS;

      sp->stallSlots = 0;
      sp->bubble = E_PIPE_ICACHE;

      /* Initialize the general-purpose physical register state */

      for (i = 0; i < NUM_GPRS; i++) {
         SamRegState *gp = &(sp->gpr[i]);
         SamRegInit(gp);
         List_InsertDebug(RegToList(gp), LIST_ATREAR(sp->gprFreeList));
      }

      /* Initialize the floating-point physical register state */

      for (i = 0; i < NUM_FPRS; i++) {
         SamRegState *fp = &(sp->fpr[i]);
         SamRegInit(fp);
         List_InsertDebug(RegToList(fp), LIST_ATREAR(sp->fprFreeList));
      }

      /* Activate and initialize the starting thread */
      /* Put the remaining threads on the free list */

      for (i = 0; i < NUM_THREADS; i++) {
         SamThreadState *th = &(sp->thread[i]);
         SamTrapClear(th->trap);
         if (List_IsEmpty(sp->activeThreads)) {
            /* Assign physical registers to architectural registers */
            int j;
            for (j = 0; j < ALPHA_NUM_GPRS; j++) {
               /* Get GPR from free list */
               SamRegState *gp;
               SamGetFreeGPR(gp, sp);
               ASSERT(gp != NULL);
               SamAcquireReg(gp);
               th->gprName[j] = gp;
            }
            for (j = 0; j < ALPHA_NUM_FPRS; j++) {
               /* Get FPR from free list */
               SamRegState *fp;
               SamGetFreeFPR(fp, sp);
               ASSERT(fp != NULL);
               SamAcquireReg(fp);
               th->fprName[j] = fp;
            }
            SamImportRegs(sp, th);
            th->pc = sp->PE->PC;
            th->status = ThActive;
            th->stallFetch = StallNone;
            th->cmov2Fetch = 0;
            List_InsertDebug(ThreadToList(th), LIST_ATREAR(sp->activeThreads));
         } else {
            th->status = ThFree;
            List_InsertDebug(ThreadToList(th), LIST_ATREAR(sp->freeThreads));
         }
      }

      /* Initialize the instruction window entries */

      for (i = 0; i < NUM_IWIN_ENTRIES; i++) {
         SamIwinEntry *ip = &(sp->iwin[i]);
         SamTrapClear(ip->trap);
         List_InsertDebug(IwinToList(ip), LIST_ATREAR(sp->freeIwinEntries));
      }

      /* Initialize the load/store buffer entries */

      for (i = 0; i < NUM_LDST_ENTRIES; i++) {
         SamLdstEntry *ls = &(sp->ldstbuf[i]);
         List_InsertDebug(LdstToList(ls), LIST_ATREAR(sp->freeLdstEntries));
      }

      /* Initialize the branch prediction state */

      ASSERT(SAM_BP_TABLE_SIZE <= SAM_MAX_BP_TABLE_SIZE);
      ASSERT(SAM_BP_RETURN_STACK <= SAM_MAX_BP_RETURN_STACK);
      sp->branchSP = 0;
      for (i = 0; i < SAM_BP_TABLE_SIZE; i++) {
         sp->branchPredictionTable[i].bits = 0x04;
         sp->branchPredictionTable[i].target = 0;
      }
      for (i = 0; i < SAM_BP_RETURN_STACK; i++) {
         sp->branchStack[i] = 0;
      }
   }
}

void
SamExit(void)
{
   int cpuNum;
   for (cpuNum = 0; cpuNum < TOTAL_CPUS; cpuNum++) {
      SamState *sp = SPE[cpuNum];
      SamThreadState *th;
      SamGetGradThread(th, sp);

      /* Record any remaining stall time */

      if (sp->stallSlots > 0) {
         if (List_IsEmpty(sp->reorderBuffer)) {
            if (sp->bubble != 0) {
               MemStatPipelineStall(cpuNum, 0, 0, sp->stallSlots, sp->bubble);
            } else {
               MemStatPipelineStall(cpuNum, 0, 0, sp->stallSlots, E_PIPE_OTHER);
            }
         } else {
            SamIwinEntry *ip = ReorderListToIwin(List_First(sp->reorderBuffer));
            int type = SamStallType(ip, sp);
            MemStatPipelineStall(cpuNum, 0, 0, sp->stallSlots, type);
         }
      }

      /* Restore current architectural state to PE array */

      SamExportRegs(sp, th);
      sp->PE->PC = th->pc;
   }
}

void 
SamExportRegs(SamState *sp, SamThreadState *th)
{
   int i;
   for (i = 0; i < ALPHA_NUM_GPRS; i++) {
      SamRegState *gp = th->gprName[i];
      sp->PE->reg[i] = gp->contents.gpr;
   }
   for (i = 0; i < ALPHA_NUM_FPRS; i++) {
      SamRegState *fp = th->fprName[i];
      sp->PE->fp[i] = fp->contents.fpr;
   }
}

void 
SamImportRegs(SamState *sp, SamThreadState *th)
{
   int i;
   for (i = 0; i < ALPHA_NUM_GPRS; i++) {
      SamRegState *gp = th->gprName[i];
      gp->contents.gpr = sp->PE->reg[i];
   }
   for (i = 0; i < ALPHA_NUM_FPRS; i++) {
      SamRegState *fp = th->fprName[i];
      fp->contents.fpr = sp->PE->fp[i];
   }
}

static void
SamCPURun(void)
{
   int cpuNum;

   /* Get current cycle count from previous simulator */
   if (CPUVec.OldCycleCount != NULL) {
      samCurrentTime = CPUVec.OldCycleCount(0);
      /* Take the minimum of all cycle counts so we don't skip any events */
      for (cpuNum = 1; cpuNum < TOTAL_CPUS; cpuNum++) {
         int cycle = CPUVec.OldCycleCount(cpuNum);
         if (cycle < samCurrentTime) {
            samCurrentTime = cycle;
         }
      }
      CPUWarning("Entering SAM at cycle %lld\n", samCurrentTime);
   } else {
      samCurrentTime = 0;
   }

   cpuNum = TOTAL_CPUS - 1;

   while (1) {
      if (readyToSwitch == TOTAL_CPUS) {
         return;
      }

      cpuNum++;
      if (cpuNum == TOTAL_CPUS) {
         cpuNum = 0;
         samCurrentTime++;
         EventPollSingleQueue(samCurrentTime);
      }
      curSP = SPE[cpuNum];
      curPE = curSP->PE;

      /* Sam main loop: run in reverse order to prevent instructions
         from completing several pipeline stages in a single cycle */

      if (!SamInterruptPending(curSP)) {
         /*    Pull in highest priority interrupt, if any */
         EV5_CheckForInterrupts(curPE, "SamCPURun");
      }

      SamPrintCycle(curSP);

      SamGraduate(curSP);
      SamExecute(curSP);
      SamMemory(curSP);
      SamIssue(curSP);
      SamPrintIwin(curSP);
      SamFetch(curSP);
   }
}

static void
SamDoneRunning(void)
{
   CPUWarning("Exiting the SAM simulator \n");
}

void
ExecuteSam(void)
{
   int cpuNum;
   
   SamEnter();

   GammaInitTables();
   SamCPUVectorInit();
   
   if (!alphaLateInitDone) {
      /*
       * Stuff that needs late initalization.
       * XXX should use a register mechanism
       */
      
      SamStatsInit();
      AnnCommonSetup();
      AnnFMInit(128);
      InstallPoller();  /* devices */
      HWEventsLateInit();
      
      /*
       * simos enter annotation
       */
      
      for (cpuNum = 0; cpuNum < TOTAL_CPUS; cpuNum++) {
         curSP = SPE[cpuNum];
         curPE = curSP->PE;
         AnnExec(AnnFind("simos","enter"));
      }

      curSP = SPE[0];
      curPE = curSP->PE;
      
      samCurrentTime = -1;
      alphaLateInitDone = 1;
   }

   SamFPInit();
   SamCPURun();

   SamDoneRunning();
   simosCPUType = exitToSimulator;
   SamExit();
   SamStatsDumpAll();
}

void
SamFreeInst(SamState *sp, SamIwinEntry *ip, bool spec)
{
   int i;
   if (spec) {
      /* Clear dependencies of this instruction on other instructions */
      if ((ip->ra.reg != NULL) && (ip->rdest != &ip->ra)) {
         SamIwinEntry *oip = ip->ra.reg->owner;
         if (oip != NULL) {
            for (i = 0; i < oip->depCount; i++) {
               if (oip->depList[i] == ip) {
                  /* Remove from dependent list */
                  oip->depList[i] = oip->depList[--oip->depCount];
               }
            }
         }
      }
      if ((ip->rb.reg != NULL) && (ip->rdest != &ip->rb)) {
         SamIwinEntry *oip = ip->rb.reg->owner;
         if (oip != NULL) {
            for (i = 0; i < oip->depCount; i++) {
               if (oip->depList[i] == ip) {
                  /* Remove from dependent list */
                  oip->depList[i] = oip->depList[--oip->depCount];
               }
            }
         }
      }
      /* Update target register */
      if (ip->iwinFlags & IwinDefine) {
         ip->rdest->reg->status &= ~RegInWin;
         SamCheckRegFree(ip->rdest->reg, ip->rdest->type, sp);
      }
   }

   ASSERT(ip->iwinFlags & IwinBusy);
   ASSERT(!(ip->iwinFlags & IwinFreed));

   ip->iwinFlags |= IwinFreed;
   SamPrintStage(sp, ip, "RB");

   if ((ip->ra.reg != NULL) && (ip->rdest != &ip->ra)) {
      ip->ra.reg->ref--;
      SamCheckRegFree(ip->ra.reg, ip->ra.type, sp);
   }
   if ((ip->rb.reg != NULL) && (ip->rdest != &ip->rb)) {
      ip->rb.reg->ref--;
      SamCheckRegFree(ip->rb.reg, ip->rb.type, sp);
   }
}

static void
SamDeissueInst(SamState *sp, SamIwinEntry *ip)
{
   if (ip->iwinFlags & IwinFreed) {
      if (ip->iwinFlags & IwinLdstBuf) {
         SamLdstBufferSquash(sp, ip);
      }
   } else {
      bool willBeFreed = FALSE;
      if (!(ip->iwinFlags & IwinExecuted)) {
         if (ip->iwinFlags & IwinIssued) {
            /* Remove from executeQueue */
            List_RemoveDebug(IwinToList(ip));
         } else {
            /* Remove from issueQueue */
            SamIssueDequeue(ip, sp);
         }
      }
      if (ip->iwinFlags & IwinLdst) {
         if (ip->iwinFlags & IwinIssued) {
            if (ip->iwinFlags & IwinLdstBuf) {
               willBeFreed = ((ip->ls->status & LdstDone) && 
                              (ip->instFlags & InstStore) &&
                              !(ip->instFlags & InstSC));
               SamLdstBufferSquash(sp, ip);
            } else if (ip->ls != NULL) {
               SamLdstBufferRelease(sp, ip);
            }
         } else {
            List_RemoveDebug(IwinToLdstList(ip));
         }
      }
      if (!willBeFreed) {
         SamFreeInst(sp, ip, TRUE);
      }
   }
   if (ip->iwinFlags & IwinScbdHi) {
      byte scbd = ~(ip->instr.word & 0xf0);
      sp->scoreboard &= ~scbd;
      ip->iwinFlags &= ~IwinScbdHi;
   }
   if (ip->iwinFlags & IwinScbdLo) {
      byte scbd = ~(ip->instr.word & 0x0f);
      sp->scoreboard &= ~scbd;
      ip->iwinFlags &= ~IwinScbdLo;
   }
}

void
SamSquashInst(SamState *sp, SamIwinEntry *ip)
{
   SamDeissueInst(sp, ip);
   ip->thread = NULL;
   ip->iwinFlags &= ~IwinSpec;
   ip->iwinFlags |= IwinSquash;

}
void
SamFlushInst(SamState *sp, SamIwinEntry *ip)
{
   /* rpb: handle trap flush like squash for now, 
    * but don't clear ip->thread (needed to update register names) */
   SamDeissueInst(sp, ip);
   ip->iwinFlags &= ~IwinSpec;
}

void
SamRegWriteback(SamState *sp, SamIwinEntry *ip)
{
   /* Keep R31/F31 0 */

   if ((ip->rdest != NULL) && (ip->rdest->areg == ZERO_REGISTER)) {
      if (ip->rdest->type == RegInt) {
         ip->rdest->reg->contents.gpr = 0;
      } else {
         ip->rdest->reg->contents.fpr = 0;
      }
   }

   if (ip->iwinFlags & IwinDefine) {
      int i;
      SamRegState *rp = ip->rdest->reg;
      ASSERT(rp != NULL);
      ASSERT(rp->owner == ip);
      
      rp->status &= ~RegInWin;
      SamCheckRegFree(rp, ip->rdest->type, sp);
      
      /* Clear dependencies of instructions waiting for this result */
      for (i = 0; i < ip->depCount; i++) {
         SamIwinEntry *dip = ip->depList[i];
         if (dip->ra.dep == ip) {
            dip->iwinFlags &= ~IwinDepA;
         }
         if (dip->rb.dep == ip) {
            dip->iwinFlags &= ~IwinDepB;
         }
         SamCheckInstAvail(dip, sp);
      }
   }

   SamFreeInst(sp, ip, FALSE);
}

/*
 * Floating-point support
 */

static INLINE bool
SamFPCRIndicatesTrap(SamIwinEntry *ip)
{
   if ((ip->pc & 1) == 0) {
      if ((ip->fpcr & FPCR_INV) && !(ip->fpcr & FPCR_INVD))
         return TRUE;
      if ((ip->fpcr & FPCR_DZE) && !(ip->fpcr & FPCR_DZED))
         return TRUE;
      if ((ip->fpcr & FPCR_OVF) && !(ip->fpcr & FPCR_OVFD))
         return TRUE;
      if ((ip->fpcr & FPCR_UNF) && !(ip->fpcr & FPCR_UNFD))
         return TRUE;
      if ((ip->fpcr & FPCR_INE) && !(ip->fpcr & FPCR_INED))
         return TRUE;
      if ((ip->fpcr & FPCR_IOV))
         return TRUE;
   }
   return FALSE;
}


/* SamSigfpeHandler: handle exceptions raised by directr opcode execution */

static void 
SamSigfpeHandler(int signal, siginfo_t *psiginfo, struct sigcontext *ctxt)
{
   SamState *sp = curSP;
   SamIwinEntry *ip = sp->executing;
   long code = psiginfo->si_code;

   switch (code) {
   /* int overflow, but caused by a floating point instruction */
   case FPE_INTOVF: ip->fpcr |= FPCR_IOV; break;
   case FPE_FLTDIV: ip->fpcr |= FPCR_DZE; break;
   case FPE_FLTOVF: ip->fpcr |= FPCR_OVF; break;
   case FPE_FLTUND: ip->fpcr |= FPCR_UNF; break;
   case FPE_FLTRES: ip->fpcr |= FPCR_INE; break;
   case FPE_FLTINV: ip->fpcr |= FPCR_INV; break;

   /*
    * integer division by zero is weird, since it tends to be caught by
    * the divide subroutine which uses a PAL_GENTRAP to fake a trap.
    * There is no FPCR bit for this condition, but checking a real FPCR
    * after a PAL_GENTRAP(GEN_INTDIV) shows that the INE bit is set, so
    * I do that here.
    */
   case FPE_INTDIV: ip->fpcr |= FPCR_INE; break;
      
   /*
    * some VAX VMS migration documentation states that there is no
    * Alpha equivalent.  I can't find any references in the source code.
    */
   case FPE_FLTSUB:
   default: NOTREACHED();
   }

   if ((curPE->operating_flags & AXPTYPE_CANTRAP) == 0) {
      CPUWarning("@@@@ SamSigfpeHandler: CANTRAP flag not set!  PC 0x%lx, instr 0x%x, "
		 "flags: 0x%x\n", ip->pc, ip->instr.word, curPE->operating_flags);
   }

   if (SamFPCRIndicatesTrap(ip)) { 
      CPUVec.Exception(sp->PE->myNum, TRAP_ARITH);
   } else {
      CPUError("In SamSigfpeHandler. No trap\n");
   }
}

static void
SamFPInit(void)
{
   struct sigaction action;
   struct sigaction oaction;

   action.sa_handler = (void(*)(int))SamSigfpeHandler;
   action.sa_mask = 0;
   action.sa_flags = SA_SIGINFO;
   sigaction(SIGFPE, &action, &oaction);
}

/*
 * Instruction and functional unit specifications
 */

/* Table of instruction info: name, decode function, execute function */

SamInstructionInfo samInstructionTable[] = {
   { "call_pal", SamDecodePAL,         SamExecutePAL         },
   { "opc01",    SamDecodeIllegal,     SamExecuteIllegal     },
   { "opc02",    SamDecodeIllegal,     SamExecuteIllegal     },
   { "opc03",    SamDecodeIllegal,     SamExecuteIllegal     },
   { "opc04",    SamDecodeIllegal,     SamExecuteIllegal     },
   { "opc05",    SamDecodeIllegal,     SamExecuteIllegal     },
   { "opc06",    SamDecodeIllegal,     SamExecuteIllegal     },
   { "opc07",    SamDecodeIllegal,     SamExecuteIllegal     },
   { "lda",      SamDecodeLoadAddress, SamExecuteLoadAddress },
   { "ldah",     SamDecodeLoadAddress, SamExecuteLoadAddress },
#if defined(op_ldbu)
   { "ldbu",     SamDecodeMemLoad,     SamExecuteMemLoad     },
#else
   { "ldbu",     SamDecodeIllegal,     SamExecuteIllegal     },
#endif
   { "ldq_u",    SamDecodeMemLoad,     SamExecuteMemLoad     },
#if defined(op_ldwu)
   { "ldwu",     SamDecodeMemLoad,     SamExecuteMemLoad     },
#else
   { "ldwu",     SamDecodeIllegal,     SamExecuteIllegal     },
#endif
#if defined(op_stw)
   { "stw",      SamDecodeMemStore,    SamExecuteMemStore    },
#else
   { "stw",      SamDecodeIllegal,     SamExecuteIllegal     },
#endif
#if defined(op_stb)
   { "stb",      SamDecodeMemStore,    SamExecuteMemStore    },
#else
   { "stb",      SamDecodeIllegal,     SamExecuteIllegal     },
#endif
   { "stq_u",    SamDecodeMemStore,    SamExecuteMemStore    },
   { "inta",     SamDecodeOperate,     SamExecuteOperate     },
   { "intl",     SamDecodeOperate,     SamExecuteOperate     },
   { "ints",     SamDecodeOperate,     SamExecuteOperate     },
   { "intm",     SamDecodeOperate,     SamExecuteOperate     },
   { "fltm",     SamDecodeFloating,    SamExecuteFloating    },
   { "fltv",     SamDecodeFloating,    SamExecuteFloating    },
   { "flti",     SamDecodeFloating,    SamExecuteFloating    },
   { "fltl",     SamDecodeFloating,    SamExecuteFloating    },
   { "misc",     SamDecodeMisc,        SamExecuteMisc        },
   { "MFPR",     SamDecodePAL,         SamExecutePAL         },
   { "jsr",      SamDecodeJump,        SamExecuteJump        },
   { "HW_LD",    SamDecodePAL,         SamExecutePAL         },
   { "intmisc",  SamDecodeOperate,     SamExecuteOperate     },
   { "MTPR",     SamDecodePAL,         SamExecutePAL         },
   { "REI",      SamDecodePAL,         SamExecutePAL         },
   { "HW_ST",    SamDecodePAL,         SamExecutePAL         },
   { "ldf",      SamDecodeMemLoad,     SamExecuteMemLoad     },
   { "ldg",      SamDecodeMemLoad,     SamExecuteMemLoad     },
   { "lds",      SamDecodeMemLoad,     SamExecuteMemLoad     },
   { "ldt",      SamDecodeMemLoad,     SamExecuteMemLoad     },
   { "stf",      SamDecodeMemStore,    SamExecuteMemStore    },
   { "stg",      SamDecodeMemStore,    SamExecuteMemStore    },
   { "sts",      SamDecodeMemStore,    SamExecuteMemStore    },
   { "stt",      SamDecodeMemStore,    SamExecuteMemStore    },
   { "ldl",      SamDecodeMemLoad,     SamExecuteMemLoad     },
   { "ldq",      SamDecodeMemLoad,     SamExecuteMemLoad     },
   { "ldl_l",    SamDecodeMemLoad,     SamExecuteMemLoad     },
   { "ldq_l",    SamDecodeMemLoad,     SamExecuteMemLoad     },
   { "stl",      SamDecodeMemStore,    SamExecuteMemStore    },
   { "stq",      SamDecodeMemStore,    SamExecuteMemStore    },
   { "stl_c",    SamDecodeMemStore,    SamExecuteMemStore    },
   { "stq_c",    SamDecodeMemStore,    SamExecuteMemStore    },
   { "br",       SamDecodeBranch,      SamExecuteBranch      },
   { "fbeq",     SamDecodeBranch,      SamExecuteBranch      },
   { "fblt",     SamDecodeBranch,      SamExecuteBranch      },
   { "fble",     SamDecodeBranch,      SamExecuteBranch      },
   { "bsr",      SamDecodeBranch,      SamExecuteBranch      },
   { "fbne",     SamDecodeBranch,      SamExecuteBranch      },
   { "fbge",     SamDecodeBranch,      SamExecuteBranch      },
   { "fbgt",     SamDecodeBranch,      SamExecuteBranch      },
   { "blbc",     SamDecodeBranch,      SamExecuteBranch      },
   { "beq",      SamDecodeBranch,      SamExecuteBranch      },
   { "blt",      SamDecodeBranch,      SamExecuteBranch      },
   { "ble",      SamDecodeBranch,      SamExecuteBranch      },
   { "blbs",     SamDecodeBranch,      SamExecuteBranch      },
   { "bne",      SamDecodeBranch,      SamExecuteBranch      },
   { "bge",      SamDecodeBranch,      SamExecuteBranch      },
   { "bgt",      SamDecodeBranch,      SamExecuteBranch      }
};

/* Functional unit specifications */

static bool
SamFixedUnit(SamIwinEntry *ip, int *latency, int *repeatRate)
{
   *latency = FIXED_UNIT_LATENCY;
   *repeatRate = FIXED_UNIT_REPEAT_RATE;
   return TRUE;
}

static bool
SamUnifiedUnit(SamIwinEntry *ip, int *latency, int *repeatRate)
{
   if (ip->instFlags & InstMemory) {
      return SamEV6LSU(ip, latency, repeatRate);
   } else if (ip->instFlags & InstFloating) {
      return SamEV6FPU(ip, latency, repeatRate);
   } else {
      return SamEV6ALU(ip, latency, repeatRate);
   }
}

static bool
SamEV6ALU(SamIwinEntry *ip, int *latency, int *repeatRate)
{
   if ((ip->instFlags & InstMemory) || (ip->instFlags & InstFloating)) {
      return FALSE;
   } else {
      *latency = ALU_LATENCY;
      *repeatRate = ALU_REPEAT_RATE;
      return TRUE;
   }
}

static bool
SamEV6LSU(SamIwinEntry *ip, int *latency, int *repeatRate)
{
   if (ip->instFlags & InstMemory) {
      *latency = LSU_LATENCY;
      *repeatRate = LSU_REPEAT_RATE;
      return TRUE;
   } else {
      return FALSE;
   }
}

static bool
SamEV6FPU(SamIwinEntry *ip, int *latency, int *repeatRate)
{
   if (ip->instFlags & InstFloating) {
      *latency = FPU_LATENCY;
      *repeatRate = FPU_REPEAT_RATE;
      return TRUE;
   } else {
      return FALSE;
   }
}
