/*
 * sam.h -- the primary Sam data structures and functions
 */

#ifndef SAM_H
#define SAM_H

#include "list.h"
#include "sim_error.h"
#include "simtypes.h"
#include "machine_params.h"
#include "alpha.h"
#include "register.h"
#include "thread.h"
#include "iwin.h"
#include "ldst.h"
#include "funcunit.h"
#include "histogram.h"
#include "sam_params.h"

#ifdef DIGITAL_UNIX_CC
#define INLINE
#else
#define INLINE inline
#endif

typedef struct {
   int bits;
   VA target;
} SamBPEntry;

typedef struct {
   long correct;
   long incorrect;
   long unpredicted;
} SamBPStats;

typedef struct {
   /* Pointer to architectural state */
   AlphaState *PE;

   /* Physical register state */
   SamRegState gpr[SAM_MAX_GPRS];
   SamRegState fpr[SAM_MAX_FPRS];
   List_Links *gprFreeList;
   List_Links *fprFreeList;

   /* Thread state */
   List_Links *activeThreads;
   List_Links *freeThreads;
   SamThreadState thread[SAM_MAX_THREADS];

   /* Branch prediction state */
   SamBPEntry branchPredictionTable[SAM_MAX_BP_TABLE_SIZE];
   Reg branchStack[SAM_MAX_BP_RETURN_STACK];
   int branchSP;

   /* Instruction window state */
   List_Links *reorderBuffer;
   List_Links *ldstQueue;
   List_Links *issueQueue;
   List_Links *executeQueue;
   List_Links *freeIwinEntries;
   SamIwinEntry iwin[SAM_MAX_IWIN_ENTRIES];

   /* Number of issue slots for the in-order processor model */
   int issueSlots;

   /* Load/store buffer state */
   List_Links *activeLdstEntries;
   List_Links *freeLdstEntries;
   SamLdstEntry ldstbuf[SAM_MAX_LDST_ENTRIES];

   /* Functional unit state */
   int numFuncUnits;
   SamFuncUnit funcUnit[SAM_MAX_FUNC_UNITS];

   /* IPR Scoreboard bits */
   byte scoreboard;

   /* Exception state */
   SamTrapState lastTrap;
   SamTrapState currentTrap;
   SamTrapState interrupt;

   /* Currently executing instruction (needed for Sigfpe handling) */
   SamIwinEntry *executing;

   /* Flag indicating that we're switching out of Sam */
  enum { NoSwitch, SwitchRequested, SwitchPending } cpuSwitch;

   /* Replay trap state */
   enum { Replay, Wait, Issue } replayStatus;

   /* Keep track of the number of load/store buffers in use */
   int activeLdstBufCount;

   /* Statistics histograms */
   SamHistogram fetchHistogram;
   SamHistogram issueHistogram;
   SamHistogram gradHistogram;
   SamHistogram ldstBufHistogram;

   SamHistogram readMissHistogram;
   SamHistogram readDMissHistogram;
   SamHistogram writeMissHistogram;
   SamHistogram cacheMissHistogram;

   /* Branch prediction statistics */
   enum { BranchTaken, BranchNotTaken, CallTaken, CallNotTaken,
          CallIndirect, ReturnIndirect, JumpIndirect, NumBranchTypes };
   SamBPStats branchStatistics[NumBranchTypes];

   /* Number of empty graduation slots since last graduated instruction */
   int stallSlots;

   /* Type of pipeline bubble occurring on this CPU */
   int bubble;
} SamState;

extern SamState *SPE[SIM_MAXCPUS];
extern SamState *curSP;
extern SimTime samCurrentTime;
extern int samPrintLevel[SIM_MAXCPUS];

/* Globals for switching from Sam to another simulator */
extern int exitToSimulator;
extern int readyToSwitch;

/* Declarations of the main pipeline stage functions */

extern void SamFetch(SamState *sp);
extern void SamDecode(SamState *sp, SamIwinEntry *ip);
extern void SamBranch(SamState *sp, SamIwinEntry *ip);
extern void SamIssue(SamState *sp);
extern void SamExecute(SamState *sp);
extern void SamMemory(SamState *sp);
extern void SamGraduate(SamState *sp);
extern void SamTrap(SamState *sp);

/* Interface to the instruction cache */

extern Result SamRefICache(SamState *sp, VA vAddr, PA pAddr, Inst *inst);

/* Functions for manipulating the load/store buffer */

extern bool SamLdstBufferReserve(SamState *sp, SamIwinEntry *ip);
extern void SamLdstBufferSquash(SamState *sp, SamIwinEntry *ip);
extern void SamLdstBufferRelease(SamState *sp, SamIwinEntry *ip);
extern int  SamLdstBufferRetire(SamState *sp, SamLdstEntry *ls);

extern void SamLdstQueue(SamState *sp, SamIwinEntry *ip, int type,
                         SamRegState *reg, SamRegState *screg);
extern int SamLdstFillLine(SamState *sp, PA pA, int type);
extern void SamLdstFlushLine(SamState *sp, PA pA, int size, int retain);

extern bool SamLdstCheckAlignment(PA pAddr, int type);
extern bool SamLdstGetByte(SamState *sp, PA pAddr, byte *buf);

/* Functions for decode of the CMOV instruction */

extern void SamDecodeCMOV1(SamState *sp, SamIwinEntry *ip);
extern void SamDecodeCMOV2(SamState *sp, SamIwinEntry *ip);

/* Functions for handling branch resolution */

extern void SamBranchResolve(SamState *sp, SamIwinEntry *ip, bool resolution);
extern void SamBranchResolveJump(SamState *sp, SamIwinEntry *ip, VA target);

/* Thread deactivation function: currently used by trap to clean up threads */

extern void SamThreadDeactivate(SamState *sp, SamThreadState *th);

/* Address calculation functions */

extern void SamExecuteAddressCalculation(SamState *sp, SamIwinEntry *ip);
extern void SamExecuteHWAddressCalculation(SamState *sp, SamIwinEntry *ip);

/* Functions for copying register state between AlphaState and SamState */

void SamImportRegs(SamState *sp, SamThreadState *th);
void SamExportRegs(SamState *sp, SamThreadState *th);

/* Function for recording traps when they occur */

void SamTrapRecord(SamState *sp, int trap);

/* Statistics initialization and dumping */

void SamStatsInit(void);
void SamStatsDumpAll(void);

/* Miscellaneous functions, presently found in sam.c */

extern void SamFreeInst(SamState *sp, SamIwinEntry *ip, bool spec);
extern void SamSquashInst(SamState *sp, SamIwinEntry *ip);
extern void SamFlushInst(SamState *sp, SamIwinEntry *ip);
extern void SamRegWriteback(SamState *sp, SamIwinEntry *ip);

#endif /* SAM_H */
