/*
 * sam_params.h -- Sam pipeline configuration, initialized from Tcl
 */

#ifndef SAM_PARAMS_H
#define SAM_PARAMS_H

#include "params.h"

#define ALPHA_NUM_GPRS             32
#define ALPHA_NUM_FPRS             32

#define SAM_MAX_GPRS              128
#define SAM_MAX_FPRS              128
#define SAM_MAX_THREADS             8
#define SAM_MAX_LDST_ENTRIES       64
#define SAM_MAX_IWIN_ENTRIES      128
#define SAM_MAX_FUNC_UNITS         16

#define SAM_MAX_BP_TABLE_SIZE    1024
#define SAM_MAX_BP_RETURN_STACK    32

typedef struct {
   /* Does the pipeline allow out-of-order instruction issue? */
   bool inOrder;
   /* Number of issue slots in the in-order model */
   int numIssueSlots;

   /* Number of system resources available */
   int numGPRs;
   int numFPRs;
   int numInstructionSlots;
   int numThreads;
   int numLdstBuffers;

   /* Pipeline stage bandwidths */
   int fetchWidth;
   int issueWidth;
   int gradWidth;

   /* Generic functional units: handle all instructions */
   int numFixedUnits;
   int fixedUnitLatency;
   int fixedUnitRepeatRate;

   /* EV6-style functional units */
   int numUnifiedUnits;
   int numALUs;
   int ALULatency;
   int ALURepeatRate;
   int numFPUs;
   int FPULatency;
   int FPURepeatRate;
   int numLSUs;
   int LSULatency;
   int LSURepeatRate;

   /* Load-store configuration */
   int loadToUseLatency;

   /* Can stores be retired before the line comes back to the cache? */
   bool retireStoresEarly;

   /* Branch prediction configuration */
   struct {
      int tableSize;
      int returnStack;
      int mispredictLatency;
   } branchPrediction;

} SamParams;

extern SamParams samParams;

#define SAM_IN_ORDER              samParams.inOrder
#define SAM_ISSUE_SLOTS           samParams.numIssueSlots

#define NUM_GPRS                  samParams.numGPRs
#define NUM_FPRS                  samParams.numFPRs
#define NUM_IWIN_ENTRIES          samParams.numInstructionSlots
#define NUM_THREADS               samParams.numThreads
#define NUM_LDST_ENTRIES          samParams.numLdstBuffers

#define FETCH_WIDTH               samParams.fetchWidth
#define ISSUE_WIDTH               samParams.issueWidth
#define GRAD_WIDTH                samParams.gradWidth

#define NUM_FIXED_UNITS           samParams.numFixedUnits
#define FIXED_UNIT_LATENCY        samParams.fixedUnitLatency
#define FIXED_UNIT_REPEAT_RATE    samParams.fixedUnitRepeatRate

#define NUM_UNIFIED_UNITS         samParams.numUnifiedUnits
#define NUM_ALUS                  samParams.numALUs
#define ALU_LATENCY        samParams.ALULatency
#define ALU_REPEAT_RATE    samParams.ALURepeatRate
#define NUM_FPUS                  samParams.numFPUs
#define FPU_LATENCY        samParams.FPULatency
#define FPU_REPEAT_RATE    samParams.FPURepeatRate
#define NUM_LSUS                  samParams.numLSUs
#define LSU_LATENCY        samParams.LSULatency
#define LSU_REPEAT_RATE    samParams.LSURepeatRate

#define SAM_LOAD_TO_USE_LATENCY   samParams.loadToUseLatency
#define SAM_RETIRE_STORES_EARLY   samParams.retireStoresEarly

#define SAM_BP_TABLE_SIZE         samParams.branchPrediction.tableSize
#define SAM_BP_RETURN_STACK       samParams.branchPrediction.returnStack
#define SAM_BP_MISPREDICT_LATENCY samParams.branchPrediction.mispredictLatency

#endif /* SAM_PARAMS_H */
