/*
 * stats.c -- support for dumping periodic stats to the log
 */

#include "sam.h"
#include "stats.h"

#include "simtypes.h"
#include "sim_error.h"

#include "eventcallback.h"
#include "cpu_stats.h"
#include "cpu_interface.h"
#include "memref.h"
#include "hw_events.h"

static struct StatsHdr {
    EventCallbackHdr hdr;
    SimTime statsInterval;
    int periodicStats;
    int intervalsLeft;

} statsEvent[SIM_MAXCPUS];

static void EnablePeriodicStats(int cpuNum, SimTime statInterval);
static INLINE char *SamStats(void) { return ""; }

void
SamStatsInit(void)
{
   MemRefInit();
   InitCPUStats();
   EnablePeriodicStats(0, SAMPLE_PC_INTERVAL);
   InitSimulatorStats(4*1024*1024, SamStats);
   /* rpb hack: used for parsing timing trees */
   CPUPrint("TIMING: SR_STALLSCALE %d\n", GRAD_WIDTH);
}

void
SamStatsDumpAll(void) 
{
   int cpuNum;
   int i;

   CPUPrint("********************DUMPING SAM STATS******************** \n");
   PrintAllCPUStats();
   CPUPrint("************\n");
   CPUPrint("Cache Stats\n");
   CPUPrint("************\n");
   for (cpuNum = 0; cpuNum < TOTAL_CPUS; cpuNum++) {
      CPUPrint("C%d Stall Time: %d\n", cpuNum, STATS_VALUE(cpuNum, stallTime));
      MemRefDumpStats(cpuNum);
   }
   for (cpuNum = 0; cpuNum < TOTAL_CPUS; cpuNum++) {
      CPUPrint("Fetch histogram for CPU %d:\n", cpuNum);
      HistogramDump(SPE[cpuNum]->fetchHistogram);
      CPUPrint("Issue histogram for CPU %d:\n", cpuNum);
      HistogramDump(SPE[cpuNum]->issueHistogram);
      CPUPrint("Graduation histogram for CPU %d:\n", cpuNum);
      HistogramDump(SPE[cpuNum]->gradHistogram);
      CPUPrint("Load/store buffer histogram for CPU %d:\n", cpuNum);
      HistogramDump(SPE[cpuNum]->ldstBufHistogram);
      CPUPrint("Cache miss histogram for CPU %d:\n", cpuNum);
      HistogramDump(SPE[cpuNum]->cacheMissHistogram);
      CPUPrint("Read miss histogram (data/inst) for CPU %d:\n", cpuNum);
      HistogramDump(SPE[cpuNum]->readMissHistogram);
      CPUPrint("Read miss histogram (data only) for CPU %d:\n", cpuNum);
      HistogramDump(SPE[cpuNum]->readDMissHistogram);
      CPUPrint("Write miss histogram for CPU %d:\n", cpuNum);
      HistogramDump(SPE[cpuNum]->writeMissHistogram);
   }

   for (cpuNum = 0; cpuNum < TOTAL_CPUS; cpuNum++) {
      SamState *sp = SPE[cpuNum];
      CPUPrint("Instructions executed by each functional unit for CPU %d:\n", cpuNum);
      for (i = 0; i < sp->numFuncUnits; i++) {
         CPUPrint("Unit %d (%s): %lld\n", i, 
                  sp->funcUnit[i].typeString,
                  sp->funcUnit[i].numExecuted);
      }
   }

   for (cpuNum = 0; cpuNum < TOTAL_CPUS; cpuNum++) {
      SamState *sp = SPE[cpuNum];
      CPUPrint("Branch prediction statistics for CPU %d:\n", cpuNum);
      for (i = 0; i < NumBranchTypes; i++) {
         SamBPStats *bpStats = sp->branchStatistics + i;
         float total = 
            bpStats->correct + bpStats->incorrect + bpStats->unpredicted;

         switch (i) {
         case BranchTaken:    CPUPrint("BranchTaken:    "); break;
         case BranchNotTaken: CPUPrint("BranchNotTaken: "); break;
         case CallTaken:      CPUPrint("CallTaken:      "); break;
         case CallNotTaken:   CPUPrint("CallNotTaken:   "); break;
         case CallIndirect:   CPUPrint("CallIndirect:   "); break;
         case ReturnIndirect: CPUPrint("ReturnIndirect: "); break;
         case JumpIndirect:   CPUPrint("JumpIndirect:   "); break;
         }
         CPUPrint("%8ld correct, %8ld incorrect, %8ld unpredicted",
                  bpStats->correct, bpStats->incorrect, bpStats->unpredicted);
         if (total > 0) {
            float successRate = 100.0 * (float) bpStats->correct / total;
            CPUPrint(" -- %0.1f%%", successRate);
         }
         CPUPrint("\n");
      }
   }
}

static void
PeriodicStats(SimTime cycleCount)
{
   int cpuNum;

   PrintPeriodicCPUStats();
   for (cpuNum = 0; cpuNum < TOTAL_CPUS; cpuNum++) {
      MemRefPeriodicStats(cpuNum);
   }
   memsysVec.MemsysDumpStats();
}

static void 
SamStatsCallback(int cpuNum, EventCallbackHdr *hdr, void *arg)
{
   int cpu;
   for (cpu = 0; cpu < TOTAL_CPUS; cpu++) {
      /* rpb: using the current PE->PC might not be right.... */
      CYCLE_SAMPLE_EVENT(CPUVec.CycleCount(cpu), cpu, SPE[cpu]->PE->PC);
   }
   
   if (--statsEvent[cpuNum].intervalsLeft <= 0) { 
      /* Things to do each STAT_INTERVAL */
      statsEvent[cpuNum].intervalsLeft = statsEvent[cpuNum].periodicStats;
      PERIODIC_EVENT;
      PeriodicStats((SimTime)CPUVec.CycleCount(cpuNum));
   }
   
   EventDoCallback(cpuNum, SamStatsCallback,
                   (EventCallbackHdr *)&statsEvent[cpuNum],
                   0, statsEvent[cpuNum].statsInterval);
}

static void 
EnablePeriodicStats(int cpuNum, SimTime statInterval)
{
   statsEvent[cpuNum].statsInterval = statInterval;
   statsEvent[cpuNum].periodicStats = STAT_INTERVAL / statInterval;
   statsEvent[cpuNum].intervalsLeft = statsEvent[cpuNum].periodicStats;
   
   EventDoCallback(0, SamStatsCallback, 
                   (EventCallbackHdr *)&statsEvent[cpuNum],
                   0, statInterval);
}
