/*
 * trap.c -- Sam trap handling code, executed during grad stage
 */

/* #define TRAP_DEBUG */

#include "sam.h"
#include "ev5.h"
#include "ev5_ipr.h"
#include "hw_events.h"

static INLINE bool 
SamIsInterruptable(SamState *sp, SamIwinEntry *ip)
{
   /* SAM actually executes uncached operations from SamMemory().
    * The operation is not executed until it reaches the head of 
    * the re-order buffer, and therefore is about to graduate. The
    * instruction will graduate on the next cycle, when SamGraduate
    * is next executed.
    * A bug occured when we issued a memory operation and then an
    * interrupt occurred on the next cycle. The interrupt handler
    * routine would incorrectly flush the uncached operation, thereby
    * causing it to be eventually re-issued. This is incorrect as
    * uncached operations can have side-effects. Technically, the
    * uncached operation should be launched directly in SamGraduate()
    * where its execution and concurrent interrupts can be managed. 
    * Instead of moving the execution code up to SamGraduate, I changed
    * this function to make the cpu as non-interruptible whenever
    * an uncached operation is at the head of the re-order buffer (and
    * therefore has just been executed). Previously, we did this for
    * uncached stores only. RJS 11/12/01
    */
#if 0
   if ((ip->iwinFlags & IwinLdst) && (ip->iwinFlags & IwinLdstBuf) &&
       (ip->iwinFlags & IwinUncached) && !(ip->instFlags & InstStore) &&
       ip->ls->dataPtr) 
#else
   if ((ip->iwinFlags & IwinLdst) && (ip->iwinFlags & IwinLdstBuf) &&
       (ip->iwinFlags & IwinUncached)) 
#endif
   {
      /* Don't allow interrupt during unfinished uncached reads */
      return FALSE;
   } else if (ip->instFlags & InstCMOV2) {
      /* Don't allow interrupt between the two halves of the CMOV instruction */
      return FALSE;
   } else if (ip->instFlags & InstREI) {
      /* EV5_HwRei can post an interrupt at execution time, but we 
       * don't want to take it until the hw_rei graduates; by preventing
       * hw_rei from being interrupted, this should work correctly */
      return FALSE;
   } else {
      return TRUE;
   }
}

void
SamTrap(SamState *sp)
{
   bool empty = List_IsEmpty(sp->reorderBuffer);
   SamIwinEntry *ip = NULL;
   SamThreadState *th;
   SamGetGradThread(th, sp);
   if (!empty) {
      ip = ReorderListToIwin(List_First(sp->reorderBuffer));
      sp->PE->PC = ip->pc;
   } else {
      sp->PE->PC = th->pc;
   }

   /* Check for ITB fault */
   if (!SamTrapPending(sp) && empty && (th->stallFetch & StallITB)) {
      SamTrapCopy(th->trap, sp->currentTrap);
   }

   /* Check for simulator switch */
   if (!SamTrapPending(sp) && (sp->cpuSwitch == SwitchRequested)) {
      if (empty || SamIsInterruptable(sp, ip)) {
          SamTrapRecord(sp, TrapSwitch);
          SamTrapCopy(sp->lastTrap, sp->currentTrap);
          sp->cpuSwitch = SwitchPending;
      }
   }

   /* Check for interrupts */
   if (!SamTrapPending(sp) && SamInterruptPending(sp) && !IS_PAL(sp->PE)) {
      if (empty || SamIsInterruptable(sp, ip)) {
         SamTrapCopy(sp->interrupt, sp->currentTrap);
      }
   } 
   
   /* If the pipeline is empty, handle the pending trap */
   if (SamTrapPending(sp) && empty) {
      /* There may be multiple active threads: get rid of all but one */
      while (List_First(sp->activeThreads) != List_Last(sp->activeThreads)) {
         th = ListToThread(List_Last(sp->activeThreads));
         SamThreadDeactivate(sp, th);
      }
      SamGetFetchThread(th, sp);

      /* Store correct PC value for return from interrupt */
      if (sp->currentTrap.trap == TRAP_INTERRUPT) {
         sp->currentTrap.pc = th->pc;
      }

      if (sp->currentTrap.trap == TrapReplay) {
         sp->replayStatus = Replay;
      } else {
         sp->replayStatus = Issue;
      }

      switch (sp->currentTrap.trap) {
      case TrapReplay:
         /* Don't need to do anything, just restart the pipeline */
         break;
      case TrapSwitch:
         readyToSwitch++;
         break;
      default:
         ASSERT(sp->currentTrap.trap != TrapNone);
         /* If there's an interrupt still pending, blow it away */
         if (SamInterruptPending(sp)) {
            SamTrapFree(sp->interrupt);
         }
         /* Set up PE state for this trap and call out to EV5 code */
         freeTrapState(sp->PE->trapState);
         sp->PE->trapState = allocTrapState(sp->currentTrap.state);
         sp->PE->PC = sp->currentTrap.pc;
         SamExportRegs(sp, th);
#ifdef TRAP_DEBUG
         CPUPrint("TRAP: type = 0x%x, PC = 0x%lx [%ld:%d]\n", 
                  sp->currentTrap.trap, sp->PE->PC,
                  samCurrentTime, sp->PE->myNum);
#endif
         EV5_Trap(sp->PE, sp->currentTrap.trap);
         SamImportRegs(sp, th);
         th->pc = sp->PE->PC;
         break;
      }

      th->stallFetch = StallNone;
      SamTrapFree(sp->currentTrap);
      sp->currentTrap.state = allocTrapState(sp->PE->trapState);
      sp->bubble = E_PIPE_TRAP;
   }
}

/* SamTrapRecord: invoked by CPUVec.Exception() when trap occurs */

void
SamTrapRecord(SamState *sp, int trap)
{
   if (trap == TRAP_INTERRUPT) {
      /* Special case: interrupt comes from EV5, not Sam, so save it away */
      ASSERT(sp->interrupt.state == NULL);
      sp->interrupt.pc = sp->PE->PC;
      sp->interrupt.trap = trap;
      sp->interrupt.state = sp->PE->trapState;
   } else {
      ASSERT(sp->lastTrap.state == NULL);
      sp->lastTrap.pc = sp->PE->PC;
      sp->lastTrap.trap = trap;
      sp->lastTrap.state = sp->PE->trapState;
   }
   sp->PE->trapState = allocTrapState(sp->currentTrap.state);
}
