/*
 * Copyright (C) 1998 by the Board of Trustees
 *    of Leland Stanford Junior University.
 * Copyright (C) 1998 Digital Equipment Corporation
 *
 * This file is part of the SimOS distribution.
 * See LICENSE file for terms of the license.
 *
 */



#include <stdlib.h>
#include <stddef.h>

#include "simtypes.h"
#include "sim_error.h"
#include "machine_params.h"
#include "machine_defs.h"
#include "registry.h"
#include "console.h"
#include "hd.h"
#include "alpha_access.h"
#include "cpu_interface.h"
#include "gamma.h"
#include "tlaser.h"
#include "ev5.h"
#include "console.h"

#include "tlaser.h"
#include "tlaser_clock.h"
#include "kbd_mouse.h"

#include "ev5_ipr.h"
#include "simether.h"


/* See registry.h for explanation of this table */ 
char *bdoorType[] = 
{ "NIL", "BAD", "LOAD08", "STORE08",
  "LOAD16", "STORE16", "BAD", "BAD",
  "LOAD32", "STORE32", "BAD", "BAD",
  "BAD", "BAD", "BAD", "BAD",
  "LOAD64", "STORE64" };


#define VA_NODE(va) (((va) >> __MAGIC_NODE_OFFS) & \
		     ((1<<__MAGIC_NODE_BITS)-1))
#define VA_ZONE(va) (((va) >> __MAGIC_ZONE_OFFS) & \
		     ((1<<__MAGIC_ZONE_BITS)-1))
#define VA_REG(va) (((va) >> __MAGIC_REG_OFFS) & \
		    ((1<<__MAGIC_REG_BITS)-1))
#define VA_GRP(va) (((va) >> __MAGIC_PPC_GRP_OFFS) & \
		    ((1<<__MAGIC_PPC_GRP_BITS)-1))
#define VA_OPC(va) (((va) >> __MAGIC_PPC_OPC_OFFS) & \
		    ((1<<__MAGIC_PPC_OPC_BITS)-1))
#define VA_SEQ(va) (((va) >> __MAGIC_PPC_SEQ_OFFS) & \
		    ((1<<__MAGIC_PPC_SEQ_BITS)-1))
#define VA_OFFS(va) ((va) & ((1<<__MAGIC_ZONE_OFFS)-1))

#define ZONE_SZ (1 << __MAGIC_ZONE_OFFS)




/*
 * EV5_access.
 * Backdoor way of passing configuration from SimOS to the console
 * Used only by the console, not by the OS
 * 
 * XXX
 * XXX Should eventually be replaced by a structure written by the 
 * XXX simulator at the top of memory, a-la Linux and pass only the
 * XXX memory size as an argument to the console. 
 */

AlphaAccess alphaAccessBlock[MAX_MACHINES];
AlphaAccess *alphaAccessArray = alphaAccessBlock;
AlphaAccess *alphaAccess;

static int EV5_access(int cpuNum, VA va, int type, long *buff)
{
   uint off = VA_OFFS(va);
   VA ra;
   int machine = M_FROM_CPU(cpuNum);
   int firstCPU = FIRST_CPU(machine);
   
   alphaAccess = alphaAccessArray + machine;

   CPUVec.GetRegister(cpuNum, REG_RA, &ra);
   ASSERT( VA_ZONE(va) ==MAGIC_ZONE_EV5_ALIAS);
   
   if (type==BDOOR_LOAD_DOUBLE || type==BDOOR_LOAD_WORD) {
      if (off==offsetof(AlphaAccess,inputChar)) {
         *(Reg*)buff = sim_console_simple_in(machine);
         return 0;
      }
      *(long*)buff = *(int*)(VA_OFFS(va)+(char*)alphaAccess);
      CPUPrint("EV5_Access: va=%llx offset=%lx val=%lx\n",
               va, VA_OFFS(va), *(long*)buff);
      return 0;
   }
   
   if (type==BDOOR_STORE_DOUBLE || type==BDOOR_STORE_WORD) {
      if (off==offsetof(AlphaAccess,diskUnit)) {
         alphaAccess->diskUnit = *buff;
         return 0;
      }
      if (off==offsetof(AlphaAccess,diskCount)) {
         alphaAccess->diskCount = *buff;
         return 0;
      }
      if (off==offsetof(AlphaAccess,diskPAddr)) {
         alphaAccess->diskPAddr = *buff;
         return 0;
      }
      if (off==offsetof(AlphaAccess,diskBlock)) {
         alphaAccess->diskBlock = *buff;
         return 0;
      }
      
      if (off==offsetof(AlphaAccess,diskOperation)) {
         alphaAccess->diskOperation = 
            sim_disk_simpleIO(cpuNum, *buff, 
                              alphaAccess->diskUnit,
                              alphaAccess->diskCount,
                              alphaAccess->diskPAddr,
                              alphaAccess->diskBlock);
         return 0;
      }
      
      if (off==offsetof(AlphaAccess,outputChar)) {
         sim_console_simple_out(machine, *(Reg*)buff);
         return 0;
      }
      
      if (off==offsetof(AlphaAccess,bootStrapImpure)) {
	 alphaAccess->bootStrapImpure = *buff;
	 return 0;
      }
      
      if (off==offsetof(AlphaAccess,bootStrapCPU)) {
	 int cpu = *buff;
	 ASSERT(cpu>=1 && cpu < NUM_CPUS(machine));
	 PE[firstCPU + cpu]->reg[REG_A0]  =  cpu;
	 PE[firstCPU + cpu]->ipr[IPR_PAL_WHAMI] = cpu;
	 PE[firstCPU + cpu]->reg[0] = cpu; /* this one is the one really needed */
	 PE[firstCPU + cpu]->reg[REG_SP]  = alphaAccess->bootStrapImpure;
	 CPUWarning("AlphaAccess. machine %d launching %d at %lx \n",machine, cpu,
                    alphaAccess->bootStrapImpure);
         
	 CPUVec.LaunchSlaveCPU(firstCPU + cpu);
	 return 0;
      }

      if ((off >= offsetof(AlphaAccess,envName)) &&
          (off <= offsetof(AlphaAccess,setEnv))) {
         
         CPUWarning("Env_Access:  PC 0x%lx RA 0x%lx va=0x%lx type=%s \n",
                    PE[cpuNum]->PC,
                    ra,
                    va,bdoorType[type]);  
         return 0;
      }
   }
   
   CPUError("EV5_access:  PC 0x%lx RA 0x%lx va=0x%lx type=%s \n",
	    PE[cpuNum]->PC,
	    ra,
	    va,bdoorType[type]);  

   return 0;
}



static void
console_interrupt(int n, int on)
{
#if 0
   CPUWarning("\n   %ld  Console interrupt %d %d \n",(uint64)CPUVec.CycleCount(0),n,on);
#endif
   int cpu = FIRST_CPU(n);

   if (on) {
      EV5_SetExtInterrupt(PE[cpu],INTLEVEL_IRQ0);
   } else {
      EV5_ClearExtInterrupt(PE[cpu],INTLEVEL_IRQ0);
   }
}



static EventCallbackHdr pollHdr;
#define DEVICE_POLL_INTERVAL (10*1000)
static void PollDevices(int cpuNum, EventCallbackHdr *hdr, void *arg)
{

  sim_console_poll();
  Simether_Poll();
  /* register callback for next poll round */
  EventDoCallback(0, PollDevices, &pollHdr, NULL, DEVICE_POLL_INTERVAL);
}


/* ****************************************************************
 * Late Initialization of Devices. Allowed to install
 * callbacks,... but not new entries in the registry
 * ****************************************************************/
  
void InstallPoller(void)
{
   int count;
   extern int restoringCpt;   /* XXX bugnion gross */
   static int installed = 0;
   if (!installed) {
      installed = 1;
      EventDoCallback(0, PollDevices, &pollHdr, NULL, DEVICE_POLL_INTERVAL);
      
      /*
       * Console is disabled by default, enabled here 
       */
      for (count = 0; count < TOTAL_CONSOLES; count++)
      {
         sim_console_int_set(count,CONS_INT_TX|CONS_INT_RX);
      }
      InitTLaserClock(restoringCpt);
   }
}



/* **************************************************************
 * Initialization of all the I/O space of the machine
 * This is:
 *   -Turbolaser devices
 *   -EV5_access function for the console
 *   -XXM stuff (only for the console I/O: to be removed
 * ************************************************************* */

void InitDevices(int restoreFromChkpt) 
{
   int n = 0;
   int machine;
   char name[64]; 

   for (machine = 0; machine < MAX_MACHINES; machine++)
   {
      alphaAccessArray[machine].diskUnit = 1; /* By default, use the first disk */
#ifdef NT_FRAMEBUFFER
    KbdMouse_Initialize(machine);
#endif
   }
   
   RegistryInit();
   sim_console_init(TOTAL_CONSOLES, console_interrupt);
   sprintf(name, "DevDiskRegisters %d", n);
   sprintf(name, "EV5 alias %d",n);
   RegistryAddRange((VA)__MAGIC_ZONE(n, 0, MAGIC_ZONE_EV5_ALIAS), ZONE_SZ,
		    REG_FUNC, (void*)EV5_access, name);
   RegistryAddRange((VA)__MAGIC_ZONE_32(n, 0, MAGIC_ZONE_EV5_ALIAS), ZONE_SZ,
		    REG_FUNC, (void*)EV5_access, name);
    
   Simether_Init(restoreFromChkpt);
#if 0
   InitXXM(TOTAL_CPUS);
#endif

   InitTLaser(); 

#if 0 
   RegistryAddRange((VA)(__MAGIC_ZONE(n, 0, MAGIC_ZONE_BDOOR_DEV) +
			 __MAGIC_BDOOR_DISKS_OFFS),
		    sizeof(DevDiskRegisters)*DEV_DISK_MAX_DISKS,
		    REG_FUNC, (void*)BDOOR_disk_access, name);
#endif

   RegistryComplete();
   RegistryDumpEntries();
}





