/*
 * Copyright (C) 1998 by the Board of Trustees
 *    of Leland Stanford Junior University.
 * Copyright (C) 1998 Digital Equipment Corporation
 *
 * This file is part of the SimOS distribution.
 * See LICENSE file for terms of the license.
 *
 */


#ifdef NT_FRAMEBUFFER


#include "simtypes.h"
#include "sim_error.h"
#include "machine_params.h"
#include "machine_defs.h"
#include "registry.h"
#include "console.h"
#include "alpha_access.h"
#include "cpu_interface.h"
#include "gamma.h"
#include "ev5.h"
#include "framebuf.h"
#include "x_framebuf.h"

#include <stddef.h>
#include <stdlib.h>

/*
 * Framebuffer state
 */ 
typedef struct FB_State { 

   unsigned int actual_width;  /* Real screen width */
   unsigned int actual_height; /* Real screen height */
   unsigned int img_width;     /* Image block width */
   unsigned int img_height;    /* Image block height */
   unsigned char *fb_data;     /* Image data block */
   unsigned int palette[256];  /* Color palette */
   
   SimOS_FB_Regs FB_reg_data;

} FB_State;

static FB_State state[MAX_MACHINES];

/*
 * Exported variables
 */
int displayActive[MAX_MACHINES];

/* The modified region */
int min_x[MAX_MACHINES];
int max_x[MAX_MACHINES];
int min_y[MAX_MACHINES];
int max_y[MAX_MACHINES];

extern char *bdoorType[];

int Framebuf_reg_access(int cpuNum, VA vA, int type, Reg *buff)
{
   VA off = vA & 0xFFFF;
   int machine = M_FROM_CPU(cpuNum);

   ASSERT( off < sizeof(SimOS_FB_Regs));
   
#ifdef dont
   CPUPrint("Framebuf_reg OFF=0x%lx PC=0x%lx vA=%lx value=0x%lx %s\n",
            off, PE[cpuNum]->PC,vA,*buff,bdoorType[type]); 
#endif

   /* 
    * Interface is write-only (for now)
    */

   if (BDOOR_IS_STORE(type)) 
   {

      /* First the commands */

      if (off==offsetof(SimOS_FB_Regs,Reset)) 
      {
         if (state[machine].fb_data != NULL)
         {
            if (!DestroyFramebuf(machine))
            {
               CPUWarning("Framebuf: DestroyFramebuf failed inside Reset\n");
            }
            
            state[machine].fb_data = NULL;
            
            displayActive[machine] = 0;

            state[machine].actual_width = state[machine].actual_height = 
               state[machine].img_width = state[machine].img_height = 0;
            
         }
         return 0;
      }

      if (off==offsetof(SimOS_FB_Regs,ChangeMode)) 
      {
         ASSERT(state[machine].fb_data == NULL); 

         /* Don't do anything until we map memory */
         return 0;
      }

      if (off==offsetof(SimOS_FB_Regs,ChangePalette)) 
      {
         int i;
         
         for (i = 0; i < 256; i++)
         {
            state[machine].palette[i] = (unsigned int)
               (state[machine].FB_reg_data.Palette[i]);
         }
         
         if (!SetPalette(machine, state[machine].palette))
         {
            CPUWarning("Framebuf: SetPalette failed\n");
         }
         
         return 0;
      }

      if (off==offsetof(SimOS_FB_Regs,MapMemory)) 
      {
         /* Don't create if we have an active framebuf */
         ASSERT(state[machine].fb_data == NULL);

         state[machine].actual_width = state[machine].FB_reg_data.XResolution;
         state[machine].actual_height = state[machine].FB_reg_data.YResolution;
         state[machine].fb_data = CreateFramebuf(machine,
                                                 state[machine].actual_width, 
                                                 state[machine].actual_height,
                                                 &(state[machine].img_width), 
                                                 &(state[machine].img_height));
         if (!state[machine].fb_data) 
         {
            CPUError("Framebuf: Framebuffer creation failed (%u x %u)\n",
                     state[machine].actual_width, state[machine].actual_height);
         }
         
         bzero((void*)(state[machine].palette), 256*sizeof(unsigned int));
         bzero((void*)(state[machine].FB_reg_data.Palette), 256*sizeof(FBReg));

         if (!SetPalette(machine, state[machine].palette))
         {
            CPUWarning("Framebuf: SetPalette failed\n");
         }

         displayActive[machine] = 1;

         return 0;
      }

      if (off==offsetof(SimOS_FB_Regs,UnMapMemory)) 
      {
         ASSERT(state[machine].fb_data != NULL);

         if (!DestroyFramebuf(machine))
         {
            CPUWarning("Framebuf: DestroyFramebuf failed\n");
         }

         state[machine].fb_data = NULL;

         displayActive[machine] = 0;

         state[machine].actual_width = state[machine].actual_height = 
            state[machine].img_width = state[machine].img_height = 0;

         return 0;
      }

      if (type==BDOOR_STORE_DOUBLE || type==BDOOR_STORE_WORD)
      {
         /* 
          * Set the variable in the data structure
          */
         ((FBReg*)&(state[machine].FB_reg_data))[off/8] = *(unsigned int*)buff;
      }
      else 
      {
         CPUWarning("Framebuf_regs: Illegal size write at offset %u\n", off);
         return 0;
      }
   }
   else
   {
      CPUWarning("Framebuf_regs: Illegal read at offset %u\n", off);
      return 0;
   }

   return 0;
}


int Framebuf_mem_access(int cpuNum, VA vA, int type, Reg *buff)
{
   int machine = M_FROM_CPU(cpuNum);
   VA off;
   int x, y;
   unsigned char *data_ptr, *buff_ptr;

   /* Don't know why we get accesses that aren't aligned, but doing
      these masks makes everything work ok. We just want to ignore those
      low order bits that shouldn't be there. Probably an artifact of
      NT's sparse addressing hack on the Alpha - BenW */
   if (type == BDOOR_STORE_DOUBLE || type == BDOOR_LOAD_DOUBLE)
   {
      off = vA & 0x7FFFFF8;
   }
   else
   {
      ASSERT(type == BDOOR_STORE_WORD || type == BDOOR_LOAD_WORD);
      off = vA & 0x7FFFFFC;
   }

   x = off % state[machine].actual_width;
   y = off / state[machine].actual_width;
   data_ptr = (unsigned char *)(state[machine].fb_data + 
                                (y*state[machine].img_width) + x);
   buff_ptr = (unsigned char *)buff;
   
   if (type==BDOOR_STORE_DOUBLE || type==BDOOR_STORE_WORD)
   {
      /* Extend the area needing X refresh */
      if (x < min_x[machine])
         min_x[machine] = x;
      if ((x+7) > max_x[machine])
         max_x[machine] = (x+7); /* Since writing blocks of up to 8 pixels */
      if (y < min_y[machine])
         min_y[machine] = y;
      if (y > max_y[machine]) 
         max_y[machine] = y;

      if (type==BDOOR_STORE_DOUBLE)
      {
         *(unsigned long*)data_ptr = *(unsigned long*)buff_ptr;
      }
      else
      {
         *(unsigned int*)data_ptr = *(unsigned int*)buff_ptr;
      }
      return 0;
   }
   else if (type==BDOOR_LOAD_DOUBLE) 
   {
      *(unsigned long*)buff_ptr = *(unsigned long*)data_ptr;
      return 0;
   } 
   else if (type==BDOOR_LOAD_WORD)
   { 
      *(unsigned int*)buff_ptr = *(unsigned int*)data_ptr;
      return 0;
   }

   CPUWarning("Framebuf_mem: PC=0x%lx vA=%lx offset=0x%lx value=0x%lx %s\n",
              PE[cpuNum]->PC,vA,off,*buff,bdoorType[type]); 
   return 0;
}


#endif /* NT_FRAMEBUFFER */

