/*
 * Copyright (C) 1998 by the Board of Trustees
 *    of Leland Stanford Junior University.
 * Copyright (C) 1998 Digital Equipment Corporation
 *
 * This file is part of the SimOS distribution.
 * See LICENSE file for terms of the license.
 *
 */


#ifdef NT_FRAMEBUFFER

#include "simtypes.h"
#include "sim_error.h"
#include "machine_params.h"
#include "machine_defs.h"
#include "registry.h"
#include "console.h"
#include "alpha_access.h"
#include "cpu_interface.h"
#include "gamma.h"
#include "ev5.h"
#include "kbd_mouse.h"

static int insertKbdBuffer(int machine, unsigned char scancode);
static int insertMouseBuffer(int machine, unsigned char scancode);
static void setKbdInterrupt(int cpuNum);
static void clearKbdInterrupt(int cpuNum);
static void setMouseInterrupt(int cpuNum);
static void clearMouseInterrupt(int cpuNum);

/* 1 if the framebuffer display is active */
extern unsigned int displayActive[MAX_MACHINES];

/*
 * Keyboard and mouse state
 */
static struct {
   int expectDataPortCommand;
   unsigned char lastCommand;
} shared_state[MAX_MACHINES];

static struct {
   int deviceEnabled;
   int interruptEnabled;
   int translateMode;
   int expectTypematic;
   int expectLED;
   int expectScanCodeSet;

   int head;
   int count;
   unsigned char buffer[KEYBOARD_BUFF_SIZE];
   int pending;
} kbd_state[MAX_MACHINES];

static struct {
   int deviceEnabled;
   int interruptEnabled;
   int streamEnabled;   /* Should we fill the buffer? */
   int expectResolution;
   int expectSampling;

   int head;
   int count;
   unsigned char buffer[MOUSE_BUFF_SIZE];
   int pending;
} mouse_state[MAX_MACHINES];

/*
 * Device Events
 */

int KbdMouse_Initialize(int machine)
{
   shared_state[machine].expectDataPortCommand = 0;
   shared_state[machine].lastCommand = 0;

   kbd_state[machine].deviceEnabled = 1;
   kbd_state[machine].interruptEnabled = 0;
   kbd_state[machine].translateMode = 1;
   kbd_state[machine].expectTypematic = 0;
   kbd_state[machine].expectLED = 0;
   kbd_state[machine].expectScanCodeSet = 0;
   kbd_state[machine].head = 0;
   kbd_state[machine].count = 0;
   kbd_state[machine].pending = 0;

   mouse_state[machine].deviceEnabled = 1;
   mouse_state[machine].interruptEnabled = 0;
   mouse_state[machine].streamEnabled = 0;
   mouse_state[machine].expectResolution = 0;
   mouse_state[machine].expectSampling = 0;
   mouse_state[machine].head = 0;
   mouse_state[machine].count = 0;
   mouse_state[machine].pending = 0;

   return 1;
}

int KbdMouse_KeyPress(int machine, unsigned char scancode)
{
#ifdef KBD_DEBUG
   CPUWarning("KbdMouse_KeyPress: machine %d scancode 0x%x\n", machine, scancode);
#endif
   
   return insertKbdBuffer(machine, scancode);
}

int KbdMouse_MouseEvent(int machine, int x_delta, int y_delta, unsigned int button_state) 
{
   unsigned char val1, val2, val3;

   val1 = (button_state & 0x0f);

   if ( (x_delta < -1) || (x_delta > 1) )
      x_delta /= 2;
   if ( (y_delta < -1) || (y_delta > 1) )
      y_delta /= 2;

   if ( (x_delta>=0) && (x_delta<=255) ) {
      val2 = x_delta;
   }
   else if ( x_delta > 255 ) {
      val2 = 0xff;
   }
   else if ( x_delta >= -256 ) {
      val2 = x_delta;
      val1 |= 0x10;
   }
   else {
      val2 = 0x00;
      val1 |= 0x10;
   }
   
   if ( (y_delta>=0) && (y_delta<=255) ) {
      val3 = y_delta;
   }
   else if ( y_delta > 255 ) {
      val3 = 0xff;
   }
   else if ( y_delta >= -256 ) {
      val3 = y_delta;
      val1 |= 0x20;
   }
   else {
      val3 = 0x00;
      val1 |= 0x20;
   }
   
   if (mouse_state[machine].streamEnabled)
   {
      insertMouseBuffer(machine, val1);
      insertMouseBuffer(machine, val2);
      insertMouseBuffer(machine, val3);
   }

   return 1;
}

int insertKbdBuffer(int machine, unsigned char scancode)
{
   int tail;

   if (kbd_state[machine].count >= KEYBOARD_BUFF_SIZE)
   {
      CPUWarning("KbdMouse: Keyboard buffer full - ignoring scancode 0x%x\n",
                 scancode);
      return 0;
   }
   
   tail = (kbd_state[machine].head + kbd_state[machine].count) % KEYBOARD_BUFF_SIZE;
   kbd_state[machine].buffer[tail] = scancode;
   kbd_state[machine].count++;
   
   setKbdInterrupt(FIRST_CPU(machine));

   return 1;
}

int insertMouseBuffer(int machine, unsigned char value)
{
   int tail;

   if (mouse_state[machine].count >= MOUSE_BUFF_SIZE)
   {
      CPUWarning("KbdMouse: Mouse buffer full - ignoring value 0x%x\n",
                 value);
      return 0;
   }
   
   tail = (mouse_state[machine].head + mouse_state[machine].count) % MOUSE_BUFF_SIZE;
   mouse_state[machine].buffer[tail] = value;
   mouse_state[machine].count++;
   
   setMouseInterrupt(FIRST_CPU(machine));

   return 1;
}

/*
 * Handler routines
 */

int KbdMouse_Status_Read(int cpuNum, VA vA, int type, Reg *buff)
{
   int machine = M_FROM_CPU(cpuNum);
   unsigned char val = 0;

#ifdef KBD_DEBUG
   CPUWarning("KbdMouse_Status_Read called\n");
#endif

   if (mouse_state[machine].count) 
   {
      val = KBM_OUTPUT_BUFFER_FULL | KBM_MOUSE_OUTPUT_BUFFER_FULL;
   }
   else if (kbd_state[machine].count) 
   {
      val = KBM_OUTPUT_BUFFER_FULL;
   }

   *buff = val;

   return 0;
}

int KbdMouse_Command_Write(int cpuNum, VA vA, int type, Reg *buff)
{
   int machine = M_FROM_CPU(cpuNum);
   unsigned char val = 0;

#ifdef KBD_DEBUG
   CPUWarning("KbdMouse_Command_Write called (value 0x%x)\n", *buff);
#endif

   shared_state[machine].lastCommand = *buff;

   switch (*buff)
   {
   case KBM_READ_CONTROLLER_COMMAND_BYTE:
      if (kbd_state[machine].interruptEnabled)
         val |= CCB_ENABLE_KEYBOARD_INTERRUPT;
      if (mouse_state[machine].interruptEnabled)
         val |= CCB_ENABLE_MOUSE_INTERRUPT;
      if (!kbd_state[machine].deviceEnabled)
         val |= CCB_DISABLE_KEYBOARD_DEVICE;
      if (!mouse_state[machine].deviceEnabled)
         val |= CCB_DISABLE_MOUSE_DEVICE;
      if (kbd_state[machine].translateMode)
         val |= CCB_KEYBOARD_TRANSLATE_MODE;
      insertKbdBuffer(machine, val);
      break;

   case KBM_WRITE_CONTROLLER_COMMAND_BYTE:
      shared_state[machine].expectDataPortCommand = 1;
      break;

   case KBM_DISABLE_MOUSE_DEVICE:
      mouse_state[machine].deviceEnabled = 0;
      break;

   case KBM_ENABLE_MOUSE_DEVICE:
      mouse_state[machine].deviceEnabled = 1;
      break;

   case KBM_AUXILIARY_DEVICE_TEST:
      ASSERT(0);
      break;

   case KBM_KEYBOARD_DEVICE_TEST:
      ASSERT(0);
      break;

   case KBM_DISABLE_KEYBOARD_DEVICE:
      kbd_state[machine].deviceEnabled = 0;
      break;

   case KBM_ENABLE_KEYBOARD_DEVICE:
      kbd_state[machine].deviceEnabled = 1;
      break;

   case KBM_WRITE_TO_AUXILIARY_DEVICE:
      shared_state[machine].expectDataPortCommand = 1;
      break;

   default:
      CPUWarning("KbdMouse: Bad command received (0x%x)\n", *buff);
      break;
   }
   
   return 0;
}

int KbdMouse_Data_Read(int cpuNum, VA vA, int type, Reg *buff)
{
   int machine = M_FROM_CPU(cpuNum);
   if (mouse_state[machine].count)
   {
      *buff = mouse_state[machine].buffer[mouse_state[machine].head];
      mouse_state[machine].count--;
      mouse_state[machine].head = (mouse_state[machine].head + 1) % MOUSE_BUFF_SIZE;
      if (KbdMouse_PendingMouse(cpuNum))
      {
         clearMouseInterrupt(FIRST_CPU(machine));
      }
   } 
   else if (kbd_state[machine].count)
   {
      *buff = kbd_state[machine].buffer[kbd_state[machine].head];
      kbd_state[machine].count--;
      kbd_state[machine].head = (kbd_state[machine].head + 1) % KEYBOARD_BUFF_SIZE;
      if (KbdMouse_PendingKbd(cpuNum))
      {
         clearKbdInterrupt(FIRST_CPU(machine));
      }
   }
   else
   {
      CPUPrint("KbdMouse: Data read with no data ready\n");
      *buff = 0;
   }
#ifdef KBD_DEBUG
   CPUWarning("KbdMouse_Data_Read called (returning 0x%x)\n", *buff);
#endif

   return 0;
}

int KbdMouse_Data_Write(int cpuNum, VA vA, int type, Reg *buff)
{
   int machine = M_FROM_CPU(cpuNum);
   unsigned char val = *buff;

#ifdef KBD_DEBUG
   CPUWarning("KbdMouse_Data_Write called (value 0x%x)\n", *buff);
#endif

   if (shared_state[machine].expectDataPortCommand) 
   {
      shared_state[machine].expectDataPortCommand = 0;
      switch (shared_state[machine].lastCommand)
      {
      case KBM_WRITE_CONTROLLER_COMMAND_BYTE:
         kbd_state[machine].interruptEnabled =
            ((val & CCB_ENABLE_KEYBOARD_INTERRUPT) != 0);
         mouse_state[machine].interruptEnabled =
            ((val & CCB_ENABLE_MOUSE_INTERRUPT) != 0);
         kbd_state[machine].deviceEnabled =
            ((val & CCB_DISABLE_KEYBOARD_DEVICE) == 0);
         mouse_state[machine].deviceEnabled =
            ((val & CCB_DISABLE_MOUSE_DEVICE) == 0);
         kbd_state[machine].translateMode =
            ((val & CCB_KEYBOARD_TRANSLATE_MODE) != 0);

#ifdef KBD_DEBUG
         CPUWarning("Keyboard: Enabled=%d Interrupts=%d TranslateMode=%d\n",
                    kbd_state[machine].deviceEnabled, kbd_state[machine].interruptEnabled,
                    kbd_state[machine].translateMode);
         CPUWarning("Mouse: Enabled=%d Interrupts=%d\n",
                    mouse_state[machine].deviceEnabled, mouse_state[machine].interruptEnabled);
#endif
         break;
      case KBM_WRITE_TO_AUXILIARY_DEVICE:
         if (mouse_state[machine].expectResolution)
         {
            mouse_state[machine].expectResolution = 0;
            CPUPrint("KbdMouse: Mouse resolution set to 0x%x\n", *buff);
            insertMouseBuffer(machine, KBM_ACKNOWLEDGE);
            return 0;
         }
         if (mouse_state[machine].expectSampling)
         {
            mouse_state[machine].expectSampling = 0;
            CPUPrint("KbdMouse: Mouse sampling set to 0x%x\n", *buff);
            insertMouseBuffer(machine, KBM_ACKNOWLEDGE);
            return 0;
         }
         switch (*buff)
         {
         case ENABLE_MOUSE_TRANSMISSION:
            mouse_state[machine].streamEnabled = 1;
            CPUPrint("KbdMouse: Mouse streaming enabled (machine %d)\n",
                     machine);
            insertMouseBuffer(machine, KBM_ACKNOWLEDGE);
            break;
         case READ_MOUSE_STATUS:
            insertMouseBuffer(machine, KBM_ACKNOWLEDGE);
            insertKbdBuffer(machine, 0x00); /* ignored */
            insertKbdBuffer(machine, 0x03); /* number of buttons */
            insertKbdBuffer(machine, 0x00); /* ignored */
            break;
         case SET_MOUSE_RESOLUTION:
            insertMouseBuffer(machine, KBM_ACKNOWLEDGE);
            mouse_state[machine].expectResolution = 1;
            break;
         case SET_MOUSE_SAMPLING_RATE:
            insertMouseBuffer(machine, KBM_ACKNOWLEDGE);
            mouse_state[machine].expectSampling = 1;
            break;
         case SET_MOUSE_SCALING_1TO1:
            insertMouseBuffer(machine, KBM_ACKNOWLEDGE);
            break;
         case GET_DEVICE_ID:
            insertMouseBuffer(machine, KBM_ACKNOWLEDGE);
            insertKbdBuffer(machine, MOUSE_ID_BYTE); 
            break;
         case MOUSE_RESET:
            mouse_state[machine].streamEnabled = 0;
            CPUPrint("KbdMouse: Mouse streaming disabled (machine %d)\n",
                     machine);
            insertMouseBuffer(machine, KBM_ACKNOWLEDGE);
            insertKbdBuffer(machine, MOUSE_COMPLETE);
            insertKbdBuffer(machine, MOUSE_ID_BYTE); 
            break;
         default:
            CPUWarning("KbdMouse: Bad write to mouse (0x%x)\n", *buff);
            insertMouseBuffer(machine, KBM_ACKNOWLEDGE);
            break;
         }
         break;
      default:
         CPUWarning("KbdMouse: Bad data port command write (0x%x after 0x%x)\n", 
                    *buff, shared_state[machine].lastCommand);
         break;
      }
   }
   else 
   {
      if (kbd_state[machine].expectTypematic)
      {
         kbd_state[machine].expectTypematic = 0;
         CPUPrint("KbdMouse: Typematic set to 0x%x\n", *buff);
         insertKbdBuffer(machine, KBM_ACKNOWLEDGE); 
         return 0;
      }
      if (kbd_state[machine].expectLED)
      {
         kbd_state[machine].expectLED = 0;
         CPUPrint("KbdMouse: Keyboard indicators set to 0x%x\n", *buff);
         insertKbdBuffer(machine, KBM_ACKNOWLEDGE); 
         return 0;
      }
      if (kbd_state[machine].expectScanCodeSet)
      {
         kbd_state[machine].expectScanCodeSet = 0;
         CPUPrint("KbdMouse: ScanCodeSet set to 0x%x\n", *buff);
         insertKbdBuffer(machine, KBM_ACKNOWLEDGE); 
         return 0;
      }
      switch (*buff)
      {
      case READ_KEYBOARD_ID:
      case SET_ALL_TYPEMATIC_MAKE_BREAK:
         CPUWarning("KbdMouse: Unimplemented write to kbd (0x%x)\n", *buff);
         insertKbdBuffer(machine, KBM_ACKNOWLEDGE);
         break;
      case SELECT_SCAN_CODE_SET:
         insertKbdBuffer(machine, KBM_ACKNOWLEDGE);
         kbd_state[machine].expectScanCodeSet = 1;
         break;
      case SET_KEYBOARD_INDICATORS:
         insertKbdBuffer(machine, KBM_ACKNOWLEDGE);
         kbd_state[machine].expectLED = 1;
         break;
      case SET_KEYBOARD_TYPEMATIC:
         insertKbdBuffer(machine, KBM_ACKNOWLEDGE);
         kbd_state[machine].expectTypematic = 1;
         break;
      case KEYBOARD_RESET:
         insertKbdBuffer(machine, KBM_ACKNOWLEDGE);
         insertKbdBuffer(machine, KEYBOARD_COMPLETE_SUCCESS);
         break;
      default:
         CPUWarning("KbdMouse: Bad write to kbd (0x%x)\n", *buff);
         insertKbdBuffer(machine, KBM_ACKNOWLEDGE);
         break;
      }
   }

   return 0;
}

int KbdMouse_PendingKbd(int cpuNum) 
{
   int machine = M_FROM_CPU(cpuNum);
   return (kbd_state[machine].pending);
}

int KbdMouse_PendingMouse(int cpuNum) 
{
   int machine = M_FROM_CPU(cpuNum);
   return (mouse_state[machine].pending);
}

/*
 * Interrupt toggle utilities
 */

void setKbdInterrupt(int cpuNum)
{
   int machine = M_FROM_CPU(cpuNum);
   int lastPending = kbd_state[machine].pending;

   if (kbd_state[machine].interruptEnabled) 
   {
      if (!displayActive[machine])
      {
         return;
      }
      if (kbd_state[machine].count && !kbd_state[machine].pending)
      {
         kbd_state[machine].pending = 1;
         EV5_SetExtInterrupt(PE[cpuNum],INTLEVEL_IRQ1);
      }
   }
#ifdef KBD_DEBUG
   CPUWarning("KbdMouse: setKbdInterrupt (%d -> %d)\n",
              lastPending, kbd_state[machine].pending);
#endif
}

void clearKbdInterrupt(int cpuNum)
{
   int machine = M_FROM_CPU(cpuNum);
   int lastPending = kbd_state[machine].pending;

   ASSERT(kbd_state[machine].pending);

   if (kbd_state[machine].count == 0)
   {
      kbd_state[machine].pending = 0;
      EV5_ClearExtInterrupt(PE[cpuNum],INTLEVEL_IRQ1);
   }
#ifdef KBD_DEBUG
   CPUWarning("KbdMouse: clearKbdInterrupt (%d -> %d)\n",
              lastPending, kbd_state[machine].pending);
#endif
}

void setMouseInterrupt(int cpuNum)
{
   int machine = M_FROM_CPU(cpuNum);
   int lastPending = mouse_state[machine].pending;

   if (mouse_state[machine].interruptEnabled) 
   {
      if (!displayActive[machine])
      {
         return;
      }
      if (mouse_state[machine].count && !mouse_state[machine].pending)
      {
         mouse_state[machine].pending = 1;
         EV5_SetExtInterrupt(PE[cpuNum],INTLEVEL_IRQ1);
      }
   }
#ifdef KBD_DEBUG
   CPUWarning("KbdMouse: setMouseInterrupt (%d -> %d)\n",
              lastPending, mouse_state[machine].pending);
#endif
}

void clearMouseInterrupt(int cpuNum)
{
   int machine = M_FROM_CPU(cpuNum);
   int lastPending = mouse_state[machine].pending;

   ASSERT(mouse_state[machine].pending);

   if (mouse_state[machine].count == 0)
   {
      mouse_state[machine].pending = 0;
      EV5_ClearExtInterrupt(PE[cpuNum],INTLEVEL_IRQ1);
   }
#ifdef KBD_DEBUG
   CPUWarning("KbdMouse: clearMouseInterrupt (%d -> %d)\n",
              lastPending, mouse_state[machine].pending);
#endif
}

#endif /* NT_FRAMEBUFFER */

