/*
 * Copyright (C) 1998 by the Board of Trustees
 *    of Leland Stanford Junior University.
 * Copyright (C) 1998 Digital Equipment Corporation
 *
 * This file is part of the SimOS distribution.
 * See LICENSE file for terms of the license.
 *
 */


/*#define HACK_PRINTINST*/ 


#include <stdio.h>
#include <unistd.h>
#include <string.h>
#include <fcntl.h>
#include <signal.h>
#include <sys/types.h>
#include <sys/mman.h>

#include "simtypes.h"
#include "sim_error.h"
#include "syslimits.h"
#include "machine_params.h"
#include "gdb_interface.h"
#include "cpu_stats.h"
#include "cpu_interface.h"
#include "params.h"

#include "hd.h"
#include "console.h"
#include "dma.h"

#include "../../memsystems/memsys.h"

#include "alpha.h"
#include "ev5_ipr.h"
#include "ipr_decode.h"
#include "alpha_params.h"
#include "tb.h"

#include "alpha_access.h"
#include "simos_scsi.h"
#include "loader.h"
#include "ev5.h"
#include "devices.h"
#include "checkpoint.h"       /* csc: */

/* ***********************************************
 * Global variables
 * ***********************************************/

int restoringCpt=0;
int simosCPUType = NO_CPU;
AlphaState *PE[SIM_MAXCPUS];
int alphaLateInitDone;

/* ****************************
 * dummy defines that enable link
 * ***************************************/

MemsysVector memsysVec;
char *EthersimHostname;
char *EtherAddress;
int inCellMode = 0;
int RestoreEthernet = 0;
int EtherSendPort = 0;
char *FPromFile = "foo";
long verboseStartTime;

/* csc: Declarations for checkpoint functions */
static CptCallback MemoryCheckpointCB;
static CptCallback TclCheckpointCB;
static CptCallback ExecStateCheckpointCB;
static CptCallback ParamCheckpointCB;

static int mfd = -1;
extern CptVersion cptVersion;
char *checkpointName;

/* csc:Structure required for command line option processing */
struct {
  int simCheckpointRestore;    /* -r selected */
  char* checkpoint_master;     /* argument to -r */
  CPUType startcpu;            /* -m, -c, -p selected */
  char *fault_file;            /* -f selected */
  char* cpualloc_hostport;     /* argument to -a, nil if -a not selected */
  char* skiplist;              /* argument to -s, nil if -s not selected */
  char* param_file;            /* File with parameter values */
  char* log_file;              /* -l option: log file name */
  char* debug_flags;
} optvals;




static void InitPE(void)
{
  int cpuNum;
  int numCPUs = TOTAL_CPUS;
   
  bzero((char*)PE,sizeof(PE));
  ASSERT( numCPUs>=0 && numCPUs<=SIM_MAXCPUS);
  for (cpuNum=0;cpuNum<numCPUs;cpuNum++) { 
    PE[cpuNum] = (AlphaState*)malloc(sizeof(AlphaState));
    EV5_InitPE(PE[cpuNum], cpuNum);
    
    STATS_SET(cpuNum, numInstructions, 0);
#if 0
    STATS_SET(cpuNum, nextInstrSample, MS_SAMPLE_INSTR_INTERVAL);
#endif
    STATS_SET(cpuNum, iReads, 0);
    STATS_SET(cpuNum, dReads, 0);
    STATS_SET(cpuNum, dWrites, 0);
    STATS_SET(cpuNum, stallTime, 0);
  }
}


/* ****************************************   */

char hackPALcode[128] = "palcode";


/* csc: Function to parse command line: copied from MIPS version */
static void
ParseOpts(int argc, char** argv)
{
  int c;
  extern char* optarg;
  extern int optind;
  
  bzero((char*)&optvals, sizeof(optvals));
  optvals.startcpu   = NO_CPU;
  optvals.param_file = "init.simos";
  optvals.log_file   = "log";
  optvals.debug_flags = "";
  
  while ((c = getopt(argc, argv, "z:r:bf:a:s:o:l:d:p:")) != EOF) {
    switch (c) {
    case 'r':
      optvals.simCheckpointRestore = 1;
      optvals.checkpoint_master = optarg;
      checkpointName = optarg;
      break;
    case 'p':
      if (strcmp(optarg, "kappa") == 0) {
        optvals.startcpu = KAPPA;
      } else if (strcmp(optarg, "delta") == 0) {
        optvals.startcpu = DELTA;
      } else if (strcmp(optarg, "gamma") == 0) {
        optvals.startcpu = GAMMA;
      } else if (strcmp(optarg, "sam") == 0) {
        optvals.startcpu = SAM;
      } else {
	fprintf(stderr, "\nMissing or unsupported CPU type specified with the -p option: %s\nCurrently supported types are kappa, gamma, delta, or sam\n", optarg);
        exit(1);
      }
      break;
    case 'z': 
      optvals.param_file = optarg;
      break;
    case 'd':
      if (!argv[optind - 1]) {
	fprintf(stderr, "SimOS: Turning all debugging flags on\n");
	optvals.debug_flags = "+";
      } else {
	optvals.debug_flags = optarg;
      }
      break;
    case 'a':
      optvals.cpualloc_hostport = optarg;
      break;
    case 's':
      optvals.skiplist = optarg;
      break;
    case 'f':
      optvals.fault_file = optarg;
      break;
    case 'o':
      if (!argv[optind - 1]) {
	fprintf(stderr, "\nmissing value for option: %s\n", optarg);
	exit(1);
      } else if ((optind < argc) && (argv[optind][0] != '-')) {
	TCLDefineOption(optarg, argv[optind]);
	optind++;
      } else {
	TCLDefineOption(optarg, "1");
      }
      break;
    case 'l':
      optvals.log_file = optarg;
      break;
    case '?':
    default:
      fprintf(stderr,
	      "\nusage: simos [-r checkpoint_masterlog]\n"
	      "    [-d <debugflags>] [-p <cpu_type>]\n"
	      "    [-z param_file_name] [-f fault_file_name]\n"
	      "    [-o tclopt [val]]  [-l log_file_name]\n");
      exit(1);
    }
  }
}


/*
 * Called on a SIGUSR1 signal. Used to enter tcl, etc. 
 */
static void 
Sigusr(int sig)
{
   if (signal(SIGUSR1, Sigusr) == SIG_ERR) perror("signal(SIGUSR1)");

   if (CPUVec.Handle_Debug_Signal) {
      CPUVec.Handle_Debug_Signal(0, (sig == SIGUSR1));
   }
   return;
}

int main(int argc, char **argv)
{
   char *init_simos_filename = "init.simos";
   int mfd = -1;
   int machNo;

   ParseOpts(argc, argv);
   restoringCpt = optvals.simCheckpointRestore;

   SimErrorInit(optvals.log_file);

   Simcpt_Init();
   Simcpt_Register("param", ParamCheckpointCB, ALL_CPUS);
   
   if (restoringCpt) { 
     Simcpt_Checkpoint(CPT_RESTORE, optvals.checkpoint_master);
     Simcpt_Restore("param");
     /* let param manager know which of the parameters have been initialized
      * so it doesn't error the user for forgetting to init them
      */
     ParamGrabbedAtStartup("PARAM(MACHINE.Count)");
     ParamGrabbedAtStartup("PARAM(CPU.Count)");
     ParamGrabbedAtStartup("PARAM(HIVE.NumCells)");
     ParamGrabbedAtStartup("PARAM(MEMSYS.MemSize)");
     ParamGrabbedAtStartup("PARAM(CONSOLE.Count)");
     ParamGrabbedAtStartup("PARAM(ETHERNET.Count)");
     ParamGrabbedAtStartup("PARAM(CLOCK.Count)");
     ParamGrabbedAtStartup("PARAM(DISK.NumControllers)");
   }
   
   CPUVec.OldCycleCount = NULL;


   DebugInit(optvals.debug_flags);
   /* Remap Vector. */
   remapVec = ZMALLOC(sizeof(RemapVector),"RemapVector");


   InitCPUModel(optvals.startcpu);
   EpsilonExternalSchedulerLink();
   TCLInit(init_simos_filename);
   Simcpt_Register("memory", MemoryCheckpointCB, ALL_CPUS);
   Simcpt_Register("Execstate",ExecStateCheckpointCB,ALL_CPUS);
   Simcpt_UseCompression("memory", GZIP, BIN_FILE_ONLY);
   Simcpt_Register("tcl",TclCheckpointCB,ALL_CPUS);
 
   InitCPUStats();
   InitIPRNames();

   EventCallbackInit(TOTAL_CPUS);
   Simdebug_init();
   if (signal(SIGUSR1, Sigusr) == SIG_ERR) perror("signal(SIGUSR1");


   /*
    * Memory File
    */
  if (restoringCpt ) {
     if (Simcpt_CanMmapMemoryCheckpoint() ) {
        mfd = Simcpt_GetMemoryCptFile();
        if( mfd < 0 ) {
           perror("Open MemFile failed");
           exit(1);
        } 
     } else {
	mfd = -1;
     }
  } else {
     mfd = -1;
  } 

  
  /*
   * Perform the mmap for each machine (out of the same memory file)
   *
   */
  {
     long totalSize=0;
     char *addr;

     ASSERT(NUM_MACHINES > 0);

     /* Initialize the base memory address for each machine
      */
     SIM_MEM_ADDR(0) = BASESIMMEMADDR;
     ASSERT(MEM_SIZE(0));
     for (machNo = 1; machNo < NUM_MACHINES; machNo++) {
	ASSERT(MEM_SIZE(machNo));
        SIM_MEM_ADDR(machNo) =
           SIM_MEM_ADDR(machNo-1) + MEM_SIZE(machNo-1);
     }

     /*
      * Perform the mmap for each machine (out of the same memory file)
      */
     for (machNo = 0; machNo < NUM_MACHINES; machNo++) {
        SIM_DEBUG(('g', "size = %d\n", MEM_SIZE(machNo)));

        addr = (char *) mmap(SIM_MEM_ADDR(machNo),
                             MEM_SIZE(machNo),
                             PROT_READ|PROT_WRITE, 
                             MAP_FIXED|MAP_PRIVATE|
                             (mfd>0?MAP_FILE:MAP_ANONYMOUS),
                             mfd, SIM_MEM_ADDR(machNo)-SIM_MEM_ADDR(0));
        CPUWarning("physicalMemory set at 0x%p memSize=%llx  \n",
                   addr, MEM_SIZE(machNo));
        if (addr == (void*)-1) { 
           perror("mmap of physical memrory: ");
           CPUError("mmap of physical memory failed. size=%llx \n",totalSize);
        }
        ASSERT(addr == SIM_MEM_ADDR(machNo));
     }
  }
 
  
  if (restoringCpt && !Simcpt_CanMmapMemoryCheckpoint() ) {
     Simcpt_Restore("memory");
  }
    
  InitZeroMem(PHYS_TO_MEMADDR(NUM_MACHINES-1, MEM_SIZE(NUM_MACHINES-1)), 
              SIM_ALPHA_ZERO_MEMEND,0);
  /*
   * Devices initializations
   */

  sim_disk_init(DISK_NODES,
                NUM_UNITS_PER_CONTROLLER(0),
                SimOS_SCSI_intr,
                restoringCpt);

  DMAInit();
  if (!strcmp(DISK_MODEL(0), "HP")) {
     /* Initialize the DISK model we are compiled for */
     DiskModelInit(SIM_MAX_DISKS, ((double)HP_DISK_SCALING(0))/100.0);
  } else {
     if (strcmp(DISK_MODEL(0), "Fixed")) {
        CPUWarning("Bad disk model: %s\n", DISK_MODEL(0));
        ASSERT(0);
     }
  }
  
  

#ifdef HACK_PRINTINST
  {
     int fd,i,len,flag;
     AlphaState *first;
     InitPE();
     first = PE[0];
     fd = open(hackPALcode,0);
     CPUWarning("HACK TO EXTRACT THE NT PALCODE \n");
     len = read(fd,alphaMemoryPtr+0x2fe00,0x9000);
     first->PC = 0x30001;
     for (i=0;i<(len-0x200);i+=4,first->PC+=4) {
        uint val = *(uint*)(alphaMemoryPtr+0x30000+i);
        if (!val) { 
           if (!flag) {
              flag = 1;
              CPUPrint("    [ .......  ]\n");
           } 
           continue;
        }
        flag = 0;            
        PrintInstruction(first,*(int*)(alphaMemoryPtr+0x30000+i));
     }
     exit(1);  
  }
#endif
  
  for (machNo = 0; machNo < NUM_MACHINES; machNo++)
  {
     AlphaAccess *alphaAccess;

     if (!restoringCpt) {
        if (!load_ultrix_aout(alphaConsoleParams.PALcodeFileName,machNo,PAL_BASE)) {
           CPUError("Could not load base PAL code image.\n");
        }
        if (!load_ultrix_coff(alphaConsoleParams.consoleFileName,machNo)) {
           CPUError("Could not load SRM firmware image.\n");
        }
        if (alphaConsoleParams.bootNT)  {
           CPUWarning("SimOS is set up for an NT boot\n");
           if (!Load_NTOSLoader(alphaConsoleParams.NT_OSLoaderFileName,machNo)) {
              CPUError("Could not load NT osloader image.\n");
           }
           if (!Load_NTOSLoader(alphaConsoleParams.NT_FirmwareFileName,machNo)) {
              CPUError("Could not load ARC firmware image.\n");
           }
        } else {
           CPUWarning("SimOS is set up for an OSF/1 or Linux boot\n");
           if (!load_ultrix_coff(alphaConsoleParams.kernelFileName,machNo)) { 
              CPUError("Could not load kernel image.\n");
           }
        }
     }

     alphaAccess = alphaAccessArray + machNo;
     alphaAccess->last_offset = sizeof(struct AlphaAccess);
     alphaAccess->version = ALPHA_ACCESS_VERSION;

     alphaAccess->numCPUs = NUM_CPUS(machNo);
     alphaAccess->mem_size = MEM_SIZE(machNo);
     alphaAccess->cpuClock = CPU_CLOCK;
     alphaAccess->intrClockFrequency = INTRCLOCK_FREQUENCY;
  }
  
  /*
   * initialize first processor of each machine
   */

  /* xxx this is second call to DebugInit here.  ??? -Jamey */
  DebugInit(optvals.debug_flags);
  InitPE();

  InitDevices(restoringCpt);
  
  if (restoringCpt) {
     Simcpt_Restore("tcl");
     /* Simcpt_Restore("diskdev"); */ /* restored by sim_disk_init() -Jamey 10/7/97 */
     Simcpt_Restore("Execstate");
  }

  if (optvals.startcpu != NO_CPU) {
    CPUWarning("Overriding SimosCPUType from command line\n");
    simosCPUType = optvals.startcpu;
  }

  while (simosCPUType != NO_CPU) {
     switch (simosCPUType) {
     case GAMMA: 
        CPUWarning("Starting GAMMA simulator \n");
        ExecuteGamma();
        break;
     case DELTA: 
        CPUWarning("Starting DELTA simulator \n");
        ExecuteDelta();
        break;
     case EPSILON: 
        CPUWarning("Starting EPSILON simulator \n");
        ExecuteEpsilon();
        break;
     case KAPPA: 
        CPUWarning("Starting KAPPA simulator \n");
        ExecuteKappa();
        break;
     case SAM: 
        CPUWarning("Starting SAM simulator \n");
        ExecuteSam();
        break;
     default: 
        CPUError("CPU Simulator not defined %d \n",simosCPUType);
     }
  }
  AnnExec(AnnFind("simos", "exit"));
  return 0;
   
}



/* *******************************************************
 * Checkpointing code (should move to a separate file)
 * *******************************************************/



static int 
ParamCheckpointCB(CptDescriptor *cptd)
{

   int machNo;

   /*
    * Establish or save the checkpoint version
    */
   
   if (cptd->mode == CPT_SAVE) {
      /* Set the current version to the desired save version */
      ASSERT(cptVersion.saveVer);
      cptVersion.ver = cptVersion.saveVer;
      cptVersion.sub = cptVersion.saveSub;
   } 
   Simcpt_CptInt(cptd, "Version", NO_INDEX, NO_INDEX,
                 &(cptVersion.ver));
   Simcpt_CptParamInt(cptd, "SubVersion", NO_INDEX, NO_INDEX,
		      &(cptVersion.sub));
        
   if (cptd->mode == CPT_SAVE) {
      CPUPrint("CPT: Saving (Format is version %d, sub-version %d)\n",
               cptVersion.ver, cptVersion.sub);
   } else {
      CPUPrint("CPT: Restoring (Format is version %d, sub-version %d)\n",
               cptVersion.ver, cptVersion.sub);
   }
      
   /*
    * Do the checkpointing
    */
   

   Simcpt_CptParamInt(cptd, "NumMachines", NO_INDEX, NO_INDEX,
		      &(NUM_MACHINES));
   
   for (machNo = 0; machNo < NUM_MACHINES; machNo++) {
      Simcpt_CptParamInt(cptd, "NumCPUs", machNo, NO_INDEX,
			 &(NUM_CPUS(machNo)));
      Simcpt_CptParamInt(cptd, "NumCells", machNo, NO_INDEX,
			 &(NUM_CELLS(machNo)));
      Simcpt_CptParamInt(cptd, "MemSize", machNo, NO_INDEX,
			 &(MEM_SIZE(machNo)));
      Simcpt_CptParamInt(cptd, "NumConsoles", machNo, NO_INDEX,
			 &(NUM_CONSOLES(machNo)));
      Simcpt_CptParamInt(cptd, "NumEtherControllers", machNo, NO_INDEX,
			 &(NUM_ETHER_CONTROLLERS(machNo)));
      {
         /* NOTE on compatibility:
          * Former checkpointing code save a total number of disk
          * controllers per machine. To preserve compatibility with
          * older checkpoints, we now save the following:
          *    NumDiskControllers[m] : 0
          *    NumDiskControllers[m,n] : <ctrls>
          *    NumUnitsPerController[m] : <upc>
          * (note the first line should be eventually eliminated).
          */
         int nc = 0;
#ifdef rpb_notdef
         Simcpt_CptParamInt(cptd, "NumDiskControllers", machNo, NO_INDEX,
                            &nc);
#endif
         if (nc != 0) {
            /* compatibility: restoring from old checkpoint */
            /* note old checkpoints can have more disk-carrying nodes
             * than processor nodes. Ugh.
             */
            int node;
            Sim_Warning("*** Restoring disk ctrls info from old ckpt ***\n");
            ASSERT(cptd->mode == CPT_RESTORE);
            for (node = 0; node < nc; node++)
	       NUM_DISK_CONTROLLERS(machNo, node) = 1;
            NUM_UNITS_PER_CONTROLLER(machNo) = 128;
         } else {
            /* new checkpoint code: save controllers/node, units/controller */
            int node;
            for (node = 0; node < NUM_CPUS(machNo); node++) {
	       Simcpt_CptParamInt(cptd, "NumDiskControllers",
				  machNo, node,
				  &NUM_DISK_CONTROLLERS(machNo, node));
            }
            Simcpt_CptParamInt(cptd, "NumUnitsPerController",
                               machNo, NO_INDEX,
                               &NUM_UNITS_PER_CONTROLLER(machNo));
            
         }
      }  
   }
   
   Simcpt_CptString(cptd, "EthersimHostname", NO_INDEX, NO_INDEX, 
		    &EthersimHostname);
   Simcpt_CptString(cptd, "EtherAddress", NO_INDEX, NO_INDEX,
		    &EtherAddress);
   
   
   /* Do console checkpointing */
   sim_console_ckpt(cptd);
   
   return 0;
}



static int
MemoryCheckpointCB(CptDescriptor *cptd)
{
  int machNo;

   /*
    * Establish or save the checkpoint version
    */

  cptVersion.ver = cptVersion.saveVer;
  cptVersion.sub = cptVersion.saveSub;

  for (machNo = 0; machNo < NUM_MACHINES; machNo++) {
     Simcpt_CptBlock(cptd, "Memory", machNo, NO_INDEX, 
		     (caddr_t) PHYS_TO_MEMADDR(machNo,0),
		     MEM_SIZE(machNo), mfd );
  }
  return 0;
}
 
static int
TclCheckpointCB(CptDescriptor *cptd) {
   if (cptd->mode == CPT_RESTORE) TclRestoreCheckpoint(cptName);
   else                           TclDoCheckpoint(cptName);
   return 0;
}


static int
ExecStateCheckpointCB(CptDescriptor *cptd)
{
   int cpu,regst,ipreg,irqlevel;

   for (cpu=0;cpu < TOTAL_CPUS; cpu++)  {
      Simcpt_CptInt(cptd,"CPUSTATE",cpu,NO_INDEX,&(PE[cpu]->cpuState));
      for (regst=0;regst < 33;regst++) {
         Simcpt_CptUlong(cptd,"REG",cpu,regst,(Reg*)&(PE[cpu]->reg[regst]));
      }

      for (regst=0;regst < 33;regst++) {
         Simcpt_CptUlong(cptd,"FP",cpu,regst,(Reg*)&(PE[cpu]->fp[regst]));
      }

      for (regst=0;regst < 16;regst++) {
         Simcpt_CptUlong(cptd,"SREG",cpu,regst,(Reg*)&(PE[cpu]->shadowRegs[regst]));
      }
    
      Simcpt_CptUlong(cptd,"FPCR",cpu,NO_INDEX,&(PE[cpu]->fpcr));      
      Simcpt_CptUlong(cptd,"PC",cpu,NO_INDEX,&(PE[cpu]->PC));
      Simcpt_CptUint(cptd,"LCKREG",cpu,NO_INDEX,&(PE[cpu]->lockedRegisters));
      Simcpt_CptInt(cptd,"DTBNLU",cpu,NO_INDEX,&(PE[cpu]->dtbNLU));
      Simcpt_CptInt(cptd,"ITBNLU",cpu,NO_INDEX,&(PE[cpu]->itbNLU));
      
      TBCheckpoint(cptd,cpu,PE[cpu]->tb);

      for(ipreg=0;ipreg < MAX_IPR;ipreg ++) {
         Simcpt_CptUlong(cptd,"IPR",cpu,ipreg,&(PE[cpu]->ipr[ipreg]));
      }
      
      Simcpt_CptInt(cptd,"LOCKFLAG",cpu,NO_INDEX,&(PE[cpu]->lock_flag));
      Simcpt_CptUlong(cptd,"LOCKEDADDR",cpu,NO_INDEX,&(PE[cpu]->locked_addr));
      
      
      Simcpt_CptUlong(cptd,"PENDINGIRQ",cpu,NO_INDEX,&(PE[cpu]->pending_irq));
      for (irqlevel = 0; irqlevel < INTLEVEL_EXTERNAL_MAX; irqlevel++) {
         Simcpt_CptInt(cptd,"PENDINGIRQCOUNT",cpu,irqlevel,&(PE[cpu]->pending_irq_count[irqlevel]));
      }
      Simcpt_CptInt(cptd,"TAKEINTERRUPT",cpu,NO_INDEX,&(PE[cpu]->takeInterrupt));      
      Simcpt_CptInt(cptd,"TAKEINTERRUPTATREI",cpu,NO_INDEX,&(PE[cpu]->takeInterruptAtRei));      
      Simcpt_CptInt(cptd,"USESHADOWREGS",cpu,NO_INDEX,&(PE[cpu]->useShadowRegisters));
   }
  
   sim_console_ckpt(cptd);
   return 0;
}

	    
	    
