/*
 * Copyright (C) 1998 by the Board of Trustees
 *    of Leland Stanford Junior University.
 * Copyright (C) 1998 Digital Equipment Corporation
 *
 * This file is part of the SimOS distribution.
 * See LICENSE file for terms of the license.
 *
 */


/* 
 * This file contains the code to interface between
 * the turbolaser simulated system and the simos 
 * ethernet device 
 */

#include "simtypes.h"
#include "sim_error.h"
#include "machine_params.h"
#include "machine_defs.h"
#include "registry.h"
#include "console.h"
#include "hd.h"
#include "alpha_access.h"
#include "cpu_interface.h"
#include "gamma.h"
#include "ev5.h"
#include "ethernet.h"
#include "simether.h"

#define DEBUG_ETHER


static int simetherPending[SIM_MAXCPUS];


/* ETHERNET device.
 *
 */

void
Simether_Init(int restoreFromChkpt)
{
   int i;
   for (i=0; i<SIM_MAXCPUS; i++) {
      simetherPending[i] = 0;
   }
   SimetherInit(restoreFromChkpt, Simether_interrupt);

}



int
Simether_access(int cpuNum, VA vA, int type, DevRegister64 *datap) 
{
   int ctrl = 0;
   VA offs = vA & 0xFFFF;

#ifdef DEBUG_ETHER
  LogEntry("Simether_access", cpuNum, "offs=0x%x t=%d data=0x%08x\n",
	   offs, type, *datap);
#endif

  /* we only handle dword accesses and buserror all others
   *ASSERT (BDOOR_SIZE(type) == sizeof(DevRegister64));
   * Seemingly bogus code in gamma.c that always sets the type
   * to BDOOR_WORD. So change the assert for now
   */
  ASSERT (BDOOR_SIZE(type) == BDOOR_WORD);

  /* NOTE: it so "happens" that the SimetherRegisters structure is
   * isomorphic to DevEtherRegisters. This helps us simplify the
   * mapping from the simulated registers to the ether simulator.
   */

  if (BDOOR_IS_LOAD(type)) {
    *datap = (DevRegister64)SimetherIO(ctrl, offs, 0, (DevRegister64)0);
  } else {
    (void)SimetherIO(ctrl, offs, 1, (DevRegister64)*datap);
  }
  return 0;
}



/*
 * check the interrupt level 
 */

void
Simether_interrupt(int n, int on) {
#ifdef DEBUG_ETHER
   CPUPrint("Simether_intr on=%d \n", on);
#endif

   /* for now we are always interrupting cpu 0 */
   if (on) {
      simetherPending[0] = 1;
      EV5_SetExtInterrupt(PE[0],INTLEVEL_IRQ1);
   } else {
      simetherPending[0] = 0;
      EV5_ClearExtInterrupt(PE[0],INTLEVEL_IRQ1);
   }
}


/* 
 * Check for pending interrupts
 */
int
Simether_Pending(int cpuNum) 
{
   return (simetherPending[cpuNum]);
}




void
Simether_Poll (void) 
{
   SimetherPoll();
}

#define offsetof(s, m)  (size_t)(&(((s *)0)->m))


/* A fundamental assumption of simether.c is that the DevEtherRegisters
 * (machine_defs.h) and the SimetherRegisters structures are isomorphic.
 * We'd better check for this.
 */
void
check_ether_offs(void)
{
  ASSERT(offsetof(DevEtherRegisters, etheraddr[0]) == 
	 offsetof(SimetherRegisters, etheraddr[0]));
  ASSERT(offsetof(DevEtherRegisters, numRcvEntries) == 
	offsetof(SimetherRegisters, numRcvEntries));
  ASSERT(offsetof(DevEtherRegisters, numSndEntries) ==
	 offsetof(SimetherRegisters, numSndEntries));
  ASSERT(offsetof(DevEtherRegisters, numSndChunks) ==
	 offsetof(SimetherRegisters, numSndChunks));
  ASSERT(offsetof(DevEtherRegisters, rcvEntries[0].pAddr) ==
	 offsetof(SimetherRegisters, rcvEntries[0].pAddr));
  ASSERT(offsetof(DevEtherRegisters, rcvEntries[0].maxLen) ==
	 offsetof(SimetherRegisters, rcvEntries[0].maxLen));
  ASSERT(offsetof(DevEtherRegisters, rcvEntries[0].len) ==
	 offsetof(SimetherRegisters, rcvEntries[0].len));
  ASSERT(offsetof(DevEtherRegisters, rcvEntries[0].flag) ==
	 offsetof(SimetherRegisters, rcvEntries[0].flag));
  ASSERT(offsetof(DevEtherRegisters, sndEntries[0].firstChunk) ==
	 offsetof(SimetherRegisters, sndEntries[0].firstChunk));
  ASSERT(offsetof(DevEtherRegisters, sndEntries[0].lastChunk) ==
	 offsetof(SimetherRegisters, sndEntries[0].lastChunk));
  ASSERT(offsetof(DevEtherRegisters, sndEntries[0].flag) ==
	 offsetof(SimetherRegisters, sndEntries[0].flag));
  ASSERT(offsetof(DevEtherRegisters, sndChunks[0].pAddr) ==
	 offsetof(SimetherRegisters, sndChunks[0].pAddr));
  ASSERT(offsetof(DevEtherRegisters, sndChunks[0].len) ==
	 offsetof(SimetherRegisters, sndChunks[0].len));
}
