/*
 * Copyright (C) 1998 by the Board of Trustees
 *    of Leland Stanford Junior University.
 * Copyright (C) 1998 Digital Equipment Corporation
 *
 * This file is part of the SimOS distribution.
 * See LICENSE file for terms of the license.
 *
 */


/* #define DEBUG_SCSI */

#include "simtypes.h"
#include "sim_error.h"
#include "machine_params.h"
#include "machine_defs.h"
#include "registry.h"
#include "console.h"
#include "hd.h"
#include "alpha_access.h"
#include "cpu_interface.h"
#include "gamma.h"
#include "ev5.h"
#include "simos_scsi.h"

#include <stddef.h>
#include <stdlib.h>


static SimOS_SCSI scsiRegs[MAX_MACHINES*SIMOS_SCSI_MAXCTLR];

static int scsiInProgress[MAX_MACHINES][SIMOS_SCSI_MAXCTLR][SIMOS_SCSI_MAXTARG];
static int numInProgress[MAX_MACHINES] = {0};


static int scsiPending[SIM_MAXCPUS];

extern char *bdoorType[];


int SimOS_SCSI_access(int cpuNum, VA vA, int type, Reg *buff)
{
   uint ctrl = ((unsigned long)vA - (unsigned long)(SIMOS_SCSI_REGS))/sizeof(SimOS_SCSI);
/*   VA off = vA & 0xFFFF; */
   VA off = ((unsigned long)vA - (unsigned long)(SIMOS_SCSI_REGS))%sizeof(SimOS_SCSI);
   int machine = M_FROM_CPU(cpuNum);
   int node = FIRST_CPU(machine);

   SimOS_SCSI* currRegs = scsiRegs + ctrl + machine*SIMOS_SCSI_MAXCTLR;

#ifdef DEBUG_SCSI
   CPUPrint("SIMOS_SCSI PC=0x%lx vA=%lx  offset=0x%lx value=0x%lx %s [%ld]\n",
            PE[cpuNum]->PC,vA,off,*buff,bdoorType[type],
            CPUVec.CycleCount(cpuNum));  
#endif  
  ASSERT( off < sizeof(SimOS_SCSI));
   if (off==offsetof(SimOS_SCSI,startIO)) {
      int i;
      unsigned char cmd[SIM_DISK_CMD_SIZE];
      PA pages[SIM_DISK_MAX_DMA_LENGTH];
      ASSERT(type==BDOOR_STORE_DOUBLE || type==BDOOR_STORE_WORD);
      /*
       * start io
       */
      CPUPrint("SCSI_IO: machine=%d ctrl=%d targ=%ld lun=%ld  cmd=0x%lx sgLen=%d\n",
                 machine, ctrl, currRegs->target,currRegs->lun,
                 currRegs->cmd[0],currRegs->sgLen);
      ASSERT (!scsiInProgress[machine][ctrl][currRegs->target]);
      scsiInProgress[machine][ctrl][currRegs->target] = 1;
      numInProgress[machine]++;
      ASSERT (currRegs->sgLen <SIM_DISK_MAX_DMA_LENGTH); 
      for (i=0;i<currRegs->sgLen;i++) {
         
#ifdef DEBUG_SCSI
         CPUPrint("        sg pAddr=0x%lx len=%d\n",
                    currRegs->sgMap[i].pAddr,
                    currRegs->sgMap[i].len);
         
#endif
         ASSERT(currRegs->sgMap[i].pAddr); 
      }
      
      /*
       * conversion to devices/controllers/hd.c format
       */
      for (i=0;i<SIM_DISK_CMD_SIZE;i++) {
         cmd[i] = currRegs->cmd[i];
      }
      for (i=0;i<currRegs->sgLen;i++) {
         pages[i] = (currRegs->sgMap[i].pAddr & ~BITMASK(PAGE_BITS)) & PA_MASK;
         ASSERT (pages[i]);
         if (i) { 
            ASSERT (!(currRegs->sgMap[i].pAddr & BITMASK(PAGE_BITS)));
         }
      }
      pages[i] = 0;
      
      sim_disk_touch(node,ctrl,currRegs->target,NULL);
      
      sim_disk_startio(node,ctrl,currRegs->target,cmd,pages,
                       currRegs->sgMap[0].pAddr & BITMASK(PAGE_BITS));
            
      
   } else if (off >= offsetof(SimOS_SCSI,done[0]) &&
              off < offsetof(SimOS_SCSI,target)) {
      
      int target = (off-offsetof(SimOS_SCSI,done[0])) / sizeof(SCSIReg);
     
      if (type==BDOOR_LOAD_DOUBLE || type==BDOOR_LOAD_WORD) { 
         int done=0;
         int bytesTr=0;
         int errno=0;
         sim_disk_status(node,ctrl,target,&done,&bytesTr,&errno);
#ifdef DEBUG_SCSI
         CPUPrint("SCSI probe DISK%d.%d done=%d bytesTr=%d errno=%d\n",
                    ctrl,target,done,bytesTr,errno);
#endif
         *buff = done;
      } else if (type==BDOOR_STORE_DOUBLE || type==BDOOR_STORE_WORD) {

#ifdef DEBUG_SCSI
         CPUPrint("SCSI ack DISK%d.%d \n",ctrl,target);
#endif
         ASSERT (scsiInProgress[machine][ctrl][target]);
         scsiInProgress[machine][ctrl][target]=0;
         numInProgress[machine]--;
	 sim_disk_iodone(node,ctrl,target);

      } else {
         ASSERT(0);
      }

   } else {
      Reg *ptr = (Reg *) currRegs;
      ptr[off/sizeof(Reg)] = *buff;
      
      ASSERT(type==BDOOR_STORE_DOUBLE || type == BDOOR_STORE_WORD);
   }
   return 0;
}


/*
 * check the interrupt level 
 */


void SimOS_SCSI_intr(int node,int ctrl, int target, int on)
{
   int cpu = FIRST_CPU(M_FROM_CPU(node));

#ifdef DEBUG_SCSI
   CPUPrint("At %ld SimOS_SCSI_intr DISK%d.%d on=%d \n",
	      (uint64)CPUVec.CycleCount(0),ctrl,target,on);
#endif

   scsiPending[cpu] = (numInProgress[M_FROM_CPU(cpu)]>0);
   if (on) {
      EV5_SetExtInterrupt(PE[cpu],INTLEVEL_IRQ1);
   } else {
      EV5_ClearExtInterrupt(PE[cpu],INTLEVEL_IRQ1);
   }
   
}

int SimOS_SCSI_Pending(int cpuNum) 
{
   return scsiPending[cpuNum];
}

