/*
 * Copyright (C) 1998 by the Board of Trustees
 *    of Leland Stanford Junior University.
 * Copyright (C) 1998 Digital Equipment Corporation
 *
 * This file is part of the SimOS distribution.
 * See LICENSE file for terms of the license.
 *
 */



/* *****************************************************************
 * Clock interrupts
 * Clock tick frequency is assumed to be 10millseconds
 * (should confirm this somewhere eventually)
 *
 * The SimOS console does not manage the TLINTRMASK register. 
 * Therefore, we start the clock simply at the start of the simulation
 * Since IPL==max at that point anyway, it does not matter.
 * *****************************************************************/


#include <stddef.h>
#include <stdlib.h>


#include "simtypes.h"
#include "sim_error.h"
#include "machine_params.h"
#include "machine_defs.h"
#include "registry.h"
#include "cpu_interface.h"
#include "gamma.h"
#include "devices.h"
#include "tlaser.h"
#include "tlaser_clock.h"
#include "ev5.h"
#include "x_framebuf.h"
#include "kbd_mouse.h"

extern char *bdoorType[];

static struct ClockCallbackHdr {
   EventCallbackHdr hdr;
   int clockIntr;
   SimTime clockInterval;
} clockCallback[SIM_MAXCPUS];

int TLaserClockIntrPending(int cpuNum)
{
   return clockCallback[cpuNum].clockIntr;
}


static void TLaserClockCallback(int cpuNum, EventCallbackHdr *hdr, void *arg)
{
   struct ClockCallbackHdr *cb = (struct ClockCallbackHdr *)hdr;

#ifdef NT_FRAMEBUFFER
   HandleFramebufEvents();
   RefreshFramebuf(M_FROM_CPU(cpuNum));
#endif
   
   if (!cb->clockIntr) {
      cb->clockIntr = 1;
      EV5_SetExtInterrupt(PE[cpuNum],INTLEVEL_IRQ2);
   }
   EventDoCallback(cpuNum,TLaserClockCallback,hdr,NULL,
		   cb->clockInterval);
}

void TLaserClockAck(int cpuNum)
{
   struct ClockCallbackHdr *cb = &clockCallback[cpuNum];
   ASSERT( cb->clockIntr);
   cb->clockIntr = 0;
   EV5_ClearExtInterrupt(PE[cpuNum],INTLEVEL_IRQ2);
}

void InitTLaserClock(int restoringCpt)
{
   int i;
   
   SimTime interval = NanoSecsToCycles( (1000*1000*1000) / INTRCLOCK_FREQUENCY);

   ASSERT(interval);
   CPUPrint("InitTLaserClock: interval = %ld cycles\n",(Reg64)interval);

   for (i=0;i<TOTAL_CPUS;i++) {
      struct ClockCallbackHdr *cb = (struct ClockCallbackHdr *)&clockCallback[i];
      cb->clockIntr = 0;
      cb->clockInterval = interval;
      if (restoringCpt) {
         EventDoCallback(i,TLaserClockCallback, &cb->hdr,NULL,cb->clockInterval);
      } else {
         /*
          * give it 1 second before timer comes up
          */
         EventDoCallback(i,TLaserClockCallback, &cb->hdr,NULL,INTRCLOCK_FREQUENCY*cb->clockInterval);
      }
   }
}

