/*
 * Copyright (C) 1998 by the Board of Trustees
 *    of Leland Stanford Junior University.
 * Copyright (C) 1998 Digital Equipment Corporation
 *
 * This file is part of the SimOS distribution.
 * See LICENSE file for terms of the license.
 *
 */


#include <stddef.h>
#include <stdlib.h>

#include "simtypes.h"
#include "sim_error.h"
#include "machine_params.h"
#include "machine_defs.h"
#include "registry.h"
#include "console.h"
#include "hd.h"
#include "alpha_access.h"
#include "cpu_interface.h"
#include "gamma.h"
#include "ev5.h"
#include "tlaser.h"
#include "tlaser_clock.h"
#include "tlaser_uart.h"
#include "devices.h"




#define IO_SHIFT 5
#define WIDTH_SHIFT ( IO_SHIFT -2)

/*
 * PZA sits on the PCI bus --> use PCI addressing
 * from io/cam/spo/sim_pza.h
 */


#define PZA_RESET_START_OFFSET	0x0000000CL
#define PZA_ASRCR_OFFSET	0x00000300L

#define AMCSR_LO_OFFSET		0x00000100L      
#define AMCSR_HI_OFFSET		0x00000104L      
#define ABBR_LO_OFFSET		0x00000108L           
#define ABBR_HI_OFFSET		0x0000010CL   
#define DAFQIR_LO_OFFSET	0x00000110L      
#define DAFQIR_HI_OFFSET	0x00000114L     
#define DACQIR_LO_OFFSET	0x00000118L      
#define DACQIR_HI_OFFSET	0x0000011cL      
#define ASR_LO_OFFSET		0x00000140L     
#define ASR_HI_OFFSET		0x00000144L      
#define AFAR_LO_OFFSET		0x00000148L     
#define AFAR_HI_OFFSET		0x0000014CL    
#define AFPR_LO_OFFSET		0x00000150L    
#define AFPR_HI_OFFSET		0x00000154L     
#define ERR_LOG_1_OFFSET	0x000001C0L
#define ERR_LOG_2_OFFSET	0x000001C4L
#define ERR_LOG_3_OFFSET	0x000001C8L
#define ERR_LOG_4_OFFSET	0x000001CCL
#define ERR_LOG_5_OFFSET	0x000001D0L
#define ERR_LOG_6_OFFSET	0x000001D4L
#define ERR_LOG_7_OFFSET	0x000001D8L
#define ERR_LOG_8_OFFSET	0x000001DCL
#define ERR_LOG_9_OFFSET	0x000001E0L
#define ERR_LOG_A_OFFSET	0x000001E4L


/*
 * from simport.h
 */


/*  
 *  SIMport Adapter Status Register Bit Fields Definitions
 *  The ASR register is a 64 bit register.  Currently SIMport
 *  does not define bits 32-63.
 *  Bits 10-30 are Should be Zero fields.
 */
#define SPO_ASR_0	0x00000001	/* Should be Zero */
#define SPO_ASR_1	0x00000002	/* Should be Zero */
#define SPO_ASR_ADSE	0x00000004  	/* Adapter Data Struture Error */
#define SPO_ASR_AMSE	0x00000008	/* Adapter Memory System Error */
#define SPO_ASR_AAC	0x00000010  	/* Adapter Abnormal Condition */ 
#define SPO_ASR_5	0x00000020	/* Should be Zero */
#define SPO_ASR_6	0x00000040	/* Should be Zero */
#define SPO_ASR_7	0x00000080	/* Should be Zero */
#define SPO_ASR_ASIC	0x00000100	/* Adapter Completion Interrupt Request */
#define SPO_ASR_UNIN	0x00000200	/* Unitialized State */
#define SPO_ASR_AME	0x00008000  	/* Adapter Maintenance Error */



struct PZAState {
   int init;
} pzaState[MAX_MACHINES][16];


int TLaserPZA_access(int cpuNum, VA vA, int type, Reg *buff)
{  
   VA base = vA & BITMASK(31); /* clear top bits */
   uint64 width = (base >> WIDTH_SHIFT) & 0x3;
   uint64 pciAddr = (base>>IO_SHIFT);
   uint64 offset  = pciAddr & 0x1ff;
   int ctrl = 0;
   int machine = M_FROM_CPU(cpuNum);
   enum {NONE,BOTH,HIGH,LOW} mask = NONE;

   if (type==BDOOR_LOAD_WORD) {
      uint64 res = 0;
      uint64 resMask;
      switch (offset) { 
      case AMCSR_LO_OFFSET: 
      case AMCSR_HI_OFFSET: 
	 res = 0;
	 mask = (offset==AMCSR_LO_OFFSET?  LOW : HIGH);
	 break;
      case ASR_LO_OFFSET: 
      case ASR_HI_OFFSET: 
	 if (!pzaState[machine][ctrl].init) res |= SPO_ASR_UNIN;
	 mask = (offset==ASR_LO_OFFSET ?  LOW : HIGH);
	 break;
      default: 
	 CPUError("TLaser_PZA PC=0x%lx vA=%lx pci=0x%lx width=0x%lx offset=0x%lx \n",
		  PE[cpuNum]->PC,vA,pciAddr,width,offset);	 
	  
      }
      switch (mask) {
      case BOTH: resMask = res; break;
      case HIGH: resMask = (res>>32) & BITMASK(32); break;
      case LOW:  resMask = res & BITMASK(32); break;
      default: ASSERT(0); resMask = 0;
      }
      /*
       * PCI->Tlaser data conversion. See io/dec/pci/pcia.c 
       */
      
      *buff = resMask<<((pciAddr&3)<<3);
      CPUWarning("TLaser_PZA PC=0x%lx vA=%lx pci=0x%lx width=0x%lx offset=0x%lx value=0x%lx->0x%lx->0x%lx \n",
		 PE[cpuNum]->PC,vA,pciAddr,width,offset,res,resMask,*buff);	 
      return 0;
   }
   
   if (type==BDOOR_STORE_WORD) { 
      switch (offset) { 
      case PZA_RESET_START_OFFSET:
	 pzaState[machine][ctrl].init = 1;
	 CPUWarning("PZA_RESET_START_OFFSET \n");
	 break;
      }
   }	 
   CPUWarning("TLaser_PZA PC=0x%lx vA=%lx pci=0x%lx width=0x%lx offset=0x%lx value=0x%lx %s\n",
	      PE[cpuNum]->PC,vA,pciAddr,width,offset,*buff,bdoorType[type] )  ;  
  
   return 0;
}

