/*
 * Copyright (C) 1998 by the Board of Trustees
 *    of Leland Stanford Junior University.
 * Copyright (C) 1998 Digital Equipment Corporation
 *
 * This file is part of the SimOS distribution.
 * See LICENSE file for terms of the license.
 *
 */


#ifdef NT_FRAMEBUFFER

#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <X11/Xos.h>
#include <X11/Xatom.h>
#include <X11/keysym.h>

#include <stdlib.h>
#include "simtypes.h"
#include "sim_error.h"
#include "machine_params.h"
#include "x_framebuf.h"
#include "kbd_mouse.h"

#define MAX_X_STRING 10

static unsigned char keysym_to_scancode[0x7f] = {
   0, 0, 0, 0, 0, 0, 0, 0,
   0, 0, 0, 0, 0, 0, 0, 0,
   0, 0, 0, 0, 0, 0, 0, 0,
   0, 0, 0, 0, 0, 0, 0, 0,
   0x39, 0x02, 0x28, 0x04, 0x05, 0x06, 0x08, 0x28, /*  !"#$%&' */
   0x0a, 0x0b, 0x09, 0x0d, 0x33, 0x0c, 0x34, 0x35, /* ()*+,-./ */
   0x0b, 0x02, 0x03, 0x04, 0x05, 0x06, 0x07, 0x08, /* 01234567 */
   0x09, 0x0a, 0x27, 0x27, 0x33, 0x0d, 0x34, 0x35, /* 89:;<=>? */
   0x03, 0x1e, 0x30, 0x2e, 0x20, 0x12, 0x21, 0x22, /* @ABCDEFG */
   0x23, 0x17, 0x24, 0x25, 0x26, 0x32, 0x31, 0x18, /* HIJKLMNO */
   0x19, 0x10, 0x13, 0x1f, 0x14, 0x16, 0x2f, 0x11, /* PQRSTUVW */
   0x2d, 0x15, 0x2c, 0x1a, 0x2b, 0x1b, 0x07, 0x0c, /* XYZ[\]^_ */
   0x29, 0x1e, 0x30, 0x2e, 0x20, 0x12, 0x21, 0x22, /* `abcdefg */
   0x23, 0x17, 0x24, 0x25, 0x26, 0x32, 0x31, 0x18, /* hijklmno */
   0x19, 0x10, 0x13, 0x1f, 0x14, 0x16, 0x2f, 0x11, /* pqrstuvw */
   0x2d, 0x15, 0x2c, 0x1a, 0x2b, 0x1b, 0x29        /* xyz{|}~  */
};

/* The region needing updating */
extern int min_x[MAX_MACHINES], max_x[MAX_MACHINES], 
   min_y[MAX_MACHINES], max_y[MAX_MACHINES];

/*
 * Internal functions
 */
static int x_CreateDisplay(void);
static int x_DestroyDisplay(void);
static int x_CreateWindow(int machine, int width, int height);
static int x_DestroyWindow(int machine);
static int x_CreateXImage(int machine);
static int x_DestroyXImage(int machine);
static void x_KeyPress(int machine, KeySym keysym, int press);
static void x_MousePress(int machine, unsigned int button, int press);
static void x_MouseMotion(int machine, int x, int y);

/* Keeps track of how many windows are using the display */
static int displayRefCount = 0;

static char window_array[64];
static char icon_array[16];
static char *window_name = window_array;
static char *icon_name = icon_array;

/*
 * Window state
 */
typedef struct WinState {
   /* X stuff */
   Window win;
   GC gc;
   XImage *ximage;
   Colormap cmap;
   int depth;
   Visual *visual;
   int seenExpose;
   
   /* General stuff */
   unsigned int actual_width, actual_height;
   unsigned int img_width, img_height;
   unsigned char* buf;
} WinState;

static Display *fb_display = NULL;
static int fb_screen_num = 0;
static WinState w_s[MAX_MACHINES];

/*
 * Creates a framebuffer for a display of size actual_x by actual_y
 * This function creates an X window of the desired size, creates an XImage
 * of the same size for copying to it, allocates memory for the XImage,
 * and other startup stuff.
 *
 * Returns a pointer to an 8-bit pseudocolor area of memory for writing
 * pixels to. This is of size img_x by img_y, which may not be exactly the
 * same as actual_x by actual_y.
 * Returns NULL if unsuccessful.
 */
unsigned char* CreateFramebuf(int machine, 
                              unsigned int actual_x, unsigned int actual_y,
                              unsigned int* img_x, unsigned int* img_y)
{
   CPUWarning("\nX_Framebuf: Creating a framebuffer of %u by %u (machine %d).\n",
              actual_x, actual_y, machine);

   ASSERT(w_s[machine].buf == NULL);

   w_s[machine].actual_width = actual_x;
   w_s[machine].actual_height = actual_y;

   if (!x_CreateWindow(machine, w_s[machine].actual_width, w_s[machine].actual_height))
   {
      CPUError("X_Framebuf: Unable to create framebuffer window (machine %d)\n",
               machine);
   }

   if (!x_CreateXImage(machine))
   {
      CPUError("X_Framebuf: Unable to create framebuffer XImage (machine %d)\n",
               machine);
   }

   *img_x = w_s[machine].img_width;
   *img_y = w_s[machine].img_height;

   min_x[machine] = INT_MAX;
   max_x[machine] = INT_MIN;
   min_y[machine] = INT_MAX;
   max_y[machine] = INT_MIN;

   return w_s[machine].buf;
}

/*
 * Destroys the current framebuffer, its window, XImage, and associated memory.
 * Returns 1 if successful, and 0 if not.
 */
int DestroyFramebuf(int machine)
{
   CPUWarning("X_Framebuf: Destroying the current framebuffer (machine %d).\n",
              machine);

   if (!w_s[machine].buf)
   {
      return 0;
   }

   x_DestroyXImage(machine);
   x_DestroyWindow(machine);

   free(w_s[machine].buf);
   w_s[machine].buf = NULL;
   return 1;
}

int x_CreateDisplay(void)
{
   ASSERT(displayRefCount >= 0);
   if (displayRefCount++ == 0)
   {
      if ( (fb_display = XOpenDisplay(NULL)) == NULL)
      {
         CPUError("X_Framebuf: Couldn't open display '%s'\n",
                  XDisplayName(NULL));
      }
      
      fb_screen_num = DefaultScreen(fb_display);
   }
   return 1;
}

int x_DestroyDisplay(void)
{
   ASSERT(displayRefCount > 0);
   if (--displayRefCount == 0)
   {
      XCloseDisplay(fb_display);
   }
   return 1;
}

int x_CreateWindow(int machine, int width, int height)
{
   int status;
   unsigned long pixels[257];
   int i;
   XWMHints wm_hints;
   XClassHint class_hints;
   XSizeHints size_hints;
   XTextProperty windowName, iconName;
   XGCValues values;

   x_CreateDisplay();

   w_s[machine].win = XCreateSimpleWindow(fb_display,
                                          RootWindow(fb_display,fb_screen_num),
                                          0,
                                          0,
                                          width,
                                          height,
                                          4,   /* border width */
                                          0,
                                          0);

   w_s[machine].depth = DefaultDepth(fb_display, fb_screen_num);
   w_s[machine].visual = DefaultVisual(fb_display, fb_screen_num);

   ASSERT(w_s[machine].depth == 8);
   ASSERT(w_s[machine].visual->class == PseudoColor);
      
   w_s[machine].cmap = XCreateColormap(fb_display,
                                       w_s[machine].win,
                                       w_s[machine].visual,
                                       AllocNone);

   if (w_s[machine].cmap == (Colormap) None)
   {
      CPUError("X_Framebuf: Couldn't create colormap (machine %d)\n", machine);
   }

   status = XAllocColorCells(fb_display,
                             w_s[machine].cmap,
                             True,
                             (unsigned long*) NULL,
                             0,
                             pixels,
                             256);

   if (status == 0) 
   {
      CPUError("X_Framebuf: Error allocating color cells. (machine %d)\n", machine);
   }

   for (i=0; i < 256; i++)
   {
      if (pixels[i] != i)
      {
         CPUError("X_Framebuf: Mismatch in allocating color cells. (machine %d)\n",
                  machine);
      }
   }

   XSetWindowColormap(fb_display, w_s[machine].win, w_s[machine].cmap);

   sprintf(window_name, "SimOS-Alpha Framebuffer (machine %d)", machine);
   sprintf(icon_name, "SimOS(%d)", machine);

   if (XStringListToTextProperty(&window_name, 1, &windowName) == 0) 
   {
      CPUError("X_Framebuf: Couldn't allocate window name (machine %d)\n", machine);
   }
   if (XStringListToTextProperty(&icon_name, 1, &iconName) == 0) 
   {
      CPUError("X_Framebuf: Couldn't allocate icon name (machine %d)\n", machine);
   }
   size_hints.flags = PPosition | PSize | PMinSize | PMaxSize;
   size_hints.max_width = size_hints.min_width = width;
   size_hints.max_height = size_hints.min_height = height;
   wm_hints.initial_state = NormalState;
   wm_hints.input = True;
   wm_hints.flags = StateHint | InputHint;
   class_hints.res_name = "SimOS-Alpha";
   class_hints.res_class = "SimOS";

   XSetWMProperties(fb_display, w_s[machine].win, &windowName, &iconName,
                    NULL, 0, &size_hints, &wm_hints, &class_hints);

   XSelectInput(fb_display, w_s[machine].win, ExposureMask | KeyPressMask | 
                KeyReleaseMask | ButtonPressMask | ButtonReleaseMask | 
                PointerMotionMask | EnterWindowMask | LeaveWindowMask);

   w_s[machine].gc = XCreateGC(fb_display, w_s[machine].win, 0, &values);

   XMapWindow(fb_display, w_s[machine].win);
   XSync(fb_display, 0);

   return 1;
}

int x_DestroyWindow(int machine)
{
   XFreeColormap(fb_display, w_s[machine].cmap);
   XDestroyWindow(fb_display, w_s[machine].win);
   w_s[machine].win = 0;
   x_DestroyDisplay();
   w_s[machine].seenExpose = 0;

   return 1;
}

int HandleFramebufEvents(void)
{
   XEvent event;
   XExposeEvent *expose_event;
   XKeyEvent *key_event;
   XButtonEvent *button_event;
   XMotionEvent *motion_event;
   XComposeStatus compose;
   KeySym keysym;
   char buffer[MAX_X_STRING];
   int bufsize = MAX_X_STRING;
   int count;
   int machine;

   while (displayRefCount && (XPending(fb_display) > 0)) 
   {
      XNextEvent(fb_display, &event);
      
      for (machine = 0; machine < NUM_MACHINES; machine++)
      {
         if (w_s[machine].win == ((XAnyEvent *)&event)->window)
         {
            ASSERT(w_s[machine].ximage);
            switch (event.type) 
            {
            case Expose:
               w_s[machine].seenExpose = 1;
               expose_event = (XExposeEvent *) &event;
               if (expose_event->x < min_x[machine])
                  min_x[machine] = expose_event->x;
               if ((expose_event->x + expose_event->width) > max_x[machine])
                  max_x[machine] = (expose_event->x + expose_event->width);
               if (expose_event->y < min_y[machine])
                  min_y[machine] = expose_event->y;
               if ((expose_event->y + expose_event->height) > max_y[machine])
                  max_y[machine] = (expose_event->y + expose_event->height);
               break;
               
            case KeyPress:
               key_event = (XKeyEvent *) &event;
               count = XLookupString(key_event, buffer, bufsize, &keysym, &compose);
               x_KeyPress(machine, keysym, 1);
               break;
               
            case KeyRelease:
               key_event = (XKeyEvent *) &event;
               count = XLookupString(key_event, buffer, bufsize, &keysym, &compose);
               x_KeyPress(machine, keysym, 0);
               break;

            case ButtonPress:
               button_event = (XButtonEvent *) &event;
               x_MousePress(machine, button_event->button, 1);  
               break;

            case ButtonRelease:
               button_event = (XButtonEvent *) &event;
               x_MousePress(machine, button_event->button, 0);  
               break;

            case MotionNotify:
               motion_event = (XMotionEvent *) &event;
               x_MouseMotion(machine, motion_event->x, motion_event->y);
               break;
               
            default:
               break;
            }
            break;
         }
      }
   }
   return 1;
}

int x_CreateXImage(int machine)
{
   ASSERT(!w_s[machine].ximage);
  
   w_s[machine].ximage = XCreateImage(fb_display, 
                                      w_s[machine].visual,
                                      w_s[machine].depth,
                                      ZPixmap,
                                      0,
                                      NULL,
                                      w_s[machine].actual_width,
                                      w_s[machine].actual_height,
                                      XBitmapPad(fb_display),
                                      0);
   
   if (!w_s[machine].ximage)
   {
      CPUError("X_Framebuf: Couldn't create XImage (machine %d)\n", machine);
   }

   w_s[machine].img_width = w_s[machine].ximage->bytes_per_line;
   w_s[machine].img_height = w_s[machine].actual_height;
   ASSERT(w_s[machine].ximage->bits_per_pixel == 8);

   w_s[machine].buf = (char *)malloc( w_s[machine].img_width * w_s[machine].img_height );
   if (!w_s[machine].buf)
   {
      CPUError("X_Framebuf: Couldn't allocate XImage memory (machine %d)\n", machine);
   }

   w_s[machine].ximage->data = w_s[machine].buf;
      
   return 1;
}

int x_DestroyXImage(int machine)
{
   ASSERT(w_s[machine].ximage);
   
   XDestroyImage(w_s[machine].ximage);

   w_s[machine].ximage = NULL;

   w_s[machine].buf = NULL;

   return 1;
}

/*
 * Sets the palette of the framebuffer to the 256 32-bit RGB values passed in.
 * Returns 1 if successful, and 0 if not.
 */
int SetPalette(int machine, unsigned col[256])
{
   XColor xcolor;
   int i;

   if (w_s[machine].win)
   {
      CPUPrint("X_Framebuf: Setting the palette (machine %d).\n", machine);
   
      for (i = 0; i < 256; i++)
      {
         xcolor.pixel = i;
         xcolor.red = ((65535 * (long)(col[i]&0xff))/256);
         xcolor.green = ((65535 * (long)((col[i]>>8)&0xff))/256);
         xcolor.blue = ((65535 * (long)((col[i]>>16)&0xff))/256);
         xcolor.flags = DoRed | DoGreen | DoBlue;
         
         XStoreColor(fb_display, w_s[machine].cmap, &xcolor);
      }
   }

   return 1;
}

/*
 * Refreshes the screen - updates the screen window from the XImage 
 */
int RefreshFramebuf(int machine)
{
   if (w_s[machine].ximage && w_s[machine].seenExpose && (min_x[machine] < INT_MAX))
   {
#ifdef FB_DEBUG
      CPUWarning("X_Framebuf: Refreshing framebuf (%dx%d to %dx%d) (machine %d).\n",
                 min_x[machine], min_y[machine], 
                 max_x[machine], max_y[machine], machine);
#endif
      if (max_x[machine] > w_s[machine].actual_width)
         max_x[machine] = w_s[machine].actual_width;
      XPutImage(fb_display, w_s[machine].win, w_s[machine].gc, w_s[machine].ximage, 
                min_x[machine], min_y[machine], 
                min_x[machine], min_y[machine], 
                max_x[machine] - min_x[machine] + 1, 
                max_y[machine] - min_y[machine] + 1);

      min_x[machine] = INT_MAX;
      max_x[machine] = INT_MIN;
      min_y[machine] = INT_MAX;
      max_y[machine] = INT_MIN;
      
      return 1;
   }
   else
   {
      return 0;
   }
}

/*
 * Generates a scancode from a keypress
 * If press is 1, then key pressed
 * If press is 0, then key released
 */
void x_KeyPress(int machine, KeySym keysym, int press)
{
   unsigned char scancode;

   if ((keysym >= XK_space) && (keysym <= XK_asciitilde))
   {
      scancode = keysym_to_scancode[keysym];
   }
   else
   {
      switch(keysym)
      {
      case XK_KP_0:
      case XK_Insert:
      case XK_KP_Insert:   scancode = 0x52; break;
      case XK_KP_1:
      case XK_End:
      case XK_KP_End:      scancode = 0x4f; break;
      case XK_KP_2:
      case XK_Down:
      case XK_KP_Down:     scancode = 0x50; break;
      case XK_KP_3:
      case XK_F35:  /* Sun keyboard */
      case XK_Page_Down:
      case XK_KP_Page_Down:scancode = 0x51; break;
      case XK_KP_4:
      case XK_Left:
      case XK_KP_Left:     scancode = 0x4b; break;
      case XK_KP_5:        scancode = 0x4c; break;
      case XK_KP_6:
      case XK_Right:
      case XK_KP_Right:    scancode = 0x4d; break;
      case XK_KP_7:
      case XK_Home:
      case XK_KP_Home:     scancode = 0x47; break;
      case XK_KP_8:
      case XK_Up:
      case XK_KP_Up:       scancode = 0x48; break;
      case XK_KP_9:
      case XK_Page_Up:
      case XK_KP_Page_Up:  scancode = 0x49; break;
      case XK_Delete:      scancode = 0x53; break;
      case XK_BackSpace:   scancode = 0x0e; break;
      case XK_Tab:         scancode = 0x0f; break;
      case XK_Return:      scancode = 0x1c; break;
      case XK_KP_Subtract: scancode = 0x4a; break;
      case XK_KP_Add:      scancode = 0x4e; break;
      case XK_Escape:      scancode = 0x01; break;
      case XK_F1:          scancode = 0x3b; break;
      case XK_F2:          scancode = 0x3c; break;
      case XK_F3:          scancode = 0x3d; break;
      case XK_F4:          scancode = 0x3e; break;
      case XK_F5:          scancode = 0x3f; break;
      case XK_F6:          scancode = 0x40; break;
      case XK_F7:          scancode = 0x41; break;
      case XK_F8:          scancode = 0x42; break;
      case XK_F9:          scancode = 0x43; break;
      case XK_F10:         scancode = 0x44; break;
      case XK_F11:         scancode = 0x57; break;
      case XK_F12:         scancode = 0x58; break;
      case XK_Control_L:   scancode = 0x1d; break;
      case XK_Shift_L:     scancode = 0x2a; break;
      case XK_Shift_R:     scancode = 0x36; break;
      case XK_Caps_Lock:   scancode = 0x3a; break;
      case XK_Num_Lock:    scancode = 0x45; break;
      case XK_Alt_L:       scancode = 0x38; break;
      default:
         CPUWarning("x_KeyPress: keysym %x not handled\n", 
                    (unsigned int)keysym);
         return;
      }
   }
   
   if (!press)
   {
      scancode |= 0x80;
   }

   KbdMouse_KeyPress(machine, scancode);
}


/*
 * Mouse support stuff
 */

#define LEFT_BUTTON_DOWN 0x01 
#define RIGHT_BUTTON_DOWN 0x02 
#define MIDDLE_BUTTON_DOWN 0x04 

static unsigned int buttonState[MAX_MACHINES];
static int pointerX[MAX_MACHINES];
static int pointerY[MAX_MACHINES];
static int buttonId[] = { 0, LEFT_BUTTON_DOWN, MIDDLE_BUTTON_DOWN, RIGHT_BUTTON_DOWN };

void x_MousePress(int machine, unsigned int button, int press)
{
   if (press)
   {
      buttonState[machine] |= buttonId[button];
   } 
   else
   {
      buttonState[machine] &= ~(buttonId[button]);
   }

#ifdef FB_DEBUG 
   CPUWarning("Machine %d: button state is [%d, %d, %d]\n",
              machine,
              (buttonState[machine]&(LEFT_BUTTON_DOWN))!=0,
              (buttonState[machine]&(MIDDLE_BUTTON_DOWN))!=0,
              (buttonState[machine]&(RIGHT_BUTTON_DOWN))!=0);
#endif

   KbdMouse_MouseEvent(machine, 0, 0, buttonState[machine]);
}

void x_MouseMotion(int machine, int x, int y)
{
   int x_delta, y_delta;

   x_delta = x - pointerX[machine];
   y_delta = y - pointerY[machine];
   pointerX[machine] = x;
   pointerY[machine] = y;

#ifdef FB_DEBUG 
   CPUWarning("Machine %d: moved %d,%d to %d,%d\n",
              machine, x_delta, y_delta, x, y);
#endif

   KbdMouse_MouseEvent(machine, x_delta, -y_delta, buttonState[machine]);
}

#endif /* NT_FRAMEBUFFER */

