/************************************************************
 *
 *  intr.c
 *
 ************************************************************/

#include <stdlib.h>
#include "simtypes.h"
#include "iodevs.h"
#include "iospace.h"
#include "intr.h"
#include "sim_error.h"

#define ICW1_ICW4           0x01
#define ICW1_SINGLE_PIC     0x02
#define ICW1_LEVEL_TRIG     0x08
#define ICW1_INIT           0x10

#define ICW2_MASK           0xf8

#define ICW3_SID_MASK       0x07

#define ICW4_8086_MODE      0x01
#define ICW4_AUTO_EOI       0x02
#define ICW4_MASTER         0x04
#define ICW4_BUFFERED       0x08
#define ICW4_FULLY_NESTED   0x10

#define NOT_REACHED  abort

/*
 *   OK, this looks backwards but you are a slave if you *have* a master,
 *   and vice versa.
 */
#define IS_MASTER(_pic)        (!(_pic)->master)

static void PIC_init(struct IODevices *iod, PIC *pic, uint8 addr, PIC *master, int cascade_IRQ);
static void PIC_recalcINT(IODevices *iod, PIC *pic);
static void PIC_raiseIRQ(IODevices *iod, PIC *pic, int local_IRQ_line);
static int PIC_ackINT(IODevices *iod, PIC *pic);

static int cmdPort(IODevices *iod, uint16 port, int size, int inout, void *byte, void *clientData);
static int maskPort(IODevices *iod, uint16 port, int size, int inout, void *byte, void *clientData);


/************************************************************
 *   bus functions
 ************************************************************/

void Intr_Init(IODevices *iod)
{
   /*
    * encodes PC motherboard wiring - pic1 is cascaded to pic0 on pic0's IRQ 2
    */
   
   PIC_init(iod, &(iod->pic[0]), 0x20, NULL, 0);
   PIC_init(iod, &(iod->pic[1]), 0xA0, &(iod->pic[0]), 2);
}

/*
 *   called by I/O devices to raise an interrupt
 */
void Intr_Raise(IODevices *iod, int IRQ)
{
   /*
    * this function encodes the wiring of the PICs on the PC mother board
    *
    *   IRQ line 2 on the bus is wired into IRQ 9 on the
    *   controllers. This is because IRQ 2 on the controllers is used for
    *   cascading the slave to the master.
    */

   ASSERT((IRQ >= 0) && (IRQ < 16));

   if (IRQ == 2) {
      PIC_raiseIRQ(iod, &(iod->pic[1]), 1);
   } else if (IRQ < 8) {
      PIC_raiseIRQ(iod, &(iod->pic[0]), IRQ);
   } else {
      PIC_raiseIRQ(iod, &(iod->pic[1]), IRQ - 8);
   }
}

/*
 *   called by the processor to check for interrupts
 */
int 
Intr_Check(IODevices *iod)
{
   int vector;

   if (iod->pic[0].INT_line) {
      vector = PIC_ackINT(iod, &(iod->pic[0]));
      return vector;
   }
   return NO_INTR_PENDING;
}


/************************************************************
 *   PIC functions - faithfully emulates the 8259 PIC
 ************************************************************/

void PIC_init(struct IODevices *iod, PIC *pic, uint8 addr, PIC *master, int cascade_IRQ)
{
   pic->ICW1 = 0;
   pic->ICW2 = 0;
   pic->ICW3 = 0;
   pic->ICW4 = 0;
   pic->OCW = 0;
   pic->IRR = 0;
   pic->IMR = 0;
   pic->ISR = 0;
   pic->INT_line = 0;
   pic->nextInitWord = 1;
   pic->master = master;
   pic->slave_chain = NULL;
   pic->cascade_IRQ = cascade_IRQ;

   if (!IS_MASTER(pic)) {
      /* I am a slave, put me on master's slave chain */
      pic->slave_chain = master->slave_chain;
      master->slave_chain = pic;
   }
   
   IOSpace_Register(iod, addr, (IOSpace_Func)cmdPort, (void*)pic);
   IOSpace_Register(iod, addr+1, (IOSpace_Func)maskPort, (void*)pic);
}

void PIC_recalcINT(IODevices *iod, PIC *pic)
{
   int i;
   uint8 ibit;
   uint8 pending;

   pending = pic->IRR & ~pic->IMR;
   if (pending) {
      for (i=0; i<8; i++) {
         ibit = (1 << i);
         if (pic->ISR & ibit) {
            break;
         } else if (pending & ibit) {
            pic->INT_line = 1;
            if (!IS_MASTER(pic)) {
               PIC_raiseIRQ(iod, pic->master, pic->cascade_IRQ);
            }
            break;
         }
      }
   }
}

void PIC_raiseIRQ(IODevices *iod, PIC *pic, int local_IRQ_line)
{
   uint8 ibit;
   ASSERT(local_IRQ_line < 8);
   ibit = 1 << local_IRQ_line;
   pic->IRR |= ibit;
   PIC_recalcINT(iod, pic);
}

int PIC_ackINT(IODevices *iod, PIC *pic)
{
   int i;
   uint8 ibit;
   uint8 pending;
   PIC *slave;

   ASSERT(pic->INT_line);
   
   pending = pic->IRR & ~pic->IMR;
   if (pending) {
      for (i=0; i<8; i++) {
         ibit = (1 << i);
         ASSERT(!(pic->ISR & ibit));
         if (pending & ibit) {
            pic->IRR &= ~ibit;
            pic->ISR |= ibit;
            pic->INT_line = 0;
            
            if (IS_MASTER(pic) && (pic->ICW3 & ibit)) {
               /* ack slaves INT request */
               for (slave = pic->slave_chain; slave; slave = slave->slave_chain) {
                  if ((slave->ICW3 & ICW3_SID_MASK) == i) {
                     ASSERT(slave->INT_line);
                     return PIC_ackINT(iod, slave);
                  }
               }

               NOT_REACHED();
               
            } else {
               ASSERT(i < 8);
               return (pic->ICW2 & ICW2_MASK) | i;
            }
         }
      }
   }

   NOT_REACHED();
   return 0;
}

int cmdPort(IODevices *iod, uint16 port, int size, int inout, void *byte, void *clientData)
{
   PIC *pic = (PIC *)clientData;
   int i;
   uint8 ibit;
   
   if (inout == PORT_OUT) {
      pic->OCW = *(uint8 *)byte;

      if ((pic->OCW & ICW1_INIT) || (pic->nextInitWord == 1)) {
         ASSERT((*(uint8 *)byte & (ICW1_LEVEL_TRIG | ICW1_INIT)) == (ICW1_INIT));
         /* reset PIC */
         pic->ICW1 = *(uint8 *)byte;
         pic->ICW2 = 0;
         pic->ICW3 = 0;
         pic->ICW4 = 0;
         pic->OCW = 0;
         pic->IRR = 0;
         pic->IMR = 0;
         pic->ISR = 0;
         pic->INT_line = 0;
         pic->nextInitWord = 2;

         PIC_recalcINT(iod, pic);

      } else if (pic->OCW == 0x20) {
         /* end of interrupt */
         for (i=0; i<8; i++) {
            ibit = (1 << i);
            if (pic->ISR & ibit) {
               pic->ISR &= ~ibit;
               break;
            }
         }

         PIC_recalcINT(iod, pic);

      } else if ((pic->OCW == 0x0A) || (pic->OCW == 0x0B)) {
         /* controls next port access - no action */

      } else {
         NOT_REACHED();
      }

   } else {
      switch (pic->OCW) {
      case 0x0A:
         *(uint8 *)byte = pic->IRR;
         break;
      case 0x0B:
         *(uint8 *)byte = pic->ISR;
         break;
      default:
         NOT_REACHED();
         break;
      }
   }

   return 0;
}

int maskPort(IODevices *iod, uint16 port, int size, int inout, void *byte, void *clientData)
{
   PIC *pic = (PIC *)clientData;

   if (pic->nextInitWord) {
      ASSERT(inout == PORT_OUT);
      switch (pic->nextInitWord) {
      case 2:
         ASSERT(!(*(uint8 *)byte & ~ICW2_MASK));
         pic->ICW2 = *(uint8 *)byte;
         pic->nextInitWord = 3;
         break;
      case 3:
         pic->ICW3 = *(uint8 *)byte;         
         pic->nextInitWord = (pic->ICW1 & ICW1_ICW4) ? 4 : 0;
         break;
      case 4:
         ASSERT(pic->ICW1 & ICW1_ICW4);
         ASSERT((*(uint8 *)byte & (ICW4_8086_MODE|ICW4_AUTO_EOI|ICW4_BUFFERED|ICW4_FULLY_NESTED))
                == ICW4_8086_MODE);
         pic->ICW4 = *(uint8 *)byte;
         pic->nextInitWord = 0;
         break;
      default:
         NOT_REACHED();
         break;
      }
      
   } else if (inout == PORT_IN) {
      *(uint8 *)byte = pic->IMR;

   } else {
      pic->IMR = *(uint8 *)byte;
      PIC_recalcINT(iod, pic);
   }

   return 0;
}

