/*
 * Copyright (C) 1996 by the Board of Trustees
 *    of Leland Stanford Junior University.
 * 
 * This file is part of the SimOS distribution. 
 * See LICENSE file for terms of the license. 
 *
 */

/*****************************************************************
 * simmisc.c
 *
 * Miscellaneous SimOS stuff which had no better place to go.
 * Dan Teodosiu, 07/96
 *
 *****************************************************************/

#include <sys/signal.h>
#ifndef i386
# include <sys/sysinfo.h>
#endif
#include <sys/times.h>
#include <sys/types.h>
#include <sys/termios.h>
#include <sys/time.h>
#include <sys/file.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <netinet/tcp.h>
#include <sys/mman.h>
#include <errno.h>
#include <unistd.h>
#include <fcntl.h>
#include <stdio.h>
#include <setjmp.h>
#include <string.h>
#include <memory.h>
#include <ctype.h>
#include <math.h>
#include <stdlib.h>
#include <sys/wait.h>
#include <sys/socket.h>
#include <netdb.h>
#include <signal.h>
#include <setjmp.h>

#include "sim.h"
#include "machine_params.h"
#include "simstats.h"
#include "cpu_interface.h"
#include "sim_error.h"
#include "cpu_interface.h"
#include "registry.h"
#include "expect.h"
#include "file_formats.h"
#include "console.h"
#include "simmisc.h"
#include "startup.h"
#include "../../caches/memref.h"
#include "../../memsystems/memsys.h"
#include "../x86sim/simos_interface.h"
#include "iospace.h"

/* Shared array of CPUStates */
extern CPUState* SBase;

char *FPromFile;

#define MAX_SCNS 50

static void  HackUpLinuxStartupEnv(int cpuNum, VA entry);
static void HackUpInstallTimer(int cpuNum, unsigned int interval, unsigned int timeLeft);

/* Storage for sim_misc catch-all structure */
SimMisc sim_misc;

bool tookCheckpoint = FALSE;
bool CPUlaunched[SIM_MAXCPUS];
CPUType myCPUType = BASE;

void Chal2CycleAdvance (void);

/*****************************************************************
 * SimulatorEnter - 
 * This guy is called on the initial switch into a detailed cpu
 * simulator. All switches after this go through SimulatorSwitch.
 *****************************************************************/
void
SimulatorEnter(CPUType simulator, int restoreCkpt, int swtch)
{
  int cpu;

  for (cpu = 0; cpu < TOTAL_CPUS; cpu++) {
    /* Print where the CPU is. */
    if ((swtch || restoreCkpt) && (SBase[cpu].cpuStatus == cpu_not_booted)) {
       /* don't change state for failed CPUs */
    } else if ((!CPUlaunched[cpu]) && (!restoreCkpt) && (!swtch) ) {
      SBase[cpu].outOfSlaveLoop = 0;
      SBase[cpu].cpuStatus      = cpu_not_booted;
    } else {
      SBase[cpu].outOfSlaveLoop = 1;
      SBase[cpu].cpuStatus      = cpu_running;
    } 
    CPUPrint("cpu%d type=%s status=%u\n",
      cpu, simName[simulator], SBase[cpu].cpuStatus);
    myCPUType = simulator;
  }
  simosCPUType = simulator;
  
  switch (simulator) {
    case NO_CPU: 
      exit (0);
      break;
    
    case X86SIM: 
      MemRefInit ();
      X86SimEnter(swtch);
      break; 

    default:
      CPUError("Bad CPU type in SimulatorEnter()");
  } 
}
  
/*****************************************************************
 * SimulatorSwitch
 * 
 * For now, the simulators just call this function to switch to each
 * other... if we get into serious sampling, we can't have the stack
 * grow that much.
 *****************************************************************/

static struct {
   jmp_buf jmpEnv;
   int defined;
} simulatorSwitch;


void 
SimulatorSwitch(CPUType oldSimulator, CPUType newSimulator)
{
   if (newSimulator != NO_CPU) {
      Sim_Warning("SimOS: SWITCHING from %s to %s\n",
                  simName[oldSimulator], simName[newSimulator]);
   } else {
      Sim_Warning("SimOS: exit\n");
   }

  ASSERT(myCPUType == oldSimulator);

  myCPUType = newSimulator;
  simosCPUType = newSimulator;

  if (!simulatorSwitch.defined) {
     setjmp(simulatorSwitch.jmpEnv);
     simulatorSwitch.defined = 1;
  } else { 
     longjmp(simulatorSwitch.jmpEnv,1);
  }

  /* becasue cpu switch causes cyclecount to drop to 0, need to adjust callbacks */
  /* xxx AdjustEventCallbackTime(); */
  SimulatorEnter(myCPUType, 0, 1);

}

/*****************************************************************
 * Initialize (some of the) values in simmisc.c
 *****************************************************************/
void
SimMiscInit(void)
{
  int i;

  for (i = 0; i < SIM_MAXCPUS; i++) 
    CPUlaunched[i] = FALSE;
}

/*****************************************************************
 *
 * Support for loading and starting a kernel image.
 *
 *****************************************************************/


/*****************************************************************
 *
 * Support for loading and starting a linux kernel image
 * (assumes elf format)
 *
 *****************************************************************/

#define MAX_STRINGLEN 512
static int
LoadHeaders(char *pathname, 
		 Elf32_Ehdr *ehdr,
                 FILHDR *fileHeader,
		 SCNHDR secHeaders[MAX_SCNS], 
		 int *fd,
                 VA *entry)
{

  int retval;
  int i;
  char strs[MAX_STRINGLEN];
  char *secName;
  Elf32_Shdr shdr[MAX_SCNS];
  Elf32_Shdr *strSecHdr;


  if ((*fd = open(pathname, O_RDONLY, 0)) < 0) {
    Sim_Warning("Problem opening kernel named %s\n", pathname);
    retval = errno;
    goto error;
  }

  retval = read(*fd,(char *)ehdr, sizeof(Elf32_Ehdr));
  if ((retval != sizeof(Elf32_Ehdr))) {
    Sim_Warning("LoadHeaders:  Bad file header\n");
    retval = ENOEXEC;
    goto error;
  }

  if (!IS_ELF(*ehdr)) {
    Sim_Warning("LoadHeaders:  Not an ELF file header\n");
    retval = ENOEXEC;
    goto error;
  }

  if ((ehdr->e_shnum) >= MAX_SCNS) {
    Sim_Warning("Not enough section headers in Elf32LoadHeaders\n");
    retval = ENOEXEC;
    goto error;
  }

  fileHeader->f_nscns = (ehdr->e_shnum);

  *entry = (ehdr->e_entry);

  if (lseek(*fd, (ehdr->e_shoff), SEEK_SET) < 0) {
     retval = errno;
     Sim_Warning("Could not lseek to section headers in ElfLoadHeaders\n");
     goto error;
  }
  retval = read(*fd,(char *)shdr, ((ehdr->e_shnum)*sizeof(Elf32_Shdr)));
  if (retval < 0) {
    retval = errno;
    Sim_Warning("Could not read section headers in Elf32LoadHeaders\n");
    goto error;
  }
  if (retval != sizeof(Elf32_Shdr) * (ehdr->e_shnum)) {
    Sim_Warning("Read too few section headers in Elf32LoadHeaders\n");
    retval = ENOEXEC;
    goto error;
  }

  strSecHdr = &shdr[(ehdr->e_shstrndx)];

  if (lseek(*fd, (strSecHdr->sh_offset), SEEK_SET) < 0) {
     retval = errno;
     Sim_Warning("Could not lseek to strings in ElfLoadHeaders\n");
     goto error;
  }

  if ( (strSecHdr->sh_size) >= sizeof(strs)) {
     Sim_Warning("String table too large if ElfLoadHeaders\n");
     retval = ENOEXEC;
     goto error;
  }
  retval = read(*fd,(char*)strs, (strSecHdr->sh_size));
  if (retval !=  (strSecHdr->sh_size)){
     if (retval < 0) { 
        retval = errno;
        Sim_Warning("Could not reads strings in ElfLoadHeaders\n");
     } else {
        retval = ENOEXEC;
        Sim_Warning("Read too few strings in Elf32LoadHeaders\n");
     }
     goto error;
  }     
  
  for (i=0; i<(ehdr->e_shnum); i++) {
    secName = strs + (shdr[i].sh_name);
    memcpy(secHeaders[i].s_name,secName,8);
    secHeaders[i].s_vaddr = (shdr[i].sh_addr);
    secHeaders[i].s_size = (shdr[i].sh_size);
    secHeaders[i].s_scnptr = (shdr[i].sh_offset);
  }

  return 0;

 error:
  (void) close(*fd);
  return retval;


}

static int
LoadImage(char* pathname,   /* boot image filename */
          int   machine,    /* machine */
	  VA    img_addr,   /* image start address */
	  VA    ld_addr,    /* load start addres (if diff than above, reloc) */
	  int   nbytes,     /* image length (max.) */
	  bool  reloc_data, /* relocate data segments? */
	  bool* elfFormat,  /* out: ELF file? */
	  VA*   entry       /* out: entry point */
	  )
{
  int       fd, retval, i;
  Elf32_Ehdr header;
  FILHDR    fileHeader;
  SCNHDR    secHeaders[MAX_SCNS];
  FILHDR*   hdrPtr;

  retval = LoadHeaders(pathname, &header, &fileHeader, secHeaders, &fd, entry);
  if (retval) return retval;

  hdrPtr = &fileHeader;
  ASSERT((hdrPtr->f_nscns) < MAX_SCNS);

  for (i = 0; i < (hdrPtr->f_nscns); i++ ) {

    PA  sec_addr  = (VA)(secHeaders[i].s_vaddr);
    MA  read_addr_ma = PHYS_TO_MEMADDR(M_FROM_CPU(0), sec_addr);
    int size         = (secHeaders[i].s_size);
    
    if ((strcmp(secHeaders[i].s_name, ".text")  == 0)  ||
	(strcmp(secHeaders[i].s_name, ".data")  == 0)  ||
	(strcmp(secHeaders[i].s_name, ".sdata") == 0)  ||
        (strcmp(secHeaders[i].s_name, ".rodata") == 0) ||
        (strcmp(secHeaders[i].s_name, ".srdata") == 0) ||
        (strcmp(secHeaders[i].s_name, ".lit8") == 0)   ||
	(strcmp(secHeaders[i].s_name, ".rdata") == 0)  ||
        (strcmp(secHeaders[i].s_name, ".got") == 0) ) {
      
      
      if (lseek(fd, (secHeaders[i].s_scnptr), SEEK_SET) !=
	  (secHeaders[i].s_scnptr)) {
	retval = errno;
	CPUError("Bad section seek in LoadImage\n");
	goto error;
      }
      
      
      if (read(fd, read_addr_ma, size) != size) {
	retval = errno;
	CPUError("Bad section read in LoadImage - "
		 "size = %d, hold = %#x errno %d\n",
		 size,  sec_addr, errno);
	goto error;
      }

    } else if ((strcmp(secHeaders[i].s_name, ".bss") == 0) ||
	       (strcmp(secHeaders[i].s_name, ".sbss") == 0)) {
      memset(read_addr_ma, 0, size);

    }

  }

  retval = 0;
  goto done;
  
error:
done:
  (void) close(fd);
  return retval;
}

#ifdef notdef

/*
 * Copy bootline arguments (for master CPU) into kernel memory so that
 * kernel will be able to read them.
 */
static void
store_mem_word(int cpu, VA addr, int val)
{
#ifdef notdef
  PromWord valBE;


  if (sizeof(PromWord) == 4) {
     valBE = HO2BE_4(val);
  } else {
     valBE = HO2BE_8(val);
  }
   
  if (sim_misc.myCPUType == MIPSY) {
     MemRefDebugWriteData(cpu,addr,K0_TO_PHYS(addr),(char*)&valBE,sizeof(PromWord));
  } else {
    *((PromWord*)K0_TO_MEMADDR(M_FROM_CPU(cpu), addr)) = valBE;
  }
#endif
}

static void
store_mem_byte(int cpu, VA addr, char c)
{
#ifdef notdef
  ASSERT(IS_KSEG0(addr));
  if (sim_misc.myCPUType == MIPSY) {
     MemRefDebugWriteData(cpu,addr,K0_TO_PHYS(addr),(char*)&c,sizeof(char));
  } else {
     *((char*)K0_TO_MEMADDR(M_FROM_CPU(cpu), addr)) = c;
  }
#endif
}
#endif
static void
copy_bootline_args(int machine, int argc, char** argv,
		   VA* argcVA, VA* argvVA)
{
#ifdef notdef
    int cpu  = FIRST_CPU(machine);
    VA argz  = (SIMCP0_KSEG0_ADDR + ARGZONE_OFFS);
    VA strp  = argz + (argc + 2)*sizeof(VA);
    VA argvp = argz + sizeof(VA);
    int i, j;
    
    *argcVA = argz;
    *argvVA = argvp;

    /* first word: argc */
    store_mem_word(cpu, argz, (PromWord)argc);
    
    for (i = 0; i < argc; i++) {
        /* store pointer in argv zone */
        store_mem_word(cpu, argvp + (sizeof(PromWord)*i), (PromWord)strp);
      
        /* copy argument string */
        for (j = 0; argv[i][j]; j++) {
            store_mem_byte(cpu, strp, argv[i][j]); 
            strp++;
        }
        store_mem_byte(cpu, (VA)strp, 0); 
        strp++;
    }
    /* store final 0 in argv */
    store_mem_word(cpu, argvp+sizeof(PromWord)*i, 0); 
#endif
}


/*
 * Read and start the program.
 */
static void
LoadNgo(char *pathname, int argc, char **argv)
{
  int            retval;
  int            firstCPU;
  VA             entry=0;
  VA             argcVA, argvVA;
#ifdef SIM_LINUX
  VA             endaddr;
#endif
  bool           elfFormat = FALSE;
  int            machine;
  
  for (machine = 0; machine < NUM_MACHINES; machine++) {

      retval = LoadImage(pathname,
                         machine, 
                         (VA)0,
			 (VA)0, 
                         MEM_SIZE(machine) / NUM_CPUS(machine),
			 FALSE,
                         &elfFormat,
                         &entry);
  
#ifdef SIM_LINUX
      if (elfFormat) {
          
          retval = 0;
          
          TagInit(machine, &tags);
          
          /* copy tag data to end of vmem --ejs */
          endaddr = (VA)SIMCP0_KSEG0_ADDR +
              (MEM_SIZE(machine) / NUM_CELLS(machine));
          if (memcpy(K0_TO_MEMADDR(machine,
                                   BE2HO_4(endaddr - sizeof(TagList))), &tags,
                     sizeof(TagList)) == NULL) {
              retval = errno;
              Sim_Warning("Bad tags copy in LoadNgo\n");
              Sim_Warning("vaddr = 0x%08x\n", BE2HO_4(endaddr - sizeof(TagList)));
              perror(pathname);
              return;
          }
      }
#endif
      
      if (retval != 0) {
          perror(pathname);
          return;
      }
      
      /* Must copy argc / argv to simulated memory so that kernel
       * can read the arguments. 
       */
      copy_bootline_args(machine, argc, argv, &argcVA, &argvVA);
      
      /* CPUWarning("Boot args copied to 0x%08x 0x%08x\n", argcVA, argvVA); */
     
#ifdef notdef 
      /* This deterministically enters detailed mode so that every */
      /* instruction is emulated */

      SBase[firstCPU].R[4] = argc;
      SBase[firstCPU].R[5] = argvVA;
      SBase[firstCPU].R[6] = 0;
#endif
      firstCPU = FIRST_CPU(machine);
      HackUpLinuxStartupEnv(firstCPU,entry);
      
      CPUWarning( "Load image machine=%d pc=%llx\n", machine,
                  (uint64)SBase[firstCPU].eip);
  
      
  }
  
  SimulatorEnter(simosCPUType, 0, 0);
  ASSERT(0);
}



/*
 * Launch slave cpu :  can get here through two different paths:
 *    from simmagic.c:MAGIC_PPC_OP_STARTSLAVENODE  for setting the
 *               start address of a slave cpu
 *    from tcl_init.c:hive launchslave  for passing arguments to
 *               a cell when booting it
 *
 * Returns:
 *  0 - success
 *  1 - failure, this CPU has already been launched.
 */

int
LaunchSlave(int cpu, VA pc,
	    int a0, int a1,int a2, int a3)
{
   ASSERT(0 <= cpu && cpu < TOTAL_CPUS);

   if (SBase[cpu].cpuStatus == cpu_running ||
       SBase[cpu].outOfSlaveLoop != 0)
     return 1; /* already running */

   /*  Sim_Warning("LaunchSlave: cpu %d PC 0x%08x" */
   /* 	       "a0 0x%08x a1 0x%08x a2 0x%08x a3 0x%08x\n", */
   /* 	       cpu, pc, a0, a1, a2, a3); */

   SBase[cpu].cpuStatus      = cpu_running;    /* mipsy polls this each cycle */
#ifdef notdef
   SBase[cpu].PC             = pc;
#endif
   SBase[cpu].outOfSlaveLoop = 1;              /* embra polls this each context switch */

   return 0;
}

/*
 * Main loop for master processor startup
 */
void
SimPromEnter(void)
{
  char line[MAXBOOTARGBYTES];
  int  retval;
  int  i;
  char prompt[64];
  static int prompt_no = 1;
  int console = 0;
  char usage[]    = "\n\rBoot PROM commands\n\r"
                    "------------------\n\r"
                    "b pathname - Boot the kernel\n\r"
                    "q - quit SimOS\n\r"
                    "h - Halt the machine\n\r"
                    "d - Enter the debugger\n\r"
                    "? - This help listing\r\n\n";

  /* indicate that this cpu has launched for other people to check */
  for (i = 0; i < NUM_MACHINES; i++) {
      sim_misc.launchAddr[FIRST_CPU(i)] = (void *)SimPromEnter; /* BOGUS */
  }

  while (1) {
      sprintf(prompt, "\rsimosboot (%d)> ", prompt_no++);
      sim_console_write(console, prompt, strlen(prompt));
      
      for (i = 0; i < MAXBOOTARGBYTES; i++) {
         /*
           * Expect hooks
           *  - expect characters take priority over user typed chars
           */
          if ((line[i] = ExpectGetChar(console))) {
              sim_console_write(console, &line[i], 1);
          } else {
              line[i] = retval = sim_console_in(console); /* hmmm */
              if (retval == -1) {
#ifdef sgi
                  sginap(0);
#elif sun
                  sleep(1);
#endif
                  i--;
                  continue;
              }
              sim_console_write(console, &line[i], 1);
          }
          if (line[i] == '\n' || line[i] == '\r') {
              i++;
              sim_console_write(console, "\r\n", 2);
              break;
          }
          if (line[i] == 'C'-'@') {
              sim_console_write(console, "\r\n", 2);
              exit(0);
          }
      }
      
      line[i] = '\000';
      
      switch (line[0]) {
      case 'b':
      case 'B': {    /* B pathname  - Boot the specified kernel */
          char *arg, *file;
          char* argv[MAXBOOTARGS+1]; /* list of arguments */
          int   argc;                /* argcount (for kernel) */
          
          for (arg = line + 1; *arg == ' '; arg++) ; /* skip spaces */
          if (!*arg) { 
              sim_console_write(console, "Must specify kernel file\n", 25);
              break;
          } else {
              file = arg;
              while (*arg && !isspace(*arg)) arg++;
              argc = 0;
              if (*arg) {
                  *arg = 0; /* 0-terminate file name */
                  arg++;
                  for(argc = 0; argc < MAXBOOTARGS; argc++) {
                      for ( ; isspace(*arg); arg++) ; /* skip spaces */
                      if (!*arg) break;
                      argv[argc] = arg;
                      while (*arg && !isspace(*arg)) arg++;
                      if (*arg) { *arg = 0; arg++; }
                  }
              }
          }
          
          LoadNgo(file, argc, argv);
          
          break;
      }
      
      case 'h':
      case 'H': {    /* H - Halt the simulated machine. */
          SimHalt();
          break;
      }
      case 'd':
      case 'D': {
          kill((int)getpid(), SIGUSR1);
          break;
      }
      case '?': {
          sim_console_write(console, usage, sizeof(usage)-1);
          break;
      }
      case 'q':
      case 'Q':
          sim_console_write(console, "\r\n", 2);
          exit(0);
          break;
      default:
          sim_console_write(console, "Unknown command\n", 16);
      }
  }
}
#ifdef notdef
/* NOTE:  for flashlite, cpuNum is the destination node,
   not the node originating the PIO */
Result
SimMagic_DoPIO(int cpuNum, uint addr, int isRead, int size, void *data)
{
   CPUError("Bad PIO access\n");
   return FAILURE;
}

void
CycleAdvance (SimTime * currentTime) {
  SimTime newTime = ++*currentTime;

  if (memsysVec.type == CHAL2)
    Chal2CycleAdvance ();

  EventPollSingleQueue (newTime);
} /* CycleAdvance () */
#endif


void
SimHalt(void)
{
  exit(0);
}

#define LINUX_KERNEL_CS  0x10
#define LINUX_KERNEL_DS  0x18

/* from linux/arch/i386/kernel/head.S */
#define CL_MAGIC_ADDR   0x90020
#define CL_MAGIC        0xA33F
#define CL_BASE_ADDR    0x90000
#define CL_OFFSET       0x90022

#define INIT_GDA_ADDR        0x2000
static uint64 init_gda[] = {
        0x0LL,   /*  Entry zero not used */
        0x0LL,   /*  Entry one not used */
       0xcdda000000ffffLL,
                  /* Entry two - Kernel code segment. 
                     limit = 0xdffff ;  base = 0; type = 0xa; S=1, DPL=2, present=1,
                     AVL=0, DB=1, gran=1  */
       0xcdd2000000ffffLL,
                  /* Entry three - Kernel data segment. 
                     limit = 0xdffff ;  base = 0; type = 0x2; S=1, DPL=2, present=1,
                     AVL=0, DB=1, gran=1  */
}; 



static void 
HackUpLinuxStartupEnv(int cpuNum, VA entry)
{
   CPUState *P = SBase + cpuNum;
   int numKB, offset;
   char *cmdline;
   uint8 byte;
   IODevices *iod = GetIODevices(cpuNum);

    P->eip = entry; 

   P->eax = P->ebx = P->ecx = P->edx = P->esi = P->edi = P->ebp = 0;

   /*
    * set stack at 0x94000
    */
   P->esp = 0x94000-12;
     
   P->cs.selector = LINUX_KERNEL_CS;
   P->cs.descriptor = init_gda[LINUX_KERNEL_CS>>3];
   P->ss.selector = LINUX_KERNEL_DS;
   P->ss.descriptor = init_gda[LINUX_KERNEL_DS>>3];

   P->ds = P->es = P->fs = P->gs = P->ss;

   P->cr[0] = 0x1;  /* in Protected mode */
   P->cr[1] =  P->cr[2] =  P->cr[3] = P->cr[4] = 0;

   P->gdtr = (INIT_GDA_ADDR<<16)|(sizeof(init_gda)-1);
   P->idtr = 0;
   P->ldtr.selector = P->tr.selector = 0;
   P->ldtr.descriptor = P->tr.descriptor = 0;

   /*
    * Initialize the GDA memory
    */
   bcopy((char *) init_gda, PHYS_TO_MEMADDR(0, INIT_GDA_ADDR), 
         sizeof(init_gda));
   /*
    * parameters at 0x90000 (no segments setup yet)
    */
   numKB = MEM_SIZE(0)/1024;
   /* extended memory in K */
   if (numKB  > 0x400) {
      *(uint16 *) PHYS_TO_MEMADDR(0, 0x90002) = (numKB - 0x400);
   }

   /* video cols */
   *(char *) PHYS_TO_MEMADDR(0, 0x90008) = 80;
   /* video rows */
   *(char *) PHYS_TO_MEMADDR(0, 0x9000e) = 24;

   /* command line */
   *(uint16 *) PHYS_TO_MEMADDR(0, CL_MAGIC_ADDR) = CL_MAGIC;
   offset = 2048;  /* XXX ? */
   *(uint16 *) PHYS_TO_MEMADDR(0, CL_OFFSET) = offset;
   cmdline = "root=/dev/sda0";
   bcopy(cmdline,  PHYS_TO_MEMADDR(0,CL_BASE_ADDR + offset), strlen(cmdline));

   
   /*
    *  Initialize the interrupt controllers.
    *
    *  This initialization is done by setup.S in 16-bit mode.
    *  Do it here for now.
    */
   byte = 0x11;
   IOSpace_InOut(iod, 0x20, 1, PORT_OUT, &byte);  /* initialization sequence */
   IOSpace_InOut(iod, 0xA0, 1, PORT_OUT, &byte);

   byte = 0x20;
   IOSpace_InOut(iod, 0x21, 1, PORT_OUT, &byte);  /* start of hardware int's (0x20) */
   byte = 0x28;
   IOSpace_InOut(iod, 0xA1, 1, PORT_OUT, &byte);  /* start of hardware int's 2 (0x28) */

   byte = 0x04;
   IOSpace_InOut(iod, 0x21, 1, PORT_OUT, &byte);  /* 8259-1 is master */
   byte = 0x02;
   IOSpace_InOut(iod, 0xA1, 1, PORT_OUT, &byte);  /* 8259-2 is slave */

   byte = 0x01;
   IOSpace_InOut(iod, 0x21, 1, PORT_OUT, &byte);  /* 8086 mode for both */
   IOSpace_InOut(iod, 0xA1, 1, PORT_OUT, &byte);

   byte = 0xFF;
   IOSpace_InOut(iod, 0xA1, 1, PORT_OUT, &byte);  /* mask off all interrupts for now */
   byte = 0xFB;
   IOSpace_InOut(iod, 0x21, 1, PORT_OUT, &byte);  /* but irq2 which is cascaded */


   /*
    *  XXX linux assumes io addr 0x80 is a dummy port.
    *  undocumented PC says it is part of the XT DMA.
    *  anyway null it out for now
    */
   IOSpace_Register(iod, 0x80, NULL, NULL);

   /*
    *  XXX linux writes to the Diskette Controller Digital Output register
    *  to disable the diskette system.
    */
   IOSpace_Register(iod, 0x3f2, NULL, NULL);

   HackUpInstallTimer(0, 10000, 10000);  /* Set fake clock to 10ms period */
}


/****************************************************************************
 *
 * Fake interrupt clock  support
 *
 ****************************************************************************/


/* eventcallback hdr for clock timer */
static EventCallbackHdr timerHdr[SIM_MAXCPUS];
/* clock interrupt counter for stats */
static uint             clockIntrs;
/* maintain last time timer was invoked so that
   we can checkpoint the correct timeleft. */
static SimTime          lastTimeout[SIM_MAXCPUS];


/* timer callback: raise interrupt and enqueue yourself for next tick */
static void
TimerCallback(int cpuNum, EventCallbackHdr *hdr, void *arg)
{
  
  lastTimeout[cpuNum] = CPUVec.CycleCount(cpuNum);
  clockIntrs++;

  Intr_Raise(GetIODevices(cpuNum), 0);
    
  EventDoCallback(cpuNum, TimerCallback, hdr, NULL,
		  CPUVec.clockInterval[cpuNum] * CPU_CLOCK);
}

/* sets up the callback data and the first callback */
static void
HackUpInstallTimer(int cpuNum, unsigned int interval, unsigned int timeLeft)
{

  CPUVec.clockInterval[cpuNum] = interval;
  
  if (CPUVec.clockStarted[cpuNum]) EventCallbackRemove(&timerHdr[cpuNum]);
  CPUVec.clockStarted[cpuNum] = 1;

  EventDoCallback(cpuNum, TimerCallback, &timerHdr[cpuNum], NULL, 
		  timeLeft * CPU_CLOCK);

#ifdef notdef
  SBase[cpuNum].clockInterval = CPUVec.clockInterval[cpuNum];
  SBase[cpuNum].clockStarted = CPUVec.clockStarted[cpuNum];
#endif
}

