/*
 * Copyright (C) 1996 by the Board of Trustees
 *    of Leland Stanford Junior University.
 * 
 * This file is part of the SimOS distribution. 
 * See LICENSE file for terms of the license. 
 *
 */

/*****************************************************************
 * simmisc.h
 *
 * Plethora of misc. definitions which don't fit anywhere else.
 * Dan Teodosiu, 07/96
 *
 ****************************************************************/

#if !defined(_SIM_MISC_H) && defined(SIM_X86)
#define _SIM_MISC_H

#include "syslimits.h"
#include "simtypes.h"

#include "file_formats.h"

/*
 * BASESIMMEMADDR definitions
 * This is the starting address of the simulated memory for the first machine.
 * Each other machine's memory starts after the end of the previous one.
 * Use the SIM_MEM_ADDR(MachNo) to get the start address for a particular machine.
 */
#if defined(sgi) || defined(sun) || defined(i386)

# define BASESIMMEMADDR ((MA)0x18000000)

#elif defined(linux)

/* Linux */
# define BASESIMMEMADDR ((MA)0x50000000)

#elif defined(__alpha)

/* Alpha */
# define BASESIMMEMADDR ((MA)0x200000000)

#endif 

#ifndef LANGUAGE_ASSEMBLY

/*
 * Macro for converting a KSEG0 address into an access to the
 * simulated memory.   
 */
#define K0_TO_MEMADDR(_m, _x) ((MA)((SIM_MEM_ADDR(_m) - SIMCP0_KSEG0_ADDR) \
                                    + (VA)(_x)))
#define MEMADDR_TO_K0(_m, _x)  ((VA)(((MA)(_x) - SIM_MEM_ADDR(_m)) \
                                     + SIMCP0_KSEG0_ADDR))

/*
 * Macro to test whether Flashlite is being used.
 */
#define USE_MAGIC()    (memsysVec.type == FLASHLITE)

/*
 * Miscellaneous definitions
 */
#define MAXBOOTARGS     32    /* max. args to kernel */
#define KNAME_MAX       64    /* max. kernel name length */
#define MAXBOOTARGBYTES 256   /* max. length of boot command line */

/* This is probably not needed any longer and should be removed */
#define SLAVELOOP_CONTINUE  0x4015

/* offset (starting from the beginning of a cell's memory range) of
 * page which is used for passing bootline arguments to the kernel.
 * NOTE: currently, kernels are compiled so that they leave 5 free
 * pages at the beginning of the cell's memory range. The 5 pages are
 * used as follows:
 *   - page 0:   remap area
 *   - page 1:   OSPC area (fake for 32-bit SimOS)
 *   - page 2:   bootline arguments
 *   - page 3,4: unused
 */
#define ARGZONE_OFFS        0x00002000 /* third page */

typedef struct Header {
union header {
  FILHDR coffHeader;
  Elf32_Ehdr elfHeader;
  Elf64_Ehdr elf64Header;
} hb;

} HeaderBuf;


/*
 * This data structure holds several not necessarily related pieces
 * of information. The reasons for this being a single data structure
 * are historical.
 */

typedef struct SimMisc {

  CPUType   enterThisCPU; /* notify the simulator to switch to this CPU type */
  CPUType   myCPUType;
  

  void*     launchAddr[SIM_MAXCPUS];   /* launch address for slaves */
  int       launchArg[SIM_MAXCPUS][4]; /* launch args (A0-A3) for slaves */

  char*     fprom[SIM_MAXCPUS];        /* fprom */
  char*     fram[SIM_MAXCPUS];         /* fram */

  char*     incoherentPages; /* array indicating incoherent pages
                              -- lost in low-level recovery */
} SimMisc;

extern bool CPUlaunched[SIM_MAXCPUS];
extern SimMisc sim_misc;

extern void SimMiscInit(void);
extern void SimPromEnter(void);
extern void SimulatorEnter(CPUType simulator, int restoreCkpt, int swtch);
extern void SimulatorSwitch(CPUType oldSimulator, CPUType newSimulator);

typedef void (*aout_entryfunc)(int, char**, char**);
typedef void (*bdload_sectionfilter)(char* startaddr, unsigned nbytes);

#ifdef notdef
extern int  LoadImage(char* pathname, int machine, VA startaddr,
                      int nbytes, bool *elfFormat,
		      VA* entry, bdload_sectionfilter filter);
#endif
extern int  LaunchSlave(int cpu, VA pc, int a0, int a1, int a2, int a3);

extern int  ResetCPUs(int firstcpu, int lastcpu);

extern void SimHalt(void);



#endif /* !LANGUAGE_ASSEMBLY */
#endif /* _SIM_MISC_H */
