/*
 * Copyright (C) 1996 by the Board of Trustees
 *    of Leland Stanford Junior University.
 * 
 * This file is part of the SimOS distribution. 
 * See LICENSE file for terms of the license. 
 *
 */

/*****************************************************************
 * startup.c
 * 
 * This is the entry point into simos. 
 *
 * $Author: nikos $
 * $Date: 1998/04/29 04:54:44 $
 *****************************************************************/

#include "sim.h"
#define _BSD_SIGNALS 
#include <sys/signal.h>
#ifndef i386
# include <sys/sysinfo.h>
#endif
#include <sys/times.h>
#include <sys/types.h>
#include <sys/errno.h>
#include <sys/termios.h>
#include <sys/time.h>
#include <sys/file.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <netinet/tcp.h>
#include <sys/mman.h>
#include <unistd.h>
#include <fcntl.h>
#include <stdio.h>
#include <setjmp.h>
#include <string.h>
#include <memory.h>
#include <ctype.h>
#include <math.h>
#include <stdlib.h>
#ifdef sgi
# include <sys/sysmp.h>
#endif
#include <sys/wait.h>
#include <netdb.h>
#include "startup.h"
#include "checkpoint.h"
#include "machine_params.h"
#include "sim_error.h"
#include "registry.h"
#include "simstats.h"
#include "annotations.h"
#include "cpu_interface.h"
#include "hd.h"
#include "ethernet.h"
#include "console.h"
#include "eventcallback.h"
#include "dma.h"
#include "../../devices/disk/simos_interface.h"
#include "simutil.h"
#include "hw_events.h"
#include "simtypes.h"
#include "remote_access.h"
#include "gdb_interface.h"
#include "tcl_init.h"
#include "params.h"
#ifdef HWBCOPY
# include "hw_bcopy.h"
#endif
#include "simmisc.h"
#include "iodevs.h"

#define SIMOS_VERSION "3.1"

#ifdef __alpha
 char *physMem = SIM_ALPHA_MEMORY_START;
#endif


/********************
 * GLOBAL VARIABLES *
 ********************/
CPUState* SBase;
CPUType   simosCPUType = NO_CPU;      /* Global notion of the current simulator */
int       restoringCpt = 0;
int       hostNumProcs = 1;
int hostCPUIdleStats[SIM_MAXHOSTCPUS]; /* %idle for each processor */
int       inCellMode = 0;              /* True if using cells (implies 1 machine) */

/**********************
 * EXTERNAL FUNCTIONS *
 **********************/
extern void TCLInit(char *initFile);
extern void InitDevices(void);

/**********************
 * EXTERNAL VARIABLES *
 **********************/
extern char **environ;
extern CptVersion cptVersion;

/********************
 * STATIC VARIABLES *
 ********************/
static int mfd;



/********************
 * STATIC FUNCTIONS *
 ********************/
static void Sigusr(int);

/* Physical memory file checkpoint */
static CptCallback MemoryCheckpointCB;
/* Execution state checkpoint */
static CptCallback ExecStateCheckpointCB;
/* Prom state checkpoint */
static CptCallback PromCheckpointCB;

/* option processing */

struct {
   int simCheckpointRestore;    /* -r selected */
   char* checkpoint_master;     /* argument to -r */
   CPUType startcpu;            /* -m, -c, -p selected */
   char* param_file;            /* File with parameter values */
   char* log_file;              /* -l option: log file name */
   char* debug_flags;           /* string passed to DebugInit */
} optvals;

/*****************************************************************
 *
 *****************************************************************/
static void
ParseOpts(int argc, char** argv)
{
   int c;
   extern char* optarg;
   extern int optind;
   
   bzero((char*)&optvals, sizeof(optvals));

   optvals.startcpu   = NO_CPU;
   optvals.param_file = "init.simos";
   optvals.log_file   = "cpu.log";
   optvals.debug_flags = "";
   optvals.simCheckpointRestore = FALSE;

   while ((c = getopt(argc, argv, "z:r:o:l:d:")) != EOF) {
       switch (c) {
       case 'z': 
	   optvals.param_file = optarg;
	   break;
       case 'r':
	   optvals.simCheckpointRestore = TRUE;
	   optvals.checkpoint_master = optarg;
	   break;
       case 'o':
           if (!argv[optind - 1]) {
	     fprintf(stderr, "\nmissing value for option: %s\n", optarg);
	     exit(1);
	   } else if ((optind < argc) && (argv[optind][0] != '-')) {
              TCLDefineOption(optarg, argv[optind]);
              optind++;
           } else {
              TCLDefineOption(optarg, "1");
           }
           break;
       case 'l':
	   optvals.log_file = optarg;
	   break;
       case 'd':
           if (!argv[optind - 1]) {
              fprintf(stderr, "SimOS: Turning all debugging flags on\n");
              optvals.debug_flags = "+";
           } else {
              optvals.debug_flags = optarg;
           }
	   break;
       case '?':
	   fprintf(stderr,
	     "\nusage: simos [-r checkpoint_masterlog] [-d <debugflags>]\n"
	     "   [-z param_file_name] [-o tclopt [val]] [-l log_file_name]\n");
           exit(1);
       }
   }
}

static SimTime CycleZero(int cpuNum) { return 0; }

/*****************************************************************
 * MAIN ENTRY POINT INTO SIMOS
 *****************************************************************/
int
main(int argc, char *argv[], char **env)
{
  char *addr;
  int i, machNo;
    
  /* Clear the machine structure */
  bzero (&machines, sizeof(MachinesStruct));
   
  ParseOpts(argc, argv);

  /* Enable debug flags */
  DebugInit(optvals.debug_flags);

  /* init log file before printing anything */
  SimErrorInit(optvals.log_file);
   
  Sim_Warning("===== SimOS Version %s =====\n", SIMOS_VERSION );


  InitDevices(); 
  restoringCpt = optvals.simCheckpointRestore;
  
  Simcpt_Init();

  Simcpt_Register("prom", PromCheckpointCB, ALL_CPUS);
  if (restoringCpt) {
     Simcpt_Checkpoint(CPT_RESTORE, optvals.checkpoint_master);
     Simcpt_Restore("prom");
     if (cptVersion.ver == 3) {
        /* XXX
	 * Note (DT): this should go!
	 */
        int m, n;
        /* Read the number of cells and ethernet ctrls (if cpt ver is 3) */
        /* If ver >= 4, then they will already have been read in the prom */
        Simcpt_Preread_Numcells();
        Simcpt_Preread_Numether();
        /* This hack is not required in version 4 cpts, since the num of disk
            controllers and clocks are saved */
        if (NUM_CELLS(0) > 1) {
	  NUM_CLOCKS(0) = NUM_CELLS(0);
	  for (n = 0; n < SIM_MAXCPUS; n++) NUM_DISK_CONTROLLERS(0,n) = 1;
        } else {
	  NUM_CLOCKS(0) = MOOSE_MAXVMS; /* XXX brain damage, REMOVE */
	  for (m = 0 ; m < MAX_MACHINES; m++)
	    for (n = 0; n < SIM_MAXCPUS; n++)
	      NUM_DISK_CONTROLLERS(m,n) = 1;
	}
     }
     
     /* let param manager know which of the parameters have been initialized
      * so it doesn't error the user for forgetting to init them
      */
     ParamGrabbedAtStartup("PARAM(MACHINE.Count)");
     ParamGrabbedAtStartup("PARAM(CPU.Count)");
     ParamGrabbedAtStartup("PARAM(HIVE.NumCells)");
     ParamGrabbedAtStartup("PARAM(MEMSYS.MemSize)");
     ParamGrabbedAtStartup("PARAM(CONSOLE.Count)");
     ParamGrabbedAtStartup("PARAM(ETHERNET.Count)");
     ParamGrabbedAtStartup("PARAM(CLOCK.Count)");
     ParamGrabbedAtStartup("PARAM(DISK.NumControllers)");
  }

  InitCPUModel(optvals.startcpu);

  TCLInit(optvals.param_file);

  if (simosCPUType == NO_CPU) {
     CPUError("No cpu model specified.\n");
  }
  
  CPUPrint("Invocation: ");
  for (i=0; i<argc; i++) CPUPrint("%s ", argv[i]);
  CPUPrint("\n");
  
  RegistryInit();
  
  /* Initialize the base memory address for each machine
   */
  SIM_MEM_ADDR(0) = BASESIMMEMADDR;
  for (machNo = 1; machNo < NUM_MACHINES; machNo++) {
      SIM_MEM_ADDR(machNo) =
          SIM_MEM_ADDR(machNo-1) + MEM_SIZE(machNo-1);
  }
      
  Simcpt_Register("memory", MemoryCheckpointCB, ALL_CPUS);
  Simcpt_UseCompression("memory", GZIP, BIN_FILE_ONLY);

  /* Remap Vector. Not exported to kernel, but in shared memory */
  /* XXX is this still needed? Don;t think so! XXX */
  remapVec = ZMALLOC(sizeof(RemapVector),"RemapVector");

  /* Initialize some values in sim_misc */
  SimMiscInit();

  /* Initialize MAGIC and devices. */
  /* sim_magic_init(restoringCpt); */

  
  /* 
   * Various  initializations. These all generally need to be done 
   * when the parameters (known when the checkpoint state is
   * restored) are known.  
   */
  
  EventCallbackInit(TOTAL_CPUS);

  /* Setup CycleCount so InitDevices can register events */
  CPUVec.CycleCount = CycleZero;

  InstallPoller();

  sim_disk_init(1,8,0,restoringCpt);

  DMAInit();
  
  /*
   * Configure disk model. By default we use an accurate model of a SCSI disk.
   * To allow some flexiblity we have the ability to scale the speed of the 
   * disk (simConfigDiskScaling is a number in precent used for scaling, 100%
   * is "real" disk speed. A scaling of 0 or less means model a fixed latency
   * disk using simConfigAsyncDiskDelay as the fixed delay in milliseconds. 
   */
  if (!strcmp(DISK_MODEL(0), "HP")) {
     /* Initialize the DISK model we are compiled for */
#ifdef notdef
     DiskModelInit(SIM_MAX_DISKS, ((double)HP_DISK_SCALING(0))/100.0);
#endif
     FIXED_DISK_DELAY(0) = 0;
     strcpy(DISK_MODEL(0), "Fixed"); 
     Sim_Warning("x86 currently only works with zero-latency disk model\n");
     Sim_Warning("Setting disk model to zero-latency\n");
  } else {
     if (strcmp(DISK_MODEL(0), "Fixed")) {
        CPUWarning("Bad disk model: %s\n", DISK_MODEL(0));
        ASSERT(0);
     }
     if (FIXED_DISK_DELAY(0)) {
        Sim_Warning("x86 currently only works with zero-latency disk model\n");
        Sim_Warning("Setting disk model to zero-latency\n");
     }
     FIXED_DISK_DELAY(0) = 0;
  }

  /* Create the physical memory for the machine. 
   * Initialize the value of memory to zero. 
   * If we are restoring from a checkpoint we do this anyway
   * to make the restore symmetric with the save. 
   */
  if (restoringCpt) {
     if (Simcpt_CanMmapMemoryCheckpoint()) {
        mfd = Simcpt_GetMemoryCptFile();
        if( mfd < 0 ) {
           perror("Open MemFile failed");
           exit(1);
        } 
     } else {
        /* Restore into a local file. */
        mfd = MakeFile(MemFileDir, NULL, 0,1);
        Simcpt_Restore("memory");
     }
     if (mfd < 0) {
        Sim_Error("startup: Can't create memory file\n");
     }  
  } else {
     /* This maps K0 out of /dev/zero which makes */
     /* simos startup and termination much faster */
#ifdef __alpha
     mfd = -1;
#else
     mfd = open("/dev/zero", O_RDWR, 0);
     if (mfd < 0) {
        Sim_Error("startup: Can't create memory file\n");
     }
#endif  
  } 
  
  /*
   * Perform the mmap for each machine (out of the same memory file)
   */
  for (machNo = 0; machNo < NUM_MACHINES; machNo++) {
      SIM_DEBUG(('g', "size = %d\n", MEM_SIZE(machNo)));

#ifdef sgi
      addr = (char *) mmap(SIM_MEM_ADDR(machNo), 
                           MEM_SIZE(machNo) & ~(SIM_PAGE_SIZE-1),
                           PROT_READ|PROT_WRITE|PROT_EXEC, 
                           MAP_PRIVATE|MAP_AUTORESRV,
                           mfd, SIM_MEM_ADDR(machNo)-SIM_MEM_ADDR(0));
      if (addr != SIM_MEM_ADDR(machNo)) {
          perror("mmap of KSEG0 (sgi)");
          exit(1);
      }
#endif
#ifdef sun
      SIM_DEBUG(('g', "map to %#x\n", SIM_MEM_ADDR(machNo)));

      addr = (char *) mmap(SIM_MEM_ADDR(machNo),
                           MEM_SIZE(machNo) & ~(SIM_PAGE_SIZE-1),
                           PROT_READ|PROT_WRITE|PROT_EXEC, 
                           MAP_FIXED|MAP_PRIVATE,
                           mfd, SIM_MEM_ADDR(machNo)-SIM_MEM_ADDR(0));
      if (addr != SIM_MEM_ADDR(machNo)) {

          fprintf (stderr,
      "addr=0x%08x, len=0x%08x, prot=0%o, flags=0x%x, fildes=%u, off=0x%08x\n", 
       (unsigned)(SIM_MEM_ADDR(machNo)),
       (unsigned)(MEM_SIZE(machNo) & ~(SIM_PAGE_SIZE-1)),
       (PROT_READ|PROT_WRITE|PROT_EXEC), (MAP_FIXED|MAP_PRIVATE), mfd,
       (unsigned)(SIM_MEM_ADDR(machNo)-SIM_MEM_ADDR(0)));
       
          perror("mmap of KSEG0");
          exit(1);
      }
#endif
#ifdef __alpha
      addr = (char *) mmap(SIM_MEM_ADDR(machNo),
                           MEM_SIZE(machNo),
                           PROT_READ|PROT_WRITE, 
                           MAP_FIXED|MAP_PRIVATE|
                             (mfd>0?MAP_FILE:MAP_ANONYMOUS),
                           mfd, SIM_MEM_ADDR(machNo)-SIM_MEM_ADDR(0));
      CPUWarning("physicalMemory set at 0x%p memSize=%d  \n",
                 addr,MEM_SIZE(machNo));
      if (addr == (void*)-1) { 
          perror("mmap of physical memory (alpha): ");
          exit(1);
      }
#endif

#ifdef linux
      addr = (char *) mmap(SIM_MEM_ADDR(machNo),
                           MEM_SIZE(machNo),
                           PROT_READ|PROT_WRITE,
                           MAP_PRIVATE,
                           mfd, SIM_MEM_ADDR(machNo)-SIM_MEM_ADDR(0));
      if (addr != SIM_MEM_ADDR(machNo)) {
          perror("mmap of KSEG0 (linux)");
          exit(1);
      }
#endif

#ifdef UWARE_HOST
      addr = (char *) mmap(SIM_MEM_ADDR(machNo),
                           MEM_SIZE(machNo),
                           PROT_READ|PROT_WRITE,
                           MAP_FIXED|MAP_PRIVATE,
                           mfd, SIM_MEM_ADDR(machNo)-SIM_MEM_ADDR(0));
      if (addr != SIM_MEM_ADDR(machNo)) {
          perror("mmap of KSEG0 (unixware)");
          exit(1);
      }
#endif /* def UWARE_HOST */
  }

  /*
   * ZeroMem initialization. 
   * InitZeroMem simply declares the address range from
   * which memory can be allocated and zeroed, potentially
   * using an mmap of /dev/null.
   */
#ifdef __alpha
  InitZeroMem(PHYS_TO_MEMADDR(NUM_MACHINES-1, MEM_SIZE(NUM_MACHINES-1)), 
              (char *) 0x300000000,0);
#else
  InitZeroMem(PHYS_TO_MEMADDR(NUM_MACHINES-1, MEM_SIZE(NUM_MACHINES-1)), 
              (char*)UINT_TO_PTRSIZE(0x70000000), 1 /* yes, redzone the next page */);
#endif
  
  /*
   * Initalize memory and interrupt handling.
   */
  
  SBase = (CPUState *) malloc(sizeof(CPUState)*(TOTAL_CPUS+1)); /* ??? why +1 */
  bzero((char*)SBase,sizeof(CPUState)*(TOTAL_CPUS+1));
  for (i = 0; i < TOTAL_CPUS; i++) {
    SBase[i].memoryPtr   = SIM_MEM_ADDR(M_FROM_CPU(i));
    SBase[i].memSize     = MEM_SIZE(M_FROM_CPU(i));
    /* TODO Invalidate TLBs */
  }

  /*
   * Startup the simulation of the MMU and exception handlers.
   */ 
  Simcpt_Register("execstate", ExecStateCheckpointCB, ALL_CPUS);

  if (restoringCpt)
    Simcpt_Restore("execstate");
  
  Simdebug_init();
  if (signal(SIGUSR1, Sigusr) == SIG_ERR)
    perror("signal(SIGUSR1)");

  RegistryDumpEntries ();

  if (restoringCpt) SimulatorEnter(simosCPUType, 1, 0); /* enter simulator */
  else              SimPromEnter();                     /* enter "bootprom" */

  exit (0);
  return 0;           /* Shut up the compiler */
}



/*
 * Called on a SIGUSR1 signal. Used to enter tcl, etc. 
 */
static void 
Sigusr(int sig)
{
   if (signal(SIGUSR1, Sigusr) == SIG_ERR) perror("signal(SIGUSR1)");

   if (CPUVec.Handle_Debug_Signal) {
      CPUVec.Handle_Debug_Signal(0, (sig == SIGUSR1));
   }
   return;
}


/***************************************************************************
 *
 * Checkpointing routines.
 *
 ***************************************************************************/

static int
MemoryCheckpointCB(CptDescriptor *cptd)
{
   int machNo;

   if (cptVersion.ver == 4 && cptVersion.sub == 0) {
      for (machNo = 0; machNo < NUM_MACHINES; machNo++) {
         Simcpt_CptBlock(cptd, "Memory", machNo, NO_INDEX, 
                         (caddr_t) SIM_MEM_ADDR(machNo),
                         MEM_SIZE(machNo), mfd );
      }
   } else if (cptVersion.ver == 3) {
      Simcpt_CptBlock(cptd, "Memory", NO_INDEX, NO_INDEX, 
                      (caddr_t) SIM_MEM_ADDR(0),
                      MEM_SIZE(0), mfd );
   } else {
      ASSERT(0);
   }
   return 0;
}

static int
ExecStateCheckpointCB(CptDescriptor *cptd)
{
  int cpu;

  /* For checkpoint ver. 3 compatibility */
  int memSize = MEM_SIZE(0);
  int cpncpu = NUM_CPUS(0);

  if (cptd->mode == CPT_SAVE) {
     /* TimerUpdateTimeLeft(); */
  }

  if (cptVersion.ver == 3) {
     Simcpt_CptInt(cptd, "MemSize", NO_INDEX, NO_INDEX, &memSize);
     ASSERT(memSize == MEM_SIZE(0));
     Simcpt_CptInt(cptd, "NumCPUs", NO_INDEX, NO_INDEX, &cpncpu);
     ASSERT(cpncpu == NUM_CPUS(0));
  }
  
  for( cpu = 0; cpu < TOTAL_CPUS; cpu++ ) {
    /* TODO checkpoint CPU */
    /* TODO checkpoint TLBs */
  }
  return 0;
}


static int
PromCheckpointCB(CptDescriptor *cptd)
{
   int machNo;

   /*
    * Establish or save the checkpoint version
    */
   
   if (cptd->mode == CPT_SAVE) {
     /* Set the current version to the desired save version */
      cptVersion.ver = cptVersion.saveVer;
      cptVersion.sub = cptVersion.saveSub;
   } 
   Simcpt_CptInt(cptd, "Version", NO_INDEX, NO_INDEX,
                 &(cptVersion.ver));
   if (cptVersion.ver >= 4) {
      Simcpt_CptParamInt(cptd, "SubVersion", NO_INDEX, NO_INDEX,
                         &(cptVersion.sub));
   } else {
      cptVersion.sub = 0;
   }
      
   if (cptd->mode == CPT_SAVE) {
      CPUPrint("CPT: Saving (Format is version %d, sub-version %d)\n",
               cptVersion.ver, cptVersion.sub);
   } else {
      CPUPrint("CPT: Restoring (Format is version %d, sub-version %d)\n",
               cptVersion.ver, cptVersion.sub);
   }
      
   /*
    * Do the checkpointing
    */
   
   if (cptVersion.ver == 4 && cptVersion.sub == 0) {

      Simcpt_CptParamInt(cptd, "NumMachines", NO_INDEX, NO_INDEX,
                         &(NUM_MACHINES));
      
      for (machNo = 0; machNo < NUM_MACHINES; machNo++) {
         Simcpt_CptParamInt(cptd, "NumCPUs", machNo, NO_INDEX,
                            &(NUM_CPUS(machNo)));
         Simcpt_CptParamInt(cptd, "NumCells", machNo, NO_INDEX,
                            &(NUM_CELLS(machNo)));
         Simcpt_CptParamInt(cptd, "MemSize", machNo, NO_INDEX,
                            &(MEM_SIZE(machNo)));
         Simcpt_CptParamInt(cptd, "NumConsoles", machNo, NO_INDEX,
                            &(NUM_CONSOLES(machNo)));
         Simcpt_CptParamInt(cptd, "NumEtherControllers", machNo, NO_INDEX,
                            &(NUM_ETHER_CONTROLLERS(machNo)));
	 {
	   /* NOTE on compatibility:
	    * Former checkpointing code save a total number of disk
	    * controllers per machine. To preserve compatibility with
	    * older checkpoints, we now save the following:
	    *    NumDiskControllers[m] : 0
	    *    NumDiskControllers[m,n] : <ctrls>
	    *    NumUnitsPerController[m] : <upc>
	    * (note the first line should be eventually eliminated).
	    */
	   int nc = 0;
	   Simcpt_CptParamInt(cptd, "NumDiskControllers", machNo, NO_INDEX,
			      &nc);
	   if (nc != 0) {
	     /* compatibility: restoring from old checkpoint */
	     /* note old checkpoints can have more disk-carrying nodes
	      * than processor nodes. Ugh.
	      */
#define DEV_DISK_MAX_UNIT            128  /* max disks per node. */
	     int node;
	     Sim_Warning("*** Restoring disk ctrls info from old ckpt ***\n");
	     ASSERT(cptd->mode == CPT_RESTORE);
	     for (node = 0; node < nc; node++)
	       NUM_DISK_CONTROLLERS(machNo, node) = 1;
	     NUM_UNITS_PER_CONTROLLER(machNo) = DEV_DISK_MAX_UNIT;
	   } else {
	     /* new checkpoint code: save controllers/node, units/controller */
	     int node;
	     for (node = 0; node < NUM_CPUS(machNo); node++) {
	       Simcpt_CptParamInt(cptd, "NumDiskControllers",
				  machNo, node,
				  &NUM_DISK_CONTROLLERS(machNo, node));
	     }
	     Simcpt_CptParamInt(cptd, "NumUnitsPerController",
				machNo, NO_INDEX,
				&NUM_UNITS_PER_CONTROLLER(machNo));
 
	   }
	 } 
         Simcpt_CptParamInt(cptd, "NumClocks", machNo, NO_INDEX,
                            &(NUM_CLOCKS(machNo)));
         if (cptd->mode == CPT_RESTORE) {
            MEM_SIZE_SPECIFIED(machNo) = MEM_SIZE(machNo) / (1024 * 1024);
         }
      }   
      
      Simcpt_CptString(cptd, "EthersimHostname", NO_INDEX, NO_INDEX, 
                       &EthersimHostname);
      Simcpt_CptString(cptd, "EtherAddress", NO_INDEX, NO_INDEX,
                       &EtherAddress);
      
   } else {
     /* Unknown checkpoint version */
      
      CPUWarning("ERROR: Can't handle checkpoints of version %d.%d\n",
                 cptVersion.ver, cptVersion.sub);
      ASSERT(0);
   }

   /* Do console checkpointing */
   sim_console_ckpt(cptd);

   return 0;
}


#ifdef sun
#include <sys/types.h>
#include <netinet/in.h>
#include <arpa/inet.h>

int
inet_aton(char *cp, struct in_addr *pin)
{
   struct hostent *hp;

   hp = gethostbyname(cp);

   if (hp == NULL) return 0;
   
   bcopy(hp->h_addr_list[0],pin, sizeof(*pin));

   return 1;
}

#endif

#ifdef NEED_IF_STATIC_LINK
int dlopen() { return 0; }
int dlsym() { return 0; }
int dlerror() { return 0; }
#endif
