/************************************************************
 *
 *  vserial.c
 *
 *  implements a 16450 port
 *  XXX loopback of status flags only
 *
 ************************************************************/

#include "simtypes.h"
#include "sim_error.h"
#include "iodevs.h"
#include "iospace.h"
#include "vserial.h"

#define SERIAL_IER_RX_AVAIL          0x01
#define SERIAL_IER_TX_EMPTY          0x02
#define SERIAL_IER_LINE_STATUS       0x04
#define SERIAL_IER_MODEM_STATUS      0x08

#define SERIAL_IIR_NO_INTER          0x01
#define SERIAL_IIR_TX_EMPTY          0x02
#define SERIAL_IIR_RX_AVAIL          0x04

#define SERIAL_LCR_LEN_MASK          0x03
#define SERIAL_LCR_LEN_SHIFT         0
#define SERIAL_LCR_2STOP_BITS        0x04
#define SERIAL_LCR_PARITY            0x08
#define SERIAL_LCR_EVEN_PARITY       0x10
#define SERIAL_LCR_STICKY            0x20
#define SERIAL_LCR_BRK_CONTROL       0x40
#define SERIAL_LCR_DLAB              0x80

#define SERIAL_MCR_DTR               0x01
#define SERIAL_MCR_RTS               0x02
#define SERIAL_MCR_OUT1              0x04
#define SERIAL_MCR_ENABLE_IRQ        0x08
#define SERIAL_MCR_LOOPBACK          0x10

#define SERIAL_LSR_DATA_READY        0x01
#define SERIAL_LSR_OVERRUN_ERROR     0x02
#define SERIAL_LSR_PARITY_ERROR      0x04
#define SERIAL_LSR_FRAMING_ERROR     0x08
#define SERIAL_LSR_BRK               0x10
#define SERIAL_LSR_TX_EMPTY          0x20
#define SERIAL_LSR_TX_SHIFT_EMPTY    0x40

#define SERIAL_MSR_CTS_CHANGE        0x01
#define SERIAL_MSR_DSR_CHANGE        0x02
#define SERIAL_MSR_RDL_CHANGE        0x04
#define SERIAL_MSR_DCD_CHANGE        0x08
#define SERIAL_MSR_CTS               0x10
#define SERIAL_MSR_DSR               0x20
#define SERIAL_MSR_RIL               0x40
#define SERIAL_MSR_DCD               0x80

#define SERIAL_MSR_SYNC_BITS         (SERIAL_MSR_CTS   \
                                      | SERIAL_MSR_DSR \
                                      | SERIAL_MSR_RIL \
                                      | SERIAL_MSR_DCD)
   
#define SERIAL_MSR_CHANGE_BITS       (SERIAL_MSR_CTS_CHANGE    \
                                      | SERIAL_MSR_DSR_CHANGE  \
                                      | SERIAL_MSR_RDL_CHANGE  \
                                      | SERIAL_MSR_DCD_CHANGE)

#define DLAB(_sp)    ((_sp)->LCR & SERIAL_LCR_DLAB)

static int Reg0(IODevices *iod, uint16 port, int size, int inout, void *byte, 
                void *clientData);
static int Reg1(IODevices *iod, uint16 port, int size, int inout, void *byte, 
                void *clientData);
static int Reg2(IODevices *iod, uint16 port, int size, int inout, void *byte, 
                void *clientData);
static int Reg3(IODevices *iod, uint16 port, int size, int inout, void *byte, 
                void *clientData);
static int Reg4(IODevices *iod, uint16 port, int size, int inout, void *byte, 
                void *clientData);
static int Reg5(IODevices *iod, uint16 port, int size, int inout, void *byte, 
                void *clientData);
static int Reg6(IODevices *iod, uint16 port, int size, int inout, void *byte, 
                void *clientData);
static int Reg7(IODevices *iod, uint16 port, int size, int inout, void *byte, 
                void *clientData);
static void CheckInter(SerialPort *sp);


void VSerial_Init(SerialPort *sp, IODevices *iod, uint16 addr, uint8 IRQ)
{
   ASSERT(!(addr & 0x3));

   sp->iod = iod;
   sp->sd = NULL;
   
   sp->baseAddr = addr;
   sp->IRQ = IRQ;
   
   sp->IER = 0;
   sp->IIR = SERIAL_IIR_NO_INTER;
   sp->LCR = 0;
   sp->MCR = 0;
   sp->LSR = SERIAL_LSR_TX_EMPTY | SERIAL_LSR_TX_SHIFT_EMPTY;
   sp->MSR = 0; /* XXX ? */
   sp->SCR = 0;
   sp->DL = 0; /* XXX ? */
   
   IOSpace_Register(iod, sp->baseAddr + 0, (IOSpace_Func)Reg0, (void*)sp);
   IOSpace_Register(iod, sp->baseAddr + 1, (IOSpace_Func)Reg1, (void*)sp);
   IOSpace_Register(iod, sp->baseAddr + 2, (IOSpace_Func)Reg2, (void*)sp);
   IOSpace_Register(iod, sp->baseAddr + 3, (IOSpace_Func)Reg3, (void*)sp);
   IOSpace_Register(iod, sp->baseAddr + 4, (IOSpace_Func)Reg4, (void*)sp);
   IOSpace_Register(iod, sp->baseAddr + 5, (IOSpace_Func)Reg5, (void*)sp);
   IOSpace_Register(iod, sp->baseAddr + 6, (IOSpace_Func)Reg6, (void*)sp);
   IOSpace_Register(iod, sp->baseAddr + 7, (IOSpace_Func)Reg7, (void*)sp);
}

void VSerial_SetIOAddr(SerialPort *sp, uint16 addr)
{
   ASSERT(sp);
   ASSERT(sp->iod);
   
   IOSpace_Unregister(sp->iod, sp->baseAddr + 0, (IOSpace_Func)Reg0);
   IOSpace_Unregister(sp->iod, sp->baseAddr + 1, (IOSpace_Func)Reg1);
   IOSpace_Unregister(sp->iod, sp->baseAddr + 2, (IOSpace_Func)Reg2);
   IOSpace_Unregister(sp->iod, sp->baseAddr + 3, (IOSpace_Func)Reg3);
   IOSpace_Unregister(sp->iod, sp->baseAddr + 4, (IOSpace_Func)Reg4);
   IOSpace_Unregister(sp->iod, sp->baseAddr + 5, (IOSpace_Func)Reg5);
   IOSpace_Unregister(sp->iod, sp->baseAddr + 6, (IOSpace_Func)Reg6);
   IOSpace_Unregister(sp->iod, sp->baseAddr + 7, (IOSpace_Func)Reg7);

   sp->baseAddr = addr;

   IOSpace_Register(sp->iod, sp->baseAddr + 0, (IOSpace_Func)Reg0, (void*)sp);
   IOSpace_Register(sp->iod, sp->baseAddr + 1, (IOSpace_Func)Reg1, (void*)sp);
   IOSpace_Register(sp->iod, sp->baseAddr + 2, (IOSpace_Func)Reg2, (void*)sp);
   IOSpace_Register(sp->iod, sp->baseAddr + 3, (IOSpace_Func)Reg3, (void*)sp);
   IOSpace_Register(sp->iod, sp->baseAddr + 4, (IOSpace_Func)Reg4, (void*)sp);
   IOSpace_Register(sp->iod, sp->baseAddr + 5, (IOSpace_Func)Reg5, (void*)sp);
   IOSpace_Register(sp->iod, sp->baseAddr + 6, (IOSpace_Func)Reg6, (void*)sp);
   IOSpace_Register(sp->iod, sp->baseAddr + 7, (IOSpace_Func)Reg7, (void*)sp);  
}

void VSerial_SetIRQ(SerialPort *sp, uint8 IRQ)
{
   ASSERT(sp);
   sp->IRQ = IRQ;
}

void VSerial_Connect(SerialPort *sp, SerialDevice *sd)
{
   ASSERT(sp);
   ASSERT(sd);
   ASSERT(!sp->sd);
   sp->sd = sd;
   sd->connect(sd, sp);
}


void VSerial_BytesWaiting(SerialPort *sp)
{
   ASSERT(sp->sd && sp->sd->RxAvail);
   sp->LSR |= SERIAL_LSR_DATA_READY;
   CheckInter(sp);
}

static void CheckInter(SerialPort *sp)
{
   if (sp->sd && sp->sd->RxAvail && (sp->IER & SERIAL_IER_RX_AVAIL)) {
      ASSERT(sp->LSR & SERIAL_LSR_DATA_READY);
      sp->IIR = SERIAL_IIR_RX_AVAIL;
      Intr_Raise(sp->iod, sp->IRQ);
      return;
   }

   if (sp->IER & SERIAL_IER_TX_EMPTY) {
      ASSERT(sp->LSR & SERIAL_LSR_TX_EMPTY);
      sp->IIR = SERIAL_IIR_TX_EMPTY;
      Intr_Raise(sp->iod, sp->IRQ);
      return;
   }

   return;
}


static void SyncMSRBits(SerialPort *sp)
{
   sp->MSR &= ~SERIAL_MSR_SYNC_BITS;

   if (sp->MCR & SERIAL_MCR_DTR) {
      sp->MSR |= SERIAL_MSR_DSR;
   }

   if (sp->MCR & SERIAL_MCR_RTS) {
      sp->MSR |= SERIAL_MSR_CTS;
   }

   if (sp->MCR & SERIAL_MCR_OUT1) {
      sp->MSR |= SERIAL_MSR_RIL;
   }

   if (sp->MCR & SERIAL_MCR_ENABLE_IRQ) {
      sp->MSR |= SERIAL_MSR_DCD;
   }
}

/*
 *  Register 0
 *
 *   (DLAB = 0) (I/O)    Transmit/Receive Buffer
 *   (DLAB = 1) (I/O)    Divisor Latch LSB
 *
 */

int Reg0(IODevices *iod, uint16 port, int size, int inout, void *byte, void *clientData)
{
   SerialPort *sp = (SerialPort *)clientData;

   ASSERT(sp);
   ASSERT(sp->baseAddr == (port - 0));
   ASSERT(sp->iod == iod);
   
   if (DLAB(sp)) {
      /* divisor latch LSB */
      if (inout == PORT_IN) {
         *(uint8 *)byte = sp->DL & 0xff;
      } else {
         sp->DL = (sp->DL & 0xff00) | *(uint8 *)byte;
      }
      
   } else {
      if (inout == PORT_IN) {
         if (sp->sd && sp->sd->RxAvail) {
            *(uint8 *)byte = sp->sd->receiveByte(sp->sd);

            if (sp->sd->RxAvail) {
               sp->LSR |= SERIAL_LSR_DATA_READY;
            } else {
               sp->LSR &= ~SERIAL_LSR_DATA_READY;
            }

         } else {
            *(uint8 *)byte = 0x0;
         }
      } else {
         if (sp->sd) {
            sp->sd->transmitByte(sp->sd, *(uint8 *)byte);
         }
      }
   }

   return 0;
}


/*
 *  Register 1
 *
 *   (DLAB = 0) (I/O)    Interrupt Enable
 *   (DLAB = 1) (I/O)    Divisor Latch MSB
 *
 */
int Reg1(IODevices *iod, uint16 port, int size, int inout, void *byte, void *clientData)
{
   SerialPort *sp = (SerialPort *)clientData;

   ASSERT(sp);
   ASSERT(sp->baseAddr == (port - 1));
   ASSERT(sp->iod == iod);

   if (DLAB(sp)) {
      /* divisor latch MSB */
      if (inout == PORT_IN) {
         *(uint8 *)byte = (sp->DL >> 8) & 0xff;
      } else {
         sp->DL = (*(uint8 *)byte << 8) | (sp->DL & 0x00ff);
      }
      
   } else {
      /* interrupt enable register */
      if (inout == PORT_IN) {
         *(uint8 *)byte = sp->IER;
      } else {
         sp->IER = *(uint8 *)byte;
         CheckInter(sp);
      }
   }

   return 0;
}


/*
 *  Register 2
 *
 *   (DLAB = 0) (Input)  Interrupt Identification
 *   (DLAB = 0) (Output) undefined (FIFO Control on 1655x only)
 *   (DLAB = 1) (I/O)    undefined (Alternate Function on 16552 only)
 *
 */

int Reg2(IODevices *iod, uint16 port, int size, int inout, void *byte, void *clientData)
{
   SerialPort *sp = (SerialPort *)clientData;

   ASSERT(sp);
   ASSERT(sp->baseAddr == (port - 2));
   ASSERT(sp->iod == iod);

   if (!DLAB(sp) && (inout == PORT_IN)) {
      *(uint8 *)byte = sp->IIR;
   } else {
      ASSERT(inout == PORT_OUT);
      /*
       * ignore - linux probes this to see if we have a 16650
       * which we are not currently emulating
       */
   }
   
   return 0;
}


/*
 *  Register 3
 *
 *   (DLAB = 0) (I/O)    Line Control  
 *
 */
int Reg3(IODevices *iod, uint16 port, int size, int inout, void *byte, void *clientData)
{
   SerialPort *sp = (SerialPort *)clientData;

   ASSERT(sp);
   ASSERT(sp->baseAddr == (port - 3));
   ASSERT(sp->iod == iod);
   /* ignore DLAB */
   
   if (inout == PORT_IN) {
      *(uint8 *)byte = sp->LCR;
   } else {
      sp->LCR = *(uint8 *)byte;
   }
   
   return 0;
}


/*
 *  Register 4
 *
 *   (DLAB = 0) (I/O)    Modem Control  
 *
 */
int Reg4(IODevices *iod, uint16 port, int size, int inout, void *byte, void *clientData)
{
   SerialPort *sp = (SerialPort *)clientData;

   ASSERT(sp);
   ASSERT(sp->baseAddr == (port - 4));
   ASSERT(sp->iod == iod);
   ASSERT(!DLAB(sp));

   if (inout == PORT_IN) {
      *(uint8 *)byte = sp->MCR;
   } else {
      sp->MCR = *(uint8 *)byte;
      if (sp->MCR & SERIAL_MCR_LOOPBACK) {
         SyncMSRBits(sp);
      }
   }   
   
   return 0;
}


/*
 *  Register 5
 *
 *   (DLAB = 0) (I/O)    Line Status
 *
 */
int Reg5(IODevices *iod, uint16 port, int size, int inout, void *byte, void *clientData)
{
   SerialPort *sp = (SerialPort *)clientData;

   ASSERT(sp);
   ASSERT(sp->baseAddr == (port - 5));
   ASSERT(sp->iod == iod);
   ASSERT(!DLAB(sp));

   ASSERT(inout == PORT_IN);
   *(uint8 *)byte = sp->LSR;
   
   return 0;
}


/*
 *  Register 6
 *
 *   (DLAB = 0) (I/O)    Modem Status
 *
 */
int Reg6(IODevices *iod, uint16 port, int size, int inout, void *byte, void *clientData)
{
   SerialPort *sp = (SerialPort *)clientData;

   ASSERT(sp);
   ASSERT(sp->baseAddr == (port - 6));
   ASSERT(sp->iod == iod);
   ASSERT(!DLAB(sp));

   if (inout == PORT_IN) {
      *(uint8 *)byte = sp->MSR;
      sp->MSR &= ~(SERIAL_MSR_CHANGE_BITS);
   } else {
      sp->MSR = *(uint8 *)byte;
      if (sp->MCR & SERIAL_MCR_LOOPBACK) {
         SyncMSRBits(sp);
      }      
   }
   
   return 0;
}


/*
 *  Register 7
 *
 *   (DLAB = 0) (I/O)    Scratch Pad
 *
 */
int Reg7(IODevices *iod, uint16 port, int size, int inout, void *byte, void *clientData)
{
   SerialPort *sp = (SerialPort *)clientData;

   ASSERT(sp);
   ASSERT(sp->baseAddr == (port - 7));
   ASSERT(sp->iod == iod);
   ASSERT(!DLAB(sp));

   if (inout == PORT_IN) {
      *(uint8 *)byte = sp->SCR;
   } else {
      sp->SCR = *(uint8 *)byte;
   }   

   return 0;
}



