/* $Id: data_stream.cc,v 1.4 1998/04/29 04:54:47 nikos Exp $
 *
 * RCS Log:
 * $Log: data_stream.cc,v $
 * Revision 1.4  1998/04/29 04:54:47  nikos
 * Added new AranaTrace command "AranaTrace  filter_thru_L1" that dumps
 * only instructions with references (pc + ea) that miss in the L1 cache.
 * MBs are always dumped.
 *
 * Revision 1.4  1998/03/25 19:16:37  nikos
 * atf_tlb_fixed
 *
// Revision 1.3  1998/03/13  17:12:23  nikos
// Fix vaddr->physaddr translation bug and tlb behavior in ATF.
// Changes affect only people who use ATF for their traces.
//
 * Revision 1.3  1998/03/13 16:36:56  nikos
 * .
 *
// Revision 1.2  1998/02/16  23:19:56  nikos
// Fixed AranaTrace command which was broken with the last commit.
// Macros in external_trace.h are set such that the name "arana"
// appears only in arana_trace.[ch] and no-where else!!!
// In the previous revision, kappa.c has not been cleaned up.
//
// If you want to use the AranaTrace command,
// compile SimOS with the following variables set :
//
// EXTERNAL_TRACE=ARANA_TRACE
// ARANA_TRACE_INC_DIR=/perf_sym/ATF/atf
// ARANA_TRACE_LIB=/perf_sym/ATF/atf/libatf.a
// ARANA_TRACE_OBJS=arana_trace.o
// ARANA_TRACE_HEADER=arana_trace.h
//
// This commit reports a bunch of modified files which are not
// really modified (only some rcs fields like $Id: data_stream.cc,v 1.4 1998/04/29 04:54:47 nikos Exp $, $Header: /wrl/proj/simos/cvsroot/simulation/cpus-x86/x86sim/data_stream.cc,v 1.4 1998/04/29 04:54:47 nikos Exp $, etc).
//
 * Revision 1.2  1998/02/11 14:24:55  nikos
 * .
 *
// Revision 1.1  1997/12/16  01:28:24  bugnion
// stanford update (x86)
//
// Revision 1.4  1997/03/18  21:34:02  sjt
// fixed stupid bug in DataStream::ReadNBytes (using wrong variable in
//   check for unread data)
//
 * Revision 1.3  1997/03/06 02:12:17  sean
 * Data Streams now keep track of how many bytes have been read from them
 *
 * Revision 1.2  1997/03/02 03:04:33  sean
 * massive modifications to balance of power between DataStream class and its
 * children, created UnRead...() calls that work.
 *
 * Revision 1.1  1997/02/11 05:08:26  sjt
 * Initial revision
 *
 */

static char rcsid[] __attribute__((unused))
        = "$Id: data_stream.cc,v 1.4 1998/04/29 04:54:47 nikos Exp $";

#include "data_stream.h"
#include "cpu.h"
#include "seg_desc.h"

// This stuff defines a general "stream"-like object that can be used to
//  read data from things.  So far, the only instantiation of this is the
//  MemoryStream class, which rips bytes from a CPU's memory.

/////////////////////////////////////////////////////////////////////////////

// These DataStream methods provide basic functionality so that a specific
//   DataStream subclass need only override the ReadNBytesFromStream() method.

DataStream::DataStream(int32 _max_unread_bytes)
{
  unread_bytes = 0;

  if((_max_unread_bytes > 0) &&
     ((unread_buffer = new int8[_max_unread_bytes]) != 0)) {
    max_unread_bytes = _max_unread_bytes;
  } else {
    max_unread_bytes = 0;
    unread_buffer = 0;
  }

  bytes_read = max_bytes_read = 0;
}

DataStream::~DataStream(void)
{
  delete[] unread_buffer;
}

bool DataStream::ReadOneByte(int8& data)
{
  return(ReadNBytes(&data,1));
}

bool DataStream::ReadTwoBytes(int16& data)
{
  int16 local;

  if(ReadNBytes((int8 *) &local, 2)) {
    data = xtoh16(local);
    return(true);
  } else
    return(false);
}

bool DataStream::ReadFourBytes(int32& data)
{
  int32 local;

  if(ReadNBytes((int8 *) &local, 4)) {
    data = xtoh32(local);
    return(true);
  } else
    return(false);
}

bool DataStream::ReadEightBytes(int64& data)
{
  int64 local;

  if(ReadNBytes((int8 *) &local, 8)) {
    data = xtoh64(local);
    return(true);
  } else
    return(false);
}

bool DataStream::ReadNBytes(int8 *buffer, int32 n)
{
  int32 to_read, n_left;

  // check the unread buffer first
  if((unread_bytes > 0) && (n > 0)) {
    to_read = MIN(unread_bytes,n);

    // read from the start of valid unread bytes (not front of buffer)
    memcpy(buffer, unread_buffer + (max_unread_bytes - unread_bytes), to_read);
    buffer += to_read;
    n_left = n - to_read;
    unread_bytes -= to_read;
  } else
    n_left = n;

  if((n_left == 0) || ReadNBytesFromStream(buffer,n_left)) {
    bytes_read += n;
    if(bytes_read > max_bytes_read) max_bytes_read = bytes_read;
    return(true);
  } else
    return(false);
}

bool DataStream::UnReadNBytes(const int8 *buffer, int32 n)
{
  // first, see if we've got space
  if((unread_bytes + n) > max_unread_bytes) return(false);

  // now just copy into the front of the valid data
  unread_bytes += n;
  memcpy(unread_buffer + (max_unread_bytes - unread_bytes), buffer, n);

  bytes_read -= n;

  return(true);
}

bool DataStream::UnReadOneByte(int8 data)
{
  // no need to fix byte order
  return(UnReadNBytes((int8 *) &data,1));
}

bool DataStream::UnReadTwoBytes(int16 data)
{
  // fix byte order
  data = xtoh16(data);

  // slam it into the unread buffer
  return(UnReadNBytes((int8 *) &data,2));
}

bool DataStream::UnReadFourBytes(int32 data)
{
  // fix byte order
  data = xtoh32(data);

  // slam it into the unread buffer
  return(UnReadNBytes((int8 *) &data,4));
}

bool DataStream::UnReadEightBytes(int64 data)
{
  // fix byte order
  data = xtoh64(data);

  // slam it into the unread buffer
  return(UnReadNBytes((int8 *) &data,8));
}

/////////////////////////////////////////////////////////////////////////////

// This class provides a DataStream-like interface for reading data from a
//  CPU's memory

MemoryStream::MemoryStream(CPU *_cpu)
{
  cpu = _cpu;
  linear = linear_avail = 0;
  physical = physical_avail = 0;
  is_stack = false;
}

MemoryStream::MemoryStream(CPU *_cpu, SegmentDescriptor& _segment, 
			   int32 _offset, int32 _max_len, bool _fetching,
			   bool _is_stack, bool _user)
{
  cpu = _cpu;
  ResetStream(_segment,_offset,_max_len,_fetching,_is_stack,_user);
}

MemoryStream::~MemoryStream(void) {}

bool MemoryStream::ResetStream(SegmentDescriptor& _segment, int32 _offset, 
			       int32 _max_len, bool _fetching, bool _is_stack,
			       bool _user)
{
  linear_avail = MIN(_max_len, 
		     cpu->TranslateVirtualToLinear(_segment,_offset,linear,
						   false,_fetching,_is_stack,
						   false));

  is_stack = _is_stack;;
  user = _user;
  physical = physical_avail = 0;

  return(linear_avail != 0);
}

bool MemoryStream::ReadNBytesFromStream(int8 *buffer, int32 n)
{
  int32 chunk_len;

  while(n > 0) {
    if(physical_avail == 0) {
      if(linear_avail != 0) {
	physical_avail = MIN(linear_avail,
			     cpu->TranslateLinearToPhysical(linear,physical,
							    user,false,false));

	if(physical_avail == 0) return(false);
      } else {
	cpu->ExceptionOccurred((is_stack ? EXC_STACK_FAULT : EXC_PROTECTION), 0);
	return(false);
      }
    }

    chunk_len = MIN(n, physical_avail);
    
    cpu->ReadFromPhysicalAddress(physical,buffer,chunk_len);
    buffer += chunk_len;
    n -= chunk_len;

    linear += chunk_len;
    linear_avail -= chunk_len;
    physical += chunk_len;
    physical_avail -= chunk_len;
  }

  return(true);
}

