/* -*- C++ -*-
 *
 * $Id: data_stream.h,v 1.4 1998/04/29 04:54:48 nikos Exp $
 * 
 * RCS Log:
 * $Log: data_stream.h,v $
 * Revision 1.4  1998/04/29 04:54:48  nikos
 * Added new AranaTrace command "AranaTrace  filter_thru_L1" that dumps
 * only instructions with references (pc + ea) that miss in the L1 cache.
 * MBs are always dumped.
 *
 * Revision 1.4  1998/03/25 19:16:38  nikos
 * atf_tlb_fixed
 *
 * Revision 1.3  1998/03/13  17:12:23  nikos
 * Fix vaddr->physaddr translation bug and tlb behavior in ATF.
 * Changes affect only people who use ATF for their traces.
 *
 * Revision 1.3  1998/03/13 16:36:57  nikos
 * .
 *
 * Revision 1.2  1998/02/16  23:19:56  nikos
 * Fixed AranaTrace command which was broken with the last commit.
 * Macros in external_trace.h are set such that the name "arana"
 * appears only in arana_trace.[ch] and no-where else!!!
 * In the previous revision, kappa.c has not been cleaned up.
 *
 * If you want to use the AranaTrace command,
 * compile SimOS with the following variables set :
 *
 * EXTERNAL_TRACE=ARANA_TRACE
 * ARANA_TRACE_INC_DIR=/perf_sym/ATF/atf
 * ARANA_TRACE_LIB=/perf_sym/ATF/atf/libatf.a
 * ARANA_TRACE_OBJS=arana_trace.o
 * ARANA_TRACE_HEADER=arana_trace.h
 *
 * This commit reports a bunch of modified files which are not
 * really modified (only some rcs fields like $Id: data_stream.h,v 1.4 1998/04/29 04:54:48 nikos Exp $, $Header: /wrl/proj/simos/cvsroot/simulation/cpus-x86/x86sim/data_stream.h,v 1.4 1998/04/29 04:54:48 nikos Exp $, etc).
 *
 * Revision 1.2  1998/02/11 14:24:55  nikos
 * .
 *
 * Revision 1.1  1997/12/16  01:28:24  bugnion
 * stanford update (x86)
 *
 * Revision 1.4  1997/03/06  02:12:17  sean
 * Data Streams now keep track of how many bytes have been read from them
 *
 * Revision 1.3  1997/03/02 03:04:33  sean
 * massive modifications to balance of power between DataStream class and its
 * children, created UnRead...() calls that work.
 *
 * Revision 1.2  1997/02/28 03:52:44  sean
 * Added (somewhat broken) prototype for un-reading data.
 *
 * Revision 1.1  1997/02/11 05:09:52  sjt
 * Initial revision
 *
 */

// This stuff defines a general "stream"-like object that can be used to
//  read data from things.  So far, the only instantiation of this is the
//  MemoryStream class, which rips bytes from a CPU's memory.

#ifndef _DATA_STREAM_H
#define _DATA_STREAM_H

#include "ints.h"

class CPU;
class SegmentDescriptor;

#define MAX_UNREAD_BYTES   16

// This is the abstract superclass for all data stream objects.
class DataStream {
 public:
  DataStream(int32 _max_unread_bytes = MAX_UNREAD_BYTES);

  virtual ~DataStream(void);

  // Each of these function reads the next 'n' bytes from the data stream,
  //  and returns them in host (not necessarily CPU) byte order.  In each
  //  case, 'data' is where the read value is placed, and the function returns
  //  true if it was able to read the necessary number of bytes, and false if
  //  any errors occurred (in which case the value of 'data' is undefined.
  virtual bool ReadOneByte(int8& data);
  virtual bool ReadTwoBytes(int16& data);
  virtual bool ReadFourBytes(int32& data);
  virtual bool ReadEightBytes(int64& data);

  // These allow data to be "unread" in the sense that they are put back
  //  into the data stream, and future Read...() calls will pull the data
  //  back out.
  virtual bool UnReadOneByte(int8 data);
  virtual bool UnReadTwoBytes(int16 data);
  virtual bool UnReadFourBytes(int32 data);
  virtual bool UnReadEightBytes(int64 data);

  inline int BytesRead(void) const { return(bytes_read); }
  inline int MaxBytesRead(void) const { return(max_bytes_read); }

 protected:
  bool ReadNBytes(int8 *buffer, int32 n);
  bool UnReadNBytes(const int8 *buffer, int32 n);

  // This is the one function that a subclass of DataStream must override.
  // This function takes care of reading 'n' bytes from the stream into
  //  the 'buffer'.  It is called by ReadNBytes(), which also takes care of
  //  reading previously un-read characters
  virtual bool ReadNBytesFromStream(int8 *buffer, int32 n) = 0;

 private:
  int8 *unread_buffer;
  int32 unread_bytes, max_unread_bytes;

  int bytes_read, max_bytes_read;
};

// This class provides a DataStream-like interface for reading data from a
//  CPU's memory
class MemoryStream : public DataStream {
 public:
  MemoryStream(CPU *_cpu);
  MemoryStream(CPU *_cpu, SegmentDescriptor& _segment, int32 _offset,
	       int32 _max_len, bool _fetching, bool _is_stack, bool _user);

  virtual ~MemoryStream(void);

  bool ResetStream(SegmentDescriptor& _segment, int32 _offset, int32 _max_len,
		   bool _fetching, bool _is_stack, bool _user);

 protected:
  virtual bool ReadNBytesFromStream(int8 *buffer, int32 n);

 private:
  CPU *cpu;
  int32 linear, linear_avail;
  int32 physical, physical_avail;

  bool is_stack, user;
};

#endif

