#include <sys/types.h>
#include <sys/stat.h>
#include <sys/mman.h>
#include <fcntl.h>
#include <stdio.h>
#include <unistd.h>
#include <string.h>
#include <stdlib.h>
#include "load_elf.h"
#include "cpu.h"
#include "debug.h"
#include "instruction.h"

#define ROUND_TO_4(x)      (((x) + 3) & ~3L)

// given an argument list (like "argv"), it determines the number of arguments
//  in the list
static int ArglistEntries(const char **arglist)
{
  int n;

  if(arglist) {
    n = 0;
    while(*arglist) {
      n++;
      arglist++;
    }
    return(n);
  } else
    return(0);
}

// given an argument list (like "argv"), this function determines the space
//  necessary to store all of the strings in the argument list (the start of
//  each string is assumed to be 4-byte aligned)
static int ArglistBytesNeeded(const char **arglist)
{
  int total;

  if(arglist) {
    total = 0;

    while(*arglist)
//      Do NOT round to 4!
//      total += ROUND_TO_4(strlen(*arglist++) + 1);
        total += strlen(*arglist++) + 1;

    return(total);
  } else
    return(0);
}

// this function copies the data in an 'arglist' (like "argv") into a 'cpu's
//  memory - the array of string pointers is placed at 'array_base', and the
//  text data is placed starting at 'text_base' (the caller needs to make
//  sure these two values are appropriate)
static int CopyArglistToCPU(const char **arglist, CPU& cpu,
			    int32 array_base, int32 text_base)
{
  int32 ptr;
  int len;

  if(arglist)
    while(*arglist) {
      // put in the pointer to the string
      ptr = htox32(text_base);
      if(!cpu.WriteToLinearAddress(array_base,&ptr,SIZE_32BIT,false)) 
	return(-1);
      array_base += SIZE_32BIT;

      // now write the string
      len = strlen(*arglist) + 1;
      if(!cpu.WriteToLinearAddress(text_base,*arglist,len,false)) return(-1);

      // Don't do this!  Linux does not round to 4!
      //text_base += ROUND_TO_4(len);
      text_base += len;
				   
      arglist++;
    }

  ptr = 0;  // no need to re-order these bytes... :)
  if(!cpu.WriteToLinearAddress(array_base,&ptr,SIZE_32BIT,false)) return(-1);

  return(0);
}

// This function takes care of putting initial data on a new program's stack.
//   In addition to the programs arguments (in 'argv') and environment (in
//   'envp'), it appears to be standard ELF behavior to put the 'filename' on
//   the very top of the stack.  Also, if a program is being loaded by an
//   interpreter (i.e. it's dynamically loaded), the fairly cryptic 'dlinfo'
//   block of information for _dl_boot() must be placed on the stack as well
static int PutArglistsOnStack(const char *filename,
			      const char **argv, const char **envp,
			      void *dlinfo, int dlinfo_size,
			      CPU& cpu, int32& stack_ptr)
{
  int argc;
  int32 fname_text;
  int32 dlinfo_ptr;
  int32 argv_array, argv_text;
  int32 envp_array, envp_text;
  int32 zpad_start, zpad_bytes;
  int8  zero;

  // step down the values in the correct order
//  No ROUND_TO_4!
//  fname_text = (stack_ptr -= ROUND_TO_4(strlen(filename)+1));

  // Linux stack appear to leave the top 4 bytes of the stack 0.
  stack_ptr -= SIZE_32BIT;

  fname_text = (stack_ptr -= strlen(filename)+1);
  envp_text  = (stack_ptr -= ArglistBytesNeeded(envp));
  argv_text  = (stack_ptr -= ArglistBytesNeeded(argv));

  // dlinfo_ptr (actually the top of the block) needs to be on a 16-byte
  //  boundary (cause the kernel code says so...) - extra bytes are 0'd out
  zpad_bytes = (stack_ptr & 15UL);
  zpad_start = (stack_ptr -= zpad_bytes);

  dlinfo_ptr = (stack_ptr -= dlinfo_size);
  envp_array = (stack_ptr -= ((ArglistEntries(envp) + 1) * SIZE_32BIT));
  argv_array = (stack_ptr -= (((argc = ArglistEntries(argv)) + 1) * SIZE_32BIT));

  zero = 0;
  while(zpad_bytes-- > 0)
    if(!cpu.WriteToLinearAddress(zpad_start++, &zero, SIZE_8BIT, false))
      return(-1);

  if(dlinfo_size && !cpu.WriteToLinearAddress(dlinfo_ptr,dlinfo,dlinfo_size,
					      false)) return(-1);

  if(CopyArglistToCPU(argv,cpu,argv_array,argv_text) < 0) return(-1);
  if(CopyArglistToCPU(envp,cpu,envp_array,envp_text) < 0) return(-1);

  if(!cpu.WriteToLinearAddress(fname_text,filename,strlen(filename)+1,false))
    return(-1);

  // now write argc (in proper byte order)
  argc = htox32(argc);
  if(!cpu.WriteToLinearAddress(stack_ptr -= SIZE_32BIT,&argc,SIZE_32BIT,false))
    return(-1);

  return(0);
}

// This function copies data from a file into memory.
//   'file_size' bytes from an offset of 'file_addr' in the file 'exec_fd' are
//   copied to 'mem_size' bytes at 'mem_addr' on the 'cpu'.  If 'file_size' is
//   greater than 'mem_size' the extra bytes are not copied.  If 'file_size' is
//   smaller than 'mem_size' the extra memory is zeroed.
static int CopyELFSegmentToCPU(int exec_fd, int32 file_addr, int32 file_size,
			       CPU& cpu, int32& mem_addr, int32 mem_size,
			       bool can_reloc,
			       MemCreateFunc create_fn, void *create_param)
{
  int8 buffer[4096];
  int len; 
  int32 total;

  if(create_fn &&
     (create_fn(mem_addr,mem_size,!can_reloc,create_param) < 0)) return(-1);
  
  lseek(exec_fd, file_addr, SEEK_SET);

  total = 0;

  while((total < file_size) && (total < mem_size)) {
    len = MIN(4096,(file_size - total));
    if((len = read(exec_fd, buffer, len)) <= 0) return(-1);

    if(!cpu.WriteToLinearAddress(mem_addr + total, buffer, len, false))
      return(-1);

    total += len;
  }

  // if memsz > filesz, zero the extra
  if(file_size < mem_size) {
    memset(buffer, 0, 4096);

    while(total < mem_size) {
      len = MIN(4096,(mem_size - total));

      if(!cpu.WriteToLinearAddress(mem_addr + total, buffer, len, false))
	return(-1);

      total += len;
    }
  }

  return(0);
}

// these are just time-savers to do in-place byte order swapping
#define ISWAP16(val)     (val) = FLIP16(val)
#define ISWAP32(val)     (val) = FLIP32(val)

// This swaps the byte order on all the relevant fields in an ELF header
static void ByteSwapELFHeader(ELF32_Header& hdr)
{
  ISWAP16(hdr.e_type);
  ISWAP16(hdr.e_machine);
  ISWAP32(hdr.e_version);
  ISWAP32(hdr.e_entry);
  ISWAP32(hdr.e_phoff);
  ISWAP32(hdr.e_shoff);
  ISWAP32(hdr.e_flags);
  ISWAP16(hdr.e_ehsize);
  ISWAP16(hdr.e_phentsize);
  ISWAP16(hdr.e_phnum);
  ISWAP16(hdr.e_shentsize);
  ISWAP16(hdr.e_shnum);
  ISWAP16(hdr.e_shstrndx);
}

static void ByteSwapELFSegmentHeader(ELF32_SegmentHeader& hdr)
{
  ISWAP32(hdr.p_type);
  ISWAP32(hdr.p_offset);
  ISWAP32(hdr.p_vaddr);
  ISWAP32(hdr.p_paddr);
  ISWAP32(hdr.p_filesz);
  ISWAP32(hdr.p_memsz);
  ISWAP32(hdr.p_flags);
  ISWAP32(hdr.p_align);
}

// This is the magic function that attempts to load the binary in 'filename'
//  onto the 'cpu'.
// Arguments:
//     cpu          - the CPU on which to load the program
//     filename     - name of the executable to load
//     load_offset  - amount to add to all load addresses (shifts image in
//                      memory, but not pointers in code, so don't make this
//                      non-zero unless you know what you're doing)
//     initial_sp   - initial top of the stack for the new program (it is
//                      assumed that space below this address has already been
//                      allocated - i.e. the 'create_fn' will NOT be called for
//                      stack addresses)
//     argv         - the arglist that should be given to the program as 'argv'
//     envp         - the arglist that should be given to the program as 'envp'
//     create_fn    - the function that should be used to make sure memory
//                      exists at a given address (can be 0 if no fn should be
//                      called)
//     create_param - the parameter that should be passed to the 'create_fn'
//                      any time it is called
//     load_lib     - true if we are expecting to load a dynamic library;
//                      false if we are expecting to load an executable
//     simulated_brk_ptr - if non-null, the brk value (first unused byte in
//                           in memory) of the simulated program is stored
// Return value:  This function returns 0 if 'filename' was loaded 
//   successfully, and a number < 0 if errors occurred.
int LoadELFBinary(CPU& cpu, const char *filename, 
		  int32 load_offset, int32 initial_sp,
		  const char *argv[], const char *envp[],
		  MemCreateFunc create_fn, void *create_param, bool load_lib,
		  int32 *simulated_brk_ptr)
{
  int exec_fd, i;
  ELF32_Header elf_hdr;
  ELF32_SegmentHeader seg_hdr;
  RegisterFile *regptr;
  int32 current_sp;
  SegmentDescriptor *seg_descs;
  char *interp;
  int keep_eip = 0;
  int32 dlinfo[24];
  bool byte_swap_headers;  // true if we need to swap byte order in headers
  int32 load_base = 0;     // where we loaded our binary
  bool load_base_set = false;
  int32 seg_load_addr;
  bool seg_can_move;
  int32 interp_load_base = 0;
  int extra_before;

  // initialize some values...
  if(simulated_brk_ptr != NULL) *simulated_brk_ptr = 0;

  // first, try to open the file
  if((exec_fd = open(filename,O_RDONLY)) < 0) {
    DEBUG(DEBUG_LOADER,"Could not open '%s'.",filename);
    return(-1);
  }

  // now, try to read an ELF header...
  if(read(exec_fd, &elf_hdr, sizeof(ELF32_Header)) < 
     ((int) sizeof(ELF32_Header))) {
    DEBUG(DEBUG_LOADER,"Could not read ELF header from '%s'.",filename);
    close(exec_fd);
    return(-1);
  }

  // make sure it's got the correct sig, and figure out if we need to swap
  //  the byte ordering in our headers
  if(!strncmp((char *) elf_hdr.e_ident,ELF_SIG_LE,ELF_SIGLEN)) {
#ifdef LITTLE_ENDIAN
    byte_swap_headers = false;
#else
    byte_swap_headers = true;
#endif
  } else
  if(!strncmp((char *) elf_hdr.e_ident,ELF_SIG_BE,ELF_SIGLEN)) {
#ifdef LITTLE_ENDIAN
    byte_swap_headers = true;
#else
    byte_swap_headers = false;
#endif
  } else {
    DEBUG(DEBUG_LOADER,"Invalid ELF signature in '%s' - not a 32-bit "
	               "ELF executable?",filename);
    close(exec_fd);
    return(-1);
  }

  if(byte_swap_headers) ByteSwapELFHeader(elf_hdr);

  // make sure sizes of headers (file and segment) check out
  if((elf_hdr.e_ehsize != sizeof(ELF32_Header)) ||
     (elf_hdr.e_phentsize != sizeof(ELF32_SegmentHeader))) {
    DEBUG(DEBUG_LOADER,"ELF header size mismatch in '%s'!",filename);
    close(exec_fd);
    return(-1);
  }

  // make sure it's an executable for the right kind of machine (i386,i486)
  if((elf_hdr.e_machine != ELFMACHINE_i386) &&
     (elf_hdr.e_machine != ELFMACHINE_i486)) {
    DEBUG(DEBUG_LOADER,"'%s' is not an i386 (or i486) ELF file.",filename);
    close(exec_fd);
    return(-1);
  }

  if(!load_lib && (elf_hdr.e_type != ELFTYPE_EXEC)) {
    DEBUG(DEBUG_LOADER,"'%s' is not an ELF executable.",filename);
    close(exec_fd);
    return(-1);
  }

  if(load_lib && (elf_hdr.e_type != ELFTYPE_DYNAMIC)) {
    DEBUG(DEBUG_LOADER,"'%s' is not an ELF dynamic library.",filename);
    close(exec_fd);
    return(-1);
  }

  // we got through all the checks, so start loading segments!
  DEBUG(DEBUG_LOADER,"'%s' is ELF, %d segments.",filename,elf_hdr.e_phnum);

  for(i = 0; i < elf_hdr.e_phnum; i++) {
    // jump to proper place in file
    lseek(exec_fd, elf_hdr.e_phoff + (i * elf_hdr.e_phentsize), SEEK_SET);

    // read in the header, checking for errors
    if(read(exec_fd, &seg_hdr, sizeof(ELF32_SegmentHeader)) <
       ((int) sizeof(ELF32_SegmentHeader))) {
      DEBUG(DEBUG_LOADER,"Could not read ELF segment header %d from '%s'.",i,filename);
      close(exec_fd);
      return(-1);
    }
    
    // swap bytes if necessary
    if(byte_swap_headers) ByteSwapELFSegmentHeader(seg_hdr);

    DEBUG(DEBUG_LOADER, "Type of segment %d is %d.", i, seg_hdr.p_type);

    // watch for INTERP segment types, which specify something to preload
    //  (like /lib/ld-linux.so.1)
    if(seg_hdr.p_type == ELFSEG_INTERP) {
      interp = new char[seg_hdr.p_filesz + 1];

      lseek(exec_fd, seg_hdr.p_offset, SEEK_SET);
      if(read(exec_fd, interp, seg_hdr.p_filesz) < (int) seg_hdr.p_filesz) {
	DEBUG(DEBUG_LOADER,"Could not read name of interpreter for  '%s'.",filename);
	close(exec_fd);
	delete[] interp;
	return(-1);
      }

      // null terminate in case file didn't
      interp[seg_hdr.p_filesz] = '\0';

      DEBUG(DEBUG_LOADER,"'%s' wants '%s' as an intepreter.",filename,interp);

      // Try to load the interpreter
      if(LoadELFBinary(cpu, interp, 0, initial_sp, NULL, NULL,
		       create_fn, create_param, true, 0) < 0) {
	close(exec_fd);
	delete[] interp;
	return(-1);
      }

      // mark that we want to keep the currently set CPU_EIP
      keep_eip = 1;

      // sort of a hack - the interpreter will leave its load address in
      //  ECX
      interp_load_base = REG32(*(cpu.GetRegisters()),REG_ECX);

      delete[] interp;
    }
    
    // we only need to look at LOAD segment types
    if(seg_hdr.p_type == ELFSEG_LOAD) {
      // figure out where we want to load this segment
      seg_load_addr = seg_hdr.p_vaddr + load_offset;
      seg_can_move = load_lib && !load_base_set;

      DEBUG(DEBUG_LOADER,"going to load (%d,%08X,%d) -> (%08X,%d,%d)",
	    exec_fd,seg_hdr.p_offset,seg_hdr.p_filesz,
	    seg_load_addr,seg_hdr.p_memsz,seg_can_move);

      extra_before = (seg_hdr.p_offset & (PAGE_SIZE - 1));
      seg_load_addr -= extra_before;

      if(CopyELFSegmentToCPU(exec_fd, seg_hdr.p_offset - extra_before, 
			     seg_hdr.p_filesz + extra_before,
			     cpu, 
			     seg_load_addr, seg_hdr.p_memsz + extra_before, 
			     seg_can_move, create_fn, create_param) < 0) {
	DEBUG(DEBUG_LOADER,"Could not read ELF segment %d from '%s'.",i,filename);
	close(exec_fd);
	return(-1);
      }

      seg_load_addr += extra_before;

      // if we need to set a brk value, check to see if this is higher than
      //  before
      if((simulated_brk_ptr != NULL) &&
	 (*simulated_brk_ptr < (seg_load_addr + seg_hdr.p_memsz))) {
	DEBUG(DEBUG_LOADER, "Simulated brk is now %08X.", 
	      (seg_load_addr + seg_hdr.p_memsz));
	*simulated_brk_ptr = (seg_load_addr + seg_hdr.p_memsz);
      }

      if(!load_base_set) {
	DEBUG(DEBUG_LOADER,"'%s' loading at %08X.",filename,seg_load_addr);
	load_base = seg_load_addr;
	if(load_lib) load_offset = seg_load_addr - seg_hdr.p_vaddr;

	load_base_set = true;  // don't want to set it twice
      }
    }
  }

  // if we are dynamicly linked, we need to provide some info...
  if(keep_eip) {
    dlinfo[ 0] = htox32(ELFAUX_PHDR);
    dlinfo[ 1] = htox32(load_base + elf_hdr.e_phoff);
    dlinfo[ 2] = htox32(ELFAUX_PHENT);
    dlinfo[ 3] = htox32(elf_hdr.e_phentsize);
    dlinfo[ 4] = htox32(ELFAUX_PHNUM);
    dlinfo[ 5] = htox32(elf_hdr.e_phnum);
    dlinfo[ 6] = htox32(ELFAUX_PAGESZ);
    dlinfo[ 7] = htox32(PAGE_SIZE);
    dlinfo[ 8] = htox32(ELFAUX_BASE);
    dlinfo[ 9] = htox32(interp_load_base);
    dlinfo[10] = htox32(ELFAUX_FLAGS);
    dlinfo[11] = htox32(0);
    dlinfo[12] = htox32(ELFAUX_ENTRY);
    dlinfo[13] = htox32(elf_hdr.e_entry);
    dlinfo[14] = htox32(ELFAUX_UID);
    dlinfo[15] = htox32(getuid());          
    dlinfo[16] = htox32(ELFAUX_EUID);
    dlinfo[17] = htox32(geteuid());
    dlinfo[18] = htox32(ELFAUX_GID);
    dlinfo[19] = htox32(getgid());
    dlinfo[20] = htox32(ELFAUX_EGID);
    dlinfo[21] = htox32(getegid());
    dlinfo[22] = htox32(ELFAUX_NULL);
    dlinfo[23] = htox32(0);
  }

  current_sp = initial_sp;
  if(PutArglistsOnStack(filename,argv,envp,
			dlinfo,(keep_eip ? sizeof(dlinfo) : 0),
			cpu,current_sp) < 0) {
    DEBUG(DEBUG_LOADER,"Could copy arguments onto stack for '%s'.",filename);
    close(exec_fd);
    return(-1);
  }

  // set up relevant registers
  regptr = cpu.GetRegisters();
  REG32(*regptr,REG_EDX) = 0;          // libc stuff wants this to be 0
  REG32(*regptr,REG_ESP) = current_sp;

  // hack here - libs should leave their load base in ECX
  if(load_lib)
    REG32(*regptr,REG_ECX) = load_offset;

  // only stick in our entry point if we didn't load an interpreter
  if(!keep_eip)
    REG32(*regptr,REG_EIP) = elf_hdr.e_entry + load_offset;

  // we need to clobber the segment descriptors to get them to point at all
  //  of our memory space...
  seg_descs = cpu.GetDescriptors();

  seg_descs[SDESC_ES ].set_base_limit(0x00000000,0x0BFFFF,
				      (0x0C0E0 | DESC_DATA_RW_ACC));
  seg_descs[SDESC_CS ].set_base_limit(0x00000000,0x0BFFFF,
				      (0x0C0E0 | DESC_CODE_ER_ACC));
  seg_descs[SDESC_SS ].set_base_limit(0x00000000,0x0BFFFF,
				      (0x0C0E0 | DESC_DATA_RW_ACC));
  seg_descs[SDESC_DS ].set_base_limit(0x00000000,0x0BFFFF,
				      (0x0C0E0 | DESC_DATA_RW_ACC));
  seg_descs[SDESC_FS ].set_base_limit(0x00000000,0x0BFFFF,
				      (0x0C0E0 | DESC_DATA_RW_ACC));
  seg_descs[SDESC_GS ].set_base_limit(0x00000000,0x0BFFFF,
				      (0x0C0E0 | DESC_DATA_RW_ACC));

  return(0);
}
