/* -*- C++ -*-
 * $Id: load_elf.h,v 1.4 1998/04/29 04:54:54 nikos Exp $
 * 
 * RCS Log:
 * $Log: load_elf.h,v $
 * Revision 1.4  1998/04/29 04:54:54  nikos
 * Added new AranaTrace command "AranaTrace  filter_thru_L1" that dumps
 * only instructions with references (pc + ea) that miss in the L1 cache.
 * MBs are always dumped.
 *
 * Revision 1.4  1998/03/25 19:16:46  nikos
 * atf_tlb_fixed
 *
 * Revision 1.3  1998/03/13  17:12:30  nikos
 * Fix vaddr->physaddr translation bug and tlb behavior in ATF.
 * Changes affect only people who use ATF for their traces.
 *
 * Revision 1.3  1998/03/13 16:37:05  nikos
 * .
 *
 * Revision 1.2  1998/02/16  23:20:03  nikos
 * Fixed AranaTrace command which was broken with the last commit.
 * Macros in external_trace.h are set such that the name "arana"
 * appears only in arana_trace.[ch] and no-where else!!!
 * In the previous revision, kappa.c has not been cleaned up.
 *
 * If you want to use the AranaTrace command,
 * compile SimOS with the following variables set :
 *
 * EXTERNAL_TRACE=ARANA_TRACE
 * ARANA_TRACE_INC_DIR=/perf_sym/ATF/atf
 * ARANA_TRACE_LIB=/perf_sym/ATF/atf/libatf.a
 * ARANA_TRACE_OBJS=arana_trace.o
 * ARANA_TRACE_HEADER=arana_trace.h
 *
 * This commit reports a bunch of modified files which are not
 * really modified (only some rcs fields like $Id: load_elf.h,v 1.4 1998/04/29 04:54:54 nikos Exp $, $Header: /wrl/proj/simos/cvsroot/simulation/cpus-x86/x86sim/load_elf.h,v 1.4 1998/04/29 04:54:54 nikos Exp $, etc).
 *
 * Revision 1.2  1998/02/11 14:24:55  nikos
 * .
 *
 * Revision 1.1  1997/12/16  01:28:28  bugnion
 * stanford update (x86)
 *
 * Revision 1.2  1997/09/24  08:11:31  sean
 * LoadELFBinary can keep track of the highest address loaded to, which is
 *  the 'brk' point for a Unix process.
 *
 * Revision 1.1  1997/03/18 19:22:32  sean
 * Initial revision
 *
 */

#ifndef _LOAD_ELF_H
#define _LOAD_ELF_H

#include "ints.h"

class CPU;

#define ELF_IDENT_LEN    16

// for now, we'll only handle 32-bit, current version
//  both little and big endian binaries are allowed though
#define ELF_SIG_LE       "\177ELF\001\001\001"
#define ELF_SIG_BE       "\177ELF\001\002\001"
#define ELF_SIGLEN        7

// relevant constants
#define ELFTYPE_EXEC      2
#define ELFTYPE_DYNAMIC   3
#define ELFMACHINE_i386   3
#define ELFMACHINE_i486   6
#define ELFSEG_LOAD       1  
#define ELFSEG_INTERP     3

/* Symbolic values for the entries in the auxiliary table
   put on the initial stack - stolen from /usr/include/linux/elf.h */
#define ELFAUX_NULL   0     /* end of vector */
#define ELFAUX_IGNORE 1     /* entry should be ignored */
#define ELFAUX_EXECFD 2     /* file descriptor of program */
#define ELFAUX_PHDR   3     /* program headers for program */
#define ELFAUX_PHENT  4     /* size of program header entry */
#define ELFAUX_PHNUM  5     /* number of program headers */
#define ELFAUX_PAGESZ 6     /* system page size */
#define ELFAUX_BASE   7     /* base address of interpreter */
#define ELFAUX_FLAGS  8     /* flags */
#define ELFAUX_ENTRY  9     /* entry point of program */
#define ELFAUX_NOTELF 10    /* program is not ELF */
#define ELFAUX_UID    11    /* real uid */
#define ELFAUX_EUID   12    /* effective uid */
#define ELFAUX_GID    13    /* real gid */
#define ELFAUX_EGID   14    /* effective gid */

typedef struct {
  int8   e_ident[ELF_IDENT_LEN]; // ELF file identifier (starts w/ '\177ELF')
  int16  e_type;                 // type of ELF file (e.g. ELFTYPE_EXEC)
  int16  e_machine;              // machine file runs on (e.g. ELFMACHINE_i386)
  int32  e_version;              // version of file
  int32  e_entry;                // entry point (initial EIP)
  int32  e_phoff;                // offset in file of segment headers
  int32  e_shoff;                // offset in file of symbol(?) headers
  int32  e_flags;                // flags
  int16  e_ehsize;               // size of this header (according to file)
  int16  e_phentsize;            // size of a segment header
  int16  e_phnum;                // number of segment headers in file
  int16  e_shentsize;            // size of a symbol(?) header
  int16  e_shnum;                // number of symbol(?) headers in file
  int16  e_shstrndx;
} ELF32_Header;

typedef struct {
  int32  p_type;                 // Type of segment (e.g. ELFSEG_LOAD)
  int32  p_offset;               // Start of segment data in file
  int32  p_vaddr;                // Virtual dest. address in memory
  int32  p_paddr;                // Physical address(?)
  int32  p_filesz;               // Size of segment data in file
  int32  p_memsz;                // Size of segment in memory (>= filesz?)
  int32  p_flags;                // Segment flags (e.g. ELFSEGFL_CAN_READ)
  int32  p_align;                // Segment alignment (not used?)
} ELF32_SegmentHeader;

// This weird looking function type is used by 'LoadELFBinary' to make sure
//   memory exists at proper places (this can be fixed by messing with the
//   CPU's page mapping, or calling something like mmap() to create memory
//   at the right place
// 'base' is the base needed, if 'fixed' is true.  (If 'fixed' is false, then
//   'base' can be considered a hint, but it does not need to be followed.)
//   'size' is the necessary size of the block, and 'param' is a parameter 
//   passed by the person who gave the pointer to this function in case the
//   function needs state.
// On success, the function returns 0 and fixes 'base' to be the correct base.
//   On failure, the function returns < 0, and shouldn't change 'base'.
typedef int (*MemCreateFunc)(int32& base, int32 size, bool fixed,
			     void *param);

// This is the magic function that attempts to load the binary in 'filename'
//  onto the 'cpu'.
// Arguments:
//     cpu          - the CPU on which to load the program
//     filename     - name of the executable to load
//     load_offset  - amount to add to all load addresses (shifts image in
//                      memory, but not pointers in code, so don't make this
//                      non-zero unless you know what you're doing)
//     initial_sp   - initial top of the stack for the new program (it is
//                      assumed that space below this address has already been
//                      allocated - i.e. the 'create_fn' will NOT be called for
//                      stack addresses)
//     argv         - the arglist that should be given to the program as 'argv'
//     envp         - the arglist that should be given to the program as 'envp'
//     create_fn    - the function that should be used to make sure memory
//                      exists at a given address (can be 0 if no fn should be
//                      called)
//     create_param - the parameter that should be passed to the 'create_fn'
//                      any time it is called
//     load_lib     - true if we are expecting to load a dynamic library;
//                      false if we are expecting to load an executable
//     simulated_brk_ptr - if non-null, the brk value (first unused byte in
//                           in memory) of the simulated program is stored
// Return value:  This function returns 0 if 'filename' was loaded 
//   successfully, and a number < 0 if errors occurred.
int LoadELFBinary(CPU& cpu, const char *filename, 
		  int32 load_offset, int32 initial_sp,
		  const char *argv[], const char *envp[],
		  MemCreateFunc create_fn, void *create_param, bool load_lib,
		  int32 *simulated_brk_ptr);

#endif
