/* $Id: lockstep.cc,v 1.4 1998/04/29 04:54:54 nikos Exp $
 *
 * RCS Log:
 * $Log: lockstep.cc,v $
 * Revision 1.4  1998/04/29 04:54:54  nikos
 * Added new AranaTrace command "AranaTrace  filter_thru_L1" that dumps
 * only instructions with references (pc + ea) that miss in the L1 cache.
 * MBs are always dumped.
 *
 * Revision 1.4  1998/03/25 19:16:46  nikos
 * atf_tlb_fixed
 *
// Revision 1.3  1998/03/13  17:12:31  nikos
// Fix vaddr->physaddr translation bug and tlb behavior in ATF.
// Changes affect only people who use ATF for their traces.
//
 * Revision 1.3  1998/03/13 16:37:05  nikos
 * .
 *
// Revision 1.2  1998/02/16  23:20:03  nikos
// Fixed AranaTrace command which was broken with the last commit.
// Macros in external_trace.h are set such that the name "arana"
// appears only in arana_trace.[ch] and no-where else!!!
// In the previous revision, kappa.c has not been cleaned up.
//
// If you want to use the AranaTrace command,
// compile SimOS with the following variables set :
//
// EXTERNAL_TRACE=ARANA_TRACE
// ARANA_TRACE_INC_DIR=/perf_sym/ATF/atf
// ARANA_TRACE_LIB=/perf_sym/ATF/atf/libatf.a
// ARANA_TRACE_OBJS=arana_trace.o
// ARANA_TRACE_HEADER=arana_trace.h
//
// This commit reports a bunch of modified files which are not
// really modified (only some rcs fields like $Id: lockstep.cc,v 1.4 1998/04/29 04:54:54 nikos Exp $, $Header: /wrl/proj/simos/cvsroot/simulation/cpus-x86/x86sim/lockstep.cc,v 1.4 1998/04/29 04:54:54 nikos Exp $, etc).
//
 * Revision 1.2  1998/02/11 14:24:55  nikos
 * .
 *
// Revision 1.1  1997/12/16  01:28:28  bugnion
// stanford update (x86)
//
// Revision 1.5  1997/09/27  18:41:51  sean
// Added ability to ignore some flags in lockstep comparison.
// Also added call to read data from lockstep'd process's memory space.
//
 * Revision 1.4  1997/09/24 08:10:41  sean
 * Tweaked call to VerifyRegisters to know what instruction caused the
 * mismatch.
 *
 * Revision 1.3  1997/09/15 19:49:27  sean
 * tweaked register verification code to allow correction of mismatches
 *  rather than just aborting.  Copying from sim-land to real-land
 *  does not work yet, and it'd be nice if a mismatch killed the repetition
 *  of a command...
 *
 * Revision 1.2  1997/09/06 00:39:51  yoel
 * Added LockStep support and some display.
 *
 * Revision 1.1  1997/04/13 23:37:31  yoel
 * Initial revision
 *
 */

static char rcsid[] __attribute__((unused))
        = "$Id: lockstep.cc,v 1.4 1998/04/29 04:54:54 nikos Exp $";

/*  Lockstep.cc, a simple program using ptrace() to run a program a single
 * instruction at a time and dump checkpoints for comparison purposes.
 *
 *  Author Jeremy Sugerman, yoel@Stanford.EDU, 3 April 1997
 */
#include <stdio.h>
#include <unistd.h>
#include <signal.h>
#include <errno.h>

#include <sys/types.h>
#include <sys/wait.h>

#include "lockstep.h"
#include "x86sim.h"
#include "cpu.h"
#include "debug.h"
#include "instruction.h"

// this must go AFTER instruction.h (instruction.h wants to define functions
//  using register names, which ptrace.h #defines away...
#include <sys/ptrace.h>

/* From values in /usr/include/asm/ptrace.h */
const char *LockStep::register_names[] = {
  "EBX", "ECX", "EDX", "ESI", "EDI", "EBP", "EAX", 
  "DS", "ES", "FS", "GS", "ORIG_EAX", "EIP", "CS", "EFL", "ESP", "SS"
};

//  Order to use when translating to Simulator order.  It might be worth
// reconciling registers.h with /usr/include/asm/ptrace.h to avoid this.
const int LockStep::sim_order[] = {
  REG_EBX, REG_ECX, REG_EDX, REG_ESI, REG_EDI, REG_EBP, REG_EAX, REG_DS,
  REG_ES, REG_FS, REG_GS, -1, REG_EIP, REG_CS, REG_EFLAGS, -1, REG_SS
};

/* Order to display.  This is to match what x86sim uses */
const int LockStep::reg_col1[6] = { CS, DS, ES, FS, GS, SS };
const int LockStep::reg_col2[6] = { EIP, EFL, EAX, EBX, ECX, EDX };
const int LockStep::reg_col3[6] = { -1, -1, UESP, EBP, ESI, EDI };

/*  Okay, to facilitate returning errors the constructor is really lame and
 * there's a static initializer LockStep::Init() that can return NULL if
 * something goes wrong.
 */
LockStep::LockStep(int _pid)
{
  pid = _pid;
}

/*  PTRACE_PEEKUSR all of the registers and return them */
int LockStep::GetRegisters(int32 registers[])
{
  int regno;

  for (regno = 0; regno < 17; ++regno) {
     errno = 0;
     registers[regno] = 
           ptrace(PTRACE_PEEKUSR, pid, regno * sizeof(int), 0);

     if ((registers[regno] == (unsigned) -1) && (errno != 0)) {
       perror("PEEKUSR");
       return -1;
     }
  }
  return 0;
}

int LockStep::ReadFromLockstepMemory(int32 address, void *buffer, int len)
{
  int val;
  int ofs;

  if((ofs = (address & 3)) != 0) {
    val = ptrace(PTRACE_PEEKTEXT, pid, address - ofs, 0);
    if((val == -1) && (errno != 0)) return(0);

    memcpy(buffer, ((void *)&val) + ofs, MIN(4 - ofs, len));
    address += ofs;
    buffer += ofs;
    len -= ofs;
  }

  while(len >= 4) {
    val = ptrace(PTRACE_PEEKTEXT, pid, address, 0);
    if((val == -1) && (errno != 0)) return(0);

    memcpy(buffer, &val, 4);
    address += 4;
    buffer += 4;
    len -= 4;
  }

  if(len > 0) {
    val = ptrace(PTRACE_PEEKTEXT, pid, address, 0);
    if((val == -1) && (errno != 0)) return(0);

    memcpy(buffer, &val, len);
  }

  return(1);
}

int LockStep::MakeRegisterFile(RegisterFile *sim_registers)
{
  int regno;
  int32 registers[17];

  // Fetch traced registers
  if (GetRegisters(registers) == -1) return -1;

  // Mask 0x100 _out_ of EFL (it's the TRAP flag)
  registers[EFL] &= ~(0x100U);

  //  Copy them into the RegisterFile, Excepting ORIG_EAX (which has no
  // meaning) and ESP (which is somewhat sacred...)
  for (regno = 0; regno < 17; regno++) {
    if (sim_order[regno] == -1) continue;
    REG32(sim_registers, sim_order[regno]) = registers[regno];
  } 

  return 0;
}

int LockStep::TraceRegisters(void)
{
  int regno;
  int32 registers[17];

  // Fetch the registers and make sure we get them okay.
  if (GetRegisters(registers) == -1) return -1;

  // Mask 0x100 _out_ of EFL (it's the TRAP flag)
  registers[EFL] &= ~(0x100U);

  // Print them out in an order matching x86sim
  for (regno = 0; regno < 6; regno++) {
    fprintf(stderr, "%s: %04X   %s: %08X", 
               register_names[reg_col1[regno]], registers[reg_col1[regno]],
	       register_names[reg_col2[regno]], registers[reg_col2[regno]]);
    if (reg_col3[regno] != -1)
      fprintf(stderr, "   %s: %08X",
               register_names[reg_col3[regno]], registers[reg_col3[regno]]);
    fprintf(stderr, "\n");
  }

  fprintf(stderr, "\n");
  return 0;
}

/*  Forks and sets up one copy to be traced.  If the exec() works, will
 * never return in the child and will return the child's pid in the parent.
 * Note that I could get cmd from argv[0], but some people like to
 * masquerade as other things (not that I would ever do such a thing...).
 * Note that the parent waits for the child to exec and be stopped before
 * returning.  This way, the child is _always_ stopped when a LockStep::
 * call returns.
 */
LockStep *LockStep::Init(const char *cmd, char *argv[])
{
  int pid;
  int status;

  switch(pid = fork()) {
  case -1:
    perror("fork");
    return NULL;
    break;
  case 0:
    /*  TRACE_ME indicates that the next time we call exec, we should be
     * stopped and our parent (who will trace us) should be awakened.
     */
    if (ptrace(PTRACE_TRACEME, 0, 0, 0) == -1) {
      perror("PTRACE_TRACEME");
      return NULL;
    }
    execv(cmd, argv);
    return NULL;				// If the exec() fails.

  default:
    if (waitpid(pid, &status, 0) == -1) {
      perror("waitpid");
      return NULL;
    }

    if (WIFEXITED(status)) {
      fprintf(stderr, "Child exited with code %d\n", WEXITSTATUS(status));
      return NULL;
    }
    return new LockStep(pid);
  }
}

/*  Allow process pid to execute exactly one instruction.  Assumes pid is
 * stopped.  Returns -1 on failure, 0 on exit, 1 on normal iteration. 
 */
int LockStep::SingleStep(void)
{
   int status;

   if (ptrace(PTRACE_SINGLESTEP, pid, 0, 0) == -1) {
     fprintf(stderr, "Status: %d\n", status);
     perror("PTRACE_SINGLESTEP");
     return -1;
   }

   if (waitpid(pid, &status, 0) == -1) {
     perror("waitpid");
     return -1;
   }

   if (WIFEXITED(status)) {
     fprintf(stderr, "Child exited with code %d\n", WEXITSTATUS(status));
     return 0;
   }
   return 1;
}

//  Whee, this function clones the stack of a running process.  This could
// be interesting...
int LockStep::CopyStack(CPU& cpu)
{
  int32 esp, len;
  int32 *new_stack;
  int32 i;

  errno = 0;
  esp = ptrace(PTRACE_PEEKUSR, pid, UESP * sizeof(int), 0);
  if ((esp == (unsigned) -1) && (errno != 0)) {
       perror("PEEKUSR");
       return -1;
  }

  // Four byte align.  Since the stack is zero below initialized data this
  // works and saves dealing with the boundary.  Round down.
  esp = esp & ~3L;
  len = (DEFAULT_LINUX_STACK_TOP - esp)>>2;
  new_stack = new int32[len];
  
  //  Slurp the old stack
  for (i = esp; i < DEFAULT_LINUX_STACK_TOP; i += 4) {
    new_stack[(i - esp)>>2] = ptrace(PTRACE_PEEKTEXT, pid, i, 0);
    if ((new_stack[i] == (unsigned) -1) && (errno != 0)) {
      perror("PEEKTEXT");
      return -1;
    }
  }

  // Change - check the stack first to see if it's ok - if not, copy
  if(!cpu.CompareStacks(esp, len, new_stack)) {
    DEBUG(DEBUG_LOADER, 
	  "Stack created by LoadELFBinary() is wrong - using lockstep stack.");

    //  Pass the stack to the cpu and let it do the copying since it's all
    // internal state.
    if (cpu.InitStack(esp, len, new_stack) == -1) {
      fprintf(stderr, "cpu.InitStack() failed!\n");
      delete[] new_stack;

      return -1;
    }
  }

  delete [] new_stack;
  return 0;
}

void LockStep::MismatchAction(int32 *sim_value, int real_regnum, 
			      int32 real_value)
{
  char line[80];

  while(1) {
    printf("Action? (c, <, >, q, ?=Help) ");

    // if we get EOF, pretend it's a Quit
    if(!fgets(line, 79, stdin)) strcpy(line, "q\n");

    switch(line[0]) {
    case 'c':  // continue
      return;
      break;

    case '<':  // copy real to simulated
      *sim_value = real_value;
      return;
      break;

    case '>':  // copy simulated to real
      fprintf(stderr, "Not yet...\n");
      ASSERT(0);
      return;
      break;

    case 'q':  // quit (Assert-fail)
      ASSERT(0);
      break;

    case '?':  // help
      printf(" c = continue, < = copy real->sim, > = copy sim->real\n");
      printf(" q = quit simulator\n");
      break;

    default:
      printf("invalid command\n");
      break;
    }
  }
}

int LockStep::VerifyRegisterFile(RegisterFile *sim_registers,
				 const Instruction *last_instr,
				         /* the one that broke */
				 int32 flag_ignore_mask 
			                 /* flags to not compare */)
{
  int regno;
  int32 cur_register;
  int32 registers[17];

  // Fetch traced registers
  if (GetRegisters(registers) == -1) return -1;

  // always ignore RF (resume flag) for now
  flag_ignore_mask |= (1 << FLAG_RF);

  // Mask 0x100 _out_ of EFL (it's the TRAP flag)
  registers[EFL] &= ~(0x100U);

  // Now do the compare...
  for (regno = 0; regno < 17; regno++) {
    if (sim_order[regno] == -1) continue;
    if (sim_order[regno] == REG_EFLAGS) continue;  // we compare flags below

    //  Use a silly temp register b/c it makes it _so_ much easier to view
    // under the debugger.
    cur_register = REG32(sim_registers, sim_order[regno]);
    if (cur_register != registers[regno]) {
      fprintf(stderr, "MISMATCH: Register %s Simulated: %X vs. Traced: %X\n",
                       register_names[regno], cur_register, registers[regno]);
      fprintf(stderr, "MISMATCH: EIP Simulated: %X / Traced: %X\n",
                       REG32(sim_registers, REG_EIP), registers[EIP]);

      if(last_instr != NULL) {
	fprintf(stderr, "MISMATCH: after ");
	last_instr->PrintSelf(stderr);
	fprintf(stderr, "\n");
      }

      MismatchAction(&REG32(sim_registers, sim_order[regno]), regno,
		     registers[regno]);
      //      ASSERT(0);
    }
  }

  // check the flag word
  if((REG32(sim_registers, REG_EFLAGS) ^ registers[EFL]) & ~flag_ignore_mask) {
    fprintf(stderr, "MISMATCH: Flags Simulated: %X(%X) vs. Traced: %X(%X)\n",
	             (REG32(sim_registers, REG_EFLAGS) & ~flag_ignore_mask),
                     REG32(sim_registers, REG_EFLAGS), 
	             (registers[EFL] & ~flag_ignore_mask),
	             registers[EFL]);
    fprintf(stderr, "MISMATCH: EIP Simulated: %X / Traced: %X\n",
                     REG32(sim_registers, REG_EIP), registers[EIP]);

    if(last_instr != NULL) {
      fprintf(stderr, "MISMATCH: after ");
      last_instr->PrintSelf(stderr);
      fprintf(stderr, "\n");
    }

    MismatchAction(&REG32(sim_registers, REG_EFLAGS), EFL,
		   registers[EFL]);
    //    ASSERT(0);
  }

  if (REG32(sim_registers, REG_ESP) != registers[UESP]) {
    fprintf(stderr, "MISMATCH: Register ESP Simulated: %X vs. Traced: %X\n",
                     REG32(sim_registers, REG_ESP), registers[UESP]);
    fprintf(stderr, "MISMATCH: EIP Simulated: %X / Traced: %X\n",
                     REG32(sim_registers, REG_EIP), registers[EIP]);

    if(last_instr != NULL) {
      fprintf(stderr, "MISMATCH: after ");
      last_instr->PrintSelf(stderr);
      fprintf(stderr, "\n");
    }

    MismatchAction(&REG32(sim_registers, REG_ESP), UESP,
		   registers[UESP]);
    //    ASSERT(0);
  }

  // finally, copy in values of don't-care bits from lockstep'd process so
  //  that flags match for next instruction
  REG32(sim_registers, REG_EFLAGS) = 
    (REG32(sim_registers, REG_EFLAGS) & ~flag_ignore_mask) |
    (registers[EFL] & flag_ignore_mask);

  return 0;
}

