/* -*- C++ -*-
 * This file is compatible with both C and C++
 *
 * $Id: simlib.h,v 1.4 1998/04/29 04:54:59 nikos Exp $
 * 
 * RCS Log:
 * $Log: simlib.h,v $
 * Revision 1.4  1998/04/29 04:54:59  nikos
 * Added new AranaTrace command "AranaTrace  filter_thru_L1" that dumps
 * only instructions with references (pc + ea) that miss in the L1 cache.
 * MBs are always dumped.
 *
 * Revision 1.4  1998/03/25 19:16:50  nikos
 * atf_tlb_fixed
 *
 * Revision 1.3  1998/03/13  17:12:35  nikos
 * Fix vaddr->physaddr translation bug and tlb behavior in ATF.
 * Changes affect only people who use ATF for their traces.
 *
 * Revision 1.3  1998/03/13 16:37:10  nikos
 * .
 *
 * Revision 1.2  1998/02/16  23:20:09  nikos
 * Fixed AranaTrace command which was broken with the last commit.
 * Macros in external_trace.h are set such that the name "arana"
 * appears only in arana_trace.[ch] and no-where else!!!
 * In the previous revision, kappa.c has not been cleaned up.
 *
 * If you want to use the AranaTrace command,
 * compile SimOS with the following variables set :
 *
 * EXTERNAL_TRACE=ARANA_TRACE
 * ARANA_TRACE_INC_DIR=/perf_sym/ATF/atf
 * ARANA_TRACE_LIB=/perf_sym/ATF/atf/libatf.a
 * ARANA_TRACE_OBJS=arana_trace.o
 * ARANA_TRACE_HEADER=arana_trace.h
 *
 * This commit reports a bunch of modified files which are not
 * really modified (only some rcs fields like $Id: simlib.h,v 1.4 1998/04/29 04:54:59 nikos Exp $, $Header: /wrl/proj/simos/cvsroot/simulation/cpus-x86/x86sim/simlib.h,v 1.4 1998/04/29 04:54:59 nikos Exp $, etc).
 *
 * Revision 1.2  1998/02/11 14:24:55  nikos
 * .
 *
 * Revision 1.1  1997/12/16  01:28:30  bugnion
 * stanford update (x86)
 *
 * Revision 1.6  1997/12/07  02:54:19  sjt
 * Bunches of changes:
 *   1) Rewritten emulation of FPU ops is half-done.  Emulation is now done in
 *        the context of the simulator (although the FPU context of the simulated
 *        program is used), which dodges virtual->phyiscal mappings of addresses.
 *      Assembler has been written to be happy under fairly aggressive (i.e. it
 *        ignores the "volatile" keyword) optimization by g++.
 *      FPU operations using memory operands are unsupported, and ignored.
 *
 *   2) Minor changes in CPU::MUL and CPU::IMUL to support 64-bit products (you
 *        have to explicitly cast the int32's to int64's before the multiply, or
 *        you get 0's for the top 32 bits (even though they're available for free
 *        - seems g++ code gen could be written better).
 *
 *   3) Changes all over to get task switching to work.  Need to fix bunches of
 *        logic errors, uses of limit() instead of offset_valid(), make TSS's
 *        writeable under certain circumstances.
 *
 * Revision 1.5  1997/10/29 22:47:45  sean
 * Changed SimulateLinuxSystemCall() to return whether or not CPU should be
 *   halted.
 * Added ExecuteBuffer() function to simulate FPU ops on the real FPU.
 *
 * Revision 1.4  1997/09/24 08:13:07  sean
 * brk and exit (sorta) system calls are wrappered, allowing better simulation
 *  of native code
 *
 * Revision 1.3  1997/03/18 19:10:03  sean
 * armored Linux-only stuff with #ifdef NATIVE_X86_HOST
 * moved the scary CopyAndSwitchCode() and CopyAndSwitchStacks() code in, so
 *   they could be with the rest of the evil assembler code
 *
 * Revision 1.2  1997/03/15 02:18:20  yoel
 * Changes start_sim() prototype to take a char *.
 *
 * Revision 1.1  1997/03/07 05:55:13  sean
 * Initial revision
 *
 *
 */

#ifndef _SIMLIB_H
#define _SIMLIB_H

/* neat trick here...  extern "C"'ing the procedure gives it a C-like 
   name, so that C programs can use what would otherwise be C++ code */

#ifdef __cplusplus
extern "C" { 
#endif

#include "ints.h"

/* The link-in simulator only exists if we are a native Linux host -
     if not, we just comment out the entire functions */
#ifdef NATIVE_X86_HOST
// we need to keep track of the simulated program's 'brk' value so that
//  we can emulate the brk system call (otherwise the simulator's brk
//  value gets used.
extern int32 simulated_brk;
extern int32 simulated_min_brk;  // smallest brk is allowed to be (end of text)

/* This function takes care of starting the simulator - it's a fairly weird
   concept, but when this function only returns in the simulated-world - the
   "real" program stays in this function.

   The return value is a void * that can be passed to stop_sim() in order to
   stop the simulator, or 0 if the simulator was not started. */
void *start_sim(const char *debug_codes); 

/* This function is used to stop the simulator.  Essentially it wakes up the
   "real" program, which was stuck in the start_sim() call, and switches it
   so that it returns from this call, and stops the simulator in the process.
   */
void stop_sim(void *sim);
#endif

/* SimulatateLinuxSystemCall exists if we are on a non-native host, but it'll
    ASSERT-fail.
   Returns non-zero if the system call (or its emulation) calls for an end to
    simulation (the idea is that the CPU will halt in such a case) */
int SimulateLinuxSystemCall(void *registers_ptr);

#ifdef NATIVE_X86_HOST
/* These functions allow on-the-fly relocation of the code and stack segments.
   CopyAndSwitchStacks() appears to be relatively stable, and needed.
   CopyAndSwitchCode() has some serious problems with pointers to the code
     segment inside of the code segment (binaries do not have relocation info).
     Luckily, if you can convince 'ld' to link the binary somewhere else to
     start with, you never need to call CopyAndSwitchCode().  (I've mostly
     included it as an example of the sick things you can do in assembler. :)
*/
void CopyAndSwitchCode(int32 old_code_base, int32 new_code_base, int32 size,
		       int32 stack_top);
void CopyAndSwitchStacks(int32 old_stack_top, int32 new_stack_top);

int ExecuteBuffer(const void *buffer, int len,
		  void *registers_ptr,
		  void *fpu_state_ptr);

int ExecuteFPUOp(const void *buffer, int len,
		 void *fpu_state_ptr,
		 int32& eax_to_use);
#endif

#ifdef __cplusplus
}
#endif

#endif
