/*
 * Copyright (C) 1996 by the Board of Trustees
 *    of Leland Stanford Junior University.
 * 
 * This file is part of the SimOS distribution. 
 * See LICENSE file for terms of the license. 
 *
 */

/*****************************************************************
 * simos_interface.c
 *
 * This is the interface from simos into the X86SIM  cpu simulator.
 * This file will handle the transision between the C++ x86sim and
 * the rest of SimOS that is written in C. 
 *
 * $Author: nikos $
 * $Date: 1998/04/29 04:54:59 $
 *****************************************************************/ 

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <signal.h>

extern "C" { 
/* SimOS header files assume to be in C. */
#include "machine_params.h"
#include "simutil.h"
#include "cpu_stats.h"
#include "cpu_state.h"
#include "sim_error.h"
#include "eventcallback.h"
#include "cpu_interface.h"
#include "simmisc.h" 
#include "annotations.h"
#include "registry.h"
#include "simos_interface.h"
#include "hw_events.h"
#include "gdb_interface.h"
#include "iodevs.h"
#include "iospace.h"
#include "intr.h"
#include "sim_error.h"
}


/* X86sim header files are written in C++ */
#include "cpu.h"

#define X86SIM_NOBREAK      (-2)
#define X86SIM_BREAKANYCPU  (-1)

int x86sim_debug_mode = 0;
int x86sim_break_nexti = 0;
int x86sim_sigusr = 0;;


/* Global Functions */

extern CPUState *SBase;  /* SimOS's CPU state */

/* Local Fuonctions */
static void ReadConfiguration(void);
static void InitStats(int cpuNum);
static void CopyOutState(CPU *cpu, int cpuNum);
static void CopyInState(CPU *cpu, int cpuNum);
static void X86SimRun(void);
static void X86CPUVectorInit(void);
Result X86TranslateVirtualNoSE(int cpunum, VA vAddr, PA *pAddr);

static char *X86Stats(void)
{
   return "";
}

CPUState *currentCPU;

/*****************************************************************
 * X86SimEnter
 *
 * This is the main interface into setting up X86Sim to run. It
 * should be called from simos to set up registers, tlb, cache,
 * memory, etc. 
 *****************************************************************/

void
X86SimEnter(int swtch)
{
   static int initialized = 0;
   CPU   *cpu;

   if (TOTAL_CPUS != 1) CPUError("X86Sim is uniprocessor only now\n!");

   X86CPUVectorInit();

   if (!initialized) {       
      initialized = 1;
      cpu = new CPU(SBase[0].memoryPtr, SBase[0].memSize);
      ASSERT(cpu);
      SBase[0].x86simCpu = cpu;
      AnnCommonSetup();
      AnnFMInit(128);
      HWEventsLateInit();
   }

   cpu = (CPU *) SBase[0].x86simCpu;

   /* Load the CPU state into X86Sim */

   CopyInState(cpu,0);

  /*
   * simos enter annotation
   */
  {
     int cpuNum;
     for (cpuNum = 0;cpuNum < TOTAL_CPUS; cpuNum++) {
	currentCPU = &SBase[cpuNum];
	AnnExec(AnnFind("simos","enter"));
     }
  }
  currentCPU = &SBase[0]; 
  currentCPU->cpuStatus = cpu_running;
  ASSERT(NUM_CPUS(0) == 1);  
/*  InitStats(0); */
   /* 
    * Jump into X86Sim 
    */ 
   X86SimRun();

#ifdef notdef
   /* Right now we dump everything when switching out of X86Sim */
   X86SimDumpAllStats();

   CPUPrint("***************  %lld  final annotations ****************\n",
            (uint64) X86SimReadTime(0));
#endif
   AnnExec(AnnFind("simos","periodic"));
}


/*
 * CopyInState - Copy the CPU state from the SimOS format into
 * the CPU class of X86Sim.
 */
static void
CopyInState(CPU *cpu, int cpuNum) 
{
  CPUState *P = SBase + cpuNum;
  RegisterFile *regs;

  regs = cpu->GetRegisters();
  REG32(regs, REG_EAX) = P->eax;
  REG32(regs, REG_ECX) = P->ecx;
  REG32(regs, REG_EDX) = P->edx;
  REG32(regs, REG_EBX) = P->ebx;
  REG32(regs, REG_ESP) = P->esp;
  REG32(regs, REG_EBP) = P->ebp;
  REG32(regs, REG_ESI) = P->esi;
  REG32(regs, REG_EDI) = P->edi;

  REG32(regs, REG_EFLAGS) = P->eflags;
  REG32(regs, REG_CR0) = P->cr[0];
  REG32(regs, REG_CR1) = P->cr[1];
  REG32(regs, REG_CR2) = P->cr[2];
  REG32(regs, REG_CR3) = P->cr[3];
  REG32(regs, REG_CR4) = P->cr[4];

  REG16(regs, REG_ES) = P->es.selector;
  REG16(regs, REG_CS) = P->cs.selector;
  REG16(regs, REG_SS) = P->ss.selector;
  REG16(regs, REG_DS) = P->ds.selector;
  REG16(regs, REG_FS) = P->fs.selector;

  REG16(regs, REG_TR) = P->tr.selector;
  REG16(regs, REG_LDT) = P->ldtr.selector;

  REG32(regs, REG_EIP) = P->eip;

  SegmentDescriptor *desc = cpu->GetDescriptors();

  if (P->es.descriptor)
     desc[SDESC_ES].extract_descriptor(&P->es.descriptor); 
  if (P->cs.descriptor)
     desc[SDESC_CS].extract_descriptor(&P->cs.descriptor); 
  if (P->ss.descriptor)
     desc[SDESC_SS].extract_descriptor(&P->ss.descriptor); 
  if (P->ds.descriptor)
     desc[SDESC_DS].extract_descriptor(&P->ds.descriptor); 
  if (P->fs.descriptor)
     desc[SDESC_FS].extract_descriptor(&P->fs.descriptor); 
  if (P->gs.descriptor)
     desc[SDESC_GS].extract_descriptor(&P->gs.descriptor); 
  if (P->gdtr) 
     desc[SDESC_GDT].set_base_limit(P->gdtr >> 16, P->gdtr & 0xffff); 
  if (P->idtr) 
     desc[SDESC_IDT].set_base_limit(P->idtr >> 16, P->idtr & 0xffff); 

  if (P->ldtr.descriptor)
     desc[SDESC_LDT].extract_descriptor(&P->ldtr.descriptor); 
  if (P->tr.descriptor)
     desc[SDESC_TR].extract_descriptor(&P->tr.descriptor); 

  REG64(regs, REG_TSC) = P->tsc;

  REG32(regs, REG_DR0) = P->dr[0];
  REG32(regs, REG_DR1) = P->dr[1];
  REG32(regs, REG_DR2) = P->dr[2];
  REG32(regs, REG_DR3) = P->dr[3];
  REG32(regs, REG_DR4) = P->dr[4];
  REG32(regs, REG_DR5) = P->dr[5];
  REG32(regs, REG_DR6) = P->dr[6];
  REG32(regs, REG_DR7) = P->dr[7];

}

/*
 * CopyOutState - Copy the CPU state intom the SimOS format from
 * the CPU class of X86Sim.
 */
static void
CopyOutState(CPU *cpu, int cpuNum) 
{
  CPUState *P = SBase + cpuNum;
  RegisterFile *regs;

  regs = cpu->GetRegisters();
  P->eax = REG32(regs, REG_EAX);
  P->ecx = REG32(regs, REG_ECX);
  P->edx = REG32(regs, REG_EDX);
  P->ebx = REG32(regs, REG_EBX);
  P->esp = REG32(regs, REG_ESP);
  P->ebp = REG32(regs, REG_EBP);
  P->esi = REG32(regs, REG_ESI);
  P->edi = REG32(regs, REG_EDI);

  P->eflags = REG32(regs, REG_EFLAGS);
  P->cr[0] = REG32(regs, REG_CR0);
  P->cr[1] = REG32(regs, REG_CR1);
  P->cr[2] = REG32(regs, REG_CR2);
  P->cr[3] =  REG32(regs, REG_CR3);
  P->cr[4] =  REG32(regs, REG_CR4);

  P->es.selector = REG16(regs, REG_ES);
  P->cs.selector = REG16(regs, REG_CS);
  P->ss.selector = REG16(regs, REG_SS);
  P->ds.selector = REG16(regs, REG_DS);
  P->fs.selector = REG16(regs, REG_FS);

  P->tr.selector = REG16(regs, REG_TR);
  P->ldtr.selector = REG16(regs, REG_LDT);

  P->eip = REG32(regs, REG_EIP);

  SegmentDescriptor *desc = cpu->GetDescriptors();

  desc[SDESC_ES].compress_descriptor(&P->es.descriptor); 
  desc[SDESC_CS].compress_descriptor(&P->cs.descriptor); 
  desc[SDESC_SS].compress_descriptor(&P->ss.descriptor); 
  desc[SDESC_DS].compress_descriptor(&P->ds.descriptor); 
  desc[SDESC_FS].compress_descriptor(&P->fs.descriptor); 
  desc[SDESC_GS].compress_descriptor(&P->gs.descriptor); 
  P->gdtr =  (((uint64)desc[SDESC_GDT].base() << 16) | 
              (desc[SDESC_GDT].limit()));
  P->idtr =  (((uint64)desc[SDESC_IDT].base() << 16) | 
              (desc[SDESC_IDT].limit()));
  desc[SDESC_LDT].compress_descriptor(&P->ldtr.descriptor); 
  desc[SDESC_TR].compress_descriptor(&P->tr.descriptor); 

  P->tsc = REG64(regs, REG_TSC);

  P->dr[0] = REG32(regs, REG_DR0);
  P->dr[1] = REG32(regs, REG_DR1);
  P->dr[2] = REG32(regs, REG_DR2);
  P->dr[3] = REG32(regs, REG_DR3);
  P->dr[4] = REG32(regs, REG_DR4);
  P->dr[5] = REG32(regs, REG_DR5);
  P->dr[6] = REG32(regs, REG_DR6);
  P->dr[7] = REG32(regs, REG_DR7);

}

SimTime x86CurrentTime; 

static void
X86SimRun(void)
{
  int cpuNum = 0;
  CPUState *P = SBase + cpuNum;
  CPU   *cpu = (CPU *) ( P->x86simCpu);
  RegisterFile *regs = cpu->GetRegisters();
  IODevices *iod = GetIODevices(cpuNum);


   while(!cpu->is_halted()) {
      int cycles;
      VA oldPC;
      EventPollSingleQueue(x86CurrentTime);

      if (x86sim_debug_mode) {
         if (x86sim_debug_mode == 1) { 
            if ((cpuNum == x86sim_break_nexti) 
                || (x86sim_break_nexti == X86SIM_BREAKANYCPU)) {
               /* X86SimDebug returns nonzero if the PC has changed */
               /* or the contents of it have been overwritten, so it */
               /* needs to be reloaded. */
               if (X86SimDebug(cpuNum, 1)) {
                  continue;
               }
            } else if (x86sim_sigusr) {
               x86sim_debug_mode = 0;
               x86sim_sigusr = 0;
               AnnExec(AnnFind("simos", "sigusr"));
            }
         } else {
            ASSERT(x86sim_debug_mode > 1);
            x86sim_debug_mode--; 
         }
      }

      oldPC = REG32(regs, REG_EIP);
      cycles = cpu->ExecuteOneInstruction(NULL);
      if (cycles <= 0) cycles = 1; /* Sometimes we get an zero cycle 
                                    * instruction here */
      x86CurrentTime += cycles;
      RUN_PC_ANNOTATIONS(oldPC, REG32(regs, REG_EIP));       
      /*  STATS_INC(0, numInstructions, 1); */
   }
   CPUError("X86Sim CPU halted\n");

}
/*****************************************************************
 * InitStats
 * 
 *****************************************************************/
static void
InitStats(int cpuNum)
{
   STATS_SET(cpuNum, numInstructions, 0);
   STATS_SET(cpuNum, nextInstrSample, MS_SAMPLE_INSTR_INTERVAL);
   STATS_SET(cpuNum, iReads, 0);
   STATS_SET(cpuNum, dReads, 0);
   STATS_SET(cpuNum, dWrites, 0);
   STATS_SET(cpuNum, stallTime, 0);


}

/*****************************************************************
 * SimosX86SimSetExitSimulator
 * 
 * This overrides the default exit simulator passed into 
 * SimosCPUEnter
 *****************************************************************/
void
X86SimSetExitSimulator(CPUType newCpu)
{
}

/*****************************************************************
 * Debug Access
 *****************************************************************/
static void 
X86HandleDebugSignal(int cpuid, int sigusr)
{
   x86sim_debug_mode = 1;
   if (sigusr) {
      x86sim_sigusr = 1;
   } else { 
      if (cpuid == -1) {
         x86sim_break_nexti = X86SIM_BREAKANYCPU;
      } else {
         x86sim_break_nexti = cpuid;
      }
   }
}

/*****************************************************************
 * X86SimDebug
 *
 * Main entry point into the debugger from normal execution. 
 *
 * NOTE: the second argument is_break_instruction is set when
 * we call in due to a single-step trap (from cpu.c) even though
 * there is no break instruction at the PC.  We need to send
 * the SIGTRAP to gdb in this case.  If we send a SIGUSR2 it
 * stops single-stepping immediately even though we may not
 * have yet reached the next line boundary (for 'n' or 's').
 *
 * Return value 0 means nothing has changed, go through normal
 * instruction completion (pc = npc, npc += 4, annotations, etc).
 *
 * Return value 1 means PC has changed or instruction at PC has
 * been overwritten.  If you entered the debugger between instructions
 * abort anything you have read about the next instruction and restart.
 * If you entered the debugger from a breakpoint, reexecute the
 * instruction at that PC.
 *
 *****************************************************************/
int 
X86SimDebug(int cpunum, int is_break_instruction)
{
   Simdebug_result res;

   res = Simdebug_run(is_break_instruction ? SIGTRAP : SIGUSR2, cpunum);
   switch (res) {
   case SD_CONTINUE:
      x86sim_break_nexti = X86SIM_NOBREAK;
      x86sim_debug_mode = 0;
      break;
   case SD_NEXTI_ANYCPU:
      x86sim_break_nexti = X86SIM_BREAKANYCPU;
      x86sim_debug_mode = 2;
      break;
   default:
      x86sim_break_nexti = res;
      x86sim_debug_mode = 2;
      break;
   }
   
   return 1;
}

/*
 * Return the current CPU being executed 
 */
int
X86CurrentCpuNum(void)
{
   return (currentCPU - SBase);
}

/*
 * Return the current PC of a processor.
 */
VA
X86CurrentPC(int cpuNum)
{

  CPU   *cpu = (CPU *) ( SBase[cpuNum].x86simCpu);
  RegisterFile *regs = cpu->GetRegisters();
  return REG32(regs, REG_EIP);
}

/*
 * Return the current instruction 
 */
Inst X86CurrentInstruction(int cpu)
{
  CPUError("X86CurrentInstructioncalled not implemented\n");
  return 0;
}


/*
 * Get a x86 register.
 */

Result X86GetReg(int cpuNum, int regnum, Reg *val)
{
   CPUState *P = SBase + cpuNum;
   CPU   *cpu = (CPU *) ( P->x86simCpu);
   RegisterFile *regs = cpu->GetRegisters();
   SegmentDescriptor *desc = cpu->GetDescriptors();
   Reg64 descr;
   
   if (regnum <= EDI_REGNUM) {
      *val = REG32(regs, regnum);
      return SUCCESS;
   }
   switch (regnum) {
   case EFLAGS_REGNUM: { *val = REG32(regs, REG_EFLAGS); return SUCCESS; }
   case EIP_REGNUM: { *val = REG32(regs, REG_EIP); return SUCCESS; }
   case CS_REGNUM: { *val = REG16(regs, REG_CS); return SUCCESS; }
   case SS_REGNUM: { *val = REG16(regs, REG_SS); return SUCCESS; }
   case DS_REGNUM: { *val = REG16(regs, REG_DS); return SUCCESS; }
   case ES_REGNUM: { *val = REG16(regs, REG_ES); return SUCCESS; }
   case FS_REGNUM: { *val = REG16(regs, REG_FS); return SUCCESS; }
   case GS_REGNUM: { *val = REG16(regs, REG_GS); return SUCCESS; }
   case LDTR_REGNUM: { *val = REG16(regs, REG_LDT); return SUCCESS; }
   case TR_REGNUM: { *val = REG16(regs, REG_TR); return SUCCESS; }
   case CR0_REGNUM: { *val = REG32(regs, REG_CR0); return SUCCESS; }
   case CR1_REGNUM: { *val = REG32(regs, REG_CR1); return SUCCESS; }
   case CR2_REGNUM: { *val = REG32(regs, REG_CR2); return SUCCESS; }
   case CR3_REGNUM: { *val = REG32(regs, REG_CR3); return SUCCESS; }
   case CR4_REGNUM: { *val = REG32(regs, REG_CR4); return SUCCESS; }
   case GDTR_BASE_REGNUM: { *val = desc[SDESC_GDT].base(); return SUCCESS; }
   case GDTR_LIMIT_REGNUM:  { *val = desc[SDESC_GDT].limit(); return SUCCESS; }
   case IDTR_BASE_REGNUM: { *val = desc[SDESC_IDT].limit(); return SUCCESS; }
   case IDTR_LIMIT_REGNUM:  { *val = desc[SDESC_IDT].limit(); return SUCCESS; }
   case CS_DESC_LO_REGNUM: 
   case CS_DESC_HI_REGNUM: {
      desc[SDESC_CS].compress_descriptor(&descr); 
      *val = (regnum == CS_DESC_LO_REGNUM) ? (descr) : (descr>>32);
      return SUCCESS;
   }
   case SS_DESC_LO_REGNUM: 
   case SS_DESC_HI_REGNUM: {
      desc[SDESC_SS].compress_descriptor(&descr); 
      *val = (regnum == SS_DESC_LO_REGNUM) ? (descr) : (descr>>32);
      return SUCCESS;
   }
   case DS_DESC_LO_REGNUM: 
   case DS_DESC_HI_REGNUM: {
      desc[SDESC_DS].compress_descriptor(&descr); 
      *val = (regnum == DS_DESC_LO_REGNUM) ? (descr) : (descr>>32);
      return SUCCESS;
   }
   case ES_DESC_LO_REGNUM: 
   case ES_DESC_HI_REGNUM: {
      desc[SDESC_ES].compress_descriptor(&descr); 
      *val = (regnum == ES_DESC_LO_REGNUM) ? (descr) : (descr>>32);
      return SUCCESS;
   }
   case FS_DESC_LO_REGNUM: 
   case FS_DESC_HI_REGNUM: {
      desc[SDESC_FS].compress_descriptor(&descr); 
      *val = (regnum == FS_DESC_LO_REGNUM) ? (descr) : (descr>>32);
      return SUCCESS;
   }
   case GS_DESC_LO_REGNUM: 
   case GS_DESC_HI_REGNUM: {
      desc[SDESC_GS].compress_descriptor(&descr); 
      *val = (regnum == GS_DESC_LO_REGNUM) ? (descr) : (descr>>32);
      return SUCCESS;
   }
   case LDTR_DESC_LO_REGNUM: 
   case LDTR_DESC_HI_REGNUM: {
      desc[SDESC_LDT].compress_descriptor(&descr); 
      *val = (regnum == LDTR_DESC_LO_REGNUM) ? (descr) : (descr>>32);
      return SUCCESS;
   }
   case TR_DESC_LO_REGNUM: 
   case TR_DESC_HI_REGNUM: {
      desc[SDESC_TR].compress_descriptor(&descr); 
      *val = (regnum == TR_DESC_LO_REGNUM) ? (descr) : (descr>>32);
      return SUCCESS;
   }
   case TSC_LO_REGNUM: { *val = REG64(regs, REG_TSC); return SUCCESS; }
   case TSC_HI_REGNUM: { *val = (REG64(regs, REG_TSC)>>32); return SUCCESS; }

   case DR0_REGNUM: { *val = REG32(regs, REG_DR0); return SUCCESS; }
   case DR1_REGNUM: { *val = REG32(regs, REG_DR1); return SUCCESS; }
   case DR2_REGNUM: { *val = REG32(regs, REG_DR2); return SUCCESS; }
   case DR3_REGNUM: { *val = REG32(regs, REG_DR3); return SUCCESS; }
   case DR4_REGNUM: { *val = REG32(regs, REG_DR4); return SUCCESS; }
   case DR5_REGNUM: { *val = REG32(regs, REG_DR5); return SUCCESS; }
   case DR6_REGNUM: { *val = REG32(regs, REG_DR6); return SUCCESS; }
   case DR7_REGNUM: { *val = REG32(regs, REG_DR7); return SUCCESS; }

   default: 
     break;
   }
   return FAILURE;
} 

/*
 * Store an x86 register 
 */
Result X86PutReg(int cpuNum, int regnum, Reg val)
{
   CPUState *P = SBase + cpuNum;
   CPU   *cpu = (CPU *) ( P->x86simCpu);
   RegisterFile *regs = cpu->GetRegisters();
   SegmentDescriptor *desc = cpu->GetDescriptors();
   Reg64 descr;
   
   if (regnum <= EDI_REGNUM) {
      REG32(regs, regnum) = val;
      return SUCCESS;
   }
   switch (regnum) {
   case EFLAGS_REGNUM: { REG32(regs, REG_EFLAGS) = val; return SUCCESS; }
   case EIP_REGNUM: { REG32(regs, REG_EIP) = val; return SUCCESS; }
   case CS_REGNUM: {  REG16(regs, REG_CS) = val; return SUCCESS; }
   case SS_REGNUM: {  REG16(regs, REG_SS) = val; return SUCCESS; }
   case DS_REGNUM: {  REG16(regs, REG_DS) = val; return SUCCESS; }
   case ES_REGNUM: {  REG16(regs, REG_ES) = val; return SUCCESS; }
   case FS_REGNUM: {  REG16(regs, REG_FS) = val; return SUCCESS; }
   case GS_REGNUM: {  REG16(regs, REG_GS) = val; return SUCCESS; }
   case LDTR_REGNUM: { REG16(regs, REG_LDT) = val; return SUCCESS; }
   case TR_REGNUM: { REG16(regs, REG_TR) = val; return SUCCESS; }
   case CR0_REGNUM: { REG32(regs, REG_CR0) = val; return SUCCESS; }
   case CR1_REGNUM: { REG32(regs, REG_CR1) = val; return SUCCESS; }
   case CR2_REGNUM: { REG32(regs, REG_CR2) = val; return SUCCESS; }
   case CR3_REGNUM: { REG32(regs, REG_CR3) = val; return SUCCESS; }
   case CR4_REGNUM: { REG32(regs, REG_CR4) = val; return SUCCESS; }
   case GDTR_BASE_REGNUM: { 
      desc[SDESC_GDT].set_base_limit(val, desc[SDESC_GDT].limit()); 
      return SUCCESS; }
   case GDTR_LIMIT_REGNUM:  {
      desc[SDESC_GDT].set_base_limit(desc[SDESC_GDT].base(),val); 
      return SUCCESS; }
   case IDTR_BASE_REGNUM: {
      desc[SDESC_IDT].set_base_limit(val, desc[SDESC_IDT].limit()); 
      return SUCCESS; }
   case IDTR_LIMIT_REGNUM:  {
      desc[SDESC_IDT].set_base_limit(desc[SDESC_IDT].base(),val); 
      return SUCCESS; }
   case CS_DESC_LO_REGNUM: 
   case CS_DESC_HI_REGNUM: {
      desc[SDESC_CS].compress_descriptor(&descr); 
      if (regnum == CS_DESC_LO_REGNUM) 
         descr = (descr & 0xffffffff00000000LL) | val;
      else
         descr = (descr & 0x00000000ffffffffLL) | ((Reg64)val<<32);
      desc[SDESC_CS].extract_descriptor(&descr); 
      return SUCCESS;
   }
   case SS_DESC_LO_REGNUM: 
   case SS_DESC_HI_REGNUM: {
      desc[SDESC_SS].compress_descriptor(&descr); 
      if (regnum == SS_DESC_LO_REGNUM) 
         descr = (descr & 0xffffffff00000000LL) | val;
      else
         descr = (descr & 0x00000000ffffffffLL) | ((Reg64)val<<32);
      desc[SDESC_SS].extract_descriptor(&descr); 
      return SUCCESS;
   }
   case DS_DESC_LO_REGNUM: 
   case DS_DESC_HI_REGNUM: {
      desc[SDESC_DS].compress_descriptor(&descr); 
      if (regnum == DS_DESC_LO_REGNUM) 
         descr = (descr & 0xffffffff00000000LL) | val;
      else
         descr = (descr & 0x00000000ffffffffLL) | ((Reg64)val<<32);
      desc[SDESC_DS].extract_descriptor(&descr); 
      return SUCCESS;
   }
   case ES_DESC_LO_REGNUM: 
   case ES_DESC_HI_REGNUM: {
      desc[SDESC_ES].compress_descriptor(&descr); 
      if (regnum == ES_DESC_LO_REGNUM) 
         descr = (descr & 0xffffffff00000000LL) | val;
      else
         descr = (descr & 0x00000000ffffffffLL) | ((Reg64)val<<32);
      desc[SDESC_ES].extract_descriptor(&descr); 
      return SUCCESS;
   }
   case FS_DESC_LO_REGNUM: 
   case FS_DESC_HI_REGNUM: {
      desc[SDESC_FS].compress_descriptor(&descr); 
      if (regnum == FS_DESC_LO_REGNUM) 
         descr = (descr & 0xffffffff00000000LL) | val;
      else
         descr = (descr & 0x00000000ffffffffLL) | ((Reg64)val<<32);
      desc[SDESC_FS].extract_descriptor(&descr); 
      return SUCCESS;
   }
   case GS_DESC_LO_REGNUM: 
   case GS_DESC_HI_REGNUM: {
      desc[SDESC_GS].compress_descriptor(&descr); 
      if (regnum == GS_DESC_LO_REGNUM) 
         descr = (descr & 0xffffffff00000000LL) | val;
      else
         descr = (descr & 0x00000000ffffffffLL) | ((Reg64)val<<32);
      desc[SDESC_GS].extract_descriptor(&descr); 
      return SUCCESS;
   }
   case LDTR_DESC_LO_REGNUM: 
   case LDTR_DESC_HI_REGNUM: {
      desc[SDESC_LDT].compress_descriptor(&descr); 
      if (regnum == LDTR_DESC_LO_REGNUM) 
         descr = (descr & 0xffffffff00000000LL) | val;
      else
         descr = (descr & 0x00000000ffffffffLL) | ((Reg64)val<<32);
      desc[SDESC_LDT].extract_descriptor(&descr); 
      return SUCCESS;
   }
   case TR_DESC_LO_REGNUM: 
   case TR_DESC_HI_REGNUM: {
      desc[SDESC_TR].compress_descriptor(&descr); 
      if (regnum == TR_DESC_LO_REGNUM) 
         descr = (descr & 0xffffffff00000000LL) | val;
      else
         descr = (descr & 0x00000000ffffffffLL) | ((Reg64)val<<32);
      desc[SDESC_TR].extract_descriptor(&descr); 
      return SUCCESS;
   }
   case TSC_LO_REGNUM: { REG32(regs, REG_TSC_LO) = val; return SUCCESS; }
   case TSC_HI_REGNUM: { REG32(regs, REG_TSC_HI) = val; return SUCCESS; }

   case DR0_REGNUM: { REG32(regs, REG_DR0) = val; return SUCCESS; }
   case DR1_REGNUM: { REG32(regs, REG_DR1) = val; return SUCCESS; }
   case DR2_REGNUM: { REG32(regs, REG_DR2) = val; return SUCCESS; }
   case DR3_REGNUM: { REG32(regs, REG_DR3) = val; return SUCCESS; }
   case DR4_REGNUM: { REG32(regs, REG_DR4) = val; return SUCCESS; }
   case DR5_REGNUM: { REG32(regs, REG_DR5) = val; return SUCCESS; }
   case DR6_REGNUM: { REG32(regs, REG_DR6) = val; return SUCCESS; }
   case DR7_REGNUM: { REG32(regs, REG_DR7) = val; return SUCCESS; }


   default: 
     break;
   }
   return FAILURE;
}


Result X86GetMem(int cpunum, VA vAddr, uint nbytes, char *buf)
{   
   uint i;
   VA vaBase;
   PA pAddr;
   CPUState *P = SBase + cpunum;
   CPU   *cpu = (CPU *) ( P->x86simCpu);
   RegisterFile *regs = cpu->GetRegisters();
   SegmentDescriptor *desc = cpu->GetDescriptors();

   buf[0] = 0;
   vaBase = desc[SDESC_CS].base();
   if (vaBase != desc[SDESC_DS].base()) {
      Sim_Warning("Problem with segments in X86GetMem\n");
   }

   for (i = 0; i < nbytes; i++) {
      if (X86TranslateVirtualNoSE(cpunum, vAddr+i, &pAddr) != SUCCESS) {
         return FAILURE;
      }
      buf[i] = *(char *)(SBase[cpunum].memoryPtr + pAddr);
   }
   return SUCCESS;
}

Result X86PutMem(int cpunum, VA vAddr, uint nbytes, char *buf)
{
   uint i;
   PA pAddr;
   CPUState *P = SBase + cpunum;
   CPU   *cpu = (CPU *) ( P->x86simCpu);
   RegisterFile *regs = cpu->GetRegisters();
   SegmentDescriptor *desc = cpu->GetDescriptors();



   for (i = 0; i < nbytes; i++) {
      if (X86TranslateVirtualNoSE(cpunum, vAddr+i, &pAddr) != SUCCESS) {
         return FAILURE;
      }
      *(char *)(SBase[cpunum].memoryPtr + pAddr) = buf[i];
   }
   return SUCCESS;
}

Result X86TranslateVirtualNoSE(int cpunum, VA vAddr, PA *pAddr)
{
   uint32 pa;
   VA vaBase;
   CPUState *P = SBase + cpunum;
   CPU   *cpu = (CPU *) ( P->x86simCpu);
   RegisterFile *regs = cpu->GetRegisters();
   SegmentDescriptor *desc = cpu->GetDescriptors(); 
   uint32 ret;

   vaBase = desc[SDESC_CS].base();
   if (vaBase != desc[SDESC_DS].base()) {
      Sim_Warning("Problem with segments in X86TranslateVirtualNoSE\n");
   }
   vAddr += vaBase;
   ret = cpu->TranslateLinearToPhysical((uint32)vAddr, pa, FALSE, FALSE, TRUE);
   if (!ret) return FAILURE;
   *pAddr = (PA)pa;
   return SUCCESS;
}

static void X86InstallMemAnnotation(VA vA)
{ ;
}

void X86DMAInval(int machine, PA* k0list)
{
   ;
}

void X86LaunchSlaveCPU(int cpuNum)
{
   CPUError("X86LaunchSlaveCPU\n");
}


static void X86DoCheckpoint(int cpuNum)
{
   CPUError("X86LaunchSlaveCPU\n");
}

static void X86Unstall(int cpuNum) 
{
   CPUError("X86Unstall\n");
}

static CPUStatus X86CPUStatus(int cpuNum)
{
   return SBase[cpuNum].cpuStatus;
}


static void X86ExitSimulator(CPUType toSim)
{
  exit(0);
}

static int 
X86SimResetCPU(int cpuNum)   
{
    CPUError("X86SimResetCPU called\n");
    return 0;
}

static SimTime
X86InstructionCount(int cpuNum)
{
  return STATS_VALUE(cpuNum, numInstructions);

}

SimTime
X86CycleCount(int CpuNum)
{
   return x86CurrentTime;
}


char *
X86PAToHostAddr(int cpu, PA pAddr, VA vAddr) 
{
   return (PHYS_TO_MEMADDR(M_FROM_CPU(cpu),pAddr) +
           (IS_REMAPPED_PADDR(pAddr,cpu) ? remapVec->NodeAddr[cpu] : 0));
}


/*****************************************************************
 * CPUVectorInit
 *****************************************************************/
void 
X86CPUVectorInit(void)
{
   /* parallel event queues */

   CPUVec.Handle_Debug_Signal  = X86HandleDebugSignal;
   CPUVec.GetRegister          = X86GetReg;
   CPUVec.PutRegister          = X86PutReg;
#ifdef notdef
   CPUVec.ClearLockFlag        = X86ClearLockFlag;
   CPUVec.GetLockFlag          = X86GetLockFlag;
   CPUVec.GetLockAddr          = X86GetLockAddr;
#endif
   CPUVec.GetMemory            = X86GetMem;
   CPUVec.PutMemory            = X86PutMem;
   CPUVec.TranslateVirtualNoSE = X86TranslateVirtualNoSE;
   CPUVec.InstallMemAnnotation = X86InstallMemAnnotation;
   CPUVec.Unstall              = X86Unstall;
   CPUVec.CPUstatus            = X86CPUStatus;
   CPUVec.CycleCount           = X86CycleCount;
   CPUVec.DMAInval             = X86DMAInval;
   CPUVec.ExitSimulator        = X86ExitSimulator;
   CPUVec.CurrentCpuNum        = X86CurrentCpuNum;
   CPUVec.CurrentPC            = X86CurrentPC;
   CPUVec.CurrentInstruction   = X86CurrentInstruction;
   CPUVec.InstrCount           = X86InstructionCount;
   CPUVec.DoCheckpoint         = X86DoCheckpoint;
#if 0
   CPUVec.TakeBusError         = X86TakeBusError;
#endif
   CPUVec.useMemRef            = FALSE;
   CPUVec.singleEventQueue     = TRUE;
   CPUVec.PAToHostAddr         = X86PAToHostAddr;
   CPUVec.LaunchSlaveCPU       = X86LaunchSlaveCPU;
}

void
X86DoPortIO(int cpuNum, bool isInInst, int port, int len, void *buffer)
{
   int direction = isInInst ? PORT_IN : PORT_OUT;
   IODevices *iod = GetIODevices(cpuNum);

   (void) IOSpace_InOut(iod, port, len, direction, buffer);
     
}

#define monitor_call_setclock      1
#define monitor_call_getrealtime   3
#define monitor_call_saveflags     4
#define monitor_call_restoreflags  5
#define monitor_call_scsi_getconfig 7
#define monitor_call_scsi_command   8
#define monitor_call_scsi_ackintr   9

#include "../simos/vscsi.h"

int
X86DoMonitorCallHack(int cpuNum)
{
  CPUState *P = SBase + cpuNum;
  CPU   *cpu = (CPU *) ( P->x86simCpu);
  RegisterFile *regs;

  regs = cpu->GetRegisters();

  switch (REG32(regs, REG_EAX)) {
  case monitor_call_setclock: {
      CPUWarning("X86DoMonitorCallHack: Ignoring mon_setclock call @ 0x%x\n",
                 REG32(regs, REG_EIP));
      break;
  }
  case monitor_call_getrealtime: {
     /* Hopefully this monitor call will disappear before we can 
      * that it is not the right time. 
      */
     REG32(regs,REG_EBX) = 0x34834e78;  /* Bogus but close */
     REG32(regs,REG_EAX) = 0;
  }
  case monitor_call_saveflags: {
      REG32(regs,REG_EBX) = REG32(regs,REG_EFLAGS);
      REG32(regs,REG_EAX) = 0;
      break;
  }
  case monitor_call_restoreflags: {
      REG32(regs,REG_EFLAGS) = REG32(regs,REG_EBX);
      REG32(regs,REG_EAX) = 0;
      break;
  }
  case monitor_call_scsi_getconfig: {
      int slot =  REG32(regs,REG_EBX);
      int retVal;

      REG32(regs,REG_EAX) = mon_scsi_getconfig(cpuNum, slot, &retVal);
      REG32(regs,REG_EBX) = retVal;
      break;
  }
  case monitor_call_scsi_command: {
      VA ecb_addr =  REG32(regs,REG_EBX);
      int retVal;

      REG32(regs,REG_EAX) = mon_scsi_command(cpuNum, ecb_addr, &retVal);
      REG32(regs,REG_EBX) = retVal;
      break;
  }
  case monitor_call_scsi_ackintr: {
      int drive =  REG32(regs,REG_EBX);
      int retVal;

      REG32(regs,REG_EAX) = mon_scsi_ackintr(cpuNum, drive, &retVal);
      REG32(regs,REG_EBX) = retVal;
      break;
  }

  default:
     CPUWarning("Unknown monitor call %d (0x%x) at 0x%x, skipping\n", 
             REG32(regs, REG_EAX), REG32(regs,REG_EBX),
             REG32(regs, REG_EIP));
  }

  return 1;
} 

bool
X86HasIntrPending(int cpuNum, uint8 *vectorNum)
{
   int intr = Intr_Check(GetIODevices(cpuNum));
   if (intr == NO_INTR_PENDING) {
       return FALSE;
   } 
   *vectorNum = (int8)intr;
   return TRUE;
}
void
X86DidIRET(int cpuNum)
{
   RFE_EVENT();
}
