/* $Id: x86sim.cc,v 1.4 1998/04/29 04:55:03 nikos Exp $
 *
 * RCS Log:
 * $Log: x86sim.cc,v $
 * Revision 1.4  1998/04/29 04:55:03  nikos
 * Added new AranaTrace command "AranaTrace  filter_thru_L1" that dumps
 * only instructions with references (pc + ea) that miss in the L1 cache.
 * MBs are always dumped.
 *
 * Revision 1.4  1998/03/25 19:16:55  nikos
 * atf_tlb_fixed
 *
// Revision 1.3  1998/03/13  17:12:40  nikos
// Fix vaddr->physaddr translation bug and tlb behavior in ATF.
// Changes affect only people who use ATF for their traces.
//
 * Revision 1.3  1998/03/13 16:37:15  nikos
 * .
 *
// Revision 1.2  1998/02/16  23:20:14  nikos
// Fixed AranaTrace command which was broken with the last commit.
// Macros in external_trace.h are set such that the name "arana"
// appears only in arana_trace.[ch] and no-where else!!!
// In the previous revision, kappa.c has not been cleaned up.
//
// If you want to use the AranaTrace command,
// compile SimOS with the following variables set :
//
// EXTERNAL_TRACE=ARANA_TRACE
// ARANA_TRACE_INC_DIR=/perf_sym/ATF/atf
// ARANA_TRACE_LIB=/perf_sym/ATF/atf/libatf.a
// ARANA_TRACE_OBJS=arana_trace.o
// ARANA_TRACE_HEADER=arana_trace.h
//
// This commit reports a bunch of modified files which are not
// really modified (only some rcs fields like $Id: x86sim.cc,v 1.4 1998/04/29 04:55:03 nikos Exp $, $Header: /wrl/proj/simos/cvsroot/simulation/cpus-x86/x86sim/x86sim.cc,v 1.4 1998/04/29 04:55:03 nikos Exp $, etc).
//
 * Revision 1.2  1998/02/11 14:24:55  nikos
 * .
 *
// Revision 1.1  1997/12/16  01:28:33  bugnion
// stanford update (x86)
//
// Revision 1.9  1997/10/29  22:44:55  sean
// added instruction counter module, option (-C) to enable it
//
 * Revision 1.8  1997/09/24 08:12:29  sean
 * glue code to pass brk value from LoadELFBinary through to simlib
 *
 * Revision 1.7  1997/09/15 02:58:21  yoel
 * Adjusted to properly call the new UI stuff.
 *
 * Revision 1.6  1997-09-14 16:19:03-07  yoel
 * Added new UI calls.
 *
 * Revision 1.5  1997-09-05 17:39:51-07  yoel
 * Added LockStep support and some display.
 *
 * Revision 1.4  1997/03/18 20:47:49  sean
 * added command-line option handling and fixed CreateMemory() for mmap()
 *   implementations that do not support MAP_ANON (all but Linux?)
 *
 * Revision 1.3  1997/03/18 19:12:57  sean
 * Massive overhaul to let it load ELF binaries, stole code from simlib.cc to
 *   have a nicer simulator (one that can be (sort of) controlled)
 *
 * Revision 1.2  1997/02/04 02:39:35  sjt
 * cosmetic changes
 *
 */

static char rcsid[] __attribute__((unused))
        = "$Id: x86sim.cc,v 1.4 1998/04/29 04:55:03 nikos Exp $";

// This is the main simulator file

#include <stdio.h>
#include <stdlib.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <fcntl.h>
#include <sys/mman.h>
#include <unistd.h>
#include "cpu.h"
#include "load_elf.h"
#include "simlib.h"
#include "debug.h"
#include "x86sim.h"
#include "lockstep.h"
#include "ui.h"
#include "counter.h"

// This is the function we'll give to LoadELFBinary in order to create memory
//   where it is needed - we'll use mmap() (it should be relatively portable)
// NOTE: Some implementations of mmap() do not allow MAP_ANON - instead you
//   have to map from /dev/zero
static int CreateMemory(int32& base, int32 size, bool fixed, void *param)
{
  int offset;
  void *result;
#ifndef MAP_ANON
  int zero_fd;

  if((zero_fd = open("/dev/zero", O_RDONLY)) < 0) return(-1);
#endif

  // need to allocate starting at a page boundary
  offset = (base % PAGE_SIZE);

  // try the mmap
#ifdef MAP_ANON
  result = mmap((char *)(base - offset), (size + offset),
		PROT_READ | PROT_WRITE | PROT_EXEC,
		MAP_ANON | MAP_PRIVATE | (fixed ? MAP_FIXED : 0), -1, 0);
#else
  result = mmap((char *)(base - offset), (size + offset),
		PROT_READ | PROT_WRITE | PROT_EXEC,
		MAP_PRIVATE | (fixed ? MAP_FIXED : 0), zero_fd, 0);
#endif

  // if we succeeded, set 'base' appropriately
  if(((int)result) != -1) {
    base = ((int32)result) + offset;

#ifndef MAP_ANON
    // need to close /dev/zero, but touch the memory first to make sure it 
    //  is writeable (not sure if this is necessary, but it doesn't hurt)
    *(int *)base = 0;
    close(zero_fd);
#endif
    return(0);
  } else {
#ifndef MAP_ANON
    close(zero_fd);
#endif
    return(-1);
  }
}

#ifdef NATIVE_X86_HOST
static int MakeNewStack(int32 new_stack_base, int32 new_stack_size)
{
  if(CreateMemory(new_stack_base,new_stack_size,false,0) < 0) {
    perror("new stack creation failed");
    return(-1);
  }

  CopyAndSwitchStacks(DEFAULT_LINUX_STACK_TOP, 
		      new_stack_base + new_stack_size);

  return(0);
}
#endif

// count  = how many cycles to run before prompting
// target = breakpoint
static int64 SimLoop(CPU& cpu, LockStep *singlestep, int count, int32 target)
{
  int64 total_cycles = 0; // total cycles executed
  RegisterFile *regptr;
  char line[40];

  regptr = cpu.GetRegisters();

  while(!cpu.is_halted()) {
    if(REG32(*regptr,REG_EIP) == target) count = 0;

    if(!count) {

      if (singlestep != NULL) singlestep->TraceRegisters();
      cpu.DumpState();
      fprintf(stderr,"> ");
      fgets(line,39,stdin);

      if(line[0] == 'q') break;

      if((line[0] != '0') || (line[1] != 'x')) {
	count = atoi(line);
	if(!count) count = 1;
	target = 0;
      } else {
	count = -1;
	target = strtol(line+2, (char **)NULL, 16);
      }

    }

    if(count == -2) cpu.DumpState();
    if(count > 0) count--;
      
    cpu.ExecuteOneInstruction(singlestep);

    total_cycles++;
  }

  return(total_cycles);
}

static void PrintUsage(const char *progname)
{
  fprintf(stderr,X86SIM_USAGE,progname);
}

static void PrintVersionInfo(const char *progname)
{
  printf(X86SIM_VERSION,rcsid);
}

int main(int argc, char *argv[], char *envp[])
{
  CPU   *cpu;
  LockStep *singlestep;
  UI	*ui;
  InstructionCounter *counter;
  bool  doLockStep = false, useCounter = false;
  int32 program_sp;
  int   optchar;
  int   initial_count = 0;
  int32 initial_bkpnt = 0;
#ifdef NATIVE_X86_HOST
  int32 simulator_stack_size = NEW_SIM_STACK_SIZE;
#else
  int32 program_stack_size = DEFAULT_PROGRAM_STACK_SIZE;
#endif

  while((optchar = getopt(argc, argv, X86SIM_OPTIONS)) != -1)
    switch(optchar) {
    case 'S':  // change size of simulator stack (native only)
#ifdef NATIVE_X86_HOST
      simulator_stack_size = strtol(optarg, NULL, 0);
      DEBUG(DEBUG_LOADER,"Size of simulator stack set to %08X.",
	    simulator_stack_size);
#else
      fprintf(stderr,"Warning: -S option has no effect in a non-native simulator.\n");
#endif
      break;

    case 'P':  // change size of program stack (non-native only)
#ifdef NATIVE_X86_HOST
      fprintf(stderr,"Warning: -P option has no effect in a native simulator.\n");
#else
      program_stack_size = strtol(optarg, NULL, 0);
      DEBUG(DEBUG_LOADER,"Size of program stack set to %08X.",
	    program_stack_size);
#endif
      break;

    case 'C': // Use the instruction counter
      useCounter = true;
      break;

    case 'L': // Enable LockStep mode
#ifdef NATIVE_X86_HOST
      doLockStep = true;
#else
      fprintf(stderr, "Warning:  Lockstep only makes sense in a native simulator.\n");
#endif
      break;

    case 'd': // turn on (or off) debug flags
      if((optarg[0] == '+') && optarg[1]) {
	optarg++;
	while(*optarg) TurnOnDebugCode(*optarg++);
      } else
      if((optarg[0] == '-') && optarg[1]) {
	optarg++;
	while(*optarg) TurnOffDebugCode(*optarg++);
      } else
	while(*optarg) TurnOnDebugCode(*optarg++);
      break;

    case 'i': // set an initial instruction count
      initial_count = strtol(optarg, NULL, 0);
      initial_bkpnt = 0;
      break;

    case 'c': // run continuously
      initial_count = -1;
      initial_bkpnt = 0;
      break;

    case 'b': // set an initial breakpoint
      initial_count = -1;
      initial_bkpnt = strtol(optarg, NULL, 0);
      break;

    case ':':
      fprintf(stderr,"Missing argument for option.\n");
      PrintUsage(argv[0]);
      exit(1);
      break;

    case 'v':
      PrintVersionInfo(argv[0]);
      exit(0);
      break;

    case 'h':
      PrintUsage(argv[0]);
      exit(0);
      break;

    case '?':
    default:
      fprintf(stderr,"Unknown option.\n");
      PrintUsage(argv[0]);
      exit(1);
      break;
    }

#ifdef NATIVE_X86_HOST
  if(MakeNewStack(NEW_SIM_STACK_START,simulator_stack_size) < 0)
    return(1);

  program_sp = DEFAULT_LINUX_STACK_TOP;
#else 
  program_sp = (int32) new int8[program_stack_size];
  if(!program_sp) {
    fprintf(stderr,"Could not allocate memory for program's stack!\n");
    exit(1);
  }

  // need to point to end of stack
  program_sp += program_stack_size;
#endif

  // make sure we got a program name to run
  if(optind >= argc) {
    PrintUsage(argv[0]);
    exit(1);
  }

  // create our CPU and give it all of memory to play with
  cpu = new CPU(0,0x0FFFFFFFF);

  if(LoadELFBinary(*cpu, argv[optind], 0, program_sp, 
		   (const char **) argv+optind, (const char **) envp,
		   CreateMemory, 0, false,
#ifdef NATIVE_X86_HOST
		   &simulated_brk
#else
		   NULL
#endif
		   ) < 0) {
    fprintf(stderr,"Could not load '%s'.\n",argv[optind]);
    exit(1);
  }

#ifdef NATIVE_X86_HOST
  // min brk value is the current brk value (end of text)
  simulated_min_brk = simulated_brk;

  // If we should be single stepping, set it up now.  After LOAD so we can
  // blow away the stack.  I hope this works.
  if (doLockStep) {
    if ((singlestep = LockStep::Init(argv[optind], &argv[optind])) == NULL) {
      fprintf(stderr, "Could not single step '%s'\n", argv[optind]);
      exit(1);
    }
    if (singlestep->MakeRegisterFile(cpu->GetRegisters())==-1) {
      fprintf(stderr, "Could not initialize Register File\n");
      exit(1);
    }
    if (singlestep->CopyStack(*cpu) == -1) {
      fprintf(stderr, "Could no clone the stack.\n");
      exit(1);
    }
  } else {
    singlestep = NULL;
  }
#else
  singlestep = NULL;
#endif

/*  printf("Done - %lld cycles executed\n",
                SimLoop(*cpu, singlestep, initial_count, initial_bkpnt));
*/

  // Create our instrunction counter.
  if(useCounter) {
    counter = new InstructionCounter;
    cpu->SetInstructionCounter(counter);
  }

  //   New UI stuff.  Note there is not yet support for initial_count or
  // initial_breakpoint, but it shouldn't be hard if we want it. --JS.
  ui = new UI(cpu, singlestep);
  printf("Done - %lld cycles executed\n", ui->SimLoop());

  if(useCounter) {
    counter->PrintReport(stdout);
    delete counter;
  }

  delete cpu;
  delete singlestep;
  delete ui;

#ifndef NATIVE_X86_HOST
  delete[] (int8 *)(program_sp - program_stack_size);
#endif
}
