/*
 * Copyright (C) 1996-1998 by the Board of Trustees
 *    of Leland Stanford Junior University.
 * 
 * This file is part of the SimOS distribution. 
 * See LICENSE file for terms of the license. 
 *
 */

 /****************************************************************
 * clock.c
 * 
 * $Author: verghese $
 * $Date: 1999/09/27 22:07:57 $
 *****************************************************************/
#include <stdio.h>
#include <string.h>
#include "simmisc.h"
#include "checkpoint.h"
#include "simmagic.h"
#include "sips.h"
#include "eventcallback.h"
#include "simutil.h"
#include "embra.h"
#include "cp0.h"
#include "clock.h"
#include "main_run.h"
#include "mem_control.h"
#include "stats.h"
#include "callout.h"
#include "translator.h"
#include "debug.h"
#include "driver.h"
#include "hw_events.h"
#include "embra_interface.h"
#include "embra_globals.h"
#include "parallel.h"
#include "tc_coherence.h"

/* **********************************************************************
 * All of the clock shared state is in shared memory. All of the callbacks
 * must be in shared memory, even non IPI as the callbacks themselves are
 * queued.
 * **********************************************************************/

#define MAX_INTERRUPT_CALLBACKS (16*SIM_MAXCPUS)


typedef struct ClockSharedState {
    EventCallbackHdr interrupts[MAX_INTERRUPT_CALLBACKS];
    int rotor;
    int lock;
    EventCallbackHdr pcSampling[SIM_MAXCPUS];
    EventCallbackHdr periodicCallback[SIM_MAXCPUS];
    EventCallbackHdr periodicAnnotationCallback;
    EventCallbackHdr statCallback[SIM_MAXCPUS];
    EventCallbackHdr bootCallback;
    EventCallbackHdr exitCallback;
    EventCallbackHdr checkInterruptCallback[SIM_MAXCPUS];
} ClockSharedState;


/* Local Data */
static ClockSharedState *clockState;
static int pc_sample_interval = 256;

extern K0A non_excepting_tv( int cpuNum, VA va );

/* Local Functions */
static void InitPeriodicCallbacks(int cpu);
static void StatCallback(int cpuNum, EventCallbackHdr *hdr,void *arg);
static void PeriodicCallback(int cpuNum, EventCallbackHdr *hdr,void *arg);
static void PeriodicAnnotationCallback(int cpuNum, EventCallbackHdr *hdr,void *arg);
static void BootCallback(int cpuNum, EventCallbackHdr *hdr,void *arg);
static void Embra_Send_Check(int cpuNum, SimTime delay);

/* XXX Obsolete:
static SimTime next_PCSample;
*/

/* Here we set up our periodic callbacks:
 * StatCallback,
 * PeriodicCallback,
 * PeriodicAnnotationCallback.
 * Question: is this called by every thread??
 * It looks like it.See Embra_Clock_Init.
 * Except... ASSERT CPU = 0???? Ugh!!!
 */
static void InitPeriodicCallbacks(int cpu)
{
  /*  ASSERT (cpu==0); */

  if (cpu != 0)
    return;
  
  if (embra.stats) {
     EventDoCallback(cpu, StatCallback,
                     &clockState->statCallback[cpu], 0,
                     embra.statInterval);
  }
  
  EventDoCallback(cpu, PeriodicCallback,
                  &clockState->periodicCallback[cpu], 0,
                  embra.miscCheckInterval);

  if (cpu == 0) {
     EventDoCallback(0, PeriodicAnnotationCallback,
                     &clockState->periodicAnnotationCallback, 0,
                     embra.periodicAnnInterval);
  }
  
}


void EmbraClosePeriodicCallbacks(void)
{
   int cpu = 0;
   if (EventCallbackActive(&clockState->statCallback[cpu])) {
      EventCallbackRemove(&clockState->statCallback[cpu]);
   }
   if (EventCallbackActive(&clockState->periodicCallback[cpu])) {
      EventCallbackRemove(&clockState->periodicCallback[cpu]);
   }
   if (EventCallbackActive(&clockState->periodicAnnotationCallback)) { 
      EventCallbackRemove(&clockState->periodicAnnotationCallback);
   }
}


void
Embra_Clock_Init(int cpuNum)
{
   /*
    * allow fast check from assembly 
    */

   callout_interval = embra.timeQuantum;

   if (embra.parallel) {
     /* Multiple queues in parallel mode */
     EMP[cpuNum].eventQueueTimePtr = &eventQueues[cpuNum].calltime;
   }
   else {
     /* Single queue otherwise */
     EMP[cpuNum].eventQueueTimePtr = SingleEventQueueCalltimeAddr;
   }
     
   EMP[cpuNum].timeQuantum = embra.timeQuantum;
   EMP[cpuNum].cycleCountdown = EMP[cpuNum].timeQuantum;

   /* Only CPU 0 gets periodic callbacks?? I don't know */
   InitPeriodicCallbacks(cpuNum);
}


int
Update_And_Check_Interrupts( int cpuNum , VA targetPC)
{
   /*CPUWarning("CPU %d Update and Check HW 0x%x\n ", 
            cpuNum,
            EMP[cpuNum].intrBitsPtr[0] ); */
   /* Update and check for interrupts */

   EMP[cpuNum].CP0[C0_CAUSE] = 
      (EMP[cpuNum].CP0[C0_CAUSE] & ~CAUSE_EXTINTBITS) | 
      ((EMP[cpuNum].intrBitsPtr[0] << CAUSE_IPSHIFT) & CAUSE_EXTINTBITS);

   if (((EMP[cpuNum].CP0[C0_CAUSE] & EMP[cpuNum].CP0[C0_SR]) & SR_IMASK) &&
        (EMP[cpuNum].CP0[C0_SR] & SR_IEC) &&
       !((EMP[cpuNum].CP0[C0_SR] & (SR_EXL|SR_ERL)))) {

      /* Processors generally take exceptions in between instructions */
      /* so when  we are stalled for a cache miss (or whatever), delay */
      /* exception (interrupt) detection */
         if ( !EMP[cpuNum].stalled ) {
            if (targetPC) { 
               ASSERT (!IN_BD(targetPC));
               ASSERT (IS_KSEG0(targetPC) || IS_KUSEG(targetPC));
               /*
                * now is the time to field the post-pc annotation,
                * since we are skipping the instruction by smashing
                * the PC.
                */
	       CalloutLock(cpuNum,1);
               AnnExec(AnnFMLookup(CLEAR_BD(EMP[cpuNum].PC),ANNFM_PC_TYPE));
	       CalloutUnlock(cpuNum,1);
               EMP[cpuNum].PC = targetPC;
            }
            Em_EXCEPTION(cpuNum,EXC_INT,0);
            EmbraSideEffect(cpuNum);
            return 1;
         }
   }
   return 0;
}

void EmIntrBitsChanged(int cpuNum)
{
  /* Note: we could optimize this??? only call if cpunum == currentCpuNum ?*/
    Embra_Send_Check( cpuNum, 0 );
}

/* Do Interrupt Now (only takes exception if interrupt is enabled) */
void CheckInterruptCallback(int cpuNum, EventCallbackHdr *hdr, void *arg)
{
#if 0
  CPUWarning("CPU %d CheckInterruptCallback\n", cpuNum);
#endif
  CalloutUnlock(cpuNum,2);
  /* Unlock while checking interrupts, in case we reenter */
  Update_And_Check_Interrupts(cpuNum,0);
  CalloutLock(cpuNum,2);
}


/* Do Interrupt Now (only takes exception if interrupt is enabled) */
static void EmbraInterruptCallback(int cpuNum, EventCallbackHdr *hdr, void *arg)
{
   IEC intrtoset = (IEC) arg;   /* keep compiler quiet by casting thru int */

   SIM_DEBUG(('i',"DEV %d %lld Send 0x%x ", 
              cpuNum, EmbraCpuCycleCount(cpuNum), intrtoset));
   RaiseIBit(cpuNum, intrtoset);
   SIM_DEBUG(('i', "Ibits 0x%x\n", EMP[cpuNum].intrBitsPtr[0]));

   Update_And_Check_Interrupts( cpuNum,0 );
}

/* Do deliver SIPS Now */
static void EmbraDeliverSIPSCallback(int cpuNum, EventCallbackHdr *hdr,void *arg)
{
   sim_sips_deliver(cpuNum, (int) arg /* chan */);
   Update_And_Check_Interrupts( cpuNum ,0);
}

/* Do Interrupt Now (only takes exception if interrupt is enabled) */
static void BootCallback(int cpuNum, EventCallbackHdr *hdr,void *arg)
{
   int cpu;

   for( cpu = 1; cpu < TOTAL_CPUS; cpu++ ) {
      uint launchAddr = (uint)sim_misc.launchAddr[cpu];
      if ( launchAddr ) {
        /* XXX - non-backdoor address is launch, backdoor address is call */
         if ( IS_BACKDOOR(launchAddr) ) {
            int64 result;
            /* Do call */
            result = ((int64 (*)(int,int,int,int))launchAddr)
               ( sim_misc.launchArg[cpu][0],
                 sim_misc.launchArg[cpu][1],
                 sim_misc.launchArg[cpu][2],
                 sim_misc.launchArg[cpu][3] );

            if ( result == SLAVELOOP_CONTINUE ) {
               /* signal init function done */
               sim_misc.launchAddr[cpu] = 0;
               /* clear so will be 0 if not set up on next call */
               sim_misc.launchArg[cpu][0] = 0;
               sim_misc.launchArg[cpu][1] = 0;
               sim_misc.launchArg[cpu][2] = 0;
               sim_misc.launchArg[cpu][3] = 0;
               CPUPut("Slave %d returning to launch wait\n",cpu );
            }
         } else {
            /* Add everybody in & let mem_control handle launch */
            int i;
            for( i = 0; i <= TOTAL_CPUS; i++ ) {
               EMP[i].next = &EMP[(i+1)%(TOTAL_CPUS+1)];
            } /* for */
            /* Don't reinstall callback */
            return;
         }
      }
   }
   EventDoCallback(cpuNum,BootCallback,hdr,0,embra.timeQuantum);
}

void Embra_Send_Check( int cpuNum, SimTime delay)
{

    EventDoCallback(cpuNum, CheckInterruptCallback,
		    &clockState->checkInterruptCallback[cpuNum],
		    0 /* no argument */, delay);
}

void Embra_Send_Interrupt( int cpuNum, IEC intrtoset, SimTime delay )
{    
    int count = 0;
    EventCallbackHdr *event;

    int execCpu = CPUVec.CurrentCpuNum();
       
    if ( delay == 0 ) {
       CalloutLock(execCpu,3);
       RaiseIBit( cpuNum, intrtoset );
       CalloutUnlock(execCpu,3);
       Update_And_Check_Interrupts( cpuNum,0 );
    } else {
      /* XXX Should be another lock, probably */
      CalloutLock(execCpu,4);
       clockState->rotor = (clockState->rotor+1) % MAX_INTERRUPT_CALLBACKS;
       while ( EventCallbackActive(&clockState->interrupts[clockState->rotor]) ){
          count++;
          ASSERT( count < MAX_INTERRUPT_CALLBACKS  );
          clockState->rotor = (clockState->rotor+1) % MAX_INTERRUPT_CALLBACKS;
       }
       event = &clockState->interrupts[ clockState->rotor];
       /* Race condition here??? hopefully not ...*/
       EventDoCallback(cpuNum, EmbraInterruptCallback, event, (void *)(int)intrtoset, delay);
       CalloutUnlock(execCpu,4);
    }
}

void Embra_Deliver_SIPS( int cpuNum, int chan, SimTime delay )
{    
    int count = 0;
    EventCallbackHdr *event;

    CalloutLock(cpuNum,5);
    clockState->rotor = (clockState->rotor+1) % MAX_INTERRUPT_CALLBACKS;
    while ( EventCallbackActive(&clockState->interrupts[clockState->rotor]) ){
        count++;
        ASSERT( count < MAX_INTERRUPT_CALLBACKS  );
        clockState->rotor = (clockState->rotor+1) % MAX_INTERRUPT_CALLBACKS;
    }
    event = &clockState->interrupts[ clockState->rotor];
    EventDoCallback(cpuNum, EmbraDeliverSIPSCallback, event, (void *)chan, delay);
    CalloutUnlock(cpuNum,5);
}

static void StatCallback(int cpuNum, EventCallbackHdr *hdr,void *arg)
{
   Print_Recent_Stats(cpuNum);
   EventDoCallback(cpuNum,StatCallback,hdr,0,embra.statInterval);
}


static void PeriodicCallback(int cpuNum, EventCallbackHdr *hdr,void *arg)
{

#if 0
    CPUPrint("callback %i %lld\n",cpuNum,CPUVec.CpuCycleCount(cpuNum));
#endif
    EventDoCallback(cpuNum, PeriodicCallback, hdr, 0, embra.miscCheckInterval);

    EmbraPollSigUsr(cpuNum);

#ifdef notdef
    /* XXX this sort of hoses us if someone other than CPU 0
     * does an exit...
     */
    if ( sim_misc.myCPUType != sim_misc.enterThisCPU) {
       EmbraExit(sim_misc.enterThisCPU);
    }
#endif

}

static void PeriodicAnnotationCallback(int cpuNum, EventCallbackHdr *hdr,void *arg)
{  
   ASSERT(cpuNum == 0);
   EventDoCallback(0, PeriodicAnnotationCallback, hdr, 0, embra.periodicAnnInterval);
   CalloutLock(cpuNum,6);
   /* This is an annotation!!! */
   PERIODIC_EVENT;
   CalloutUnlock(cpuNum,6);
}


/* ****************************************************************
 * Periodic_Callout is called very frequently. Keep it as short as 
 * possible. 
 * ****************************************************************/

void IncCC(int cpuNum, int inc)
{
   EMP[cpuNum].cycleCount += EMP[cpuNum].timeQuantum;
}

/*#define DEBUG_EVENTPOLL*/
#ifdef DEBUG_EVENTPOLL
static struct {
   uint cycleCountdown;
   uint PC;
   uint64 cycleCount;
   uint stalled;
} debug_evpoll[SIM_MAXCPUS];
static SimTime last_call;
#endif

void GeneralEmEventPoll(int cpuNum)
{
   /*CPUPrint("%d GeneralEmEventPoll\n", cpuNum);*/
   ASSERT(cpuNum < TOTAL_CPUS);
#ifdef notdef
#ifdef STALL_BY_REMOVING_FROM_LOOP
   /* This needs justification */
   /*EMP[TOTAL_CPUS].cycleCount += EMP[0].timeQuantum;*/
   /*EventPollSingleQueue(EMP[TOTAL_CPUS].cycleCount);*/
#else
   ASSERT(cpuNum == 0 );
#endif
#endif /* notdef */
   /* Value passed to EventPoll is a hint telling it how far down the
      callback queue to look.  We know all cycle counts are within 1
      quantum, so this should suffice */
#ifdef DEBUG_EVENTPOLL
   ASSERT( EmbraCpuCycleCount(0) + 4*EMP[0].timeQuantum > last_call);
#endif
   CalloutLock(cpuNum,7);
   EventPollSingleQueue(EmbraCpuCycleCount(0) + 4*EMP[0].timeQuantum);
   CalloutUnlock(cpuNum,7);
#ifdef DEBUG_EVENTPOLL
   debug_evpoll[cpuNum].cycleCountdown = EMP[cpuNum].cycleCountdown;
   debug_evpoll[cpuNum].PC = EMP[cpuNum].PC;
   debug_evpoll[cpuNum].cycleCount = EMP[cpuNum].cycleCount;
   debug_evpoll[cpuNum].stalled = EMP[cpuNum].stalled;
   last_call = EmbraCpuCycleCount(0) + 4*EMP[0].timeQuantum;
#endif
   /* some event hung off the cycle count may have
    * changed memory (eg. the debugger).  This checks that.
    */
   FlushTCIfNecessary(cpuNum);
   ReenterTC(EMP[TOTAL_CPUS].next);
   /* NOT REACHED */
}

/* This is for parallel mode */
EmbraState* CEmEventPoll(int cpuNum)
{
   EmbraGlobal *g = GetGlobals();

   /* Value passed to EventPoll is a hint telling it how far down the
      callback queue to look.  We know all cycle counts are within 1
      quantum, so this should suffice */
   SimTime now_ish = EmbraCpuCycleCount(cpuNum);

   g->curEmp = &EMP[cpuNum];

   if (eventQueues[cpuNum].calltime + 100000 < now_ish) {
      CPUWarning("CEmEventPoll missing callouts calltime=%lld, nowish=%lld \n",
                 eventQueues->calltime, now_ish);
      EventQueueDump(TOTAL_CPUS);
   }
   
#ifdef DEBUG_EVENTPOLL
   ASSERT( now_ish >= last_call );
#endif

   if ( EventPendingMultipleQueue( cpuNum, now_ish) ) {
     if (cpuNum != 0 && 0 )
       CPUWarning("CEmEventPoll event pending on cpu %d\n",cpuNum);
     /* NO!!! can cause exception ??? */
      CalloutLock(cpuNum,2);
     EventProcess( cpuNum, now_ish );
     CalloutUnlock(cpuNum,2);

     if (eventQueues[cpuNum].calltime + 100000 < now_ish) {
	CPUWarning("CEmEventPoll calltime is still sluggish after "
		   " EventProcess.... calltime=%lld, nowish=%lld \n",
		   eventQueues->calltime, now_ish);
     }

      /* Code is written to assume that this will happen because
         exceptions can be raised in a callback */
     /* XXX BL note: we can return *unless* our PC has been
	changed, right??? This would be more efficient.
     */
      ReenterTC( &EMP[cpuNum] );
      /* NOT REACHED */
   }
   
#ifdef DEBUG_EVENTPOLL
   debug_evpoll[cpuNum].cycleCountdown = EMP[cpuNum].cycleCountdown;
   debug_evpoll[cpuNum].PC = EMP[cpuNum].PC;
   debug_evpoll[cpuNum].cycleCount = EMP[cpuNum].cycleCount;
   debug_evpoll[cpuNum].stalled = EMP[cpuNum].stalled;
   last_call = EmbraCpuCycleCount(0) + 4*EMP[0].timeQuantum;
#endif
   
   /* ReenterTC(EMP[TOTAL_CPUS].next);*/
   /* NOT REACHED */
   return &EMP[0];
}

/* This is called when a single event queue is used TOTAL_CPUS == 1 || MPinUP */
EmbraState* CEmEventPollUP(EmbraState *curEmp)
{

   /* Value passed to EventPoll is a hint telling it how far down the
      callback queue to look.  We know all cycle counts are within 1
      quantum, so this should suffice */
   SimTime now_ish = EmbraCpuCycleCount(0);
   if (eventQueues->calltime + 100000 < now_ish) {
      CPUWarning("CEmEventPollUP missing callouts calltime=%lld, nowish=%lld \n",
                 eventQueues->calltime, now_ish);
   }
   
         
#ifdef DEBUG_EVENTPOLL
   ASSERT( now_ish >= last_call );
#endif

   if( EventPendingSingleQueue(now_ish) ) {
      CalloutLock(curEmp->myNum, 2);
      EventProcessSingleQueue( now_ish );
      CalloutLock(curEmp->myNum, 2);
      /* Code is written to assume that this will happen because
         exceptions can be raised in a callback */
      ReenterTC( &EMP[0] );
      /* NOT REACHED */
   }

#ifdef DEBUG_EVENTPOLL
   debug_evpoll[cpuNum].cycleCountdown = EMP[cpuNum].cycleCountdown;
   debug_evpoll[cpuNum].PC = EMP[cpuNum].PC;
   debug_evpoll[cpuNum].cycleCount = EMP[cpuNum].cycleCount;
   debug_evpoll[cpuNum].stalled = EMP[cpuNum].stalled;
   last_call = EmbraCpuCycleCount(0) + 4*EMP[0].timeQuantum;
#endif
   /* ReenterTC(EMP[TOTAL_CPUS].next);*/
   /* NOT REACHED */
   return &EMP[0];
}

/* XXX
 * This is not the most efficient thing in the world, but we are
 * using it for now so that we can better understand the behavior
 * of parallel embra. Anyway, we will be called periodically and
 * will end up here, where we can check for events, and also check
 * for a barrier?? (e.g. here is where we can implement periodic
 * barrier. Note we can optimize things later.
 */

#define PERIODIC_BARRIER_INTERVAL 10000000

static SimTime nextBarrier = PERIODIC_BARRIER_INTERVAL;

/* Flags: if barrierFlag is raised, then PeriodicCallout
 * will call DoPeriodicBarrier
 */

int barrierFlag = 0;

/* If one of the following flags is set, we will:
 * wait at a barrier for everyone to catch up;
 * flush the tc at said barrier;
 * exit embra and terminate all of our threads
 * take a checkpoint
 */

int tcFlushFlag = 0;
int embraExitFlag = 0;
int embraExitThread = 0;
int checkpointFlag = 0;

void DoPeriodicBarrier(EmbraState *curEmp)
{
   EmbraGlobal *globals = GetGlobals();

   int localExitFlag = 0;
   
   /* Wait for everyone else to get here */
#ifdef DEBUG_BARRIER
   CPUWarning("Thread %d waiting at barrier(%d)\n", globals->id
	      embra.parallel);
#endif

   /* Everybody wait */
   PeriodicBarrier(embra.parallel);
   
   /* Only thread 0 gets to do single-threaded work. */
   if (globals->id == 0) {
      int i, syncFlag;
      /* Flush TC if necessary */
      if (tcFlushFlag) {
	 Do_Clear_Translation_State(TCFLUSH_ALL, globals->curEmp);
	 tcFlushFlag = 0;
      }
      /* Do checkpoint if necessary */
      if (checkpointFlag) {
	 EmbraDoCheckpoint(globals->curEmp->myNum);
	 checkpointFlag = 0;
      }
      /* Check to see if everyone is in sync */
      for (i = 0, syncFlag = 0; i < TOTAL_CPUS; i++)
	 syncFlag |= (EMP[i].cycleCount < nextBarrier);
      /* Update barrier interval if so */
      if (!syncFlag)
	 nextBarrier += PERIODIC_BARRIER_INTERVAL;
   }

   /* Exit if we're done, and not the exiting thread */
   if (embraExitFlag)
      localExitFlag = 1;
   
   /* Wait until thread 0 has done its thing */
   PeriodicBarrier(embra.parallel);

   if (localExitFlag) {
      if (globals->id != 0)
	 ExitThread(globals->id);
      else
	 EmbraDoExit();
   }

   /* Now we can all continue */
#ifdef DEBUG_BARRIER
   CPUWarning("Thread %d continues\n", globals->id);
   fflush(stdout);
#endif

}

void Periodic_Callout(EmbraState *curEmp)
{
  SimTime *eventQueueTimePtr = curEmp->eventQueueTimePtr;
  EmbraGlobal *g = GetGlobals();
  EmbraState *emp;
  
  /* Hooray! Both cpu's are getting here!! */

  ASSERT(!curEmp->stalled);

  /* if it's time, poll for events */
  if (eventQueueTimePtr &&
      *eventQueueTimePtr <= curEmp->cycleCount) {
    CEmEventPoll(curEmp->myNum);
  }

  /* Do we need to wait at a barrier? */
  if (tcFlushFlag || checkpointFlag || embraExitFlag) {
    DoPeriodicBarrier(curEmp);
    /* TC was flushed, so call ReEnterTC */
  }


  /* BARRIER SYNCHRONIZATION */
  /* Deschedule if we've reached the barrier interval */
  if (0 && curEmp->cycleCount > nextBarrier && !curEmp->stalled) {
     curEmp->stalled = 1;
     g->stalledCount++;
     /* If everyone is stalled, wait at barrier */
     if (g->stalledCount == g->cpuCount) {
	double ProcTimeDiff(int proc), MIPS;
	if (0) { 
	   if (nextBarrier == g->lastBarrier)
	      CPUError("UGH. Thread %d waiting at same barrier.\n", g->id);
	   g->lastBarrier = nextBarrier;
	   CPUWarning("Thread %d waiting at sync barrier %lld, %lld\n",
		      g->id, nextBarrier, curEmp->cycleCount);
	   MIPS = ProcTimeDiff(g->id);
	   CPUWarning("mips(%d) = %.2f\n", g->id, MIPS);
	}
	/* wait until we're ready to go */
	while (curEmp->cycleCount > nextBarrier)
	   DoPeriodicBarrier(curEmp);
	/* Restart everyone */
	for (emp = curEmp; g->stalledCount; emp = emp->next, g->stalledCount-- ) {
	      emp->stalled = 0;
	}
     }

     ASSERT(!curEmp->stalled);
  }

  /* SKEW SYNCHRONIZATION (not exact) */
  if (0) {
     /* for now, just sync with CPU 0 */
     static int slowest = 0;
     static long long SKEW = 100000000;
     long long delta = curEmp->cycleCount - EMP[slowest].cycleCount;

     if (delta < 0)
	slowest = curEmp->myNum;
     else if (delta > SKEW) {
	CPUWarning("Thread %d waiting for slow cpu %d...\n", g->id, slowest);
	/* Wait for slowest to catch up */
	do {
	   curEmp = curEmp->next;
	   if (curEmp->cycleCount < EMP[slowest].cycleCount)
	      slowest = curEmp->myNum;
	} while  (curEmp->cycleCount > EMP[slowest].cycleCount);
	g->curEmp = curEmp;
	CPUWarning("Thread %d continuing...\n", g->id);
	ReenterTC(curEmp);
     }
  }

  /* switch and continue */
  ReenterTC_CX(curEmp);

}


#define MAXPROC 64
#define MAXCPU 128

double ProcTimeDiff(int proc)
{
   static int init=0;
   static int first[MAXPROC];
   static long long cycles[MAXCPU];
   static struct timeval old[MAXPROC];
   struct timeval new[MAXPROC];
   double diff[MAXPROC];

   EmbraState *emp;
   int i;
   double cycleTotal;
   EmbraGlobal *g = &EmbraGlobals[proc];
   
   double MIPS;
   
   gettimeofday(&new[proc],NULL);

   if (init == 0) {
      bzero(first, sizeof(first));
      bzero(cycles, sizeof(cycles));
   }

   /* calculate time */
   if (first[proc]) {
      diff[proc]=new[proc].tv_sec - old[proc].tv_sec; 
      diff[proc] += 0.000001 * (new[proc].tv_usec - old[proc].tv_usec);
   } else { 
      diff[proc]= 0.0;;
   }
   init = 1;
   first[proc] = 1;
   old[proc] = new[proc];

   /* calculate cycles */
   for (i = 0, emp = g->curEmp, cycleTotal = 0;
	i < g->cpuCount; i++, emp = emp->next) {
      cycleTotal += (double) (emp->cycleCount - cycles[emp->myNum]);
      cycles[emp->myNum] = emp->cycleCount;
   }
   
   /* calculate MIPS */
   MIPS =  cycleTotal/diff[proc]/1e6;

   if (MIPS < 1.0) {
      CPUWarning("MIPS is low.\n");
      for (i = 0, emp = g->curEmp, cycleTotal = 0;
	   i < g->cpuCount; i++, emp = emp->next) {
	 CPUWarning("cpu %d cycles = %lld, nextBarrier = %lld\n",
		    emp->myNum, emp->cycleCount, nextBarrier);
      }
   }
      
   return MIPS;

}

void NonReturningProcReturned(void)
{
   VASSERT(0,("Non-returning procedure returned.  This is bad.\n"));
}

SimTime EmbraReadTime( void )
{
   if ( embra.MPinUP ) {
      return EmbraCpuCycleCount(0);
   } else {
	return EmbraCpuCycleCount(GetGlobals()->curEmp->myNum);
   }
}


#define DEBUG_CLOCK 0

#if DEBUG_CLOCK
static struct {
   uint cycleCountdown;
   uint PC;
   uint64 cycleCount;
   uint64 clock_val;
} debug_clock[SIM_MAXCPUS];
#endif

SimTime EmbraCpuCycleCount(int cpuNum)
{
   static SimTime last_returned_value;
   int64 new_value;

   SimTime	cycleCount = EMP[cpuNum].cycleCount;
   int		cycleCountDown = EMP[cpuNum].cycleCountdown;
   int		timeQuantum = EMP[cpuNum].timeQuantum;

   
   new_value = cycleCount + timeQuantum - cycleCountDown;

   /* Avoid race condition for foreign CPU */
   if (cpuNum != GetGlobals()->curEmp->myNum)
      new_value = cycleCount;
   
#if DEBUG_CLOCK
   if (new_value < debug_clock[cpuNum].clock_val) {
      CPUWarning("cpu %d retrograde clock %lld->%lld, staying at max\n",
                 cpuNum, debug_clock[cpuNum].clock_val, new_value);
      CPUWarning("cycleCount = %lld, cyclecountDown = %d\n",cycleCount, cycleCountDown);
      CPUWarning("old cycleCount = %lld, old cycleCountdown = %d\n",
		 debug_clock[cpuNum].cycleCount,debug_clock[cpuNum].cycleCountdown);
      if (cpuNum != GetGlobals()->curEmp->myNum)
	CPUWarning("foreign cpu.\n");
      new_value = debug_clock[cpuNum].clock_val;
   }
   
   debug_clock[cpuNum].cycleCountdown = cycleCountDown;
   debug_clock[cpuNum].cycleCount = cycleCount;
   debug_clock[cpuNum].PC = EMP[cpuNum].PC;
   debug_clock[cpuNum].clock_val = new_value;
#endif

   return new_value;
}

/* Try just returning the count for cpu 0, so that
 * we have a monotonically increasing cycle count
 * in parallel mode.
 */
SimTime EmbraCpuCycleCountZero(int dummy)
{
  return EmbraCpuCycleCount(0);
}

/* XXX THIS BOMBS HORRIBLY  */
SimTime EmbraCpuCycleCountVerify(int clockCpu)
{
  return EmbraCpuCycleCount(clockCpu);
}

/* Make the .cycleCount field the actual current time */

void EmbraFixCycleCounts(void)
{
   int cpuNum;
   SimTime maxTime = 0;
   for (cpuNum=0;cpuNum<TOTAL_CPUS;cpuNum++) { 
      EMP[cpuNum].cycleCount += EMP[cpuNum].timeQuantum - 
         EMP[cpuNum].cycleCountdown;
      EMP[cpuNum].cycleCountdown = EMP[cpuNum].timeQuantum;
      if ( EMP[cpuNum].cycleCount > maxTime) { 
         maxTime = EMP[cpuNum].cycleCount;
      }
   }
   for (cpuNum=0;cpuNum<TOTAL_CPUS;cpuNum++) { 
      EMP[cpuNum].cycleCount = maxTime;
   }
}


void EmbraClockInit(void)
{
    if ( !clockState ) {
        clockState = MallocShared(sizeof(ClockSharedState),"EmClock");
    }
    ASSERT( clockState );
}

/******************************************************************
 * Write a call to the system call exit on the user's stack.  Then 
 * jump to it 
*****************************************************************/
/**** XXX OBSOLETE???? ****/
void 
EmbraMakeProcExit(int cpuNum)
{
   EmbraState *P = &EMP[cpuNum];
   uint *pPtr;
   VA vSP;

   /* Insure that we are not writting over a page boundary by using */
   /* the beginning of the page*/  
   vSP = FORM_ADDR( PAGE_NUMBER(P->R[REG_SP]), 0 );
   pPtr = (uint*) non_excepting_tv( cpuNum, vSP );
   if ( !pPtr ) {
      CPUWarning("Failure to translate stack addr in MakeProcExit\n");
      return;
   }

   CPUWarning("PROCexit: CPU %d smashing address %#x\n", P->myNum, vSP);

   *pPtr++ = CIi( addiu_op, A0, G0, 0 ); /* li a0, 0 */
   *pPtr++ = CIi( addiu_op, V0, G0, 1001 ); /* li v0, 1001 */

   /* This trick won't work in base mode both because of this opcode */
   /* because we are not flushing the data cache*/
   *pPtr++ = CIs( syscall_op, 0, 0, 0); /* syscall */
   P->PC = vSP;
   return;
}

