/*
 * Copyright (C) 1996-1998 by the Board of Trustees
 *    of Leland Stanford Junior University.
 * 
 * This file is part of the SimOS distribution. 
 * See LICENSE file for terms of the license. 
 *
 */

/****************************************************************
 * driver.c

 * This file contains the entry points to embra.  The init functions
 * in this file are responsible for initializing all of the needed
 * data structures.

 * We also try to concentrate the policy #defines in this file.
 * 
 * Author: $Author: stets $
 * Date:   $Date: 2000/03/21 06:08:51 $
 *****************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <stddef.h>
#include <strings.h>
#include <unistd.h>
#include <sys/mman.h>
#include <sys/types.h>
#include <sys/time.h>
#include <fcntl.h>
#include <setjmp.h>

#include "simmisc.h"
#include "annotations.h"
#include "checkpoint.h"

#include "embra.h"
#include "driver.h"
#include "tc.h"
#include "translator.h"
#include "mem_control.h"
#include "callout.h"
#include "main_run.h"
#include "cp0.h"
#include "clock.h"
#include "qc.h"
#include "cache.h"
#include "stats.h"
#include "embra_interface.h"
#include "directory.h"
#include "decoder.h"
#include "stats.h"
#include "clock.h"
#include "simutil.h"
#include "tcl_init.h"
#include "debug.h"
#include "tc_coherence.h"
#include "parallel.h"
#include "embra_globals.h"

/* Locals */

#if 0 
int EmUseETLB = 0;
int barrCPUs = 0;
int MPinUP = 0;
int useVQC = 0;
int K0FromDevZero = 0;
int alreadyAlloc = 0;
#endif

/* Hmm, what to do about this??? */

int* gBarrier = 0;
int* debugBarrier = 0;
int* checkpointBarrier = 0;
int* tnsLock = 0;
int* exitBarrier = 0;

int deschedNumCPUs = -1;

/* XXX - this is really dependent on CPUType */
int log2sizeofcputype = 2;


/* Wouldn't it be nice if we lived in a world where compilers could
   deal with alignment information?  That way maybe our much abused
   typing system in C could stand a ghost of a chance */
#ifdef COMMENTOUT
unsigned PE_space[ ( SIM_MAXCPUS * sizeof(EmbraState) + DEFAULT_PAGESZ)/
                   sizeof(unsigned)];
#endif

/* Local function prototypes */
static void EmbraConfigureTC(void);
static void VerifyOffsets(void);
static void EmbraSchedule(void);
void Embra_Parallel_Run(void *);

/* On startup the mmu gets initialized, so we must go through
 *  and zero out the appropriate pages.  We could just rely
 *  on the fact that qc_map_page won't map pages with annotations,
 *  but this will make sure kernel addresses are zeroed out.
*/
void
EmbraInstallMemAnnotation(VA vAddr) 
{
   uint page = PAGE_NUMBER(vAddr);

   if (embra.emode == EMBRA_PAGE) {
         int cpu;
         for (cpu = 0; cpu < TOTAL_CPUS; cpu++) {
#ifdef EMBRA_USE_QC64
            qc64_remove_addr(cpu,vAddr);
#else
            EMP[cpu].mmu[page] = 0;
#endif
         }
      }
}

C_LINK int _is_mips2(void);

void Embra_Init(int swtch)
{
   static int embra_initialized = 0;

   int cpuNum;

   if ( !embra_initialized ) {
      embra_initialized = 1;

      /* Make sure that offet constants into EmbraState are correct */
      VerifyOffsets();

      /* Quick runtime way of deciding if we are on an R4000 */
      if ( !_is_mips2() ) {
         CPUPut("Embra only runs on machines that supports at least the mips2 ISA\n");
         exit( 1 );
      }

      EmbraClockInit();

      /* Schedule CPUs */
      EmbraSchedule();

      /* Initialize translation caches & space for dynamic interface code */
      EmbraConfigureTC();

      
      /* Initialize interface code and register allocation stuff */
      Translator_Init();

      /* Initialize maxInGroup */
      DecoderInit();

      /* Allocate cache miss table */
      emSMHT = ZMALLOC(TOTAL_CPUS * sizeof(EmSMHT),"Embra:SMHT");
 
      TCcoherence_init(SIM_MEM_ADDR(0));

   }

   /* Initialize statistics */
   Stat_Init();

   /* CPU-specific initialization */
   for (cpuNum = 0; cpuNum < TOTAL_CPUS; cpuNum++) {
     Embra_Clock_Init(cpuNum);
     Cache_Init(cpuNum);
     Em_Tlb_Init(cpuNum, swtch);
   }

   /* Initialize cache directory */
   Directory_Init();

   if (embra.emode == EMBRA_PAGE) {
     AnnFMInit(DEFAULT_PAGESZ);
   } else {
     /* embra.emode == EMBRA_CACHE */
     AnnFMInit(SCACHE_LINE_SIZE);
   }
   if (!swtch) {
     AnnCommonSetup();
   }

   /* Parallel initialization */
   if ( embra.parallel ) {
     InitParallel();
   }
}


/* EmbraSchedule(): Assign CPU's to processes;
 * Currently this is in a round-robin list with a virtual (fake)
 * processor at the end;
 *
 * This is for the first scheduler bring-up and is a general-purpose
 * scheduler. After measuring the overhead, performance debugging,
 * and other debugging, we may wish to reinstate special modes for
 * serial-only (1-on-1) and straight parallel (N-on-N) execution.
 */

#define TOTAL_PROCS embra.parallel

static void
EmbraSchedule(void)
{
   int cpu;
   int proc;
   int count, leftover;
   int i;

   EmbraState *emp;

   if (TOTAL_PROCS <= 0) TOTAL_PROCS = 1;
   
   CPUPrint("EmbraSchedule: running %d cpus on %d processes\n",
	      TOTAL_CPUS, TOTAL_PROCS);
   
   if (TOTAL_PROCS > TOTAL_CPUS) {
      CPUWarning("EmbraSchedule: setting TOTAL_PROCS from %d to %d\n",
		 TOTAL_PROCS, TOTAL_CPUS);
      TOTAL_PROCS = TOTAL_CPUS;
   }

   count = TOTAL_CPUS/TOTAL_PROCS;		/* basic amount */
   leftover = TOTAL_CPUS % TOTAL_PROCS;

   /* This is 123, 456, 789 assignment */
   if (0) {
      /* Assign CPU's to processors; each processor gets a
       * ring of CPU's.
       */
      cpu = 0;
      for (proc = 0; proc < TOTAL_PROCS; proc++) {
	 EmbraGlobal *g = &EmbraGlobals[proc];
	 /* How many CPU's do I get ? */
	 g->cpuCount = (proc < leftover ? count + 1: count);
	 /* Head of run queue */
	 g->curEmp = &EMP[cpu];
	 /* Thread count-1 processors */
	 for (i = 1; i < g->cpuCount; i++) {
	    EMP[cpu].next = &EMP[cpu+1];
	    EMP[cpu+1].prev = &EMP[cpu];
	    cpu++;
	 }
	 /* Point last CPU in queue to first */
	 EMP[cpu].next = g->curEmp;
	 g->curEmp->prev = &EMP[cpu];
	 cpu++;
      }
   }

   /* This is 531, 642 assignment */

 /* Clear lists */
   for (proc = 0; proc < TOTAL_PROCS; proc++)
      EmbraGlobals[proc].curEmp = 0;
   /* Distribute CPU's across processors */
   proc = 0;
   for (cpu = 0; cpu < TOTAL_CPUS; cpu++) {
      EmbraGlobal *g = &EmbraGlobals[proc];
      if (g->curEmp) {
	 /* Yes this is N^2, but it's clear */
	 for (emp = g->curEmp; emp->next; emp = emp->next);
	 EMP[cpu].prev = emp;
	 emp->next = &EMP[cpu];
      }
      else g->curEmp = &EMP[cpu];
      g->cpuCount++;
      proc = (proc + 1) % TOTAL_PROCS;
   }
   /* Link up ends */
   for (proc = 0; proc < TOTAL_PROCS; proc++) {
      for (emp = EmbraGlobals[proc].curEmp; emp->next; emp = emp->next);
      emp->next = EmbraGlobals[proc].curEmp;
   }
      
   /* Print out schedule to log */
   CPUPrint("EmbraSchedule: CPU assignments:\n");
   for (proc = 0; proc < TOTAL_PROCS; proc++) {
      CPUPrint("EmbraSchedule: Process %d has CPU: ", proc);
      emp = EmbraGlobals[proc].curEmp;
      do {
	 CPUPrint("%d ",emp->myNum);
	 emp = emp->next;
      } while (emp != EmbraGlobals[proc].curEmp);
      CPUPrint("\n");
   }

   /* Initialize pseudo-processor for MPinUP */
   /* For now, this never happens!! */
   if ( 0 && embra.MPinUP ) {
      /* XXX - Note the space for the extra processor is not actually */
      /* mapped shared. It should be allocated elsewhere */
      /* The mythical extra processor just holds the callback address */
      EMP[TOTAL_CPUS].PC = (Reg)EmEventPoll;
      EMP[TOTAL_CPUS].R[31] = (Reg)EmEventPoll;
      /* This needs to be set for the code in main_run which tries to do a */
      /* lookup on the PC, but it is a backdoor, so that look up will */
      /* fail.   May want to check for backdoor directly in that code */
      EMP[TOTAL_CPUS].mmu = EMP[0].mmu;
      /* Give it an event queue */
      EMP[TOTAL_CPUS].eventQueueTimePtr = SingleEventQueueCalltimeAddr;
      /* Have to activate this processor */
      EMP[TOTAL_CPUS].outOfSlaveLoop = 1;
   }

}

void Init_Proc_State(int cpu )
{
  /* cache hit counters */
   EMP[cpu].Sdhit_count = 0;
   EMP[cpu].Sihit_count = 0;
   /* Initialize MMU_MASK register */
   if ( (embra.emode == EMBRA_PAGE) || (!embra.useVQC) )
       EMP[cpu].Ssreg2 = 0x7fffffff;
}

/* This setjmp allows us to unwind the stack (e.g. after we've flushed
 * the translation cache, taken exceptions, etc.) and return to
 * running in the TC
*/


/* We call this routine to unwind the stack and continue running
 * in the TC
 */
void ReenterTC(EmbraState *emp)
{
  EmbraGlobal *g = GetGlobals();
  g->curEmp = emp;
  longjmp(g->Embra_Run_Setjmp, 1);
  /* Not Reached */
  ASSERT(0);
}

/* We call this routine to unwind the stack and continue running
 * in the TC
 */

void ReenterTC_CX(EmbraState *emp)
{
  EmbraGlobal *g = GetGlobals();
  g->curEmp = emp;

  longjmp(g->Embra_Run_Setjmp, 2); /* 2 -> do context switch */
  /* Not Reached */
  ASSERT(0);
}


/* This is the entry point to Embra. Embra_Init should be called */
/* before calling this function */
void
Embra_Run( int cpuNum, int clobber_machine_regs )
{
   int i; 

   /* Initialize CPU's */
   for (i = 0; i < TOTAL_CPUS; i++) {
      Init_Proc_State(i);
   }
      
   /* Spawn off threads if we are running in parallel */
   for (i = 1; i < embra.parallel; i++ ) {
      /* Start up a thread with our particular state */
      CreateThread(Embra_Parallel_Run, &EmbraGlobals[i]);
   }
   
   /* Use this thread to run cpu 0 */
   Embra_Parallel_Run(&EmbraGlobals[0]);
   
   /* Never returns */
   ASSERT(0);
}

/* This is the main entry point for threads in parallel Embra
*/
void
Embra_Parallel_Run(void *globalPtr)
{
   
  EmbraGlobal *global = (EmbraGlobal *)globalPtr;
  EmbraState *curEmp = global->curEmp;

  /* This is bogus, but for now... */
  SetGlobals((int) (global - EmbraGlobals), global);

  CPUPrint("PARALLEL: thread %d (EmbraGlobals[%d]) entering TC at 0x%x\n",
	     global->id, global-EmbraGlobals, curEmp,
	     curEmp->PC);
  gettimeofday(&run_start);
  /* Save where we are; we return here on a ReenterTC */
  if (setjmp(global->Embra_Run_Setjmp) != 2)
    /* Enter TC - never returns*/
    EnterTC(global->curEmp);
  else
    EnterTC_CX(global->curEmp); /* Enter TC and switch */
  /* Never reached */
  ASSERT(0);

}

void
ProcRemove( int cpuNum )
{
   ASSERT( embra.MPinUP );
   ASSERT( cpuNum < TOTAL_CPUS );
   ((EmbraState*)EMP[cpuNum].prev)->next = EMP[cpuNum].next;
   ((EmbraState*)EMP[cpuNum].next)->prev = EMP[cpuNum].prev;
   /* Aid Debugging -- don't reset next because I need it to schedule */
   /* the next guy */
   EMP[cpuNum].prev = 0;
}
void
ProcInsertFirst( int cpuNum )
{
   ASSERT( embra.MPinUP );
   ASSERT( cpuNum < TOTAL_CPUS );
   EMP[cpuNum].next = EMP[TOTAL_CPUS].next;
   EMP[cpuNum].prev = &EMP[TOTAL_CPUS];
   ((EmbraState*)EMP[TOTAL_CPUS].next)->prev = &EMP[cpuNum];
   EMP[TOTAL_CPUS].next = &EMP[cpuNum];
}

void 
ProcInsert( int cpuNum )
{
   EmbraState *t = &EMP[TOTAL_CPUS];
   ASSERT( cpuNum < TOTAL_CPUS );
   ASSERT( !EMP[cpuNum].prev );
   while (cpuNum > t->next->myNum ) {
      t = t->next;
   }
   EMP[cpuNum].next = t->next;
   EMP[cpuNum].prev = t;
   t->next->prev = & EMP[cpuNum];
   t->next = & EMP[cpuNum];
}


int
ProcListEmpty(void)
{
   ASSERT( embra.MPinUP );
   return EMP[TOTAL_CPUS].next == &EMP[TOTAL_CPUS];
}

/* **********************************************
 * Descheduling to speed up Embra boot
 * **********************************************/


void EmbraPromReleaseInit(void)
{
   deschedNumCPUs = 0;
}

void EmbraPromRelease(void *p1, void *p2)
{
   int i;
   CPUWarning("Releasing CPUs. \n");
   for(i=0;i<TOTAL_CPUS;i++) {
      if ( !EMP[i].outOfSlaveLoop && sim_misc.launchAddr[i]) {
         CPUWarning("EmbraPromRelease, launching %i \n",i);
         deschedNumCPUs++;
         ProcInsert(i);
      }
   }
}




/* ******************************************************************
 * EmbraConfigureTC
 * ******************************************************************/

#define LOG_TC_USER_SIZE 22
#define LOG_TC_KERN_SIZE 22
#define LOG_TC_GLUE_SIZE 12

#define TC_USER_NUM (1<<LOG_TC_USER_SIZE)
#define TC_KERN_NUM (1<<LOG_TC_KERN_SIZE)
#define TC_GLUE_NUM (1<<LOG_TC_GLUE_SIZE)

#define TC_USER_SIZE (TC_USER_NUM * 4 )
#define TC_KERN_SIZE (TC_KERN_NUM * 4 )
#define TC_GLUE_SIZE (TC_GLUE_NUM * 4 )

#define TC_ANNOTATION_BUFFER (32*1024)

#define MIPS_COMPOSE_I_OP(opCode, rt, rs, immed) ((opCode) << 26 | (rs) << 21 | (rt) << 16 | ((immed) & 0xffff ))
#define TC_BREAKOPCODE  MIPS_COMPOSE_I_OP(cop0_op,0,mfc_op,0);

/* Note we also allocate space for our dynamically
 * generated interface code (see main_run.c) as
 * well as the actual translation caches;
 * we hope that these are within jump range
 */
static void EmbraConfigureTC(void)
{
   int tcSize[3];
   int tcAnnSize[3];
   int pctcSize[3];
   Inst breakOp = TC_BREAKOPCODE;

   /*
    * init TC
    */
   
   tcSize[0] = TC_USER_SIZE;
   tcSize[1] = TC_KERN_SIZE;
   tcSize[2] = TC_GLUE_SIZE;
   tcAnnSize[0] = TC_ANNOTATION_BUFFER;
   tcAnnSize[1] = TC_ANNOTATION_BUFFER;
   tcAnnSize[2] = 2;
   pctcSize[0] = TC_USER_SIZE / (2*8); /* was: 4 * 8 */
   pctcSize[1] = TC_KERN_SIZE / (2*8);
   pctcSize[2] = 2;
   
   TC_init(3,tcSize,pctcSize,tcAnnSize,breakOp);
}

/* This clears all data structures associated with a translation chache */
void Do_Clear_Translation_State(int cache, EmbraState *curEmp )
{
   int i;
   /* Need to clear (now stale) return addresses from inside CPU state */
   /* structures */ 
   for( i = 0; i < TOTAL_CPUS; i++ ) {
      EMP[i].jumpPC = (Reg) continue_run_without_chaining;
   }
   if (cache==TCFLUSH_ALL) {
      /*
       * conflict detected. flush everything.
       * Only at this point can be clear the tccoherence information
       * XXX maybe we should split it by TC cache
       */
      TCcoherence_flush();
      /* For serial (or initialization) code, just flush 'em */
      /* Note we don't want to flush our dynamic code!!
       * Only flush User and Kernel TC's
       */
      TC_flush(TCFLUSH_USER);
      TC_flush(TCFLUSH_KERN);
   }
   else {
      CPUWarning("Do_Clear_Translation_State:only flushing one cache...\n");
      TC_flush(cache);
   }
}

/* This clears all data structures associated with a translation cache */
void Clear_Translation_State(int cache, EmbraState *curEmp )
{
   if (parallel_running) {
      extern int tcFlushFlag;
      /* Wait at barrier!! and flush caches */
      tcFlushFlag = 1;
      DoPeriodicBarrier(curEmp);
   }
   else {
      Do_Clear_Translation_State(cache, curEmp);
   }
}

/*****************************************************************/

/* Debugging function called from continue_run */
void print_pc( int cpuNum, unsigned pc )
{
   if ( EMP[cpuNum].outOfSlaveLoop ) {
      CPUPrint("%d 0x%x\n", EMP[cpuNum].myNum, pc );
      fflush( stdout );
   }
}

void AssertNotStalled(int cpuNum)
{
   VASSERT(!EMP[cpuNum].stalled, ("CpuNum %d\n", cpuNum) );
}

/* This is only used when simulating R3k locking */
/* It doesn't work anymore because of cpuNum */
void TNS( int cpuNum )
{
   int retval;
   volatile int* lock_ptr;

   CalloutLock(cpuNum,12);
   
   lock_ptr = (volatile int*)(EMP[cpuNum].R[4]);
   if ( *lock_ptr & 1 ){
      retval = 0;
   } else {
      *lock_ptr = EMP[cpuNum].R[8];
      retval = 1;
   }
   EMP[cpuNum].R[8] = retval;
   
   CalloutUnlock(cpuNum,12);
}


/* Called when a checkpoint is requested.*/
int EmbraDoCheckpoint(int cpuNum)
{
   VA oldPC[SIM_MAXCPUS];
   int i, result=0;
   
   for (i=0;i<TOTAL_CPUS;i++) { 
      oldPC[i] = EMP[i].PC;
      if ( IS_BACKDOOR(EMP[i].PC) ) {
         /* PC was in the backdoor. In embra, this can only be the
          * case for the processor requesting the checkpoint. 
          * We save the RA rather than the PC
          */
	 EMP[i].PC  = EMP[i].R[REG_RA];
      }
      if (IN_BD(EMP[i].PC)) {
         if ( EmbraAnnType(cpuNum) == ANNFM_PC_TYPE) {
            CPUError("CANNOT TAKE CHECKPOINT IN A BRANCH DELAY with a post-pc annotation -- at least not now \n");
         } 
         EMP[i].PC = CLEAR_BD(EMP[i].PC)-INST_SIZE;
      } else { 
         if (i==CPUVec.CurrentCpuNum() &&
             EmbraAnnType(cpuNum )== ANNFM_PC_TYPE ) {
            EMP[i].PC += INST_SIZE;
         }
      }  
   }

   result = Simcpt_Checkpoint(CPT_SAVE, NULL);
   
   CPUWarning("\n-> Simcheckpoint: Returning from checkpoint save... result=%d\n", result);
   for (i=0;i<TOTAL_CPUS;i++) {
      EMP[i].PC = oldPC[i];
   }

   return result;
}


/* Called when a checkpoint is requested.*/
void EmbraCheckpoint(int cpuNum)
{
   extern int checkpointFlag;

   /* Code doesn't pass in the correct cpuNum */
   cpuNum = GetGlobals()->curEmp->myNum;
   
   /* Do checkpoint at barrier */
   checkpointFlag = 1;
   CalloutUnlock(cpuNum, 0);
   DoPeriodicBarrier(&EMP[cpuNum]);
   CalloutLock(cpuNum, 0); /* Get our lock back! */
}                     


/* *********************************************************************
 * Embra_DoAnn: (TC callout) 
 * Handles PC annotations
 * *********************************************************************/
void Embra_DoAnn( int cpuNum) 
{
   VA annAddr = CLEAR_BD(EMP[cpuNum].PC);
   AnnPtr aptr;

   GetGlobals()->curEmp = &EMP[cpuNum];
   
   /* Why this memory remapping ??? */
#if defined(SIM_MIPS64xxx) 
   if (IS_CKSEG0(annAddr)) {
      annAddr = CKSEG0_TO_XKPHYS(annAddr);
   }
#endif
   CalloutLock(cpuNum,13);
   aptr = AnnFMLookup(annAddr, ANNFM_PC_TYPE);
   ASSERT(aptr);
   /* 
    * Call EmbraAnnExec to trap the case where
    * the annotation has side-effects
    */
   EmbraAnnExec(cpuNum, aptr, ANNFM_PC_TYPE);
   CalloutUnlock(cpuNum,13);
}

void Embra_DoPrePCAnn( int cpuNum) 
{
   VA annAddr = CLEAR_BD(EMP[cpuNum].PC);
   AnnPtr aptr;

   CalloutLock(cpuNum,14);
   aptr = AnnFMLookup(annAddr,ANNFM_PRE_PC_TYPE);
   CalloutUnlock(cpuNum,14);
   ASSERT(aptr);
   /* 
    * Call EmbraAnnExec to trap the case where
    * the annotation has side-effects
    */
   CalloutLock(cpuNum,15);
   EmbraAnnExec(cpuNum,aptr,ANNFM_PRE_PC_TYPE);
   CalloutUnlock(cpuNum,15);
}

static void VerifyOffsets(void)
{
   /* checking a couple of the fields of CPUState */

   ASSERT(  GP_OFF == offsetof(CPUState,R));
   ASSERT(  FP_OFF == offsetof(CPUState,FPR));
   ASSERT(  FCR_OFF == offsetof(CPUState,FCR));
   ASSERT(  CP0_OFF == offsetof(CPUState,CP0));
   ASSERT(  PC_OFF == offsetof(CPUState,PC));
   ASSERT(  HI_OFF == offsetof(CPUState,HI));
   ASSERT(  LO_OFF == offsetof(CPUState,LO));
   ASSERT(  CCD_OFF == offsetof(CPUState,cycleCountdown));
   ASSERT(  BCCD_OFF == offsetof(CPUState,blockCycleCountdown));
   ASSERT(  TQ_OFF == offsetof(CPUState,timeQuantum));
   ASSERT(  MYNUM_OFF == offsetof(CPUState,myNum));
   ASSERT(  MMU_OFF == offsetof(CPUState,mmu));
   ASSERT(  CACHE_OFF == offsetof(CPUState,cache_tag));
   ASSERT(  SDHIT_OFF == offsetof(CPUState,Sdhit_count));
   ASSERT(  SIHIT_OFF == offsetof(CPUState,Sihit_count));
   ASSERT(  SSREG2_OFF == offsetof(CPUState,Ssreg2));
   ASSERT(  NEXT_OFF == offsetof(CPUState,next));
   ASSERT(  OLDPC_OFF == offsetof(CPUState,oldPC));
   ASSERT(  JUMPPC_OFF == offsetof(CPUState,jumpPC));
   ASSERT(  CALL_OFF == offsetof(CPUState,eventQueueTimePtr));
   ASSERT(  HICC_OFF == offsetof(CPUState,cycleCount));
   ASSERT(  LOCC_OFF == offsetof(CPUState,cycleCount)+4);
   ASSERT(  CACHE_AX_OFF == offsetof(CPUState,cache_ax));
   ASSERT(  LLCONTENTS_OFF == offsetof(CPUState,LLContents));
   ASSERT(  LLADDR_OFF == offsetof(CPUState,LLAddr));
   ASSERT(  HACKADDR_OFF == offsetof(CPUState,hackedSavedVaddr));
   ASSERT(  LLBIT_OFF == offsetof(CPUState,LLBit));
   ASSERT(  FPLOADED_OFF == offsetof(CPUState,fpLoaded));
   ASSERT(  OUTTC_OFF == offsetof(CPUState,outTC));
   ASSERT(  STALLED_OFF == offsetof(CPUState,stalled));
   ASSERT(  QCRA_OFF  == offsetof(CPUState,qcra));

}


#define PRINTOFFSET(_f, _off) printf("%s == %d\n", # _f , (int) _off)

void DumpOffsets(void)
{
   /* print offsets of the fields of CPUState */

   PRINTOFFSET(  GP_OFF , offsetof(CPUState,R));
   PRINTOFFSET(  FP_OFF , offsetof(CPUState,FPR));
   PRINTOFFSET(  FCR_OFF , offsetof(CPUState,FCR));
   PRINTOFFSET(  CP0_OFF , offsetof(CPUState,CP0));
   PRINTOFFSET(  PC_OFF , offsetof(CPUState,PC));
   PRINTOFFSET(  HI_OFF , offsetof(CPUState,HI));
   PRINTOFFSET(  LO_OFF , offsetof(CPUState,LO));
   PRINTOFFSET(  CCD_OFF , offsetof(CPUState,cycleCountdown));
   PRINTOFFSET(  BCCD_OFF , offsetof(CPUState,blockCycleCountdown));
   PRINTOFFSET(  TQ_OFF , offsetof(CPUState,timeQuantum));
   PRINTOFFSET(  MYNUM_OFF , offsetof(CPUState,myNum));
   PRINTOFFSET(  MMU_OFF , offsetof(CPUState,mmu));
   PRINTOFFSET(  CACHE_OFF , offsetof(CPUState,cache_tag));
   PRINTOFFSET(  SDHIT_OFF , offsetof(CPUState,Sdhit_count));
   PRINTOFFSET(  SIHIT_OFF , offsetof(CPUState,Sihit_count));
   PRINTOFFSET(  SSREG2_OFF , offsetof(CPUState,Ssreg2));
   PRINTOFFSET(  OLDPC_OFF , offsetof(CPUState,oldPC));
   PRINTOFFSET(  JUMPPC_OFF , offsetof(CPUState,jumpPC));
   PRINTOFFSET(  CALL_OFF , offsetof(CPUState,eventQueueTimePtr));
   PRINTOFFSET(  NEXT_OFF , offsetof(CPUState,next));
   PRINTOFFSET(  HICC_OFF , offsetof(CPUState,cycleCount));
   PRINTOFFSET(  LOCC_OFF , offsetof(CPUState,cycleCount)+4);
   PRINTOFFSET(  QCRA_OFF  , offsetof(CPUState,qcra));
   PRINTOFFSET(  CACHE_AX_OFF , offsetof(CPUState,cache_ax));
   PRINTOFFSET(  LLCONTENTS_OFF , offsetof(CPUState,LLContents));
   PRINTOFFSET(  LLADDR_OFF , offsetof(CPUState,LLAddr));
   PRINTOFFSET(  HACKADDR_OFF , offsetof(CPUState,hackedSavedVaddr));
   PRINTOFFSET(  LLBIT_OFF , offsetof(CPUState,LLBit));
   PRINTOFFSET(  FPLOADED_OFF , offsetof(CPUState,fpLoaded));
   PRINTOFFSET(  OUTTC_OFF , offsetof(CPUState,outTC));
   PRINTOFFSET(  STALLED_OFF , offsetof(CPUState,stalled));
      
}

