/*
 * Copyright (C) 1997 by the Board of Trustees
 *    of Leland Stanford Junior University.
 * 
 * This file is part of the SimOS distribution. 
 * See LICENSE file for terms 
 * of the license. 
 *
 */

/****************************************************************
 * parallel.c
 *
 * This file contains routines which handle thread-level
 * parallelism in embra. Note that since we use macros (our
 * own implementation of the ANL macros) this file actually
 * needs to be run through m4 as part of the build process.
 *
 * We provide a level of indirection so that we can change
 * the parallel implementation as desired.
 * 
 * $Author: verghese $
 * $Date: 1999/09/27 22:08:04 $
 *****************************************************************/


/*
 * Parallel Computation Models
 *
 * For now, we are going to support two different methods on SGI
 * systems. Pthreads are portable, and low-overhead.
 * Sprocs are non-portable and high-overhead, but they support
 * gang scheduling on Irix, which should be a huge win!!
 * I don't know why pthreads don't support gang scheduling.
 * However, gang scheduling is broken on Irix 6.4 anyway...
 */

#define SPROCS 1
#define PTHREADS 0

#define PTHREAD_LOCKS 0

#include <stdlib.h>
#include <errno.h>

#if PTHREADS
#include <pthread.h>
#endif

#if SPROCS
#include <sys/types.h>
#include <sys/prctl.h>
#include <sys/schedctl.h>
#include <sys/sysmp.h>
#include <ulocks.h>
#endif

#include "parallel.h"
#include "embra_globals.h"

/* Constants */

#define NOLOCKEDCPU -2

/* Prototypes that should be elsewhere. Here for now */

extern void _LockEnter(void *lock);
extern void _Unlock(void *lock);
extern void _Barrier(void *barrier, int num_procs);

/* I believe this is the line size which is required by asm_support.s */
#define SOLO_CACHE_LINE_SIZE 128

/*** GLOBALS ***/

#if PTHREADS

pthread_key_t globalKey;	/* thread-specific key for global pointer */
pthread_t	threads[MAX_EMBRA_THREADS];	/* allocate them statically for now??? */

#endif

int parallel_running = 0; /* are we parallel yet? */

static EmbraGlobal *temp_globals = 0;


/* Periodic barrier to prevent us from getting too far out of
 * sync
 */

BARDEC(periodic_barrier)

   /* Lock declaration for synchronizing callouts */

   LOCKDEC(callout_lock)
   LOCKDEC(trans_lock)
   LOCKDEC(tc_lock)
   LOCKDEC(event_lock)
   LOCKDEC(cpucount_lock)
   LOCKDEC(mmu_lock)
   
   /*** Implementation ***/

/* InitParallel: perform initialization of synchronization data structures */
void InitParallel(void)
{
   int i;

   CPUPrint("InitParallel: Initializing\n");
   BARINIT(periodic_barrier);		/* periodic sync barrier */
   LOCKINIT(callout_lock);		/* lock for callouts */
   LOCKINIT(trans_lock);			/* lock for translate()/vcode */
   LOCKINIT(tc_lock);			/* lock for Translation Cache */
   LOCKINIT(event_lock);			/* lock for event queue */
   LOCKINIT(cpucount_lock);		/* lock for cpu count */
   LOCKINIT(mmu_lock)
#if SPROCS
   if (usconfig(CONF_INITUSERS, 128) < 0) {
      CPUError("InitParallel: Error calling usconfig\n");
   }
#endif
#if PTHREADS
   pthread_key_create(&globalKey, NULL);	/* access to globals */
#endif

}


/* CreateThread: Create a new thread starting at function void *start() */

void CreateThread(void (*start)(void *), EmbraGlobal *globalPtr)
{
   int id = globalPtr - EmbraGlobals; /* XXX BL: fix this!!! cheesoid */

#if PTHREADS
   /* it's nice to be able to look at these statically */
   pthread_t *thread = &threads[id];
#endif

   /* Give the thread its globals! */
   SetGlobals(id, globalPtr);

   /* Spawn a new thread */
   globalPtr->id = id;
   CPUPrint("CreateThread: spawning thread %d\n", globalPtr->id);
  
   /* This would be a race condition, but we're not parallel yet! */
   parallel_running = 1;
  
#if PTHREADS
   pthread_create(thread, NULL, start, (void *) globalPtr);
#endif
  
#if SPROCS
   {
      int pid;

      /* Make an Sproc */
      pid = sproc(start,  PR_SALL, (void *) globalPtr);
      if (pid < 0) {
	 CPUError("CreateThread: sproc failed with error %d (%s)\n",oserror(),
		  strerror(oserror()));
      }

      /* This seems to slow things down */
      if (0) {
	 /* Run thread id on cpu id, always! */
	 sysmp(MP_MUSTRUN_PID, id, pid);
	 /* run me wherever I belong */
	 sysmp(MP_MUSTRUN, GetGlobals()->id);
      }

      /* Gang scheduling */
      if (1) {
	 schedctl(SCHEDMODE, SGS_GANG);
      }
      
   }
#endif
  
}

void ExitThread(int cpu)
{
   EmbraGlobal *g = GetGlobals();
   CPUPrint("ExitThread: Thread %d exiting\n", g->id);
#if PTHREADS
   /* Detach for now */
   pthread_detach(threads[g->id]);
   /* Terminate the calling thread */
   pthread_exit((void *)g);
#elif SPROCS
   exit(0);
#endif
}

 
/* SetGlobals(): set pointer to thread-specific globals */
void SetGlobals (int cpu, EmbraGlobal *globals)
{
#if PTHREADS
   pthread_setspecific(globalKey, (void *) globals);
#elif SPROCS
   /* SGI SPECIFIC: store global pointer in process/thread-local data area */
   *((EmbraGlobal **)PRDA->usr_prda.fill) = globals;
#endif
}

/* GetGlobals(); return a pointer to thread-specific globals */
EmbraGlobal *GetGlobals(void)
{
#if PTHREADS
   EmbraGlobal *globals = (EmbraGlobal *) pthread_getspecific(globalKey);
   if (!globals) globals = &EmbraGlobals[0];
#elif SPROCS
   /* SGI SPECIFIC: get global pointer from process/thread-local data area */
   EmbraGlobal *globals = *((EmbraGlobal **)PRDA->usr_prda.fill);
#endif
   return globals;
}

/* These variables are for the reentrant CalloutLock() routine;
 * Also we keep track of who owns the lock, which allows callouts
 * to query the "current" CPU.
 */

int lockCPU = NOLOCKEDCPU;
int lockWhere[100];
int lockDepth = 0;


#if PTHREAD_LOCKS
pthread_mutex_t callout_mutex = PTHREAD_MUTEX_INITIALIZER;
#endif

/* CalloutLock: acquire callout lock
 * XXX This is our main callout path from Embra.
 * Note that this lock is reentrant (and expensive.)
 */
void CalloutLock(int cpu, int where)
{
   /* Are we not maintaining curEmp correctly?? */
   GetGlobals()->curEmp = &EMP[cpu];
   /* Allow us to reenter this lock!! ;-( */
   if (lockCPU != cpu) {
      /* Acquire the lock */
#if PTHREAD_LOCKS
      pthread_mutex_lock(&callout_mutex);
#else
      LOCK(callout_lock);
#endif
      lockCPU = cpu;
   }
   lockWhere[lockDepth] = where;
   lockDepth++;
    
}

void MMULock(void)
{
   LOCK(mmu_lock);
}

void MMUUnlock(void)
{
   UNLOCK(mmu_lock);
}

/* CalloutUnlock: release callout lock */
void CalloutUnlock(int cpu, int where)
{
   ASSERT(lockCPU == cpu);
   lockDepth--;
   if (lockWhere[lockDepth] != where) {
      CPUPrint("CalloutUnlock: unlock site (%d) != lock site (%d)\n",
		 lockWhere[lockDepth],where);
   }
   ASSERT(lockDepth >= 0);
   if (lockDepth == 0) {
      lockCPU = NOLOCKEDCPU;
#if PTHREAD_LOCKS
      pthread_mutex_unlock(&callout_mutex);
#else
      UNLOCK(callout_lock);
#endif
   }
}

  
/* XXX This is a villainous function whose purpose is to satisfy
 * those wicked callouts which require a notion of "current CPU"
 * So, we return the CPU who owns the callout lock
 */
int calloutCpuNum(void)
{
   ASSERT(lockCPU != NOLOCKEDCPU);
   return lockCPU;
}

#if PTHREAD_LOCKS
pthread_mutex_t event_mutex = PTHREAD_MUTEX_INITIALIZER;
#endif

/* unfortunately, NOLOCKEDCPU = serial mode */

#define EVENT_BOGUS -2
int event_cpu = EVENT_BOGUS;

void EventLock(int cpu)
{
   /*  CalloutLock(); */
   if (!parallel_running) return;
  
   ASSERT(cpu != event_cpu);
#if PTHREAD_LOCKS
   pthread_mutex_lock(&event_mutex);
#else
   LOCK(event_lock)
#endif
      event_cpu = cpu;
}

void EventUnlock(int cpu)
{
   if (!parallel_running) return;
  
   ASSERT(cpu == event_cpu);
   event_cpu = EVENT_BOGUS;
#if PTHREAD_LOCKS
   pthread_mutex_unlock(&event_mutex);
#else
   UNLOCK(event_lock)
#endif

      /*     CalloutUnlock(); */
  
      }

/*
 * OK, for now we are going to have separate locks for
 * translation (i.e. vcode, etc.)
 * and messing with the translation cache itself!!
 *
 * I'm still not happy about FindTCA having to go through
 * a lock - ugh!!!
 */

#if PTHREAD_LOCKS
pthread_mutex_t trans_mutex = PTHREAD_MUTEX_INITIALIZER;
#endif
int trans_who = NOLOCKEDCPU;
int trans_where = NOLOCKEDCPU;

void TransLock(int cpu, int where)
{
   /*   CalloutLock(); */
  
   if (trans_who == cpu)
      CPUError("TransLock: trans_who %d == cpu %d, trans_where = %d, "
	       "%where = %d\n",
	       trans_who, cpu, trans_where, where);
#if PTHREAD_LOCKS
   pthread_mutex_lock(&trans_mutex);
#else
   LOCK(trans_lock)
#endif
      trans_who = cpu;
   trans_where = where;
}

void TransUnlock(int cpu, int where)
{
   ASSERT(trans_who == cpu && trans_where == where);
   trans_who = NOLOCKEDCPU;
#if PTHREAD_LOCKS
   pthread_mutex_unlock(&trans_mutex);
#else
   UNLOCK(trans_lock)
#endif
      /*    CalloutUnlock(); */
  
      }

#if PTHREAD_LOCKS
pthread_mutex_t tc_mutex = PTHREAD_MUTEX_INITIALIZER;
#endif

void TCLock(int cpu, int where)
{
#if PTHREAD_LOCKS 
   pthread_mutex_lock(&tc_mutex);
#else
   LOCK(tc_lock);
#endif
}

void TCUnlock(int cpu, int where)
{
#if PTHREAD_LOCKS  
   pthread_mutex_unlock(&tc_mutex);
#else
   UNLOCK(tc_lock);
#endif
}

/* Periodic barrier; every so often, we wait to make sure that
 * the other CPU's catch up to a certain time (in cycles.) This
 * guarantees that we won't get too far out of sync.
 *
 */

void PeriodicBarrier(int threadCount)
{
   /* We may be hosed if this doesn't work with threads */
   /* Note that this barrier waits until threadCount threads
    * have reached the barrier
    */
   BARRIER(periodic_barrier, embra.parallel);
}

/* Active cpu count -> how many cpu's are running???
 * (i.e. not in slave loop); note this is *different* than
 * the number of threads that are running. We may have no
 * threads at all, but lots of active cpu's which have
 * already been launched
 */

int activeCpuCount = 1; /* one is running to start off */

#if PTHREAD_LOCKS
pthread_mutex_t active_mutex = PTHREAD_MUTEX_INITIALIZER;
#endif

void IncActiveCpuCount(void)
{
#if PTHREAD_LOCKS
   pthread_mutex_lock(&active_mutex);
#else
   LOCK(cpucount_lock);
#endif
   activeCpuCount++;
#if PTHREAD_LOCKS
   pthread_mutex_unlock(&active_mutex);
#else
   UNLOCK(cpucount_lock);
#endif
}


/* I think this routine is useless */
void DecActiveCpuCount(void)
{
#if PTHREAD_LOCKS
   pthread_mutex_lock(&active_mutex);
#else
   LOCK(cpucount_lock);
#endif
   activeCpuCount--;
#if PTHREAD_LOCKS
   pthread_mutex_unlock(&active_mutex);
#else
   UNLOCK(cpucount_lock);
#endif
}


