/*
 * Copyright (C) 1996-1998 by the Board of Trustees
 *    of Leland Stanford Junior University.
 * 
 * This file is part of the SimOS distribution. 
 * See LICENSE file for terms of the license. 
 *
 */

/*****************************************************************
 * fpu.c
 * 
 * $Author: verghese $
 * $Date: 1999/09/27 22:08:13 $
 *****************************************************************/
#include <math.h>
#include <limits.h>
#ifdef __alpha
#include <float.h>
#else
# ifndef i386
# include <ieeefp.h>
# endif
#endif
#ifdef sparc
#define sqrtf(_x) ((float) sqrt((double) _x))
#endif
#include "mipsy.h"
#include "cpu.h"
#include "cpu_state.h"
#include "sim_error.h"
#include "machine_params.h"
#include "cp0.h"
#include "fpu.h"
#include "cpu_stats.h"

extern Result MipsyPrefetch(CPUState *P, VA vAddr, uint hint);
/*
 * CHECK_64BIT_ALLOWED() - Check to see if 64bit instructions are permitted,
 *                         if not raise an exception. 
 */
#if defined(SIM_MIPS64)
#define CHECK_64BIT_ALLOWED(_p)  \
      if ((_p)->is32bitMode && !IS_KERNEL_MODE((_p))) {  \
         EXCEPTION((_p), EXC_II);                 \
         return C1_FAILURE;                             \
      }
#else
#define CHECK_64BIT_ALLOWED(_p) \
      if (1) {                                           \
         EXCEPTION((_p), EXC_II);                 \
         return C1_FAILURE;                             \
      }

#endif

static Result CompareOp(CPUState *P, int cond, int fcc);

/* Enumeration of floating point exception types */
typedef enum { FPE_I = 0, FPE_U, FPE_O, FPE_Z, FPE_V, FPE_E, FPECount } FPUException;
static int FPCSRFlag[FPECount] = {
   0x00000004, 0x00000008, 0x00000010, 0x00000020, 0x00000040, 0x00000000
};
static int FPCSREnable[FPECount] = {
   0x00000080, 0x00000100, 0x00000200, 0x00000400, 0x00000800, 0x00000000
};
static int FPCSRCause[FPECount] = {
   0x00001000, 0x00002000, 0x00004000, 0x00008000, 0x00010000, 0x00020000
};

/*****************************************************************
 * WARNING: An ugly feature of the sgi IEEE fp implementation greatly
 * complicates this c-based fpu model. FP operations on doubles take
 * place in a reversed order in the registers than they would in
 * memory. As a result, operations on doubles in my memory-based fp
 * registers would all be word-swapped. As a result, I swap every
 * single float reference so that reg 1 is in reg 0's place in memory
 * and vice versa. It's a pain, but will speed up the double
 * operations. 
 * 
 * There are a few side effects of this change:
 * -> The "real" fpu state must be copied into mipsy's special
 *    structure and out of it on entry and exit.
 * -> Gdb get reg needs to be aware of the special state
 * -> People debugging the fpu should be aware that FPR doesn't
 *    have the most up-to-date copy!!!
 ****************************************************************/
 
void
MipsyInitFPU(void)
{
   CPUState *P;
   int i,j;
   
   for (i=0; i < TOTAL_CPUS; i++) {
      P = &PE[i];
      for (j=0; j < NUM_FP_REGS; j++) {
#ifdef SIM_MIPS32
         IntReg(j) = P->FPR[j];
#else
         LongLongReg(j) = P->FPR[j];
#endif
      }
   }
}

void 
MipsyExitFPU(void)
{
   CPUState *P;
   int i,j;
   
   for (i=0; i < TOTAL_CPUS; i++) {
      P = &PE[i];
      for (j=0; j < NUM_FP_REGS; j++) {
#ifdef SIM_MIPS32
         P->FPR[j] = IntReg(j);
#else
         P->FPR[j] = LongLongReg(j);
#endif
      }
   }
}

static Result
MipsyFPUException(CPUState *P, FPUException type)
{
   P->FCR[31] |= FPCSRFlag[type];
   P->FCR[31] |= FPCSRCause[type];
   if (P->FCR[31] & FPCSREnable[type]) {
      EXCEPTION(P, EXC_FPE);
      return C1_FAILURE;
   } else {
      return C1_SUCCESS;
   }
}

typedef Result (*C1OpcodeFPtr)(CPUState *, Inst, uint, uint, uint);
typedef Result (*C1XOpcodeFPtr)(CPUState *, Inst, uint, uint, uint, uint);

typedef struct c1op_desc {
   char        *opname;
   C1OpcodeFPtr func;
   uint         num;
} MipsC1OpcodeDesc;

typedef struct c1xop_desc {
   char         *opname;
   C1XOpcodeFPtr func;
   uint          num;
} MipsC1XOpcodeDesc;

MipsC1OpcodeDesc mipsC1SPOpcodes[];
MipsC1OpcodeDesc mipsC1DPOpcodes[];
MipsC1XOpcodeDesc mipsC1XOpcodes[];

#define COP1OP(NAME) \
Result NAME(CPUState *P, Inst instr, uint fd, uint fs, uint ft)

#define COP1XOP(NAME) \
Result NAME(CPUState *P, Inst instr, uint fd, uint fr, uint fs, uint ft)

/*****************************************************************
 * Single-precision floating-point opcodes
 *****************************************************************/

COP1OP(abs_sOp)
{
   float contents_fs = FloatReg(fs);
   FloatReg(fd) = (contents_fs >= 0.0 ? contents_fs : -contents_fs);
   SIM_DEBUG(('f', "C1:ABS.S: abs(f%d %e) = f%d %e\n",
              fs, contents_fs, fd, FloatReg(fd)));
   return C1_SUCCESS;
}
COP1OP(add_sOp)
{
   float contents_fs = FloatReg(fs);
   float contents_ft = FloatReg(ft);
   FloatReg(fd) = contents_fs + contents_ft;
   SIM_DEBUG(('f', "C1:ADD.S: f%d %e + f%d %e = f%d %e\n",
              fs, contents_fs, ft, contents_ft, fd, FloatReg(fd)));
   return C1_SUCCESS;
}
COP1OP(cvt_d_sOp)
{
   float contents_fs = FloatReg(fs);
   DoubleReg(fd) = contents_fs;
   SIM_DEBUG(('f', "C1:CVT.D.S: f%d %e --> f%d %e\n",
              fs, contents_fs, ft, DoubleReg(fd)));
   return C1_SUCCESS;
}
COP1OP(cvt_l_sOp)
{
   float contents_fs = FloatReg(fs);
   uint roundMode = P->FCR[31] & 0x03;
   float rounded;
   CHECK_64BIT_ALLOWED(P);
   switch (roundMode) {
   case 0:                 /* RN: Round to nearest         */
      rounded = (contents_fs > 0.0 ? contents_fs + 0.5 : contents_fs - 0.5);
      break;
   case 1:                 /* RZ: Round toward 0           */
      rounded = contents_fs;
      break;
   case 2:                 /* RP: Round to +infinity       */
      rounded = (contents_fs >= 0.0 ? contents_fs + SP_NEAR_ONE : contents_fs);
      break;
   case 3:                 /* RM: Round to -infinity       */
      rounded = (contents_fs >= 0.0 ? contents_fs : contents_fs - SP_NEAR_ONE);
      break;
   }

   LongLongReg(fd) = rounded;
   SIM_DEBUG(('f', "C1:CVT.L.S: f%d %e --> f%d %lld rm = %d\n", 
              fs, contents_fs, fd, LongLongReg(fd), roundMode));

   /* Check for valid operand */
   if ((rounded > (float)LONGLONG_MAX) || (rounded < (float)LONGLONG_MIN)) {
      Result ret = MipsyFPUException(P, FPE_V);
      return ret;
   }
   return C1_SUCCESS;
}   

COP1OP(cvt_w_sOp)
{
   float contents_fs = FloatReg(fs);
   uint roundMode = P->FCR[31] & 0x03;
   float rounded;
   switch (roundMode) {
   case 0:                 /* RN: Round to nearest         */
      rounded = (contents_fs > 0.0 ? contents_fs + 0.5 : contents_fs - 0.5);
      break;
   case 1:                 /* RZ: Round toward 0           */
      rounded = contents_fs;
      break;
   case 2:                 /* RP: Round to +infinity       */
      rounded = (contents_fs >= 0.0 ? contents_fs + SP_NEAR_ONE : contents_fs);
      break;
   case 3:                 /* RM: Round to -infinity       */
      rounded = (contents_fs >= 0.0 ? contents_fs : contents_fs - SP_NEAR_ONE);
      break;
   }

   IntReg(fd) = rounded;
   SIM_DEBUG(('f', "C1:CVT.W.S: f%d %e --> f%d %d rm = %d\n", 
              fs, contents_fs, fd, IntReg(fd), roundMode));

   /* Check for valid operand */
   if ((rounded > (float)INT_MAX) || (rounded < (float)INT_MIN)) {
      Result ret = MipsyFPUException(P, FPE_V);
      return ret;
   }
   return C1_SUCCESS;
}

COP1OP(div_sOp)
{
   float contents_fs = FloatReg(fs);
   float contents_ft = FloatReg(ft);
   if (contents_ft == 0.0) {
      CPUWarning("C1:DIV.S: divide by zero\n");
   }
   FloatReg(fd) = contents_fs / contents_ft;
   SIM_DEBUG(('f', "C1:DIV.S: f%d %e / f%d %e = f%d %e\n", 
              fs, contents_fs, ft, contents_ft, fd, FloatReg(fd)));
   return C1_SUCCESS;
}

COP1OP(mov_sOp)
{
   float contents_fs = FloatReg(fs);
   FloatReg(fd) = contents_fs;
   SIM_DEBUG(('f', "C1:MOV.S: f%d %e --> f%d %e\n", 
              fs, contents_fs, fd, FloatReg(fd)));
   return C1_SUCCESS;
}
COP1OP(movcf_sOp)
{
   float contents_fs = FloatReg(fs);
   int fcc = MOVC_FCC(instr);
   
   /* Move conditional on FP condition */
   if (TF(instr)) { 
      if (P->FCR[31] & M_csr_fcc(fcc)) {
         FloatReg(fd) = contents_fs;
         SIM_DEBUG(('f', "C1:MOVT.S: fcc %d T, f%d %e --> f%d %e\n", 
                    fcc, fs, contents_fs, fd, FloatReg(fd)));
      } else {
         SIM_DEBUG(('f', "C1:MOVT.S: fcc %d F, f%d %e unchanged\n",
                    fcc, fd, FloatReg(fd)));
      }
   } else {
      if (!(P->FCR[31] & M_csr_fcc(fcc))) {
         FloatReg(fd) = contents_fs;
         SIM_DEBUG(('f', "C1:MOVF.S: fcc %d F, f%d %e --> f%d %e\n", 
                    fcc, fs, contents_fs, fd, FloatReg(fd)));
      } else {
         SIM_DEBUG(('f', "C1:MOVF.S: fcc %d T, f%d %e unchanged\n",
                    fcc, fd, FloatReg(fd)));
      }
   }
   return C1_SUCCESS;
}

COP1OP(movn_sOp)
{
   float contents_fs = FloatReg(fs);
   Reg contents_rt = P->R[ft];
   if (contents_rt != 0) {
      FloatReg(fd) = contents_fs;
      SIM_DEBUG(('f', "C1:MOVN.S: r%d %lld, f%d %e to f%d %e\n", 
                 ft, (uint64) contents_rt, 
                 fs, contents_fs, fd, FloatReg(fd)));
   } else {
      SIM_DEBUG(('f', "C1:MOVN.S: r%d %lld, f%d %e unchanged\n", 
                 ft, (int64) contents_rt, fd, FloatReg(fd)));
   }
   return C1_SUCCESS;
}

COP1OP(movz_sOp)
{
   float contents_fs = FloatReg(fs);
   Reg contents_rt = P->R[ft];
   if (contents_rt == 0) {
      FloatReg(fd) = contents_fs;
      SIM_DEBUG(('f', "C1:MOVZ.S: r%d %lld, f%d %e to f%d %e\n", 
                 ft, (int64) contents_rt, 
                 fs, contents_fs, fd, FloatReg(fd)));
   } else {
      SIM_DEBUG(('f', "C1:MOVZ.S: r%d %lld, f%d %e unchanged\n", 
                 ft, (uint64) contents_rt, fd, FloatReg(fd)));
   }
   return C1_SUCCESS;
}

COP1OP(mul_sOp)
{
   float contents_fs = FloatReg(fs);
   float contents_ft = FloatReg(ft);
   FloatReg(fd) = contents_fs * contents_ft;
   SIM_DEBUG(('f', "C1:MUL.S: f%d %e * f%d %e = f%d %e\n", 
              fs, contents_fs, ft, contents_ft, fd, FloatReg(fd)));
   return C1_SUCCESS;
}

COP1OP(neg_sOp)
{
   float contents_fs = FloatReg(fs);
   FloatReg(fd) = -contents_fs;
   SIM_DEBUG(('f', "C1:NEG.S: f%d %e --> f%d %e\n", 
              fs, contents_fs, fd, FloatReg(fd)));
   return C1_SUCCESS;
}

COP1OP(sub_sOp)
{
   float contents_fs = FloatReg(fs);
   float contents_ft = FloatReg(ft);
   FloatReg(fd) = contents_fs - contents_ft;
   SIM_DEBUG(('f', "C1:SUB.S: f%d %e - f%d %e = f%d %e\n", 
              fs, contents_fs, ft, contents_ft, fd, FloatReg(fd)));
   return C1_SUCCESS;
}

COP1OP(sqrt_sOp)
{
   float contents_fs = FloatReg(fs);
   FloatReg(fd) = sqrtf(contents_fs);
   SIM_DEBUG(('f', "C1:SQRT.S: sqrt(f%d %e) = f%d %e\n", 
              fs, contents_fs, fd, FloatReg(fd)));
   return C1_SUCCESS;
}

COP1OP(trunc_w_sOp)
{
   float contents_fs = FloatReg(fs);
   IntReg(fd) = contents_fs;
   SIM_DEBUG(('f', "C1:TRUNC.W.S: f%d %e --> f%d %d\n", 
              fs, contents_fs, fd, IntReg(fd)));
   return C1_SUCCESS;
}

COP1OP(trunc_l_sOp)
{
   float contents_fs = FloatReg(fs);
   LongLongReg(fd) = contents_fs;
   SIM_DEBUG(('f', "C1:TRUNC.L.S: f%d %e --> f%d %lld\n", 
              fs, contents_fs, fd, LongLongReg(fd)));
   return C1_SUCCESS;
}

COP1OP(floor_w_sOp)
{
   float contents_fs = FloatReg(fs);
   IntReg(fd) = (contents_fs >= 0.0 ? 
                       contents_fs : 
                       contents_fs - SP_NEAR_ONE);
   SIM_DEBUG(('f', "C1:FLOOR.W.S: f%d %e --> f%d %d\n", 
              fs, contents_fs, fd, IntReg(fd)));
   return C1_SUCCESS;
}

COP1OP(floor_l_sOp)
{
   float contents_fs = FloatReg(fs);
   LongLongReg(fd) = (contents_fs >= 0.0 ? 
                            contents_fs : 
                            contents_fs - SP_NEAR_ONE);
   SIM_DEBUG(('f', "C1:FLOOR.L.S: f%d %e --> f%d %lld\n", 
              fs, contents_fs, fd, LongLongReg(fd)));
   return C1_SUCCESS;
}

COP1OP(recip_sOp)
{
   float contents_fs = FloatReg(fs);
   if (contents_fs == 0.0) {
      CPUWarning("C1:RECIP.S: divide by zero\n");
   }
   FloatReg(fd) = 1.0 / contents_fs;
   SIM_DEBUG(('f', "C1:RECIP.S: 1.0 / f%d %e = f%d %e\n", 
              fs, contents_fs, fd, FloatReg(fd)));
   return C1_SUCCESS;
}

COP1OP(rsqrt_sOp)
{
   float contents_fs = FloatReg(fs);
   if (contents_fs == 0.0) {
      CPUWarning("C1:RSQRT.S: divide by zero\n");
   }
   FloatReg(fd) = 1.0 / sqrtf(contents_fs);
   SIM_DEBUG(('f', "C1:RSQRT.S: 1.0 / sqrt(f%d %e) = f%d %e\n", 
              fs, contents_fs, fd, FloatReg(fd)));
   return C1_SUCCESS;
}

COP1OP(round_w_sOp)
{
   float contents_fs = FloatReg(fs);
   IntReg(fd) = (contents_fs > 0.0 ? 
                       contents_fs + 0.5 : 
                       contents_fs - 0.5);
   SIM_DEBUG(('f', "C1:ROUND.W.S: f%d %e --> f%d %d\n", 
              fs, contents_fs, fd, IntReg(fd)));
   return C1_SUCCESS;
}

COP1OP(round_l_sOp)
{
   float contents_fs = FloatReg(fs);
   LongLongReg(fd) = (contents_fs > 0.0 ? 
                            contents_fs + 0.5 : 
                            contents_fs - 0.5);
   SIM_DEBUG(('f', "C1:ROUND.W.S: f%d %e --> f%d %lld\n", 
              fs, contents_fs, fd, LongLongReg(fd)));
   return C1_SUCCESS;
}

COP1OP(ceil_w_sOp)
{
   float contents_fs = FloatReg(fs);
   IntReg(fd) = (contents_fs > 0.0 ? 
                       contents_fs + SP_NEAR_ONE : 
                       contents_fs);
   SIM_DEBUG(('f', "C1:CEIL.W.S: f%d %e --> f%d %d\n", 
              fs, contents_fs, fd, IntReg(fd)));
   return C1_SUCCESS;
}

COP1OP(ceil_l_sOp)
{
   float contents_fs = FloatReg(fs);
   LongLongReg(fd) = (contents_fs > 0.0 ? 
                            contents_fs + SP_NEAR_ONE : 
                            contents_fs);
   SIM_DEBUG(('f', "C1:CEIL.L.S: f%d %e --> f%d %lld\n", 
              fs, contents_fs, fd, LongLongReg(fd)));
   return C1_SUCCESS;
}

COP1OP(c_eq_sOp)
{
   float contents_fs = FloatReg(fs);
   float contents_ft = FloatReg(ft);
   int cond = (contents_fs == contents_ft);
   int fcc = C_FCC(instr);
   SIM_DEBUG(('f', "C1:C.EQ.S: f%d %e, f%d %e --> fcc%d %d\n",
              fs, contents_fs, ft, contents_ft, fcc, cond));
   return CompareOp(P, cond, fcc);
}

COP1OP(c_f_sOp)
{
   float contents_fs = FloatReg(fs);
   float contents_ft = FloatReg(ft);
   int cond = 0;
   int fcc = C_FCC(instr);
   SIM_DEBUG(('f', "C1:C.F.S: f%d %e, f%d %e --> fcc%d %d\n",
              fs, contents_fs, ft, contents_ft, fcc, cond));
   return CompareOp(P, cond, fcc);
}

COP1OP(c_le_sOp)
{
   float contents_fs = FloatReg(fs);
   float contents_ft = FloatReg(ft);
   int cond = (contents_fs <= contents_ft);
   int fcc = C_FCC(instr);
   SIM_DEBUG(('f', "C1:C.LE.S: f%d %e, f%d %e --> fcc%d %d\n",
              fs, contents_fs, ft, contents_ft, fcc, cond));
   return CompareOp(P, cond, fcc);
}

COP1OP(c_lt_sOp)
{
   float contents_fs = FloatReg(fs);
   float contents_ft = FloatReg(ft);
   int cond = (contents_fs < contents_ft);
   int fcc = C_FCC(instr);
   SIM_DEBUG(('f', "C1:C.LT.S: f%d %e, f%d %e --> fcc%d %d\n",
              fs, contents_fs, ft, contents_ft, fcc, cond));
   return CompareOp(P, cond, fcc);
}

COP1OP(c_nge_sOp)
{
   float contents_fs = FloatReg(fs);
   float contents_ft = FloatReg(ft);
   int cond = (contents_fs <= contents_ft);
   int fcc = C_FCC(instr);
   SIM_DEBUG(('f', "C1:C.NGE.S: f%d %e, f%d %e --> fcc%d %d\n",
              fs, contents_fs, ft, contents_ft, fcc, cond));
   return CompareOp(P, cond, fcc);
}

COP1OP(c_ngl_sOp)
{
   float contents_fs = FloatReg(fs);
   float contents_ft = FloatReg(ft);
   int cond = (contents_fs == contents_ft);
   int fcc = C_FCC(instr);
   SIM_DEBUG(('f', "C1:C.NGL.S: f%d %e, f%d %e --> fcc%d %d\n",
              fs, contents_fs, ft, contents_ft, fcc, cond));
   return CompareOp(P, cond, fcc);
}

COP1OP(c_ngle_sOp)
{
   float contents_fs = FloatReg(fs);
   float contents_ft = FloatReg(ft);
   int cond = 0;
   int fcc = C_FCC(instr);
   SIM_DEBUG(('f', "C1:C.NGLE.S: f%d %e, f%d %e --> fcc%d %d\n",
              fs, contents_fs, ft, contents_ft, fcc, cond));
   return CompareOp(P, cond, fcc);
}

COP1OP(c_ngt_sOp)
{
   float contents_fs = FloatReg(fs);
   float contents_ft = FloatReg(ft);
   int cond = (contents_fs <= contents_ft);
   int fcc = C_FCC(instr);
   SIM_DEBUG(('f', "C1:C.NGT.S: f%d %e, f%d %e --> fcc%d %d\n",
              fs, contents_fs, ft, contents_ft, fcc, cond));
   return CompareOp(P, cond, fcc);
}

COP1OP(c_ole_sOp)
{
   float contents_fs = FloatReg(fs);
   float contents_ft = FloatReg(ft);
   int cond = (contents_fs < contents_ft);
   int fcc = C_FCC(instr);
   SIM_DEBUG(('f', "C1:C.OLE.S: f%d %e, f%d %e --> fcc%d %d\n",
              fs, contents_fs, ft, contents_ft, fcc, cond));
   return CompareOp(P, cond, fcc);
}

COP1OP(c_olt_sOp)
{
   float contents_fs = FloatReg(fs);
   float contents_ft = FloatReg(ft);
   int cond = (contents_fs < contents_ft);
   int fcc = C_FCC(instr);
   SIM_DEBUG(('f', "C1:C.OLT.S: f%d %e, f%d %e --> fcc%d %d\n",
              fs, contents_fs, ft, contents_ft, fcc, cond));
   return CompareOp(P, cond, fcc);
}

COP1OP(c_seq_sOp)
{
   float contents_fs = FloatReg(fs);
   float contents_ft = FloatReg(ft);
   int cond = (contents_fs == contents_ft);
   int fcc = C_FCC(instr);
   SIM_DEBUG(('f', "C1:C.SEQ.S: f%d %e, f%d %e --> fcc%d %d\n",
              fs, contents_fs, ft, contents_ft, fcc, cond));
   return CompareOp(P, cond, fcc);
}

COP1OP(c_sf_sOp)
{
   float contents_fs = FloatReg(fs);
   float contents_ft = FloatReg(ft);
   int cond = 0;
   int fcc = C_FCC(instr);
   SIM_DEBUG(('f', "C1:C.SF.S: f%d %e, f%d %e --> fcc%d %d\n",
              fs, contents_fs, ft, contents_ft, fcc, cond));
   return CompareOp(P, cond, fcc);
}

COP1OP(c_ueq_sOp)
{
   float contents_fs = FloatReg(fs);
   float contents_ft = FloatReg(ft);
   int cond = (contents_fs == contents_ft);
   int fcc = C_FCC(instr);
   SIM_DEBUG(('f', "C1:C.UEQ.S: f%d %e, f%d %e --> fcc%d %d\n",
              fs, contents_fs, ft, contents_ft, fcc, cond));
   return CompareOp(P, cond, fcc);
}

COP1OP(c_ule_sOp)
{
   float contents_fs = FloatReg(fs);
   float contents_ft = FloatReg(ft);
   int cond = (contents_fs <= contents_ft);
   int fcc = C_FCC(instr);
   SIM_DEBUG(('f', "C1:C.ULE.S: f%d %e, f%d %e --> fcc%d %d\n",
              fs, contents_fs, ft, contents_ft, fcc, cond));
   return CompareOp(P, cond, fcc);
}

COP1OP(c_ult_sOp)
{
   float contents_fs = FloatReg(fs);
   float contents_ft = FloatReg(ft);
   int cond = (contents_fs < contents_ft);
   int fcc = C_FCC(instr);
   SIM_DEBUG(('f', "C1:C.ULT.S: f%d %e, f%d %e --> fcc%d %d\n",
              fs, contents_fs, ft, contents_ft, fcc, cond));
   return CompareOp(P, cond, fcc);
}

COP1OP(c_un_sOp)
{
   float contents_fs = FloatReg(fs);
   float contents_ft = FloatReg(ft);
   int cond = 0;
   int fcc = C_FCC(instr);
   SIM_DEBUG(('f', "C1:C.UN.S: f%d %e, f%d %e --> fcc%d %d\n",
              fs, contents_fs, ft, contents_ft, fcc, cond));
   return CompareOp(P, cond, fcc);
}

/*****************************************************************
 * Double-precision floating-point opcodes
 *****************************************************************/

COP1OP(abs_dOp)
{
   double contents_fs = DoubleReg(fs);
   DoubleReg(fd) = (contents_fs >= 0.0 ? contents_fs : -contents_fs);
   SIM_DEBUG(('f', "C1:ABS.D: abs(f%d %e) = f%d %e\n",
              fs, contents_fs, fd, DoubleReg(fd)));
   return C1_SUCCESS;
}
COP1OP(add_dOp)
{
   double contents_fs = DoubleReg(fs);
   double contents_ft = DoubleReg(ft);
   DoubleReg(fd) = contents_fs + contents_ft;
   SIM_DEBUG(('f', "C1:ADD.D: f%d %e + f%d %e = f%d %e\n",
              fs, contents_fs, ft, contents_ft, fd, DoubleReg(fd)));
   return C1_SUCCESS;
}
COP1OP(cvt_s_dOp)
{
   double contents_fs = DoubleReg(fs);
   FloatReg(fd) = contents_fs;
   SIM_DEBUG(('f', "C1:CVT.S.D: f%d %e --> f%d %e\n",
              fs, contents_fs, ft, FloatReg(fd)));
   return C1_SUCCESS;
}
COP1OP(cvt_l_dOp)
{
   double contents_fs = DoubleReg(fs);
   uint roundMode = P->FCR[31] & 0x03;
   double rounded;
   CHECK_64BIT_ALLOWED(P);
   switch (roundMode) {
   case 0:                 /* RN: Round to nearest         */
      rounded = (contents_fs > 0.0 ? contents_fs + 0.5 : contents_fs - 0.5);
      break;
   case 1:                 /* RZ: Round toward 0           */
      rounded = contents_fs;
      break;
   case 2:                 /* RP: Round to +infinity       */
      rounded = (contents_fs >= 0.0 ? contents_fs + SP_NEAR_ONE : contents_fs);
      break;
   case 3:                 /* RM: Round to -infinity       */
      rounded = (contents_fs >= 0.0 ? contents_fs : contents_fs - SP_NEAR_ONE);
      break;
   }

   LongLongReg(fd) = rounded;
   SIM_DEBUG(('f', "C1:CVT.L.D: f%d %e --> f%d %lld rm = %d\n", 
              fs, contents_fs, fd, LongLongReg(fd), roundMode));

   /* Check for valid operand */
   if ((rounded > (float)LONGLONG_MAX) || (rounded < (float)LONGLONG_MIN)) {
      Result ret = MipsyFPUException(P, FPE_V);
      return ret;
   }
   return C1_SUCCESS;
}   

COP1OP(cvt_w_dOp)
{
   double contents_fs = DoubleReg(fs);
   uint roundMode = P->FCR[31] & 0x03;
   double rounded;
   switch (roundMode) {
   case 0:                 /* RN: Round to nearest         */
      rounded = (contents_fs > 0.0 ? contents_fs + 0.5 : contents_fs - 0.5);
      break;
   case 1:                 /* RZ: Round toward 0           */
      rounded = contents_fs;
      break;
   case 2:                 /* RP: Round to +infinity       */
      rounded = (contents_fs >= 0.0 ? contents_fs + SP_NEAR_ONE : contents_fs);
      break;
   case 3:                 /* RM: Round to -infinity       */
      rounded = (contents_fs >= 0.0 ? contents_fs : contents_fs - SP_NEAR_ONE);
      break;
   }

   IntReg(fd) = rounded;
   SIM_DEBUG(('f', "C1:CVT.W.D: f%d %e --> f%d %d rm = %d\n", 
              fs, contents_fs, fd, IntReg(fd), roundMode));

   /* Check for valid operand */
   if ((rounded > (float)INT_MAX) || (rounded < (float)INT_MIN)) {
      Result ret = MipsyFPUException(P, FPE_V);
      return ret;
   }
   return C1_SUCCESS;
}

COP1OP(div_dOp)
{
   double contents_fs = DoubleReg(fs);
   double contents_ft = DoubleReg(ft);
   if (contents_ft == 0.0) {
      CPUWarning("C1:DIV.D: divide by zero\n");
   }
   DoubleReg(fd) = contents_fs / contents_ft;
   SIM_DEBUG(('f', "C1:DIV.D: f%d %e / f%d %e = f%d %e\n", 
              fs, contents_fs, ft, contents_ft, fd, DoubleReg(fd)));
   return C1_SUCCESS;
}

COP1OP(mov_dOp)
{
   double contents_fs = DoubleReg(fs);
   DoubleReg(fd) = contents_fs;
   SIM_DEBUG(('f', "C1:MOV.D: f%d %e --> f%d %e\n", 
              fs, contents_fs, fd, DoubleReg(fd)));
   return C1_SUCCESS;
}
COP1OP(movcf_dOp)
{
   double contents_fs = DoubleReg(fs);
   int fcc = MOVC_FCC(instr);
   
   /* Move conditional on FP condition */
   if (TF(instr)) { 
      if (P->FCR[31] & M_csr_fcc(fcc)) {
         DoubleReg(fd) = contents_fs;
         SIM_DEBUG(('f', "C1:MOVT.D: fcc %d T, f%d %e --> f%d %e\n", 
                    fcc, fs, contents_fs, fd, DoubleReg(fd)));
      } else {
         SIM_DEBUG(('f', "C1:MOVT.D: fcc %d F, f%d %e unchanged\n",
                    fcc, fd, DoubleReg(fd)));
      }
   } else {
      if (!(P->FCR[31] & M_csr_fcc(fcc))) {
         DoubleReg(fd) = contents_fs;
         SIM_DEBUG(('f', "C1:MOVF.D: fcc %d F, f%d %e --> f%d %e\n", 
                    fcc, fs, contents_fs, fd, DoubleReg(fd)));
      } else {
         SIM_DEBUG(('f', "C1:MOVF.D: fcc %d T, f%d %e unchanged\n",
                    fcc, fd, DoubleReg(fd)));
      }
   }
   return C1_SUCCESS;
}

COP1OP(movn_dOp)
{
   double contents_fs = DoubleReg(fs);
   Reg contents_rt = P->R[ft];
   if (contents_rt != 0) {
      DoubleReg(fd) = contents_fs;
      SIM_DEBUG(('f', "C1:MOVN.D: r%d %lld, f%d %e to f%d %e\n", 
                 ft, (uint64) contents_rt, 
                 fs, contents_fs, fd, DoubleReg(fd)));
   } else {
      SIM_DEBUG(('f', "C1:MOVN.D: r%d %lld, f%d %e unchanged\n", 
                 ft, (int64) contents_rt, fd, DoubleReg(fd)));
   }
   return C1_SUCCESS;
}

COP1OP(movz_dOp)
{
   double contents_fs = DoubleReg(fs);
   Reg contents_rt = P->R[ft];
   if (contents_rt == 0) {
      DoubleReg(fd) = contents_fs;
      SIM_DEBUG(('f', "C1:MOVZ.D: r%d %lld, f%d %e to f%d %e\n", 
                 ft, (int64) contents_rt, 
                 fs, contents_fs, fd, DoubleReg(fd)));
   } else {
      SIM_DEBUG(('f', "C1:MOVZ.D: r%d %lld, f%d %e unchanged\n", 
                 ft, (uint64) contents_rt, fd, DoubleReg(fd)));
   }
   return C1_SUCCESS;
}

COP1OP(mul_dOp)
{
   double contents_fs = DoubleReg(fs);
   double contents_ft = DoubleReg(ft);
   DoubleReg(fd) = contents_fs * contents_ft;
   SIM_DEBUG(('f', "C1:MUL.D: f%d %e * f%d %e = f%d %e\n", 
              fs, contents_fs, ft, contents_ft, fd, DoubleReg(fd)));
   return C1_SUCCESS;
}

COP1OP(neg_dOp)
{
   double contents_fs = DoubleReg(fs);
   DoubleReg(fd) = -contents_fs;
   SIM_DEBUG(('f', "C1:NEG.D: f%d %e --> f%d %e\n", 
              fs, contents_fs, fd, DoubleReg(fd)));
   return C1_SUCCESS;
}

COP1OP(sub_dOp)
{
   double contents_fs = DoubleReg(fs);
   double contents_ft = DoubleReg(ft);
   DoubleReg(fd) = contents_fs - contents_ft;
   SIM_DEBUG(('f', "C1:SUB.D: f%d %e - f%d %e = f%d %e\n", 
              fs, contents_fs, ft, contents_ft, fd, DoubleReg(fd)));
   return C1_SUCCESS;
}

COP1OP(sqrt_dOp)
{
   double contents_fs = DoubleReg(fs);
   DoubleReg(fd) = sqrt(contents_fs);
   SIM_DEBUG(('f', "C1:SQRT.D: sqrt(f%d %e) = f%d %e\n", 
              fs, contents_fs, fd, DoubleReg(fd)));
   return C1_SUCCESS;
}

COP1OP(trunc_w_dOp)
{
   double contents_fs = DoubleReg(fs);
   IntReg(fd) = contents_fs;
   SIM_DEBUG(('f', "C1:TRUNC.W.D: f%d %e --> f%d %d\n", 
              fs, contents_fs, fd, IntReg(fd)));
   return C1_SUCCESS;
}

COP1OP(trunc_l_dOp)
{
   double contents_fs = DoubleReg(fs);
   LongLongReg(fd) = contents_fs;
   SIM_DEBUG(('f', "C1:TRUNC.L.D: f%d %e --> f%d %lld\n", 
              fs, contents_fs, fd, LongLongReg(fd)));
   return C1_SUCCESS;
}

COP1OP(floor_w_dOp)
{
   double contents_fs = DoubleReg(fs);
   IntReg(fd) = (contents_fs >= 0.0 ? 
                       contents_fs : 
                       contents_fs - SP_NEAR_ONE);
   SIM_DEBUG(('f', "C1:FLOOR.W.D: f%d %e --> f%d %d\n", 
              fs, contents_fs, fd, IntReg(fd)));
   return C1_SUCCESS;
}

COP1OP(floor_l_dOp)
{
   double contents_fs = DoubleReg(fs);
   LongLongReg(fd) = (contents_fs >= 0.0 ? 
                            contents_fs : 
                            contents_fs - SP_NEAR_ONE);
   SIM_DEBUG(('f', "C1:FLOOR.L.D: f%d %e --> f%d %lld\n", 
              fs, contents_fs, fd, LongLongReg(fd)));
   return C1_SUCCESS;
}

COP1OP(recip_dOp)
{
   double contents_fs = DoubleReg(fs);
   if (contents_fs == 0.0) {
      CPUWarning("C1:RECIP.D: divide by zero\n");
   }
   DoubleReg(fd) = 1.0 / contents_fs;
   SIM_DEBUG(('f', "C1:RECIP.D: 1.0 / f%d %e = f%d %e\n", 
              fs, contents_fs, fd, DoubleReg(fd)));
   return C1_SUCCESS;
}

COP1OP(rsqrt_dOp)
{
   double contents_fs = DoubleReg(fs);
   if (contents_fs == 0.0) {
      CPUWarning("C1:RSQRT.D: divide by zero\n");
   }
   DoubleReg(fd) = 1.0 / sqrt(contents_fs);
   SIM_DEBUG(('f', "C1:RSQRT.D: 1.0 / sqrt(f%d %e) = f%d %e\n", 
              fs, contents_fs, fd, DoubleReg(fd)));
   return C1_SUCCESS;
}

COP1OP(round_w_dOp)
{
   double contents_fs = DoubleReg(fs);
   IntReg(fd) = (contents_fs > 0.0 ? 
                       contents_fs + 0.5 : 
                       contents_fs - 0.5);
   SIM_DEBUG(('f', "C1:ROUND.W.D: f%d %e --> f%d %d\n", 
              fs, contents_fs, fd, IntReg(fd)));
   return C1_SUCCESS;
}

COP1OP(round_l_dOp)
{
   double contents_fs = DoubleReg(fs);
   LongLongReg(fd) = (contents_fs > 0.0 ? 
                            contents_fs + 0.5 : 
                            contents_fs - 0.5);
   SIM_DEBUG(('f', "C1:ROUND.W.D: f%d %e --> f%d %lld\n", 
              fs, contents_fs, fd, LongLongReg(fd)));
   return C1_SUCCESS;
}

COP1OP(ceil_w_dOp)
{
   double contents_fs = DoubleReg(fs);
   IntReg(fd) = (contents_fs > 0.0 ? 
                       contents_fs + SP_NEAR_ONE : 
                       contents_fs);
   SIM_DEBUG(('f', "C1:CEIL.W.D: f%d %e --> f%d %d\n", 
              fs, contents_fs, fd, IntReg(fd)));
   return C1_SUCCESS;
}

COP1OP(ceil_l_dOp)
{
   double contents_fs = DoubleReg(fs);
   LongLongReg(fd) = (contents_fs > 0.0 ? 
                            contents_fs + SP_NEAR_ONE : 
                            contents_fs);
   SIM_DEBUG(('f', "C1:CEIL.L.D: f%d %e --> f%d %lld\n", 
              fs, contents_fs, fd, LongLongReg(fd)));
   return C1_SUCCESS;
}

COP1OP(c_eq_dOp)
{
   double contents_fs = DoubleReg(fs);
   double contents_ft = DoubleReg(ft);
   int cond = (contents_fs == contents_ft);
   int fcc = C_FCC(instr);
   SIM_DEBUG(('f', "C1:C.EQ.D: f%d %e, f%d %e --> fcc%d %d\n",
              fs, contents_fs, ft, contents_ft, fcc, cond));
   return CompareOp(P, cond, fcc);
}

COP1OP(c_f_dOp)
{
   double contents_fs = DoubleReg(fs);
   double contents_ft = DoubleReg(ft);
   int cond = 0;
   int fcc = C_FCC(instr);
   SIM_DEBUG(('f', "C1:C.F.D: f%d %e, f%d %e --> fcc%d %d\n",
              fs, contents_fs, ft, contents_ft, fcc, cond));
   return CompareOp(P, cond, fcc);
}

COP1OP(c_le_dOp)
{
   double contents_fs = DoubleReg(fs);
   double contents_ft = DoubleReg(ft);
   int cond = (contents_fs <= contents_ft);
   int fcc = C_FCC(instr);
   SIM_DEBUG(('f', "C1:C.LE.D: f%d %e, f%d %e --> fcc%d %d\n",
              fs, contents_fs, ft, contents_ft, fcc, cond));
   return CompareOp(P, cond, fcc);
}

COP1OP(c_lt_dOp)
{
   double contents_fs = DoubleReg(fs);
   double contents_ft = DoubleReg(ft);
   int cond = (contents_fs < contents_ft);
   int fcc = C_FCC(instr);
   SIM_DEBUG(('f', "C1:C.LT.D: f%d %e, f%d %e --> fcc%d %d\n",
              fs, contents_fs, ft, contents_ft, fcc, cond));
   return CompareOp(P, cond, fcc);
}

COP1OP(c_nge_dOp)
{
   double contents_fs = DoubleReg(fs);
   double contents_ft = DoubleReg(ft);
   int cond = (contents_fs <= contents_ft);
   int fcc = C_FCC(instr);
   SIM_DEBUG(('f', "C1:C.NGE.D: f%d %e, f%d %e --> fcc%d %d\n",
              fs, contents_fs, ft, contents_ft, fcc, cond));
   return CompareOp(P, cond, fcc);
}

COP1OP(c_ngl_dOp)
{
   double contents_fs = DoubleReg(fs);
   double contents_ft = DoubleReg(ft);
   int cond = (contents_fs == contents_ft);
   int fcc = C_FCC(instr);
   SIM_DEBUG(('f', "C1:C.NGL.D: f%d %e, f%d %e --> fcc%d %d\n",
              fs, contents_fs, ft, contents_ft, fcc, cond));
   return CompareOp(P, cond, fcc);
}

COP1OP(c_ngle_dOp)
{
   double contents_fs = DoubleReg(fs);
   double contents_ft = DoubleReg(ft);
   int cond = 0;
   int fcc = C_FCC(instr);
   SIM_DEBUG(('f', "C1:C.NGLE.D: f%d %e, f%d %e --> fcc%d %d\n",
              fs, contents_fs, ft, contents_ft, fcc, cond));
   return CompareOp(P, cond, fcc);
}

COP1OP(c_ngt_dOp)
{
   double contents_fs = DoubleReg(fs);
   double contents_ft = DoubleReg(ft);
   int cond = (contents_fs <= contents_ft);
   int fcc = C_FCC(instr);
   SIM_DEBUG(('f', "C1:C.NGT.D: f%d %e, f%d %e --> fcc%d %d\n",
              fs, contents_fs, ft, contents_ft, fcc, cond));
   return CompareOp(P, cond, fcc);
}

COP1OP(c_ole_dOp)
{
   double contents_fs = DoubleReg(fs);
   double contents_ft = DoubleReg(ft);
   int cond = (contents_fs < contents_ft);
   int fcc = C_FCC(instr);
   SIM_DEBUG(('f', "C1:C.OLE.D: f%d %e, f%d %e --> fcc%d %d\n",
              fs, contents_fs, ft, contents_ft, fcc, cond));
   return CompareOp(P, cond, fcc);
}

COP1OP(c_olt_dOp)
{
   double contents_fs = DoubleReg(fs);
   double contents_ft = DoubleReg(ft);
   int cond = (contents_fs < contents_ft);
   int fcc = C_FCC(instr);
   SIM_DEBUG(('f', "C1:C.OLT.D: f%d %e, f%d %e --> fcc%d %d\n",
              fs, contents_fs, ft, contents_ft, fcc, cond));
   return CompareOp(P, cond, fcc);
}

COP1OP(c_seq_dOp)
{
   double contents_fs = DoubleReg(fs);
   double contents_ft = DoubleReg(ft);
   int cond = (contents_fs == contents_ft);
   int fcc = C_FCC(instr);
   SIM_DEBUG(('f', "C1:C.SEQ.D: f%d %e, f%d %e --> fcc%d %d\n",
              fs, contents_fs, ft, contents_ft, fcc, cond));
   return CompareOp(P, cond, fcc);
}

COP1OP(c_sf_dOp)
{
   double contents_fs = DoubleReg(fs);
   double contents_ft = DoubleReg(ft);
   int cond = 0;
   int fcc = C_FCC(instr);
   SIM_DEBUG(('f', "C1:C.SF.D: f%d %e, f%d %e --> fcc%d %d\n",
              fs, contents_fs, ft, contents_ft, fcc, cond));
   return CompareOp(P, cond, fcc);
}

COP1OP(c_ueq_dOp)
{
   double contents_fs = DoubleReg(fs);
   double contents_ft = DoubleReg(ft);
   int cond = (contents_fs == contents_ft);
   int fcc = C_FCC(instr);
   SIM_DEBUG(('f', "C1:C.UEQ.D: f%d %e, f%d %e --> fcc%d %d\n",
              fs, contents_fs, ft, contents_ft, fcc, cond));
   return CompareOp(P, cond, fcc);
}

COP1OP(c_ule_dOp)
{
   double contents_fs = DoubleReg(fs);
   double contents_ft = DoubleReg(ft);
   int cond = (contents_fs <= contents_ft);
   int fcc = C_FCC(instr);
   SIM_DEBUG(('f', "C1:C.ULE.D: f%d %e, f%d %e --> fcc%d %d\n",
              fs, contents_fs, ft, contents_ft, fcc, cond));
   return CompareOp(P, cond, fcc);
}

COP1OP(c_ult_dOp)
{
   double contents_fs = DoubleReg(fs);
   double contents_ft = DoubleReg(ft);
   int cond = (contents_fs < contents_ft);
   int fcc = C_FCC(instr);
   SIM_DEBUG(('f', "C1:C.ULT.D: f%d %e, f%d %e --> fcc%d %d\n",
              fs, contents_fs, ft, contents_ft, fcc, cond));
   return CompareOp(P, cond, fcc);
}

COP1OP(c_un_dOp)
{
   double contents_fs = DoubleReg(fs);
   double contents_ft = DoubleReg(ft);
   int cond = 0;
   int fcc = C_FCC(instr);
   SIM_DEBUG(('f', "C1:C.UN.D: f%d %e, f%d %e --> fcc%d %d\n",
              fs, contents_fs, ft, contents_ft, fcc, cond));
   return CompareOp(P, cond, fcc);
}

/*****************************************************************
 * Fixed-point opcodes
 *****************************************************************/
COP1OP(cvt_s_wOp)
{
   int contents_fs = IntReg(fs);
   FloatReg(fd) = contents_fs;
   SIM_DEBUG(('f', "C1:CVT.S.W: f%d %d --> f%d %e\n",
              fs, contents_fs, ft, FloatReg(fd)));
   return C1_SUCCESS;
}

COP1OP(cvt_d_wOp)
{
   int contents_fs = IntReg(fs);
   DoubleReg(fd) = contents_fs;
   SIM_DEBUG(('f', "C1:CVT.D.W: f%d %d --> f%d %e\n",
              fs, contents_fs, ft, DoubleReg(fd)));
   return C1_SUCCESS;
}

/*****************************************************************
 * Long fixed-point opcodes
 *****************************************************************/
COP1OP(cvt_s_lOp)
{
   int64 contents_fs = LongLongReg(fs);
   FloatReg(fd) = contents_fs;
   SIM_DEBUG(('f', "C1:CVT.S.L: f%d %lld --> f%d %e\n",
              fs, contents_fs, ft, FloatReg(fd)));
   return C1_SUCCESS;
}

COP1OP(cvt_d_lOp)
{
   int contents_fs = LongLongReg(fs);
   DoubleReg(fd) = contents_fs;
   SIM_DEBUG(('f', "C1:CVT.D.L: f%d %lld --> f%d %e\n",
              fs, contents_fs, ft, DoubleReg(fd)));
   return C1_SUCCESS;
}

static Result
CompareOp(CPUState *P, int cond, int fcc)
{
   if (cond) {
      P->FCR[31] |= M_csr_fcc(fcc);
   } else {
      P->FCR[31] &= ~M_csr_fcc(fcc);
   }
   return C1_SUCCESS;
}

/*****************************************************************
 * COP1X opcodes -- single-precision floating-point
 *****************************************************************/

COP1XOP(madd_sOp)
{
   float contents_fr = FloatReg(fr);
   float contents_fs = FloatReg(fs);
   float contents_ft = FloatReg(ft);

   FloatReg(fd) = (contents_fs * contents_ft) + contents_fr;
   SIM_DEBUG(('f', "C1:MADD.S: (f%d %e * f%d %e) + f%d %e = f%d %e\n", 
              fs, contents_fs, ft, contents_ft, fr, contents_fr, 
              fd, FloatReg(fd)));
   return C1_SUCCESS;
}

COP1XOP(msub_sOp)
{
   float contents_fr = FloatReg(fr);
   float contents_fs = FloatReg(fs);
   float contents_ft = FloatReg(ft);

   FloatReg(fd) = (contents_fs * contents_ft) - contents_fr;
   SIM_DEBUG(('f', "C1:MSUB.S: (f%d %e * f%d %e) - f%d %e = f%d %e\n", 
              fs, contents_fs, ft, contents_ft, fr, contents_fr, 
              fd, FloatReg(fd)));
   return C1_SUCCESS;
}

COP1XOP(nmadd_sOp)
{
   float contents_fr = FloatReg(fr);
   float contents_fs = FloatReg(fs);
   float contents_ft = FloatReg(ft);

   FloatReg(fd) = -((contents_fs * contents_ft) + contents_fr);
   SIM_DEBUG(('f', "C1:NMADD.S: -((f%d %e * f%d %e) + f%d %e) = f%d %e\n", 
              fs, contents_fs, ft, contents_ft, fr, contents_fr, 
              fd, FloatReg(fd)));
   return C1_SUCCESS;
}

COP1XOP(nmsub_sOp)
{
   float contents_fr = FloatReg(fr);
   float contents_fs = FloatReg(fs);
   float contents_ft = FloatReg(ft);

   FloatReg(fd) = -((contents_fs * contents_ft) - contents_fr);
   SIM_DEBUG(('f', "C1:NMSUB.S: -((f%d %e * f%d %e) - f%d %e) = f%d %e\n", 
              fs, contents_fs, ft, contents_ft, fr, contents_fr, 
              fd, FloatReg(fd)));
   return C1_SUCCESS;
}

/*****************************************************************
 * COP1X opcodes -- double-precision floating-point
 *****************************************************************/

COP1XOP(madd_dOp)
{
   double contents_fr = DoubleReg(fr);
   double contents_fs = DoubleReg(fs);
   double contents_ft = DoubleReg(ft);

   DoubleReg(fd) = (contents_fs * contents_ft) + contents_fr;
   SIM_DEBUG(('f', "C1:MADD.D: (f%d %e * f%d %e) + f%d %e = f%d %e\n", 
              fs, contents_fs, ft, contents_ft, fr, contents_fr, 
              fd, DoubleReg(fd)));
   return C1_SUCCESS;
}

COP1XOP(msub_dOp)
{
   double contents_fr = DoubleReg(fr);
   double contents_fs = DoubleReg(fs);
   double contents_ft = DoubleReg(ft);

   DoubleReg(fd) = (contents_fs * contents_ft) - contents_fr;
   SIM_DEBUG(('f', "C1:MSUB.D: (f%d %e * f%d %e) - f%d %e = f%d %e\n", 
              fs, contents_fs, ft, contents_ft, fr, contents_fr, 
              fd, DoubleReg(fd)));
   return C1_SUCCESS;
}

COP1XOP(nmadd_dOp)
{
   double contents_fr = DoubleReg(fr);
   double contents_fs = DoubleReg(fs);
   double contents_ft = DoubleReg(ft);

   DoubleReg(fd) = -((contents_fs * contents_ft) + contents_fr);
   SIM_DEBUG(('f', "C1:NMADD.D: -((f%d %e * f%d %e) + f%d %e) = f%d %e\n", 
              fs, contents_fs, ft, contents_ft, fr, contents_fr, 
              fd, DoubleReg(fd)));
   return C1_SUCCESS;
}

COP1XOP(nmsub_dOp)
{
   double contents_fr = DoubleReg(fr);
   double contents_fs = DoubleReg(fs);
   double contents_ft = DoubleReg(ft);

   DoubleReg(fd) = -((contents_fs * contents_ft) - contents_fr);
   SIM_DEBUG(('f', "C1:NMSUB.D: -((f%d %e * f%d %e) - f%d %e) = f%d %e\n", 
              fs, contents_fs, ft, contents_ft, fr, contents_fr, 
              fd, DoubleReg(fd)));
   return C1_SUCCESS;
}

/*****************************************************************
 * COP1X opcodes -- indexed memory ops -- currently unimplemented
 *****************************************************************/

/*****************************************************************
 * Catch-all routines for unimplemented FPU opcodes
 *****************************************************************/
COP1OP(unknownC1Op)
{
   return C1_FAILURE;
}

COP1XOP(unknownC1XOp)
{
   return C1_FAILURE;
}

/*****************************************************************
 * MipsyFPU
 *****************************************************************/
static Result 
MipsyFPU(CPUState *P, Inst instr) 
{
   uint fmt = FORMAT(instr);
   uint func = FUNC(instr);
   uint fs = FS(instr);
   uint ft = FT(instr);
   uint fd = FD(instr);
   Result ret;

   switch (fmt) {
   case FMT_S:
      ret = mipsC1SPOpcodes[func].func(P, instr, fd, fs, ft);
      break;
   case FMT_D:
      ret = mipsC1DPOpcodes[func].func(P, instr, fd, fs, ft);
      break;
   case FMT_W:
      /* Only W-format opcodes are cvt.s.w and cvt.d.w */
      if (func == fcvts_op) {
         ret = cvt_s_wOp(P, instr, fd, fs, ft);
      } else if (func == fcvtd_op) {
         ret = cvt_d_wOp(P, instr, fd, fs, ft);
      } else {
         ret = C1_FAILURE;
      }
      break;
   case FMT_L:
      /* Only L-format opcodes are cvt.s.l and cvt.d.l */
      if (func == fcvts_op) {
         ret = cvt_s_lOp(P, instr, fd, fs, ft);
      } else if (func == fcvtd_op) {
         ret = cvt_d_lOp(P, instr, fd, fs, ft);
      } else {
         ret = C1_FAILURE;
      }
      break;
   default:
      ret = C1_FAILURE;
      break;
   }
   if (ret == C1_FAILURE) {
      EXCEPTION(P, EXC_II);
      return C1_FAILURE;
   }
   return ret;
}

/*****************************************************************
 * MipsyFPUx
 *****************************************************************/
static Result 
MipsyFPUx(CPUState *P, Inst instr) 
{
   uint fs  = FS(instr); /* 1st source reg */
   uint ft  = FT(instr); /* 2nd source register */
   uint fr  = FORMAT(instr); /* 3rd source reg. */
   uint fd  = FD(instr); /* destionation */
   uint func = FUNC(instr);

   Result ret = mipsC1XOpcodes[func].func(P, instr, fd, fr, fs, ft);

   if (ret == C1_FAILURE) {
      EXCEPTION(P, EXC_II);
      return C1_FAILURE;
   }
   return ret;
}

/*****************************************************************
 * Set the rounding mode of the host processor to match that of the
 * simulated machine. This is needed since we're actually executing the
 * instructions on the host platform. 
 *****************************************************************/
static void 
SetRoundingMode(int modeNum)
{
#if defined(sgi) || defined(sparc)
   fp_rnd newMode = FP_RN; 
   
   switch (modeNum) {
   case 0:
      newMode = FP_RN;
      break;
   case 1:
      newMode = FP_RZ;
      break;
   case 2:
      newMode = FP_RP;
      break;
   case 3:
      newMode = FP_RM;
      break;
   default:
      CPUWarning("Bad rounding mode set in FCR31 (%d)\n", modeNum);
      break;
   }
   fpsetround(newMode);
#endif
#ifdef __alpha
   write_rnd(modeNum);
#endif
#ifdef linux
   {
      static int fpwarned = 0;
      if (!fpwarned) {
         CPUWarning("MIPSY: could not figure out how to fpsetround on x86 \n");
         fpwarned = 1;
      }
   }
#endif
}

/*****************************************************************
 * All fpu (COP1) insts come through here.
 *****************************************************************/
int 
ExecuteC1Instruction(CPUState *P, Inst instr)
{
   int immediate   = IMMED(instr);
   uint fs  = FS(instr); /* 1st source reg */
   uint ft  = FT(instr); /* 2nd source register */
   uint sub_value = FORMAT(instr);
   uint br_value  = ft;
   uint major_op;
   Result result = C1_SUCCESS;
   uint func = FUNC(instr);
   StatusReg statusReg;
   
   /* The kernel sets the cop1 usable flag to 0 when it */
   /* first starts a new context to limit the saving of */
   /* FPU registers to those procs who use them. The */
   /* first time a process uses a FPU instruction it */
   /* needs to trap into the kernel for it to set this */
   /* bit. */
   
   statusReg.ts_data = P->CP0[C0_SR];

   if (!(statusReg.s32.ts_cu1)) {
      CauseReg  causeReg;
      
      causeReg.tc_data = P->CP0[C0_CAUSE];
      causeReg.s32.tc_ce = 1;
      P->CP0[C0_CAUSE] = causeReg.tc_data;
      
      EXCEPTION(P, EXC_CPU); 
      return C1_FAILURE;
   }

   SetRoundingMode(P->FCR[31] & 0x3);

   major_op = MAJOR_OP(instr);
   if (major_op == cop1_op) { 

      switch(sub_value) {

      case bc_op:
         if (br_value & 0x1) { /* BC1T */
            /* Check if the condition bit in the csr is set */
            if (P->FCR[31] & M_csr_fcc(BC_FCC(instr))) {
               SIM_DEBUG(('f', "C1:BC1T: fcc %d true, branch taken\n",
                          BC_FCC(instr)));
               P->branchTarget = P->nPC + SIGN_EXTEND(18, (immediate << 2)); 
               P->branchStatus = BranchStatus_taken;
            }  else { 
               SIM_DEBUG(('f', "C1:BC1T: fcc %d false, branch not taken\n",
                          BC_FCC(instr)));
                if( br_value & 0x2 ) { /* branch likely */
                    /* Skip instruction in delay slot */
                    P->nPC = P->nPC + 4;
                } else {
                    P->branchStatus = BranchStatus_nottaken;
                } 
            }
         } else { /* BC1F */
            /* Check if condition bit not set */
            if (!(P->FCR[31] & M_csr_fcc(BC_FCC(instr)))) {
               SIM_DEBUG(('f', "C1:BC1F: fcc %d false, branch taken\n",
                          BC_FCC(instr)));
               P->branchTarget = P->nPC + SIGN_EXTEND(18, (immediate << 2));
               P->branchStatus = BranchStatus_taken;
            }  else { 
               SIM_DEBUG(('f', "C1:BC1F: fcc %d true, branch not taken\n",
                          BC_FCC(instr)));
                if( br_value & 0x2 ) { /* branch likely */
                    /* Skip instruction in delay slot */
                    P->nPC = P->nPC + 4;
                } else {
                    P->branchStatus = BranchStatus_nottaken;
                }
            }
         }
         break;
      
      case cfc_op:
         if (fs == 31) {
            P->R[ft] = P->FCR[fs];
         } else {
            P->R[ft] = (3 << 8); /* R3010 FPU */
         }
         SIM_DEBUG(('f', "C1: CFC1 got %#x from fcr %d\n", P->R[ft], fs));
         break;
      
      case ctc_op:
         P->FCR[fs] = P->R[ft];
         SIM_DEBUG(('f', "C1: CTC1 put %#x into fcr %d\n", P->R[ft], fs));
         break;
      
      case mfc_op: 
         P->R[ft] = (Reg32_s)IntReg(fs);
         SIM_DEBUG(('f', "C1: mfc1 put f%d %d into r%d \n", fs, IntReg(fs), ft));
         break;
      
      case mtc_op:
         IntReg(fs) = (Reg32)P->R[ft];
         SIM_DEBUG(('f', "C1: mtc1 put r%d %d into f%d -> %d\n", ft, P->R[ft], 
               fs, IntReg(fs)));
         break;
  
      case dmfc_op: 
         CHECK_64BIT_ALLOWED(P)
         P->R[ft] = LongLongReg(fs);
         SIM_DEBUG(('f', "C1: dmfc1 put f%d %lld into r%d \n", 
                    fs, LongLongReg(fs), ft));
         break;
      
      case dmtc_op:
         CHECK_64BIT_ALLOWED(P)
         LongLongReg(fs) = P->R[ft];
         SIM_DEBUG(('f', "C1: dmtc1 put r%d %lld into f%d -> %lld\n", 
                    ft, P->R[ft], fs, LongLongReg(fs)));
         break;
      default: 
         /* S, D, W format instructions */
         switch (func) {
         case frecip_op:
         case frsqrt_op:
         case fmovc_op:
         case fmovn_op:
         case fmovz_op: {
            /* Make sure we can execute mips4 instructions */
            if (IS_USER_MODE(P) && !(statusReg.s32.ts_xx)) {
               EXCEPTION(P, EXC_II); 
               return C1_FAILURE;
            }
            STATS_INC(P->myNum, numFPOps, 1);
            CPUVec.IncC0PerfCtr(P->myNum, 21); /* FP instructions graduated */
            break;
         }
         default:
            STATS_INC(P->myNum, numFPOps, 1);
            CPUVec.IncC0PerfCtr(P->myNum, 21); /* FP instructions graduated */
            break;
         } 
         /* Clear the exception flags from the FCR */
         P->FCR[31] &= ~M_csr_causes;
         result = MipsyFPU(P, instr);
         break;
      } /* switch subvalue */
   } else if (major_op == cop1x_op) {
      if (IS_USER_MODE(P) && !(statusReg.s32.ts_xx)) {
         EXCEPTION(P, EXC_II); 
         return C1_FAILURE;
      }
      STATS_INC(P->myNum, numFPOps, 1);
      CPUVec.IncC0PerfCtr(P->myNum, 21); /* FP instructions graduated */
      /* Clear the exception flags from the FCR */
      P->FCR[31] &= ~M_csr_causes;
      result = MipsyFPUx(P,instr);
      
   } else if (major_op == spec_op) { 
      /* Must be on the MOVC instructions */
      ASSERT(FUNC(instr) == movc_op);
      /* Move conditional on FP condition */
      if (TF(instr)) { 
         if (P->FCR[31] & M_csr_fcc(MOVC_FCC(instr))) {
            P->R[RD(instr)] = P->R[RS(instr)];
         }
      } else {
         if (!(P->FCR[31] & M_csr_fcc(MOVC_FCC(instr)))) {
            P->R[RD(instr)] = P->R[RS(instr)];
         }
      }
   } else {
      ASSERT(0); /* unknown C1 instruction */
   }
      
   return result;
}

MipsC1OpcodeDesc mipsC1SPOpcodes[] = {
   { "add.s", add_sOp, 0x00 },
   { "sub.s", sub_sOp, 0x01 },
   { "mul.s", mul_sOp, 0x02 },
   { "div.s", div_sOp, 0x03 },
   { "sqrt.s", sqrt_sOp, 0x04 },
   { "abs.s", abs_sOp, 0x05 },
   { "mov.s", mov_sOp, 0x06 },
   { "neg.s", neg_sOp, 0x07 },
   { "round.l.s", round_l_sOp, 0x08 },
   { "trunc.l.s", trunc_l_sOp, 0x09 },
   { "ceil.l.s", ceil_l_sOp, 0x0a },
   { "floor.l.s", floor_l_sOp, 0x0b },
   { "round.w.s", round_w_sOp, 0x0c },
   { "trunc.w.s", trunc_w_sOp, 0x0d },
   { "ceil.w.s", ceil_w_sOp, 0x0e },
   { "floor.w.s", floor_w_sOp, 0x0f },
   { "unknownC1", unknownC1Op, 0x10 },
   { "movcf.s", movcf_sOp, 0x11 },
   { "movz.s", movz_sOp, 0x12 },
   { "movn.s", movn_sOp, 0x13 },
   { "unknownC1", unknownC1Op, 0x14 },
   { "recip.s", recip_sOp, 0x15 },
   { "rsqrt.s", rsqrt_sOp, 0x16 },
   { "unknownC1", unknownC1Op, 0x17 },
   { "unknownC1", unknownC1Op, 0x18 },
   { "unknownC1", unknownC1Op, 0x19 },
   { "unknownC1", unknownC1Op, 0x1a },
   { "unknownC1", unknownC1Op, 0x1b },
   { "unknownC1", unknownC1Op, 0x1c },
   { "unknownC1", unknownC1Op, 0x1d },
   { "unknownC1", unknownC1Op, 0x1e },
   { "unknownC1", unknownC1Op, 0x1f },
   { "unknownC1", unknownC1Op, 0x20 },
   { "cvt.d.s", cvt_d_sOp, 0x21 },
   { "unknownC1", unknownC1Op, 0x22 },
   { "unknownC1", unknownC1Op, 0x23 },
   { "cvt.w.s", cvt_w_sOp, 0x24 },
   { "cvt.l.s", cvt_l_sOp, 0x25 },
   { "unknownC1", unknownC1Op, 0x26 },
   { "unknownC1", unknownC1Op, 0x27 },
   { "unknownC1", unknownC1Op, 0x28 },
   { "unknownC1", unknownC1Op, 0x29 },
   { "unknownC1", unknownC1Op, 0x2a },
   { "unknownC1", unknownC1Op, 0x2b },
   { "unknownC1", unknownC1Op, 0x2c },
   { "unknownC1", unknownC1Op, 0x2d },
   { "unknownC1", unknownC1Op, 0x2e },
   { "unknownC1", unknownC1Op, 0x2f },
   { "c.f.s", c_f_sOp, 0x30 },
   { "c.un.s", c_un_sOp, 0x31 },
   { "c.eq.s", c_eq_sOp, 0x32 },
   { "c.ueq.s", c_ueq_sOp, 0x33 },
   { "c.olt.s", c_olt_sOp, 0x34 },
   { "c.ult.s", c_ult_sOp, 0x35 },
   { "c.ole.s", c_ole_sOp, 0x36 },
   { "c.ule.s", c_ule_sOp, 0x37 },
   { "c.sf.s", c_sf_sOp, 0x38 },
   { "c.ngle.s", c_ngle_sOp, 0x39 },
   { "c.seq.s", c_seq_sOp, 0x3a },
   { "c.ngl.s", c_ngl_sOp, 0x3b },
   { "c.lt.s", c_lt_sOp, 0x3c },
   { "c.nge.s", c_nge_sOp, 0x3d },
   { "c.le.s", c_le_sOp, 0x3e },
   { "c.ngt.s", c_ngt_sOp, 0x3f }
};

MipsC1OpcodeDesc mipsC1DPOpcodes[] = {
   { "add.d", add_dOp, 0x00 },
   { "sub.d", sub_dOp, 0x01 },
   { "mul.d", mul_dOp, 0x02 },
   { "div.d", div_dOp, 0x03 },
   { "sqrt.d", sqrt_dOp, 0x04 },
   { "abs.d", abs_dOp, 0x05 },
   { "mov.d", mov_dOp, 0x06 },
   { "neg.d", neg_dOp, 0x07 },
   { "round.l.d", round_l_dOp, 0x08 },
   { "trunc.l.d", trunc_l_dOp, 0x09 },
   { "ceil.l.d", ceil_l_dOp, 0x0a },
   { "floor.l.d", floor_l_dOp, 0x0b },
   { "round.w.d", round_w_dOp, 0x0c },
   { "trunc.w.d", trunc_w_dOp, 0x0d },
   { "ceil.w.d", ceil_w_dOp, 0x0e },
   { "floor.w.d", floor_w_dOp, 0x0f },
   { "unknownC1", unknownC1Op, 0x10 },
   { "movcf.d", movcf_dOp, 0x11 },
   { "movz.d", movz_dOp, 0x12 },
   { "movn.d", movn_dOp, 0x13 },
   { "unknownC1", unknownC1Op, 0x14 },
   { "recip.d", recip_dOp, 0x15 },
   { "rsqrt.d", rsqrt_dOp, 0x16 },
   { "unknownC1", unknownC1Op, 0x17 },
   { "unknownC1", unknownC1Op, 0x18 },
   { "unknownC1", unknownC1Op, 0x19 },
   { "unknownC1", unknownC1Op, 0x1a },
   { "unknownC1", unknownC1Op, 0x1b },
   { "unknownC1", unknownC1Op, 0x1c },
   { "unknownC1", unknownC1Op, 0x1d },
   { "unknownC1", unknownC1Op, 0x1e },
   { "unknownC1", unknownC1Op, 0x1f },
   { "cvt.s.d", cvt_s_dOp, 0x20 },
   { "unknownC1", unknownC1Op, 0x21 },
   { "unknownC1", unknownC1Op, 0x22 },
   { "unknownC1", unknownC1Op, 0x23 },
   { "cvt.w.d", cvt_w_dOp, 0x24 },
   { "cvt.l.d", cvt_l_dOp, 0x25 },
   { "unknownC1", unknownC1Op, 0x26 },
   { "unknownC1", unknownC1Op, 0x27 },
   { "unknownC1", unknownC1Op, 0x28 },
   { "unknownC1", unknownC1Op, 0x29 },
   { "unknownC1", unknownC1Op, 0x2a },
   { "unknownC1", unknownC1Op, 0x2b },
   { "unknownC1", unknownC1Op, 0x2c },
   { "unknownC1", unknownC1Op, 0x2d },
   { "unknownC1", unknownC1Op, 0x2e },
   { "unknownC1", unknownC1Op, 0x2f },
   { "c.f.d", c_f_dOp, 0x30 },
   { "c.un.d", c_un_dOp, 0x31 },
   { "c.eq.d", c_eq_dOp, 0x32 },
   { "c.ueq.d", c_ueq_dOp, 0x33 },
   { "c.olt.d", c_olt_dOp, 0x34 },
   { "c.ult.d", c_ult_dOp, 0x35 },
   { "c.ole.d", c_ole_dOp, 0x36 },
   { "c.ule.d", c_ule_dOp, 0x37 },
   { "c.df.d", c_sf_dOp, 0x38 },
   { "c.ngle.d", c_ngle_dOp, 0x39 },
   { "c.deq.d", c_seq_dOp, 0x3a },
   { "c.ngl.d", c_ngl_dOp, 0x3b },
   { "c.lt.d", c_lt_dOp, 0x3c },
   { "c.nge.d", c_nge_dOp, 0x3d },
   { "c.le.d", c_le_dOp, 0x3e },
   { "c.ngt.d", c_ngt_dOp, 0x3f }
};

MipsC1XOpcodeDesc mipsC1XOpcodes[] = {
   { "lwxc1", unknownC1XOp, 0x00 },
   { "ldxc1", unknownC1XOp, 0x01 },
   { "unknownC1X", unknownC1XOp, 0x02 },
   { "unknownC1X", unknownC1XOp, 0x03 },
   { "unknownC1X", unknownC1XOp, 0x04 },
   { "unknownC1X", unknownC1XOp, 0x05 },
   { "unknownC1X", unknownC1XOp, 0x06 },
   { "unknownC1X", unknownC1XOp, 0x07 },
   { "unknownC1X", unknownC1XOp, 0x08 },
   { "unknownC1X", unknownC1XOp, 0x09 },
   { "unknownC1X", unknownC1XOp, 0x0a },
   { "unknownC1X", unknownC1XOp, 0x0b },
   { "unknownC1X", unknownC1XOp, 0x0c },
   { "unknownC1X", unknownC1XOp, 0x0d },
   { "unknownC1X", unknownC1XOp, 0x0e },
   { "prefx", unknownC1XOp, 0x0f },
   { "unknownC1X", unknownC1XOp, 0x10 },
   { "unknownC1X", unknownC1XOp, 0x11 },
   { "unknownC1X", unknownC1XOp, 0x12 },
   { "unknownC1X", unknownC1XOp, 0x13 },
   { "unknownC1X", unknownC1XOp, 0x14 },
   { "unknownC1X", unknownC1XOp, 0x15 },
   { "unknownC1X", unknownC1XOp, 0x16 },
   { "unknownC1X", unknownC1XOp, 0x17 },
   { "unknownC1X", unknownC1XOp, 0x18 },
   { "unknownC1X", unknownC1XOp, 0x19 },
   { "unknownC1X", unknownC1XOp, 0x1a },
   { "unknownC1X", unknownC1XOp, 0x1b },
   { "unknownC1X", unknownC1XOp, 0x1c },
   { "unknownC1X", unknownC1XOp, 0x1d },
   { "unknownC1X", unknownC1XOp, 0x1e },
   { "unknownC1X", unknownC1XOp, 0x1f },
   { "madd.s", madd_sOp, 0x20 },
   { "madd.d", madd_dOp, 0x21 },
   { "unknownC1X", unknownC1XOp, 0x22 },
   { "unknownC1X", unknownC1XOp, 0x23 },
   { "unknownC1X", unknownC1XOp, 0x24 },
   { "unknownC1X", unknownC1XOp, 0x25 },
   { "unknownC1X", unknownC1XOp, 0x26 },
   { "unknownC1X", unknownC1XOp, 0x27 },
   { "msub.s", msub_sOp, 0x28 },
   { "msub.d", msub_dOp, 0x29 },
   { "unknownC1X", unknownC1XOp, 0x2a },
   { "unknownC1X", unknownC1XOp, 0x2b },
   { "unknownC1X", unknownC1XOp, 0x2c },
   { "unknownC1X", unknownC1XOp, 0x2d },
   { "unknownC1X", unknownC1XOp, 0x2e },
   { "unknownC1X", unknownC1XOp, 0x2f },
   { "nmadd.s", nmadd_sOp, 0x30 },
   { "nmadd.d", nmadd_dOp, 0x31 },
   { "unknownC1X", unknownC1XOp, 0x32 },
   { "unknownC1X", unknownC1XOp, 0x33 },
   { "unknownC1X", unknownC1XOp, 0x34 },
   { "unknownC1X", unknownC1XOp, 0x35 },
   { "unknownC1X", unknownC1XOp, 0x36 },
   { "unknownC1X", unknownC1XOp, 0x37 },
   { "nmsub.s", nmsub_sOp, 0x38 },
   { "nmsub.d", nmsub_dOp, 0x39 },
   { "unknownC1X", unknownC1XOp, 0x3a },
   { "unknownC1X", unknownC1XOp, 0x3b },
   { "unknownC1X", unknownC1XOp, 0x3c },
   { "unknownC1X", unknownC1XOp, 0x3d },
   { "unknownC1X", unknownC1XOp, 0x3e },
   { "unknownC1X", unknownC1XOp, 0x3f }
};
