/*
 * Copyright (C) 1996-1998 by the Board of Trustees
 *    of Leland Stanford Junior University.
 * 
 * This file is part of the SimOS distribution. 
 * See LICENSE file for terms of the license. 
 *
 */

/****************************************************************
 * opcodes.c
 * 
 * Implementation of each MIPS instruction and declaration of the
 * table used to decode them.
 * 
 * $Author: verghese $
 * $Date: 1999/09/27 22:08:15 $
 *****************************************************************/
#include "simstats.h"
#include "opcodes.h"
#include "cpu.h"
#include "cpu_state.h"
#include "sim_error.h"
#include "cp0.h"
#include "fpu.h"
#include "memref.h"
#include "cpu_stats.h"
#include "hw_events.h"
#include "mdmx.h"
#ifdef SOLO
#  include "solo.h"
#endif

#define OP(NAME) \
Result NAME(Inst instr, CPUState *P)

#ifdef JUMP_ANNS
VA hideousJumpAnnotationExceptionHack[SIM_MAXCPUS];
#endif

MipsOpcodeDesc mipsSpecOpcodes[];
MipsOpcodeDesc mipsBcondOpcodes[];

/*
 * CHECK_64BIT_ALLOWED() - Check to see if 64bit instructions are permitted,
 *                         if not raise an exception. 
 */
#if defined(SIM_MIPS64)
#define CHECK_64BIT_ALLOWED(_p)  \
      if ((_p)->is32bitMode && !IS_KERNEL_MODE((_p))) {  \
         EXCEPTION((_p), EXC_II);                 \
         return FAILURE;                           \
      }
#else
#define CHECK_64BIT_ALLOWED(_p) \
      if (1) {                                           \
         EXCEPTION((_p), EXC_II);                 \
         return FAILURE;                                       \
      }
#endif

#define CURRENT_PC(_cpu) (pePtr[_cpu]->PC)

extern Result MipsyReadMem(VA, void *, RefSize, RefFlavor);
extern Result MipsyWriteMem(VA, void *, RefSize, RefFlavor);

/*
  extern Result MipsyWriteByte(VA, byte); 
  extern Result MipsyWriteHalf(VA, short); 
  extern Result MipsyWriteWord(VA, uint data, RefFlavor); 
  extern Result MipsyWriteDouble(VA, uint64, RefFlavor); 
  */
extern int FlashliteWQFull(int);

static void   LongMultiply32 (CPUState *, Reg32, Reg32);
static void   LongMultiply64 (CPUState *, Reg64, Reg64);
static Result Prefetch(CPUState *, VA, uint);
static Result CacheOp(CPUState *, uint, VA, PA);

extern PA  LLAddrs[MIPSY_MAX_CPUS];
extern int numLLactive;


/****************************************************************
 * M A J O R   O P C O D E S
 *****************************************************************/

OP(specOp) 
{
   uint   specialFunc = FUNC(instr);
   return mipsSpecOpcodes[specialFunc].func(instr, P);
}


OP(bcondOp) 
{
   uint bcondFunc    = RT(instr);
   return mipsBcondOpcodes[bcondFunc].func(instr, P);
}

OP(addiOp)
{
   int immediate     = IMMED(instr);
   Reg_s contents_rs = P->R[RS(instr)];
   Reg32_s result32  = (Reg32_s)contents_rs + SIGN_EXTEND32(16, immediate);
   
   if (ARITH_OVFL32(result32, (Reg32_s)contents_rs, 
                    SIGN_EXTEND32(16, immediate))) {
      EXCEPTION(P, EXC_OV);
      return FAILURE;
   }
   /* Don't write this value if there was an overflow */
   P->R[RT(instr)] = result32;
   return SUCCESS;
}

OP(addiuOp)
{
   int immediate     = IMMED(instr);
   Reg_s contents_rs = P->R[RS(instr)];
   P->R[RT(instr)] = (Reg32_s)contents_rs + SIGN_EXTEND32(16, immediate);
   return SUCCESS;

}

OP(andiOp)
{
   int immediate     = IMMED(instr);
   Reg_s contents_rs = P->R[RS(instr)];
   P->R[RT(instr)] = contents_rs & (0x0000ffff & immediate);
   return SUCCESS;
}

OP(beqOp)
{
   int immediate     = IMMED(instr);
   Reg_s contents_rs = P->R[RS(instr)];
   if (contents_rs == P->R[RT(instr)]) {
      P->branchTarget = P->nPC + (SIGN_EXTEND32(18, (immediate << 2)));
      P->branchStatus = BranchStatus_taken;
   } else { 
      P->branchStatus = BranchStatus_nottaken;
   }
   CPUVec.IncC0PerfCtr(P->myNum, 6); /* Branches resolved */
   return SUCCESS;
}

OP(beqlOp)
{
   int immediate     = IMMED(instr);
   Reg_s contents_rs = P->R[RS(instr)];
   if (contents_rs == P->R[RT(instr)]) {
      P->branchTarget = P->nPC + (SIGN_EXTEND32(18, (immediate << 2)));
      P->branchStatus = BranchStatus_taken;
   } else {
      /* Skip instruction in delay slot */
      P->nPC = P->nPC + 4;  
   }
   CPUVec.IncC0PerfCtr(P->myNum, 6); /* Branches resolved */
   return SUCCESS;
}

OP(bgtzOp)
{   
   int immediate     = IMMED(instr);
   Reg_s contents_rs = P->R[RS(instr)];
   if (!(SIGN_BIT(contents_rs)) && (contents_rs != 0)) {
      P->branchTarget = P->nPC + (SIGN_EXTEND32(18, (immediate << 2)));
      P->branchStatus = BranchStatus_taken;
   }  else { 
      P->branchStatus = BranchStatus_nottaken;
   }
   CPUVec.IncC0PerfCtr(P->myNum, 6); /* Branches resolved */
   return SUCCESS;
}

OP(bgtzlOp)
{
   int immediate     = IMMED(instr);
   Reg_s contents_rs = P->R[RS(instr)];
   if (!(SIGN_BIT(contents_rs)) && (contents_rs != 0)) {
      P->branchTarget = P->nPC + (SIGN_EXTEND32(18, (immediate << 2)));
      P->branchStatus = BranchStatus_taken;
   } else {
      /* Skip instruction in delay slot */
      P->nPC = P->nPC + 4;  
   }
   CPUVec.IncC0PerfCtr(P->myNum, 6); /* Branches resolved */
   return SUCCESS;
}

OP(blezOp)
{
   int immediate     = IMMED(instr);
   Reg_s contents_rs = P->R[RS(instr)];
   if ((SIGN_BIT(contents_rs) || (contents_rs == 0))) {
      P->branchTarget = P->nPC + (SIGN_EXTEND32(18, (immediate << 2)));
      P->branchStatus = BranchStatus_taken;
   }  else { 
      P->branchStatus = BranchStatus_nottaken;
   }
   CPUVec.IncC0PerfCtr(P->myNum, 6); /* Branches resolved */
   return SUCCESS;
}

OP(blezlOp)
{
   int immediate     = IMMED(instr);
   Reg_s contents_rs = P->R[RS(instr)];
   if ((SIGN_BIT(contents_rs) || (contents_rs == 0))) {
      P->branchTarget = P->nPC + (SIGN_EXTEND32(18, (immediate << 2)));
      P->branchStatus = BranchStatus_taken;
   } else {
      /* Skip instruction in delay slot */
      P->nPC = P->nPC + 4;  
   }
   CPUVec.IncC0PerfCtr(P->myNum, 6); /* Branches resolved */
   return SUCCESS;
}

OP(bneOp)
{
   int immediate     = IMMED(instr);
   Reg_s contents_rs = P->R[RS(instr)];
   if (contents_rs != P->R[RT(instr)]) {
      P->branchTarget = P->nPC + (SIGN_EXTEND32(18, (immediate << 2)));
      P->branchStatus = BranchStatus_taken;
   } else { 
      P->branchStatus = BranchStatus_nottaken;
   }
   CPUVec.IncC0PerfCtr(P->myNum, 6); /* Branches resolved */
   return SUCCESS;
}

OP(bnelOp)
{
   int immediate     = IMMED(instr);
   Reg_s contents_rs = P->R[RS(instr)];
   if (contents_rs != P->R[RT(instr)]) {
      P->branchTarget = P->nPC + (SIGN_EXTEND32(18, (immediate << 2)));
      P->branchStatus = BranchStatus_taken;
   } else {
      /* Skip instruction in delay slot */
      P->nPC = P->nPC + 4;  
   }
   CPUVec.IncC0PerfCtr(P->myNum, 6); /* Branches resolved */
   return SUCCESS;
}

OP(cacheOp)
{
   int immediate     = IMMED(instr);
   /* Be careful with this one. Mipsy's caches are
      physically indexed. */
   VA   vAddr;
   PA   pAddr;
   uint tlbFlavor = TLB_READING;
   void *bdoorAddr = 0;
   Reg_s contents_rs = P->R[RS(instr)];
   
#ifdef IRIX6_4               
   if (!IS_KERNEL_MODE(P)) {                 
      /*  is unusable only in kernel and or when the cu0 bit is set. */
      CauseReg  causeReg;
      StatusReg statusReg;
      statusReg.ts_data = P->CP0[C0_SR];
      if (!statusReg.s32.ts_cu0) {
         causeReg.tc_data = P->CP0[C0_CAUSE];
         causeReg.s32.tc_ce = 0;
         P->CP0[C0_CAUSE] = causeReg.tc_data;
         EXCEPTION(P, EXC_CPU);
         return FAILURE;
      }
   }
#endif
   /* This is a reserved instruction, but I'm allowing it
      to be used in user mode. */
   vAddr = SIGN_EXTEND(16, immediate) + contents_rs;
   
   if (TranslateVirtual(P, vAddr, &pAddr, &tlbFlavor,&bdoorAddr) != SUCCESS) {
      if (P->cpuStatus != cpu_running) {
         P->stalledInst = instr;
         STATS_SET(P->myNum, stallStart, MipsyReadTime(P->myNum));
      }
      return FAILURE;
   }
   if (CacheOp(P, instr, vAddr, pAddr) != SUCCESS) {
      return FAILURE;
   }
   CPUVec.IncC0PerfCtr(P->myNum, 2); /* Load/prefetch/sync/CacheOp issued */
   CPUVec.IncC0PerfCtr(P->myNum, 18); /* Load/prefetch/sync/CacheOp graduated */
   return SUCCESS;
}

OP(daddiOp) 
{
   int immediate     = IMMED(instr);
   Reg_s contents_rs = P->R[RS(instr)];
   Reg64_s result64 = contents_rs + SIGN_EXTEND64(16, immediate);

   CHECK_64BIT_ALLOWED(P);
   if (ARITH_OVFL64(result64, contents_rs,
                    SIGN_EXTEND64(16, immediate))) {
      EXCEPTION(P, EXC_OV);
      return FAILURE;
   }
   /* Don't write this value if there was an overflow */
   P->R[RT(instr)] = result64;
   return SUCCESS;
}

OP(daddiuOp)
{
   int immediate     = IMMED(instr);
   Reg_s contents_rs = P->R[RS(instr)];
   CHECK_64BIT_ALLOWED(P);
   P->R[RT(instr)] = contents_rs + SIGN_EXTEND64(16,immediate);
   return SUCCESS;
   
}

OP(jOp)
{
   int immediate     = IMMED(instr);
   Reg_s contents_rs = P->R[RS(instr)];
   P->branchTarget = (P->nPC & (Reg32_s)0xf0000000) | (TARGET(instr) << 2);
   P->branchStatus = BranchStatus_taken;
   CPUVec.IncC0PerfCtr(P->myNum, 6); /* Branches resolved */
   return SUCCESS;
}

OP(jalOp)
{
   int immediate     = IMMED(instr);
   Reg_s contents_rs = P->R[RS(instr)];
   P->R[REG_RA] = P->nPC + 4;
   P->branchTarget = (P->nPC & (Reg32_s)0xf0000000) | (TARGET(instr) << 2);
   P->branchStatus = BranchStatus_taken;
#ifdef JUMP_ANNS
   if (P->PC == hideousJumpAnnotationExceptionHack[P->myNum]) {
      hideousJumpAnnotationExceptionHack[P->myNum] = 0;
   } else {
      char target[32];
      sprintf(target, "0x%llx", (uint64)P->branchTarget);
      Tcl_SetVar(TCLInterp, "JumpTarget", target, TCL_GLOBAL_ONLY);
      AnnExec(annJAL);
   }
#endif
   CPUVec.IncC0PerfCtr(P->myNum, 6); /* Branches resolved */
   return SUCCESS;
   
}

OP(lbOp)
{
   int immediate     = IMMED(instr);
   Reg_s contents_rs = P->R[RS(instr)];
   byte data;
   VA   vAddr;
   Reg base = P->R[RS(instr)];
   
   vAddr = SIGN_EXTEND(16, immediate) + base; 
   
   if (MipsyReadMem(vAddr, &data, BYTE_SZ, NO_FLAVOR) != SUCCESS) {
      if (P->cpuStatus != cpu_running) {
         P->stalledInst = instr;
         STATS_SET(P->myNum, stallStart, MipsyReadTime(P->myNum));
      }
      return FAILURE;
   } else {
      P->R[RT(instr)]   = SIGN_EXTEND(8, data);
      CPUVec.IncC0PerfCtr(P->myNum, 2); /* Load/prefetch/sync/CacheOp issued */
      CPUVec.IncC0PerfCtr(P->myNum, 18); /* Load/prefetch/sync/CacheOp graduated */
      return SUCCESS;
   }
}

OP(lbuOp)
{ 
   int immediate     = IMMED(instr);
   Reg_s contents_rs = P->R[RS(instr)];
   byte data;
   VA   vAddr;
   Reg base = P->R[RS(instr)];
   
   vAddr = SIGN_EXTEND(16, immediate) + base;
   
   if (MipsyReadMem(vAddr, &data, BYTE_SZ, NO_FLAVOR) != SUCCESS) {
      if (P->cpuStatus != cpu_running) {
         P->stalledInst = instr;
         STATS_SET(P->myNum, stallStart, MipsyReadTime(P->myNum));
      }
      return FAILURE;
   } else {
      P->R[RT(instr)]   = 0x000000ff & data;
      CPUVec.IncC0PerfCtr(P->myNum, 2); /* Load/prefetch/sync/CacheOp issued */
      CPUVec.IncC0PerfCtr(P->myNum, 18); /* Load/prefetch/sync/CacheOp graduated */
      return SUCCESS;
   }
}

OP(lhOp)
{
   int immediate     = IMMED(instr);
   short data;
   VA    vAddr;
   Reg   base = P->R[RS(instr)];
   Reg_s contents_rs = P->R[RS(instr)];

   vAddr = SIGN_EXTEND(16, immediate) + base;
   
   if (vAddr & 0x00000001) {
      RECORD_EXCEPTION(P, EXC_RADE, E_VEC, vAddr,
                       P->CP0[C0_TLBHI], P->CP0[C0_CTXT], P->CP0[C0_XCTXT]);
      /* EXCEPTION(P->myNum,EXC_RADE); */
      return FAILURE;
   }
   
   if (MipsyReadMem(vAddr, &data, HALF_SZ, NO_FLAVOR) != SUCCESS) {
      if (P->cpuStatus != cpu_running) {
         P->stalledInst = instr;
         STATS_SET(P->myNum, stallStart, MipsyReadTime(P->myNum));
      }
      return FAILURE; 
   } else {
      P->R[RT(instr)]   = SIGN_EXTEND(16,data);
      CPUVec.IncC0PerfCtr(P->myNum, 2); /* Load/prefetch/sync/CacheOp issued */
      CPUVec.IncC0PerfCtr(P->myNum, 18); /* Load/prefetch/sync/CacheOp graduated */
      return SUCCESS;
   }
}

OP(lhuOp)
{
   int immediate     = IMMED(instr);
   Reg_s contents_rs = P->R[RS(instr)];
   short data;
   VA    vAddr;
   Reg   base = P->R[RS(instr)];
   
   vAddr = SIGN_EXTEND(16, immediate) + base;
   
   if (vAddr & 0x00000001) {
      RECORD_EXCEPTION(P, EXC_RADE, E_VEC, vAddr,
                       P->CP0[C0_TLBHI], P->CP0[C0_CTXT], P->CP0[C0_XCTXT]);
      /* EXCEPTION(P->myNum,EXC_RADE); */
      return FAILURE;
   }
   
   if (MipsyReadMem(vAddr, &data, HALF_SZ, NO_FLAVOR) != SUCCESS) {
      if (P->cpuStatus != cpu_running) {
         P->stalledInst = instr;
         STATS_SET(P->myNum, stallStart, MipsyReadTime(P->myNum));
      }
      return FAILURE;
   } else {
      P->R[RT(instr)]   = 0x0000ffff & data;
      CPUVec.IncC0PerfCtr(P->myNum, 2); /* Load/prefetch/sync/CacheOp issued */
      CPUVec.IncC0PerfCtr(P->myNum, 18); /* Load/prefetch/sync/CacheOp graduated */
      return SUCCESS;
   }
}

OP(luiOp)
{
   int immediate     = IMMED(instr);
   P->R[RT(instr)] = immediate << 16;
   return SUCCESS;
}

OP(llOp)
{
   int immediate     = IMMED(instr);
   Reg32_s data;
   VA   vAddr;
   Reg base = P->R[RS(instr)];
   
   vAddr = SIGN_EXTEND(16, immediate) + base;
   
   if (vAddr & 0x00000003) {
      RECORD_EXCEPTION(P, EXC_RADE, E_VEC, vAddr,
                       P->CP0[C0_TLBHI], P->CP0[C0_CTXT], P->CP0[C0_XCTXT]);
      /* EXCEPTION(P->myNum,EXC_RADE); */
      return FAILURE;
   }
   
   if (UNCACHED_LL_SC) {
      if (!P->LLbit) {
         numLLactive++;
      }
   }
   
   P->LLbit = 1; 
   if (MipsyReadMem(vAddr, &data, WORD_SZ, LL_FLAVOR) != SUCCESS) {
      if (P->cpuStatus != cpu_running) {
         P->stalledInst = instr;
         STATS_SET(P->myNum, stallStart, MipsyReadTime(P->myNum));
         STATS_SET(P->myNum, syncStallStart, MipsyReadTime(P->myNum));
         if (UNCACHED_LL_SC) {
            LLAddrs[P->myNum] = (PA) -1;
            if (P->LLbit) {
               numLLactive--; 
            }
         }
         
         P->LLbit = 0; 
      }
      return FAILURE;
   } else {
      P->R[RT(instr)] = data;
      STATS_INC(P->myNum, syncStats.lls, 1);
      STATS_ADD_INTERVAL(P->myNum, syncStats.llStallTime, syncStallStart);
      if (data & 1) 
         STATS_INC(P->myNum, syncStats.llNonZero, 1);
   }
   CPUVec.IncC0PerfCtr(P->myNum, 2); /* Load/prefetch/sync/CacheOp issued */
   CPUVec.IncC0PerfCtr(P->myNum, 18); /* Load/prefetch/sync/CacheOp graduated */
   return SUCCESS;
}

OP(lldOp)
{
   int immediate     = IMMED(instr);
   Reg64  data;
   VA   vAddr;
   Reg base = P->R[RS(instr)];
   
   CHECK_64BIT_ALLOWED(P);
   vAddr = SIGN_EXTEND(16, immediate) + base;
   
   if (vAddr & 0x00000007) {
      RECORD_EXCEPTION(P, EXC_RADE, E_VEC, vAddr,
                       P->CP0[C0_TLBHI], P->CP0[C0_CTXT], P->CP0[C0_XCTXT]);
      return FAILURE;
   }
   
   if (UNCACHED_LL_SC) {
      if (!P->LLbit) {
         numLLactive++;
      }
   }
   
   P->LLbit = 1; 
   if (MipsyReadMem(vAddr, &data, DOUBLE_SZ, LL_FLAVOR) != SUCCESS) {
      if (P->cpuStatus != cpu_running) {
         P->stalledInst = instr;
         STATS_SET(P->myNum, stallStart, MipsyReadTime(P->myNum));
         STATS_SET(P->myNum, syncStallStart, MipsyReadTime(P->myNum));
         if (UNCACHED_LL_SC) {
            LLAddrs[P->myNum] = (PA) -1;
            if (P->LLbit) {
               numLLactive--; 
            }
         }
         
         P->LLbit = 0; 
      }
      return FAILURE;
   } else {
      P->R[RT(instr)] = data;
      STATS_INC(P->myNum, syncStats.lls, 1);
      STATS_ADD_INTERVAL(P->myNum, syncStats.llStallTime, syncStallStart);
      if (data & 1) 
         STATS_INC(P->myNum, syncStats.llNonZero, 1);
   }
   CPUVec.IncC0PerfCtr(P->myNum, 2); /* Load/prefetch/sync/CacheOp issued */
   CPUVec.IncC0PerfCtr(P->myNum, 18); /* Load/prefetch/sync/CacheOp graduated */
   return SUCCESS;
}

OP(lwOp)
{
   int immediate     = IMMED(instr);
   Reg32_s data;
   VA   vAddr;
   Reg base = P->R[RS(instr)];
   
   vAddr  = SIGN_EXTEND(16, immediate) + base;
#ifndef SOLO
   /* check for unaligned loads of vaddr 3 -
    * this is a syscall exception left over from base mode days
    * and can be removed when all the old disks are gone
    */
   if ((vAddr == 3) && (RS(instr) == 0)) {
      int syscallNum = P->R[REG_V0] - 1000;
      
      STATS_INC(P->myNum, numSyscalls, 1);
      if ((syscallNum < MAX_SYSCALL) && (syscallNum >= 0)) { 
         STATS_INC(P->myNum, syscallCount[syscallNum], 1);
      }
      EXCEPTION(P, EXC_SYSCALL);
      return FAILURE;
   }
#endif
   
   if (vAddr & 0x00000003) {
      RECORD_EXCEPTION(P, EXC_RADE, E_VEC, vAddr,
                       P->CP0[C0_TLBHI], P->CP0[C0_CTXT], P->CP0[C0_XCTXT]);
      return FAILURE;
   }
   
   if (MipsyReadMem(vAddr, &data, WORD_SZ, NO_FLAVOR) != SUCCESS) {
      if (P->cpuStatus != cpu_running) {
         P->stalledInst = instr;
         STATS_SET(P->myNum, stallStart, MipsyReadTime(P->myNum));
      }
      return FAILURE;
   } else {
      P->R[RT(instr)]   = data;
      CPUVec.IncC0PerfCtr(P->myNum, 2); /* Load/prefetch/sync/CacheOp issued */
      CPUVec.IncC0PerfCtr(P->myNum, 18); /* Load/prefetch/sync/CacheOp graduated */
      return SUCCESS;
   }
}

OP(ldOp)
{
   int immediate     = IMMED(instr);
   uint64  data;
   VA   vAddr;
   Reg base = P->R[RS(instr)];
   
   CHECK_64BIT_ALLOWED(P);
   vAddr = SIGN_EXTEND(16, immediate) + base;
   if (vAddr & 0x00000007) {
      RECORD_EXCEPTION(P, EXC_RADE, E_VEC, vAddr,
                       P->CP0[C0_TLBHI], P->CP0[C0_CTXT], P->CP0[C0_XCTXT]);
      return FAILURE;
   }
   if (MipsyReadMem(vAddr, &data, DOUBLE_SZ, NO_FLAVOR) != SUCCESS) {
      if (P->cpuStatus != cpu_running) {
         P->stalledInst = instr;
         STATS_SET(P->myNum, stallStart, MipsyReadTime(P->myNum));
      }
      return FAILURE;
   } else {
      P->R[RT(instr)] = data;
   }
   CPUVec.IncC0PerfCtr(P->myNum, 2); /* Load/prefetch/sync/CacheOp issued */
   CPUVec.IncC0PerfCtr(P->myNum, 18); /* Load/prefetch/sync/CacheOp graduated */
   return SUCCESS;
}

OP(ldc2Op)
{
   int immediate     = IMMED(instr);
   /* NOTE: This is not a real "ld" instruction since a real 
    * ld instruction would load a 64bit value into a 64bit 
    * registers. Since our registers are only 32bits we
    * load into two adjacent registers. 
    * BEWARE: THIS IS A HACK.
    */
   uint64 data;
   VA   vAddr;
   Reg base = P->R[RS(instr)];
   
   vAddr  = SIGN_EXTEND(16, immediate) + base;
   
   if (vAddr & 0x00000007) {
      RECORD_EXCEPTION(P, EXC_RADE, E_VEC, vAddr,
                       P->CP0[C0_TLBHI], P->CP0[C0_CTXT], P->CP0[C0_XCTXT]);
      /* EXCEPTION(P->myNum,EXC_RADE); */
      return FAILURE;
   }
   if (MipsyReadMem(vAddr, &data, DOUBLE_SZ, NO_FLAVOR) != SUCCESS) {
      if (P->cpuStatus != cpu_running) {
         P->stalledInst = instr;
         STATS_SET(P->myNum, stallStart, MipsyReadTime(P->myNum));
      }
      return FAILURE;
   } else {
      if (SAME_ENDIAN) { 
         P->R[RT(instr)]   = ((Reg32_s *)&data)[0];
         P->R[RT(instr)+1] = ((Reg32_s *)&data)[1];
      } else { 
         P->R[RT(instr)]   = ((Reg32_s *)&data)[1];
         P->R[RT(instr)+1] = ((Reg32_s *)&data)[0];
      }
      
   }
   return SUCCESS;
}

OP(lwc1Op)
{
   int immediate     = IMMED(instr);
   Reg32_s data;
   VA   vAddr;
   Reg base = P->R[RS(instr)];
   StatusReg statusReg;
   statusReg.ts_data = P->CP0[C0_SR];
   if (!(statusReg.s32.ts_cu1)) {
      CauseReg  causeReg;
      
      causeReg.tc_data = P->CP0[C0_CAUSE];
      causeReg.s32.tc_ce = 1;
      P->CP0[C0_CAUSE] = causeReg.tc_data;
      EXCEPTION(P,EXC_CPU);
      return FAILURE;
   }
   
   vAddr = SIGN_EXTEND(16, immediate) + base;
   if (vAddr & 0x00000003) {
      RECORD_EXCEPTION(P, EXC_RADE, E_VEC, vAddr,
                       P->CP0[C0_TLBHI], P->CP0[C0_CTXT], P->CP0[C0_XCTXT]);
      /* EXCEPTION(P->myNum,EXC_RADE); */
      return FAILURE;
   }
   
   if (MipsyReadMem(vAddr, &data, WORD_SZ, NO_FLAVOR) != SUCCESS) {
      if (P->cpuStatus != cpu_running) {
         P->stalledInst = instr;
         STATS_SET(P->myNum, stallStart, MipsyReadTime(P->myNum));
      }
      return FAILURE;
   } else {
      IntReg(RT(instr)) = data;
      CPUVec.IncC0PerfCtr(P->myNum, 2); /* Load/prefetch/sync/CacheOp issued */
      CPUVec.IncC0PerfCtr(P->myNum, 18); /* Load/prefetch/sync/CacheOp graduated */
      return SUCCESS;
   }
}

OP(ldc1Op)
{
   int immediate     = IMMED(instr);
   uint64  data;
   VA   vAddr;
   Reg base = P->R[RS(instr)];
   StatusReg statusReg;
   
   statusReg.ts_data = P->CP0[C0_SR];
   if (!(statusReg.s32.ts_cu1)) {
      CauseReg  causeReg;
      
      causeReg.tc_data = P->CP0[C0_CAUSE];
      causeReg.s32.tc_ce = 1;
      P->CP0[C0_CAUSE] = causeReg.tc_data;
      EXCEPTION(P,EXC_CPU);
      return FAILURE;
   }
   vAddr = SIGN_EXTEND(16, immediate) + base;
   if (vAddr & 0x00000007) {
      RECORD_EXCEPTION(P, EXC_RADE, E_VEC, vAddr,
                       P->CP0[C0_TLBHI], P->CP0[C0_CTXT], P->CP0[C0_XCTXT]);
      /* EXCEPTION(P->myNum,EXC_RADE); */
      return FAILURE;
   }
   if (MipsyReadMem(vAddr, &data, DOUBLE_SZ, NO_FLAVOR) != SUCCESS) {
      if (P->cpuStatus != cpu_running) {
         P->stalledInst = instr;
         STATS_SET(P->myNum, stallStart, MipsyReadTime(P->myNum));
      }
      return FAILURE;
   } else {
      DoubleReg(RT(instr)) = *((double *)&data);
      CPUVec.IncC0PerfCtr(P->myNum, 2); /* Load/prefetch/sync/CacheOp issued */
      CPUVec.IncC0PerfCtr(P->myNum, 18); /* Load/prefetch/sync/CacheOp graduated */
      return SUCCESS;
   }
}

OP(ldlOp)
{
   int immediate     = IMMED(instr);
   int  byte;
   Reg64 data;
   VA   vAddr;
   Reg base = P->R[RS(instr)];
   Reg  contents_rt = P->R[RT(instr)];
   
   CHECK_64BIT_ALLOWED(P);
      vAddr = SIGN_EXTEND(16, immediate) + base;
   byte  = vAddr & 0x7;
   
   vAddr = vAddr & (Reg32_s)0xfffffff8;
   if (MipsyReadMem(vAddr, &data, DOUBLE_SZ, NO_FLAVOR) != SUCCESS) {
      if (P->cpuStatus != cpu_running) {
         P->stalledInst = instr;
         STATS_SET(P->myNum, stallStart, MipsyReadTime(P->myNum));
      }
      return FAILURE;
   } else {
      /* Big Endian */
      switch (byte) {
      case 0:
         P->R[RT(instr)] = data;
         break;
      case 1:
         P->R[RT(instr)] = ((data & 0x00ffffffffffffffLL) << 8) | 
            (contents_rt & 0x00000000000000ffLL);
         break;
      case 2:
         P->R[RT(instr)] = ((data & 0x0000ffffffffffffLL) << 16) | 
            (contents_rt & 0x000000000000ffffLL);
         break;
      case 3:
         P->R[RT(instr)] = ((data & 0x000000ffffffffffLL) << 24) | 
            (contents_rt & 0x0000000000ffffffLL);
         break;
      case 4:
         P->R[RT(instr)] = ((data & 0x00000000ffffffffLL) << 32) | 
            (contents_rt & 0x00000000ffffffffLL);
         break;
      case 5:
         P->R[RT(instr)] = ((data & 0x0000000000ffffffLL) << 40) | 
            (contents_rt & 0x000000ffffffffffLL);
         break;
      case 6:
         P->R[RT(instr)] = ((data & 0x000000000000ffffLL) << 48) | 
            (contents_rt & 0x0000ffffffffffffLL);
         break;
      case 7:
         P->R[RT(instr)] = ((data & 0x00000000000000ffLL) << 56) | 
            (contents_rt & 0x00ffffffffffffffLL);
         break;
      }
   }
   CPUVec.IncC0PerfCtr(P->myNum, 2); /* Load/prefetch/sync/CacheOp issued */
   CPUVec.IncC0PerfCtr(P->myNum, 18); /* Load/prefetch/sync/CacheOp graduated */
   return SUCCESS;
}

OP(ldrOp)
{
   int immediate     = IMMED(instr);
   Reg64 data;
   int  byte;
   VA   vAddr;
   Reg  contents_rt = P->R[RT(instr)];
   Reg base = P->R[RS(instr)];
   
   CHECK_64BIT_ALLOWED(P);
   vAddr = SIGN_EXTEND(16, immediate) + base;
   byte  = vAddr & 0x7;
   
   vAddr = vAddr & (Reg32_s)0xfffffff8;
   
   if (MipsyReadMem(vAddr, &data, DOUBLE_SZ, NO_FLAVOR) != SUCCESS) {
      if (P->cpuStatus != cpu_running) {
         P->stalledInst = instr;
         STATS_SET(P->myNum, stallStart, MipsyReadTime(P->myNum));
      }
      return FAILURE;
   } else {
      /* Big Endian */
      switch (byte) {
      case 0:
         P->R[RT(instr)] = ((data & 0xff00000000000000LL) >> 56) | 
            (contents_rt & 0xffffffffffffff00LL);
         break;
      case 1:
         P->R[RT(instr)] = ((data & 0xffff000000000000LL) >> 48) | 
            (contents_rt & 0xffffffffffff0000LL);
         break;
      case 2:
         P->R[RT(instr)] = ((data & 0xffffff0000000000LL) >> 40) | 
            (contents_rt & 0xffffffffff000000LL);
         break;
      case 3:
         P->R[RT(instr)] = ((data & 0xffffffff00000000LL) >> 32) | 
            (contents_rt & 0xffffffff00000000LL);
         break;
      case 4:
         P->R[RT(instr)] = ((data & 0xffffffffff000000LL) >> 24) | 
            (contents_rt & 0xffffff0000000000LL);
         break;
      case 5:
         P->R[RT(instr)] = ((data & 0xffffffffffff0000LL) >> 16) | 
            (contents_rt & 0xffff000000000000LL);
         break;
      case 6:
         P->R[RT(instr)] = ((data & 0xffffffffffffff00LL) >> 8) | 
            (contents_rt & 0xff00000000000000LL);
         break;
      case 7:
         P->R[RT(instr)] = data;
         break;
      }
   }
   CPUVec.IncC0PerfCtr(P->myNum, 2); /* Load/prefetch/sync/CacheOp issued */
   CPUVec.IncC0PerfCtr(P->myNum, 18); /* Load/prefetch/sync/CacheOp graduated */
   return SUCCESS;
}

OP(lwc2Op)
{
   int immediate     = IMMED(instr);
   CPUError("LWC2 operations not implemented\n"); 
   return FAILURE;
}

OP(lwlOp) 
{
   int immediate     = IMMED(instr);
   int  byte;
   Reg32_s data;
   VA   vAddr;
   Reg base = P->R[RS(instr)];
   Reg  contents_rt = P->R[RT(instr)];
   
   vAddr = SIGN_EXTEND(16, immediate) + base;
   byte  = vAddr & 0x3;
   
   vAddr = vAddr & (Reg32_s)0xfffffffc;
   if (MipsyReadMem(vAddr, &data, WORD_SZ, NO_FLAVOR) != SUCCESS) {
      if (P->cpuStatus != cpu_running) {
         P->stalledInst = instr;
         STATS_SET(P->myNum, stallStart, MipsyReadTime(P->myNum));
      }
      return FAILURE;
   } else {
      /* Big Endian */
      switch (byte) {
      case 0:
         P->R[RT(instr)] = data;
            break;
      case 1:
         P->R[RT(instr)] = ((data & 0x00ffffff) << 8) | 
            ((Reg32_s)contents_rt & 0x000000ff);
         break;
      case 2:
         P->R[RT(instr)] = ((data & 0x0000ffff) << 16) | 
            ((Reg32_s)contents_rt & 0x0000ffff);
         break;
      case 3:
         P->R[RT(instr)] = ((data & 0x000000ff) << 24) | 
            ((Reg32_s)contents_rt & 0x00ffffff);
         break;
      }
   }
   CPUVec.IncC0PerfCtr(P->myNum, 2); /* Load/prefetch/sync/CacheOp issued */
   CPUVec.IncC0PerfCtr(P->myNum, 18); /* Load/prefetch/sync/CacheOp graduated */
   return SUCCESS;
}

OP(lwrOp) 
{
   int immediate     = IMMED(instr);
   Reg32_s data;
   int  byte;
   VA   vAddr;
   Reg  contents_rt = P->R[RT(instr)];
   Reg base = P->R[RS(instr)];
   
   vAddr = SIGN_EXTEND(16, immediate) + base;
   byte  = vAddr & 0x3;
   
   vAddr = vAddr & (Reg32_s)0xfffffffc;
   
   if (MipsyReadMem(vAddr, &data, WORD_SZ, NO_FLAVOR) != SUCCESS) {
      if (P->cpuStatus != cpu_running) {
         P->stalledInst = instr;
         STATS_SET(P->myNum, stallStart, MipsyReadTime(P->myNum));
      }
      return FAILURE;
   } else {
      /* Big Endian */
      switch (byte) {
      case 0:
         P->R[RT(instr)] = ((data & 0xff000000) >> 24) | 
            ((Reg32_s)contents_rt & 0xffffff00);
         break;
      case 1:
         P->R[RT(instr)] = ((data & 0xffff0000) >> 16) | 
            ((Reg32_s)contents_rt & 0xffff0000);
         break;
         
      case 2:
         P->R[RT(instr)] = ((data & 0xffffff00) >> 8) | 
            ((Reg32_s)contents_rt & 0xff000000);
         break;
         
      case 3:
         P->R[RT(instr)] = data;
         break;
      }
   }
   CPUVec.IncC0PerfCtr(P->myNum, 2); /* Load/prefetch/sync/CacheOp issued */
   CPUVec.IncC0PerfCtr(P->myNum, 18); /* Load/prefetch/sync/CacheOp graduated */
   return SUCCESS;
}

OP(lwuOp)
{
   int immediate     = IMMED(instr);
   VA   vAddr;
   Reg base = P->R[RS(instr)];
   Reg32 data = 0;
   
   CHECK_64BIT_ALLOWED(P);
   vAddr  = SIGN_EXTEND(16, immediate) + base;
   
   if (vAddr & 0x00000003) {
      RECORD_EXCEPTION(P, EXC_RADE, E_VEC, vAddr,
                       P->CP0[C0_TLBHI], P->CP0[C0_CTXT], P->CP0[C0_XCTXT]);
      return FAILURE;
   }
   
   if (MipsyReadMem(vAddr, &data, WORD_SZ, NO_FLAVOR) != SUCCESS) {
      if (P->cpuStatus != cpu_running) {
         P->stalledInst = instr;
         STATS_SET(P->myNum, stallStart, MipsyReadTime(P->myNum));
      }
      return FAILURE;
   } else {
      P->R[RT(instr)]   = data;
   }
   return SUCCESS;
}

OP(prefOp)
{
   int immediate     = IMMED(instr);
   VA   vAddr;
   Result tRet;
   Reg base = P->R[RS(instr)];
   uint  hint = RT(instr);
   
   vAddr = (SIGN_EXTEND(16, immediate) + base);
   
   tRet = Prefetch(P,vAddr, hint);
   
   if (tRet == STALL) {
      P->cpuStatus = cpu_stalled;
      P->stalledInst = instr;
      STATS_SET(P->myNum, stallStart, MipsyReadTime(P->myNum));
      return FAILURE;
   }
   CPUVec.IncC0PerfCtr(P->myNum, 2); /* Load/prefetch/sync/CacheOp issued */
   CPUVec.IncC0PerfCtr(P->myNum, 18); /* Load/prefetch/sync/CacheOp graduated */
   return SUCCESS;
}
            
OP(oriOp)
{
   int immediate     = IMMED(instr);
   Reg_s contents_rs = P->R[RS(instr)];
   P->R[RT(instr)] = contents_rs | (0x0000ffff & immediate);
   return SUCCESS;
}

OP(sbOp)
{
   int immediate     = IMMED(instr);
   VA  vAddr;
   Reg base = P->R[RS(instr)];
   byte data = P->R[RT(instr)];

   vAddr = SIGN_EXTEND(16, immediate) + base;
   
   if (MipsyWriteMem(vAddr, &data, BYTE_SZ, NO_FLAVOR) 
       != SUCCESS) {
      if (P->cpuStatus != cpu_running) {
         P->stalledInst = instr;
         STATS_SET(P->myNum, stallStart, MipsyReadTime(P->myNum));
      }
      return FAILURE;
   }
   CPUVec.IncC0PerfCtr(P->myNum, 3); /* Stores issued */
   CPUVec.IncC0PerfCtr(P->myNum, 19); /* Stores graduated */
   return SUCCESS;
}

OP(shOp)
{
   int immediate     = IMMED(instr);
   VA  vAddr;
   Reg base = P->R[RS(instr)];
   short data = P->R[RT(instr)];

   vAddr = SIGN_EXTEND(16, immediate) + base;
   if (vAddr & 0x00000001) {
      EXCEPTION(P,EXC_WADE);
      return FAILURE;
   }
   if (MipsyWriteMem(vAddr, &data, HALF_SZ, NO_FLAVOR) 
       != SUCCESS) {
      if (P->cpuStatus != cpu_running) {
         P->stalledInst = instr;
         STATS_SET(P->myNum, stallStart, MipsyReadTime(P->myNum));
      }
      return FAILURE;
   }
   CPUVec.IncC0PerfCtr(P->myNum, 3); /* Stores issued */
   CPUVec.IncC0PerfCtr(P->myNum, 19); /* Stores graduated */
   return SUCCESS;
}

OP(sltiOp)
{
   int immediate     = IMMED(instr);
   Reg_s contents_rs = P->R[RS(instr)];
   if (contents_rs < (Reg32_s)SIGN_EXTEND(16, immediate)) {
      P->R[RT(instr)] = 1;
   } else {
      P->R[RT(instr)] = 0;
   }
   return SUCCESS;
}

OP(sltiuOp)
{
   int immediate     = IMMED(instr);
   Reg_s contents_rs = P->R[RS(instr)];
   if ((Reg)contents_rs < (Reg)(Reg32_s)SIGN_EXTEND(16, immediate)) {
      P->R[RT(instr)] = 1;
   } else {
      P->R[RT(instr)] = 0;
   }
   return SUCCESS;
}

OP(scOp)
{
   int immediate     = IMMED(instr);
   VA  vAddr;
   bool scSuccess = FALSE;
   Reg base = P->R[RS(instr)];
   Result ret;
   uint data = P->R[RT(instr)];
   
   vAddr = SIGN_EXTEND(16, immediate) + base;
   if (vAddr & 0x00000003) {
      EXCEPTION(P,EXC_WADE);
      return FAILURE;
   }
   
   if (UNCACHED_LL_SC) {  
      if (P->LLbit) {
         /* This is a successful SC */
         scSuccess = TRUE;
         numLLactive--;
         LLAddrs[P->myNum] = (PA) -1;
      }
   }
   
   ret = MipsyWriteMem(vAddr, &data, WORD_SZ, SC_FLAVOR);
   if (ret == STALL) { 
      if (P->cpuStatus != cpu_running) {
         P->stalledInst = instr;
         STATS_SET(P->myNum, stallStart, MipsyReadTime(P->myNum));
         STATS_SET(P->myNum, syncStallStart, MipsyReadTime(P->myNum));
      }
      return FAILURE;
   } else if (ret == FAILURE || ret == BUSERROR) {
      STATS_SET(P->myNum, syncStallStart, MipsyReadTime(P->myNum));
      return FAILURE;
   }
   
   P->LLbit = 0;
   
   P->R[RT(instr)] = (ret == SUCCESS) ? 1 : 0;
   
   if (UNCACHED_LL_SC) {  
      if (scSuccess) {
         P->R[RT(instr)] = 1;
      } else {
         P->R[RT(instr)] = 0;
      }
   }
   
   STATS_INC(P->myNum, syncStats.scs, 1);
   if (P->R[RT(instr)] == 0) {
      STATS_INC(P->myNum, syncStats.scFailed, 1);
   }
   STATS_ADD_INTERVAL(P->myNum, syncStats.scStallTime, syncStallStart);

   CPUVec.IncC0PerfCtr(P->myNum, 3); /* Stores issued */
   CPUVec.IncC0PerfCtr(P->myNum, 19); /* Stores graduated */
   CPUVec.IncC0PerfCtr(P->myNum, 4); /* SCs issued */
   CPUVec.IncC0PerfCtr(P->myNum, 20); /* SCs graduated */
   if (!scSuccess) {
      CPUVec.IncC0PerfCtr(P->myNum, 5); /* Failed SCs */
   }

   return SUCCESS;
}

OP(scdOp)
{
   int immediate     = IMMED(instr);
   VA  vAddr;
   bool scSuccess = FALSE;
   Reg base = P->R[RS(instr)];
   Result ret;
   uint64 data;

   CHECK_64BIT_ALLOWED(P);
   vAddr = SIGN_EXTEND(16, immediate) + base;
   if (vAddr & 0x00000007) {
      EXCEPTION(P,EXC_WADE);
      return FAILURE;
   }
   
   if (UNCACHED_LL_SC) {  
      if (P->LLbit) {
         /* This is a successful SC */
         scSuccess = TRUE;
         numLLactive--;
         LLAddrs[P->myNum] = (PA) -1;
      }
   }

   data = P->R[RT(instr)];
   ret = MipsyWriteMem(vAddr, &data, DOUBLE_SZ, SC_FLAVOR);
   if (ret == STALL) { 
      if (P->cpuStatus != cpu_running) {
         P->stalledInst = instr;
         
         STATS_SET(P->myNum, stallStart, MipsyReadTime(P->myNum));
         STATS_SET(P->myNum, syncStallStart, MipsyReadTime(P->myNum));
      }
      return FAILURE;
   } else if (ret == FAILURE || ret == BUSERROR) {
      STATS_SET(P->myNum, syncStallStart, MipsyReadTime(P->myNum));
      return FAILURE;
   }
   
   P->LLbit = 0;
   
   P->R[RT(instr)] = (ret == SUCCESS) ? 1 : 0;
   
   if (UNCACHED_LL_SC) {  
      if (scSuccess) {
         P->R[RT(instr)] = 1;
      } else {
         P->R[RT(instr)] = 0;
      }
   }
   
   STATS_INC(P->myNum, syncStats.scs, 1);
   if (P->R[RT(instr)] == 0) {
      STATS_INC(P->myNum, syncStats.scFailed, 1);
   }
   STATS_ADD_INTERVAL(P->myNum, syncStats.scStallTime, syncStallStart);

   CPUVec.IncC0PerfCtr(P->myNum, 3); /* Stores issued */
   CPUVec.IncC0PerfCtr(P->myNum, 19); /* Stores graduated */
   CPUVec.IncC0PerfCtr(P->myNum, 4); /* SCs issued */
   CPUVec.IncC0PerfCtr(P->myNum, 20); /* SCs graduated */
   if (!scSuccess) {
      CPUVec.IncC0PerfCtr(P->myNum, 5); /* Failed SCs */
   }

   return SUCCESS;
}

OP(swOp)
{
   int immediate     = IMMED(instr);
   VA  vAddr;
   Reg base = P->R[RS(instr)];
   uint data = P->R[RT(instr)];

   vAddr = SIGN_EXTEND(16, immediate) + base;
   if (vAddr & 0x00000003) {
      EXCEPTION(P,EXC_WADE);
      return FAILURE;
   }
   
   if (MipsyWriteMem(vAddr, &data, WORD_SZ, NO_FLAVOR) != SUCCESS) {
      if (P->cpuStatus != cpu_running) {
         P->stalledInst = instr;
         STATS_SET(P->myNum, stallStart, MipsyReadTime(P->myNum));
      }
      return FAILURE;
   }
   CPUVec.IncC0PerfCtr(P->myNum, 3); /* Stores issued */
   CPUVec.IncC0PerfCtr(P->myNum, 19); /* Stores graduated */
   return SUCCESS;
}


OP(sdOp)
{
   int immediate     = IMMED(instr);
   VA  vAddr;
   Reg base = P->R[RS(instr)];
   uint64 data = 0;
   
   CHECK_64BIT_ALLOWED(P);
   vAddr = SIGN_EXTEND(16, immediate) + base;
   
   if (vAddr & 0x00000007) {
      EXCEPTION(P,EXC_WADE);
      return FAILURE;
   } else {
      data = P->R[RT(instr)];
      if (MipsyWriteMem(vAddr, &data, DOUBLE_SZ, NO_FLAVOR) != SUCCESS) {
         if (P->cpuStatus != cpu_running) {
            P->stalledInst = instr;
            STATS_SET(P->myNum, stallStart, MipsyReadTime(P->myNum));
         }
         return FAILURE;
      }
   }
   CPUVec.IncC0PerfCtr(P->myNum, 3); /* Stores issued */
   CPUVec.IncC0PerfCtr(P->myNum, 19); /* Stores graduated */
   return SUCCESS;
}

OP(sdc2Op)
{
   int immediate     = IMMED(instr);
   /* 
    * This instruction is different from the real machine.
    * see note under ld_op for details.
    */
   VA  vAddr;
   Reg base = P->R[RS(instr)];
   uint64 data;
   
   vAddr = SIGN_EXTEND(16, immediate) + base;
   if (vAddr & 0x00000007) {
      EXCEPTION(P,EXC_WADE);
      return FAILURE;
   }
   if (SAME_ENDIAN) { 
      ((uint *)&data)[0] = P->R[RT(instr)];
      ((uint *)&data)[1] = P->R[RT(instr)+1];
   } else { 
      ((uint *)&data)[1] = P->R[RT(instr)];
      ((uint *)&data)[0] = P->R[RT(instr)+1];
   } 
   
   
   if (MipsyWriteMem(vAddr, &data, DOUBLE_SZ, NO_FLAVOR) != SUCCESS) {
      if (P->cpuStatus != cpu_running) {
         P->stalledInst = instr;
         STATS_SET(P->myNum, stallStart, MipsyReadTime(P->myNum));
      }
      return FAILURE;
   }
   CPUVec.IncC0PerfCtr(P->myNum, 3); /* Stores issued */
   CPUVec.IncC0PerfCtr(P->myNum, 19); /* Stores graduated */
   return SUCCESS;
}

OP(sdc1Op)
{
   int immediate     = IMMED(instr);
   VA  vAddr;
   Reg base = P->R[RS(instr)];
   uint64 data = 0;
   StatusReg statusReg;
   
   statusReg.ts_data = P->CP0[C0_SR];
   if (!(statusReg.s32.ts_cu1)) {
      CauseReg  causeReg;
      
      causeReg.tc_data = P->CP0[C0_CAUSE];
      causeReg.s32.tc_ce = 1;
      P->CP0[C0_CAUSE] = causeReg.tc_data;
      EXCEPTION(P,EXC_CPU);
      return FAILURE;
   }
   
   vAddr = SIGN_EXTEND(16, immediate) + base;
   
   if (vAddr & 0x00000007) {
      EXCEPTION(P,EXC_WADE);
      return FAILURE;
   } else {
      *((double *)&data) = DoubleReg(RT(instr));
      if (MipsyWriteMem(vAddr, &data, DOUBLE_SZ, NO_FLAVOR) != SUCCESS) {
         if (P->cpuStatus != cpu_running) {
            P->stalledInst = instr;
            STATS_SET(P->myNum, stallStart, MipsyReadTime(P->myNum));
         }
         return FAILURE;
      }
   }
   CPUVec.IncC0PerfCtr(P->myNum, 3); /* Stores issued */
   CPUVec.IncC0PerfCtr(P->myNum, 19); /* Stores graduated */
   return SUCCESS;
}

OP(sdlOp)
{
   int immediate     = IMMED(instr);
   Reg64  data;
   int  byte;
   VA   vAddr;
   Reg base = P->R[RS(instr)];
   Reg64  contents_rt = P->R[RT(instr)];
   
   CHECK_64BIT_ALLOWED(P);
   vAddr = SIGN_EXTEND(16, immediate) + base;
   byte  = vAddr & 0x7;
   vAddr &= (Reg32_s)~0x7; 
   if (MipsyReadMem(vAddr, &data, DOUBLE_SZ, GETX_FLAVOR) != SUCCESS) {
      if (P->cpuStatus != cpu_running) {
         P->stalledInst = instr;
         STATS_SET(P->myNum, stallStart, MipsyReadTime(P->myNum));
      }
      return FAILURE;
   } else {
      /* Big Endian */
      switch (byte) {
      case 0:
         data = contents_rt;
         break;
      case 1:
         data =             (data & 0xff00000000000000LL) 
            | (contents_rt >> 8 & 0x00ffffffffffffffLL);
         break;
      case 2:
         data =             (data & 0xffff000000000000LL) 
            | (contents_rt >> 16 & 0x0000ffffffffffffLL);
         break;
      case 3:
         data =             (data & 0xffffff0000000000LL) 
            | (contents_rt >> 24 & 0x000000ffffffffffLL);
         break;
      case 4:
         data =             (data & 0xffffffff00000000LL) 
            | (contents_rt >> 32 & 0x00000000ffffffffLL);
         break;
      case 5:
         data =             (data & 0xffffffffff000000LL) 
            | (contents_rt >> 40 & 0x0000000000ffffffLL);
         break;
      case 6:
         data =             (data & 0xffffffffffff0000LL) 
            | (contents_rt >> 48 & 0x000000000000ffffLL);
         break;
      case 7:
         data =             (data & 0xffffffffffffff00LL) 
            | (contents_rt >> 56 & 0x00000000000000ffLL);
         break;
         
      }
      if (MipsyWriteMem(vAddr, &data, DOUBLE_SZ, NO_FLAVOR) != SUCCESS) {
         if (P->cpuStatus != cpu_running) {
            P->stalledInst = instr;
            STATS_SET(P->myNum, stallStart, MipsyReadTime(P->myNum));
         }
         return FAILURE;
      } 
   }
   CPUVec.IncC0PerfCtr(P->myNum, 3); /* Stores issued */
   CPUVec.IncC0PerfCtr(P->myNum, 19); /* Stores graduated */
   return SUCCESS;
}

OP(sdrOp)
{
   int immediate     = IMMED(instr);
   Reg64 data;
   uint byte;
   VA   vAddr;
   Reg base = P->R[RS(instr)];
   Reg64  contents_rt = P->R[RT(instr)];
   
   CHECK_64BIT_ALLOWED(P);
   vAddr = SIGN_EXTEND(16, immediate) + base;
   byte = vAddr & 0x7;
   vAddr &= (Reg32_s)~0x7; 
   if (MipsyReadMem(vAddr, &data, DOUBLE_SZ, GETX_FLAVOR) != SUCCESS) {
      if (P->cpuStatus != cpu_running) {
         P->stalledInst = instr;
         STATS_SET(P->myNum, stallStart, MipsyReadTime(P->myNum));
      }
      return FAILURE;
   } else {
      /* Big Endian */
      switch (byte) {
      case 0:
         data = ((contents_rt << 56) & 0xff00000000000000LL) 
            | (data & 0x00ffffffffffffffLL);
         break;
      case 1:
         data = ((contents_rt << 48) & 0xffff000000000000LL) 
            | (data & 0x0000ffffffffffffLL);
         break;
      case 2:
         data = ((contents_rt << 40) & 0xffffff0000000000LL) 
            | (data & 0x000000ffffffffffLL);
         break;
      case 3:
         data = ((contents_rt << 32) & 0xffffffff00000000LL) 
            | (data & 0x00000000ffffffffLL);
         break;
      case 4:
         data = ((contents_rt << 24) & 0xffffffffff000000LL) 
            | (data & 0x0000000000ffffffLL);
         break;
      case 5:
         data = ((contents_rt << 16) & 0xffffffffffff0000LL) 
            | (data & 0x000000000000ffffLL);
         break;
      case 6:
         data = ((contents_rt << 8) &  0xffffffffffffff00LL) 
            | (data & 0x00000000000000ffLL);
         break;
      case 7:
         data = contents_rt;
         break;
      }
      
      if (MipsyWriteMem(vAddr, &data, DOUBLE_SZ, NO_FLAVOR) != SUCCESS) {
         if (P->cpuStatus != cpu_running) {
            P->stalledInst = instr;
            STATS_SET(P->myNum, stallStart, MipsyReadTime(P->myNum));
         }
         return FAILURE;
      }
   }
   CPUVec.IncC0PerfCtr(P->myNum, 3); /* Stores issued */
   CPUVec.IncC0PerfCtr(P->myNum, 19); /* Stores graduated */
   return SUCCESS;
}

OP(swc1Op)
{
   int immediate     = IMMED(instr);
   VA  vAddr;
   Reg base = P->R[RS(instr)];
   StatusReg statusReg;
   uint data = IntReg(RT(instr));

   statusReg.ts_data = P->CP0[C0_SR];

   if (!(statusReg.s32.ts_cu1)) {
      CauseReg  causeReg;
      
      causeReg.tc_data = P->CP0[C0_CAUSE];
      causeReg.s32.tc_ce = 1;
      P->CP0[C0_CAUSE] = causeReg.tc_data;
      EXCEPTION(P,EXC_CPU);
      return FAILURE;
   }
   
   vAddr = SIGN_EXTEND(16, immediate) + base;
   if (vAddr & 0x00000003) {
      EXCEPTION(P,EXC_WADE);
      return FAILURE;
   } else {
      if (MipsyWriteMem(vAddr, &data, WORD_SZ, NO_FLAVOR) != SUCCESS) {
         if (P->cpuStatus != cpu_running) {
            P->stalledInst = instr;
            STATS_SET(P->myNum, stallStart, MipsyReadTime(P->myNum));
         }
         return FAILURE;
      }
   }
   CPUVec.IncC0PerfCtr(P->myNum, 3); /* Stores issued */
   CPUVec.IncC0PerfCtr(P->myNum, 19); /* Stores graduated */
   return SUCCESS;
}

OP(swc2Op)
{
   int immediate     = IMMED(instr);
   CPUError("SWCx not implemented!\n");
   return FAILURE;
}

OP(swlOp)
{
   int immediate     = IMMED(instr);
   Reg32_s data;
   int  byte;
   VA   vAddr;

   Reg base = P->R[RS(instr)];
   Reg  contents_rt = P->R[RT(instr)];
   
   vAddr = SIGN_EXTEND(16, immediate) + base;
   byte  = vAddr & 0x3;
   vAddr &= (Reg32_s)~0x3; 
   if (MipsyReadMem(vAddr, &data, WORD_SZ, GETX_FLAVOR) != SUCCESS) {
      if (P->cpuStatus != cpu_running) {
         P->stalledInst = instr;
         STATS_SET(P->myNum, stallStart, MipsyReadTime(P->myNum));
      }
      return FAILURE;
   } else {
      /* Big Endian */
      switch (byte) {
      case 0:
         data = contents_rt;
         break;
      case 1:
         data = (data & 0xff000000) | (contents_rt >> 8 & 0xffffff);
         break;
      case 2:
         data = (data & 0xffff0000) | (contents_rt >> 16 & 0xffff);
         break;
      case 3:
         data = (data & 0xffffff00) | (contents_rt >> 24 & 0xff);
         break;
      }
      if (MipsyWriteMem(vAddr, &data, WORD_SZ, NO_FLAVOR) != SUCCESS) {
         if (P->cpuStatus != cpu_running) {
            P->stalledInst = instr;
            STATS_SET(P->myNum, stallStart, MipsyReadTime(P->myNum));
         }
         return FAILURE;
      } 
   }
   CPUVec.IncC0PerfCtr(P->myNum, 3); /* Stores issued */
   CPUVec.IncC0PerfCtr(P->myNum, 19); /* Stores graduated */
   return SUCCESS;
}

OP(swrOp)
{
   int immediate     = IMMED(instr);
   Reg32_s data;
   uint byte;
   VA   vAddr;
   Reg base = P->R[RS(instr)];
   Reg  contents_rt = P->R[RT(instr)];
   
   vAddr = SIGN_EXTEND(16, immediate) + base;
   byte = vAddr & 0x3;
   vAddr &= (Reg32_s)~0x3; 
   if (MipsyReadMem(vAddr, &data, WORD_SZ, GETX_FLAVOR) != SUCCESS) {
      if (P->cpuStatus != cpu_running) {
         P->stalledInst = instr;
         STATS_SET(P->myNum, stallStart, MipsyReadTime(P->myNum));
      }
      return FAILURE;
   } else {
      /* Big Endian */
      switch (byte) {
      case 0:
         data = ((contents_rt << 24) & 0xff000000) | (data & 0xffffff);
         break;
      case 1:
         data = ((contents_rt << 16) & 0xffff0000) | (data & 0xffff);
         break;
      case 2:
         data = ((contents_rt << 8) & 0xffffff00)  | (data & 0xff) ;
         break;
      case 3:
         data = contents_rt;
         break;
      }
      
      if (MipsyWriteMem(vAddr, &data, WORD_SZ, NO_FLAVOR) != SUCCESS) {
         if (P->cpuStatus != cpu_running) {
            P->stalledInst = instr;
            STATS_SET(P->myNum, stallStart, MipsyReadTime(P->myNum));
         }
         return FAILURE;
      }
   }
   CPUVec.IncC0PerfCtr(P->myNum, 3); /* Stores issued */
   CPUVec.IncC0PerfCtr(P->myNum, 19); /* Stores graduated */
   return SUCCESS;
}

OP(xoriOp)
{
   int immediate     = IMMED(instr);
   Reg_s contents_rs = P->R[RS(instr)];
   P->R[RT(instr)] = contents_rs ^ (0x0000ffff & immediate);
   return SUCCESS;
}

OP(cop0Op)
{
   int immediate     = IMMED(instr);
   int retval = 0;
   VA  pcVal = P->PC;
   
   retval = ExecuteC0Instruction(P, instr);
   
   if (retval != C0_SUCCESS) {
      if (retval == C0_ILLEGAL_INST) {
         CPUWarning("Illegal inst at %#llx on CPU %d RA %#llx\n", 
                    (Reg64)P->PC, P->myNum, (Reg64)P->R[REG_RA]);
         ASSERT(0);
         EXCEPTION(P, EXC_II);
      }
      ASSERT(retval == C0_CONTINUE);
      return FAILURE;
   }
   return SUCCESS;
}

OP(cop1Op)
{
   int retval    =  ExecuteC1Instruction(P, instr);
   
   if (retval != C1_SUCCESS) {
      ASSERT(retval == C1_FAILURE);
      return FAILURE;
   }
   return SUCCESS;
}

OP(cop1xOp)
{
   int immediate     = IMMED(instr);
   uint rt = RT(instr);
   int retval = 0;
   StatusReg statusReg;
   
   /* Need to make sure xx is enable if not in kernel mode */
   statusReg.ts_data = P->CP0[C0_SR];
   if (IS_USER_MODE(P) && !(statusReg.s32.ts_xx)) {
      EXCEPTION(P, EXC_II);
      return FAILURE;
   }
   if (!(statusReg.s32.ts_cu1)) {
      CauseReg  causeReg;
      
      causeReg.tc_data = P->CP0[C0_CAUSE];
      causeReg.s32.tc_ce = 1;
      P->CP0[C0_CAUSE] = causeReg.tc_data;
      EXCEPTION(P, EXC_CPU);
      return FAILURE;
   }
   
   switch (FUNC(instr)) {
      
   case lwxc1_op: {
      Reg32_s data;
      VA   vAddr;
      Reg base = P->R[RS(instr)];
      
      vAddr = base + P->R[RT(instr)];
      if (vAddr & 0x00000003) {
         RECORD_EXCEPTION(P, EXC_RADE, E_VEC, vAddr,
                          P->CP0[C0_TLBHI], P->CP0[C0_CTXT], P->CP0[C0_XCTXT]);
         return FAILURE;
      }
      
      if (MipsyReadMem(vAddr, &data, WORD_SZ, NO_FLAVOR) != SUCCESS) {
         if (P->cpuStatus != cpu_running) {
            P->stalledInst = instr;
            STATS_SET(P->myNum, stallStart, MipsyReadTime(P->myNum));
         }
         return FAILURE;
      } else {
         IntReg(rt) = data;
         CPUVec.IncC0PerfCtr(P->myNum, 2); /* Load/prefetch/sync/CacheOp issued */
         CPUVec.IncC0PerfCtr(P->myNum, 18); /* Load/prefetch/sync/CacheOp graduated */
         break;
      }
   }
   
   case ldxc1_op: 
      {
         Reg64 data;
         VA   vAddr;
         Reg base = P->R[RS(instr)];
         
         vAddr = base + P->R[RT(instr)];
         if (vAddr & 0x00000007) {
            RECORD_EXCEPTION(P, EXC_RADE, E_VEC, vAddr,
                             P->CP0[C0_TLBHI], P->CP0[C0_CTXT], P->CP0[C0_XCTXT]);
            return FAILURE;
         }
         
         if (MipsyReadMem(vAddr, &data, DOUBLE_SZ, NO_FLAVOR) != SUCCESS) {
            if (P->cpuStatus != cpu_running) {
               P->stalledInst = instr;
               STATS_SET(P->myNum, stallStart, MipsyReadTime(P->myNum));
            }
            return FAILURE;
         } else {
            DoubleReg(rt) = *((double *)&data);
            CPUVec.IncC0PerfCtr(P->myNum, 2); /* Load/prefetch/sync/CacheOp issued */
            CPUVec.IncC0PerfCtr(P->myNum, 18); /* Load/prefetch/sync/CacheOp graduated */
            break;
         }
      }
   
   case swxc1_op:
      {
         VA  vAddr;
         Reg base = P->R[RS(instr)];
         uint data = IntReg(rt);
         
         vAddr = base + P->R[RT(instr)];
         if (vAddr & 0x00000003) {
            RECORD_EXCEPTION(P, EXC_WADE, E_VEC, vAddr,
                             P->CP0[C0_TLBHI], P->CP0[C0_CTXT], P->CP0[C0_XCTXT]);
            return FAILURE;
         } else {
            if (MipsyWriteMem(vAddr, &data, WORD_SZ, NO_FLAVOR) != SUCCESS) {
               if (P->cpuStatus != cpu_running) {
                  P->stalledInst = instr;
                  STATS_SET(P->myNum, stallStart, MipsyReadTime(P->myNum));
               }
               return FAILURE;
            }
         }
         CPUVec.IncC0PerfCtr(P->myNum, 3); /* Stores issued */
         CPUVec.IncC0PerfCtr(P->myNum, 19); /* Stores graduated */
         break;
      }
   case sdxc1_op:
      {
         VA  vAddr;
         Reg base = P->R[RS(instr)];
         uint64 data = LongLongReg(rt);

         vAddr = base + P->R[RT(instr)];
         if (vAddr & 0x00000007) {
            RECORD_EXCEPTION(P, EXC_WADE, E_VEC, vAddr,
                             P->CP0[C0_TLBHI], P->CP0[C0_CTXT], P->CP0[C0_XCTXT]);
            return FAILURE;
         } else {
            if (MipsyWriteMem(vAddr, &data, DOUBLE_SZ, NO_FLAVOR) != SUCCESS) {
               if (P->cpuStatus != cpu_running) {
                  P->stalledInst = instr;
                  STATS_SET(P->myNum, stallStart, MipsyReadTime(P->myNum));
               }
               return FAILURE;
            }
         }
         CPUVec.IncC0PerfCtr(P->myNum, 3); /* Stores issued */
         CPUVec.IncC0PerfCtr(P->myNum, 19); /* Stores graduated */
         break;
      }
   
   case pfetch_op: {
      VA  vAddr;
      Reg base = P->R[RS(instr)];
      uint hint = RD(instr);
      Result ret;
      
      vAddr = (base + P->R[RT(instr)]);
      ret = Prefetch(P,vAddr, hint);
      if (ret == STALL) {
         P->cpuStatus = cpu_stalled;
         P->stalledInst = instr;
         STATS_SET(P->myNum, stallStart, MipsyReadTime(P->myNum));
         return FAILURE;
      }
      CPUVec.IncC0PerfCtr(P->myNum, 2); /* Load/prefetch/sync/CacheOp issued */
      CPUVec.IncC0PerfCtr(P->myNum, 18); /* Load/prefetch/sync/CacheOp graduated */
      break;
   }
   default:
      
      retval = ExecuteC1Instruction(P, instr);
      
      if (retval != C1_SUCCESS) {
         ASSERT(retval == C1_FAILURE);
         return FAILURE;
      }
      break;
   }
   return SUCCESS;
}


OP(cop2Op)
{
   CPUWarning("COP2 (MDMX) instruction at PC %#llx\n", (Reg64)P->PC);
   return ExecuteMDMXInstruction(P, instr);
}



OP(undefOp)
{
   CPUWarning("Illegal inst at %#llx on CPU %d RA %#llx\n", 
              (Reg64)P->PC, P->myNum, (Reg64)P->R[REG_RA]);
   EXCEPTION(P, EXC_II);
   return FAILURE;
}


/*****************************************************************
 * S P E C   O P C O D E S
 *****************************************************************/

OP(addOp)
{
   Reg32_s result32;
   uint  rd          = RD(instr);
   Reg_s contents_rs = P->R[RS(instr)];
   Reg_s contents_rt = P->R[RT(instr)];

   result32 = (Reg32_s) contents_rs + (Reg32_s)contents_rt;
   if (ARITH_OVFL32 (result32, (Reg32_s)contents_rs, 
                     (Reg32_s)contents_rt)) {
      EXCEPTION(P, EXC_OV);
      return FAILURE;
   }
   /* Don't write this if there was an exception */
   P->R[rd] = result32;
   return SUCCESS;
}

OP(adduOp)
{
   uint  rd          = RD(instr);
   Reg_s contents_rs = P->R[RS(instr)];
   Reg_s contents_rt = P->R[RT(instr)];

   P->R[rd] = (Reg32_s)contents_rs + (Reg32_s)contents_rt;
   return SUCCESS;
}

OP(andOp)
{
   uint  rd          = RD(instr);
   Reg_s contents_rs = P->R[RS(instr)];
   Reg_s contents_rt = P->R[RT(instr)];

   P->R[rd] = contents_rs & contents_rt;
   return SUCCESS;
}

OP(breakOp) 
{
   uint  rd          = RD(instr);
   Reg_s contents_rs = P->R[RS(instr)];
   Reg_s contents_rt = P->R[RT(instr)];

#define BREAK_CODE(_inst) (((unsigned)_inst >> 16 ) & 0x3ff )
   
   if (IS_KERNEL_MODE(P) ||
       ((IS_SUPERV_MODE(P) && (BREAK_CODE(instr) < 0x99)))) {
      rd = MipsyDebug(P->myNum, 1);
      if (rd) {
         /* this is an alternate path around the end of the loop
          * if the PC should not increment.  The only case that uses
          * it currently is when a break instruction is hit.
          *
          * We also skip the check for post-PC annotations: their
          * exactly-once semantics means we should execute them
          * after the reexecution.
          */
         CPUError("BREAK OP at %#x\n", P->PC);
#ifdef NOTDEF
         FIX_REG_ZERO;
         INSTRUCTION_DONE_EVENT();
#endif
         return FAILURE;
      }
   } else { 
      EXCEPTION(P, EXC_BREAK);
      return FAILURE;
   }
   return SUCCESS;
}
     
OP(daddOp)
{
   Reg64_s result64;
   uint  rd          = RD(instr);
   Reg_s contents_rs = P->R[RS(instr)];
   Reg_s contents_rt = P->R[RT(instr)];

   CHECK_64BIT_ALLOWED(P);
   result64 =  contents_rs + contents_rt;
   if (ARITH_OVFL64 (result64, contents_rs, contents_rt)) {
      EXCEPTION(P, EXC_OV);
      return FAILURE;
   }
   /* Don't write this if there was an exception */
   P->R[rd] =  result64;
   return SUCCESS;
}

OP(dadduOp)
{
   uint  rd          = RD(instr);
   Reg_s contents_rs = P->R[RS(instr)];
   Reg_s contents_rt = P->R[RT(instr)];
   
   CHECK_64BIT_ALLOWED(P);
   P->R[rd] =  contents_rs + contents_rt;
   return SUCCESS;
}

OP(ddivOp)
{
   uint  rd          = RD(instr);
   Reg_s contents_rs = P->R[RS(instr)];
   Reg_s contents_rt = P->R[RT(instr)];

   CHECK_64BIT_ALLOWED(P);
   if (contents_rt != 0) {
      P->LO = (Reg_s)contents_rs / (Reg_s)contents_rt;
      P->HI = (Reg_s)contents_rs % (Reg_s)contents_rt; 
   }
   return SUCCESS;
}

OP(ddivuOp)
{
   uint  rd          = RD(instr);
   Reg_s contents_rs = P->R[RS(instr)];
   Reg_s contents_rt = P->R[RT(instr)];
   
   CHECK_64BIT_ALLOWED(P);
   if (contents_rt != 0) {
      P->LO = (Reg)contents_rs / (Reg)contents_rt;
      P->HI = (Reg)contents_rs % (Reg)contents_rt; 
   }
   return SUCCESS;
}

OP(divOp)
{
   uint  rd          = RD(instr);
   Reg_s contents_rs = P->R[RS(instr)];
   Reg_s contents_rt = P->R[RT(instr)];

   if (contents_rt != 0) {
      P->LO = (Reg32_s)contents_rs / (Reg32_s)contents_rt;
      P->HI = (Reg32_s)contents_rs % (Reg32_s)contents_rt;
   }
   return SUCCESS;
}

OP(divuOp)
{
   uint  rd          = RD(instr);
   Reg_s contents_rs = P->R[RS(instr)];
   Reg_s contents_rt = P->R[RT(instr)];

   if (contents_rt != 0) {
      P->LO = (Reg32)contents_rs / (Reg32)contents_rt;
      P->HI = (Reg32)contents_rs % (Reg32)contents_rt;
   }
   return SUCCESS;
}

OP(dmultOp)
{
   uint  rd          = RD(instr);
   Reg_s contents_rs = P->R[RS(instr)];
   Reg_s contents_rt = P->R[RT(instr)];
   bool negative = FALSE;

   CHECK_64BIT_ALLOWED(P);
   
   if ((Reg64_s)contents_rs < 0) {
      negative = TRUE;
      contents_rs = -contents_rs;
   }
   if ((Reg64_s)contents_rt < 0) {
      negative = !negative;
      contents_rt = -contents_rt;
   }
   LongMultiply64(P,contents_rs,contents_rt);
   if (negative) {
      P->LO = ~P->LO;
      P->HI = ~P->HI;
      P->LO += 1;
      if (P->LO == 0)
         P->HI += 1;
   }
   return SUCCESS;
}

OP(dmultuOp)
{
   uint  rd          = RD(instr);
   Reg_s contents_rs = P->R[RS(instr)];
   Reg_s contents_rt = P->R[RT(instr)];

   CHECK_64BIT_ALLOWED(P);
   LongMultiply64(P,(Reg64)contents_rs, (Reg64)contents_rt);
   return SUCCESS;
}

OP(dsllOp)
{
   uint  rd          = RD(instr);
   Reg_s contents_rs = P->R[RS(instr)];
   Reg_s contents_rt = P->R[RT(instr)];
   
   CHECK_64BIT_ALLOWED(P);
   P->R[rd] = (Reg_s)contents_rt << SHAMT(instr);
   return SUCCESS;
}

OP(dsllvOp)
{
   uint  rd          = RD(instr);
   Reg_s contents_rs = P->R[RS(instr)];
   Reg_s contents_rt = P->R[RT(instr)];

   CHECK_64BIT_ALLOWED(P);
   P->R[rd] = (Reg_s)contents_rt << (contents_rs & 0x3f);
   return SUCCESS;
}

OP(dsll32Op)
{
   uint  rd          = RD(instr);
   Reg_s contents_rs = P->R[RS(instr)];
   Reg_s contents_rt = P->R[RT(instr)];
   
   CHECK_64BIT_ALLOWED(P);
   P->R[rd] = (Reg_s)contents_rt << (SHAMT(instr)+32);
   return SUCCESS;
}

OP(dsraOp)
{
   uint  rd          = RD(instr);
   Reg_s contents_rs = P->R[RS(instr)];
   Reg_s contents_rt = P->R[RT(instr)];

   CHECK_64BIT_ALLOWED(P);
   P->R[rd] = (Reg64_s)contents_rt >> SHAMT(instr);
   return SUCCESS;
}

OP(dsravOp)
{
   uint  rd          = RD(instr);
   Reg_s contents_rs = P->R[RS(instr)];
   Reg_s contents_rt = P->R[RT(instr)];

   CHECK_64BIT_ALLOWED(P);
   P->R[rd] = (Reg64_s)contents_rt >> (contents_rs & 0x3f);
   return SUCCESS;
}

OP(dsra32Op)
{
   uint  rd          = RD(instr);
   Reg_s contents_rs = P->R[RS(instr)];
   Reg_s contents_rt = P->R[RT(instr)];

   CHECK_64BIT_ALLOWED(P);
   P->R[rd] = (Reg64_s)contents_rt >> (SHAMT(instr)+32);
   return SUCCESS;
}

OP(dsrlOp)
{
   uint  rd          = RD(instr);
   Reg_s contents_rs = P->R[RS(instr)];
   Reg_s contents_rt = P->R[RT(instr)];

   CHECK_64BIT_ALLOWED(P);
   P->R[rd] = ((Reg64) contents_rt) >> SHAMT(instr);
   return SUCCESS;
}

OP(dsrlvOp)
{
   uint  rd          = RD(instr);
   Reg_s contents_rs = P->R[RS(instr)];
   Reg_s contents_rt = P->R[RT(instr)];

   CHECK_64BIT_ALLOWED(P);
   P->R[rd] = ((Reg64) contents_rt) >> (contents_rs & 0x3f);
   return SUCCESS;
}

OP(dsrl32Op)
{
   uint  rd          = RD(instr);
   Reg_s contents_rs = P->R[RS(instr)];
   Reg_s contents_rt = P->R[RT(instr)];

   CHECK_64BIT_ALLOWED(P);
   P->R[rd] = ((Reg64) contents_rt) >> (SHAMT(instr)+32);
   return SUCCESS;
}

OP(dsubOp)
{
   uint  rd          = RD(instr);
   Reg_s contents_rs = P->R[RS(instr)];
   Reg_s contents_rt = P->R[RT(instr)];
   Reg64_s diff;

   CHECK_64BIT_ALLOWED(P);
   diff = contents_rs - contents_rt;
   if (SIGN_BIT64(contents_rs) != SIGN_BIT64(contents_rt)
             && SIGN_BIT64(contents_rs) != SIGN_BIT64(diff)) {
      EXCEPTION(P, EXC_OV);
      return FAILURE;
   }
   P->R[rd] = diff;
   return SUCCESS;
}

OP(dsubuOp)
{
   uint  rd          = RD(instr);
   Reg_s contents_rs = P->R[RS(instr)];
   Reg_s contents_rt = P->R[RT(instr)];

   CHECK_64BIT_ALLOWED(P);
   P->R[rd] = ((Reg64)contents_rs - (Reg64)contents_rt);
   return SUCCESS;

}

OP(jalrOp)
{
   uint  rd          = RD(instr);
   Reg_s contents_rs = P->R[RS(instr)];
   Reg_s contents_rt = P->R[RT(instr)];

   P->R[rd] = P->nPC + 4;
   P->branchTarget = contents_rs;
   P->branchStatus = BranchStatus_taken;
#ifdef JUMP_ANNS
   if (P->PC == hideousJumpAnnotationExceptionHack[P->myNum]) {
      hideousJumpAnnotationExceptionHack[P->myNum] = 0;
   } else {
      char target[32];
      sprintf(target, "0x%llx", (uint64)P->branchTarget);
      Tcl_SetVar(TCLInterp, "JumpTarget", target, TCL_GLOBAL_ONLY);
      AnnExec(annJAL);
   }
#endif
   CPUVec.IncC0PerfCtr(P->myNum, 6); /* Branches resolved */
   return SUCCESS;
}

OP(jrOp)
{
   uint  rd          = RD(instr);
   Reg_s contents_rs = P->R[RS(instr)];

   P->branchTarget = contents_rs;
   P->branchStatus = BranchStatus_taken;
#ifdef JUMP_ANNS
   if (P->PC == hideousJumpAnnotationExceptionHack[P->myNum]) {
      hideousJumpAnnotationExceptionHack[P->myNum] = 0;
   } else {
      if (RS(instr) == 31) {
         AnnExec(annJR);
      }
   }
#endif
   CPUVec.IncC0PerfCtr(P->myNum, 6); /* Branches resolved */
   return SUCCESS;
}

OP(mfhiOp)
{
   uint  rd          = RD(instr);

   P->R[rd] = P->HI;
   return SUCCESS;
}

OP(mfloOp)
{
   uint  rd          = RD(instr);

   P->R[rd] = P->LO;
   return SUCCESS;
}

OP(movcOp)
{
   uint  rd          = RD(instr);
   int retval = 0;
   StatusReg statusReg;
   
   /* Need to make sure xx is enable if not in kernel mode */
   statusReg.ts_data = P->CP0[C0_SR];
   if (IS_USER_MODE(P) && !(statusReg.s32.ts_xx)) {
      EXCEPTION(P, EXC_II);
      return FAILURE;
   }
   
   retval = ExecuteC1Instruction(P, instr);
   
   if (retval != C1_SUCCESS) {
      ASSERT(retval == C1_FAILURE);
      return FAILURE;
   }
   return SUCCESS; 
}

OP(movnOp)
{
   uint  rd          = RD(instr);
   Reg_s contents_rs = P->R[RS(instr)];
   Reg_s contents_rt = P->R[RT(instr)];
   StatusReg statusReg;
               
   /* Need to make sure xx is enable if not in kernel mode */
   statusReg.ts_data = P->CP0[C0_SR];
   if (IS_USER_MODE(P) && !(statusReg.s32.ts_xx)) {
      EXCEPTION(P, EXC_II);
      return FAILURE;
   }
   
   if (contents_rt != 0) {
      P->R[rd] = contents_rs;
   }
   return SUCCESS; 
}

OP(movzOp)
{
   uint  rd          = RD(instr);
   Reg_s contents_rs = P->R[RS(instr)];
   Reg_s contents_rt = P->R[RT(instr)];
   StatusReg statusReg;
   
   /* Need to make sure xx is enable if not in kernel mode */
   statusReg.ts_data = P->CP0[C0_SR];
   if (IS_USER_MODE(P) && !(statusReg.s32.ts_xx)) {
      EXCEPTION(P, EXC_II);
      return FAILURE;
   }
   
   if (contents_rt == 0) {
      P->R[rd] = contents_rs;
   }
   return SUCCESS; 
}

OP(mthiOp)
{
   uint  rd          = RD(instr);
   Reg_s contents_rs = P->R[RS(instr)];

   P->HI = contents_rs;
   return SUCCESS;
}

OP(mtloOp)
{
   Reg_s contents_rs = P->R[RS(instr)];

   P->LO = contents_rs;
   return SUCCESS;
}

OP(multOp)
{
   uint  rd          = RD(instr);
   Reg_s contents_rs = P->R[RS(instr)];
   Reg_s contents_rt = P->R[RT(instr)];
   bool negative = FALSE;
            
   if (contents_rs < 0) {
      negative = TRUE;
      contents_rs = -contents_rs;
   }
   if (contents_rt < 0) {
      negative = !negative;
      contents_rt = -contents_rt;
   }
   LongMultiply32(P,(Reg32)contents_rs, (Reg32)contents_rt);
   if (negative) {
      P->LO = ~P->LO;
      P->HI = ~P->HI;
      P->LO += 1;
      if (P->LO == 0)
         P->HI += 1;
   }
   return SUCCESS;
}

OP(multuOp)
{
   Reg_s contents_rs = P->R[RS(instr)];
   Reg_s contents_rt = P->R[RT(instr)];

   LongMultiply32(P,(Reg32)contents_rs, (Reg32)contents_rt);
   return SUCCESS;
}

OP(norOp)
{
   uint  rd          = RD(instr);
   Reg_s contents_rs = P->R[RS(instr)];
   Reg_s contents_rt = P->R[RT(instr)];

   P->R[rd] =  ~ (contents_rs | contents_rt);
   return SUCCESS;
}

OP(orOp)
{
   uint  rd          = RD(instr);
   Reg_s contents_rs = P->R[RS(instr)];
   Reg_s contents_rt = P->R[RT(instr)];

   P->R[rd] = contents_rs | contents_rt;
   return SUCCESS;
}

OP(sllOp)
{
   uint  rd          = RD(instr);
   Reg_s contents_rs = P->R[RS(instr)];
   Reg_s contents_rt = P->R[RT(instr)];

   P->R[rd] = (Reg32_s)contents_rt << SHAMT(instr);
   return SUCCESS;
}

OP(sllvOp)
{
   uint  rd          = RD(instr);
   Reg_s contents_rs = P->R[RS(instr)];
   Reg_s contents_rt = P->R[RT(instr)];

   P->R[rd] = (Reg32_s)contents_rt << (contents_rs & 0x1f);
   return SUCCESS;
}

OP(sltOp)
{
   uint  rd          = RD(instr);
   Reg_s contents_rs = P->R[RS(instr)];
   Reg_s contents_rt = P->R[RT(instr)];

   if (contents_rs < contents_rt)
      P->R[rd] = 1;
   else
      P->R[rd] = 0;
   return SUCCESS;
}

OP(sltuOp)
{
   uint  rd          = RD(instr);
   Reg_s contents_rs = P->R[RS(instr)];
   Reg_s contents_rt = P->R[RT(instr)];

   if ((Reg) contents_rs < (Reg) contents_rt)
      P->R[rd] = 1;
   else
      P->R[rd] = 0;
   return SUCCESS;
}

OP(sraOp)
{
   uint  rd          = RD(instr);
   Reg_s contents_rs = P->R[RS(instr)];
   Reg_s contents_rt = P->R[RT(instr)];

   P->R[rd] = (Reg32_s)contents_rt >> SHAMT(instr);
   return SUCCESS;
}

OP(sravOp)
{
   uint  rd          = RD(instr);
   Reg_s contents_rs = P->R[RS(instr)];
   Reg_s contents_rt = P->R[RT(instr)];

   P->R[rd] = (Reg32_s)contents_rt >> (contents_rs & 0x1f);
   return SUCCESS;
}

OP(srlOp)
{
   uint  rd          = RD(instr);
   Reg_s contents_rs = P->R[RS(instr)];
   Reg_s contents_rt = P->R[RT(instr)];

   P->R[rd] = (Reg32_s)(((Reg32) contents_rt) >> SHAMT(instr));
   return SUCCESS;
}

OP(srlvOp)
{
   uint  rd          = RD(instr);
   Reg_s contents_rs = P->R[RS(instr)];
   Reg_s contents_rt = P->R[RT(instr)];

   P->R[rd] = (Reg32_s)((Reg32) contents_rt >> (contents_rs & 0x1f));
   return SUCCESS;
}

OP(subOp)
{
   uint  rd          = RD(instr);
   Reg_s contents_rs = P->R[RS(instr)];
   Reg_s contents_rt = P->R[RT(instr)];
   Reg32_s diff;
   
   diff = (Reg32_s)contents_rs - (Reg32_s)contents_rt;
   if (SIGN_BIT32(contents_rs) != SIGN_BIT32(contents_rt)
       && SIGN_BIT32(contents_rs) != SIGN_BIT32(diff)) {
      EXCEPTION(P, EXC_OV);
      return FAILURE;
   }
   P->R[rd] = diff;
   return SUCCESS;
}

OP(subuOp)
{
   uint  rd          = RD(instr);
   Reg_s contents_rs = P->R[RS(instr)];
   Reg_s contents_rt = P->R[RT(instr)];

   P->R[rd] = (Reg32_s)((Reg32) contents_rs - (Reg32) contents_rt);
   return SUCCESS;
}

OP(syncOp)
{
   uint  rd          = RD(instr);
   Reg_s contents_rs = P->R[RS(instr)];
   Reg_s contents_rt = P->R[RT(instr)];

   if (SKIP_CACHES(P)) {
      CPUPrint("Sync called at %#x with caches off\n", P->PC);
   } else {
      Result ret = MemRefSync(P->myNum);
      
      if (ret == STALL) {
         if (STATS_VALUE(P->myNum, syncOpStallStart) == 0) {
            STATS_SET(P->myNum, syncOpStallStart, MipsyReadTime(P->myNum));
         }
         P->stalledInst = instr;
         /* Retry on the next cycle */
         return FAILURE;
      }
   }
   /* A sync has now completed */
   STATS_INC(P->myNum, syncOps, 1);
   STATS_ADD_INTERVAL(P->myNum, syncOpStallTime, syncOpStallStart);
   CPUVec.IncC0PerfCtr(P->myNum, 2); /* Load/prefetch/sync/CacheOp issued */
   CPUVec.IncC0PerfCtr(P->myNum, 18); /* Load/prefetch/sync/CacheOp graduated */
   return SUCCESS;
}

OP(syscallOp)
{
   uint  rd          = RD(instr);
   Reg_s contents_rs = P->R[RS(instr)];
   Reg_s contents_rt = P->R[RT(instr)];

   /* This value is added to system calls to avoid conflicts with BSD */
#define SYSVoffset 1000
   int syscallNum = P->R[REG_V0] - SYSVoffset;
   
   
#ifdef SOLO
   if (syscallNum == 1) {
      CPUWarning("Exit syscall hit\n");
      MipsyExit(BASE);
#ifdef notdef
      return FAILURE;          /* MipsyRun will return to "main" in SOLO */
#endif
   } else {
      /* Let the underlying OS handle the system call */
      if (SoloEmulateSyscall(P->myNum,syscallNum) == STALL) {
         if (P->cpuStatus != cpu_running) {
            P->stalledInst = instr;
            STATS_SET(P->myNum, stallStart, MipsyReadTime(P->myNum));
            SIM_DEBUG(('o', "Syscall stalled\n"));
         }
         return FAILURE;     /* If system call emulation is not complete, 
                                we need to restart this instruction 
                                after the stall comes back!! */
      }
   }
   SIM_DEBUG(('o', "Syscall %s by CPU %d\n",
              syscallName[syscallNum], P->myNum)); 
   STATS_INC(P->myNum, numSyscalls, 1);
   STATS_INC(P->myNum, syscallCount[syscallNum], 1);
   return SUCCESS;
#else
   STATS_INC(P->myNum, numSyscalls, 1);
   if ((syscallNum < MAX_SYSCALL) && (syscallNum >= 0)) { 
      STATS_INC(P->myNum, syscallCount[syscallNum], 1);
   }
   EXCEPTION(P, EXC_SYSCALL);
   return FAILURE;
#endif
}


OP(teqOp)
{
   uint  rd          = RD(instr);
   Reg_s contents_rs = P->R[RS(instr)];
   Reg_s contents_rt = P->R[RT(instr)];

   if (contents_rs == contents_rt) {
      EXCEPTION(P, EXC_TRAP);
      return FAILURE;
   }
   return SUCCESS;
}

OP(tgeOp)
{
   Reg_s contents_rs = P->R[RS(instr)];
   Reg_s contents_rt = P->R[RT(instr)];
   
   if (contents_rs >= contents_rt) {
      EXCEPTION(P, EXC_TRAP);
      return FAILURE;
   }
   return SUCCESS;
}

OP(tgeuOp)
{
   uint  rd          = RD(instr);
   Reg_s contents_rs = P->R[RS(instr)];
   Reg_s contents_rt = P->R[RT(instr)];
   
   if ((Reg)contents_rs >= (Reg)contents_rt) {
      EXCEPTION(P, EXC_TRAP);
      return FAILURE;
   }
   return SUCCESS;
}

OP(tltOp)
{
   Reg_s contents_rs = P->R[RS(instr)];
   Reg_s contents_rt = P->R[RT(instr)];
   
   if (contents_rs < contents_rt) {
      EXCEPTION(P, EXC_TRAP);
      return FAILURE;
   }
   return SUCCESS;
}

OP(tltuOp)
{
   uint  rd          = RD(instr);
   Reg_s contents_rs = P->R[RS(instr)];
   Reg_s contents_rt = P->R[RT(instr)];

   if ((Reg)contents_rs < (Reg)contents_rt) {
      EXCEPTION(P, EXC_TRAP);
      return FAILURE;
   } 
   return SUCCESS;
}


OP(tneOp)
{
   uint  rd          = RD(instr);
   Reg_s contents_rs = P->R[RS(instr)];
   Reg_s contents_rt = P->R[RT(instr)];

   if (contents_rs != contents_rt) {
      EXCEPTION(P, EXC_TRAP);
      return FAILURE;
   }
   return SUCCESS;
}

OP(xorOp)
{
   uint  rd          = RD(instr);
   Reg_s contents_rs = P->R[RS(instr)];
   Reg_s contents_rt = P->R[RT(instr)];

   P->R[rd] = contents_rs ^ contents_rt;
   return SUCCESS;
}

OP(unknownSpecOp) 
{
   CPUWarning("Illegal inst at %#llx on CPU %d RA %#llx\n", 
              (Reg64)P->PC, P->myNum, (Reg64)P->R[REG_RA]);
   EXCEPTION(P, EXC_II);
   return FAILURE;
}


/*****************************************************************
 * B C O N D    O P C O D E S
 *****************************************************************/

OP(bgezOp)
{
   Reg_s contents_rs = P->R[RS(instr)];
   int immediate     = IMMED(instr);
   Reg32_s offset    = SIGN_EXTEND32(18, (immediate << 2));
      
   if (!(SIGN_BIT(contents_rs))) {
      P->branchTarget = P->nPC + offset;
      P->branchStatus = BranchStatus_taken;
   }  else { 
      P->branchStatus = BranchStatus_nottaken;
   }
   CPUVec.IncC0PerfCtr(P->myNum, 6); /* Branches resolved */
   return SUCCESS;
}
      
OP(bgezlOp)
{
   Reg_s contents_rs = P->R[RS(instr)];
   int immediate     = IMMED(instr);
   Reg32_s offset    = SIGN_EXTEND32(18, (immediate << 2));
      
   if (!(SIGN_BIT(contents_rs))) {
      P->branchTarget = P->nPC + offset;
      P->branchStatus = BranchStatus_taken;
   } else {
      /* Skip instruction in delay slot */
      P->nPC = P->nPC + 4;
   }
   CPUVec.IncC0PerfCtr(P->myNum, 6); /* Branches resolved */
   return SUCCESS;
}
      
OP(bgezalOp)
{
   Reg_s contents_rs = P->R[RS(instr)];
   int immediate     = IMMED(instr);
   Reg32_s offset    = SIGN_EXTEND32(18, (immediate << 2));

   P->R[REG_RA] = P->nPC + 4;
   if (!(SIGN_BIT(contents_rs))) {
      P->branchTarget = P->nPC + offset;
      P->branchStatus = BranchStatus_taken;
#ifdef JUMP_ANNS
      if (P->PC == hideousJumpAnnotationExceptionHack[P->myNum]) {
         hideousJumpAnnotationExceptionHack[P->myNum] = 0;
      } else {
         char target[32];
         sprintf(target, "0x%llx", (uint64)P->branchTarget);
         Tcl_SetVar(TCLInterp, "JumpTarget", target, TCL_GLOBAL_ONLY);
         AnnExec(annJAL);
      }
#endif
   }   else { 
      P->branchStatus = BranchStatus_nottaken;
   }
   CPUVec.IncC0PerfCtr(P->myNum, 6); /* Branches resolved */
   return SUCCESS;
}
      
OP(bgezallOp)
{
   Reg_s contents_rs = P->R[RS(instr)];
   int immediate     = IMMED(instr);
   Reg32_s offset    = SIGN_EXTEND32(18, (immediate << 2));

   P->R[REG_RA] = P->nPC + 4;
   if (!(SIGN_BIT(contents_rs))) {
      P->branchTarget = P->nPC + offset;
      P->branchStatus = BranchStatus_taken;
#ifdef JUMP_ANNS
      if (P->PC == hideousJumpAnnotationExceptionHack[P->myNum]) {
         hideousJumpAnnotationExceptionHack[P->myNum] = 0;
      } else {
         char target[32];
         sprintf(target, "0x%llx", (uint64)P->branchTarget);
         Tcl_SetVar(TCLInterp, "JumpTarget", target, TCL_GLOBAL_ONLY);
         AnnExec(annJAL);
      }
#endif
   } else {
      /* Skip instruction in delay slot */
      P->nPC = P->nPC + 4;
   }
   CPUVec.IncC0PerfCtr(P->myNum, 6); /* Branches resolved */
   return SUCCESS;
}
      
OP(bltzOp)
{
   Reg_s contents_rs = P->R[RS(instr)];
   int immediate     = IMMED(instr);
   Reg32_s offset    = SIGN_EXTEND32(18, (immediate << 2));

   if (SIGN_BIT(contents_rs)) {
      P->branchTarget = P->nPC + offset;
      P->branchStatus = BranchStatus_taken;
   }   else { 
      P->branchStatus = BranchStatus_nottaken;
   }
   CPUVec.IncC0PerfCtr(P->myNum, 6); /* Branches resolved */
   return SUCCESS;
}
      
OP(bltzlOp)
{
   Reg_s contents_rs = P->R[RS(instr)];
   int immediate     = IMMED(instr);
   Reg32_s offset    = SIGN_EXTEND32(18, (immediate << 2));
      
   if (SIGN_BIT(contents_rs)) {
      P->branchTarget = P->nPC + offset;
      P->branchStatus = BranchStatus_taken;
   } else {
      /* Skip instruction in delay slot */
      P->nPC = P->nPC + 4;  
   }
   CPUVec.IncC0PerfCtr(P->myNum, 6); /* Branches resolved */
   return SUCCESS;
}
      
OP(bltzalOp)
{
   Reg_s contents_rs = P->R[RS(instr)];
   int immediate     = IMMED(instr);
   Reg32_s offset    = SIGN_EXTEND32(18, (immediate << 2));

   P->R[REG_RA] = P->nPC + 4;
   if ((SIGN_BIT(contents_rs))) {
      P->branchTarget = P->nPC + offset;
      P->branchStatus = BranchStatus_taken;
#ifdef JUMP_ANNS
      if (P->PC == hideousJumpAnnotationExceptionHack[P->myNum]) {
         hideousJumpAnnotationExceptionHack[P->myNum] = 0;
      } else {
         char target[32];
         sprintf(target, "0x%llx", (uint64)P->branchTarget);
         Tcl_SetVar(TCLInterp, "JumpTarget", target, TCL_GLOBAL_ONLY);
         AnnExec(annJAL);
      }
#endif
   }  else { 
      P->branchStatus = BranchStatus_nottaken;
   }
   CPUVec.IncC0PerfCtr(P->myNum, 6); /* Branches resolved */
   return SUCCESS;
}
      
OP(bltzallOp)
{
   Reg_s contents_rs = P->R[RS(instr)];
   int immediate     = IMMED(instr);
   Reg32_s offset    = SIGN_EXTEND32(18, (immediate << 2));
      
   P->R[REG_RA] = P->nPC + 4;
   if ((SIGN_BIT(contents_rs))) {
      P->branchTarget = P->nPC + offset;
      P->branchStatus = BranchStatus_taken;
#ifdef JUMP_ANNS
      if (P->PC == hideousJumpAnnotationExceptionHack[P->myNum]) {
         hideousJumpAnnotationExceptionHack[P->myNum] = 0;
      } else {
         char target[32];
         sprintf(target, "0x%llx", (uint64)P->branchTarget);
         Tcl_SetVar(TCLInterp, "JumpTarget", target, TCL_GLOBAL_ONLY);
         AnnExec(annJAL);
      }
#endif
   } else {
      /* Skip instruction in delay slot */
      P->nPC = P->nPC + 4;  
   }
   CPUVec.IncC0PerfCtr(P->myNum, 6); /* Branches resolved */
   return SUCCESS;
}

OP(teqiOp)
{
   Reg_s contents_rs = P->R[RS(instr)];
   int immediate     = IMMED(instr);
   Reg32_s offset    = SIGN_EXTEND32(18, (immediate << 2));
      
   if (contents_rs == SIGN_EXTEND(16,immediate)) {
      EXCEPTION(P, EXC_TRAP);
      return FAILURE;
   }
   return SUCCESS;
}
      
OP(tgeiOp)
{
   Reg_s contents_rs = P->R[RS(instr)];
   int immediate     = IMMED(instr);
   Reg32_s offset    = SIGN_EXTEND32(18, (immediate << 2));
      
   if (contents_rs >= SIGN_EXTEND(16,immediate)) {
      EXCEPTION(P, EXC_TRAP);
      return FAILURE;
   }
   return SUCCESS;
}
      
OP(tgeiuOp)
{
   Reg_s contents_rs = P->R[RS(instr)];
   int immediate     = IMMED(instr);
   Reg32_s offset    = SIGN_EXTEND32(18, (immediate << 2));

   if ((Reg)contents_rs >= (Reg)SIGN_EXTEND(16,immediate)) {
      EXCEPTION(P, EXC_TRAP);
      return FAILURE;
   }
   return SUCCESS;
}
      
OP(tltiOp)
{
   Reg_s contents_rs = P->R[RS(instr)];
   int immediate     = IMMED(instr);
   Reg32_s offset    = SIGN_EXTEND32(18, (immediate << 2));

   if (contents_rs < SIGN_EXTEND(16,immediate)) {
      EXCEPTION(P, EXC_TRAP);
      return FAILURE;
   }
   return SUCCESS;
}
      
OP(tltiuOp)
{
   Reg_s contents_rs = P->R[RS(instr)];
   int immediate     = IMMED(instr);
   Reg32_s offset    = SIGN_EXTEND32(18, (immediate << 2));

   if ((Reg)contents_rs < (Reg)SIGN_EXTEND(16,immediate)) {
      EXCEPTION(P, EXC_TRAP);
      return FAILURE;
   }
   return SUCCESS;
}
      
OP(tneiOp)
{
   Reg_s contents_rs = P->R[RS(instr)];
   int immediate     = IMMED(instr);
   Reg32_s offset    = SIGN_EXTEND32(18, (immediate << 2));
      

   if (contents_rs != SIGN_EXTEND(16,immediate)) {
      EXCEPTION(P, EXC_TRAP);
      return FAILURE;
   }
   return SUCCESS;
}
      
OP(unknownBcondOp)
{
   CPUWarning("Illegal inst at %#llx on CPU %d RA %#llx\n", 
              (Reg64)P->PC, P->myNum, (Reg64)P->R[REG_RA]);
   ASSERT(0);
   EXCEPTION(P, EXC_II);
   return FAILURE;
}





/*****************************************************************
 * LongMultiply64 -
 * Multiply two 64 bit numbers into the HI and LO registers.
 *****************************************************************/
static void 
LongMultiply64(CPUState *P, Reg64 v_1, Reg64 v_2)
{
   Reg64 a, b, c, d;
   Reg64 bd, ad, cb, ac;
   Reg64 mid, mid2, carry_mid = 0;

#define _HI32(_x) (((_x)>>32) & 0xffffffff)
#define _LO32(_x) ((_x) & 0xffffffff)
  
   a = _HI32(v_1);
   b = _LO32(v_1);
   c = _HI32(v_2);
   d = _LO32(v_2);
  
   bd = b * d;
   ad = a * d;
   cb = c * b;
   ac = a * c;
  
   mid = ad + cb;
   if (mid < ad || mid < cb)
      /* Arithmetic overflow or carry-out */
      carry_mid = 1;
  
   mid2 = mid + _HI32(bd);
   if (mid2 < mid || mid2 < _HI32(bd))
      /* Arithmetic overflow or carry-out */
      carry_mid += 1;
   
   P->LO =  _LO32(bd) | (_LO32(mid2)<<32);
   P->HI =  ac + (carry_mid << 32) + _HI32(mid2);
}

/*****************************************************************
 * LongMultiply32 -
 * Multiply two 32 bit numbers into the HI and LO registers.
 *****************************************************************/
static void 
LongMultiply32(CPUState *P,Reg32 v_1, Reg32 v_2)
{
   Reg64 product64;

   product64 = (Reg64)v_1 * (Reg64)v_2;   

#ifdef SIM_MIPS64
   /* Compiler bug workaround */
   P->LO = SIGN_EXTEND(32, product64);
   P->HI = SIGN_EXTEND(32, (product64>>32));
#else
   P->LO = (Reg32_s) product64;
   P->HI = (Reg32_s) (product64>>32);
#endif
}


static Result
Prefetch(CPUState *P, VA vAddr, uint hint)
{
   PA   pAddr;
   Result tRet;
   uint tlbFlavor;
   void *bdoorAddr;

   if (SKIP_CACHES(P)) {
      return SUCCESS;
   }

   /* Align to a cache line */
   vAddr &= ~(SCACHE_LINE_SIZE - 1);

   PREFETCH_ISSUE_EVENT(P->myNum,P->PC,vAddr);

   tlbFlavor = TLB_DFETCH | TLB_PREFETCH;
   tRet = TranslateVirtual(P, vAddr, &pAddr, &tlbFlavor, &bdoorAddr);

   if (tRet == SUCCESS) {
      if (MemRefPrefetch(P->myNum, vAddr, pAddr, hint)) {
         return STALL;
      } else {
         return SUCCESS;
      }
   } else {
      /* TLB miss on a prefetch - prefetch is dropped */
      if (hint & 0x1) {
         STATS_INC(P->myNum, prefStats.prefXs, 1);
         STATS_INC(P->myNum, prefStats.prefXTLBMisses, 1);
      } else {
         STATS_INC(P->myNum, prefStats.prefs, 1);
         STATS_INC(P->myNum, prefStats.prefTLBMisses, 1);
      }
#ifdef SOLO
      CPUWarning("SOLO: prefetch translation failed\n");
      ASSERT(0);
#endif
   }
   return SUCCESS;
}

/*****************************************************************
 * CacheOp
 *
 * Handle several of the cache operations found in the R4000.
 *****************************************************************/
static Result
CacheOp(CPUState *P, uint instr, VA vAddr, PA pAddr) 
{
   Result res = SUCCESS;
   uint targetCache = TARGET_CACHE(instr);
   int cacheOp = CACHE_OP(instr);
   int cpuNum = P->myNum;

   if (SKIP_CACHES(P)) {
      return SUCCESS;
   }
   switch(targetCache) {
   case CACH_PI:
      if (cacheOp == C_IINV) { 
      } else if (cacheOp == C_HINV) { 
      } else if (cacheOp == C_BARRIER) { 
         /* Cache barrier - mipsy doesn't do anything */
      } else {
         CPUWarning("Bad op 0x%x for the PICache at %#llx\n", cacheOp,
                    (Reg64)CURRENT_PC(cpuNum));
      }
      break;
      
   case CACH_PD:
      if (cacheOp == C_IWBINV) { 
      } else if (cacheOp == C_HWBINV) { 
      } else if (cacheOp == C_HWB) { 

      } else {
         CPUWarning("Bad op 0x%x for the PDCache at %#llx\n", cacheOp,
                    (Reg64)CURRENT_PC(cpuNum));
      }
      break;
      
   case CACH_SI:
       CPUWarning("Bad SI target for cache_op at %#llx\n",
                  (Reg64)CURRENT_PC(cpuNum));
      break;
      
   case CACH_SD:
      if (FlashliteWQFull(cpuNum)) {
         SET_CPU_STATUS(cpuNum, cpu_stalled);
         return  STALL;
      }
      switch (cacheOp) {
      case C_IWBINV:
         SET_CPU_STATUS(cpuNum, cpu_running);         
         res = MemRefIndexHitWBInval(cpuNum, vAddr, pAddr);
         break;
         
      case C_HWBINV:
         SET_CPU_STATUS(cpuNum, cpu_running);
         res = MemRefHitWBInval(cpuNum, vAddr, pAddr);
         break;

      case C_ILT:
      case C_IST:
         break;

      default:
         CPUWarning("BAD CACHE OP at (0x%x) %#llx\n", cacheOp,
                    (Reg64)CURRENT_PC(cpuNum));
         break;
      }
      break;

   default:
      CPUWarning("BAD TARGET CACHE at %#llx\n", (Reg64)CURRENT_PC(cpuNum));
      break;
   }
   if (res == STALL) {
      SET_CPU_STATUS(cpuNum, cpu_stalled);
   }
   return res;
}




/*****************************************************************
 * O P C O D E   T A B L E S
 *****************************************************************/ 

MipsOpcodeDesc mipsOpcodes[] = {
   { "spec", specOp, 0x00 }, 
   { "bcond", bcondOp, 0x01 },
   { "j", jOp, 0x02 },
   { "jal", jalOp, 0x03 },
   { "beq", beqOp, 0x04 },
   { "bne", bneOp, 0x05 },
   { "blez", blezOp, 0x06 },
   { "bgtz", bgtzOp, 0x07 },
   { "addi", addiOp, 0x08 },
   { "addiu", addiuOp, 0x09 },
   { "slti", sltiOp, 0x0A },
   { "sltiu", sltiuOp, 0x0B },
   { "andi", andiOp, 0x0C },
   { "ori", oriOp, 0x0D },
   { "xori", xoriOp, 0x0E },
   { "lui", luiOp, 0x0F },  
   { "cop0", cop0Op, 0x10 },
   { "cop1", cop1Op, 0x11 },
   { "cop2", cop2Op, 0x12 },
   { "cop1x", cop1xOp, 0x13 },
   { "beql", beqlOp, 0x14 },
   { "bnel", bnelOp, 0x15 },
   { "blezl", blezlOp, 0x16 },
   { "bgtzl", bgtzlOp, 0x17 },
   { "daddi", daddiOp, 0x18 },
   { "daddiu", daddiuOp, 0x19 },
   { "ldl", ldlOp, 0x1A },
   { "ldr", ldrOp, 0x1B },
   { "undef", undefOp, 0x1C },
   { "undef", undefOp, 0x1D },
   { "undef", undefOp, 0x1E },
   { "undef", undefOp, 0x1F },
   { "lb", lbOp, 0x20 },
   { "lh", lhOp, 0x21 },
   { "lwl", lwlOp, 0x22 },
   { "lw", lwOp, 0x23 },
   { "lbu", lbuOp, 0x24 },
   { "lhu", lhuOp, 0x25 },
   { "lwr", lwrOp, 0x26 },
   { "lwu", lwuOp, 0x27 },
   { "sb", sbOp, 0x28 },
   { "sh", shOp, 0x29 },
   { "swl", swlOp, 0x2A },
   { "sw", swOp, 0x2B },
   { "sdl", sdlOp, 0x2C },
   { "sdr", sdrOp, 0x2D },
   { "swr", swrOp, 0x2E },
   { "cache", cacheOp, 0x2F },
   { "ll", llOp, 0x30 }, 
   { "lwc1", lwc1Op, 0x31 },
   { "lwc2", lwc2Op, 0x32 },
   { "pref", prefOp, 0x33 },
   { "lld", lldOp, 0x34 },
   { "ldc1", ldc1Op, 0x35 },
   { "ldc2", ldc2Op, 0x36 },
   { "ld", ldOp, 0x37 },
   { "sc", scOp, 0x38 },
   { "swc1", swc1Op, 0x39 },
   { "swc2", swc2Op, 0x3A },
   { "undef", undefOp, 0x3B },
   { "scd", scdOp, 0x3C },
   { "sdc1", sdc1Op, 0x3D },
   { "sdc2", sdc2Op, 0x3E },
   { "sd", sdOp, 0x3F },
};

/* Based on the func field of spec opcode */
MipsOpcodeDesc mipsSpecOpcodes[] = {
   { "sll", sllOp, 0x00 },
   { "movc", movcOp, 0x01 },
   { "srl", srlOp, 0x02 },
   { "sra", sraOp, 0x03 },
   { "sllv", sllvOp, 0x04 },
   { "unknownSpec", unknownSpecOp, 0x05 },
   { "srlv", srlvOp, 0x06 },
   { "srav", sravOp, 0x07 },
   { "jr", jrOp, 0x08 },
   { "jalr", jalrOp, 0x09 },
   { "movz", movzOp, 0x0A },
   { "movn", movnOp, 0x0B },
   { "syscall", syscallOp, 0x0C },
   { "break", breakOp, 0x0D },
   { "spim", unknownSpecOp, 0x0E },
   { "sync", syncOp, 0x0F },
   { "mfhi", mfhiOp, 0x10 },
   { "mthi", mthiOp, 0x11 },
   { "mflo", mfloOp, 0x12 },
   { "mtlo", mtloOp, 0x13 },
   { "dsllv", dsllvOp, 0x14 },
   { "unknownSpec", unknownSpecOp, 0x15 },
   { "dsrlv", dsrlvOp, 0x16 },
   { "dsrav", dsravOp, 0x17 },
   { "mult", multOp, 0x18 },
   { "multu", multuOp, 0x19 },
   { "div", divOp, 0x1A },
   { "divu", divuOp, 0x1B },
   { "dmult", dmultOp, 0x1C },
   { "dmultu", dmultuOp, 0x1D },
   { "ddiv", ddivOp, 0x1E },
   { "ddivu", ddivuOp, 0x1F },
   { "add", addOp, 0x20 },
   { "addu", adduOp, 0x21 },
   { "sub", subOp, 0x22 },
   { "subu", subuOp, 0x23 },
   { "and", andOp, 0x24 },
   { "or", orOp, 0x25 },
   { "xor", xorOp, 0x26 },
   { "nor", norOp, 0x27 },
   { "unknownSpec", unknownSpecOp, 0x28 },
   { "unknownSpec", unknownSpecOp, 0x29 },
   { "slt", sltOp, 0x2A },
   { "sltu", sltuOp, 0x2B },
   { "dadd", daddOp, 0x2C },
   { "daddu", dadduOp, 0x2D },
   { "dsub", dsubOp, 0x2E },
   { "dsubu", dsubuOp, 0x2F },
   { "tge", tgeOp, 0x30 },
   { "tgeu", tgeuOp, 0x31 },
   { "tlt", tltOp, 0x32 },
   { "tltu", tltuOp, 0x33 },
   { "teq", teqOp, 0x34 },
   { "unknownSpec", unknownSpecOp, 0x35 },
   { "tne", tneOp, 0x36 },
   { "unknownSpec", unknownSpecOp, 0x37 },
   { "dsll", dsllOp, 0x38 },
   { "unknownSpec", unknownSpecOp, 0x39 },
   { "dsrl", dsrlOp, 0x3A },
   { "dsra", dsraOp, 0x3B },
   { "dsll32", dsll32Op, 0x3C },
   { "unknownSpec", unknownSpecOp, 0x3D },
   { "dsrl32", dsrl32Op, 0x3E },
   { "dsra32", dsra32Op, 0x3F }
};


/* Based on the rt field of bcond opcodes */
MipsOpcodeDesc mipsBcondOpcodes[] = {
   { "bltz", bltzOp, 0x00 },
   { "bgez", bgezOp, 0x01 },
   { "bltzl", bltzlOp, 0x02 },
   { "bgezl", bgezlOp, 0x03 },
   { "spimi", unknownBcondOp, 0x04 },
   { "unknownBcond", unknownBcondOp, 0x05 },
   { "unknownBcond", unknownBcondOp, 0x06 },
   { "unknownBcond", unknownBcondOp, 0x07 },
   { "tgei", tgeiOp, 0x08 },
   { "tgeiu", tgeiuOp, 0x09 },
   { "tlti", tltiOp, 0x0A },
   { "tltiu", tltiuOp, 0x0B },
   { "teqi", teqiOp, 0x0C },
   { "unknownBcond", unknownBcondOp, 0x0D },
   { "tnei", tneiOp, 0x0E },
   { "unknownBcond", unknownBcondOp, 0x0F },
   { "bltzal", bltzalOp, 0x10 },
   { "bgezal", bgezalOp, 0x11 },
   { "bltzall", bltzallOp, 0x12 },
   { "bgezall", bgezallOp, 0x13 },
   { "unknownBcond", unknownBcondOp, 0x14 },
   { "unknownBcond", unknownBcondOp, 0x15 },
   { "unknownBcond", unknownBcondOp, 0x16 },
   { "unknownBcond", unknownBcondOp, 0x17 },
   { "unknownBcond", unknownBcondOp, 0x18 },
   { "unknownBcond", unknownBcondOp, 0x19 },
   { "unknownBcond", unknownBcondOp, 0x1A },
   { "unknownBcond", unknownBcondOp, 0x1B },
   { "unknownBcond", unknownBcondOp, 0x1C },
   { "unknownBcond", unknownBcondOp, 0x1D },
   { "unknownBcond", unknownBcondOp, 0x1E },
   { "unknownBcond", unknownBcondOp, 0x1F }
};
