/*
 * Copyright (C) 1996-1998 by the Board of Trustees
 *    of Leland Stanford Junior University.
 * 
 * This file is part of the SimOS distribution. 
 * See LICENSE file for terms of the license. 
 *
 */


	/*
	 *  ms  -  The simulator component of MXS.
	 *
	 *	Simulates a variety of architectures, from a statically
	 *	scheduled R3000-like processor, to a dynamically
	 *	scheduled processor with speculative execution.
	 *
	 *	Jim Bennett
	 *	1993, 1994, 1995
	 */

#include <sys/types.h>
#include "ms.h"
#include "ms_opcodes.h"

#include "cpu_state.h"
#include "cpu_stats.h"
#include "pcache.h"
#include "cp0.h"
#include "cpu.h"

int ms_writeback (struct s_cpu_state *st);
void ms_issue (struct s_cpu_state *st);
void ms_execute (struct s_cpu_state *st);
void ms_ldst_dep (struct s_cpu_state *st);
void ms_memory (struct s_cpu_state *st);


void free_inst (struct s_cpu_state *st, int inum);


/*
 *  ms_cycle_once  -  Step the given processor forward by one
 *			cycle.
 */

void ms_cycle_once (struct s_cpu_state *st)
{
   WorkList *w;
   CPUState *P = (CPUState *)(st->mipsyPtr);

   /* Main loop.  Each time around the loop we emulate a single CPU cycle, */
   /* so this should be coded as tightly as possible.			  */

   IncStat(ST_RUN_CYCLES);
   st->work_cycle++;
   if (st->work_cycle >= TICK_COUNT) {
      /* Every 100,000 cycles, reset	*/
      /* the work clock and put out	*/
      /* a tick mark.			*/
      st->work_cycle -= TICK_COUNT;
      st->work_ticks++;
      if (st->icount == STATS_VALUE(P->myNum, numInstructions)) {
         /* No isntruction graded since last time - trouble */        
         CPUWarning("MXS: No graduation last %d cycles on CPU %d\n",
                    TICK_COUNT, P->myNum);
      }
      st->icount = STATS_VALUE(P->myNum, numInstructions);
      for (w=st->work_head; (w); w = w->next)
         w->cycle -= TICK_COUNT;
   }

   /* Perform the per cycle operations: Writeback, Fetch, Issue,	*/
   /* Execute, and Memory access.					*/

   if (printIwinMask[P->myNum]) {
      CPUPrint("<IWIN> begin %d %lld\n", P->myNum, (uint64)CPUVec.CycleCount(P->myNum));
   }
#ifdef REVERSE_PIPELINE
   /* RPB -- execute the pipeline in reverse order so that 
    * instructions can't fly through the pipeline in one cycle */
   /* Phase 4: Graduate & write back */
   ms_writeback (st);
   ms_graduate (st);
   ms_except (st);
   /* Phase 3: Execute/Load-Store */
   ms_execute (st);
   ms_ldst_dep (st);     /* Update load/store dependencies	*/
   ms_memory (st);
   /* Phase 2: Issue */
   ms_issue (st);

   /* Dump pipeline state */
   if (printIwinMask[P->myNum]) {
      int count = 0;
      int inum;
      for (inum = st->iwin_headgrad; inum >= 0; inum = st->iwin_grad[inum]) {
         dump_iwin_entry(st, inum);
         count++;
      }
      CPUPrint("<IWIN> iwin %d\n", count);
   }

   /* Phase 1: Fetch */
   ms_fetch (st);
#else
   ms_writeback (st);
   ms_fetch (st);
   ms_issue (st);
   ms_execute (st);
   ms_ldst_dep (st);     /* Update load/store dependencies	*/
   ms_memory (st);
   ms_graduate (st);
   ms_except (st);
#endif
   
   if (printIwinMask[P->myNum]) {
      CPUPrint("<IWIN> end %d %lld\n", P->myNum, (uint64)CPUVec.CycleCount(P->myNum));
   }
}

/*
 *  ms_writeback  -  WRITEBACK the results of previous operations,
 *		and unlock registers.  Handled by the work list
 *		mechanism (see the routine reg_writeback, below,
 *		for the actual operations).
 *
 *	As an indicator for the watchdog timer, ms_writeback
 *	returns the amount of work performed.
 */

int ms_writeback (struct s_cpu_state *st)
{
   int	loop_counter;
   WorkList *w;

   w=st->work_head;
   st->wbacks = 0;
   loop_counter = 0;
   while ((w) && (st->work_cycle >= w->cycle)) {
      loop_counter++;
      /* Perform work list item if	*/
      /* it is due this cycle.	*/
      (*w->f)((void *)st, w->arg2);
      st->free_tail->next = w;
      st->free_tail = w;
      w = w->next;
      st->free_tail->next = NULL;
      if ((st->work_head = w) == NULL)
         st->work_tail = NULL;
   }
   return (loop_counter);
}


/*
 *  update_excuse  -  Change the reason why a given register
 *			is not available.
 */

void update_excuse (void *st, MXSClientData ix)
{
   int	reg = (int)ix;

   STX->reg_excuse[reg] = STX->new_excuse[reg];
   STX->new_excuse[reg] = ST_NO_EXCUSE;
}


/*
 *  unstall_fetch  -  Clear a branch stall condition for a thread
 *			
 */

void unstall_fetch (void *st, MXSClientData ix)
{
   THREAD	*th;

   th = &STX->threads[(int)ix];
   th->stall_branch = 0;
   UpdateStallFetch (th);
}


/*
 *  finish_call  -  Complete call processing by unstalling the
 *		    fetch unit and writing back the link pointer
 */

void finish_call (void *st, MXSClientData ix)
{
   int r;

   INST *ip = (INST *)ix;
   int inum = ip - STX->iwin;
   int branch_node = STX->iwin_br_node[inum];
   if (branch_node >= 0) {
      BrTREE *br = &STX->branch_tree[branch_node];
      THREAD *th = &STX->threads[br->thread];
      th->stall_branch = 0;
      UpdateStallFetch (th);
      prune_branch (STX, branch_node);
   }
   reg_writeback(st, inum);
}


/*
 *  finish_branch  -  Complete branch processing by unstalling the
 *		      fetch unit and freeing the branch instruction
 */

void finish_branch (void *st, MXSClientData ix)
{
   INST *ip = (INST *)ix;
   int inum = ip - STX->iwin;
   int branch_node = STX->iwin_br_node[inum];
   if (branch_node >= 0) {
      BrTREE *br = &STX->branch_tree[branch_node];
      THREAD *th = &STX->threads[br->thread];
      th->stall_branch = 0;
      UpdateStallFetch (th);
      prune_branch (STX, branch_node);
   }
   free_inst (STX, inum);
}


/*
 *  free_spec_inst  -  Free up instruction slot belonging to a
 *			speculated instruction.
 *
 *	Just like free_inst, except that this instruction might
 *	be on someone's dependents list, so we have to fix that.
 *	Also, the register owned by this instruction needs to
 *	be freed.
 */

void free_spec_inst (struct s_cpu_state *st, int this_inst)
{
   int	r, i, j;

   /* Update the dependent lists.				*/
   INST *ip = &st->iwin[this_inst];
   for (r = 0; r < MAX_SRC_REGS; r++) {
      RegInfo *ri = &ip->src[r];
      if (ri->flags != 0) {
         int inum = st->reg_owner[ri->preg];
         for (i=0, j=0; i<st->iwin_index[r][inum]; j++) {
            st->iwin_dep[r][inum][i] = st->iwin_dep[r][inum][j];
            if (st->iwin_dep[r][inum][j] == this_inst) {
               st->iwin_index[r][inum]--;
            } else {
               i++;
            }
         }
      }
   }
   /* Take care of the target (destination) register(s).	*/
   for (r = 0; r < MAX_DST_REGS; r++) {
      RegInfo *ri = &ip->dst[r];
      if (ri->preg > 0) {
         REGSTAT *rs = &st->reg_rstat[ri->preg];
         rs->reg_status &= ~REG_IN_WIN;
         st->reg_excuse[ri->preg] = ST_NO_EXCUSE;
         st->new_excuse[ri->preg] = ST_NO_EXCUSE;
         CheckRegFree (st, rs, ri->preg);
      }
   }

   /* Then free like a regular instruction.			*/
   free_inst (st, this_inst);
}


/*
 *  reg_writeback  -  Perform WRITEBACK actions:
 *
 *	1.  Free slot of completing instruction
 *	2.  Decrement ref counts of source registers for
 *		completing instruction
 *	3.  Clear dependent flags of instructions waiting for
 *		this result
 *	4.  Update register status (no longer "in window")
 *	5.  Free load/store buffer entry of loads
 */

void reg_writeback (void *st, MXSClientData ix)
{
   REGSTAT *rs;
   int inum = (int) ix;
   int r, rd;
   INST *ip = &STX->iwin[inum];

   /* If we haven't exceeded allowed write backs, unlock the registers */
   if ((STX->wbacks + ip->ndst) > WBACK_WIDTH) {
      Add_to_worklist(STX, 1, reg_writeback, ix);
   } else {
      for (rd = 0; rd < MAX_DST_REGS; rd++) {
         RegInfo *ri = &STX->iwin[inum].dst[rd];
         if (ri->flags == 0) {
            continue;
         }
         /* Keep R0 0 */
         if (ri->preg == 0) {
            GPR(0) = 0;
            continue;
         }
         rs = &STX->reg_rstat[ri->preg];
         ASSERT(inum == STX->reg_owner[ri->preg]);

         STX->wbacks++;
         /* Turn off REG_IN_WIN status of this	*/
         /* reg and see if it is now free.	*/
         rs->reg_status &= ~REG_IN_WIN;
         STX->reg_excuse[ri->preg] = ST_NO_EXCUSE;
         STX->new_excuse[ri->preg] = ST_NO_EXCUSE;
         CheckRegFree (STX, rs, ri->preg);

         /* Turn off dependent flags of inst's	*/
         /* waiting for this result.		*/
         for (r = 0; r < MAX_SRC_REGS; r++) {
            int dwi;
            for (dwi = 0; dwi < STX->iwin_index[r][inum]; dwi++) {
               int inum2 = STX->iwin_dep[r][inum][dwi];
               STX->iwin_flags[inum2] &= ~IWIN_DEP(r);
               CheckInstAvail(STX, inum2);
            }
         }
      }
      /* Then release this instruction slot.	*/
      free_inst(STX, inum);
   }
}

/*
 *  free_inst  -  Free up instruction slot in instruction window
 *
 *	For the precise interrupts model, just set a flag to
 *	indicate that the instruction can be freed.
 */

void free_inst (struct s_cpu_state *st, int inum)
{
   int r;
   INST *ip;

#ifdef DEBUG_CHECKS
   if (!(st->iwin_flags[inum] & IWIN_BUSY)) {
      CPUError("Error: instruction slot already free\n");
   }
#ifdef PRECISE
   if (st->iwin_flags[inum] & IWIN_FREED) {
      CPUError("Error: instruction slot already free\n");
   }
#endif /* PRECISE */
/* These checks are somewhat expensive.  Comment out for now. */
#ifdef DEBUG_CHECKS_EXPENSIVE
   {
      int i, pri;
      
      for (i=0, pri=st->iwin_head_ldst; (i<10) && (pri >= 0);
           i++, pri = st->iwin_ldst[pri]) {
         if (pri == inum) {
            CPUError("Error: freed instruction still in use\n");
         }
      }
      
      for (i=0, pri=st->iwin_headpri; (i<10) && (pri >= 0);
           i++, pri = st->iwin_pri[pri]) {
         if (pri == inum) {
            CPUError("Error: freed instruction still in use\n");
         }
      }
      
      for (i=0, pri=st->iwin_next_avail; (i<10) && (pri >= 0);
           i++, pri = st->iwin_avail_list[pri]) {
         if (pri == inum) {
            CPUError("Error: freed instruction still in use\n");
         }
      }
      
      for (i=0, pri=st->iwin_nextfree; (i<10) && (pri >= 0);
           i++, pri = st->iwin_freelist[pri]) {
         if (pri == inum) {
            CPUError("Error: Error: instruction slot already free\n");
         }
      }
   }
#endif /* notdef */
#endif /* DEBUG_CHECKS */

#ifdef PRECISE
   st->iwin_flags[inum] |= IWIN_FREED;
#else
   UnthreadInst (st, inum);
#endif

   /* Decrement reference counts of source	*/
   /* registers for this instruction.		*/
   ip = &st->iwin[inum];
   for (r = 0; r < MAX_SRC_REGS; r++) {
      RegInfo *ri = &ip->src[r];
      if (ri->flags != 0) {
         REGSTAT *rs = &st->reg_rstat[ri->preg];
         rs->reg_ref--;
         CheckRegFree(st, rs, ri->preg);
      }
   }
}


#ifndef INLINE

/*
 *  UnthreadInst  -  Remove instruction from its thread and
 *	put it back on the free list.
 */

void UnthreadInst (struct s_cpu_state *st, int inum)
{
   int br_node;
   int r;

   /* Remove from the instruction chain of its thread	*/
   br_node = st->iwin_br_node[inum];
   if (br_node >= 0) {
      int	tmpi;

      BrTREE *br = &st->branch_tree[br_node];
      if ((tmpi = st->iwin_bthread[inum]) >= 0)
         st->iwin_thread[tmpi] = st->iwin_thread[inum];
      else
         br->iwin_head_th = st->iwin_thread[inum];
      
      if ((tmpi = st->iwin_thread[inum]) >= 0)
         st->iwin_bthread[tmpi] = st->iwin_bthread[inum];
      else
         br->iwin_tail_th = st->iwin_bthread[inum];
   }

   /* Put instruction back on free list	*/
   st->iwin_freelist[inum] = st->iwin_nextfree;
   st->iwin_nextfree = inum;
   st->iwin_flags[inum] = 0;
   for (r = 0; r < MAX_SRC_REGS; r++) {
      st->iwin_index[r][inum] = 0;
   }
}


/*
 *  AcquireRegMap  -  Mark physical register mapped, and track
 *			number of times it is mapped.
 */


void AcquireRegMap (REGSTAT *rs, THREAD *th, int lreg)
{
   rs->reg_nmap++;
   rs->reg_status |= REG_MAPPED;
}


/*
	 *  ReleaseRegMap  -  Decrement map count for register and mark
	 *			it unmapped when the count reaches 0.
	 */

void ReleaseRegMap (struct s_cpu_state *st, THREAD *th, int preg, int lreg)
{
   REGSTAT *rs = &st->reg_rstat[preg];
   rs->reg_nmap--;
   if (rs->reg_nmap == 0) {
      rs->reg_status &= ~REG_MAPPED;
      CheckRegFree (st, rs, preg);
   }
}


/*
 *  CheckRegFree  -  Update register state
 *
 *  For normal registers, they become free when the number of
 *  references drops to zero, and there is no other reason to
 *  stay busy (i.e. they aren't still mapped and aren't in the
 *  instruction window).
 *
 *  In the PRECISE case, when the writer of a register graduates,
 *  the prior name for that register is available.  This is
 *  indicated by setting the REG_FREED flag.
 */

void CheckRegFree (struct s_cpu_state *st, REGSTAT *rs, int preg)
{
#ifdef PRECISE
   if (rs->reg_status & REG_FREED) {
      if (rs->reg_status & REG_MAPPED) {
         rs->reg_status = REG_MAPPED;
      } else {
         rs->reg_status = 0;
         st->reg_freelist[preg] = st->reg_nextfree;
         st->reg_nextfree = preg;
      }
   }
#else
   if ((rs->reg_ref <= 0) && ((rs->reg_status & REG_IN_WIN) == 0)) {
      if ((rs->reg_status & (REG_BUSY | REG_IN_WIN | REG_MAPPED)) == REG_BUSY) {
         rs->reg_status = 0;
         st->reg_freelist[preg] = st->reg_nextfree;
         st->reg_nextfree = preg;
      }
   }
#endif /* PRECISE */
}


/*
	 *  CheckInstAvail  -  Check if the instruction is now available for
	 *			execution
	 *
	 *	Inum is the index of the instruction to check
	 */

void CheckInstAvail (struct s_cpu_state *st, int inum)
{

#ifdef DEBUG_CHECKS
   if (!(st->iwin_flags[inum] & IWIN_BUSY)) {
      CPUError("Instruction slot not in use\n");
   }
#endif

   /* Check if the instruction is unencumbered	*/
   if (st->iwin_flags[inum] &
       (IWIN_DEPS | IWIN_FLUSH | IWIN_AVAIL | IWIN_LDST_DEP | IWIN_BRDLY))
      return;

   /* Don't allow speculative writes to special registers	*/
   if ((st->iwin_flags[inum] & (IWIN_SPEC | IWIN_CTL)) ==
       (IWIN_SPEC | IWIN_CTL))
      return;
   
   /* Then mark it available and add it to the available	*/
   /* list.							*/
   st->iwin_flags[inum] |= IWIN_AVAIL;

#ifdef PRIORITIZE_AVAIL_LIST
   /* Insert the instruction into the available list in	*/
   /* priority order.					*/
   pri = st->iwin_headpri;
   avail = st->iwin_next_avail;
   last_avail = -1;
   while (pri != inum) {
#ifdef DEBUG_CHECKS
      if (pri < 0) {
         CPUError("Error - Available instruction not in priority list\n");
      }
#endif
      if (pri == avail) {
         last_avail = avail;
         avail = st->iwin_avail_list[avail];
      }
      pri = st->iwin_pri[pri];
   }

   st->iwin_avail_list[inum] = avail;
   if (last_avail < 0) {
      st->iwin_next_avail = inum;
   } else {
      st->iwin_avail_list[last_avail] = inum;
   }

#else
   /* Otherwise just add newly available instructions	*/
   /* to the tail of the available list.		*/
   if (st->iwin_next_avail >= 0) {
      st->iwin_avail_list[st->iwin_last_avail] = inum;
      st->iwin_last_avail = inum;
   } else {
      st->iwin_next_avail = inum;
      st->iwin_last_avail = inum;
   }
   st->iwin_avail_list[inum] = -1;
#endif
}

#endif /* !INLINE */


/*
 *  ms_issue  -  ISSUE instructions that are ready to go.
 *		 Chase down the priority chain, issuing the first
 *		 available instructions.
 */

void ms_issue (struct s_cpu_state *st)
{
   int	pri, prev_pri, ex_count, issues;

   /* If issue is stalled, skip it and retry execution.	*/
   /* Pending interrupts stall issue.			*/
   if (st->exception_pending) {
      AddToStat (ST_ISSUE_STALLED, ISSUE_WIDTH);
      AddToStat (ST_EXCEPT, ISSUE_WIDTH);
      return;
   }

   if (st->stall_issue) {
      AddToStat (ST_ISSUE_STALLED, ISSUE_WIDTH);
      AddToStat (st->stall_type, ISSUE_WIDTH);
#ifdef DEBUG_CHECKS
      if (st->stall_type == ST_NO_EXCUSE) {
         CPUError("No excuse stall type\n");
      }
#endif
      return;
   }

   AddToStat (ST_ISSUE_TRIED, ISSUE_WIDTH);
   for (issues=0, ex_count=0;
        (issues<ISSUE_WIDTH) && (ex_count<(2*ISSUE_WIDTH)); ) {
      /* Get next available instruction	*/
      pri = st->iwin_next_avail;
      prev_pri = -1;
      while (pri >= 0) {
         if ((st->iwin_flags[pri] & IWIN_LDST) && (!PhaseA (pri))) {
            /* If it's a memory operation, make	*/
            /* sure that there is room for it.	*/
            if (ldst_buffer_reserve (st) >= 0) break;

            /* No room, proceed to next instruction	*/
            prev_pri = pri;
            pri = st->iwin_avail_list[pri];
            continue;
         }
         break;
      }
      if (pri < 0) break;

      if (pri == st->iwin_next_avail)
         st->iwin_next_avail = st->iwin_avail_list[pri];
      else
         st->iwin_avail_list[prev_pri] = st->iwin_avail_list[pri];

      if (pri == st->iwin_last_avail)
         st->iwin_last_avail = prev_pri;

#ifdef DEBUG_CHECKS
      if (!(st->iwin_flags[pri] & IWIN_BUSY)) {
         CPUError("Inconsistent instruction window state\n");
      }
#endif
      
      /* Then issue to execution unit	*/
      st->ex[ex_count++] = pri;
      st->iwin_flags[pri] |= IWIN_ISSUED;
      
      /* and remove from priority chain,	*/
      /* except for Phase A of loads/stores.	*/
      if (st->iwin_flags[pri] & IWIN_LDST) {
         if (!PhaseA (pri)) {
            ms_ldst_dequeue (st, pri);
            ms_pri_dequeue (st, pri);
            IncStat(ST_EXECUTED);
            issues++;
            
#ifdef ONE_PHASE_LS
            /* Update store count and set flag to	*/
            /* force update of load/store status.	*/
            if (st->iwin_flags[pri] & IWIN_STORE)
               st->iwin_nstores--;
#endif
         } else {
            /* Don't count Phase A as issue	*/
            IncStat(ST_PHASE_A);
         }
      } else {
         ms_pri_dequeue (st, pri);
         IncStat(ST_EXECUTED);
         issues++;
      }
   }
   
   /* Special check for flush instructions (not issued	*/
   /* until all preceeding instructions have issued).	*/
   if (ex_count == 0) {
      /* If there is exactly one instruction in the	*/
      /* buffer, and it is waiting for the window to	*/
      /* flush, and it belongs to branch node 0 (=>	*/
      /* it is no longer speculative), and it has no	*/
      /* unresolved dependencies, then issue it.	*/
      pri = st->iwin_headpri;
      if ((pri >= 0) && (st->iwin_flags[pri] & IWIN_FLUSH) &&
          (pri == st->iwin_tailpri) &&
#ifdef PRECISE
          (pri == st->iwin_headgrad) &&
#endif
          (st->iwin_br_node[pri] == 0) &&
          (!(st->iwin_flags[pri] & IWIN_DEPS))) {
         st->ex[ex_count++] = pri;
         st->iwin_flags[pri] |= IWIN_ISSUED;
         if (st->iwin_flags[pri] & IWIN_LDST) {
            ms_ldst_dequeue (st, pri);
         }
         ms_pri_dequeue (st, pri);
         IncStat(ST_EXECUTED);
         issues++;
      }
   }

   /* Update statistics of successful and unsuccessful	*/
   /* instruction issue tries.				*/
   for (pri = st->iwin_headpri; issues < ISSUE_WIDTH;
        pri = st->iwin_pri[pri] ) {
      if (pri < 0) {
         AddToStat (ST_IWIN_EMPTY, ISSUE_WIDTH-issues);
         break;
      }
      
      if (!(st->iwin_flags[pri] & IWIN_ISSUED) || 
          (st->iwin_flags[pri] & IWIN_LDST_DEP)) {
         /* If it isn't issued because it	*/
         /* depends on an inst still in the	*/
         /* window, skip it.			*/
         int r;
         int found = 0;
         for (r = 0; r < MAX_SRC_REGS; r++) {
            if (st->iwin_flags[pri] & IWIN_DEP(r)) {
               int preg = st->iwin[pri].src[r].preg;
               if (st->reg_excuse[preg] == ST_NOT_ISSUED) {
                  found = 1;
                  break;
               }
            }
         }
         if (found) {
            continue;
         }
         issues++;
         IncStat(ST_INST_STALLED);
         if (st->iwin_flags[pri] & IWIN_LDST_DEP) {
            IncStat(ST_LDST_DEP);
         } else {
            for (r = 0; r < MAX_SRC_REGS; r++) {
               if (st->iwin_flags[pri] & IWIN_DEP(r)) {
                  int preg = st->iwin[pri].src[r].preg;
                  IncStat(ST_REG_DEP);
                  IncStat(st->reg_excuse[preg]);
#ifdef DEBUG_CHECKS
                  if (st->reg_excuse[preg] == ST_NO_EXCUSE) {
                     CPUError("Register absent without excuse\n");
                  }
#endif
                  found = 1;
                  break;
               }
            }
            if (found) {
               continue;
            } else if (st->iwin_flags[pri] & IWIN_FLUSH) {
               IncStat(ST_IWIN_FLUSH);
            } else if (st->iwin_flags[pri] & IWIN_BRDLY) {
               IncStat(ST_BR_DLY);
            } else if ((st->iwin_flags[pri] & (IWIN_SPEC | IWIN_CTL)) ==
                       (IWIN_SPEC | IWIN_CTL)) {
               IncStat(ST_SPEC_CTL);
            } else if (st->iwin_flags[pri] & IWIN_LDST) {
               if (ldst_buffer_reserve (st) >= 0) {
                  ldst_buffer_release (st);
                  IncStat(ST_NO_EXCUSE);
               } else {
                  IncStat(ST_LDST_FULL);
               }
            } else {
               IncStat(ST_NO_EXCUSE);
#ifdef DEBUG_CHECKS
               if ((ex_count<(2*ISSUE_WIDTH)) &&
                   (st->stall_type == ST_NO_EXCUSE)) {
                  CPUError("No excuse for unissued instruction\n");
               }
#endif
            }
         }
      }
   }
      
   st->ex_count = ex_count;
}

/*
 *  ms_pri_dequeue  -  Remove instruction from priority queue
 */

void ms_pri_dequeue (struct s_cpu_state *st, int pri)
{
   int inum;

   st->iwin_ninst--;
   if ((inum = st->iwin_bpri[pri]) >= 0) {
      st->iwin_pri[inum] = st->iwin_pri[pri];
   } else {
      st->iwin_headpri = st->iwin_pri[pri];
   }
   if ((inum = st->iwin_pri[pri]) >= 0) {
      st->iwin_bpri[inum] = st->iwin_bpri[pri];
   } else {
      st->iwin_tailpri = st->iwin_bpri[pri];
   }
}


/*
 *  ms_pri_enqueue  -  Add instruction to priority queue
 */

void ms_pri_enqueue (struct s_cpu_state *st, int pri)
{
   st->iwin_ninst++;
   if (st->iwin_tailpri >= 0) {
      st->iwin_pri[st->iwin_tailpri] = pri;
      st->iwin_bpri[pri] = st->iwin_tailpri;
      st->iwin_pri[pri] = -1;
      st->iwin_tailpri = pri;
   } else {
      st->iwin_headpri = pri;
      st->iwin_tailpri = pri;
      st->iwin_bpri[pri] = -1;
      st->iwin_pri[pri] = -1;
   }
}



/*
 *  ms_ldst_dequeue  -  Remove instruction from load/store queue
 */

void ms_ldst_dequeue (struct s_cpu_state *st, int inum)
{
   int tmpi;

#ifdef DEBUG_CHECKS
   st->iwin_nldst--;
   if (st->iwin_nldst < 0) {
      CPUError("Underflow in load/store queue\n");
   }
#endif
   if ((tmpi = st->iwin_bldst[inum]) >= 0) {
      st->iwin_ldst[tmpi] = st->iwin_ldst[inum];
   } else {
      st->iwin_head_ldst = st->iwin_ldst[inum];
   }
   if ((tmpi = st->iwin_ldst[inum]) >= 0) {
      st->iwin_bldst[tmpi] = st->iwin_bldst[inum];
   } else {
      st->iwin_tail_ldst = st->iwin_bldst[inum];
   }
}


/*
	 *  ms_ldst_enqueue  -  Add instruction to load/store queue
	 */

void ms_ldst_enqueue (struct s_cpu_state *st, int inum)
{
#ifdef DEBUG_CHECKS
   st->iwin_nldst++;
   if (st->iwin_nldst > IWIN_SIZE) {
      CPUError("Overflow in load/store queue\n");
   }
#endif
   st->iwin_flags[inum] |= IWIN_LDST;
   if (st->iwin_tail_ldst >= 0) {
      st->iwin_ldst[st->iwin_tail_ldst] = inum;
      st->iwin_bldst[inum] = st->iwin_tail_ldst;
      st->iwin_ldst[inum] = -1;
      st->iwin_tail_ldst = inum;
   } else {
      st->iwin_head_ldst = inum;
      st->iwin_tail_ldst = inum;
      st->iwin_bldst[inum] = -1;
      st->iwin_ldst[inum] = -1;
   }
}

/*
 *  ms_execute  -  EXECUTE the instructions in the execution unit.
 */
 
void ms_execute (struct s_cpu_state *st)
{
   INST   *ip;
   BrTREE *br;
   THREAD *th;
   int     inum;

   int ex_count = st->ex_count;
   int exi = (st->stall_issue ? st->stall_issue - 1 : 0);
   for (; exi < ex_count; exi++) {
      st->stall_issue = 0;
      st->stall_type = ST_NO_EXCUSE;
      inum = st->ex[exi];
      ip = &st->iwin[inum];
      if (st->iwin_flags[inum] & IWIN_SQUASH) {
         if (st->iwin_flags[inum] & IWIN_LDST) {
            if (PhaseA(inum)) {
               ms_ldst_dequeue (st, inum);
               ms_pri_dequeue (st, inum);
               IncStat(ST_PRUNED);
            } else {
               ldst_buffer_release (st);
            }
         }
         free_spec_inst (st, inum);
         continue;
      }

      br = &st->branch_tree[st->iwin_br_node[inum]];
      th = &st->threads[br->thread];
      if (PhaseA (inum)) {
         /* Special case for Phase A issue of	*/
         /* load/store instructions.		*/
         /* Do the address computation		*/
#ifndef ONE_PHASE_LS
         Reg base = GPR(ip->src[0].preg);
         VA vAddr = SIGN_EXTEND(16, IMMED(ip->instr)) + base;
         st->iwin_addr[inum] = vAddr;
         if (!(ip->flags & (INST_PREFETCH|INST_CACHE)) && (ip->ndst > 0)) {
            st->reg_excuse[ip->dst[0].preg] = ST_PC_DLY;
         }
         {
            PA paddr;
            uint tlbFlavor;
            Result ret;
            void *bdoorAddr = 0;

            if (ip->flags & INST_PREFETCH) {
               tlbFlavor = TLB_PREFETCH;
               ret = TranslateVirtual((CPUState *)st->mipsyPtr,
                                      (VA) st->iwin_addr[inum],
                                      &paddr, &tlbFlavor, &bdoorAddr);
               if (ret != SUCCESS)
                  st->iwin_paddr[inum] = (PA) -1;
               else
                  st->iwin_paddr[inum] = paddr;
               st->iwin_flags[inum] |= (IWIN_ADDR_VALID | IWIN_LDST_DEP);
               st->iwin_flags[inum] &= ~IWIN_AVAIL;
            } else {
               if (ip->flags & INST_STORE)
                  tlbFlavor = TLB_WRITING;
               else
                  tlbFlavor = TLB_READING;

               ret = TranslateVirtual((CPUState *)st->mipsyPtr,
                                      (VA) st->iwin_addr[inum],
                                      &paddr, &tlbFlavor, &bdoorAddr);
               if (ret != SUCCESS) {
                  st->iwin_except[inum] = GetLastException(st);
                  st->iwin_paddr[inum] = (PA) -1;
                  st->iwin_flags[inum] |=
                     (IWIN_TLBFAULT | IWIN_ADDR_VALID | IWIN_LDST_DEP);
                  st->iwin_flags[inum] &= ~IWIN_AVAIL;
               } else {
                  st->iwin_paddr[inum] = paddr;
                  st->iwin_flags[inum] |=
                     (IWIN_ADDR_VALID | IWIN_LDST_DEP);
                  st->iwin_flags[inum] &= ~IWIN_AVAIL;
                  
                  /* BDOOR accesses currently use uncached interface */
                  if (tlbFlavor & TLB_UNCACHED) {
                     st->iwin_flags[inum] |= IWIN_UNCACHED; 
                  } else if (tlbFlavor & TLB_BDOOR_FUNC) {
                     st->iwin_flags[inum] |= IWIN_BDOOR_FUNC;
                     st->iwin_paddr[inum] = (Reg) bdoorAddr;
                  } else if (tlbFlavor & TLB_BDOOR_DATA) {
                     st->iwin_flags[inum] |= IWIN_BDOOR_DATA;
                     st->iwin_paddr[inum] = (Reg) bdoorAddr;
                  }
               }
            }
         }
#endif /* !ONE_PHASE_LS */
      } else {
         uint op = MAJOR_OP(ip->instr);
         /* Need to check for ILLEGAL since an instruction with a legal opcode
          * may still be illegal -- for example, a branch in a delay slot */
         if (ip->flags & INST_ILLEGAL) {
            int inum = ip - st->iwin;
            CPUState *P = (CPUState *)st->mipsyPtr;
            RECORD_EXCEPTION(P, EXC_II, E_VEC, 0, P->CP0[C0_TLBHI],
                             P->CP0[C0_CTXT], P->CP0[C0_XCTXT]);
            st->iwin_except[inum] = GetLastException(st);
            st->iwin_flags[inum] |= IWIN_FAULT;
            CheckSquash(st, ip);
            th->stall_except = 1;
            UpdateStallFetch(th);
         } else if (ip->flags & INST_BDOOR) {
            bdoorEX(st, ip, th);
         } else {
            Result ret = mxsOpcodes[op].execute(st, ip, th);
            GPR(0) = 0;
         }

         /* If stalled, remember where.		*/
         if (st->stall_issue) {
            st->stall_issue = exi+1;
            break;
         }
         
         /* Writeback reg if single cycle inst	*/
         /* and free up instruction slot.	*/
         /* No need or want to writeback faulted inst or
          * load/store instructions . */
         if (!(st->iwin_flags[inum] & (IWIN_FAULT|IWIN_LDST))) {
            if ((ip->flags & INST_BDOOR) || (ip->opcode->latency == 1)) {
               reg_writeback(st, inum);
            } else {
               int r;
               ASSERT(ip->opcode->latency > 1);
               for (r = 0; r < ip->ndst; r++) {
                  st->reg_excuse[ip->dst[r].preg] = ST_FUNCUNIT;
               }
               Add_to_worklist(st, ip->opcode->latency, reg_writeback, inum);
            }
         }
      }
   }
   if (exi == st->ex_count)
      st->ex_count = 0;
}
