/*
 * Copyright (C) 1996-1998 by the Board of Trustees
 *    of Leland Stanford Junior University.
 * 
 * This file is part of the SimOS distribution. 
 * See LICENSE file for terms of the license. 
 *
 */

/*
 *  ms.h  -  Generic definitions for the MXS simulator component
 *
 *	Jim Bennett
 *	1993, 1994, 1995
 */

#include <sys/types.h>
#include "sim_error.h"
#include "simtypes.h"
#include "annotations.h"
#include "syslimits.h"
#include "cpu_state.h"

#include <stdio.h>
#include "ms_param.h"
#include "ms_cache.h"
#include "ms_wrklst.h"
#include "mc_opc.h"

#ifdef PRINT_INSTRUCTIONS
extern int enable_iprint;  /* Instruction print control */
extern int enable_fprint;  /* Branch processing print ctrl */
extern int enable_eprint;  /* Exception processing  "   " */
extern int iprint_frequency; /* Sampling frequency for iprint */
#endif

/*
 *  Common parameters and fixed definitions.  Typically the
 * parameters are set large enough so that they don't cause
 * any slowdown.
 */

#define MAX_WORK_ITEMS 128  /* Size of worklist array */

#define MIN_IVAR TOT_REG
#define MAX_VAR  TOT_REG

/* Bound on physical registers */
extern int mxsPhysRegs;
#define MAX_PHYS_REGS ((TOT_REG + MAX_IWIN_SIZE)*2)
#define NUM_PHYS_REGS (mxsPhysRegs)

#define PC_INC 4

/************************************************************************/
/*                                                                      */
/* Definitions for statistics gathering.                                */
/*                                                                      */
/************************************************************************/

#define ST_NO_EXCUSE   0 /* Stalled for no reason */

/* Fetch efficiency statistics */

#define ST_FETCHES_TRIED  1 /* Fetches attempted  */
#define ST_FETCHES        2 /* Successful fetches  */
#define ST_FETCH_EXPEND   3 /* Failure  - Exception pending. */
#define ST_FETCH_IW_FULL  4 /* Failures - issue window full */
#define ST_FETCH_RB_FULL  5 /* Failures - reorder buf. full */
#define ST_FETCH_THREAD   6 /* Ran out of threads  */
#define ST_FETCH_NAME     7 /* Rename failed, due to one of: */
#define ST_REG_ERROR      8 /* Reference to poisoned reg. */
#define ST_REG_BUSY       9 /* All registers in use  */
#define ST_REG_CLAIMED   10 /* Register preclaimed  */
#define ST_REG_SPEC      11 /* Special register still in use */

#define ST_FETCH_STALL       20 /* Failures - threads stalled */
#define ST_FETCH_STALLTHREAD 21 /* Fetch stall - not enough threads */
#define ST_FETCH_STALLBRANCH 22
#define ST_FETCH_STALLFPC    23
#define ST_FETCH_STALLICACHE 24
#define ST_FETCH_STALLITLB   25
#define ST_FETCH_STALLEXCEPT 26
#define ST_FETCH_STALLSYS    27
#define ST_FETCH_STALLCP0    28
#define ST_FETCH_STALLSC     29

#define ST_FETCH_STALLSYNC      30
#define ST_FETCH_STALL_OTH      31 /* Fetch stall - unidentified */
#define ST_FETCH_STALL_INV      32 /* Fetch stall - invalid addr */
#define ST_FETCH_CACHEFAILURE   33

/* Of the fetched instructions, what is their disposition? */

#define ST_EXECUTED  40 /* Instructions executed */
#define ST_EXBRANCH  41 /* Executed in branch unit */
#define ST_PRUNED    42 /* Instructions squashed */
#define ST_EXPRUNE   43 /* Executed, then squashed */

/* Issue efficiency statistics */

#define ST_ISSUE_STALLED 44 /* No issues this cycle  */
#define ST_EXCEPT        45 /* Interrupt stall cycles */
#define ST_LSQ_FULL      46 /* Load/store queue is full */
#define ST_SYSCALL       47 /* Force synchronization for syscall */
#define ST_FPC           48 /* Force sync. for FP control write */

/* Of the attempted issues, either they are executed */
/* (ST_EXECUTED, above), or they are PhaseA issues, */
/* or they fail, either due to an empty instruction */
/* window, or to a stall waiting for results.  */

#define ST_ISSUE_TRIED   49 /* Count of issues attempted */
#define ST_PHASE_A       50 /* Number of Phase A issues */
#define ST_IWIN_EMPTY    51 /* No instructions available */
#define ST_INST_STALLED  52 /* Instruction not ready */

/* Reasons why instruction not ready */

#define ST_IWIN_FLUSH  53 /* Waiting for IWIN to empty out */
#define ST_LDST_DEP    54 /* Waiting on load/store dependency */
#define ST_LDST_FULL   55 /* Couldn't reserve ldst buffer */
#define ST_BR_DLY      56 /* Need to fetch opcode in branch delay */
#define ST_SPEC_CTL    57 /* Speculative control write */
#define ST_REG_DEP     58 /* Stalled on register dependency */

/* Reasons for register dependencies */

#define ST_NOT_ISSUED  60 /* Source inst not issued yet */
#define ST_PC_DLY      61 /* Primary cache delay slot */
#define ST_IMUL        62 /* Waiting for integer multiply */
#define ST_IDIV        63 /* Waiting for integer divide */
#define ST_FPADD       64 /* Waiting for FP add operation */
#define ST_FPMUL       65 /* Waiting for FP multiply */
#define ST_FPDIV       66 /* Waiting for FP divide */
#define ST_FPSQRT      67 /* Waiting for FP square root */
#define ST_FPCMP       68 /* Waiting for FP compare */
#define ST_FPABS       69 /* Waiting for FP absolute value */
#define ST_FPNEG       70 /* Waiting for FP negate */
#define ST_FPCVT       71 /* Waiting for FP convert */
#define ST_FPMADD      72 /* Waiting for FP muladd  */
#define ST_FPRECIP     73 /* Waiting for FP reciprocal */
#define ST_FPRSQRT     74 /* Waiting for FP recip-sqrt */
#define ST_CACHE_LSQ   75 /* Waiting for load in queue */
#define ST_CACHE_LBSY  76 /* Waiting in cache line busy q */
#define ST_CACHE_BUS   77 /* Waiting for bus completion */

/* Wasted or useless instructions */

#define ST_NOP         78 /* No-op instruction  */
#define ST_BR_NOP      79 /* No-op in branch delay slot */

/* Branch prediction statistics  */

#define ST_UNCOND_BR   80 /* Unconditional branch  */
#define ST_COND_BR     81 /* Number of conditional branches */
#define ST_IND_BR      82 /* Number of indirect branches */

#define ST_CORRECT_FALLTHRU   83 /* Correctly predicted branches */
#define ST_CORRECT_W_FALLTHRU 84
#define ST_CORRECT_W_TAKEN    85
#define ST_CORRECT_TAKEN      86
#define ST_CORRECT_IND_BR     87
#define ST_CORRECT_JRET       88
#define ST_CORRECT_LIKELY     89

#define ST_INCORRECT_FALLTHRU   90 /* Incorrectly predicted branches */
#define ST_INCORRECT_W_FALLTHRU 91
#define ST_INCORRECT_TAKEN      92
#define ST_INCORRECT_W_TAKEN    93
#define ST_INCORRECT_IND_BR     94
#define ST_INCORRECT_JRET       95
#define ST_INCORRECT_LIKELY     96

/* Memory subsystem statistics  */

#define ST_NLOADS               100 /* Total number of loads */
#define ST_NSTORES              101 /* Total number of stores */
#define ST_LSOP_UNCACHED_LOAD   103
#define ST_LSOP_UNCACHED_STORE  104

/* From standalone cache subsystem */

#define ST_DREADMISS   105 /* Dirty read misses  */
#define ST_DWRITEMISS  106 /* Dirty write misses  */
#define ST_READMISS    107 /* Total read misses  */
#define ST_WRITEMISS   108 /* Total write misses  */

/* Load/store buffer statistics */

#define ST_LDSTDEP_CONISSUE 110
#define ST_LDSTDEP          111
#define ST_LDSTDEP_FAIL     112
#define ST_LDSTDEP_CONFLICT 113

#define ST_LDST_RETIRE_LOAD    114
#define ST_LDST_RETIRE_STORE   115
#define ST_LDST_INVALIDLOAD    116
#define ST_LDST_WAIT           117
#define ST_LDST_RETIRE_SQUASH  118
#define ST_LDST_REFETCH        119
#define ST_LDST_BUF_SQUASH_ACT 120
#define ST_LDST_BUF_SQUASH     121

/* Calls to ms_cache */

#define ST_CACHE               130
#define ST_CACHE_SQUASHED      131
#define ST_CACHE_CONFLICT      132
#define ST_CACHE_LOAD          133
#define ST_CACHE_STORE         134
#define ST_CACHE_HIT           135
#define ST_CACHE_MISS          136
#define ST_CACHE_FAILURE       137
#define ST_CACHE_LOADHIT       138
#define ST_CACHE_SCFAILURE     139


#define ST_VCYCS               140 /* Number of cycles in valid */
                                   /* (renameable) procedures. */
#define ST_RUN_CYCLES          141 /* Number of calls to ms_cycle_once */

/* Exception handling statistics */

#define ST_EXCEPT_INST         151
#define ST_EXCEPT_LDST         152
#define ST_EXCEPT_SQUASH       153

#define ST_EXCEPT_WAIT         154 /* Waiting with except pending */
#define ST_EXCEPT_ITLB         155
#define ST_EXCEPT_INTR         156
#define ST_EXCEPT_TAKEN        157
#define ST_EXCEPT_COHERENCY    158
#define ST_EXCEPT_SWITCH       159
#define ST_EXCEPT_SWITCH_INST  160
#define ST_EXCEPT_COHER_INST   161
#define ST_EXCEPT_COHER_LDST   162
#define ST_EXCEPT_COHER_SQUASH 163

/* Graduation statistics */

#define ST_GRAD_PHASEASTALL    166
#define ST_GRAD_ISSUESTALL     167
#define ST_GRAD_ICACHESTALL    168
#define ST_GRAD_FLUSHSTALL     169
#define ST_GRADS_TRIED         170
#define ST_GRAD_EXPEND         171
#define ST_GRAD_LDSTALL        172
#define ST_GRAD_SQUASHED       173
#define ST_GRAD_SQUASHED_LDST  174
#define ST_GRAD_INST           175
#define ST_GRAD_INST_LDST      176
#define ST_GRAD_STSTALL        177
#define ST_GRAD_EMPTYSTALL     178
#define ST_GRAD_MISCSTALL      179

#define ST_FUNCUNIT            180
#define ST_NTYPES              181 /* Number of stall types */


/*
 *  Macro for incrementing a statistics counter
 */

#define TICK_COUNT 100000  /* Number of cycles per tick mark */
#define TICK_DIGITS 5      /* Digits in tick count  */

#define CPUNUM(_st) (((CPUState *)((_st)->mipsyPtr))->myNum)
#define bp_pc_to_index(_pc) (((_pc)>>2)&(BP_TABLE_SIZE-1))

#define IncStat(sttype)  st->stats[(sttype)]++;
#define AddToStat(sttype,cnt) st->stats[(sttype)] += (cnt);    

/************************************************************************/
/*                                                                      */
/*  Simulated machine state                                             */
/*                                                                      */
/************************************************************************/

typedef struct {      /* Branch tree structure  */
   int thread;        /* Thread number of this node */
   int thread_st;     /* Thread status   */
   int lchild;        /* Indices of nodes children */
   int rchild;
   /* Condition holds target PC for unconditional indirect jumps */
   Reg condition;     /* How to resolve branch  */
   int resolution;
   int indirect;      /* If indirect branch  */
   int jret;          /* If subroutine return  */
   int call;          /* If a call   */
   int uncond;        /* If immediate unconditional */
   int restore;       /* If branch stack needs restore */
   int iwin_head_th;  /* Chain of instructions for */
   int iwin_tail_th;  /* this node.   */
} BrTREE;


typedef struct s_thread {      /* Per thread of control state */
   Reg pc;                     /* Program counter (per thread) */
   int thread_st;              /* Thread status   */
   int stall_fetch;            /* If instruction fetch stalled */
   int branch_dly;             /* In branch delay slot  */
   int branch_likely;          /* Ditto for branch likely inst's  */
   int stall_fpc;              /* If stall due to FP control write  */
   int stall_branch;           /* Stalled waiting for branch */
   int stall_thread;           /* Stalled waiting for branch - */
                               /* no free threads   */
   int stall_icache;           /* Problem fetching inst. from cache */
   int icache_stall_reason;    /* Cause of icache stall   */
   int stall_except;           /* Stall due to exception  */
   int stall_itlbmiss;         /* Stall due to ITLB miss  */
   int stall_sys;              /* Stall due to system call  */
   int stall_sc;               /* Stall due to an SC  */
   int stall_cp0;              /* Stall due to an CP0 inst  */
   int stall_sync;             /* Stall due to a SYNC inst  */
   Reg returnpc;               /* Return address PC  */
   int branch_sp;              /* Branch stack pointer  */
   Reg old_prediction;         /* Previous value of branch stack  */
   int branch_inum;            /* Index of branch inst.  */
   int branch_node;            /* Index of branch tree node */
   Reg branch_pc;              /* Branch target address  */
   Reg branch_likely_pc;       /* Ditto for branch likely inst's  */
   int except;                 /* Type of exception  */
   struct s_thread *active_thread; /* Next active thread */
   int regnames [MAX_VAR];     /* Register names, indexed by */
                               /* logical register (from inst) */
} THREAD;

#define TH_ACTIVE 0x01   /* If thread active */
#define TH_SPEC  0x02   /* If thread speculative */
#define TH_BRANCH 0x04   /* If branch graduated */

#define UpdateStallFetch(th)                                    \
(th)->stall_fetch = (th)->stall_fpc || (th)->stall_branch ||    \
(th)->stall_icache || (th)->stall_except ||                     \
(th)->stall_itlbmiss || (th)->stall_sys ||                      \
(th)->stall_sync || (th)->stall_sc || (th)->stall_cp0

#define REGNAME_MASK 0x0ffff
#define REGNAME_CLAIM 0x10000

typedef struct {    /* Register state needed for fast access */
   int reg_status;  /* Status flags    */
   int reg_ref;     /* Reference count for register  */
   int reg_nmap;    /* Map count for register   */
} REGSTAT;

#define REG_BUSY 0x0001   /* Physical registers  */
#define REG_IN_WIN 0x0002
#define REG_MAPPED 0x0004

#define REG_ERROR 0x0010  /* Indicates a fault  */
#define REG_FREED 0x0040  /* Reg should be freed  */


#define ACT_NULL           0 /* Do nothing on callback */
#define ACT_UNSTALL_ICACHE 1 /* Unstall icache on callback */
#define ACT_DCACHE_MISS    2 /* A dcache miss finishes */
#define ACT_MHT_EMPTY      3 /* The MHT is empty */

struct s_cpu_state {                    /* Holds all processor state.  */
   Reg     gpr [MAX_PHYS_REGS];
   FPReg   fpr [MAX_PHYS_REGS];
   
   THREAD threads [MAX_THREAD_WIDTH];   /* Threads of execution  */
   THREAD *active_thread;               /* First active thread  */
   int free_thread;                     /* Free list for threads */
   int nthreads;                        /* Number of threads  */
   int nactive;                         /* Number of active threads */
   
   BrTREE branch_tree [2*MAX_THREAD_WIDTH]; /* Active branch tree */
   int free_branch_node;                /* Free list for branch tree */
   
   int round_mode;                      /* FP rounding mode  */
   int stall_issue;                     /* If instruction issue stalled */
   int stall_type;                      /* Reason for inst issue stall */
   int wbacks;                          /* Count of write backs this cycle */
   
   /* Physical register status */
   int reg_owner [MAX_PHYS_REGS];       /* Owner of this register */
   REGSTAT reg_rstat [MAX_PHYS_REGS];   /* Status of register  */
   int reg_excuse [MAX_PHYS_REGS];      /* Reason why reg not yet available */
   int new_excuse [MAX_PHYS_REGS];      /* Yet another reason why reg ... */
   int reg_freelist [MAX_PHYS_REGS];
   int reg_nextfree;
   
   /* Instruction window, handles dependencies between */
   /* instructions, and issues instructions out-of-order. */
   /* Includes instructions in execution also.  */
   
   int iwin_ninst;                      /* Number of instructions in */
                                        /* window and not yet executing */
   INST iwin [MAX_IWIN_SIZE];           /* Instruction window  */
   VA iwin_addr [MAX_IWIN_SIZE];        /* Address field for load/stores */
   PA iwin_paddr [MAX_IWIN_SIZE];       /* paddr field for load/stores */
   int iwin_flags [MAX_IWIN_SIZE];
   int iwin_freelist [MAX_IWIN_SIZE];
   /* Instructions dependent on */
   /* this one.   */
   int iwin_dep [MAX_SRC_REGS][MAX_IWIN_SIZE][MAX_IWIN_SIZE];
   int iwin_index  [MAX_SRC_REGS][MAX_IWIN_SIZE];
   
   int iwin_pri [MAX_IWIN_SIZE];        /* Priority chain for insts */
   int iwin_bpri [MAX_IWIN_SIZE];       /* Back pointers in chain */
   
#ifdef PRECISE
   THREAD grad;                         /* Thread state at graduation point */
   
   int iwin_headgrad;                   /* Head and tail of graduation queue */
   int iwin_tailgrad;
   int iwin_grad [MAX_IWIN_SIZE];       /* Graduation queue */
   int iwin_bgrd [MAX_IWIN_SIZE];
#endif
   int iwin_avail_list [MAX_IWIN_SIZE]; /* Instructions ready */
   /* to be executed. */
   Reg iwin_pc [MAX_IWIN_SIZE];
   int iwin_nextfree;
   int iwin_headpri;                    /* Prioritized list of instructions */
   int iwin_tailpri;
   int iwin_next_avail;                 /* Head and tail of available list */
   int iwin_last_avail;
   
   /* Special support for speculative execution.  */
   
   int iwin_br_node [MAX_IWIN_SIZE];    /* Branch tree node */
   int iwin_thread [MAX_IWIN_SIZE];     /* Thread inst chain */
   Reg iwin_branch_pc [MAX_IWIN_SIZE];  /* PC of branch target */
   int iwin_bthread [MAX_IWIN_SIZE];
   
   /* Support for memory callbacks into MXS  */
   
   int ms_action;                       /* Action to take on callback */
   
   /* Special support for load/store ordering. */
   
   int iwin_ldst [MAX_IWIN_SIZE];       /* Load/store chain  */
   int iwin_bldst [MAX_IWIN_SIZE];
   int iwin_head_ldst;
   int iwin_tail_ldst;
   int iwin_nldst;                      /* Number of items in ldst chain */
#ifdef ONE_PHASE_LS
   int iwin_nstores;
#endif
   
   /* Support for exceptions   */
   
   int iwin_except [MAX_IWIN_SIZE];
   int exception_pending;
   int except;
   
   /* Branch prediction counters and addresses */
   
   int bp_bits [BP_TABLE_SIZE];
   Reg bp_targets [BP_TABLE_SIZE];
   Reg branch_stack[BP_RETURN_STACK];
   int branch_sp;
   
   /* Execution unit, pointers to executing instructions */
   
   int ex [2*MAX_ISSUE_WIDTH];
   int ex_count;
   
   /* Worklist mechanism, handles instruction completion */
   
   int work_cycle;                      /* Total cycles   */
   int work_ticks;                      /* Total ticks (TICK_COUNT cycles) */
   
   WorkList *work_head;                 /* Prioritized queue of time */
   WorkList *work_tail;                 /* delayed work items (e.g. */
   WorkList *free_head;                 /* FP operation write back, */
   WorkList *free_tail;                 /* or memory load completion). */
   WorkList worklist [MAX_WORK_ITEMS];
   
   /* Load/store buffer, for the memory subsystem  */
   
   struct s_ldst_buffer ldst_buffer[MAX_LDST_BUFFER_SIZE];
   struct s_lsq lsq [MAX_LDST_BUFFER_SIZE];
   struct s_ldst_buffer *ldst_nextAvail;
   struct s_ldst_buffer *ldst_nextReserved;
   struct s_ldst_buffer *ldst_head;
   struct s_ldst_buffer *ldst_tail;
   struct s_ldst_buffer *inum2ldst[MAX_IWIN_SIZE];
   
   /* Cycle by cycle statistics    */
   uint64 stats[ST_NTYPES];
   
   /* Sample stats */
   struct {
      uint64 samples;
      uint64 nthreads_hist[MAX_THREAD_WIDTH+1];
      uint64 reg_hist[MAX_PHYS_REGS+1];
      struct {
         uint64 inst;                   /* Number of instruction in window */
         uint64 specInst;               /* Number of speculative instructions */
         uint64 ldstInst;               /* Number of load/store instructions */
         uint64 ldstdepInst;            /* Nubmer of instructions with LDST_DEP */
         uint64 regdepInst;             /* Number of inst with REG_DEP* */
         uint64 squashInst;             /* Number of squashed instructions */
      } iwin_hist[MAX_IWIN_SIZE+1];
      struct {
         uint64 inst;                   /* Number of inst in ldst buffer */
         uint64 doneInst;               /* Number of finished inst in ldst buffer */
         uint64 pendInst;               /* Number of pending inst */
         uint64 conflictInst;           /* Number of conflicted inst */
         uint64 failedInst;             /* Number of failed inst. */
         uint64 stallInst;              /* Number of stalled inst. */
      } ldst_hist[MAX_LDST_BUFFER_SIZE+1]; 
   } sample_stats;
   
   int icount;
   void *mipsyPtr;
};

extern struct s_cpu_state cpu_state;

/*
 *  Status bit definitions for the various components of the model
 */

#define IWIN_BUSY       0x0001   /* Inst window: slot busy */
#define IWIN_BDOOR      0x0002
#define IWIN_BDOOR_FUNC 0x0004
#define IWIN_BDOOR_DATA 0x0008
#define IWIN_FLUSH      0x0010   /* Force instruction */
                                 /* window to empty out. */

#define IWIN_AVAIL      0x0020   /* Ready to be executed */
#define IWIN_LDST       0x0040   /* Is a load or store */
#define IWIN_ISSUED     0x0080   /* Inst has been issued */
#define IWIN_BRDLY      0x0100   /* Waiting for branch */
/* delay to be fetched */

#define IWIN_SPEC       0x0200   /* Speculative inst. */
#define IWIN_SQUASH     0x0400   /* Squashed inst. */
#define IWIN_STORE      0x0800   /* Is a store inst. */
#define IWIN_CTL        0x1000   /* Dest is a special reg */

#define IWIN_FAULT      0x2000   /* Instruction faulted */
#define IWIN_LDST_DEP   0x4000   /* Load/store dependency */
#define IWIN_ADDR_VALID 0x8000   /* Address field is valid */

/* For precise interrupts: */
#define IWIN_FREED      0x10000   /* Instruction can be freed */
#define IWIN_DEFINE     0x20000   /* Instruction defines a reg */

#define IWIN_BRANCH     0x40000   /* Branch instruction */
#define IWIN_UNCACHED   0x80000   /* Is an uncached load/store */
#define IWIN_TLBFAULT   0x100000  /* Op suffered a TLB fault */
#define IWIN_LDSTBUF    0x200000  /* In ldst_buffer */
#define IWIN_TAKENBR    0x800000  /* If branch, was taken. */

#define IWIN_DEP(I)     (0x01000000 << (I))
#if MAX_SRC_REGS > 4
#error IWIN_DEPS is too small
#else
#define IWIN_DEPS       (0x0f000000)
#endif

#define BP_TAKEN   0x04   /* Branch prediction bits */
#define BP_BOTH    0x02   /* Follow both threads? */
#define BP_MAX_VAL 0x07   /* Clamp to this value */

#define PredictTaken(pbits, likely) \
(((pbits) & BP_TAKEN) || ((likely) && ((pbits) >= BP_BOTH)))

#define PRUNE_RIGHT 0x00   /* Which branch to prune? */
#define PRUNE_LEFT  0x01

   /* Floating point condition code bit. */
   
   /* In MIPS4, there are 8 FP condition codes:
    * store the code number in ip->r3
    * force it to be negative to avoid register renaming code
    */
   
#define CC2REG(CC) (-1 - CC)
#define REG2CC(R)  (-1 - R)
#define CONDBIT(CC) (M_csr_fcc(CC))
   
   /*
    *  PhaseA  -  Determine if the instruction is a load/store which
    *  hasn't yet completed Phase A execution.  In a one
    *  phase load operation, Phase A is always false.
    */
   
#ifdef ONE_PHASE_LS

#define PhaseA(inum) 0

#else

#define PhaseA(inum) \
   ((st->iwin_flags [(inum)] & (IWIN_LDST|IWIN_ADDR_VALID)) == IWIN_LDST)

#endif

#define GPR(_r) (((struct s_cpu_state*)st)->gpr[_r])

extern void unstall_fetch (void *st, MXSClientData ix);
extern void finish_call (void *st, MXSClientData ix);
extern void finish_branch (void *st, MXSClientData ix);
extern void update_excuse (void *st, MXSClientData ix);
extern void reg_writeback (void *st, MXSClientData ix);
extern void free_spec_inst (struct s_cpu_state *st, int inum);

extern void ms_graduate (struct s_cpu_state *st);
extern void ms_grad_dequeue (struct s_cpu_state *st, int inum);
extern void ms_grad_enqueue (struct s_cpu_state *st, int inum);
extern void ms_except (struct s_cpu_state *st);

extern void ms_pri_dequeue (struct s_cpu_state *st, int pri);
extern void ms_pri_enqueue (struct s_cpu_state *st, int pri);
extern void ms_ldst_dequeue (struct s_cpu_state *st, int inum);
extern void ms_ldst_enqueue (struct s_cpu_state *st, int inum);

extern void ms_st_init (struct s_cpu_state *st);
extern void ms_iwin_init (struct s_cpu_state *st);
extern void ms_reset_stats(struct s_cpu_state *st);

extern void ms_branch (struct s_cpu_state *st, THREAD *th, int likely);
extern void prune_branch (struct s_cpu_state *st, int branch_node);
extern void speculate_reparent (struct s_cpu_state *st,
                                int branch_node, int id_node);
extern void speculate_flush (struct s_cpu_state *st, int branch_node);
extern void CheckSquash (struct s_cpu_state *st, INST *ip);

extern void ms_fetch(struct s_cpu_state *st);
extern void free_inst(struct s_cpu_state *st, int inum);

extern int  ms_decode(struct s_cpu_state *st, THREAD *th, int inum, int bdoor);
extern int  ms_rename(struct s_cpu_state *st, INST *ip, THREAD *th, int inum);

/* The load/store buffer */
extern void ms_lsq (struct s_cpu_state *st, VA addr,
                    RegInfo *reg, RegInfo *reg2, int lstype, int inum);

extern void ms_cycle_once (struct s_cpu_state *st);
extern void DoMxsIntervention(struct s_cpu_state *st, PA paddr, int size, 
                              int writebackonly);
extern int  CheckIntervention (struct s_cpu_state *st,
                              void *mtag, PA paddr, int size);
extern void DoMxsAction(struct s_cpu_state *st, void *mtag, int ms_action, int success);
extern int  GetMxsAction(struct s_cpu_state *st);

extern void MxsSampleStats(struct s_cpu_state *st);
extern int  MxsApproveImiss(struct s_cpu_state *st, PA paddr);

extern void ms_cycle_loop (struct s_cpu_state *st);
extern void init_cache (struct s_cpu_state *st);

extern void compile_inst (uint inst, VA addr, INST *ip,
                          int *squashed, int *branch_shadow);

extern void set_rm (int mode);
extern void print_stats (struct s_cpu_state *st);
extern void dump_stats (struct s_cpu_state *st);
extern void MxsPrintStatus (struct s_cpu_state *st);

extern void dump_iwin_entry (struct s_cpu_state *st, int inum);
extern char printIwinMask[SIM_MAXCPUS];

extern void ms_bus_init (struct s_cpu_state *st);
extern void ms_debus (void *st, void *unused);
extern void flush_cache (struct s_cpu_state *st,
                         int addr, int cnt, int empty);
extern int  ms_bus_busy (void);

extern void ldst_init (struct s_cpu_state *st);
extern int  ldst_buffer_reserve (struct s_cpu_state *st);
extern void ldst_buffer_release (struct s_cpu_state *st);
extern int  ldst_buffer_retire (struct s_cpu_state *st, int inum);
extern void ldst_buffer_squash (struct s_cpu_state *st, int inum);

extern void record_ldst_stall (struct s_cpu_state *st, int inum, int stall);
extern void record_icache_stall (struct s_cpu_state *st, int pc, int stall);
extern void record_pipeline_stall (struct s_cpu_state *st, int pc, int stall);
extern void record_grad_inst(struct s_cpu_state *st, int pc, int stall);

/* STATRECORD ROUTINES */
#define MS_GRAD_EVENT(_cpu, _pc, _stall) \
   ms_attribute_grad(_cpu, _pc, _stall);

#define MS_STALL_EVENT(_cpu, _vAddr, _pc, _stall, _type) \
   ms_attribute_stall(_cpu, _vAddr, _pc, _stall, _type);

#define MS_BRANCH_EVENT(_cpu, _pc, _type) \
   ms_attribute_branch(_cpu, _pc, _type);

extern void ms_events_init(void);
extern void ms_attribute_grad(int cpuNum, int pc, int stall);
extern void ms_attribute_branch(int cpuNum, int pc, int type);
extern void ms_attribute_stall(int cpuNum, int vAddr, int pc, 
                               int stall, int stallType);


extern void UnthreadInst (struct s_cpu_state *st, int inum);
extern void AcquireRegMap (REGSTAT *rs, THREAD *th, int lreg);
extern void ReleaseRegMap (struct s_cpu_state *st, THREAD *th,
                           int preg, int lreg);
extern void CheckRegFree(struct s_cpu_state *st, REGSTAT *rs, int preg);
extern void CheckInstAvail (struct s_cpu_state *st, int inum);

extern void Add_to_worklist(struct s_cpu_state *st, int inc,
                            void (*func)(void *st, MXSClientData a2),
                            MXSClientData argument2);

