/*
 * Copyright (C) 1996-1998 by the Board of Trustees
 *    of Leland Stanford Junior University.
 * 
 * This file is part of the SimOS distribution. 
 * See LICENSE file for terms of the license. 
 *
 */


	/*
	 *  ms_fetch.c  -  Handle instruction fetching for the MXS simulator
	 *
	 *	Jim Bennett
	 *	1994, 1995
	 */

#include <sys/types.h>
#include <stdlib.h>
#include "ms.h"
#include "common_cache.h"
#include "cpu_state.h"
#include "mipsy.h"
#include "annotations.h"
#include "cp0.h"
#include "pcache.h"

/*
 *  ms_fetch  -  Fetch instructions from the text segment and add
 *		 them to the instruction window
 */

void 
ms_fetch (struct s_cpu_state *st)
{
   INST	*ip;
   THREAD *th;
   BrTREE *br;
   int	this_inst, fetches;
   int	nfetch;			/* Number of successful fetches	*/
   int  r;
   uint	inst;
   int  iwinDump = printIwinMask[((CPUState *)(st->mipsyPtr))->myNum];

   AddToStat (ST_FETCHES_TRIED, FETCH_WIDTH);
   if (st->exception_pending) {	
      /* Pending exception stalls all threads	*/
      AddToStat (ST_FETCH_EXPEND, FETCH_WIDTH);
      if (iwinDump) {
         CPUPrint("<IWIN> fetch 0 EXCEPTION\n");
      }
      return;
   }
   nfetch = 0;
#if MAX_ACT_THREADS == 1
   for (fetches=0;fetches<FETCH_WIDTH;fetches++)
#else
   for (th = st->active_thread, fetches=0;
        fetches<FETCH_WIDTH;
        fetches++, th = th->active_thread)
#endif 
      {
         /* If instruction window is full,	*/
         /* then proceed to next unit.		*/
         if (st->iwin_ninst >= IWIN_SIZE) {
            AddToStat (ST_FETCH_IW_FULL, FETCH_WIDTH-fetches);
            if (iwinDump) {
               CPUPrint("<IWIN> fetch %d IWIN_FULL\n", fetches);
            }
            break;
         }
         this_inst = st->iwin_nextfree;
         if (this_inst < 0) {
            AddToStat (ST_FETCH_RB_FULL, FETCH_WIDTH-fetches);
            if (iwinDump) {
               CPUPrint("<IWIN> fetch %d RB_FULL\n", fetches);
            }
            break;
         }

         /* If there are no active unstalled threads,	*/
         /* then proceed to next unit.			*/
#if MAX_ACT_THREADS == 1
         th = st->active_thread;
         if (!th) {
            AddToStat (ST_FETCH_STALL_OTH,FETCH_WIDTH-fetches);
            AddToStat (ST_FETCH_STALL,FETCH_WIDTH-fetches);
            if (iwinDump) {
               CPUPrint("<IWIN> fetch %d NO_THREAD\n", fetches);
            }
            break;
         }
         if (th->stall_fetch) {
            if (th->stall_branch) {
               if (th->stall_thread) {
                  AddToStat (ST_FETCH_STALLTHREAD,FETCH_WIDTH-fetches);
                  if (iwinDump) {
                     CPUPrint("<IWIN> fetch %d THREAD\n", fetches);
                  }
               } else {
                  AddToStat (ST_FETCH_STALLBRANCH,FETCH_WIDTH-fetches);
                  if (iwinDump) {
                     CPUPrint("<IWIN> fetch %d BRANCH\n", fetches);
                  }
               }
            } else if (th->stall_fpc) {
               AddToStat (ST_FETCH_STALLFPC,FETCH_WIDTH-fetches);
               if (iwinDump) {
                  CPUPrint("<IWIN> fetch %d FPCSR\n", fetches);
               }
            } else if (th->stall_icache) {
               AddToStat (ST_FETCH_STALLICACHE,FETCH_WIDTH-fetches);
               if (iwinDump) {
                  CPUPrint("<IWIN> fetch %d ICACHE\n", fetches);
               }
            } else if (th->stall_itlbmiss) {
               AddToStat (ST_FETCH_STALLITLB,FETCH_WIDTH-fetches);
               if (iwinDump) {
                  CPUPrint("<IWIN> fetch %d ITLBMISS\n", fetches - fetches);
               }
            } else if (th->stall_except) {
               AddToStat (ST_FETCH_STALLEXCEPT,FETCH_WIDTH-fetches);
               if (iwinDump) {
                  CPUPrint("<IWIN> fetch %d EXCEPTION\n", fetches);
               }
            } else if (th->stall_sys) {
               AddToStat (ST_FETCH_STALLSYS,FETCH_WIDTH-fetches);
               if (iwinDump) {
                  CPUPrint("<IWIN> fetch %d SYS_OP\n", fetches);
               }
            } else if (th->stall_cp0) {
               AddToStat (ST_FETCH_STALLCP0,FETCH_WIDTH-fetches);
               if (iwinDump) {
                  CPUPrint("<IWIN> fetch %d CP0_OP\n", fetches);
               }
            } else if (th->stall_sc) {
               AddToStat (ST_FETCH_STALLSC,FETCH_WIDTH-fetches);
               if (iwinDump) {
                  CPUPrint("<IWIN> fetch %d SC_OP\n", fetches);
               }
            } else if (th->stall_sync) {
               AddToStat (ST_FETCH_STALLSYNC,FETCH_WIDTH-fetches);
               if (iwinDump) {
                  CPUPrint("<IWIN> fetch %d SYNC_OP\n", fetches);
               }
            } else {
               AddToStat (ST_FETCH_STALL_OTH,FETCH_WIDTH-fetches);
               if (iwinDump) {
                  CPUPrint("<IWIN> fetch %d OTHER\n", fetches);
               }
            }
            AddToStat (ST_FETCH_STALL,FETCH_WIDTH-fetches);
            break;
         }
#else
         while ((th) && th->stall_fetch)
            th = th->active_thread;
         if (th == NULL) {
            th = st->active_thread;
            while ((th) && th->stall_fetch)
               th = th->active_thread;
            if (th == NULL) {
               AddToStat (ST_FETCH_STALL_OTH, FETCH_WIDTH-fetches);
               AddToStat (ST_FETCH_STALL, FETCH_WIDTH-fetches);
               break;
            }
         }
#endif

         ip = &st->iwin [this_inst];

         /* Fetch the instruction to this thread	*/
         {
            PA	paddr;
            int	ret, dum1, dum2, bdoor;
            uint    tlbFlavor = TLB_IFETCH|TLB_READING;
            CPUState *P = (CPUState *) (st->mipsyPtr);
         
            if ((th->pc & 3) == 0) { 
               void *bdoorAddr = 0;
               ret = TranslateVirtual(P, th->pc, &paddr, &tlbFlavor, &bdoorAddr);
            } else {
               RECORD_EXCEPTION(P, EXC_RADE, E_VEC,th->pc, 
                                P->CP0[C0_TLBHI], P->CP0[C0_CTXT],
                                P->CP0[C0_XCTXT]);  
               ret = FAILURE;
            }
            if (ret != SUCCESS) {
               th->except = GetLastException(st);
               th->stall_itlbmiss = 1;
               UpdateStallFetch (th);
               IncStat (ST_FETCH_STALLITLB);
               IncStat (ST_FETCH_STALL);
               continue;
            }
            bdoor = (tlbFlavor & TLB_BDOOR);
            if (!bdoor) {
               st->ms_action = ACT_UNSTALL_ICACHE;
               P->PC = th->pc;
               ret = ReadICache(P->myNum, th->pc, paddr, &inst);
               if (ret != SUCCESS) {
                  if (ret == FAILURE) {
                     /* Cache wouldn't accept req - stall for a cycle. */
                     IncStat (ST_FETCH_CACHEFAILURE);
                     IncStat (ST_FETCH_STALLICACHE);
                     IncStat (ST_FETCH_STALL);
                     continue;
                  }
                  th->stall_icache = 1;
                  th->icache_stall_reason = MxsClassifyMiss(P->myNum, th->pc, paddr, TRUE);
                  UpdateStallFetch (th);
                  IncStat (ST_FETCH_STALLICACHE);
                  IncStat (ST_FETCH_STALL);
                  continue;
               }
            }
            ip->instr = inst;
            ms_decode(st, th, this_inst, bdoor);
         }

         st->iwin_br_node [this_inst] = th->branch_node;

         /* Check if there's a thread available for a branch */
         if ((ip->flags & INST_LIKELY) || ((th->branch_dly == 1) && !th->branch_likely)) {
            if (st->nthreads >= THREAD_WIDTH) {
               AddToStat (ST_FETCH_STALLTHREAD,FETCH_WIDTH-fetches);
               if (iwinDump) {
                  CPUPrint("<IWIN> fetch %d THREAD\n", fetches);
               }
               break;
            }
         }

         /* Convert illegal instruction sequence to	*/
         /* illegal instruction, for exception handling	*/
         if (th->branch_dly && (ip->flags & (INST_BRANCH | INST_CALL))) {
            ip->flags = INST_ILLEGAL;
         }

         /* Rename logical registers to use real	*/
         /* (physical) registers.			*/
         if (ms_rename (st, ip, th, this_inst) < 0) {
            IncStat (ST_FETCH_NAME);
            continue;
         }

         /* Rename succeeded, so it is OK to update the	*/
         /* machine state now.				*/
         ms_grad_enqueue(st, this_inst);
         st->iwin_pc[this_inst] = th->pc;
         st->iwin_flags[this_inst] |= IWIN_BUSY;

         if (th->thread_st & TH_SPEC)
            st->iwin_flags [this_inst] |= IWIN_SPEC;
         st->iwin_nextfree = st->iwin_freelist [this_inst];

         th->pc = th->pc + PC_INC;

         /* Add instruction to list for this thread	*/
         br = &st->branch_tree [th->branch_node];
         if (br->iwin_tail_th >= 0) {
            st->iwin_thread [br->iwin_tail_th] = this_inst;
            st->iwin_bthread [this_inst] = br->iwin_tail_th;
            st->iwin_thread [this_inst] = -1;
            br->iwin_tail_th = this_inst;
         } else {
            br->iwin_head_th = this_inst;
            br->iwin_tail_th = this_inst;
            st->iwin_bthread [this_inst] = -1;
            st->iwin_thread [this_inst] = -1;
         }

         /* Add load/store instructions to load/store chain */
         if (ip->flags & INST_LDST) {
            int store_inst = (ip->dst[0].flags & REG_STORE);
            if (store_inst) st->iwin_flags [this_inst] |= IWIN_STORE;
            ms_ldst_enqueue (st, this_inst);
            
#ifdef ONE_PHASE_LS
            /* Loads and stores must wait if there is a	*/
            /* preceeding store.				*/
            if (st->iwin_nstores > 0)
               st->iwin_flags [this_inst] |= IWIN_LDST_DEP;
            
            /* A store must wait if there are any		*/
            /* preceeding loads or stores.			*/
            if (store_inst) {
               if (this_inst != st->iwin_head_ldst)
                  st->iwin_flags [this_inst] |= IWIN_LDST_DEP;
               st->iwin_nstores++;
            }
#endif
         }
         /* Add this instruction to the tail	*/
         /* of the priority chain in the window	*/
         nfetch++;
         ms_pri_enqueue(st, this_inst);

         /* If we've just fetched some kind of branch,	*/
         /* then stop fetching until the branch is	*/
         /* resolved.					*/
         if (ip->flags & (INST_BRANCH | INST_CALL)) {
            if (th->branch_dly) {
               CPUError("No branches allowed in branch shadow!!\n");
            }
            
            /* IWIN_BRDLY is used to hold off issue	*/
            /* of this inst until branch delay	*/
            /* slots have been fetched.		*/
            st->iwin_flags [this_inst] |= IWIN_BRDLY + IWIN_BRANCH;
            th->branch_inum = this_inst;
            th->branch_dly = BRANCH_SLOTS + 1;
            th->returnpc = th->pc + (BRANCH_SLOTS*PC_INC);
            
            /* Handle branch likely instructions	*/
            if (ip->flags & INST_LIKELY) {
               th->stall_branch = 1;
               UpdateStallFetch (th);
               ms_branch (st, th, 1);
            }
         } else if (ip->flags & INST_SYS) {
            /* If it is a system call or a write to the FP	*/
            /* control register, stall fetching until all	*/
            /* prior instructions are executed.		*/
            if (ip->flags & INST_SC) {
               th->stall_sc = 1;
            } else if (ip->flags & (INST_CP0 | INST_CACHE)) {
               th->stall_cp0 = 1;
            } else if (ip->flags & INST_SYNC) {
               th->stall_sync = 1;
            } else if (ip->flags & INST_FPCTL) {
               th->stall_fpc = 1;
            } else {
               th->stall_sys = 1;
            }
            UpdateStallFetch (th);
            if (!(ip->flags & INST_SC))
               st->iwin_flags [this_inst] |= IWIN_FLUSH;
         }

         /* If there is branch processing in progress	*/
         /* for this thread, check if we have finished	*/
         /* fetching the branch delay instructions.	*/
         /* If so, then OK to process the branch.	*/
         /* In the case of the branch likely, the branch	*/
         /* has already been processed, so just update	*/
         /* the PC to the branch target.			*/

         if (th->branch_dly) {
            th->branch_dly--;
            if (th->branch_dly == 0) {
               if (th->branch_likely) {
                  th->pc = th->branch_likely_pc;
                  th->branch_likely = 0;
               } else {
                  th->stall_branch = 1;
                  UpdateStallFetch (th);
                  ms_branch (st, th, 0);
               }
            }
         }

         CheckInstAvail (st, this_inst);
         if (ip->flags & INST_ILLEGAL) {
            /* Current off the fetching for threads that encounter an
             * illegal instruction. 
             */
            th->stall_except = 1;
            UpdateStallFetch (th);
         }

         if (iwinDump) {
            dump_iwin_entry(st, this_inst);
         }
      }

   AddToStat (ST_FETCHES, nfetch);
   if (iwinDump && (fetches == FETCH_WIDTH)) {
      CPUPrint("<IWIN> fetch %d ACTIVE\n", fetches);
   }
}
