/*
 * Copyright (C) 1996-1998 by the Board of Trustees
 *    of Leland Stanford Junior University.
 * 
 * This file is part of the SimOS distribution. 
 * See LICENSE file for terms of the license. 
 *
 */

/*****************************************************************
 * fpu.c
 * 
 * $Author: verghese $
 * $Date: 1999/09/27 22:08:23 $
 *****************************************************************/
#include <math.h>
#include <limits.h>
#ifdef __alpha
#include <float.h>
#else
# ifndef i386
# include <ieeefp.h>
# endif
#endif

#include "ms.h"
#include "ms_fpu.h"

#include "mipsy.h"
#include "cpu.h"
#include "cpu_state.h"
#include "sim_error.h"
#include "cp0.h"

MXSOpcodeDesc mxsC1SPOpcodes[];
MXSOpcodeDesc mxsC1DPOpcodes[];
MXSOpcodeDesc mxsC1XOpcodes[];
MXSOpcodeDesc mxsBCOpcode, mxsCFCOpcode, mxsCTCOpcode;
MXSOpcodeDesc mxsMFCOpcode, mxsMTCOpcode, mxsDMFCOpcode, mxsDMTCOpcode;
MXSOpcodeDesc mxsCVTSWOpcode, mxsCVTDWOpcode, mxsCVTSLOpcode, mxsCVTDLOpcode;

#define ID(NAME) \
void NAME##ID(INST *ip, Reg pc)

#define COP1EX(NAME) \
Result NAME##EX(struct s_cpu_state *st, INST *ip, THREAD *th)

/*****************************************************************
 * Single-precision floating-point opcodes
 *****************************************************************/

ID(abs_s) { 
   ip->dst[0].flags = REG_FPS; ip->dst[0].reg = FDRI(ip->instr);
   ip->src[0].flags = REG_FPS; ip->src[0].reg = FSRI(ip->instr);
}
COP1EX(abs_s)
{
   int fd = ip->dst[0].fpreg;
   int fs = ip->src[0].fpreg;
   float contents_fs = FPR_S(fs);
   FPR_S(fd) = (contents_fs >= 0.0 ? contents_fs : -contents_fs);
   SIM_DEBUG(('f', "C1:ABS.S: abs(f%d %e) = f%d %e\n",
              fs, contents_fs, fd, FPR_S(fd)));
   return C1_SUCCESS;
}
ID(add_s) {
   ip->dst[0].flags = REG_FPS; ip->dst[0].reg = FDRI(ip->instr);
   ip->src[0].flags = REG_FPS; ip->src[0].reg = FSRI(ip->instr);
   ip->src[1].flags = REG_FPS; ip->src[1].reg = FTRI(ip->instr);
}
COP1EX(add_s)
{
   int fd = ip->dst[0].fpreg;
   int fs = ip->src[0].fpreg;
   int ft = ip->src[1].fpreg;
   float contents_fs = FPR_S(fs);
   float contents_ft = FPR_S(ft);
   FPR_S(fd) = contents_fs + contents_ft;
   SIM_DEBUG(('f', "C1:ADD.S: f%d %e + f%d %e = f%d %e\n",
              fs, contents_fs, ft, contents_ft, fd, FPR_S(fd)));
   return C1_SUCCESS;
}
ID(cvt_d_s) {
   ip->dst[0].flags = REG_FPD; ip->dst[0].reg = FDRI(ip->instr);
   ip->src[0].flags = REG_FPS; ip->src[0].reg = FSRI(ip->instr);
}
COP1EX(cvt_d_s)
{
   int fd = ip->dst[0].fpreg;
   int fs = ip->src[0].fpreg;
   float contents_fs = FPR_S(fs);
   FPR_D(fd) = contents_fs;
   SIM_DEBUG(('f', "C1:CVT.D.S: f%d %e --> f%d %e\n",
              fs, contents_fs, fd, FPR_D(fd)));
   return C1_SUCCESS;
}
ID(cvt_l_s) {
   ip->dst[0].flags = REG_FPL; ip->dst[0].reg = FDRI(ip->instr);
   ip->src[0].flags = REG_FPS; ip->src[0].reg = FSRI(ip->instr);
}
COP1EX(cvt_l_s)
{
   int fd = ip->dst[0].fpreg;
   int fs = ip->src[0].fpreg;
   float contents_fs = FPR_S(fs);
   uint roundMode = ROUND_MODE(st);
   float rounded;
   if (!ENABLE_64BIT(st)) {
      int inum = ip - st->iwin;
      EX_RECORD_EXCEPTION(st, inum, EXC_II, 0);
      return C1_FAILURE;
   }
   switch (roundMode) {
   case 0:                 /* RN: Round to nearest         */
      rounded = (contents_fs > 0.0 ? contents_fs + 0.5 : contents_fs - 0.5);
      break;
   case 1:                 /* RZ: Round toward 0           */
      rounded = contents_fs;
      break;
   case 2:                 /* RP: Round to +infinity       */
      rounded = (contents_fs >= 0.0 ? contents_fs + SP_NEAR_ONE : contents_fs);
      break;
   case 3:                 /* RM: Round to -infinity       */
      rounded = (contents_fs >= 0.0 ? contents_fs : contents_fs - SP_NEAR_ONE);
      break;
   }

   FPR_L(fd) = rounded;
   SIM_DEBUG(('f', "C1:CVT.L.S: f%d %e --> f%d %lld rm = %d\n", 
              fs, contents_fs, fd, FPR_L(fd), roundMode));

   /* Check for valid operand */
   if ((rounded > (float)LONGLONG_MAX) || (rounded < (float)LONGLONG_MIN)) {
      ip->fp_exc = MXS_FPE_V;
   }
   return C1_SUCCESS;
}   

ID(cvt_w_s) {
   ip->dst[0].flags = REG_FPW; ip->dst[0].reg = FDRI(ip->instr);
   ip->src[0].flags = REG_FPS; ip->src[0].reg = FSRI(ip->instr);
}
COP1EX(cvt_w_s)
{
   int fd = ip->dst[0].fpreg;
   int fs = ip->src[0].fpreg;
   float contents_fs = FPR_S(fs);
   uint roundMode = ROUND_MODE(st);
   float rounded;
   if (!ENABLE_64BIT(st)) {
      int inum = ip - st->iwin;
      EX_RECORD_EXCEPTION(st, inum, EXC_II, 0);
      return C1_FAILURE;
   }
   switch (roundMode) {
   case 0:                 /* RN: Round to nearest         */
      rounded = (contents_fs > 0.0 ? contents_fs + 0.5 : contents_fs - 0.5);
      break;
   case 1:                 /* RZ: Round toward 0           */
      rounded = contents_fs;
      break;
   case 2:                 /* RP: Round to +infinity       */
      rounded = (contents_fs >= 0.0 ? contents_fs + SP_NEAR_ONE : contents_fs);
      break;
   case 3:                 /* RM: Round to -infinity       */
      rounded = (contents_fs >= 0.0 ? contents_fs : contents_fs - SP_NEAR_ONE);
      break;
   }

   FPR_W(fd) = rounded;
   SIM_DEBUG(('f', "C1:CVT.W.S: f%d %e --> f%d %d rm = %d\n", 
              fs, contents_fs, fd, FPR_W(fd), roundMode));

   /* Check for valid operand */
   if ((rounded > (float)INT_MAX) || (rounded < (float)INT_MIN)) {
      ip->fp_exc = MXS_FPE_V;
   }
   return C1_SUCCESS;
}

ID(div_s) {
   ip->dst[0].flags = REG_FPS; ip->dst[0].reg = FDRI(ip->instr);
   ip->src[0].flags = REG_FPS; ip->src[0].reg = FSRI(ip->instr);
   ip->src[1].flags = REG_FPS; ip->src[1].reg = FTRI(ip->instr);
}
COP1EX(div_s)
{
   int fd = ip->dst[0].fpreg;
   int fs = ip->src[0].fpreg;
   int ft = ip->src[1].fpreg;
   float contents_fs = FPR_S(fs);
   float contents_ft = FPR_S(ft);
   if (contents_ft == 0.0) {
      CPUWarning("C1:DIV.S: divide by zero\n");
   }
   FPR_S(fd) = contents_fs / contents_ft;
   SIM_DEBUG(('f', "C1:DIV.S: f%d %e / f%d %e = f%d %e\n", 
              fs, contents_fs, ft, contents_ft, fd, FPR_S(fd)));
   return C1_SUCCESS;
}

ID(mov_s) {
   ip->dst[0].flags = REG_FPS; ip->dst[0].reg = FDRI(ip->instr);
   ip->src[0].flags = REG_FPS; ip->src[0].reg = FSRI(ip->instr);
}
COP1EX(mov_s)
{
   int fd = ip->dst[0].fpreg;
   int fs = ip->src[0].fpreg;
   float contents_fs = FPR_S(fs);
   FPR_S(fd) = contents_fs;
   SIM_DEBUG(('f', "C1:MOV.S: f%d %e --> f%d %e\n", 
              fs, contents_fs, fd, FPR_S(fd)));
   return C1_SUCCESS;
}
ID(movcf_s)
{ 
   ip->dst[0].flags = REG_FPS; ip->dst[0].reg = FDRI(ip->instr);
   ip->src[0].flags = REG_FPS; ip->src[0].reg = FSRI(ip->instr);
   ip->src[1].flags = REG_FPS; ip->src[1].reg = FDRI(ip->instr);
   ip->src[2].flags = REG_INT; ip->src[2].reg = MXS_REG_FPCSR;
}
COP1EX(movcf_s)
{
   int fd = ip->dst[0].fpreg;
   int fs = ip->src[0].fpreg;
   int old_fd = ip->src[1].fpreg;
   int fpcsr = ip->src[2].preg;
   Reg contents_fpcsr = GPR(fpcsr);
   float contents_fs = FPR_S(fs);
   float contents_fd = FPR_S(old_fd);
   int fcc = MOVC_FCC(ip->instr);
   int tf = TF(ip->instr);
   
   if (!ENABLE_MIPS4(st)) {
      int inum = ip - st->iwin;
      EX_RECORD_EXCEPTION(st, inum, EXC_II, 0); 
      return C1_FAILURE;
   }

   /* Move conditional on FP condition */
   if (tf) { 
      if (contents_fpcsr & M_csr_fcc(fcc)) {
         FPR_S(fd) = contents_fs;
         SIM_DEBUG(('f', "C1:MOVT.S: fcc %d T, f%d %e --> f%d %e\n", 
                    fcc, fs, contents_fs, fd, FPR_S(fd)));
      } else {
         FPR_S(fd) = contents_fd;
         SIM_DEBUG(('f', "C1:MOVT.S: fcc %d F, f%d %e unchanged\n",
                    fcc, fd, FPR_S(fd)));
      }
   } else {
      if (!(contents_fpcsr & M_csr_fcc(fcc))) {
         FPR_S(fd) = contents_fs;
         SIM_DEBUG(('f', "C1:MOVF.S: fcc %d F, f%d %e --> f%d %e\n", 
                    fcc, fs, contents_fs, fd, FPR_S(fd)));
      } else {
         FPR_S(fd) = contents_fd;
         SIM_DEBUG(('f', "C1:MOVF.S: fcc %d T, f%d %e unchanged\n",
                    fcc, fd, FPR_S(fd)));
      }
   }
   return C1_SUCCESS;
}

ID(movn_s) {
   ip->dst[0].flags = REG_FPS; ip->dst[0].reg = FDRI(ip->instr);
   ip->src[0].flags = REG_FPS; ip->src[0].reg = FSRI(ip->instr);
   ip->src[1].flags = REG_INT; ip->src[1].reg = RTRI(ip->instr);
   ip->src[2].flags = REG_FPS; ip->src[2].reg = FDRI(ip->instr);
}
COP1EX(movn_s)
{
   int fd = ip->dst[0].fpreg;
   int fs = ip->src[0].fpreg;
   int rt = ip->src[1].preg;
   int old_fd = ip->src[2].fpreg;
   float contents_fs = FPR_S(fs);
   Reg contents_rt = GPR(rt);
   float contents_fd = FPR_S(old_fd);

   if (!ENABLE_MIPS4(st)) {
      int inum = ip - st->iwin;
      EX_RECORD_EXCEPTION(st, inum, EXC_II, 0); 
      return C1_FAILURE;
   }

   if (contents_rt != 0) {
      FPR_S(fd) = contents_fs;
      SIM_DEBUG(('f', "C1:MOVN.S: r%d %lld, f%d %e to f%d %e\n", 
                 rt, (uint64) contents_rt, 
                 fs, contents_fs, fd, FPR_S(fd)));
   } else {
      FPR_S(fd) = contents_fd;
      SIM_DEBUG(('f', "C1:MOVN.S: r%d %lld, f%d %e unchanged\n", 
                 rt, (int64) contents_rt, fd, FPR_S(fd)));
   }
   return C1_SUCCESS;
}

ID(movz_s) {
   ip->dst[0].flags = REG_FPS; ip->dst[0].reg = FDRI(ip->instr);
   ip->src[0].flags = REG_FPS; ip->src[0].reg = FSRI(ip->instr);
   ip->src[1].flags = REG_INT; ip->src[1].reg = RTRI(ip->instr);
   ip->src[2].flags = REG_FPS; ip->src[2].reg = FDRI(ip->instr);
}
COP1EX(movz_s)
{
   int fd = ip->dst[0].fpreg;
   int fs = ip->src[0].fpreg;
   int rt = ip->src[1].preg;
   int old_fd = ip->src[2].fpreg;
   float contents_fs = FPR_S(fs);
   Reg contents_rt = GPR(rt);
   float contents_fd = FPR_S(old_fd);

   if (!ENABLE_MIPS4(st)) {
      int inum = ip - st->iwin;
      EX_RECORD_EXCEPTION(st, inum, EXC_II, 0); 
      return C1_FAILURE;
   }

   if (contents_rt == 0) {
      FPR_S(fd) = contents_fs;
      SIM_DEBUG(('f', "C1:MOVZ.S: r%d %lld, f%d %e to f%d %e\n", 
                 rt, (int64) contents_rt, 
                 fs, contents_fs, fd, FPR_S(fd)));
   } else {
      FPR_S(fd) = contents_fd;
      SIM_DEBUG(('f', "C1:MOVZ.S: r%d %lld, f%d %e unchanged\n", 
                 rt, (uint64) contents_rt, fd, FPR_S(fd)));
   }
   return C1_SUCCESS;
}

ID(mul_s) {
   ip->dst[0].flags = REG_FPS; ip->dst[0].reg = FDRI(ip->instr);
   ip->src[0].flags = REG_FPS; ip->src[0].reg = FSRI(ip->instr);
   ip->src[1].flags = REG_FPS; ip->src[1].reg = FTRI(ip->instr);
}
COP1EX(mul_s)
{
   int fd = ip->dst[0].fpreg;
   int fs = ip->src[0].fpreg;
   int ft = ip->src[1].fpreg;
   float contents_fs = FPR_S(fs);
   float contents_ft = FPR_S(ft);
   FPR_S(fd) = contents_fs * contents_ft;
   SIM_DEBUG(('f', "C1:MUL.S: f%d %e * f%d %e = f%d %e\n", 
              fs, contents_fs, ft, contents_ft, fd, FPR_S(fd)));
   return C1_SUCCESS;
}

ID(neg_s) {
   ip->dst[0].flags = REG_FPS; ip->dst[0].reg = FDRI(ip->instr);
   ip->src[0].flags = REG_FPS; ip->src[0].reg = FSRI(ip->instr);
}
COP1EX(neg_s)
{
   int fd = ip->dst[0].fpreg;
   int fs = ip->src[0].fpreg;
   float contents_fs = FPR_S(fs);
   FPR_S(fd) = -contents_fs;
   SIM_DEBUG(('f', "C1:NEG.S: f%d %e --> f%d %e\n", 
              fs, contents_fs, fd, FPR_S(fd)));
   return C1_SUCCESS;
}

ID(sub_s) {
   ip->dst[0].flags = REG_FPS; ip->dst[0].reg = FDRI(ip->instr);
   ip->src[0].flags = REG_FPS; ip->src[0].reg = FSRI(ip->instr);
   ip->src[1].flags = REG_FPS; ip->src[1].reg = FTRI(ip->instr);
}
COP1EX(sub_s)
{
   int fd = ip->dst[0].fpreg;
   int fs = ip->src[0].fpreg;
   int ft = ip->src[1].fpreg;
   float contents_fs = FPR_S(fs);
   float contents_ft = FPR_S(ft);
   FPR_S(fd) = contents_fs - contents_ft;
   SIM_DEBUG(('f', "C1:SUB.S: f%d %e - f%d %e = f%d %e\n", 
              fs, contents_fs, ft, contents_ft, fd, FPR_S(fd)));
   return C1_SUCCESS;
}

ID(sqrt_s) {
   ip->dst[0].flags = REG_FPS; ip->dst[0].reg = FDRI(ip->instr);
   ip->src[0].flags = REG_FPS; ip->src[0].reg = FSRI(ip->instr);
}
COP1EX(sqrt_s)
{
   int fd = ip->dst[0].fpreg;
   int fs = ip->src[0].fpreg;
   float contents_fs = FPR_S(fs);
   FPR_S(fd) = sqrtf(contents_fs);
   SIM_DEBUG(('f', "C1:SQRT.S: sqrt(f%d %e) = f%d %e\n", 
              fs, contents_fs, fd, FPR_S(fd)));
   return C1_SUCCESS;
}

ID(trunc_w_s) {
   ip->dst[0].flags = REG_FPW; ip->dst[0].reg = FDRI(ip->instr);
   ip->src[0].flags = REG_FPS; ip->src[0].reg = FSRI(ip->instr);
}
COP1EX(trunc_w_s)
{
   int fd = ip->dst[0].fpreg;
   int fs = ip->src[0].fpreg;
   float contents_fs = FPR_S(fs);
   FPR_W(fd) = contents_fs;
   SIM_DEBUG(('f', "C1:TRUNC.W.S: f%d %e --> f%d %d\n", 
              fs, contents_fs, fd, FPR_W(fd)));
   return C1_SUCCESS;
}

ID(trunc_l_s) {
   ip->dst[0].flags = REG_FPL; ip->dst[0].reg = FDRI(ip->instr);
   ip->src[0].flags = REG_FPS; ip->src[0].reg = FSRI(ip->instr);
}
COP1EX(trunc_l_s)
{
   int fd = ip->dst[0].fpreg;
   int fs = ip->src[0].fpreg;
   float contents_fs = FPR_S(fs);
   FPR_L(fd) = contents_fs;
   SIM_DEBUG(('f', "C1:TRUNC.L.S: f%d %e --> f%d %lld\n", 
              fs, contents_fs, fd, FPR_L(fd)));
   return C1_SUCCESS;
}

ID(floor_w_s) {
   ip->dst[0].flags = REG_FPW; ip->dst[0].reg = FDRI(ip->instr);
   ip->src[0].flags = REG_FPS; ip->src[0].reg = FSRI(ip->instr);
}
COP1EX(floor_w_s)
{
   int fd = ip->dst[0].fpreg;
   int fs = ip->src[0].fpreg;
   float contents_fs = FPR_S(fs);
   FPR_W(fd) = (contents_fs >= 0.0 ? 
                       contents_fs : 
                       contents_fs - SP_NEAR_ONE);
   SIM_DEBUG(('f', "C1:FLOOR.W.S: f%d %e --> f%d %d\n", 
              fs, contents_fs, fd, FPR_W(fd)));
   return C1_SUCCESS;
}

ID(floor_l_s) {
   ip->dst[0].flags = REG_FPL; ip->dst[0].reg = FDRI(ip->instr);
   ip->src[0].flags = REG_FPS; ip->src[0].reg = FSRI(ip->instr);
}
COP1EX(floor_l_s)
{
   int fd = ip->dst[0].fpreg;
   int fs = ip->src[0].fpreg;
   float contents_fs = FPR_S(fs);
   FPR_L(fd) = (contents_fs >= 0.0 ? 
                            contents_fs : 
                            contents_fs - SP_NEAR_ONE);
   SIM_DEBUG(('f', "C1:FLOOR.L.S: f%d %e --> f%d %lld\n", 
              fs, contents_fs, fd, FPR_L(fd)));
   return C1_SUCCESS;
}

ID(recip_s) {
   ip->dst[0].flags = REG_FPS; ip->dst[0].reg = FDRI(ip->instr);
   ip->src[0].flags = REG_FPS; ip->src[0].reg = FSRI(ip->instr);
}
COP1EX(recip_s)
{
   int fd = ip->dst[0].fpreg;
   int fs = ip->src[0].fpreg;
   float contents_fs = FPR_S(fs);
   if (!ENABLE_MIPS4(st)) {
      int inum = ip - st->iwin;
      EX_RECORD_EXCEPTION(st, inum, EXC_II, 0); 
      return C1_FAILURE;
   }
   if (contents_fs == 0.0) {
      CPUWarning("C1:RECIP.S: divide by zero\n");
   }
   FPR_S(fd) = 1.0 / contents_fs;
   SIM_DEBUG(('f', "C1:RECIP.S: 1.0 / f%d %e = f%d %e\n", 
              fs, contents_fs, fd, FPR_S(fd)));
   return C1_SUCCESS;
}

ID(rsqrt_s) {
   ip->dst[0].flags = REG_FPS; ip->dst[0].reg = FDRI(ip->instr);
   ip->src[0].flags = REG_FPS; ip->src[0].reg = FSRI(ip->instr);
}
COP1EX(rsqrt_s)
{
   int fd = ip->dst[0].fpreg;
   int fs = ip->src[0].fpreg;
   float contents_fs = FPR_S(fs);
   if (!ENABLE_MIPS4(st)) {
      int inum = ip - st->iwin;
      EX_RECORD_EXCEPTION(st, inum, EXC_II, 0); 
      return C1_FAILURE;
   }
   if (contents_fs == 0.0) {
      CPUWarning("C1:RSQRT.S: divide by zero\n");
   }
   FPR_S(fd) = 1.0 / sqrtf(contents_fs);
   SIM_DEBUG(('f', "C1:RSQRT.S: 1.0 / sqrt(f%d %e) = f%d %e\n", 
              fs, contents_fs, fd, FPR_S(fd)));
   return C1_SUCCESS;
}

ID(round_w_s) {
   ip->dst[0].flags = REG_FPW; ip->dst[0].reg = FDRI(ip->instr);
   ip->src[0].flags = REG_FPS; ip->src[0].reg = FSRI(ip->instr);
}
COP1EX(round_w_s)
{
   int fd = ip->dst[0].fpreg;
   int fs = ip->src[0].fpreg;
   float contents_fs = FPR_S(fs);
   FPR_W(fd) = (contents_fs > 0.0 ? 
                       contents_fs + 0.5 : 
                       contents_fs - 0.5);
   SIM_DEBUG(('f', "C1:ROUND.W.S: f%d %e --> f%d %d\n", 
              fs, contents_fs, fd, FPR_W(fd)));
   return C1_SUCCESS;
}

ID(round_l_s) {
   ip->dst[0].flags = REG_FPL; ip->dst[0].reg = FDRI(ip->instr);
   ip->src[0].flags = REG_FPS; ip->src[0].reg = FSRI(ip->instr);
}
COP1EX(round_l_s)
{
   int fd = ip->dst[0].fpreg;
   int fs = ip->src[0].fpreg;
   float contents_fs = FPR_S(fs);
   FPR_L(fd) = (contents_fs > 0.0 ? 
                            contents_fs + 0.5 : 
                            contents_fs - 0.5);
   SIM_DEBUG(('f', "C1:ROUND.W.S: f%d %e --> f%d %lld\n", 
              fs, contents_fs, fd, FPR_L(fd)));
   return C1_SUCCESS;
}

ID(ceil_w_s) {
   ip->dst[0].flags = REG_FPW; ip->dst[0].reg = FDRI(ip->instr);
   ip->src[0].flags = REG_FPS; ip->src[0].reg = FSRI(ip->instr);
}
COP1EX(ceil_w_s)
{
   int fd = ip->dst[0].fpreg;
   int fs = ip->src[0].fpreg;
   float contents_fs = FPR_S(fs);
   FPR_W(fd) = (contents_fs > 0.0 ? 
                       contents_fs + SP_NEAR_ONE : 
                       contents_fs);
   SIM_DEBUG(('f', "C1:CEIL.W.S: f%d %e --> f%d %d\n", 
              fs, contents_fs, fd, FPR_W(fd)));
   return C1_SUCCESS;
}

ID(ceil_l_s) {
   ip->dst[0].flags = REG_FPL; ip->dst[0].reg = FDRI(ip->instr);
   ip->src[0].flags = REG_FPS; ip->src[0].reg = FSRI(ip->instr);
}
COP1EX(ceil_l_s)
{
   int fd = ip->dst[0].fpreg;
   int fs = ip->src[0].fpreg;
   float contents_fs = FPR_S(fs);
   FPR_L(fd) = (contents_fs > 0.0 ? 
                            contents_fs + SP_NEAR_ONE : 
                            contents_fs);
   SIM_DEBUG(('f', "C1:CEIL.L.S: f%d %e --> f%d %lld\n", 
              fs, contents_fs, fd, FPR_L(fd)));
   return C1_SUCCESS;
}

/* Common decode routine for all c_cond_s opcodes */
ID(c_cond_s)
{
   ip->dst[0].flags = REG_INT; ip->dst[0].reg = MXS_REG_FPCSR;
   ip->src[0].flags = REG_FPS; ip->src[0].reg = FSRI(ip->instr);
   ip->src[1].flags = REG_FPS; ip->src[1].reg = FTRI(ip->instr);
}

ID(c_eq_s) { c_cond_sID(ip, pc); }
COP1EX(c_eq_s)
{
   CPUState *P = (CPUState *)(st->mipsyPtr);
   int fs = ip->src[0].fpreg;
   int ft = ip->src[1].fpreg;
   float contents_fs = FPR_S(fs);
   float contents_ft = FPR_S(ft);
   int cond = (contents_fs == contents_ft);
   int fcc = C_FCC(ip->instr);
   int fpcsr = ip->dst[0].preg;
   SIM_DEBUG(('f', "C1:C.EQ.S: f%d %e, f%d %e --> fcc%d %d\n",
              fs, contents_fs, ft, contents_ft, fcc, cond));
   return CompareOp(st, cond, fcc, fpcsr);
}

ID(c_f_s) { c_cond_sID(ip, pc); }
COP1EX(c_f_s)
{
   CPUState *P = (CPUState *)(st->mipsyPtr);
   int fs = ip->src[0].fpreg;
   int ft = ip->src[1].fpreg;
   float contents_fs = FPR_S(fs);
   float contents_ft = FPR_S(ft);
   int cond = 0;
   int fcc = C_FCC(ip->instr);
   int fpcsr = ip->dst[0].preg;
   SIM_DEBUG(('f', "C1:C.F.S: f%d %e, f%d %e --> fcc%d %d\n",
              fs, contents_fs, ft, contents_ft, fcc, cond));
   return CompareOp(st, cond, fcc, fpcsr);
}

ID(c_le_s) { c_cond_sID(ip, pc); }
COP1EX(c_le_s)
{
   CPUState *P = (CPUState *)(st->mipsyPtr);
   int fs = ip->src[0].fpreg;
   int ft = ip->src[1].fpreg;
   float contents_fs = FPR_S(fs);
   float contents_ft = FPR_S(ft);
   int cond = (contents_fs <= contents_ft);
   int fcc = C_FCC(ip->instr);
   int fpcsr = ip->dst[0].preg;
   SIM_DEBUG(('f', "C1:C.LE.S: f%d %e, f%d %e --> fcc%d %d\n",
              fs, contents_fs, ft, contents_ft, fcc, cond));
   return CompareOp(st, cond, fcc, fpcsr);
}

ID(c_lt_s) { c_cond_sID(ip, pc); }
COP1EX(c_lt_s)
{
   CPUState *P = (CPUState *)(st->mipsyPtr);
   int fs = ip->src[0].fpreg;
   int ft = ip->src[1].fpreg;
   float contents_fs = FPR_S(fs);
   float contents_ft = FPR_S(ft);
   int cond = (contents_fs < contents_ft);
   int fcc = C_FCC(ip->instr);
   int fpcsr = ip->dst[0].preg;
   SIM_DEBUG(('f', "C1:C.LT.S: f%d %e, f%d %e --> fcc%d %d\n",
              fs, contents_fs, ft, contents_ft, fcc, cond));
   return CompareOp(st, cond, fcc, fpcsr);
}

ID(c_nge_s) { c_cond_sID(ip, pc); }
COP1EX(c_nge_s)
{
   CPUState *P = (CPUState *)(st->mipsyPtr);
   int fs = ip->src[0].fpreg;
   int ft = ip->src[1].fpreg;
   float contents_fs = FPR_S(fs);
   float contents_ft = FPR_S(ft);
   int cond = (contents_fs <= contents_ft);
   int fcc = C_FCC(ip->instr);
   int fpcsr = ip->dst[0].preg;
   SIM_DEBUG(('f', "C1:C.NGE.S: f%d %e, f%d %e --> fcc%d %d\n",
              fs, contents_fs, ft, contents_ft, fcc, cond));
   return CompareOp(st, cond, fcc, fpcsr);
}

ID(c_ngl_s) { c_cond_sID(ip, pc); }
COP1EX(c_ngl_s)
{
   CPUState *P = (CPUState *)(st->mipsyPtr);
   int fs = ip->src[0].fpreg;
   int ft = ip->src[1].fpreg;
   float contents_fs = FPR_S(fs);
   float contents_ft = FPR_S(ft);
   int cond = (contents_fs == contents_ft);
   int fcc = C_FCC(ip->instr);
   int fpcsr = ip->dst[0].preg;
   SIM_DEBUG(('f', "C1:C.NGL.S: f%d %e, f%d %e --> fcc%d %d\n",
              fs, contents_fs, ft, contents_ft, fcc, cond));
   return CompareOp(st, cond, fcc, fpcsr);
}

ID(c_ngle_s) { c_cond_sID(ip, pc); }
COP1EX(c_ngle_s)
{
   CPUState *P = (CPUState *)(st->mipsyPtr);
   int fs = ip->src[0].fpreg;
   int ft = ip->src[1].fpreg;
   float contents_fs = FPR_S(fs);
   float contents_ft = FPR_S(ft);
   int cond = 0;
   int fcc = C_FCC(ip->instr);
   int fpcsr = ip->dst[0].preg;
   SIM_DEBUG(('f', "C1:C.NGLE.S: f%d %e, f%d %e --> fcc%d %d\n",
              fs, contents_fs, ft, contents_ft, fcc, cond));
   return CompareOp(st, cond, fcc, fpcsr);
}

ID(c_ngt_s) { c_cond_sID(ip, pc); }
COP1EX(c_ngt_s)
{
   CPUState *P = (CPUState *)(st->mipsyPtr);
   int fs = ip->src[0].fpreg;
   int ft = ip->src[1].fpreg;
   float contents_fs = FPR_S(fs);
   float contents_ft = FPR_S(ft);
   int cond = (contents_fs <= contents_ft);
   int fcc = C_FCC(ip->instr);
   int fpcsr = ip->dst[0].preg;
   SIM_DEBUG(('f', "C1:C.NGT.S: f%d %e, f%d %e --> fcc%d %d\n",
              fs, contents_fs, ft, contents_ft, fcc, cond));
   return CompareOp(st, cond, fcc, fpcsr);
}

ID(c_ole_s) { c_cond_sID(ip, pc); }
COP1EX(c_ole_s)
{
   CPUState *P = (CPUState *)(st->mipsyPtr);
   int fs = ip->src[0].fpreg;
   int ft = ip->src[1].fpreg;
   float contents_fs = FPR_S(fs);
   float contents_ft = FPR_S(ft);
   int cond = (contents_fs < contents_ft);
   int fcc = C_FCC(ip->instr);
   int fpcsr = ip->dst[0].preg;
   SIM_DEBUG(('f', "C1:C.OLE.S: f%d %e, f%d %e --> fcc%d %d\n",
              fs, contents_fs, ft, contents_ft, fcc, cond));
   return CompareOp(st, cond, fcc, fpcsr);
}

ID(c_olt_s) { c_cond_sID(ip, pc); }
COP1EX(c_olt_s)
{
   CPUState *P = (CPUState *)(st->mipsyPtr);
   int fs = ip->src[0].fpreg;
   int ft = ip->src[1].fpreg;
   float contents_fs = FPR_S(fs);
   float contents_ft = FPR_S(ft);
   int cond = (contents_fs < contents_ft);
   int fcc = C_FCC(ip->instr);
   int fpcsr = ip->dst[0].preg;
   SIM_DEBUG(('f', "C1:C.OLT.S: f%d %e, f%d %e --> fcc%d %d\n",
              fs, contents_fs, ft, contents_ft, fcc, cond));
   return CompareOp(st, cond, fcc, fpcsr);
}

ID(c_seq_s) { c_cond_sID(ip, pc); }
COP1EX(c_seq_s)
{
   CPUState *P = (CPUState *)(st->mipsyPtr);
   int fs = ip->src[0].fpreg;
   int ft = ip->src[1].fpreg;
   float contents_fs = FPR_S(fs);
   float contents_ft = FPR_S(ft);
   int cond = (contents_fs == contents_ft);
   int fcc = C_FCC(ip->instr);
   int fpcsr = ip->dst[0].preg;
   SIM_DEBUG(('f', "C1:C.SEQ.S: f%d %e, f%d %e --> fcc%d %d\n",
              fs, contents_fs, ft, contents_ft, fcc, cond));
   return CompareOp(st, cond, fcc, fpcsr);
}

ID(c_sf_s) { c_cond_sID(ip, pc); }
COP1EX(c_sf_s)
{
   CPUState *P = (CPUState *)(st->mipsyPtr);
   int fs = ip->src[0].fpreg;
   int ft = ip->src[1].fpreg;
   float contents_fs = FPR_S(fs);
   float contents_ft = FPR_S(ft);
   int cond = 0;
   int fcc = C_FCC(ip->instr);
   int fpcsr = ip->dst[0].preg;
   SIM_DEBUG(('f', "C1:C.SF.S: f%d %e, f%d %e --> fcc%d %d\n",
              fs, contents_fs, ft, contents_ft, fcc, cond));
   return CompareOp(st, cond, fcc, fpcsr);
}

ID(c_ueq_s) { c_cond_sID(ip, pc); }
COP1EX(c_ueq_s)
{
   CPUState *P = (CPUState *)(st->mipsyPtr);
   int fs = ip->src[0].fpreg;
   int ft = ip->src[1].fpreg;
   float contents_fs = FPR_S(fs);
   float contents_ft = FPR_S(ft);
   int cond = (contents_fs == contents_ft);
   int fcc = C_FCC(ip->instr);
   int fpcsr = ip->dst[0].preg;
   SIM_DEBUG(('f', "C1:C.UEQ.S: f%d %e, f%d %e --> fcc%d %d\n",
              fs, contents_fs, ft, contents_ft, fcc, cond));
   return CompareOp(st, cond, fcc, fpcsr);
}

ID(c_ule_s) { c_cond_sID(ip, pc); }
COP1EX(c_ule_s)
{
   CPUState *P = (CPUState *)(st->mipsyPtr);
   int fs = ip->src[0].fpreg;
   int ft = ip->src[1].fpreg;
   float contents_fs = FPR_S(fs);
   float contents_ft = FPR_S(ft);
   int cond = (contents_fs <= contents_ft);
   int fcc = C_FCC(ip->instr);
   int fpcsr = ip->dst[0].preg;
   SIM_DEBUG(('f', "C1:C.ULE.S: f%d %e, f%d %e --> fcc%d %d\n",
              fs, contents_fs, ft, contents_ft, fcc, cond));
   return CompareOp(st, cond, fcc, fpcsr);
}

ID(c_ult_s) { c_cond_sID(ip, pc); }
COP1EX(c_ult_s)
{
   CPUState *P = (CPUState *)(st->mipsyPtr);
   int fs = ip->src[0].fpreg;
   int ft = ip->src[1].fpreg;
   float contents_fs = FPR_S(fs);
   float contents_ft = FPR_S(ft);
   int cond = (contents_fs < contents_ft);
   int fcc = C_FCC(ip->instr);
   int fpcsr = ip->dst[0].preg;
   SIM_DEBUG(('f', "C1:C.ULT.S: f%d %e, f%d %e --> fcc%d %d\n",
              fs, contents_fs, ft, contents_ft, fcc, cond));
   return CompareOp(st, cond, fcc, fpcsr);
}

ID(c_un_s) { c_cond_sID(ip, pc); }
COP1EX(c_un_s)
{
   CPUState *P = (CPUState *)(st->mipsyPtr);
   int fs = ip->src[0].fpreg;
   int ft = ip->src[1].fpreg;
   float contents_fs = FPR_S(fs);
   float contents_ft = FPR_S(ft);
   int cond = 0;
   int fcc = C_FCC(ip->instr);
   int fpcsr = ip->dst[0].preg;
   SIM_DEBUG(('f', "C1:C.UN.S: f%d %e, f%d %e --> fcc%d %d\n",
              fs, contents_fs, ft, contents_ft, fcc, cond));
   return CompareOp(st, cond, fcc, fpcsr);
}

/*****************************************************************
 * Double-precision floating-point opcodes
 *****************************************************************/

ID(abs_d) {
   ip->dst[0].flags = REG_FPD; ip->dst[0].reg = FDRI(ip->instr);
   ip->src[0].flags = REG_FPD; ip->src[0].reg = FSRI(ip->instr);
}
COP1EX(abs_d)
{
   int fd = ip->dst[0].fpreg;
   int fs = ip->src[0].fpreg;
   double contents_fs = FPR_D(fs);
   FPR_D(fd) = (contents_fs >= 0.0 ? contents_fs : -contents_fs);
   SIM_DEBUG(('f', "C1:ABS.D: abs(f%d %e) = f%d %e\n",
              fs, contents_fs, fd, FPR_D(fd)));
   return C1_SUCCESS;
}
ID(add_d) {
   ip->dst[0].flags = REG_FPD; ip->dst[0].reg = FDRI(ip->instr);
   ip->src[0].flags = REG_FPD; ip->src[0].reg = FSRI(ip->instr);
   ip->src[1].flags = REG_FPD; ip->src[1].reg = FTRI(ip->instr);
}
COP1EX(add_d)
{
   int fd = ip->dst[0].fpreg;
   int fs = ip->src[0].fpreg;
   int ft = ip->src[1].fpreg;
   double contents_fs = FPR_D(fs);
   double contents_ft = FPR_D(ft);
   FPR_D(fd) = contents_fs + contents_ft;
   SIM_DEBUG(('f', "C1:ADD.D: f%d %e + f%d %e = f%d %e\n",
              fs, contents_fs, ft, contents_ft, fd, FPR_D(fd)));
   return C1_SUCCESS;
}
ID(cvt_s_d) {
   ip->dst[0].flags = REG_FPS; ip->dst[0].reg = FDRI(ip->instr);
   ip->src[0].flags = REG_FPD; ip->src[0].reg = FSRI(ip->instr);
}
COP1EX(cvt_s_d)
{
   int fd = ip->dst[0].fpreg;
   int fs = ip->src[0].fpreg;
   double contents_fs = FPR_D(fs);
   FPR_S(fd) = contents_fs;
   SIM_DEBUG(('f', "C1:CVT.S.D: f%d %e --> f%d %e\n",
              fs, contents_fs, fd, FPR_S(fd)));
   return C1_SUCCESS;
}
ID(cvt_l_d) {
   ip->dst[0].flags = REG_FPL; ip->dst[0].reg = FDRI(ip->instr);
   ip->src[0].flags = REG_FPD; ip->src[0].reg = FSRI(ip->instr);
}
COP1EX(cvt_l_d)
{
   int fd = ip->dst[0].fpreg;
   int fs = ip->src[0].fpreg;
   double contents_fs = FPR_D(fs);
   uint roundMode = ROUND_MODE(st);
   double rounded;
   if (!ENABLE_64BIT(st)) {
      int inum = ip - st->iwin;
      EX_RECORD_EXCEPTION(st, inum, EXC_II, 0);
      return C1_FAILURE;
   }
   switch (roundMode) {
   case 0:                 /* RN: Round to nearest         */
      rounded = (contents_fs > 0.0 ? contents_fs + 0.5 : contents_fs - 0.5);
      break;
   case 1:                 /* RZ: Round toward 0           */
      rounded = contents_fs;
      break;
   case 2:                 /* RP: Round to +infinity       */
      rounded = (contents_fs >= 0.0 ? contents_fs + SP_NEAR_ONE : contents_fs);
      break;
   case 3:                 /* RM: Round to -infinity       */
      rounded = (contents_fs >= 0.0 ? contents_fs : contents_fs - SP_NEAR_ONE);
      break;
   }

   FPR_L(fd) = rounded;
   SIM_DEBUG(('f', "C1:CVT.L.D: f%d %e --> f%d %lld rm = %d\n", 
              fs, contents_fs, fd, FPR_L(fd), roundMode));

   /* Check for valid operand */
   if ((rounded > (float)LONGLONG_MAX) || (rounded < (float)LONGLONG_MIN)) {
      ip->fp_exc = MXS_FPE_V;
   }
   return C1_SUCCESS;
}   

ID(cvt_w_d) {
   ip->dst[0].flags = REG_FPW; ip->dst[0].reg = FDRI(ip->instr);
   ip->src[0].flags = REG_FPD; ip->src[0].reg = FSRI(ip->instr);
}
COP1EX(cvt_w_d)
{
   int fd = ip->dst[0].fpreg;
   int fs = ip->src[0].fpreg;
   double contents_fs = FPR_D(fs);
   uint roundMode = ROUND_MODE(st);
   double rounded;
   if (!ENABLE_64BIT(st)) {
      int inum = ip - st->iwin;
      EX_RECORD_EXCEPTION(st, inum, EXC_II, 0);
      return C1_FAILURE;
   }
   switch (roundMode) {
   case 0:                 /* RN: Round to nearest         */
      rounded = (contents_fs > 0.0 ? contents_fs + 0.5 : contents_fs - 0.5);
      break;
   case 1:                 /* RZ: Round toward 0           */
      rounded = contents_fs;
      break;
   case 2:                 /* RP: Round to +infinity       */
      rounded = (contents_fs >= 0.0 ? contents_fs + SP_NEAR_ONE : contents_fs);
      break;
   case 3:                 /* RM: Round to -infinity       */
      rounded = (contents_fs >= 0.0 ? contents_fs : contents_fs - SP_NEAR_ONE);
      break;
   }

   FPR_W(fd) = rounded;
   SIM_DEBUG(('f', "C1:CVT.W.D: f%d %e --> f%d %d rm = %d\n", 
              fs, contents_fs, fd, FPR_W(fd), roundMode));

   /* Check for valid operand */
   if ((rounded > (float)INT_MAX) || (rounded < (float)INT_MIN)) {
      ip->fp_exc = MXS_FPE_V;
   }
   return C1_SUCCESS;
}

ID(div_d) {
   ip->dst[0].flags = REG_FPD; ip->dst[0].reg = FDRI(ip->instr);
   ip->src[0].flags = REG_FPD; ip->src[0].reg = FSRI(ip->instr);
   ip->src[1].flags = REG_FPD; ip->src[1].reg = FTRI(ip->instr);
}
COP1EX(div_d)
{
   int fd = ip->dst[0].fpreg;
   int fs = ip->src[0].fpreg;
   int ft = ip->src[1].fpreg;
   double contents_fs = FPR_D(fs);
   double contents_ft = FPR_D(ft);
   if (contents_ft == 0.0) {
      CPUWarning("C1:DIV.D: divide by zero\n");
   }
   FPR_D(fd) = contents_fs / contents_ft;
   SIM_DEBUG(('f', "C1:DIV.D: f%d %e / f%d %e = f%d %e\n", 
              fs, contents_fs, ft, contents_ft, fd, FPR_D(fd)));
   return C1_SUCCESS;
}

ID(mov_d) {
   ip->dst[0].flags = REG_FPD; ip->dst[0].reg = FDRI(ip->instr);
   ip->src[0].flags = REG_FPD; ip->src[0].reg = FSRI(ip->instr);
}
COP1EX(mov_d)
{
   int fd = ip->dst[0].fpreg;
   int fs = ip->src[0].fpreg;
   double contents_fs = FPR_D(fs);
   FPR_D(fd) = contents_fs;
   SIM_DEBUG(('f', "C1:MOV.D: f%d %e --> f%d %e\n", 
              fs, contents_fs, fd, FPR_D(fd)));
   return C1_SUCCESS;
}

ID(movcf_d)
{
   ip->dst[0].flags = REG_FPD; ip->dst[0].reg = FDRI(ip->instr);
   ip->src[0].flags = REG_FPD; ip->src[0].reg = FSRI(ip->instr);
   ip->src[1].flags = REG_FPD; ip->src[1].reg = FDRI(ip->instr);
   ip->src[2].flags = REG_INT; ip->src[2].reg = MXS_REG_FPCSR;
}
COP1EX(movcf_d)
{
   int fd = ip->dst[0].fpreg;
   int fs = ip->src[0].fpreg;
   int old_fd = ip->src[1].fpreg;
   int fpcsr = ip->src[2].preg;
   Reg contents_fpcsr = GPR(fpcsr);
   double contents_fs = FPR_D(fs);
   double contents_fd = FPR_S(fd);
   int fcc = MOVC_FCC(ip->instr);
   int tf = TF(ip->instr);

   if (!ENABLE_MIPS4(st)) {
      int inum = ip - st->iwin;
      EX_RECORD_EXCEPTION(st, inum, EXC_II, 0); 
      return C1_FAILURE;
   }

   /* Move conditional on FP condition */
   if (tf) { 
      if (contents_fpcsr & M_csr_fcc(fcc)) {
         FPR_D(fd) = contents_fs;
         SIM_DEBUG(('f', "C1:MOVT.D: fcc %d T, f%d %e --> f%d %e\n", 
                    fcc, fs, contents_fs, fd, FPR_D(fd)));
      } else {
         FPR_D(fd) = contents_fd;
         SIM_DEBUG(('f', "C1:MOVT.D: fcc %d F, f%d %e unchanged\n",
                    fcc, fd, FPR_D(fd)));
      }
   } else {
      if (!(contents_fpcsr & M_csr_fcc(fcc))) {
         FPR_D(fd) = contents_fs;
         SIM_DEBUG(('f', "C1:MOVF.D: fcc %d F, f%d %e --> f%d %e\n", 
                    fcc, fs, contents_fs, fd, FPR_D(fd)));
      } else {
         FPR_D(fd) = contents_fd;
         SIM_DEBUG(('f', "C1:MOVF.D: fcc %d T, f%d %e unchanged\n",
                    fcc, fd, FPR_D(fd)));
      }
   }
   return C1_SUCCESS;
}

ID(movn_d) {
   ip->dst[0].flags = REG_FPD; ip->dst[0].reg = FDRI(ip->instr);
   ip->src[0].flags = REG_FPD; ip->src[0].reg = FSRI(ip->instr);
   ip->src[1].flags = REG_INT; ip->src[1].reg = RTRI(ip->instr);
   ip->src[2].flags = REG_FPD; ip->src[1].reg = FDRI(ip->instr);
}
COP1EX(movn_d)
{
   int fd = ip->dst[0].fpreg;
   int fs = ip->src[0].fpreg;
   int rt = ip->src[1].preg;
   int old_fd = ip->src[2].fpreg;
   double contents_fs = FPR_D(fs);
   Reg contents_rt = GPR(rt);
   double contents_fd = FPR_S(old_fd);

   if (!ENABLE_MIPS4(st)) {
      int inum = ip - st->iwin;
      EX_RECORD_EXCEPTION(st, inum, EXC_II, 0); 
      return C1_FAILURE;
   }

   if (contents_rt != 0) {
      FPR_D(fd) = contents_fs;
      SIM_DEBUG(('f', "C1:MOVN.D: r%d %lld, f%d %e to f%d %e\n", 
                 rt, (uint64) contents_rt, 
                 fs, contents_fs, fd, FPR_D(fd)));
   } else {
      FPR_D(fd) = contents_fd;
      SIM_DEBUG(('f', "C1:MOVN.D: r%d %lld, f%d %e unchanged\n", 
                 rt, (int64) contents_rt, fd, FPR_D(fd)));
   }
   return C1_SUCCESS;
}

ID(movz_d) 
{ 
   ip->dst[0].flags = REG_FPD; ip->dst[0].reg = FDRI(ip->instr);
   ip->src[0].flags = REG_FPD; ip->src[0].reg = FSRI(ip->instr);
   ip->src[1].flags = REG_INT; ip->src[1].reg = RTRI(ip->instr);
   ip->src[2].flags = REG_FPD; ip->src[1].reg = FDRI(ip->instr);
}
COP1EX(movz_d)
{
   int fd = ip->dst[0].fpreg;
   int fs = ip->src[0].fpreg;
   int rt = ip->src[1].preg;
   int old_fd = ip->src[2].fpreg;
   double contents_fs = FPR_D(fs);
   Reg contents_rt = GPR(rt);
   double contents_fd = FPR_S(old_fd);

   if (!ENABLE_MIPS4(st)) {
      int inum = ip - st->iwin;
      EX_RECORD_EXCEPTION(st, inum, EXC_II, 0); 
      return C1_FAILURE;
   }

   if (contents_rt == 0) {
      FPR_D(fd) = contents_fs;
      SIM_DEBUG(('f', "C1:MOVZ.D: r%d %lld, f%d %e to f%d %e\n", 
                 rt, (int64) contents_rt, 
                 fs, contents_fs, fd, FPR_D(fd)));
   } else {
      FPR_D(fd) = contents_fd;
      SIM_DEBUG(('f', "C1:MOVZ.D: r%d %lld, f%d %e unchanged\n", 
                 rt, (uint64) contents_rt, fd, FPR_D(fd)));
   }
   return C1_SUCCESS;
}

ID(mul_d) {
   ip->dst[0].flags = REG_FPD; ip->dst[0].reg = FDRI(ip->instr);
   ip->src[0].flags = REG_FPD; ip->src[0].reg = FSRI(ip->instr);
   ip->src[1].flags = REG_FPD; ip->src[1].reg = FTRI(ip->instr);
}
COP1EX(mul_d)
{
   int fd = ip->dst[0].fpreg;
   int fs = ip->src[0].fpreg;
   int ft = ip->src[1].fpreg;
   double contents_fs = FPR_D(fs);
   double contents_ft = FPR_D(ft);
   FPR_D(fd) = contents_fs * contents_ft;
   SIM_DEBUG(('f', "C1:MUL.D: f%d %e * f%d %e = f%d %e\n", 
              fs, contents_fs, ft, contents_ft, fd, FPR_D(fd)));
   return C1_SUCCESS;
}

ID(neg_d) {
   ip->dst[0].flags = REG_FPD; ip->dst[0].reg = FDRI(ip->instr);
   ip->src[0].flags = REG_FPD; ip->src[0].reg = FSRI(ip->instr);
}
COP1EX(neg_d)
{
   int fd = ip->dst[0].fpreg;
   int fs = ip->src[0].fpreg;
   double contents_fs = FPR_D(fs);
   FPR_D(fd) = -contents_fs;
   SIM_DEBUG(('f', "C1:NEG.D: f%d %e --> f%d %e\n", 
              fs, contents_fs, fd, FPR_D(fd)));
   return C1_SUCCESS;
}

ID(sub_d) {
   ip->dst[0].flags = REG_FPD; ip->dst[0].reg = FDRI(ip->instr);
   ip->src[0].flags = REG_FPD; ip->src[0].reg = FSRI(ip->instr);
   ip->src[1].flags = REG_FPD; ip->src[1].reg = FTRI(ip->instr);
}
COP1EX(sub_d)
{
   int fd = ip->dst[0].fpreg;
   int fs = ip->src[0].fpreg;
   int ft = ip->src[1].fpreg;
   double contents_fs = FPR_D(fs);
   double contents_ft = FPR_D(ft);
   FPR_D(fd) = contents_fs - contents_ft;
   SIM_DEBUG(('f', "C1:SUB.D: f%d %e - f%d %e = f%d %e\n", 
              fs, contents_fs, ft, contents_ft, fd, FPR_D(fd)));
   return C1_SUCCESS;
}

ID(sqrt_d) {
   ip->dst[0].flags = REG_FPD; ip->dst[0].reg = FDRI(ip->instr);
   ip->src[0].flags = REG_FPD; ip->src[0].reg = FSRI(ip->instr);
}
COP1EX(sqrt_d)
{
   int fd = ip->dst[0].fpreg;
   int fs = ip->src[0].fpreg;
   double contents_fs = FPR_D(fs);
   FPR_D(fd) = sqrt(contents_fs);
   SIM_DEBUG(('f', "C1:SQRT.D: sqrt(f%d %e) = f%d %e\n", 
              fs, contents_fs, fd, FPR_D(fd)));
   return C1_SUCCESS;
}

ID(trunc_w_d) {
   ip->dst[0].flags = REG_FPW; ip->dst[0].reg = FDRI(ip->instr);
   ip->src[0].flags = REG_FPD; ip->src[0].reg = FSRI(ip->instr);
}
COP1EX(trunc_w_d)
{
   int fd = ip->dst[0].fpreg;
   int fs = ip->src[0].fpreg;
   double contents_fs = FPR_D(fs);
   FPR_W(fd) = contents_fs;
   SIM_DEBUG(('f', "C1:TRUNC.W.D: f%d %e --> f%d %d\n", 
              fs, contents_fs, fd, FPR_W(fd)));
   return C1_SUCCESS;
}

ID(trunc_l_d) {
   ip->dst[0].flags = REG_FPL; ip->dst[0].reg = FDRI(ip->instr);
   ip->src[0].flags = REG_FPD; ip->src[0].reg = FSRI(ip->instr);
}
COP1EX(trunc_l_d)
{
   int fd = ip->dst[0].fpreg;
   int fs = ip->src[0].fpreg;
   double contents_fs = FPR_D(fs);
   FPR_L(fd) = contents_fs;
   SIM_DEBUG(('f', "C1:TRUNC.L.D: f%d %e --> f%d %lld\n", 
              fs, contents_fs, fd, FPR_L(fd)));
   return C1_SUCCESS;
}

ID(floor_w_d) {
   ip->dst[0].flags = REG_FPW; ip->dst[0].reg = FDRI(ip->instr);
   ip->src[0].flags = REG_FPD; ip->src[0].reg = FSRI(ip->instr);
}
COP1EX(floor_w_d)
{
   int fd = ip->dst[0].fpreg;
   int fs = ip->src[0].fpreg;
   double contents_fs = FPR_D(fs);
   FPR_W(fd) = (contents_fs >= 0.0 ? 
                       contents_fs : 
                       contents_fs - SP_NEAR_ONE);
   SIM_DEBUG(('f', "C1:FLOOR.W.D: f%d %e --> f%d %d\n", 
              fs, contents_fs, fd, FPR_W(fd)));
   return C1_SUCCESS;
}

ID(floor_l_d) {
   ip->dst[0].flags = REG_FPL; ip->dst[0].reg = FDRI(ip->instr);
   ip->src[0].flags = REG_FPD; ip->src[0].reg = FSRI(ip->instr);
}
COP1EX(floor_l_d)
{
   int fd = ip->dst[0].fpreg;
   int fs = ip->src[0].fpreg;
   double contents_fs = FPR_D(fs);
   FPR_L(fd) = (contents_fs >= 0.0 ? 
                            contents_fs : 
                            contents_fs - SP_NEAR_ONE);
   SIM_DEBUG(('f', "C1:FLOOR.L.D: f%d %e --> f%d %lld\n", 
              fs, contents_fs, fd, FPR_L(fd)));
   return C1_SUCCESS;
}

ID(recip_d) {
   ip->dst[0].flags = REG_FPD; ip->dst[0].reg = FDRI(ip->instr);
   ip->src[0].flags = REG_FPD; ip->src[0].reg = FSRI(ip->instr);
}
COP1EX(recip_d)
{
   int fd = ip->dst[0].fpreg;
   int fs = ip->src[0].fpreg;
   double contents_fs = FPR_D(fs);
   if (!ENABLE_MIPS4(st)) {
      int inum = ip - st->iwin;
      EX_RECORD_EXCEPTION(st, inum, EXC_II, 0); 
      return C1_FAILURE;
   }
   if (contents_fs == 0.0) {
      CPUWarning("C1:RECIP.D: divide by zero\n");
   }
   FPR_D(fd) = 1.0 / contents_fs;
   SIM_DEBUG(('f', "C1:RECIP.D: 1.0 / f%d %e = f%d %e\n", 
              fs, contents_fs, fd, FPR_D(fd)));
   return C1_SUCCESS;
}

ID(rsqrt_d) {
   ip->dst[0].flags = REG_FPD; ip->dst[0].reg = FDRI(ip->instr);
   ip->src[0].flags = REG_FPD; ip->src[0].reg = FSRI(ip->instr);
}
COP1EX(rsqrt_d)
{
   int fd = ip->dst[0].fpreg;
   int fs = ip->src[0].fpreg;
   double contents_fs = FPR_D(fs);
   if (!ENABLE_MIPS4(st)) {
      int inum = ip - st->iwin;
      EX_RECORD_EXCEPTION(st, inum, EXC_II, 0); 
      return C1_FAILURE;
   }
   if (contents_fs == 0.0) {
      CPUWarning("C1:RSQRT.D: divide by zero\n");
   }
   FPR_D(fd) = 1.0 / sqrt(contents_fs);
   SIM_DEBUG(('f', "C1:RSQRT.D: 1.0 / sqrt(f%d %e) = f%d %e\n", 
              fs, contents_fs, fd, FPR_D(fd)));
   return C1_SUCCESS;
}

ID(round_w_d) {
   ip->dst[0].flags = REG_FPW; ip->dst[0].reg = FDRI(ip->instr);
   ip->src[0].flags = REG_FPD; ip->src[0].reg = FSRI(ip->instr);
}
COP1EX(round_w_d)
{
   int fd = ip->dst[0].fpreg;
   int fs = ip->src[0].fpreg;
   double contents_fs = FPR_D(fs);
   FPR_W(fd) = (contents_fs > 0.0 ? 
                       contents_fs + 0.5 : 
                       contents_fs - 0.5);
   SIM_DEBUG(('f', "C1:ROUND.W.D: f%d %e --> f%d %d\n", 
              fs, contents_fs, fd, FPR_W(fd)));
   return C1_SUCCESS;
}

ID(round_l_d) {
   ip->dst[0].flags = REG_FPL; ip->dst[0].reg = FDRI(ip->instr);
   ip->src[0].flags = REG_FPD; ip->src[0].reg = FSRI(ip->instr);
}
COP1EX(round_l_d)
{
   int fd = ip->dst[0].fpreg;
   int fs = ip->src[0].fpreg;
   double contents_fs = FPR_D(fs);
   FPR_L(fd) = (contents_fs > 0.0 ? 
                            contents_fs + 0.5 : 
                            contents_fs - 0.5);
   SIM_DEBUG(('f', "C1:ROUND.W.D: f%d %e --> f%d %lld\n", 
              fs, contents_fs, fd, FPR_L(fd)));
   return C1_SUCCESS;
}

ID(ceil_w_d) {
   ip->dst[0].flags = REG_FPW; ip->dst[0].reg = FDRI(ip->instr);
   ip->src[0].flags = REG_FPD; ip->src[0].reg = FSRI(ip->instr);
}
COP1EX(ceil_w_d)
{
   int fd = ip->dst[0].fpreg;
   int fs = ip->src[0].fpreg;
   double contents_fs = FPR_D(fs);
   FPR_W(fd) = (contents_fs > 0.0 ? 
                       contents_fs + SP_NEAR_ONE : 
                       contents_fs);
   SIM_DEBUG(('f', "C1:CEIL.W.D: f%d %e --> f%d %d\n", 
              fs, contents_fs, fd, FPR_W(fd)));
   return C1_SUCCESS;
}

ID(ceil_l_d) {
   ip->dst[0].flags = REG_FPL; ip->dst[0].reg = FDRI(ip->instr);
   ip->src[0].flags = REG_FPD; ip->src[0].reg = FSRI(ip->instr);
}
COP1EX(ceil_l_d)
{
   int fd = ip->dst[0].fpreg;
   int fs = ip->src[0].fpreg;
   double contents_fs = FPR_D(fs);
   FPR_L(fd) = (contents_fs > 0.0 ? 
                            contents_fs + SP_NEAR_ONE : 
                            contents_fs);
   SIM_DEBUG(('f', "C1:CEIL.L.D: f%d %e --> f%d %lld\n", 
              fs, contents_fs, fd, FPR_L(fd)));
   return C1_SUCCESS;
}

/* Common decode routine for all c_cond_d opcodes */
ID(c_cond_d)
{
   ip->dst[0].flags = REG_INT; ip->dst[0].reg = MXS_REG_FPCSR;
   ip->src[0].flags = REG_FPD; ip->src[0].reg = FSRI(ip->instr);
   ip->src[1].flags = REG_FPD; ip->src[1].reg = FTRI(ip->instr);
}

ID(c_eq_d) { c_cond_dID(ip, pc); }
COP1EX(c_eq_d)
{
   CPUState *P = (CPUState *)(st->mipsyPtr);
   int fs = ip->src[0].fpreg;
   int ft = ip->src[1].fpreg;
   double contents_fs = FPR_D(fs);
   double contents_ft = FPR_D(ft);
   int cond = (contents_fs == contents_ft);
   int fcc = C_FCC(ip->instr);
   int fpcsr = ip->dst[0].preg;
   SIM_DEBUG(('f', "C1:C.EQ.D: f%d %e, f%d %e --> fcc%d %d\n",
              fs, contents_fs, ft, contents_ft, fcc, cond));
   return CompareOp(st, cond, fcc, fpcsr);
}

ID(c_f_d) { c_cond_dID(ip, pc); }
COP1EX(c_f_d)
{
   CPUState *P = (CPUState *)(st->mipsyPtr);
   int fs = ip->src[0].fpreg;
   int ft = ip->src[1].fpreg;
   double contents_fs = FPR_D(fs);
   double contents_ft = FPR_D(ft);
   int cond = 0;
   int fcc = C_FCC(ip->instr);
   int fpcsr = ip->dst[0].preg;
   SIM_DEBUG(('f', "C1:C.F.D: f%d %e, f%d %e --> fcc%d %d\n",
              fs, contents_fs, ft, contents_ft, fcc, cond));
   return CompareOp(st, cond, fcc, fpcsr);
}

ID(c_le_d) { c_cond_dID(ip, pc); }
COP1EX(c_le_d)
{
   CPUState *P = (CPUState *)(st->mipsyPtr);
   int fs = ip->src[0].fpreg;
   int ft = ip->src[1].fpreg;
   double contents_fs = FPR_D(fs);
   double contents_ft = FPR_D(ft);
   int cond = (contents_fs <= contents_ft);
   int fcc = C_FCC(ip->instr);
   int fpcsr = ip->dst[0].preg;
   SIM_DEBUG(('f', "C1:C.LE.D: f%d %e, f%d %e --> fcc%d %d\n",
              fs, contents_fs, ft, contents_ft, fcc, cond));
   return CompareOp(st, cond, fcc, fpcsr);
}

ID(c_lt_d) { c_cond_dID(ip, pc); }
COP1EX(c_lt_d)
{
   CPUState *P = (CPUState *)(st->mipsyPtr);
   int fs = ip->src[0].fpreg;
   int ft = ip->src[1].fpreg;
   double contents_fs = FPR_D(fs);
   double contents_ft = FPR_D(ft);
   int cond = (contents_fs < contents_ft);
   int fcc = C_FCC(ip->instr);
   int fpcsr = ip->dst[0].preg;
   SIM_DEBUG(('f', "C1:C.LT.D: f%d %e, f%d %e --> fcc%d %d\n",
              fs, contents_fs, ft, contents_ft, fcc, cond));
   return CompareOp(st, cond, fcc, fpcsr);
}

ID(c_nge_d) { c_cond_dID(ip, pc); }
COP1EX(c_nge_d)
{
   CPUState *P = (CPUState *)(st->mipsyPtr);
   int fs = ip->src[0].fpreg;
   int ft = ip->src[1].fpreg;
   double contents_fs = FPR_D(fs);
   double contents_ft = FPR_D(ft);
   int cond = (contents_fs <= contents_ft);
   int fcc = C_FCC(ip->instr);
   int fpcsr = ip->dst[0].preg;
   SIM_DEBUG(('f', "C1:C.NGE.D: f%d %e, f%d %e --> fcc%d %d\n",
              fs, contents_fs, ft, contents_ft, fcc, cond));
   return CompareOp(st, cond, fcc, fpcsr);
}

ID(c_ngl_d) { c_cond_dID(ip, pc); }
COP1EX(c_ngl_d)
{
   CPUState *P = (CPUState *)(st->mipsyPtr);
   int fs = ip->src[0].fpreg;
   int ft = ip->src[1].fpreg;
   double contents_fs = FPR_D(fs);
   double contents_ft = FPR_D(ft);
   int cond = (contents_fs == contents_ft);
   int fcc = C_FCC(ip->instr);
   int fpcsr = ip->dst[0].preg;
   SIM_DEBUG(('f', "C1:C.NGL.D: f%d %e, f%d %e --> fcc%d %d\n",
              fs, contents_fs, ft, contents_ft, fcc, cond));
   return CompareOp(st, cond, fcc, fpcsr);
}

ID(c_ngle_d) { c_cond_dID(ip, pc); }
COP1EX(c_ngle_d)
{
   CPUState *P = (CPUState *)(st->mipsyPtr);
   int fs = ip->src[0].fpreg;
   int ft = ip->src[1].fpreg;
   double contents_fs = FPR_D(fs);
   double contents_ft = FPR_D(ft);
   int cond = 0;
   int fcc = C_FCC(ip->instr);
   int fpcsr = ip->dst[0].preg;
   SIM_DEBUG(('f', "C1:C.NGLE.D: f%d %e, f%d %e --> fcc%d %d\n",
              fs, contents_fs, ft, contents_ft, fcc, cond));
   return CompareOp(st, cond, fcc, fpcsr);
}

ID(c_ngt_d) { c_cond_dID(ip, pc); }
COP1EX(c_ngt_d)
{
   CPUState *P = (CPUState *)(st->mipsyPtr);
   int fs = ip->src[0].fpreg;
   int ft = ip->src[1].fpreg;
   double contents_fs = FPR_D(fs);
   double contents_ft = FPR_D(ft);
   int cond = (contents_fs <= contents_ft);
   int fcc = C_FCC(ip->instr);
   int fpcsr = ip->dst[0].preg;
   SIM_DEBUG(('f', "C1:C.NGT.D: f%d %e, f%d %e --> fcc%d %d\n",
              fs, contents_fs, ft, contents_ft, fcc, cond));
   return CompareOp(st, cond, fcc, fpcsr);
}

ID(c_ole_d) { c_cond_dID(ip, pc); }
COP1EX(c_ole_d)
{
   CPUState *P = (CPUState *)(st->mipsyPtr);
   int fs = ip->src[0].fpreg;
   int ft = ip->src[1].fpreg;
   double contents_fs = FPR_D(fs);
   double contents_ft = FPR_D(ft);
   int cond = (contents_fs < contents_ft);
   int fcc = C_FCC(ip->instr);
   int fpcsr = ip->dst[0].preg;
   SIM_DEBUG(('f', "C1:C.OLE.D: f%d %e, f%d %e --> fcc%d %d\n",
              fs, contents_fs, ft, contents_ft, fcc, cond));
   return CompareOp(st, cond, fcc, fpcsr);
}

ID(c_olt_d) { c_cond_dID(ip, pc); }
COP1EX(c_olt_d)
{
   CPUState *P = (CPUState *)(st->mipsyPtr);
   int fs = ip->src[0].fpreg;
   int ft = ip->src[1].fpreg;
   double contents_fs = FPR_D(fs);
   double contents_ft = FPR_D(ft);
   int cond = (contents_fs < contents_ft);
   int fcc = C_FCC(ip->instr);
   int fpcsr = ip->dst[0].preg;
   SIM_DEBUG(('f', "C1:C.OLT.D: f%d %e, f%d %e --> fcc%d %d\n",
              fs, contents_fs, ft, contents_ft, fcc, cond));
   return CompareOp(st, cond, fcc, fpcsr);
}

ID(c_seq_d) { c_cond_dID(ip, pc); }
COP1EX(c_seq_d)
{
   CPUState *P = (CPUState *)(st->mipsyPtr);
   int fs = ip->src[0].fpreg;
   int ft = ip->src[1].fpreg;
   double contents_fs = FPR_D(fs);
   double contents_ft = FPR_D(ft);
   int cond = (contents_fs == contents_ft);
   int fcc = C_FCC(ip->instr);
   int fpcsr = ip->dst[0].preg;
   SIM_DEBUG(('f', "C1:C.SEQ.D: f%d %e, f%d %e --> fcc%d %d\n",
              fs, contents_fs, ft, contents_ft, fcc, cond));
   return CompareOp(st, cond, fcc, fpcsr);
}

ID(c_sf_d) { c_cond_dID(ip, pc); }
COP1EX(c_sf_d)
{
   CPUState *P = (CPUState *)(st->mipsyPtr);
   int fs = ip->src[0].fpreg;
   int ft = ip->src[1].fpreg;
   double contents_fs = FPR_D(fs);
   double contents_ft = FPR_D(ft);
   int cond = 0;
   int fcc = C_FCC(ip->instr);
   int fpcsr = ip->dst[0].preg;
   SIM_DEBUG(('f', "C1:C.SF.D: f%d %e, f%d %e --> fcc%d %d\n",
              fs, contents_fs, ft, contents_ft, fcc, cond));
   return CompareOp(st, cond, fcc, fpcsr);
}

ID(c_ueq_d) { c_cond_dID(ip, pc); }
COP1EX(c_ueq_d)
{
   CPUState *P = (CPUState *)(st->mipsyPtr);
   int fs = ip->src[0].fpreg;
   int ft = ip->src[1].fpreg;
   double contents_fs = FPR_D(fs);
   double contents_ft = FPR_D(ft);
   int cond = (contents_fs == contents_ft);
   int fcc = C_FCC(ip->instr);
   int fpcsr = ip->dst[0].preg;
   SIM_DEBUG(('f', "C1:C.UEQ.D: f%d %e, f%d %e --> fcc%d %d\n",
              fs, contents_fs, ft, contents_ft, fcc, cond));
   return CompareOp(st, cond, fcc, fpcsr);
}

ID(c_ule_d) { c_cond_dID(ip, pc); }
COP1EX(c_ule_d)
{
   CPUState *P = (CPUState *)(st->mipsyPtr);
   int fs = ip->src[0].fpreg;
   int ft = ip->src[1].fpreg;
   double contents_fs = FPR_D(fs);
   double contents_ft = FPR_D(ft);
   int cond = (contents_fs <= contents_ft);
   int fcc = C_FCC(ip->instr);
   int fpcsr = ip->dst[0].preg;
   SIM_DEBUG(('f', "C1:C.ULE.D: f%d %e, f%d %e --> fcc%d %d\n",
              fs, contents_fs, ft, contents_ft, fcc, cond));
   return CompareOp(st, cond, fcc, fpcsr);
}

ID(c_ult_d) { c_cond_dID(ip, pc); }
COP1EX(c_ult_d)
{
   CPUState *P = (CPUState *)(st->mipsyPtr);
   int fs = ip->src[0].fpreg;
   int ft = ip->src[1].fpreg;
   double contents_fs = FPR_D(fs);
   double contents_ft = FPR_D(ft);
   int cond = (contents_fs < contents_ft);
   int fcc = C_FCC(ip->instr);
   int fpcsr = ip->dst[0].preg;
   SIM_DEBUG(('f', "C1:C.ULT.D: f%d %e, f%d %e --> fcc%d %d\n",
              fs, contents_fs, ft, contents_ft, fcc, cond));
   return CompareOp(st, cond, fcc, fpcsr);
}

ID(c_un_d) { c_cond_dID(ip, pc); }
COP1EX(c_un_d)
{
   CPUState *P = (CPUState *)(st->mipsyPtr);
   int fs = ip->src[0].fpreg;
   int ft = ip->src[1].fpreg;
   double contents_fs = FPR_D(fs);
   double contents_ft = FPR_D(ft);
   int cond = 0;
   int fcc = C_FCC(ip->instr);
   int fpcsr = ip->dst[0].preg;
   SIM_DEBUG(('f', "C1:C.UN.D: f%d %e, f%d %e --> fcc%d %d\n",
              fs, contents_fs, ft, contents_ft, fcc, cond));
   return CompareOp(st, cond, fcc, fpcsr);
}

/*****************************************************************
 * Fixed-point opcodes
 *****************************************************************/
ID(cvt_s_w) {
   ip->dst[0].flags = REG_FPS; ip->dst[0].reg = FDRI(ip->instr);
   ip->src[0].flags = REG_FPW; ip->src[0].reg = FSRI(ip->instr);
}
COP1EX(cvt_s_w)
{
   int fd = ip->dst[0].fpreg;
   int fs = ip->src[0].fpreg;
   int contents_fs = FPR_W(fs);
   FPR_S(fd) = contents_fs;
   SIM_DEBUG(('f', "C1:CVT.S.W: f%d %d --> f%d %e\n",
              fs, contents_fs, fd, FPR_S(fd)));
   return C1_SUCCESS;
}

ID(cvt_d_w) {
   ip->dst[0].flags = REG_FPD; ip->dst[0].reg = FDRI(ip->instr);
   ip->src[0].flags = REG_FPW; ip->src[0].reg = FSRI(ip->instr);
}
COP1EX(cvt_d_w)
{
   int fd = ip->dst[0].fpreg;
   int fs = ip->src[0].fpreg;
   int contents_fs = FPR_W(fs);
   FPR_D(fd) = contents_fs;
   SIM_DEBUG(('f', "C1:CVT.D.W: f%d %d --> f%d %e\n",
              fs, contents_fs, fd, FPR_D(fd)));
   return C1_SUCCESS;
}

/*****************************************************************
 * Long fixed-point opcodes
 *****************************************************************/
ID(cvt_s_l) {
   ip->dst[0].flags = REG_FPS; ip->dst[0].reg = FDRI(ip->instr);
   ip->src[0].flags = REG_FPL; ip->src[0].reg = FSRI(ip->instr);
}
COP1EX(cvt_s_l)
{
   int fd = ip->dst[0].fpreg;
   int fs = ip->src[0].fpreg;
   int64 contents_fs = FPR_L(fs);
   FPR_S(fd) = contents_fs;
   SIM_DEBUG(('f', "C1:CVT.S.L: f%d %lld --> f%d %e\n",
              fs, contents_fs, fd, FPR_S(fd)));
   return C1_SUCCESS;
}

ID(cvt_d_l) {
   ip->dst[0].flags = REG_FPD; ip->dst[0].reg = FDRI(ip->instr);
   ip->src[0].flags = REG_FPL; ip->src[0].reg = FSRI(ip->instr);
}
COP1EX(cvt_d_l)
{
   int fd = ip->dst[0].fpreg;
   int fs = ip->src[0].fpreg;
   int64 contents_fs = FPR_L(fs);
   FPR_D(fd) = contents_fs;
   SIM_DEBUG(('f', "C1:CVT.D.L: f%d %lld --> f%d %e\n",
              fs, contents_fs, fd, FPR_D(fd)));
   return C1_SUCCESS;
}

static Result
CompareOp(struct s_cpu_state *st, int cond, int fcc, int fpcsr)
{
   if (cond) {
      GPR(fpcsr) |= M_csr_fcc(fcc);
   } else {
      GPR(fpcsr) &= ~M_csr_fcc(fcc);
   }
   return C1_SUCCESS;
}

/*****************************************************************
 * COP1X opcodes -- single-precision floating-point
 *****************************************************************/

ID(madd_s) {
   ip->dst[0].flags = REG_FPS; ip->dst[0].reg = FDRI(ip->instr);
   ip->src[0].flags = REG_FPS; ip->src[0].reg = FSRI(ip->instr);
   ip->src[1].flags = REG_FPS; ip->src[1].reg = FTRI(ip->instr);
   ip->src[2].flags = REG_FPS; ip->src[2].reg = FRRI(ip->instr);
}
COP1EX(madd_s)
{
   int fd = ip->dst[0].fpreg;
   int fs = ip->src[0].fpreg;
   int ft = ip->src[1].fpreg;
   int fr = ip->src[2].fpreg;
   float contents_fr = FPR_S(fr);
   float contents_fs = FPR_S(fs);
   float contents_ft = FPR_S(ft);

   FPR_S(fd) = (contents_fs * contents_ft) + contents_fr;
   SIM_DEBUG(('f', "C1:MADD.S: (f%d %e * f%d %e) + f%d %e = f%d %e\n", 
              fs, contents_fs, ft, contents_ft, fr, contents_fr, 
              fd, FPR_S(fd)));
   return C1_SUCCESS;
}

ID(msub_s) {
   ip->dst[0].flags = REG_FPS; ip->dst[0].reg = FDRI(ip->instr);
   ip->src[0].flags = REG_FPS; ip->src[0].reg = FSRI(ip->instr);
   ip->src[1].flags = REG_FPS; ip->src[1].reg = FTRI(ip->instr);
   ip->src[2].flags = REG_FPS; ip->src[2].reg = FRRI(ip->instr);
}
COP1EX(msub_s)
{
   int fd = ip->dst[0].fpreg;
   int fs = ip->src[0].fpreg;
   int ft = ip->src[1].fpreg;
   int fr = ip->src[2].fpreg;
   float contents_fr = FPR_S(fr);
   float contents_fs = FPR_S(fs);
   float contents_ft = FPR_S(ft);

   FPR_S(fd) = (contents_fs * contents_ft) - contents_fr;
   SIM_DEBUG(('f', "C1:MSUB.S: (f%d %e * f%d %e) - f%d %e = f%d %e\n", 
              fs, contents_fs, ft, contents_ft, fr, contents_fr, 
              fd, FPR_S(fd)));
   return C1_SUCCESS;
}

ID(nmadd_s) {
   ip->dst[0].flags = REG_FPS; ip->dst[0].reg = FDRI(ip->instr);
   ip->src[0].flags = REG_FPS; ip->src[0].reg = FSRI(ip->instr);
   ip->src[1].flags = REG_FPS; ip->src[1].reg = FTRI(ip->instr);
   ip->src[2].flags = REG_FPS; ip->src[2].reg = FRRI(ip->instr);
}
COP1EX(nmadd_s)
{
   int fd = ip->dst[0].fpreg;
   int fs = ip->src[0].fpreg;
   int ft = ip->src[1].fpreg;
   int fr = ip->src[2].fpreg;
   float contents_fr = FPR_S(fr);
   float contents_fs = FPR_S(fs);
   float contents_ft = FPR_S(ft);

   FPR_S(fd) = -((contents_fs * contents_ft) + contents_fr);
   SIM_DEBUG(('f', "C1:NMADD.S: -((f%d %e * f%d %e) + f%d %e) = f%d %e\n", 
              fs, contents_fs, ft, contents_ft, fr, contents_fr, 
              fd, FPR_S(fd)));
   return C1_SUCCESS;
}

ID(nmsub_s) {
   ip->dst[0].flags = REG_FPS; ip->dst[0].reg = FDRI(ip->instr);
   ip->src[0].flags = REG_FPS; ip->src[0].reg = FSRI(ip->instr);
   ip->src[1].flags = REG_FPS; ip->src[1].reg = FTRI(ip->instr);
   ip->src[2].flags = REG_FPS; ip->src[2].reg = FRRI(ip->instr);
}
COP1EX(nmsub_s)
{
   int fd = ip->dst[0].fpreg;
   int fs = ip->src[0].fpreg;
   int ft = ip->src[1].fpreg;
   int fr = ip->src[2].fpreg;
   float contents_fr = FPR_S(fr);
   float contents_fs = FPR_S(fs);
   float contents_ft = FPR_S(ft);

   FPR_S(fd) = -((contents_fs * contents_ft) - contents_fr);
   SIM_DEBUG(('f', "C1:NMSUB.S: -((f%d %e * f%d %e) - f%d %e) = f%d %e\n", 
              fs, contents_fs, ft, contents_ft, fr, contents_fr, 
              fd, FPR_S(fd)));
   return C1_SUCCESS;
}

/*****************************************************************
 * COP1X opcodes -- double-precision floating-point
 *****************************************************************/

ID(madd_d) {
   ip->dst[0].flags = REG_FPD; ip->dst[0].reg = FDRI(ip->instr);
   ip->src[0].flags = REG_FPD; ip->src[0].reg = FSRI(ip->instr);
   ip->src[1].flags = REG_FPD; ip->src[1].reg = FTRI(ip->instr);
   ip->src[2].flags = REG_FPD; ip->src[2].reg = FRRI(ip->instr);
}
COP1EX(madd_d)
{
   int fd = ip->dst[0].fpreg;
   int fs = ip->src[0].fpreg;
   int ft = ip->src[1].fpreg;
   int fr = ip->src[2].fpreg;
   double contents_fr = FPR_D(fr);
   double contents_fs = FPR_D(fs);
   double contents_ft = FPR_D(ft);

   FPR_D(fd) = (contents_fs * contents_ft) + contents_fr;
   SIM_DEBUG(('f', "C1:MADD.D: (f%d %e * f%d %e) + f%d %e = f%d %e\n", 
              fs, contents_fs, ft, contents_ft, fr, contents_fr, 
              fd, FPR_D(fd)));
   return C1_SUCCESS;
}

ID(msub_d) {
   ip->dst[0].flags = REG_FPD; ip->dst[0].reg = FDRI(ip->instr);
   ip->src[0].flags = REG_FPD; ip->src[0].reg = FSRI(ip->instr);
   ip->src[1].flags = REG_FPD; ip->src[1].reg = FTRI(ip->instr);
   ip->src[2].flags = REG_FPD; ip->src[2].reg = FRRI(ip->instr);
}
COP1EX(msub_d)
{
   int fd = ip->dst[0].fpreg;
   int fs = ip->src[0].fpreg;
   int ft = ip->src[1].fpreg;
   int fr = ip->src[2].fpreg;
   double contents_fr = FPR_D(fr);
   double contents_fs = FPR_D(fs);
   double contents_ft = FPR_D(ft);

   FPR_D(fd) = (contents_fs * contents_ft) - contents_fr;
   SIM_DEBUG(('f', "C1:MSUB.D: (f%d %e * f%d %e) - f%d %e = f%d %e\n", 
              fs, contents_fs, ft, contents_ft, fr, contents_fr, 
              fd, FPR_D(fd)));
   return C1_SUCCESS;
}

ID(nmadd_d) {
   ip->dst[0].flags = REG_FPD; ip->dst[0].reg = FDRI(ip->instr);
   ip->src[0].flags = REG_FPD; ip->src[0].reg = FSRI(ip->instr);
   ip->src[1].flags = REG_FPD; ip->src[1].reg = FTRI(ip->instr);
   ip->src[2].flags = REG_FPD; ip->src[2].reg = FRRI(ip->instr);
}
COP1EX(nmadd_d)
{
   int fd = ip->dst[0].fpreg;
   int fs = ip->src[0].fpreg;
   int ft = ip->src[1].fpreg;
   int fr = ip->src[2].fpreg;
   double contents_fr = FPR_D(fr);
   double contents_fs = FPR_D(fs);
   double contents_ft = FPR_D(ft);

   FPR_D(fd) = -((contents_fs * contents_ft) + contents_fr);
   SIM_DEBUG(('f', "C1:NMADD.D: -((f%d %e * f%d %e) + f%d %e) = f%d %e\n", 
              fs, contents_fs, ft, contents_ft, fr, contents_fr, 
              fd, FPR_D(fd)));
   return C1_SUCCESS;
}

ID(nmsub_d) {
   ip->dst[0].flags = REG_FPD; ip->dst[0].reg = FDRI(ip->instr);
   ip->src[0].flags = REG_FPD; ip->src[0].reg = FSRI(ip->instr);
   ip->src[1].flags = REG_FPD; ip->src[1].reg = FTRI(ip->instr);
   ip->src[2].flags = REG_FPD; ip->src[2].reg = FRRI(ip->instr);
}
COP1EX(nmsub_d)
{
   int fd = ip->dst[0].fpreg;
   int fs = ip->src[0].fpreg;
   int ft = ip->src[1].fpreg;
   int fr = ip->src[2].fpreg;
   double contents_fr = FPR_D(fr);
   double contents_fs = FPR_D(fs);
   double contents_ft = FPR_D(ft);

   FPR_D(fd) = -((contents_fs * contents_ft) - contents_fr);
   SIM_DEBUG(('f', "C1:NMSUB.D: -((f%d %e * f%d %e) - f%d %e) = f%d %e\n", 
              fs, contents_fs, ft, contents_ft, fr, contents_fr, 
              fd, FPR_D(fd)));
   return C1_SUCCESS;
}

/*****************************************************************
 * COP1X opcodes -- indexed memory ops -- currently unimplemented
 *****************************************************************/

/*****************************************************************
 * Miscellaneous COP1 opcodes
 *****************************************************************/
ID(bc)
{
   uint likely = FT(ip->instr) & 0x2;
   ip->src[0].flags = REG_INT; ip->src[0].reg = MXS_REG_FPCSR;
   ip->target = (pc + 4) + SIGN_EXTEND32(18, IMMED(ip->instr) << 2);
   ip->flags = INST_BRANCH | INST_CONDITIONAL;
   if (likely) {
      ip->flags |= INST_LIKELY;
   }
}
COP1EX(bc)
{
   int fpcsr = ip->src[0].preg;
   Reg contents_fpcsr = GPR(fpcsr);
   uint br_value = FT(ip->instr);
   uint fcc      = BC_FCC(ip->instr);
   uint likely   = (br_value & 0x2);
   uint fcc_set  = contents_fpcsr & M_csr_fcc(fcc);
   if (br_value & 0x1) { /* BC1T */
      BranchOp(st, ip, th, fcc_set, likely);
      SIM_DEBUG(('f', "C1:BC1T%s: fcc %d = %d, branch %s%s\n",
                 likely ? "L" : "", fcc, (fcc_set != 0),
                 fcc_set ? "taken" : "not taken"));
   } else { /* BC1F */
      BranchOp(st, ip, th, !fcc_set, likely);
      SIM_DEBUG(('f', "C1:BC1F%s: fcc %d = %d, branch %s%s\n",
                 likely ? "L" : "", fcc, (fcc_set != 0),
                 !fcc_set ? "taken" : "not taken"));
   }
   return C1_SUCCESS;
}

ID(cfc)
{
   ip->dst[0].flags = REG_INT; ip->dst[0].reg = RTRI(ip->instr);
   if (FS(ip->instr) & 0x1) {
      ip->src[0].flags = REG_INT; ip->src[0].reg = MXS_REG_FPCSR;
   } else {
      ip->src[0].flags = REG_INT; ip->src[0].reg = MXS_REG_FPIRR;
   }
   ip->flags = INST_FPCTL;
}
COP1EX(cfc)
{
   int rt = ip->dst[0].preg;
   int rd = ip->src[0].preg;
   int inum = ip - st->iwin;

   if (!AllsQuiet (st, ip, ST_FPC))
      return C1_SUCCESS;
   /* Need to get the state of the real register, not the renamed one */
   if (ip->src[0].reg == MXS_REG_FPCSR) {
      CPUState *P = (CPUState*)(st->mipsyPtr);
      GPR(rt) = P->FCR[31];
   } else {
      GPR(rt) = (3 << 8); /* R3010 FPU */
   }
   SIM_DEBUG(('f', "C1: CFC1 got %#x from fcr %d\n", GPR(rt), rd));
   ClearFPCStall(st, th, inum);
   return C1_SUCCESS;
}

ID(ctc)
{
   if (FS(ip->instr) & 0x1) {
      ip->dst[0].flags = REG_INT; ip->dst[0].reg = MXS_REG_FPCSR;
   } else {
      ip->dst[0].flags = REG_INT; ip->dst[0].reg = MXS_REG_FPIRR;
   }
   ip->src[0].flags = REG_INT; ip->src[0].reg = RTRI(ip->instr);
   ip->flags = INST_FPCTL;
}
COP1EX(ctc)
{
   int rs = ip->src[0].preg;
   int rt = ip->dst[0].preg;
   int inum = ip - st->iwin;

   if (!AllsQuiet (st, ip, ST_FPC))
      return C1_SUCCESS;
   if (ip->dst[0].reg == MXS_REG_FPCSR) {
      GPR(rt) = GPR(rs);
   }
   SIM_DEBUG(('f', "C1: CTC1 put %#x into fcr %d\n", GPR(rs), rt));
   ClearFPCStall(st, th, inum);
   return C1_SUCCESS;
}

ID(mfc)
{
   ip->dst[0].flags = REG_INT; ip->dst[0].reg = RTRI(ip->instr);
   ip->src[0].flags = REG_FPW; ip->src[0].reg = FSRI(ip->instr);
}
COP1EX(mfc)
{
   int rt = ip->dst[0].preg;
   int fs = ip->src[0].fpreg;
   GPR(rt) = (Reg32_s)FPR_W(fs);
   SIM_DEBUG(('f', "C1: mfc1 put f%d %d into r%d \n", fs, FPR_W(fs), rt));
   return C1_SUCCESS;
}

ID(mtc)
{
   ip->dst[0].flags = REG_FPW; ip->dst[0].reg = FSRI(ip->instr);
   ip->src[0].flags = REG_INT; ip->src[0].reg = RTRI(ip->instr);
   /* Need the other half as a source/dest if we're using register pairs */
#ifdef rpb_notdef
   ip->dst[1].flags = REG_FPW; ip->dst[1].reg = REG_PAIR(FSRI(ip->instr)); 
#endif
   ip->src[1].flags = REG_FPW; ip->src[1].reg = REG_PAIR(FSRI(ip->instr)); 
   ip->flags |= INST_HALF_DEF;
}
COP1EX(mtc)
{
   int fs = ip->dst[0].fpreg;
   int rt = ip->src[0].preg;
   if ((ip->flags & INST_HALF_DEF) && MXS_NFRB) {
      int pair = REG_PAIR(ip->dst[0].fpreg);
      int old_pair = ip->src[1].fpreg;
      Reg32 contents_pair = FPR_W(old_pair);
      FPR_W(pair) = contents_pair;
      SIM_DEBUG(('f', "C1: mtc1 put pair f%d %d into f%d -> %d\n",
                 old_pair, contents_pair, pair, FPR_W(pair)));
   }
   FPR_W(fs) = (Reg32)GPR(rt);
   SIM_DEBUG(('f', "C1: mtc1 put r%d %d into f%d -> %d\n", rt, GPR(rt), 
              fs, FPR_W(fs)));
   return C1_SUCCESS;
}

ID(dmfc)
{
   ip->dst[0].flags = REG_INT; ip->dst[0].reg = RTRI(ip->instr);
   ip->src[0].flags = REG_FPL; ip->src[0].reg = FSRI(ip->instr);
}
COP1EX(dmfc)
{
   int rt = ip->dst[0].preg;
   int fs = ip->src[0].fpreg;
   if (!ENABLE_64BIT(st)) {
      int inum = ip - st->iwin;
      EX_RECORD_EXCEPTION(st, inum, EXC_II, 0);
      return C1_FAILURE;
   }
   GPR(rt) = FPR_L(fs);
   SIM_DEBUG(('f', "C1: dmfc1 put f%d %lld into r%d \n", 
              fs, FPR_L(fs), rt));
   return C1_SUCCESS;
}

ID(dmtc)
{
   ip->dst[0].flags = REG_FPL; ip->dst[0].reg = FSRI(ip->instr);
   ip->src[0].flags = REG_INT; ip->src[0].reg = RTRI(ip->instr);
}
COP1EX(dmtc)
{
   int fs = ip->dst[0].fpreg;
   int rt = ip->src[0].preg;
   if (!ENABLE_64BIT(st)) {
      int inum = ip - st->iwin;
      EX_RECORD_EXCEPTION(st, inum, EXC_II, 0);
      return C1_FAILURE;
   }
   FPR_L(fs) = GPR(rt);
   SIM_DEBUG(('f', "C1: dmtc1 put r%d %lld into f%d -> %lld\n", 
              rt, GPR(rt), fs, FPR_L(fs)));
   return C1_SUCCESS;
}

ID(movci)
{
   ip->dst[0].flags = REG_INT; ip->dst[0].reg = RDRI(ip->instr);
   ip->src[0].flags = REG_INT; ip->src[0].reg = RSRI(ip->instr);
   ip->src[1].flags = REG_INT; ip->src[1].reg = RDRI(ip->instr);
   ip->src[2].flags = REG_INT; ip->src[2].reg = MXS_REG_FPCSR;
}
COP1EX(movci)
{
   Reg contents_rs    = GPR(ip->src[0].preg);
   Reg contents_rd    = GPR(ip->src[1].preg);
   Reg contents_fpcsr = GPR(ip->src[2].preg);
   int fcc = MOVC_FCC(ip->instr);
   int rd  = ip->dst[0].preg;

   if (TF(ip->instr)) { 
      if (contents_fpcsr & M_csr_fcc(fcc)) {
         GPR(rd) = contents_rs;
      } else {
         GPR(rd) = contents_rd;
      }
   } else {
      if (!(contents_fpcsr & M_csr_fcc(fcc))) {
         GPR(rd) = contents_rs;
      } else {
         GPR(rd) = contents_rd;
      }
   }
   return C1_SUCCESS;
}

/*****************************************************************
 * Catch-all routines for unimplemented FPU opcodes
 *****************************************************************/
ID(unknownC1) { }
COP1EX(unknownC1)
{
   int inum = ip - st->iwin;
   ip->flags = INST_ILLEGAL;
   EX_RECORD_EXCEPTION(st, inum, EXC_II, 0);
   return C1_FAILURE;
}

/*****************************************************************
 * Set the rounding mode of the host processor to match that of the
 * simulated machine. This is needed since we're actually executing the
 * instructions on the host platform. 
 *****************************************************************/
static void 
SetRoundingMode(int modeNum)
{
#if defined(sgi) || defined(sparc)
   fp_rnd newMode = FP_RN; 
   
   switch (modeNum) {
   case 0:
      newMode = FP_RN;
      break;
   case 1:
      newMode = FP_RZ;
      break;
   case 2:
      newMode = FP_RP;
      break;
   case 3:
      newMode = FP_RM;
      break;
   default:
      CPUWarning("Bad rounding mode set in FCR31 (%d)\n", modeNum);
      break;
   }
   fpsetround(newMode);
#endif
#ifdef __alpha
   write_rnd(modeNum);
#endif
#ifdef linux
   {
      static int fpwarned = 0;
      if (!fpwarned) {
         CPUWarning("MIPSY: could not figure out how to fpsetround on x86 \n");
         fpwarned = 1;
      }
   }
#endif
}

/*****************************************************************
 * All fpu (COP1) insts come through here.
 *****************************************************************/
void
MXSDecodeC1Instruction(INST *ip, Reg pc)
{
   uint major_op = MAJOR_OP(ip->instr);
   uint sub_value = FORMAT(ip->instr);
   uint func = FUNC(ip->instr);
   uint fmt = FORMAT(ip->instr);

   if (major_op == cop1_op) { 
      switch(sub_value) {
      case bc_op:
         ip->opcode = &mxsBCOpcode;
         bcID(ip, pc);
         break;
      case cfc_op:
         ip->opcode = &mxsCFCOpcode;
         cfcID(ip, pc);
         break;
      case ctc_op:
         ip->opcode = &mxsCTCOpcode;
         ctcID(ip, pc);
         break;
      case mfc_op: 
         ip->opcode = &mxsMFCOpcode;
         mfcID(ip, pc);
         break;
      case mtc_op:
         ip->opcode = &mxsMTCOpcode;
         mtcID(ip, pc);
         break;
      case dmfc_op: 
         ip->opcode = &mxsDMFCOpcode;
         dmfcID(ip, pc);
         break;
      case dmtc_op:
         ip->opcode = &mxsDMTCOpcode;
         dmtcID(ip, pc);
         break;
      default: 
         /* S, D, W, L format instructions */
         switch (fmt) {
         case FMT_S:
            ip->opcode = &mxsC1SPOpcodes[func];
            mxsC1SPOpcodes[func].decode(ip, pc);
            break;
         case FMT_D:
            ip->opcode = &mxsC1DPOpcodes[func];
            mxsC1DPOpcodes[func].decode(ip, pc);
            break;
         case FMT_W:
            /* Only W-format opcodes are cvt.s.w and cvt.d.w */
            if (func == fcvts_op) {
               ip->opcode = &mxsCVTSWOpcode;
               cvt_s_wID(ip, pc);
            } else if (func == fcvtd_op) {
               ip->opcode = &mxsCVTDWOpcode;
               cvt_d_wID(ip, pc);
            } else {
               return;
            }
            break;
         case FMT_L:
            /* Only L-format opcodes are cvt.s.l and cvt.d.l */
            if (func == fcvts_op) {
               ip->opcode = &mxsCVTSLOpcode;
               cvt_s_lID(ip, pc);
            } else if (func == fcvtd_op) {
               ip->opcode = &mxsCVTDLOpcode;
               cvt_d_lID(ip, pc);
            } else {
               return;
            }
            break;
         default:
            return;
         } /* switch format */
         break;
      } /* switch subvalue */
   } else if (major_op == cop1x_op) {
      mxsC1XOpcodes[func].decode(ip, pc);
   } else if (major_op == spec_op) { 
      /* Must be on the MOVC (move conditional on FP condition) */
      ASSERT(FUNC(ip->instr) == movc_op);
      movciID(ip, pc);
   } else {
      return;
   }
}

int
MXSExecuteC1Instruction(struct s_cpu_state *st, INST *ip, THREAD *th)
{
   uint fmt = FORMAT(ip->instr);
   uint sub_value = fmt;
   uint major_op;
   Result result = C1_SUCCESS;
   uint func = FUNC(ip->instr);
   CPUState *P = (CPUState *)(st->mipsyPtr);
   int inum = ip - st->iwin;

   if (!ENABLE_FPU(st)) {
      EX_RECORD_EXCEPTION(st, inum, EXC_CPU, 1);
      return C1_FAILURE;
   }

   SetRoundingMode(ROUND_MODE(st));

   major_op = MAJOR_OP(ip->instr);
   if (major_op == cop1_op) { 
      switch(sub_value) {
      case bc_op:
         result = bcEX(st, ip, th);
         break;
      case cfc_op:
         result = cfcEX(st, ip, th);
         break;
      case ctc_op:
         result = ctcEX(st, ip, th);
         break;
      case mfc_op: 
         result = mfcEX(st, ip, th);
         break;
      case mtc_op:
         result = mtcEX(st, ip, th);
         break;
      case dmfc_op: 
         result = dmfcEX(st, ip, th);
         break;
      case dmtc_op:
         result = dmtcEX(st, ip, th);
         break;
      default: 
         /* S, D, W format instructions */
         switch (fmt) {
         case FMT_S:
            result = mxsC1SPOpcodes[func].execute(st, ip, th);
            break;
         case FMT_D:
            result = mxsC1DPOpcodes[func].execute(st, ip, th);
            break;
         case FMT_W:
            /* Only W-format opcodes are cvt.s.w and cvt.d.w */
            if (func == fcvts_op) {
               result = cvt_s_wEX(st, ip, th);
            } else if (func == fcvtd_op) {
               result = cvt_d_wEX(st, ip, th);
            } else {
               EX_RECORD_EXCEPTION(st, inum, EXC_II, 0);
               return C1_FAILURE;
            }
            break;
         case FMT_L:
            /* Only L-format opcodes are cvt.s.l and cvt.d.l */
            if (func == fcvts_op) {
               result = cvt_s_lEX(st, ip, th);
            } else if (func == fcvtd_op) {
               result = cvt_d_lEX(st, ip, th);
            } else {
               EX_RECORD_EXCEPTION(st, inum, EXC_II, 0);
               return C1_FAILURE;
            }
            break;
         default:
            EX_RECORD_EXCEPTION(st, inum, EXC_II, 0);
            return C1_FAILURE;
         }
         CPUVec.IncC0PerfCtr(P->myNum, 21); /* FP instructions graduated */
         break; /* switch format */
      } /* switch subvalue */
   } else if (major_op == cop1x_op) {
      if (!ENABLE_MIPS4(st)) {
         EX_RECORD_EXCEPTION(st, inum, EXC_II, 0); 
         return C1_FAILURE;
      }
      result = mxsC1XOpcodes[func].execute(st, ip, th);
      CPUVec.IncC0PerfCtr(P->myNum, 21); /* FP instructions graduated */
   } else if (major_op == spec_op) { 
      /* Must be on the MOVC (move conditional on FP condition) */
      ASSERT(FUNC(ip->instr) == movc_op);
      result = movciEX(st, ip, th);
   } else {
      EX_RECORD_EXCEPTION(st, inum, EXC_II, 0); 
      return C1_FAILURE;
   }
   return result;
}

MXSOpcodeDesc mxsC1SPOpcodes[] = {
   { "add.s",     add_sID,     add_sEX,     0x00, MXS_FP_UNIT,  2,  1 },
   { "sub.s",     sub_sID,     sub_sEX,     0x01, MXS_FP_UNIT,  2,  1 },
   { "mul.s",     mul_sID,     mul_sEX,     0x02, MXS_FP_UNIT,  2,  1 },
   { "div.s",     div_sID,     div_sEX,     0x03, MXS_FP_UNIT, 12,  1 },
   { "sqrt.s",    sqrt_sID,    sqrt_sEX,    0x04, MXS_FP_UNIT, 18,  1 },
   { "abs.s",     abs_sID,     abs_sEX,     0x05, MXS_FP_UNIT,  2,  1 },
   { "mov.s",     mov_sID,     mov_sEX,     0x06, MXS_FP_UNIT,  2,  1 },
   { "neg.s",     neg_sID,     neg_sEX,     0x07, MXS_FP_UNIT,  2,  1 },
   { "round.l.s", round_l_sID, round_l_sEX, 0x08, MXS_FP_UNIT,  2,  1 },
   { "trunc.l.s", trunc_l_sID, trunc_l_sEX, 0x09, MXS_FP_UNIT,  2,  1 },
   { "ceil.l.s",  ceil_l_sID,  ceil_l_sEX,  0x0a, MXS_FP_UNIT,  2,  1 },
   { "floor.l.s", floor_l_sID, floor_l_sEX, 0x0b, MXS_FP_UNIT,  2,  1 },
   { "round.w.s", round_w_sID, round_w_sEX, 0x0c, MXS_FP_UNIT,  2,  1 },
   { "trunc.w.s", trunc_w_sID, trunc_w_sEX, 0x0d, MXS_FP_UNIT,  2,  1 },
   { "ceil.w.s",  ceil_w_sID,  ceil_w_sEX,  0x0e, MXS_FP_UNIT,  2,  1 },
   { "floor.w.s", floor_w_sID, floor_w_sEX, 0x0f, MXS_FP_UNIT,  2,  1 },
   { "unknownC1", unknownC1ID, unknownC1EX, 0x10, MXS_FP_UNIT,  2,  1 },
   { "movcf.s",   movcf_sID,   movcf_sEX,   0x11, MXS_FP_UNIT,  2,  1 },
   { "movz.s",    movz_sID,    movz_sEX,    0x12, MXS_FP_UNIT,  2,  1 },
   { "movn.s",    movn_sID,    movn_sEX,    0x13, MXS_FP_UNIT,  2,  1 },
   { "unknownC1", unknownC1ID, unknownC1EX, 0x14, MXS_FP_UNIT,  2,  1 },
   { "recip.s",   recip_sID,   recip_sEX,   0x15, MXS_FP_UNIT, 12,  1 },
   { "rsqrt.s",   rsqrt_sID,   rsqrt_sEX,   0x16, MXS_FP_UNIT, 30,  1 },
   { "unknownC1", unknownC1ID, unknownC1EX, 0x17, MXS_FP_UNIT,  2,  1 },
   { "unknownC1", unknownC1ID, unknownC1EX, 0x18, MXS_FP_UNIT,  2,  1 },
   { "unknownC1", unknownC1ID, unknownC1EX, 0x19, MXS_FP_UNIT,  2,  1 },
   { "unknownC1", unknownC1ID, unknownC1EX, 0x1a, MXS_FP_UNIT,  2,  1 },
   { "unknownC1", unknownC1ID, unknownC1EX, 0x1b, MXS_FP_UNIT,  2,  1 },
   { "unknownC1", unknownC1ID, unknownC1EX, 0x1c, MXS_FP_UNIT,  2,  1 },
   { "unknownC1", unknownC1ID, unknownC1EX, 0x1d, MXS_FP_UNIT,  2,  1 },
   { "unknownC1", unknownC1ID, unknownC1EX, 0x1e, MXS_FP_UNIT,  2,  1 },
   { "unknownC1", unknownC1ID, unknownC1EX, 0x1f, MXS_FP_UNIT,  2,  1 },
   { "unknownC1", unknownC1ID, unknownC1EX, 0x20, MXS_FP_UNIT,  2,  1 },
   { "cvt.d.s",   cvt_d_sID,   cvt_d_sEX,   0x21, MXS_FP_UNIT,  2,  1 },
   { "unknownC1", unknownC1ID, unknownC1EX, 0x22, MXS_FP_UNIT,  2,  1 },
   { "unknownC1", unknownC1ID, unknownC1EX, 0x23, MXS_FP_UNIT,  2,  1 },
   { "cvt.w.s",   cvt_w_sID,   cvt_w_sEX,   0x24, MXS_FP_UNIT,  2,  1 },
   { "cvt.l.s",   cvt_l_sID,   cvt_l_sEX,   0x25, MXS_FP_UNIT,  2,  1 },
   { "unknownC1", unknownC1ID, unknownC1EX, 0x26, MXS_FP_UNIT,  2,  1 },
   { "unknownC1", unknownC1ID, unknownC1EX, 0x27, MXS_FP_UNIT,  2,  1 },
   { "unknownC1", unknownC1ID, unknownC1EX, 0x28, MXS_FP_UNIT,  2,  1 },
   { "unknownC1", unknownC1ID, unknownC1EX, 0x29, MXS_FP_UNIT,  2,  1 },
   { "unknownC1", unknownC1ID, unknownC1EX, 0x2a, MXS_FP_UNIT,  2,  1 },
   { "unknownC1", unknownC1ID, unknownC1EX, 0x2b, MXS_FP_UNIT,  2,  1 },
   { "unknownC1", unknownC1ID, unknownC1EX, 0x2c, MXS_FP_UNIT,  2,  1 },
   { "unknownC1", unknownC1ID, unknownC1EX, 0x2d, MXS_FP_UNIT,  2,  1 },
   { "unknownC1", unknownC1ID, unknownC1EX, 0x2e, MXS_FP_UNIT,  2,  1 },
   { "unknownC1", unknownC1ID, unknownC1EX, 0x2f, MXS_FP_UNIT,  2,  1 },
   { "c.f.s",     c_f_sID,     c_f_sEX,     0x30, MXS_FP_UNIT,  2,  1 },
   { "c.un.s",    c_un_sID,    c_un_sEX,    0x31, MXS_FP_UNIT,  2,  1 },
   { "c.eq.s",    c_eq_sID,    c_eq_sEX,    0x32, MXS_FP_UNIT,  2,  1 },
   { "c.ueq.s",   c_ueq_sID,   c_ueq_sEX,   0x33, MXS_FP_UNIT,  2,  1 },
   { "c.olt.s",   c_olt_sID,   c_olt_sEX,   0x34, MXS_FP_UNIT,  2,  1 },
   { "c.ult.s",   c_ult_sID,   c_ult_sEX,   0x35, MXS_FP_UNIT,  2,  1 },
   { "c.ole.s",   c_ole_sID,   c_ole_sEX,   0x36, MXS_FP_UNIT,  2,  1 },
   { "c.ule.s",   c_ule_sID,   c_ule_sEX,   0x37, MXS_FP_UNIT,  2,  1 },
   { "c.sf.s",    c_sf_sID,    c_sf_sEX,    0x38, MXS_FP_UNIT,  2,  1 },
   { "c.ngle.s",  c_ngle_sID,  c_ngle_sEX,  0x39, MXS_FP_UNIT,  2,  1 },
   { "c.seq.s",   c_seq_sID,   c_seq_sEX,   0x3a, MXS_FP_UNIT,  2,  1 },
   { "c.ngl.s",   c_ngl_sID,   c_ngl_sEX,   0x3b, MXS_FP_UNIT,  2,  1 },
   { "c.lt.s",    c_lt_sID,    c_lt_sEX,    0x3c, MXS_FP_UNIT,  2,  1 },
   { "c.nge.s",   c_nge_sID,   c_nge_sEX,   0x3d, MXS_FP_UNIT,  2,  1 },
   { "c.le.s",    c_le_sID,    c_le_sEX,    0x3e, MXS_FP_UNIT,  2,  1 },
   { "c.ngt.s",   c_ngt_sID,   c_ngt_sEX,   0x3f, MXS_FP_UNIT,  2,  1 }
};

MXSOpcodeDesc mxsC1DPOpcodes[] = {
   { "add.d",     add_dID,     add_dEX,     0x00, MXS_FP_UNIT,  2,  1 },
   { "sub.d",     sub_dID,     sub_dEX,     0x01, MXS_FP_UNIT,  2,  1 },
   { "mul.d",     mul_dID,     mul_dEX,     0x02, MXS_FP_UNIT,  2,  1 },
   { "div.d",     div_dID,     div_dEX,     0x03, MXS_FP_UNIT, 19,  1 },
   { "sqrt.d",    sqrt_dID,    sqrt_dEX,    0x04, MXS_FP_UNIT, 33,  1 },
   { "abs.d",     abs_dID,     abs_dEX,     0x05, MXS_FP_UNIT,  2,  1 },
   { "mov.d",     mov_dID,     mov_dEX,     0x06, MXS_FP_UNIT,  2,  1 },
   { "neg.d",     neg_dID,     neg_dEX,     0x07, MXS_FP_UNIT,  2,  1 },
   { "round.l.d", round_l_dID, round_l_dEX, 0x08, MXS_FP_UNIT,  2,  1 },
   { "trunc.l.d", trunc_l_dID, trunc_l_dEX, 0x09, MXS_FP_UNIT,  2,  1 },
   { "ceil.l.d",  ceil_l_dID,  ceil_l_dEX,  0x0a, MXS_FP_UNIT,  2,  1 },
   { "floor.l.d", floor_l_dID, floor_l_dEX, 0x0b, MXS_FP_UNIT,  2,  1 },
   { "round.w.d", round_w_dID, round_w_dEX, 0x0c, MXS_FP_UNIT,  2,  1 },
   { "trunc.w.d", trunc_w_dID, trunc_w_dEX, 0x0d, MXS_FP_UNIT,  2,  1 },
   { "ceil.w.d",  ceil_w_dID,  ceil_w_dEX,  0x0e, MXS_FP_UNIT,  2,  1 },
   { "floor.w.d", floor_w_dID, floor_w_dEX, 0x0f, MXS_FP_UNIT,  2,  1 },
   { "unknownC1", unknownC1ID, unknownC1EX, 0x10, MXS_FP_UNIT,  2,  1 },
   { "movcf.d",   movcf_dID,   movcf_dEX,   0x11, MXS_FP_UNIT,  2,  1 },
   { "movz.d",    movz_dID,    movz_dEX,    0x12, MXS_FP_UNIT,  2,  1 },
   { "movn.d",    movn_dID,    movn_dEX,    0x13, MXS_FP_UNIT,  2,  1 },
   { "unknownC1", unknownC1ID, unknownC1EX, 0x14, MXS_FP_UNIT,  2,  1 },
   { "recip.d",   recip_dID,   recip_dEX,   0x15, MXS_FP_UNIT, 19,  1 },
   { "rsqrt.d",   rsqrt_dID,   rsqrt_dEX,   0x16, MXS_FP_UNIT, 52,  1 },
   { "unknownC1", unknownC1ID, unknownC1EX, 0x17, MXS_FP_UNIT,  2,  1 },
   { "unknownC1", unknownC1ID, unknownC1EX, 0x18, MXS_FP_UNIT,  2,  1 },
   { "unknownC1", unknownC1ID, unknownC1EX, 0x19, MXS_FP_UNIT,  2,  1 },
   { "unknownC1", unknownC1ID, unknownC1EX, 0x1a, MXS_FP_UNIT,  2,  1 },
   { "unknownC1", unknownC1ID, unknownC1EX, 0x1b, MXS_FP_UNIT,  2,  1 },
   { "unknownC1", unknownC1ID, unknownC1EX, 0x1c, MXS_FP_UNIT,  2,  1 },
   { "unknownC1", unknownC1ID, unknownC1EX, 0x1d, MXS_FP_UNIT,  2,  1 },
   { "unknownC1", unknownC1ID, unknownC1EX, 0x1e, MXS_FP_UNIT,  2,  1 },
   { "unknownC1", unknownC1ID, unknownC1EX, 0x1f, MXS_FP_UNIT,  2,  1 },
   { "cvt.s.d",   cvt_s_dID,   cvt_s_dEX,   0x20, MXS_FP_UNIT,  2,  1 },
   { "unknownC1", unknownC1ID, unknownC1EX, 0x21, MXS_FP_UNIT,  2,  1 },
   { "unknownC1", unknownC1ID, unknownC1EX, 0x22, MXS_FP_UNIT,  2,  1 },
   { "unknownC1", unknownC1ID, unknownC1EX, 0x23, MXS_FP_UNIT,  2,  1 },
   { "cvt.w.d",   cvt_w_dID,   cvt_w_dEX,   0x24, MXS_FP_UNIT,  2,  1 },
   { "cvt.l.d",   cvt_l_dID,   cvt_l_dEX,   0x25, MXS_FP_UNIT,  2,  1 },
   { "unknownC1", unknownC1ID, unknownC1EX, 0x26, MXS_FP_UNIT,  2,  1 },
   { "unknownC1", unknownC1ID, unknownC1EX, 0x27, MXS_FP_UNIT,  2,  1 },
   { "unknownC1", unknownC1ID, unknownC1EX, 0x28, MXS_FP_UNIT,  2,  1 },
   { "unknownC1", unknownC1ID, unknownC1EX, 0x29, MXS_FP_UNIT,  2,  1 },
   { "unknownC1", unknownC1ID, unknownC1EX, 0x2a, MXS_FP_UNIT,  2,  1 },
   { "unknownC1", unknownC1ID, unknownC1EX, 0x2b, MXS_FP_UNIT,  2,  1 },
   { "unknownC1", unknownC1ID, unknownC1EX, 0x2c, MXS_FP_UNIT,  2,  1 },
   { "unknownC1", unknownC1ID, unknownC1EX, 0x2d, MXS_FP_UNIT,  2,  1 },
   { "unknownC1", unknownC1ID, unknownC1EX, 0x2e, MXS_FP_UNIT,  2,  1 },
   { "unknownC1", unknownC1ID, unknownC1EX, 0x2f, MXS_FP_UNIT,  2,  1 },
   { "c.f.d",     c_f_dID,     c_f_dEX,     0x30, MXS_FP_UNIT,  2,  1 },
   { "c.un.d",    c_un_dID,    c_un_dEX,    0x31, MXS_FP_UNIT,  2,  1 },
   { "c.eq.d",    c_eq_dID,    c_eq_dEX,    0x32, MXS_FP_UNIT,  2,  1 },
   { "c.ueq.d",   c_ueq_dID,   c_ueq_dEX,   0x33, MXS_FP_UNIT,  2,  1 },
   { "c.olt.d",   c_olt_dID,   c_olt_dEX,   0x34, MXS_FP_UNIT,  2,  1 },
   { "c.ult.d",   c_ult_dID,   c_ult_dEX,   0x35, MXS_FP_UNIT,  2,  1 },
   { "c.ole.d",   c_ole_dID,   c_ole_dEX,   0x36, MXS_FP_UNIT,  2,  1 },
   { "c.ule.d",   c_ule_dID,   c_ule_dEX,   0x37, MXS_FP_UNIT,  2,  1 },
   { "c.df.d",    c_sf_dID,    c_sf_dEX,    0x38, MXS_FP_UNIT,  2,  1 },
   { "c.ngle.d",  c_ngle_dID,  c_ngle_dEX,  0x39, MXS_FP_UNIT,  2,  1 },
   { "c.deq.d",   c_seq_dID,   c_seq_dEX,   0x3a, MXS_FP_UNIT,  2,  1 },
   { "c.ngl.d",   c_ngl_dID,   c_ngl_dEX,   0x3b, MXS_FP_UNIT,  2,  1 },
   { "c.lt.d",    c_lt_dID,    c_lt_dEX,    0x3c, MXS_FP_UNIT,  2,  1 },
   { "c.nge.d",   c_nge_dID,   c_nge_dEX,   0x3d, MXS_FP_UNIT,  2,  1 },
   { "c.le.d",    c_le_dID,    c_le_dEX,    0x3e, MXS_FP_UNIT,  2,  1 },
   { "c.ngt.d",   c_ngt_dID,   c_ngt_dEX,   0x3f, MXS_FP_UNIT,  2,  1 }
};

MXSOpcodeDesc mxsC1XOpcodes[] = {
   { "lwxc1",      unknownC1ID,  unknownC1EX,  0x00, MXS_FP_UNIT,  2,  1 },
   { "ldxc1",      unknownC1ID,  unknownC1EX,  0x01, MXS_FP_UNIT,  2,  1 },
   { "unknownC1",  unknownC1ID,  unknownC1EX,  0x02, MXS_FP_UNIT,  2,  1 },
   { "unknownC1",  unknownC1ID,  unknownC1EX,  0x03, MXS_FP_UNIT,  2,  1 },
   { "unknownC1",  unknownC1ID,  unknownC1EX,  0x04, MXS_FP_UNIT,  2,  1 },
   { "unknownC1",  unknownC1ID,  unknownC1EX,  0x05, MXS_FP_UNIT,  2,  1 },
   { "unknownC1",  unknownC1ID,  unknownC1EX,  0x06, MXS_FP_UNIT,  2,  1 },
   { "unknownC1",  unknownC1ID,  unknownC1EX,  0x07, MXS_FP_UNIT,  2,  1 },
   { "unknownC1",  unknownC1ID,  unknownC1EX,  0x08, MXS_FP_UNIT,  2,  1 },
   { "unknownC1",  unknownC1ID,  unknownC1EX,  0x09, MXS_FP_UNIT,  2,  1 },
   { "unknownC1",  unknownC1ID,  unknownC1EX,  0x0a, MXS_FP_UNIT,  2,  1 },
   { "unknownC1",  unknownC1ID,  unknownC1EX,  0x0b, MXS_FP_UNIT,  2,  1 },
   { "unknownC1",  unknownC1ID,  unknownC1EX,  0x0c, MXS_FP_UNIT,  2,  1 },
   { "unknownC1",  unknownC1ID,  unknownC1EX,  0x0d, MXS_FP_UNIT,  2,  1 },
   { "unknownC1",  unknownC1ID,  unknownC1EX,  0x0e, MXS_FP_UNIT,  2,  1 },
   { "prefx",      unknownC1ID,  unknownC1EX,  0x0f, MXS_FP_UNIT,  2,  1 },
   { "unknownC1",  unknownC1ID,  unknownC1EX,  0x10, MXS_FP_UNIT,  2,  1 },
   { "unknownC1",  unknownC1ID,  unknownC1EX,  0x11, MXS_FP_UNIT,  2,  1 },
   { "unknownC1",  unknownC1ID,  unknownC1EX,  0x12, MXS_FP_UNIT,  2,  1 },
   { "unknownC1",  unknownC1ID,  unknownC1EX,  0x13, MXS_FP_UNIT,  2,  1 },
   { "unknownC1",  unknownC1ID,  unknownC1EX,  0x14, MXS_FP_UNIT,  2,  1 },
   { "unknownC1",  unknownC1ID,  unknownC1EX,  0x15, MXS_FP_UNIT,  2,  1 },
   { "unknownC1",  unknownC1ID,  unknownC1EX,  0x16, MXS_FP_UNIT,  2,  1 },
   { "unknownC1",  unknownC1ID,  unknownC1EX,  0x17, MXS_FP_UNIT,  2,  1 },
   { "unknownC1",  unknownC1ID,  unknownC1EX,  0x18, MXS_FP_UNIT,  2,  1 },
   { "unknownC1",  unknownC1ID,  unknownC1EX,  0x19, MXS_FP_UNIT,  2,  1 },
   { "unknownC1",  unknownC1ID,  unknownC1EX,  0x1a, MXS_FP_UNIT,  2,  1 },
   { "unknownC1",  unknownC1ID,  unknownC1EX,  0x1b, MXS_FP_UNIT,  2,  1 },
   { "unknownC1",  unknownC1ID,  unknownC1EX,  0x1c, MXS_FP_UNIT,  2,  1 },
   { "unknownC1",  unknownC1ID,  unknownC1EX,  0x1d, MXS_FP_UNIT,  2,  1 },
   { "unknownC1",  unknownC1ID,  unknownC1EX,  0x1e, MXS_FP_UNIT,  2,  1 },
   { "unknownC1",  unknownC1ID,  unknownC1EX,  0x1f, MXS_FP_UNIT,  2,  1 },
   { "madd.s",     madd_sID,     madd_sEX,     0x20, MXS_FP_UNIT,  4,  1 },
   { "madd.d",     madd_dID,     madd_dEX,     0x21, MXS_FP_UNIT,  4,  1 },
   { "unknownC1",  unknownC1ID,  unknownC1EX,  0x22, MXS_FP_UNIT,  2,  1 },
   { "unknownC1",  unknownC1ID,  unknownC1EX,  0x23, MXS_FP_UNIT,  2,  1 },
   { "unknownC1",  unknownC1ID,  unknownC1EX,  0x24, MXS_FP_UNIT,  2,  1 },
   { "unknownC1",  unknownC1ID,  unknownC1EX,  0x25, MXS_FP_UNIT,  2,  1 },
   { "unknownC1",  unknownC1ID,  unknownC1EX,  0x26, MXS_FP_UNIT,  2,  1 },
   { "unknownC1",  unknownC1ID,  unknownC1EX,  0x27, MXS_FP_UNIT,  2,  1 },
   { "msub.s",     msub_sID,     msub_sEX,     0x28, MXS_FP_UNIT,  4,  1 },
   { "msub.d",     msub_dID,     msub_dEX,     0x29, MXS_FP_UNIT,  4,  1 },
   { "unknownC1",  unknownC1ID,  unknownC1EX,  0x2a, MXS_FP_UNIT,  2,  1 },
   { "unknownC1",  unknownC1ID,  unknownC1EX,  0x2b, MXS_FP_UNIT,  2,  1 },
   { "unknownC1",  unknownC1ID,  unknownC1EX,  0x2c, MXS_FP_UNIT,  2,  1 },
   { "unknownC1",  unknownC1ID,  unknownC1EX,  0x2d, MXS_FP_UNIT,  2,  1 },
   { "unknownC1",  unknownC1ID,  unknownC1EX,  0x2e, MXS_FP_UNIT,  2,  1 },
   { "unknownC1",  unknownC1ID,  unknownC1EX,  0x2f, MXS_FP_UNIT,  2,  1 },
   { "nmadd.s",    nmadd_sID,    nmadd_sEX,    0x30, MXS_FP_UNIT,  4,  1 },
   { "nmadd.d",    nmadd_dID,    nmadd_dEX,    0x31, MXS_FP_UNIT,  4,  1 },
   { "unknownC1",  unknownC1ID,  unknownC1EX,  0x32, MXS_FP_UNIT,  2,  1 },
   { "unknownC1",  unknownC1ID,  unknownC1EX,  0x33, MXS_FP_UNIT,  2,  1 },
   { "unknownC1",  unknownC1ID,  unknownC1EX,  0x34, MXS_FP_UNIT,  2,  1 },
   { "unknownC1",  unknownC1ID,  unknownC1EX,  0x35, MXS_FP_UNIT,  2,  1 },
   { "unknownC1",  unknownC1ID,  unknownC1EX,  0x36, MXS_FP_UNIT,  2,  1 },
   { "unknownC1",  unknownC1ID,  unknownC1EX,  0x37, MXS_FP_UNIT,  2,  1 },
   { "nmsub.s",    nmsub_sID,    nmsub_sEX,    0x38, MXS_FP_UNIT,  4,  1 },
   { "nmsub.d",    nmsub_dID,    nmsub_dEX,    0x39, MXS_FP_UNIT,  4,  1 },
   { "unknownC1",  unknownC1ID,  unknownC1EX,  0x3a, MXS_FP_UNIT,  2,  1 },
   { "unknownC1",  unknownC1ID,  unknownC1EX,  0x3b, MXS_FP_UNIT,  2,  1 },
   { "unknownC1",  unknownC1ID,  unknownC1EX,  0x3c, MXS_FP_UNIT,  2,  1 },
   { "unknownC1",  unknownC1ID,  unknownC1EX,  0x3d, MXS_FP_UNIT,  2,  1 },
   { "unknownC1",  unknownC1ID,  unknownC1EX,  0x3e, MXS_FP_UNIT,  2,  1 },
   { "unknownC1",  unknownC1ID,  unknownC1EX,  0x3f, MXS_FP_UNIT,  2,  1 }
};

MXSOpcodeDesc mxsBCOpcode = 
   { "bc",         bcID,         bcEX,         0x0,  MXS_FP_UNIT,  2,  1 };
MXSOpcodeDesc mxsCFCOpcode =
   { "cfc1",       cfcID,        cfcEX,        0x0,  MXS_FP_UNIT,  2,  1 };
MXSOpcodeDesc mxsCTCOpcode =
   { "ctc1",       ctcID,        ctcEX,        0x0,  MXS_FP_UNIT,  2,  1 };
MXSOpcodeDesc mxsMFCOpcode =
   { "mfc1",       mfcID,        mfcEX,        0x0,  MXS_FP_UNIT,  2,  1 };
MXSOpcodeDesc mxsMTCOpcode =
   { "mtc1",       mtcID,        mtcEX,        0x0,  MXS_FP_UNIT,  3,  1 };
MXSOpcodeDesc mxsDMFCOpcode =
   { "dmfc1",      dmfcID,       dmfcEX,       0x0,  MXS_FP_UNIT,  2,  1 };
MXSOpcodeDesc mxsDMTCOpcode =
   { "dmtc1",      dmtcID,       dmtcEX,       0x0,  MXS_FP_UNIT,  3,  1 };
MXSOpcodeDesc mxsCVTSWOpcode =
   { "cvt.s.w",    cvt_s_wID,    cvt_s_wEX,    0x0,  MXS_FP_UNIT,  4,  1 };
MXSOpcodeDesc mxsCVTDWOpcode =
   { "cvt.d.w",    cvt_d_wID,    cvt_d_wEX,    0x0,  MXS_FP_UNIT,  2,  1 };
MXSOpcodeDesc mxsCVTSLOpcode =
   { "cvt.s.l",    cvt_s_lID,    cvt_s_lEX,    0x0,  MXS_FP_UNIT,  4,  1 };
MXSOpcodeDesc mxsCVTDLOpcode =
   { "cvt.d.l",    cvt_d_lID,    cvt_d_lEX,    0x0,  MXS_FP_UNIT,  2,  1 };
