/*
 * Copyright (C) 1996-1998 by the Board of Trustees
 *    of Leland Stanford Junior University.
 * 
 * This file is part of the SimOS distribution. 
 * See LICENSE file for terms of the license. 
 *
 */


/*
 *  ms_grad.c  -  Handle instruction graduation for the MXS simulator
 *
 *	Only needed to implement precise interrupts
 *	Also handles exceptions for the MIPSY version of MXS
 *
 *	Jim Bennett
 *	1994, 1995
 */

#include <sys/types.h>
#include "ms.h"
#include "ms_fpu.h"
#include "mipsy.h"
#include "cpu_state.h"
#include "cpu_stats.h"
#include "cpu_interface.h"
#include "cp0.h"
#include "hw_events.h"
#include "machine_defs.h"
#ifdef PRINT_INSTRUCTIONS
#include "print_insts.h"
#endif

#define COHERENCY_EXCEPTION -1
#define SWITCH2MIPSY_EXCEPTION -2

static void RecordExceptionStats(struct s_cpu_state *st);
static int InstIsInterruptable(struct s_cpu_state *st);

/*
 *  ms_graduate  -  Graduate instructions at the tail of the
 *		    graduation queue that have been freed.
 *
 *	This routine maintains the state of the processor at
 *	the graduation point, which is the only point at which
 *	consistency is assured.
 */

void 
ms_graduate (struct s_cpu_state *st)
{
#ifdef PRECISE
   int	inum, flush_inst;
   int	i, r;
   THREAD	*th;
   INST	*ip;
   int	lreg, halfreg;
   
   CPUState *P = (CPUState *)(st->mipsyPtr);
   /* Dump instruction window */
   int cpuNum = P->myNum;
   int iwinDump = printIwinMask[cpuNum];

   AddToStat (ST_GRADS_TRIED, GRAD_WIDTH);
   if (st->exception_pending) {
      AddToStat (ST_GRAD_EXPEND, GRAD_WIDTH);
      record_pipeline_stall(st, st->grad.pc,GRAD_WIDTH);
      if (iwinDump) {
         CPUPrint("<IWIN> grad 0 EXCEPT\n", 0);
      }
      return;
   }

   th = &st->grad;
   for (i=0; i<GRAD_WIDTH; i++) {
      inum = st->iwin_headgrad;
      if ((inum >= 0) && ((st->iwin_flags[inum] & IWIN_FREED) ||
                          ((st->iwin_flags[inum] & IWIN_FAULT) && 
                           !(st->iwin_flags[inum] & IWIN_SPEC)))) {
         if (st->iwin_br_node [inum] == 0) {
            /* If the instruction at the head of the graduation	*/
            /* queue caused an exception, then bail out now.	*/
            if (st->iwin_flags[inum] & IWIN_FAULT) {
               st->exception_pending = 1;
               st->except = st->iwin_except [inum];
               AddToStat (ST_GRAD_EXPEND, GRAD_WIDTH-i);
               record_pipeline_stall(st, th->pc,GRAD_WIDTH-i);
               if (iwinDump) {
                  CPUPrint("<IWIN> grad %d EXCEPT\n", i);
               }
               return;
            }

            /* Retire stores out of the store buffer, if we can,	*/
            /* bail out if we can't.					*/
            
            if (st->iwin_flags[inum] & IWIN_LDST) { 
               if (st->iwin_flags[inum] & IWIN_LDSTBUF) {
                  int ret = ldst_buffer_retire(st,inum);
                  if (ret != 0) {
                     /* Line hasn't come back to cache - try again later */
                     if (ret > 0) { 
                        if (st->iwin_flags[inum] & IWIN_STORE) { 
                           AddToStat (ST_GRAD_STSTALL, GRAD_WIDTH-i);
                        } else {
                           AddToStat (ST_GRAD_LDSTALL, GRAD_WIDTH-i);
                        }
                        record_ldst_stall(st, inum, GRAD_WIDTH-i);
                        if (iwinDump) {
                           CPUPrint("<IWIN> grad %d LDST\n", i);
                        }
                        return; 
                     }
                     /* Line was nuked from the cache -  backup and retry */
                     st->exception_pending = 1;
                     st->except = COHERENCY_EXCEPTION;
                     AddToStat (ST_GRAD_EXPEND, GRAD_WIDTH-i);
                     record_pipeline_stall(st, th->pc,GRAD_WIDTH-i);
                     if (iwinDump) {
                        CPUPrint("<IWIN> grad %d EXCEPT\n", i);
                     }
                     return;
                  }
               }
               flush_inst = (st->iwin_flags[inum] & IWIN_FLUSH);
               IncStat(ST_GRAD_INST_LDST);
               if (st->iwin[inum].flags & INST_SC) {
                  THREAD *sth = &st->threads[st->branch_tree[0].thread];
                  /* Continue the instruction fetcher after an SC */
                  sth->stall_sc = 0;
                  UpdateStallFetch (sth);
               }
            } else {
               flush_inst = (st->iwin_flags[inum] & IWIN_FLUSH);
               if (st->iwin[inum].flags & INST_SYNC) {
                  THREAD *sth = &st->threads[st->branch_tree[0].thread];
                  /* Continue the instruction fetcher after a sync */
                  sth->stall_sync = 0;
                  UpdateStallFetch (sth);
               }
            }

            /* We've got an instruction that is all done, so	*/
            /* graduate it.					*/

            if (iwinDump) {
               dump_iwin_entry(st, inum);
            }
            if (th->thread_st & TH_BRANCH) {
               th->pc = th->branch_pc;
               th->thread_st &= ~TH_BRANCH;
            } else {
               th->pc = th->pc + PC_INC;
            }

            if (st->iwin_flags[inum] & IWIN_BRANCH) {
               if ((st->iwin[inum].flags & INST_LIKELY) &&
                   !(st->iwin_flags[inum] & IWIN_TAKENBR)) {
                  /* The next PC of a not taken branch likely 
                   * instruction skips over the delay slot. */
                  th->pc = st->iwin_branch_pc[inum];
               } else {
                  /* Regular branches update after delay slot inst */
                  th->thread_st |= TH_BRANCH;
                  th->branch_pc = st->iwin_branch_pc [inum];
               }
            }
#ifndef R3K
            if (st->iwin[inum].flags & INST_ERET) {
               /* R4K eret doesn't have a delay slot either. */
               th->pc = st->iwin_branch_pc[inum];
            }
#endif
            IncStat(ST_GRAD_INST);
            record_grad_inst(st, th->pc, 1);
            
            /* Have to dequeue before updating the register name	*/
            /* table, so that old name of register can be released	*/

            ms_grad_dequeue (st, inum);

            /* The thread state at the graduation point needs to	*/
            /* correctly track the register name table.  This code	*/
            /* duplicates the actions that the fetch unit took when	*/
            /* this instruction was fetched.				*/

            ip = &st->iwin[inum];
            for (r = 0; r < MAX_DST_REGS; r++) {
               RegInfo *ri = &ip->dst[r];
               if (ri->flags != 0) {
                  if (ri->flags & REG_FPU) {
                     uint fpr = ri->reg & ~MXS_NFRB;
                     th->regnames[fpr] = ri->preg;
                  } else {
                     th->regnames[ri->reg] = ri->preg;
                  }
               }
            }

            /* Update the FP control/status register, take an exception if necessary */
            if (ip->dst[0].reg == MXS_REG_FPCSR) {
               /* Need to copy condition codes over, unless this is a CTC, 
                * in which case we get the whole thing */
               if (ip->flags & INST_FPCTL) {
                  P->FCR[31] = GPR(ip->dst[0].preg);
               } else {
                  P->FCR[31] &= ~M_csr_fccs;
                  P->FCR[31] |= GPR(ip->dst[0].preg) & M_csr_fccs;
               }
            }
            if ((ip->fp_exc != 0) || (ip->flags & INST_FPCTL)) {
               /* There's no E enable/flag, just a cause */
               Reg flags = ip->fp_exc & ~MXS_FPE_E;
               Reg causes = ip->fp_exc;
               Reg enables;
               P->FCR[31] |= (causes << S_csr_causes) | (flags << S_csr_flags);
               /* Check to see if we need to take an exception */
               causes = (P->FCR[31] & M_csr_causes) >> S_csr_causes;
               enables = (P->FCR[31] & M_csr_enables) >> S_csr_enables;
               if (causes & enables) {
                  st->exception_pending = 1;
                  st->except = st->iwin_except[inum];
                  AddToStat (ST_GRAD_EXPEND, GRAD_WIDTH-i);
                  record_pipeline_stall(st, th->pc,GRAD_WIDTH-i);
                  if (iwinDump) {
                     CPUPrint("<IWIN> grad %d EXCEPT\n", i);
                  }
                  return;
               }
            }

            /* Check for load/store/PC annotations */

            P->PC = st->iwin_pc[inum];
            if (annLoads && (st->iwin[inum].flags & INST_LOAD)) {
               VA vaddr = st->inum2ldst[inum]->ls->addr;
               PA paddr = st->inum2ldst[inum]->ls->paddr;
               AnnPtr ptr = AnnFMLookup(vaddr, ANNFM_LD_TYPE);
               if (ptr) {
                  AnnExec(ptr);
               }
               ptr = AnnFMLookup(paddr, ANNFM_LD_TYPE);
               if (ptr) {
                  AnnExec(ptr);
               }
            } else if (annStores && (st->iwin[inum].flags & INST_STORE)) {
               VA vaddr = st->inum2ldst[inum]->ls->addr;
               PA paddr = st->inum2ldst[inum]->ls->paddr;
               AnnPtr ptr = AnnFMLookup(vaddr, ANNFM_ST_TYPE);
               if (ptr) {
                  AnnExec(ptr);
               }
               ptr = AnnFMLookup(paddr, ANNFM_ST_TYPE);
               if (ptr) {
                  AnnExec(ptr);
               }
            }

            STATS_INC(P->myNum, numInstructions, 1);
            if (STATS_VALUE(P->myNum, numInstructions) > 
                STATS_VALUE(P->myNum, nextInstrSample)) {
               INST_SAMPLE_EVENT(MipsyReadTime(P->myNum), P->myNum, P->PC);
               STATS_INC(P->myNum, nextInstrSample, 
                         MS_SAMPLE_INSTR_INTERVAL);
            }
#ifdef PRINT_INSTRUCTIONS
            P->PC = st->iwin_pc[inum];
            PRINT_INSTRUCTION(st->iwin[inum].instr);
#endif
            RunPCAnnotations(st->iwin_pc[inum], 0);
         } else {
            if (iwinDump) {
               dump_iwin_entry(st, inum);
            }
            /* If it's completed, delete it regardless of which	*/
            /* branch node it is on.				*/
            if (st->iwin_flags[inum] & IWIN_LDST) {
#ifdef DEBUG_CHECKS
               if (st->iwin_flags[inum] & IWIN_LDSTBUF) {
                  CPUError("Squashed entry still in ldst buf\n");
               }
#endif
               IncStat(ST_GRAD_SQUASHED_LDST);
            }
            IncStat(ST_GRAD_SQUASHED);
            flush_inst = 0;
            ms_grad_dequeue (st, inum);
            record_pipeline_stall(st, th->pc,1);
         }
         continue;	/* Keep on graduating instructions,	*/
      }			/* since this one graduated.		*/
      if (inum < 0) { 
         /* Pipeline is totally empty */
         if (st->active_thread && st->active_thread->stall_icache) {
            /* Nothing done this cycle - record this as memory stall */
            record_icache_stall(st, th->pc,GRAD_WIDTH-i);
            AddToStat(ST_GRAD_ICACHESTALL, GRAD_WIDTH-i);
            if (iwinDump) {
               CPUPrint("<IWIN> grad %d ICACHE\n", i);
            }
         } else if (flush_inst) { 
            record_pipeline_stall(st, th->pc,GRAD_WIDTH-i);
            AddToStat(ST_GRAD_FLUSHSTALL, GRAD_WIDTH-i);
            if (iwinDump) {
               CPUPrint("<IWIN> grad %d FLUSH\n", i);
            }
         } else {
            record_pipeline_stall(st, th->pc,GRAD_WIDTH-i);
            AddToStat (ST_GRAD_EMPTYSTALL, GRAD_WIDTH-i);
            if (iwinDump) {
               CPUPrint("<IWIN> grad %d EMPTY\n", i);
            }
         }
      } else if ((st->iwin_flags[inum] & IWIN_LDST) &&
                 (st->iwin_flags[inum] & IWIN_LDSTBUF) &&
                 (st->inum2ldst[inum]->ls->status & LS_ST_STALL)) {
         record_ldst_stall(st, inum, GRAD_WIDTH-i);
         if (st->iwin_flags[inum] & IWIN_STORE)  {
            AddToStat (ST_GRAD_STSTALL, GRAD_WIDTH-i);
         } else { 
            AddToStat (ST_GRAD_LDSTALL, GRAD_WIDTH-i);
         }
         if (iwinDump) {
            CPUPrint("<IWIN> grad %d LDST\n", i);
         }
      } else {
         record_pipeline_stall(st, th->pc,GRAD_WIDTH-i);
         if (!(st->iwin_flags[inum] & IWIN_ISSUED)) { 
            if (st->iwin_flags[inum] & IWIN_FLUSH) { 
               AddToStat(ST_GRAD_FLUSHSTALL, GRAD_WIDTH-i);
               if (iwinDump) {
                  CPUPrint("<IWIN> grad %d FLUSH\n", i);
               }
            } else {
               AddToStat(ST_GRAD_ISSUESTALL, GRAD_WIDTH-i);
               if (iwinDump) {
                  CPUPrint("<IWIN> grad %d ISSUE\n", i);
               }
            }
         } else if (st->iwin_flags[inum] & IWIN_LDST) {
            AddToStat(ST_GRAD_PHASEASTALL, GRAD_WIDTH-i);
            if (iwinDump) {
               CPUPrint("<IWIN> grad %d PHASE_A\n", i);
            }
         } else {
            AddToStat(ST_GRAD_MISCSTALL, GRAD_WIDTH-i);
            if (iwinDump) {
               CPUPrint("<IWIN> grad %d FUNC_UNIT\n", i);
            }
         }
      }
      break;		/* Rest of instructions not eligible	*/
   }			/* for graduation, quit.		*/
   if (iwinDump && (i == GRAD_WIDTH)) {
      CPUPrint("<IWIN> grad %d ACTIVE\n", i);
   }
#endif /* PRECISE */
}

/*
 *  ms_grad_dequeue  -  Remove instruction from graduation queue
 *
 *	Also update the status of the destination register and
 *	put the instruction slot on the free list.
 */

void 
ms_grad_dequeue (struct s_cpu_state *st, int inum)
{
#ifdef PRECISE
   int	tmpi, preg;
   REGSTAT	*rs;

   tmpi = st->iwin_bgrd[inum];

#ifdef DEBUG_CHECKS
   if ((st->iwin_flags [inum] & IWIN_FREED) == 0) {
      CPUError("Attempt to graduate a live instruction\r\n");
   }
   if ((inum != st->iwin_headgrad) || (tmpi >= 0)) {
      CPUError("Out of order graduation\r\n");
   }
#endif /* DEBUG_CHECKS */

   if (tmpi >= 0)
      st->iwin_grad[tmpi] = st->iwin_grad[inum];
   else
      st->iwin_headgrad = st->iwin_grad[inum];

   if ((tmpi = st->iwin_grad[inum]) >= 0)
      st->iwin_bgrd[tmpi] = st->iwin_bgrd[inum];
   else
      st->iwin_tailgrad = st->iwin_bgrd[inum];

   if (st->iwin_flags [inum] & IWIN_DEFINE) {
      int r;
      for (r = 0; r < MAX_DST_REGS; r++) {
         RegInfo *ri = &st->iwin[inum].dst[r];
         if (ri->flags != 0) {
            if (st->iwin_br_node [inum] == 0) {
               /* On graduation of a non-speculative redefinition of a	*/
               /* register, the previous name for the register should	*/
               /* be freed.						*/
               if (ri->flags & REG_FPU) {
                  uint fpr = ri->reg & ~MXS_NFRB;
                  preg = st->grad.regnames[fpr];
               } else {
                  preg = st->grad.regnames[ri->reg];
               }
            } else {
               /* On graduation of a speculative redefinition of a	*/
               /* register, the current name should be freed.		*/
               preg = ri->preg;
            }
            rs = &st->reg_rstat [preg];
            rs->reg_status |= REG_FREED;
            CheckRegFree (st, rs, preg);
         }
      }
   }
   UnthreadInst (st, inum);
#endif /* PRECISE */
}


/*
 *  ms_grad_enqueue  -  Add instruction to graduation queue
 */

void 
ms_grad_enqueue (struct s_cpu_state *st, int inum)
{
#ifdef PRECISE
   if (st->iwin_tailgrad >= 0) {
      st->iwin_grad [st->iwin_tailgrad] = inum;
      st->iwin_bgrd [inum] = st->iwin_tailgrad;
      st->iwin_grad [inum] = -1;
      st->iwin_tailgrad = inum;
   } else {
      st->iwin_headgrad = inum;
      st->iwin_tailgrad = inum;
      st->iwin_bgrd [inum] = -1;
      st->iwin_grad [inum] = -1;
   }
#endif /* PRECISE */
}

static int pipeline_empty (struct s_cpu_state *st);

/*
 *  ms_except  -  Check for exceptions, and handle them for MXS
 */

void
ms_except (struct s_cpu_state *st)
{
   CPUState *P = (CPUState *)(st->mipsyPtr);
   THREAD *th;
   int	inum;

   /* Have we got an ITLB miss?  Can recognize this when the graduation	*/
   /* queue is empty and the thread corresponding to the root of the	*/
   /* branch tree is stalled on an ITLB miss.				*/

   inum = st->iwin_headgrad;
   if (inum < 0) {
      th = &st->threads [st->branch_tree [0].thread];
      if ((!st->exception_pending) && th->stall_itlbmiss) {
         st->exception_pending = 1;
         st->except = th->except;
         IncStat(ST_EXCEPT_ITLB);
      }
   } else if (!st->exception_pending) {
      /* Check for the case of a ITLB miss in the delay slot of a branch */
      th = &st->threads [st->branch_tree [0].thread];
      if (th->stall_itlbmiss && (st->iwin_br_node[inum] == 0) && 
          (st->iwin_flags[inum] & IWIN_BRANCH) &&
          (th->pc == (st->iwin_pc[inum]+4))) {
         st->exception_pending = 1;
         st->except = th->except;
         IncStat(ST_EXCEPT_ITLB);
      }
   } 

   /* Is there an interrupt hanging around?  If so, recognize it.	*/
   if (!st->exception_pending && InterruptIsPending (st) &&
       InstIsInterruptable(st)) {
      RECORD_EXCEPTION(P, EXC_INT, E_VEC, 0, 
                       P->CP0[C0_TLBHI],P->CP0[C0_CTXT],P->CP0[C0_XCTXT]);
      st->exception_pending = 1;
      st->except = GetLastException (st);
      IncStat(ST_EXCEPT_INTR);
   }

   th = &st->grad;
   if (P->switchToMIPSY && !st->exception_pending) {
      if ((st->grad.pc < (K2BASE + K2SIZE)) && (!(th->thread_st & TH_BRANCH))) { 
         st->exception_pending = 1;
         st->except = SWITCH2MIPSY_EXCEPTION;
      }
   }

   if (st->exception_pending && pipeline_empty (st)) {
      RecordExceptionStats(st);
      HandleException(st, st->except, (th->thread_st & TH_BRANCH));
#ifdef PRINT_INSTRUCTIONS
      if (enable_eprint) {
         PrintException(st, st->except);
      }
#endif
      st->exception_pending = 0;
   } else if (st->exception_pending) {
      IncStat(ST_EXCEPT_WAIT);
   }
}


/*
 *  pipeline_empty  -  Return TRUE if there are no outstanding
 *			operations in the pipeline.
 */

static int 
pipeline_empty (struct s_cpu_state *st)
{
   if ((st->ex_count > 0) || (st->work_head)) {
      return 0;
   } else {
      return 1;
   }
}

static int 
InstIsInterruptable(struct s_cpu_state *st) 
{
   int inum = st->iwin_headgrad;

   if (inum >= 0) { 
      if (((st->iwin_flags[inum] & 
            (IWIN_LDST|IWIN_UNCACHED|IWIN_LDSTBUF)) ==
           (IWIN_LDST|IWIN_UNCACHED|IWIN_LDSTBUF)) &&
          IsLoad(st->inum2ldst[inum]->ls->lstype) &&
          st->inum2ldst[inum]->dataPtr) {
         return 0; /* Don't abort unfinished uncached reads */
      }
      if ((st->iwin[inum].flags & (INST_LDHACK | INST_SDHACK)) &&
          (st->iwin_flags[inum] & IWIN_ISSUED)) {
         return 0; /* Handle ld/sd hack instruction */
      }
   }
   return 1;
}
         

static void 
RecordExceptionStats(struct s_cpu_state *st)
{
   int inum, inwin = 0, ldst = 0, squashed = 0;

   for (inum = st->iwin_headgrad; inum >= 0; inum = st->iwin_grad[inum]) {
      inwin++;
      if (st->iwin_flags[inum] & IWIN_LDST) 
         ldst++;
      if (st->iwin_flags[inum] & IWIN_SQUASH) 
         squashed++;
   }

   IncStat(ST_EXCEPT_TAKEN);
   if (st->except == COHERENCY_EXCEPTION) {
      IncStat(ST_EXCEPT_COHERENCY);
      AddToStat(ST_EXCEPT_COHER_INST, inwin);
      AddToStat(ST_EXCEPT_COHER_SQUASH, squashed);
      AddToStat(ST_EXCEPT_COHER_LDST, ldst);
   } else if (st->except == SWITCH2MIPSY_EXCEPTION)  {
      IncStat(ST_EXCEPT_SWITCH);
      AddToStat(ST_EXCEPT_SWITCH_INST, inwin);
      AddToStat(ST_EXCEPT_SQUASH, squashed);
      AddToStat(ST_EXCEPT_LDST, ldst);
   } else {
      AddToStat(ST_EXCEPT_INST, inwin);
      AddToStat(ST_EXCEPT_SQUASH, squashed);
      AddToStat(ST_EXCEPT_LDST, ldst);
   }
}
