/*
 * Copyright (C) 1996-1998 by the Board of Trustees
 *    of Leland Stanford Junior University.
 * 
 * This file is part of the SimOS distribution. 
 * See LICENSE file for terms of the license. 
 *
 */

/****************************************************************
 * ms_opcodes.c
 * 
 * Based on Mipsy's opcodes.c
 * Implementation of each MIPS instruction and declaration of the
 * table used to decode them.
 * 
 * $Author: verghese $
 * $Date: 1999/09/27 22:08:25 $
 *****************************************************************/

#include "ms.h"
#include "ms_opcodes.h"
#include "ms_fpu.h"

#include "cpu.h"
#include "cpu_state.h"
#include "sim_error.h"
#include "cp0.h"

#define ID(NAME) \
void NAME##ID(INST *ip, Reg pc)
#define EX(NAME) \
Result NAME##EX(struct s_cpu_state *st, INST *ip, THREAD *th)

MXSOpcodeDesc mxsSpecOpcodes[];
MXSOpcodeDesc mxsBcondOpcodes[];

/*
 * LOAD_STATS() - Common routine for updating load statistics
 */
#if PC_LATENCY > 1
#define LOAD_STATS(_st, _reg)                                   \
      if (!_st->stall_issue) {                                  \
         _st->reg_excuse[_reg] = ST_PC_DLY;                     \
         _st->new_excuse[_reg] = ST_CACHE_LSQ;                  \
         Add_to_worklist(_st, PC_LATENCY, update_excuse, _reg); \
      }
#else
#define LOAD_STATS(_st, _reg)                                   \
      if (!_st->stall_issue) {                                  \
         _st->reg_excuse[_reg] = ST_CACHE_LSQ;                  \
         _st->new_excuse[_reg] = ST_NO_EXCUSE;                  \
      }
#endif

#define CURRENT_PC(_cpu) (pePtr[_cpu]->PC)

static void   LongMultiply32 (struct s_cpu_state *, Reg32, Reg32, uint, uint);
static void   LongMultiply64 (struct s_cpu_state *, Reg64, Reg64, uint, uint);

static void   CallOp(struct s_cpu_state *st, INST *ip,
                     THREAD *th, int cond, int likely);
static void   TrapOp(struct s_cpu_state *st, INST *ip,
                     THREAD *th, int cond);

/****************************************************************
 * M A J O R   O P C O D E S
 *****************************************************************/

ID(spec) 
{
   uint   specialFunc = FUNC(ip->instr);
   ip->opcode = &mxsSpecOpcodes[specialFunc];
   mxsSpecOpcodes[specialFunc].decode(ip, pc);
}
EX(spec) 
{
   uint   specialFunc = FUNC(ip->instr);
   mxsSpecOpcodes[specialFunc].execute(st, ip, th);
   return SUCCESS;
}


ID(bcond) 
{
   uint bcondFunc    = RT(ip->instr);
   ip->opcode = &mxsSpecOpcodes[bcondFunc];
   mxsBcondOpcodes[bcondFunc].decode(ip, pc);
}
EX(bcond) 
{
   uint bcondFunc    = RT(ip->instr);
   mxsBcondOpcodes[bcondFunc].execute(st, ip, th);
   return SUCCESS;
}

ID(addi)
{
   ip->dst[0].flags = REG_INT; ip->dst[0].reg = RTRI(ip->instr);
   ip->src[0].flags = REG_INT; ip->src[0].reg = RSRI(ip->instr); 
}
EX(addi)
{
   int immediate     = IMMED(ip->instr);
   uint  rt          = ip->dst[0].preg;
   Reg_s contents_rs = GPR(ip->src[0].preg);
   Reg32_s result32  = (Reg32_s)contents_rs + SIGN_EXTEND32(16, immediate);
   if (ARITH_OVFL32(result32, (Reg32_s)contents_rs, 
                    SIGN_EXTEND32(16, immediate))) {
      int inum = ip - st->iwin;
      EX_RECORD_EXCEPTION(st, inum, EXC_OV, 0);
      EX_HANDLE_EXCEPTION(st, ip, th);
      return SUCCESS;
   }
   /* Don't write this value if there was an overflow */
   GPR(rt) = result32;
   return SUCCESS;
}

ID(addiu)
{
   ip->dst[0].flags = REG_INT; ip->dst[0].reg = RTRI(ip->instr);
   ip->src[0].flags = REG_INT; ip->src[0].reg = RSRI(ip->instr); 
}
EX(addiu)
{
   int immediate     = IMMED(ip->instr);
   uint  rt          = ip->dst[0].preg;
   Reg_s contents_rs = GPR(ip->src[0].preg);
   GPR(rt) = (Reg32_s)contents_rs + SIGN_EXTEND32(16, immediate);
   return SUCCESS;
}

ID(andi)
{
   ip->dst[0].flags = REG_INT; ip->dst[0].reg = RTRI(ip->instr);
   ip->src[0].flags = REG_INT; ip->src[0].reg = RSRI(ip->instr); 
}
EX(andi)
{
   int immediate     = IMMED(ip->instr);
   uint  rt          = ip->dst[0].preg;
   Reg_s contents_rs = GPR(ip->src[0].preg);
   GPR(rt) = contents_rs & (0x0000ffff & immediate);
   return SUCCESS;
}

ID(beq)
{
   int immediate     = IMMED(ip->instr);
   ip->src[0].flags = REG_INT; ip->src[0].reg = RTRI(ip->instr);
   ip->src[1].flags = REG_INT; ip->src[1].reg = RSRI(ip->instr); 
   ip->target = (pc + 4) + SIGN_EXTEND32(18, immediate << 2);
   if ((ip->src[0].reg == 0) && (ip->src[1].reg == 0)) {
      ip->flags = INST_BRANCH;
   } else {
      ip->flags = INST_BRANCH | INST_CONDITIONAL;
   }
}
EX(beq)
{
   Reg_s contents_rs = GPR(ip->src[1].preg);
   Reg_s contents_rt = GPR(ip->src[0].preg);
   int cond = (contents_rs == contents_rt);
   BranchOp(st, ip, th, cond, 0);
   return SUCCESS;
}

ID(beql)
{
   int immediate     = IMMED(ip->instr);
   ip->src[0].flags = REG_INT; ip->src[0].reg = RTRI(ip->instr);
   ip->src[1].flags = REG_INT; ip->src[1].reg = RSRI(ip->instr); 
   ip->target = (pc + 4) + SIGN_EXTEND32(18, immediate << 2);
   ip->flags = INST_BRANCH | INST_CONDITIONAL | INST_LIKELY;
}
EX(beql)
{
   Reg_s contents_rs = GPR(ip->src[1].preg);
   Reg_s contents_rt = GPR(ip->src[0].preg);
   int cond = (contents_rs == contents_rt);
   BranchOp(st, ip, th, cond, 1);
   return SUCCESS;
}

ID(bgtz)
{
   int immediate     = IMMED(ip->instr);
   ip->src[0].flags = REG_INT; ip->src[0].reg = RSRI(ip->instr); 
   ip->target = (pc + 4) + SIGN_EXTEND32(18, immediate << 2);
   ip->flags = INST_BRANCH | INST_CONDITIONAL;
}
EX(bgtz)
{   
   Reg_s contents_rs = GPR(ip->src[0].preg);
   int cond = (!(SIGN_BIT(contents_rs)) && (contents_rs != 0));
   BranchOp(st, ip, th, cond, 0);
   return SUCCESS;
}

ID(bgtzl)
{
   int immediate     = IMMED(ip->instr);
   ip->src[0].flags = REG_INT; ip->src[0].reg = RSRI(ip->instr); 
   ip->target = (pc + 4) + SIGN_EXTEND32(18, immediate << 2);
   ip->flags = INST_BRANCH | INST_CONDITIONAL | INST_LIKELY;
}
EX(bgtzl)
{
   Reg_s contents_rs = GPR(ip->src[0].preg);
   int cond = (!(SIGN_BIT(contents_rs)) && (contents_rs != 0));
   BranchOp(st, ip, th, cond, 1);
   return SUCCESS;
}

ID(blez)
{
   int immediate     = IMMED(ip->instr);
   ip->src[0].flags = REG_INT; ip->src[0].reg = RSRI(ip->instr); 
   ip->target = (pc + 4) + SIGN_EXTEND32(18, immediate << 2);
   if (ip->src[0].reg == 0) {
      ip->flags = INST_BRANCH;
   } else {
      ip->flags = INST_BRANCH | INST_CONDITIONAL;
   }
}
EX(blez)
{
   Reg_s contents_rs = GPR(ip->src[0].preg);
   int cond = ((SIGN_BIT(contents_rs) || (contents_rs == 0)));
   BranchOp(st, ip, th, cond, 0);
   return SUCCESS;
}

ID(blezl)
{
   int immediate     = IMMED(ip->instr);
   ip->src[0].flags = REG_INT; ip->src[0].reg = RSRI(ip->instr); 
   ip->target = (pc + 4) + SIGN_EXTEND32(18, immediate << 2);
   ip->flags = INST_BRANCH | INST_CONDITIONAL | INST_LIKELY;
}
EX(blezl)
{
   Reg_s contents_rs = GPR(ip->src[0].preg);
   int cond = ((SIGN_BIT(contents_rs) || (contents_rs == 0)));
   BranchOp(st, ip, th, cond, 1);
   return SUCCESS;
}

ID(bne)
{
   int immediate     = IMMED(ip->instr);
   ip->src[0].flags = REG_INT; ip->src[0].reg = RTRI(ip->instr);
   ip->src[1].flags = REG_INT; ip->src[1].reg = RSRI(ip->instr); 
   ip->target = (pc + 4) + SIGN_EXTEND32(18, immediate << 2);
   ip->flags = INST_BRANCH | INST_CONDITIONAL;
}
EX(bne)
{
   Reg_s contents_rs = GPR(ip->src[1].preg);
   Reg_s contents_rt = GPR(ip->src[0].preg);
   int cond = (contents_rs != contents_rt);
   BranchOp(st, ip, th, cond, 0);
   return SUCCESS;
}

ID(bnel)
{
   int immediate     = IMMED(ip->instr);
   ip->src[0].flags = REG_INT; ip->src[0].reg = RTRI(ip->instr);
   ip->src[1].flags = REG_INT; ip->src[1].reg = RSRI(ip->instr); 
   ip->target = (pc + 4) + SIGN_EXTEND32(18, immediate << 2);
   ip->flags = INST_BRANCH | INST_CONDITIONAL | INST_LIKELY;
}
EX(bnel)
{
   Reg_s contents_rs = GPR(ip->src[1].preg);
   Reg_s contents_rt = GPR(ip->src[0].preg);
   int cond = (contents_rs != contents_rt);
   BranchOp(st, ip, th, cond, 1);
   return SUCCESS;
}

ID(cache)
{
   ip->src[0].flags = REG_INT; ip->src[0].reg = RSRI(ip->instr); 
   ip->flags = INST_CACHE;
}
EX(cache)
{
#ifdef rpb_notdef
   /* Be careful with this one. Mipsy's caches are
      physically indexed. */
   VA   vAddr;
   PA   pAddr;
   CPUState *P = (CPUState*)(st->mipsyPtr);
   Reg_s base      = GPR(ip->src[0].preg);
   int immediate   = IMMED(ip->instr);
   int inum = ip - st->iwin;
   if (!AllsQuiet(st, ip, ST_SYSCALL))
      return SUCCESS;

#ifdef IRIX6_4               
   if (!IS_KERNEL_MODE(P)) {                 
      /*  is unusable only in kernel and or when the cu0 bit is set. */
      CauseReg  causeReg;
      StatusReg statusReg;
      statusReg.ts_data = P->CP0[C0_SR];
      if (!statusReg.s32.ts_cu0) {
         int inum = ip - st->iwin;
         causeReg.tc_data = P->CP0[C0_CAUSE];
         causeReg.s32.tc_ce = 0;
         P->CP0[C0_CAUSE] = causeReg.tc_data;
         EX_RECORD_EXCEPTION(st, inum, EXC_CPU, 0);
         EX_HANDLE_EXCEPTION(st, ip, th);
         return SUCCESS;
      }
   }
#endif
   /* This is a reserved instruction, but I'm allowing it
      to be used in user mode. */
   vAddr = SIGN_EXTEND(16, immediate) + base;
   ms_lsq (st, vAddr, &ip->src[0], 0, CACHE_INST, inum);
#else
   int inum = ip - st->iwin;
   CPUPrint("MXS Warning: cache opcode is currently a nop\n");
   th->stall_cp0 = 0;	
   ClearFPCStall(st, th, inum);
#endif
   return SUCCESS;
}

ID(daddi) 
{
   ip->dst[0].flags = REG_INT; ip->dst[0].reg = RTRI(ip->instr);
   ip->src[0].flags = REG_INT; ip->src[0].reg = RSRI(ip->instr); 
}
EX(daddi) 
{
   int immediate     = IMMED(ip->instr);
   uint  rt          = ip->dst[0].preg;
   Reg_s contents_rs = GPR(ip->src[0].preg);
   Reg64_s result64 = contents_rs + SIGN_EXTEND64(16, immediate);
   int inum = ip - st->iwin;

   if (!ENABLE_64BIT(st)) {
      EX_RECORD_EXCEPTION(st, inum, EXC_II, 0);
   } else if (ARITH_OVFL64(result64, contents_rs,
                    SIGN_EXTEND64(16, immediate))) {
      EX_RECORD_EXCEPTION(st, inum, EXC_OV, 0);
   }
   if (st->iwin_flags[inum] & IWIN_FAULT) {
      EX_HANDLE_EXCEPTION(st, ip, th);
      return SUCCESS;
   }
   /* Don't write this value if there was an overflow */
   GPR(rt) = result64;
   return SUCCESS;
}

ID(daddiu)
{
   ip->dst[0].flags = REG_INT; ip->dst[0].reg = RTRI(ip->instr);
   ip->src[0].flags = REG_INT; ip->src[0].reg = RSRI(ip->instr); 
}
EX(daddiu)
{
   int immediate     = IMMED(ip->instr);
   uint  rt          = ip->dst[0].preg;
   Reg_s contents_rs = GPR(ip->src[0].preg);
   if (!ENABLE_64BIT(st)) {
      int inum = ip - st->iwin;
      EX_RECORD_EXCEPTION(st, inum, EXC_II, 0);
      EX_HANDLE_EXCEPTION(st, ip, th);
      return SUCCESS;
   }
   GPR(rt) = contents_rs + SIGN_EXTEND64(16,immediate);
   return SUCCESS;
}

ID(j)
{
   ip->target = ((pc + 4) & (Reg32_s)0xf0000000) | (TARGET(ip->instr) << 2);
   ip->flags = INST_BRANCH;
}
EX(j)
{
   int inum = ip - st->iwin;
   th->pc = ip->target;
#if BRANCH_LATENCY > 1
   Add_to_worklist (st, BRANCH_LATENCY, finish_branch, (MXSClientData)ip);
#else
   th->stall_branch = 0;
   UpdateStallFetch (th);
   prune_branch (st, st->iwin_br_node[inum]);
#endif
   return SUCCESS;
}

ID(jal)
{
   ip->dst[0].flags = REG_INT; ip->dst[0].reg = REG_RA;
   ip->target = ((pc + 4) & (Reg32_s)0xf0000000) | (TARGET(ip->instr) << 2);
   ip->flags = INST_CALL;
}
EX(jal)
{
   uint  ra = ip->dst[0].preg;
   int inum = ip - st->iwin;
   BrTREE *br = &st->branch_tree [st->iwin_br_node[inum]];
   GPR(ra) = th->returnpc;
   br->resolution = PRUNE_LEFT;
   st->iwin_branch_pc [inum] = ip->target;

   if (br->lchild == -1) {
      th->pc = ip->target;
   }

#if BRANCH_LATENCY > 1
   Add_to_worklist (st, BRANCH_LATENCY, finish_call, (MXSClientData)ip);
#else
   th->stall_branch = 0;
   UpdateStallFetch (th);
   prune_branch (st, st->iwin_br_node[inum]);
#endif
   return SUCCESS;
}

ID(lb)
{
   ip->dst[0].flags = REG_INT; ip->dst[0].reg = RTRI(ip->instr);
   ip->src[0].flags = REG_INT; ip->src[0].reg = RSRI(ip->instr); 
   ip->flags = INST_LOAD;
}
EX(lb)
{
   int immediate     = IMMED(ip->instr);
   Reg  base = GPR(ip->src[0].preg);
   VA   vAddr = SIGN_EXTEND(16, immediate) + base; 
   
   int inum = ip - st->iwin;	    
   if (st->iwin_flags[inum] & IWIN_TLBFAULT) {
      st->iwin_flags[inum] |= IWIN_FAULT;
      EX_HANDLE_EXCEPTION(st, ip, th);
      return SUCCESS;
   }
   
   ms_lsq (st, vAddr, &ip->dst[0], 0, LD_INTEGER_B, inum);
   LOAD_STATS(st, ip->dst[0].preg);
   return SUCCESS;
}

ID(lbu)
{
   ip->dst[0].flags = REG_INT; ip->dst[0].reg = RTRI(ip->instr);
   ip->src[0].flags = REG_INT; ip->src[0].reg = RSRI(ip->instr); 
   ip->flags = INST_LOAD;
}
EX(lbu)
{ 
   int immediate     = IMMED(ip->instr);
   Reg  base = GPR(ip->src[0].preg);
   VA   vAddr = SIGN_EXTEND(16, immediate) + base; 
   
   int inum = ip - st->iwin;	    
   if (st->iwin_flags[inum] & IWIN_TLBFAULT) {
      st->iwin_flags[inum] |= IWIN_FAULT;
      EX_HANDLE_EXCEPTION(st, ip, th);
      return SUCCESS;
   }
   ms_lsq (st, vAddr, &ip->dst[0], 0, LD_UNSIGNED_B, inum);
   LOAD_STATS(st, ip->dst[0].preg);
   return SUCCESS;
}

ID(lh)
{
   ip->dst[0].flags = REG_INT; ip->dst[0].reg = RTRI(ip->instr);
   ip->src[0].flags = REG_INT; ip->src[0].reg = RSRI(ip->instr); 
   ip->flags = INST_LOAD;
}
EX(lh)
{
   int immediate     = IMMED(ip->instr);
   Reg  base = GPR(ip->src[0].preg);
   VA   vAddr = SIGN_EXTEND(16, immediate) + base; 
   
   int inum = ip - st->iwin;	    
   if (st->iwin_addr[inum] & 0x1) {
      EX_RECORD_EXCEPTION(st, inum, EXC_RADE, vAddr);
   } else if (st->iwin_flags[inum] & IWIN_TLBFAULT) {
      st->iwin_flags[inum] |= IWIN_FAULT;
   }
   if (st->iwin_flags[inum] & IWIN_FAULT) {
      EX_HANDLE_EXCEPTION(st, ip, th);
      return SUCCESS;
   }
   
   ms_lsq (st, vAddr, &ip->dst[0], 0, LD_INTEGER_H, inum);
   LOAD_STATS(st, ip->dst[0].preg);
   return SUCCESS;
}

ID(lhu)
{
   ip->dst[0].flags = REG_INT; ip->dst[0].reg = RTRI(ip->instr);
   ip->src[0].flags = REG_INT; ip->src[0].reg = RSRI(ip->instr); 
   ip->flags = INST_LOAD;
}
EX(lhu)
{
   int immediate     = IMMED(ip->instr);
   Reg  base = GPR(ip->src[0].preg);
   VA   vAddr = SIGN_EXTEND(16, immediate) + base; 
   
   int inum = ip - st->iwin;	    
   if (st->iwin_addr[inum] & 0x1) {
      EX_RECORD_EXCEPTION(st, inum, EXC_RADE, vAddr);
   } else if (st->iwin_flags[inum] & IWIN_TLBFAULT) {
      st->iwin_flags[inum] |= IWIN_FAULT;
   }
   if (st->iwin_flags[inum] & IWIN_FAULT) {
      EX_HANDLE_EXCEPTION(st, ip, th);
      return SUCCESS;
   }
   
   ms_lsq (st, vAddr, &ip->dst[0], 0, LD_UNSIGNED_H, inum);
   LOAD_STATS(st, ip->dst[0].preg);
   return SUCCESS;
}

ID(lui)
{
   ip->dst[0].flags = REG_INT; ip->dst[0].reg = RTRI(ip->instr);
}
EX(lui)
{
   int immediate     = IMMED(ip->instr);
   uint  rt          = ip->dst[0].preg;
   GPR(rt) = immediate << 16;
   return SUCCESS;
}

ID(ll)
{
   ip->dst[0].flags = REG_INT; ip->dst[0].reg = RTRI(ip->instr);
   ip->src[0].flags = REG_INT; ip->src[0].reg = RSRI(ip->instr); 
   ip->flags = INST_LOAD;
}
EX(ll)
{
   int immediate     = IMMED(ip->instr);
   Reg base = GPR(ip->src[0].preg);
   VA vAddr = SIGN_EXTEND(16, immediate) + base;
   int inum = ip - st->iwin;
   if ((st->iwin_addr[inum] & 0x3) || 
       (st->iwin_flags[inum] & IWIN_UNCACHED)) {	    
      EX_RECORD_EXCEPTION(st, inum, EXC_RADE, vAddr);
   } else if (st->iwin_flags[inum] & IWIN_TLBFAULT) {
      st->iwin_flags[inum] |= IWIN_FAULT;
   }
   if (st->iwin_flags[inum] & IWIN_FAULT) {
      EX_HANDLE_EXCEPTION(st, ip, th);
      return SUCCESS;	    
   }	
   ms_lsq (st,vAddr, &ip->dst[0], 0, LD_INTEGER_LL, inum);
   LOAD_STATS(st, ip->dst[0].preg);
   return SUCCESS;
}

ID(lld)
{
   ip->dst[0].flags = REG_INT; ip->dst[0].reg = RTRI(ip->instr);
   ip->src[0].flags = REG_INT; ip->src[0].reg = RSRI(ip->instr); 
   ip->flags = INST_LOAD;
}
EX(lld)
{
   int immediate     = IMMED(ip->instr);
   Reg base = GPR(ip->src[0].preg);
   VA vAddr = SIGN_EXTEND(16, immediate) + base;
   int inum = ip - st->iwin;
   if ((st->iwin_addr[inum] & 0x7) || 
       (st->iwin_flags[inum] & IWIN_UNCACHED)) {	    
      EX_RECORD_EXCEPTION(st, inum, EXC_RADE, vAddr);
   } else if (st->iwin_flags[inum] & IWIN_TLBFAULT) {
      st->iwin_flags[inum] |= IWIN_FAULT;
   }
   if (st->iwin_flags[inum] & IWIN_FAULT) {
      EX_HANDLE_EXCEPTION(st, ip, th);
      return SUCCESS;	    
   }	
   ms_lsq (st,vAddr, &ip->dst[0], 0, LD_INTEGER_LLD, inum);
   LOAD_STATS(st, ip->dst[0].preg);
   return SUCCESS;
}

ID(lw)
{
   ip->dst[0].flags = REG_INT; ip->dst[0].reg = RTRI(ip->instr);
   ip->src[0].flags = REG_INT; ip->src[0].reg = RSRI(ip->instr); 
   ip->flags = INST_LOAD;
}
EX(lw)
{
   int immediate     = IMMED(ip->instr);
   Reg base = GPR(ip->src[0].preg);
   VA vAddr  = SIGN_EXTEND(16, immediate) + base;
   int inum = ip - st->iwin;
   if (st->iwin_addr[inum] & 0x3) {	    
      EX_RECORD_EXCEPTION(st, inum, EXC_RADE, vAddr);
   } else if (st->iwin_flags[inum] & IWIN_TLBFAULT) {
      st->iwin_flags[inum] |= IWIN_FAULT;
   }
   if (st->iwin_flags[inum] & IWIN_FAULT) {
      EX_HANDLE_EXCEPTION(st, ip, th);
      return SUCCESS;
   }
   ms_lsq (st, vAddr, &ip->dst[0], 0, LD_INTEGER_W, inum);
   LOAD_STATS(st, ip->dst[0].preg);
   return SUCCESS;
}

ID(ld)
{
   ip->dst[0].flags = REG_INT; ip->dst[0].reg = RTRI(ip->instr);
   ip->src[0].flags = REG_INT; ip->src[0].reg = RSRI(ip->instr); 
   ip->flags = INST_LOAD;
}
EX(ld)
{
   int immediate     = IMMED(ip->instr);
   Reg base = GPR(ip->src[0].preg);
   VA vAddr = SIGN_EXTEND(16, immediate) + base;
   int inum = ip - st->iwin;
   
   if (!ENABLE_64BIT(st)) {
      EX_RECORD_EXCEPTION(st, inum, EXC_II, 0);
   } else if (st->iwin_addr[inum] & 0x7) {
      EX_RECORD_EXCEPTION(st, inum, EXC_RADE, vAddr);
   } else if (st->iwin_flags[inum] & IWIN_TLBFAULT) {
      st->iwin_flags[inum] |= IWIN_FAULT;
   }
   if (st->iwin_flags[inum] & IWIN_FAULT) {
      EX_HANDLE_EXCEPTION(st, ip, th);
      return SUCCESS;
   }
   ms_lsq (st, vAddr, &ip->dst[0], 0, LD_INTEGER_D,inum);
   LOAD_STATS(st, ip->dst[0].preg);
   return SUCCESS;
}

ID(ldc2)
{
   ip->flags = INST_ILLEGAL;
}
EX(ldc2)
{
   ASSERT(0);
   return SUCCESS;
}

ID(lwc1)
{
   ip->dst[0].flags = REG_FPS; ip->dst[0].reg = FTRI(ip->instr);
   ip->src[0].flags = REG_INT; ip->src[0].reg = RSRI(ip->instr); 
   ip->flags = INST_LOAD;
   /* Need the other half as a source if we're using register pairs */
#ifdef rpb_notdef
   ip->dst[1].flags = REG_FPW; ip->dst[1].reg = REG_PAIR(FTRI(ip->instr)); 
#endif
   ip->src[1].flags = REG_FPW; ip->src[1].reg = REG_PAIR(FTRI(ip->instr)); 
   ip->flags |= INST_HALF_DEF;
}
EX(lwc1)
{
   int immediate     = IMMED(ip->instr);
   int inum = ip - st->iwin;
   Reg base = GPR(ip->src[0].preg);
   VA vAddr = SIGN_EXTEND(16, immediate) + base;

   if (!ENABLE_FPU(st)) {   
      EX_RECORD_EXCEPTION(st, inum, EXC_CPU, 1);
   } else if (st->iwin_addr[inum] & 0x3) {
      EX_RECORD_EXCEPTION(st, inum, EXC_RADE, vAddr);
   } else if (st->iwin_flags[inum] & IWIN_TLBFAULT) {
      st->iwin_flags[inum] |= IWIN_FAULT;
   }
   if (st->iwin_flags[inum] & IWIN_FAULT) {
      EX_HANDLE_EXCEPTION(st, ip, th);
      return SUCCESS;
   }

   if ((ip->flags & INST_HALF_DEF) && MXS_NFRB) {
      int pair = REG_PAIR(ip->dst[0].fpreg);
      int old_pair = ip->src[1].fpreg;
      Reg32 contents_pair = FPR_W(old_pair);
      FPR_W(pair) = contents_pair;
      SIM_DEBUG(('f', "C1: lwc1 put pair f%d %d into f%d -> %d\n",
                 old_pair, contents_pair, pair, FPR_W(pair)));
   }

   ms_lsq (st, vAddr, &ip->dst[0], 0, LD_FLOAT, inum);
   LOAD_STATS(st, ip->dst[0].preg);
   return SUCCESS;
}

ID(ldc1)
{
   ip->dst[0].flags = REG_FPD; ip->dst[0].reg = FTRI(ip->instr);
   ip->src[0].flags = REG_INT; ip->src[0].reg = RSRI(ip->instr); 
   ip->flags = INST_LOAD;
}
EX(ldc1)
{
   int immediate     = IMMED(ip->instr);
   int inum = ip - st->iwin;
   Reg base = GPR(ip->src[0].preg);
   VA vAddr = SIGN_EXTEND(16, immediate) + base;
   
   if (!ENABLE_FPU(st)) {   
      EX_RECORD_EXCEPTION(st, inum, EXC_CPU, 1);
   } else if (st->iwin_addr[inum] & 0x7) {
      EX_RECORD_EXCEPTION(st, inum, EXC_RADE, vAddr);
   } else if (st->iwin_flags[inum] & IWIN_TLBFAULT) {
      st->iwin_flags[inum] |= IWIN_FAULT;
   }
   if (st->iwin_flags[inum] & IWIN_FAULT) {
      EX_HANDLE_EXCEPTION(st, ip, th);
      return SUCCESS;
   }
   ms_lsq (st, vAddr, &ip->dst[0], 0, LD_DOUBLE, inum);
   LOAD_STATS(st, ip->dst[0].preg);
   return SUCCESS;
}

ID(ldl)
{
   ip->dst[0].flags = REG_INT; ip->dst[0].reg = RTRI(ip->instr);
   ip->src[0].flags = REG_INT; ip->src[0].reg = RSRI(ip->instr); 
   ip->src[1].flags = REG_INT; ip->src[1].reg = RTRI(ip->instr);
   ip->flags = INST_LOAD;
}
EX(ldl)
{
   int immediate     = IMMED(ip->instr);
   Reg  base = GPR(ip->src[0].preg);
   VA   vAddr = SIGN_EXTEND(16, immediate) + base;
   int inum = ip - st->iwin;

   if (!ENABLE_64BIT(st)) {
      EX_RECORD_EXCEPTION(st, inum, EXC_II, 0);
   } else if (st->iwin_flags[inum] & IWIN_TLBFAULT) {
      st->iwin_flags[inum] |= IWIN_FAULT;
   }
   if (st->iwin_flags[inum] & IWIN_FAULT) {
      EX_HANDLE_EXCEPTION(st, ip, th);
      return SUCCESS;
   }
   ms_lsq (st, vAddr, &ip->dst[0], &ip->src[1], LD_UNSIGNED_DL, inum);
   LOAD_STATS(st, ip->dst[0].preg);
   return SUCCESS;
}

ID(ldr)
{
   ip->dst[0].flags = REG_INT; ip->dst[0].reg = RTRI(ip->instr);
   ip->src[0].flags = REG_INT; ip->src[0].reg = RSRI(ip->instr); 
   ip->src[1].flags = REG_INT; ip->src[1].reg = RTRI(ip->instr);
   ip->flags = INST_LOAD;
}
EX(ldr)
{
   int immediate     = IMMED(ip->instr);
   Reg base = GPR(ip->src[0].preg);
   VA  vAddr = SIGN_EXTEND(16, immediate) + base;
   int inum = ip - st->iwin;

   if (!ENABLE_64BIT(st)) {
      EX_RECORD_EXCEPTION(st, inum, EXC_II, 0);
   } else if (st->iwin_flags[inum] & IWIN_TLBFAULT) {
      st->iwin_flags[inum] |= IWIN_FAULT;
   }
   if (st->iwin_flags[inum] & IWIN_FAULT) {
      EX_HANDLE_EXCEPTION(st, ip, th);
      return SUCCESS;
   }
   
   ms_lsq (st, vAddr, &ip->dst[0], &ip->src[1], LD_UNSIGNED_DR, inum);
   LOAD_STATS(st, ip->dst[0].preg);
   return SUCCESS;
}

ID(lwc2)
{
   ip->flags = INST_ILLEGAL;
}
EX(lwc2)
{
   ASSERT(0);
   return SUCCESS;
}

ID(lwl) 
{
   ip->dst[0].flags = REG_INT; ip->dst[0].reg = RTRI(ip->instr);
   ip->src[0].flags = REG_INT; ip->src[0].reg = RSRI(ip->instr); 
   ip->src[1].flags = REG_INT; ip->src[1].reg = RTRI(ip->instr);
   ip->flags = INST_LOAD;
}
EX(lwl) 
{
   int immediate     = IMMED(ip->instr);
   Reg base = GPR(ip->src[0].preg);
   VA  vAddr = SIGN_EXTEND(16, immediate) + base;
   int inum = ip - st->iwin;
   
   if (st->iwin_flags[inum] & IWIN_TLBFAULT) {
      st->iwin_flags[inum] |= IWIN_FAULT;
      EX_HANDLE_EXCEPTION(st, ip, th);
      return SUCCESS;
   }
   
   ms_lsq (st, vAddr, &ip->dst[0], &ip->src[1], LD_UNSIGNED_L, inum);
   LOAD_STATS(st, ip->dst[0].preg);
   return SUCCESS;
}

ID(lwr) 
{
   ip->dst[0].flags = REG_INT; ip->dst[0].reg = RTRI(ip->instr);
   ip->src[0].flags = REG_INT; ip->src[0].reg = RSRI(ip->instr); 
   ip->src[1].flags = REG_INT; ip->src[1].reg = RTRI(ip->instr);
   ip->flags = INST_LOAD;
}
EX(lwr) 
{
   int immediate     = IMMED(ip->instr);
   Reg base = GPR(ip->src[0].preg);
   VA vAddr = SIGN_EXTEND(16, immediate) + base;
   int inum = ip - st->iwin;
   
   if (st->iwin_flags[inum] & IWIN_TLBFAULT) {
      st->iwin_flags[inum] |= IWIN_FAULT;
      EX_HANDLE_EXCEPTION(st, ip, th);
      return SUCCESS;
   }
   ms_lsq (st, vAddr, &ip->dst[0], &ip->src[1], LD_UNSIGNED_R, inum);
   LOAD_STATS(st, ip->dst[0].preg);
   return SUCCESS;
}

ID(lwu)
{
   ip->dst[0].flags = REG_INT; ip->dst[0].reg = RTRI(ip->instr);
   ip->src[0].flags = REG_INT; ip->src[0].reg = RSRI(ip->instr); 
   ip->flags = INST_LOAD;
}
EX(lwu)
{
   int immediate     = IMMED(ip->instr);
   Reg base = GPR(ip->src[0].preg);
   VA vAddr = SIGN_EXTEND(16, immediate) + base;
   int inum = ip - st->iwin;

   if (!ENABLE_64BIT(st)) {
      EX_RECORD_EXCEPTION(st, inum, EXC_II, 0);
   } else if (st->iwin_addr[inum] & 0x3) {
      EX_RECORD_EXCEPTION(st, inum, EXC_RADE, vAddr);
   } else if (st->iwin_flags[inum] & IWIN_TLBFAULT) {
      st->iwin_flags[inum] |= IWIN_FAULT;
   }
   if (st->iwin_flags[inum] & IWIN_FAULT) {
      EX_HANDLE_EXCEPTION(st, ip, th);
      return SUCCESS;
   }
   ms_lsq (st, vAddr, &ip->dst[0], 0, LD_UNSIGNED_W, inum);
   LOAD_STATS(st, ip->dst[0].preg);
   return SUCCESS;
}

ID(pref)
{
   ip->src[0].flags = REG_INT; ip->src[0].reg = RSRI(ip->instr); 
   ip->flags = INST_PREFETCH;
}
EX(pref)
{
   int immediate     = IMMED(ip->instr);
   Reg base = GPR(ip->src[0].preg);
   VA vAddr = (SIGN_EXTEND(16, immediate) + base);
   int inum = ip - st->iwin;
   ms_lsq (st, vAddr, &ip->src[0], 0, PREFETCH, inum);
   return SUCCESS;
}
            
ID(ori)
{
   ip->dst[0].flags = REG_INT; ip->dst[0].reg = RTRI(ip->instr);
   ip->src[0].flags = REG_INT; ip->src[0].reg = RSRI(ip->instr); 
}
EX(ori)
{
   int immediate     = IMMED(ip->instr);
   uint  rt          = ip->dst[0].preg;
   Reg_s contents_rs = GPR(ip->src[0].preg);
   GPR(rt) = contents_rs | (0x0000ffff & immediate);
   return SUCCESS;
}

ID(sb)
{
   ip->src[0].flags = REG_INT; ip->src[0].reg = RSRI(ip->instr); 
   ip->src[1].flags = REG_INT; ip->src[1].reg = RTRI(ip->instr);
   ip->src[1].flags |= REG_STORE;
   ip->flags = INST_STORE;
}
EX(sb)
{
   int immediate     = IMMED(ip->instr);
   Reg base = GPR(ip->src[0].preg);
   VA vAddr = SIGN_EXTEND(16, immediate) + base;
   int inum = ip - st->iwin;
   if (st->iwin_flags[inum] & IWIN_TLBFAULT) {
      st->iwin_flags[inum] |= IWIN_FAULT;
      EX_HANDLE_EXCEPTION(st, ip, th);
      return SUCCESS;
   }
   ms_lsq (st, vAddr, &ip->src[1], 0, ST_INTEGER_B, inum);
   return SUCCESS;
}

ID(sh)
{
   ip->src[0].flags = REG_INT; ip->src[0].reg = RSRI(ip->instr); 
   ip->src[1].flags = REG_INT; ip->src[1].reg = RTRI(ip->instr);
   ip->src[1].flags |= REG_STORE;
   ip->flags = INST_STORE;
}
EX(sh)
{
   int immediate     = IMMED(ip->instr);
   Reg base = GPR(ip->src[0].preg);
   VA vAddr = SIGN_EXTEND(16, immediate) + base;
   int inum = ip - st->iwin;
   if (st->iwin_addr[inum] & 0x1) {
      EX_RECORD_EXCEPTION(st, inum, EXC_WADE, vAddr);
   } else if (st->iwin_flags[inum] & IWIN_TLBFAULT) {
      st->iwin_flags[inum] |= IWIN_FAULT;
   }
   
   if (st->iwin_flags[inum] & IWIN_FAULT) {
      EX_HANDLE_EXCEPTION(st, ip, th);
      return SUCCESS;
   }
   ms_lsq (st, vAddr, &ip->src[1], 0, ST_INTEGER_H, inum);
   return SUCCESS;
}
ID(slti)
{
   ip->dst[0].flags = REG_INT; ip->dst[0].reg = RTRI(ip->instr);
   ip->src[0].flags = REG_INT; ip->src[0].reg = RSRI(ip->instr); 
}
EX(slti)
{
   int immediate     = IMMED(ip->instr);
   uint  rt          = ip->dst[0].preg;
   Reg_s contents_rs = GPR(ip->src[0].preg);
   if (contents_rs < (Reg32_s)SIGN_EXTEND(16, immediate)) {
      GPR(rt) = 1;
   } else {
      GPR(rt) = 0;
   }
   return SUCCESS;
}

ID(sltiu)
{
   ip->dst[0].flags = REG_INT; ip->dst[0].reg = RTRI(ip->instr);
   ip->src[0].flags = REG_INT; ip->src[0].reg = RSRI(ip->instr); 
}
EX(sltiu)
{
   int immediate     = IMMED(ip->instr);
   uint  rt          = ip->dst[0].preg;
   Reg_s contents_rs = GPR(ip->src[0].preg);
   if ((Reg)contents_rs < (Reg)(Reg32_s)SIGN_EXTEND(16, immediate)) {
      GPR(rt) = 1;
   } else {
      GPR(rt) = 0;
   }
   return SUCCESS;
}

ID(sc)
{
   ip->dst[0].flags = REG_INT; ip->dst[0].reg = RTRI(ip->instr);
   ip->src[0].flags = REG_INT; ip->src[0].reg = RSRI(ip->instr); 
   ip->src[1].flags = REG_INT; ip->src[1].reg = RTRI(ip->instr);
   ip->src[1].flags |= REG_STORE;
   ip->flags = INST_STORE | INST_SC;
}
EX(sc)
{
   int immediate     = IMMED(ip->instr);
   Reg base = GPR(ip->src[0].preg);
   VA vAddr = SIGN_EXTEND(16, immediate) + base;
   int inum = ip - st->iwin;

   if ((st->iwin_addr[inum] & 0x3) || 
       (st->iwin_flags[inum] & IWIN_UNCACHED)) {	    
      EX_RECORD_EXCEPTION(st, inum, EXC_WADE, vAddr);
   } else if (st->iwin_flags[inum] & IWIN_TLBFAULT) {
      st->iwin_flags[inum] |= IWIN_FAULT;
   }
   if (st->iwin_flags[inum] & IWIN_FAULT) {
      th->stall_fpc = 0;
      EX_HANDLE_EXCEPTION(st, ip, th);
      return SUCCESS;
   }	
   
   ms_lsq (st, vAddr, &ip->src[1], &ip->dst[0], ST_INTEGER_SC, inum);
   return SUCCESS;
}

ID(scd)
{
   ip->dst[0].flags = REG_INT; ip->dst[0].reg = RTRI(ip->instr);
   ip->src[0].flags = REG_INT; ip->src[0].reg = RSRI(ip->instr); 
   ip->src[1].flags = REG_INT; ip->src[1].reg = RTRI(ip->instr);
   ip->src[1].flags |= REG_STORE;
   ip->flags = INST_STORE | INST_SC;
}
EX(scd)
{
   int immediate     = IMMED(ip->instr);
   Reg base = GPR(ip->src[0].preg);
   VA vAddr = SIGN_EXTEND(16, immediate) + base;
   int inum = ip - st->iwin;

   if (!ENABLE_64BIT(st)) {
      EX_RECORD_EXCEPTION(st, inum, EXC_II, 0);
   } else if ((st->iwin_addr[inum] & 0x7) || 
       (st->iwin_flags[inum] & IWIN_UNCACHED)) {	    
      EX_RECORD_EXCEPTION(st, inum, EXC_WADE, vAddr);
   } else if (st->iwin_flags[inum] & IWIN_TLBFAULT) {
      st->iwin_flags[inum] |= IWIN_FAULT;
   }
   if (st->iwin_flags[inum] & IWIN_FAULT) {
      th->stall_fpc = 0;
      EX_HANDLE_EXCEPTION(st, ip, th);
      return SUCCESS; 
   }	
   
   ms_lsq (st, vAddr, &ip->src[1], &ip->dst[0], ST_INTEGER_SCD, inum);
   return SUCCESS;
}

ID(sw)
{
   ip->src[0].flags = REG_INT; ip->src[0].reg = RSRI(ip->instr); 
   ip->src[1].flags = REG_INT; ip->src[1].reg = RTRI(ip->instr);
   ip->src[1].flags |= REG_STORE;
   ip->flags = INST_STORE;
}
EX(sw)
{
   int immediate     = IMMED(ip->instr);
   Reg base = GPR(ip->src[0].preg);
   VA vAddr = SIGN_EXTEND(16, immediate) + base;
   int inum = ip - st->iwin;
   if (st->iwin_addr[inum] & 0x3) {
      EX_RECORD_EXCEPTION(st, inum, EXC_WADE, vAddr);
   } else if (st->iwin_flags[inum] & IWIN_TLBFAULT) {
      st->iwin_flags[inum] |= IWIN_FAULT;
   }
   
   if (st->iwin_flags[inum] & IWIN_FAULT) {
      EX_HANDLE_EXCEPTION(st, ip, th);
      return SUCCESS;
   }
   ms_lsq (st, vAddr, &ip->src[1], 0, ST_INTEGER_W, inum);
   return SUCCESS;
}


ID(sd)
{
   ip->src[0].flags = REG_INT; ip->src[0].reg = RSRI(ip->instr); 
   ip->src[1].flags = REG_INT; ip->src[1].reg = RTRI(ip->instr);
   ip->src[1].flags |= REG_STORE;
   ip->flags = INST_STORE;
}
EX(sd)
{
   int immediate     = IMMED(ip->instr);
   Reg base = GPR(ip->src[0].preg);
   VA vAddr = SIGN_EXTEND(16, immediate) + base;
   int inum = ip - st->iwin;
   
   if (!ENABLE_64BIT(st)) {
      EX_RECORD_EXCEPTION(st, inum, EXC_II, 0);
   } else if (st->iwin_addr[inum] & 0x7) {
      EX_RECORD_EXCEPTION(st, inum, EXC_WADE, vAddr);
   } else if (st->iwin_flags[inum] & IWIN_TLBFAULT) {
      st->iwin_flags[inum] |= IWIN_FAULT;
   }
   if (st->iwin_flags[inum] & IWIN_FAULT) {
      EX_HANDLE_EXCEPTION(st, ip, th);
      return SUCCESS;
   }
   ms_lsq (st, vAddr, &ip->src[1], 0, ST_INTEGER_D, inum);
   return SUCCESS;
}

ID(sdc2)
{
   ip->flags = INST_ILLEGAL;
}
EX(sdc2)
{
   ASSERT(0);
   return SUCCESS;
}

ID(sdc1)
{
   ip->src[0].flags = REG_INT; ip->src[0].reg = RSRI(ip->instr); 
   ip->src[1].flags = REG_FPD; ip->src[1].reg = FTRI(ip->instr);
   ip->src[1].flags |= REG_STORE;
   ip->flags = INST_STORE;
}
EX(sdc1)
{
   int immediate     = IMMED(ip->instr);
   Reg base = GPR(ip->src[0].preg);
   VA vAddr = SIGN_EXTEND(16, immediate) + base;
   int inum = ip - st->iwin;

   if (!ENABLE_FPU(st)) {   
      EX_RECORD_EXCEPTION(st, inum, EXC_CPU, 1);
   } else if (st->iwin_addr[inum] & 0x7) {
      EX_RECORD_EXCEPTION(st, inum, EXC_WADE, vAddr);
   } else if (st->iwin_flags[inum] & IWIN_TLBFAULT) {
      st->iwin_flags[inum] |= IWIN_FAULT;
   }
   if (st->iwin_flags[inum] & IWIN_FAULT) {
      EX_HANDLE_EXCEPTION(st, ip, th);
      return SUCCESS;
   }
   ms_lsq (st, vAddr, &ip->src[1], 0, ST_DOUBLE, inum);
   return SUCCESS;
}

ID(sdl)
{
   ip->src[0].flags = REG_INT; ip->src[0].reg = RSRI(ip->instr); 
   ip->src[1].flags = REG_INT; ip->src[1].reg = RTRI(ip->instr);
   ip->src[1].flags |= REG_STORE;
   ip->flags = INST_STORE;
}
EX(sdl)
{
   int immediate     = IMMED(ip->instr);
   Reg base = GPR(ip->src[0].preg);
   VA vAddr = SIGN_EXTEND(16, immediate) + base;
   int inum = ip - st->iwin;
   
   if (!ENABLE_64BIT(st)) {
      EX_RECORD_EXCEPTION(st, inum, EXC_II, 0);
   } else if (st->iwin_flags[inum] & IWIN_TLBFAULT) {
      st->iwin_flags[inum] |= IWIN_FAULT;
   }
   if (st->iwin_flags[inum] & IWIN_FAULT) {
      EX_HANDLE_EXCEPTION(st, ip, th);
      return SUCCESS;
   }

   ms_lsq (st, vAddr, &ip->src[1], 0, ST_INTEGER_DL, inum);
   return SUCCESS;
}

ID(sdr)
{
   ip->src[0].flags = REG_INT; ip->src[0].reg = RSRI(ip->instr); 
   ip->src[1].flags = REG_INT; ip->src[1].reg = RTRI(ip->instr);
   ip->src[1].flags |= REG_STORE;
   ip->flags = INST_STORE;
}
EX(sdr)
{
   int immediate     = IMMED(ip->instr);
   Reg base = GPR(ip->src[0].preg);
   VA vAddr = SIGN_EXTEND(16, immediate) + base;
   int inum = ip - st->iwin;
   
   if (!ENABLE_64BIT(st)) {
      EX_RECORD_EXCEPTION(st, inum, EXC_II, 0);
   } else if (st->iwin_flags[inum] & IWIN_TLBFAULT) {
      st->iwin_flags[inum] |= IWIN_FAULT;
   }
   if (st->iwin_flags[inum] & IWIN_FAULT) {
      EX_HANDLE_EXCEPTION(st, ip, th);
      return SUCCESS;
   }
   ms_lsq (st, vAddr, &ip->src[1], 0, ST_INTEGER_DR, inum);
   return SUCCESS;
}

ID(swc1)
{
   ip->src[0].flags = REG_INT; ip->src[0].reg = RSRI(ip->instr); 
   ip->src[1].flags = REG_FPS; ip->src[1].reg = FTRI(ip->instr);
   ip->src[1].flags |= REG_STORE;
   ip->flags = INST_STORE;
}
EX(swc1)
{
   int immediate     = IMMED(ip->instr);
   Reg base = GPR(ip->src[0].preg);
   VA vAddr = SIGN_EXTEND(16, immediate) + base;
   int inum = ip - st->iwin;

   if (!ENABLE_FPU(st)) {   
      EX_RECORD_EXCEPTION(st, inum, EXC_CPU, 1);
   } else if (st->iwin_addr[inum] & 0x3) {
      EX_RECORD_EXCEPTION(st, inum, EXC_WADE, vAddr);
   } else if (st->iwin_flags[inum] & IWIN_TLBFAULT) {
      st->iwin_flags[inum] |= IWIN_FAULT;
   }
   if (st->iwin_flags[inum] & IWIN_FAULT) {
      EX_HANDLE_EXCEPTION(st, ip, th);
      return SUCCESS;
   }
   ms_lsq (st, vAddr, &ip->src[1], 0, ST_FLOAT, inum);
   return SUCCESS;
}

ID(swc2)
{
   ip->flags = INST_ILLEGAL;
}
EX(swc2)
{
   ASSERT(0);
   return SUCCESS;
}

ID(swl)
{
   ip->src[0].flags = REG_INT; ip->src[0].reg = RSRI(ip->instr); 
   ip->src[1].flags = REG_INT; ip->src[1].reg = RTRI(ip->instr);
   ip->src[1].flags |= REG_STORE;
   ip->flags = INST_STORE;
}
EX(swl)
{
   int immediate     = IMMED(ip->instr);
   Reg base = GPR(ip->src[0].preg);
   VA vAddr = SIGN_EXTEND(16, immediate) + base;
   int inum = ip - st->iwin;

   if (st->iwin_flags[inum] & IWIN_TLBFAULT) {
      st->iwin_flags[inum] |= IWIN_FAULT;
      EX_HANDLE_EXCEPTION(st, ip, th);
      return SUCCESS;
   }

   ms_lsq (st, vAddr, &ip->src[1], 0, ST_INTEGER_L, inum);
   return SUCCESS;
}

ID(swr)
{
   ip->src[0].flags = REG_INT; ip->src[0].reg = RSRI(ip->instr); 
   ip->src[1].flags = REG_INT; ip->src[1].reg = RTRI(ip->instr);
   ip->src[1].flags |= REG_STORE;
   ip->flags = INST_STORE;
}
EX(swr)
{
   int immediate     = IMMED(ip->instr);
   Reg base = GPR(ip->src[0].preg);
   VA vAddr = SIGN_EXTEND(16, immediate) + base;
   int inum = ip - st->iwin;

   if (st->iwin_flags[inum] & IWIN_TLBFAULT) {
      st->iwin_flags[inum] |= IWIN_FAULT;
      EX_HANDLE_EXCEPTION(st, ip, th);
      return SUCCESS;
   }
   ms_lsq (st, vAddr, &ip->src[1], 0, ST_INTEGER_R, inum);
   return SUCCESS;
}

ID(xori)
{
   ip->dst[0].flags = REG_INT; ip->dst[0].reg = RTRI(ip->instr);
   ip->src[0].flags = REG_INT; ip->src[0].reg = RSRI(ip->instr); 
}
EX(xori)
{
   int immediate     = IMMED(ip->instr);
   uint  rt          = ip->dst[0].preg;
   Reg_s contents_rs = GPR(ip->src[0].preg);
   GPR(rt) = contents_rs ^ (0x0000ffff & immediate);
   return SUCCESS;
}

ID(cop0)
{
   switch (RSRI(ip->instr)) {
   case 0: /* mfc0 */
   case 1: /* dmfc0 */
      ip->dst[0].flags = REG_INT; ip->dst[0].reg = RTRI(ip->instr);
      break;
   case 4: /* mtc0 */
   case 5: /* dmtc0 */
      ip->src[0].flags = REG_INT; ip->src[0].reg = RTRI(ip->instr);
      break;
   default:
      break;
   }
   ip->flags = INST_CP0;
   if ((ip->instr & 0x3f) == 0x18) {
      ip->flags |= INST_ERET;
   }
}
EX(cop0)
{
   Reg srcreg, dstreg;
   bool hack, err;
   int inum = ip - st->iwin;

   if (!AllsQuiet(st, ip, ST_SYSCALL))
      return SUCCESS;
#ifdef PRECISE
   if (st->iwin_headgrad != (ip - st->iwin)) {
      st->stall_type = ST_SYSCALL;
      st->stall_issue = 1;
      return SUCCESS;
   }
#endif

   if (ip->src[0].flags != 0) {
      int rt = ip->src[0].preg;
      srcreg = GPR(rt);
   } else {
      srcreg = 0;
   }

   err = DoPrivInst(st, ip->instr, srcreg, &dstreg, &hack);
   if (err) {
      st->iwin_except [inum] = GetLastException(st);
      st->iwin_flags[inum] |= IWIN_FAULT;
      EX_HANDLE_EXCEPTION(st, ip, th);
   }
#ifndef R3K
   if (hack) {
      CPUState *P = (CPUState *)st->mipsyPtr;
      st->iwin_branch_pc[inum] = P->nPC;
      th->pc = P->nPC;
   }        
#endif
   if (ip->dst[0].flags != 0) {
      int rt = ip->dst[0].preg;
      GPR(rt) = dstreg;
   }
   th->stall_cp0 = 0;	
   ClearFPCStall(st, th, inum);
   return SUCCESS;
}

ID(cop1)
{
   MXSDecodeC1Instruction(ip, pc);
}
EX(cop1)
{
   int retval = MXSExecuteC1Instruction(st, ip, th);
   int inum = ip - st->iwin;
   if (st->iwin_flags[inum] & IWIN_FAULT) {
      ASSERT(retval == C1_FAILURE);
      EX_HANDLE_EXCEPTION(st, ip, th);
   } else {
      ASSERT(retval == C1_SUCCESS);
   }
   return SUCCESS;
   return SUCCESS;
}

ID(cop1x)
{
   MXSDecodeC1Instruction(ip, pc);
}
EX(cop1x)
{
   int retval = MXSExecuteC1Instruction(st, ip, th);
   int inum = ip - st->iwin;
   if (st->iwin_flags[inum] & IWIN_FAULT) {
      ASSERT(retval == C1_FAILURE);
      EX_HANDLE_EXCEPTION(st, ip, th);
   } else {
      ASSERT(retval == C1_SUCCESS);
   }
   return SUCCESS;
   return SUCCESS;
}


ID(cop2)
{
   ip->flags = INST_ILLEGAL;
}
EX(cop2)
{
   ASSERT(0);
   return SUCCESS;
}



ID(undef)
{
   ip->flags = INST_ILLEGAL;
}
EX(undef)
{
   int inum = ip - st->iwin;
   EX_RECORD_EXCEPTION(st, inum, EXC_II, 0);
   EX_HANDLE_EXCEPTION(st, ip, th);
   return SUCCESS;
   return SUCCESS;
}


/*****************************************************************
 * S P E C   O P C O D E S
 *****************************************************************/

ID(add)
{
   ip->dst[0].flags = REG_INT; ip->dst[0].reg = RDRI(ip->instr);
   ip->src[0].flags = REG_INT; ip->src[0].reg = RSRI(ip->instr); 
   ip->src[1].flags = REG_INT; ip->src[1].reg = RTRI(ip->instr);
}
EX(add)
{
   uint  rd          = ip->dst[0].preg;
   Reg_s contents_rs = GPR(ip->src[0].preg);
   Reg_s contents_rt = GPR(ip->src[1].preg);

   Reg32_s result32 = (Reg32_s) contents_rs + (Reg32_s)contents_rt;
   if (ARITH_OVFL32 (result32, (Reg32_s)contents_rs, 
                     (Reg32_s)contents_rt)) {
      int inum = ip - st->iwin;
      EX_RECORD_EXCEPTION(st, inum,  EXC_OV, 0);
      EX_HANDLE_EXCEPTION(st, ip, th);
      return SUCCESS;
   }
   /* Don't write this if there was an exception */
   GPR(rd) = result32;
   return SUCCESS;
}

ID(addu)
{
   ip->dst[0].flags = REG_INT; ip->dst[0].reg = RDRI(ip->instr);
   ip->src[0].flags = REG_INT; ip->src[0].reg = RSRI(ip->instr); 
   ip->src[1].flags = REG_INT; ip->src[1].reg = RTRI(ip->instr);
}
EX(addu)
{
   uint  rd          = ip->dst[0].preg;
   Reg_s contents_rs = GPR(ip->src[0].preg);
   Reg_s contents_rt = GPR(ip->src[1].preg);

   GPR(rd) = (Reg32_s)contents_rs + (Reg32_s)contents_rt;
   return SUCCESS;
}

ID(and)
{
   ip->dst[0].flags = REG_INT; ip->dst[0].reg = RDRI(ip->instr);
   ip->src[0].flags = REG_INT; ip->src[0].reg = RSRI(ip->instr); 
   ip->src[1].flags = REG_INT; ip->src[1].reg = RTRI(ip->instr);
}
EX(and)
{
   uint  rd          = ip->dst[0].preg;
   Reg_s contents_rs = GPR(ip->src[0].preg);
   Reg_s contents_rt = GPR(ip->src[1].preg);

   GPR(rd) = contents_rs & contents_rt;
   return SUCCESS;
}

ID(break) 
{
   ip->flags = INST_BREAK;
}
#define BREAK_CODE(_inst) (((unsigned)_inst >> 16 ) & 0x3ff )
EX(break) 
{
   int inum = ip - st->iwin;
   EX_RECORD_EXCEPTION(st, inum,  EXC_BREAK, 0);
   EX_HANDLE_EXCEPTION(st, ip, th);
   return SUCCESS;
}
     
ID(dadd)
{
   ip->dst[0].flags = REG_INT; ip->dst[0].reg = RDRI(ip->instr);
   ip->src[0].flags = REG_INT; ip->src[0].reg = RSRI(ip->instr); 
   ip->src[1].flags = REG_INT; ip->src[1].reg = RTRI(ip->instr);
}
EX(dadd)
{
   uint  rd          = ip->dst[0].preg;
   Reg_s contents_rs = GPR(ip->src[0].preg);
   Reg_s contents_rt = GPR(ip->src[1].preg);
   Reg64_s result64  = contents_rs + contents_rt;
   int inum = ip - st->iwin;

   if (!ENABLE_64BIT(st)) {
      EX_RECORD_EXCEPTION(st, inum, EXC_II, 0);
   } else if (ARITH_OVFL64 (result64, contents_rs, contents_rt)) {
      EX_RECORD_EXCEPTION(st, inum, EXC_OV, 0);
   }
   if (st->iwin_flags[inum] & IWIN_FAULT) {
      EX_HANDLE_EXCEPTION(st, ip, th);
      return SUCCESS;
   }

   /* Don't write this if there was an exception */
   GPR(rd) =  result64;
   return SUCCESS;
}

ID(daddu)
{
   ip->dst[0].flags = REG_INT; ip->dst[0].reg = RDRI(ip->instr);
   ip->src[0].flags = REG_INT; ip->src[0].reg = RSRI(ip->instr); 
   ip->src[1].flags = REG_INT; ip->src[1].reg = RTRI(ip->instr);
}
EX(daddu)
{
   uint  rd          = ip->dst[0].preg;
   Reg_s contents_rs = GPR(ip->src[0].preg);
   Reg_s contents_rt = GPR(ip->src[1].preg);
   
   if (!ENABLE_64BIT(st)) {
      int inum = ip - st->iwin;
      EX_RECORD_EXCEPTION(st, inum, EXC_II, 0);
      EX_HANDLE_EXCEPTION(st, ip, th);
      return SUCCESS;
   }
   GPR(rd) =  contents_rs + contents_rt;
   return SUCCESS;
}

ID(ddiv)
{
   ip->dst[0].flags = REG_INT; ip->dst[0].reg = MXS_REG_HI;
   ip->dst[1].flags = REG_INT; ip->dst[1].reg = MXS_REG_LO;
   ip->src[0].flags = REG_INT; ip->src[0].reg = RSRI(ip->instr); 
   ip->src[1].flags = REG_INT; ip->src[1].reg = RTRI(ip->instr);
}
EX(ddiv)
{
   uint  hi          = ip->dst[0].preg;
   uint  lo          = ip->dst[1].preg;
   Reg_s contents_rs = GPR(ip->src[0].preg);
   Reg_s contents_rt = GPR(ip->src[1].preg);

   if (!ENABLE_64BIT(st)) {
      int inum = ip - st->iwin;
      EX_RECORD_EXCEPTION(st, inum, EXC_II, 0);
      EX_HANDLE_EXCEPTION(st, ip, th);
      return SUCCESS;
   }
   if (contents_rt != 0) {
      GPR(lo) = (Reg_s)contents_rs / (Reg_s)contents_rt;
      GPR(hi) = (Reg_s)contents_rs % (Reg_s)contents_rt; 
   }
   return SUCCESS;
}

ID(ddivu)
{
   ip->dst[0].flags = REG_INT; ip->dst[0].reg = MXS_REG_HI;
   ip->dst[1].flags = REG_INT; ip->dst[1].reg = MXS_REG_LO;
   ip->src[0].flags = REG_INT; ip->src[0].reg = RSRI(ip->instr); 
   ip->src[1].flags = REG_INT; ip->src[1].reg = RTRI(ip->instr);
}
EX(ddivu)
{
   uint  hi          = ip->dst[0].preg;
   uint  lo          = ip->dst[1].preg;
   Reg_s contents_rs = GPR(ip->src[0].preg);
   Reg_s contents_rt = GPR(ip->src[1].preg);
   
   if (!ENABLE_64BIT(st)) {
      int inum = ip - st->iwin;
      EX_RECORD_EXCEPTION(st, inum, EXC_II, 0);
      EX_HANDLE_EXCEPTION(st, ip, th);
      return SUCCESS;
   }
   if (contents_rt != 0) {
      GPR(lo) = (Reg)contents_rs / (Reg)contents_rt;
      GPR(hi) = (Reg)contents_rs % (Reg)contents_rt; 
   }
   return SUCCESS;
}

ID(div)
{
   ip->dst[0].flags = REG_INT; ip->dst[0].reg = MXS_REG_HI;
   ip->dst[1].flags = REG_INT; ip->dst[1].reg = MXS_REG_LO;
   ip->src[0].flags = REG_INT; ip->src[0].reg = RSRI(ip->instr); 
   ip->src[1].flags = REG_INT; ip->src[1].reg = RTRI(ip->instr);
}
EX(div)
{
   uint  hi          = ip->dst[0].preg;
   uint  lo          = ip->dst[1].preg;
   Reg_s contents_rs = GPR(ip->src[0].preg);
   Reg_s contents_rt = GPR(ip->src[1].preg);

   if (contents_rt != 0) {
      GPR(lo) = (Reg32_s)contents_rs / (Reg32_s)contents_rt;
      GPR(hi) = (Reg32_s)contents_rs % (Reg32_s)contents_rt;
   }
   return SUCCESS;
}

ID(divu)
{
   ip->dst[0].flags = REG_INT; ip->dst[0].reg = MXS_REG_HI;
   ip->dst[1].flags = REG_INT; ip->dst[1].reg = MXS_REG_LO;
   ip->src[0].flags = REG_INT; ip->src[0].reg = RSRI(ip->instr); 
   ip->src[1].flags = REG_INT; ip->src[1].reg = RTRI(ip->instr);
}
EX(divu)
{
   uint  hi          = ip->dst[0].preg;
   uint  lo          = ip->dst[1].preg;
   Reg_s contents_rs = GPR(ip->src[0].preg);
   Reg_s contents_rt = GPR(ip->src[1].preg);

   if (contents_rt != 0) {
      GPR(lo) = (Reg32)contents_rs / (Reg32)contents_rt;
      GPR(hi) = (Reg32)contents_rs % (Reg32)contents_rt;
   }
   return SUCCESS;
}

ID(dmult)
{
   ip->dst[0].flags = REG_INT; ip->dst[0].reg = MXS_REG_HI;
   ip->dst[1].flags = REG_INT; ip->dst[1].reg = MXS_REG_LO;
   ip->src[0].flags = REG_INT; ip->src[0].reg = RTRI(ip->instr); 
   ip->src[1].flags = REG_INT; ip->src[1].reg = RSRI(ip->instr);
}
EX(dmult)
{
   uint  hi          = ip->dst[0].preg;
   uint  lo          = ip->dst[1].preg;
   Reg_s contents_rs = GPR(ip->src[1].preg);
   Reg_s contents_rt = GPR(ip->src[0].preg);
   bool negative = FALSE;

   if (!ENABLE_64BIT(st)) {
      int inum = ip - st->iwin;
      EX_RECORD_EXCEPTION(st, inum, EXC_II, 0);
      EX_HANDLE_EXCEPTION(st, ip, th);
      return SUCCESS;
   }
   
   if ((Reg64_s)contents_rs < 0) {
      negative = TRUE;
      contents_rs = -contents_rs;
   }
   if ((Reg64_s)contents_rt < 0) {
      negative = !negative;
      contents_rt = -contents_rt;
   }
   LongMultiply64(st, contents_rs, contents_rt, lo, hi);
   if (negative) {
      GPR(lo) = ~GPR(lo);
      GPR(hi) = ~GPR(hi);
      GPR(lo) += 1;
      if (GPR(lo) == 0)
         GPR(hi) += 1;
   }
   return SUCCESS;
}

ID(dmultu)
{
   ip->dst[0].flags = REG_INT; ip->dst[0].reg = MXS_REG_HI;
   ip->dst[1].flags = REG_INT; ip->dst[1].reg = MXS_REG_LO;
   ip->src[0].flags = REG_INT; ip->src[0].reg = RTRI(ip->instr); 
   ip->src[1].flags = REG_INT; ip->src[1].reg = RSRI(ip->instr);
}
EX(dmultu)
{
   uint  hi          = ip->dst[0].preg;
   uint  lo          = ip->dst[1].preg;
   Reg_s contents_rs = GPR(ip->src[1].preg);
   Reg_s contents_rt = GPR(ip->src[0].preg);

   if (!ENABLE_64BIT(st)) {
      int inum = ip - st->iwin;
      EX_RECORD_EXCEPTION(st, inum, EXC_II, 0);
      EX_HANDLE_EXCEPTION(st, ip, th);
      return SUCCESS;
   }
   LongMultiply64(st, (Reg64)contents_rs, (Reg64)contents_rt, lo, hi);
   return SUCCESS;
}

ID(dsll)
{
   ip->dst[0].flags = REG_INT; ip->dst[0].reg = RDRI(ip->instr);
   ip->src[0].flags = REG_INT; ip->src[0].reg = RTRI(ip->instr); 
}
EX(dsll)
{
   uint  rd          = ip->dst[0].preg;
   Reg_s contents_rt = GPR(ip->src[0].preg);
   
   if (!ENABLE_64BIT(st)) {
      int inum = ip - st->iwin;
      EX_RECORD_EXCEPTION(st, inum, EXC_II, 0);
      EX_HANDLE_EXCEPTION(st, ip, th);
      return SUCCESS;
   }
   GPR(rd) = (Reg_s)contents_rt << SHAMT(ip->instr);
   return SUCCESS;
}

ID(dsllv)
{
   ip->dst[0].flags = REG_INT; ip->dst[0].reg = RDRI(ip->instr);
   ip->src[0].flags = REG_INT; ip->src[0].reg = RTRI(ip->instr); 
   ip->src[1].flags = REG_INT; ip->src[1].reg = RSRI(ip->instr);
}
EX(dsllv)
{
   uint  rd          = ip->dst[0].preg;
   Reg_s contents_rs = GPR(ip->src[1].preg);
   Reg_s contents_rt = GPR(ip->src[0].preg);

   if (!ENABLE_64BIT(st)) {
      int inum = ip - st->iwin;
      EX_RECORD_EXCEPTION(st, inum, EXC_II, 0);
      EX_HANDLE_EXCEPTION(st, ip, th);
      return SUCCESS;
   }
   GPR(rd) = (Reg_s)contents_rt << (contents_rs & 0x3f);
   return SUCCESS;
}

ID(dsll32)
{
   ip->dst[0].flags = REG_INT; ip->dst[0].reg = RDRI(ip->instr);
   ip->src[0].flags = REG_INT; ip->src[0].reg = RTRI(ip->instr); 
}
EX(dsll32)
{
   uint  rd          = ip->dst[0].preg;
   Reg_s contents_rt = GPR(ip->src[0].preg);
   
   if (!ENABLE_64BIT(st)) {
      int inum = ip - st->iwin;
      EX_RECORD_EXCEPTION(st, inum, EXC_II, 0);
      EX_HANDLE_EXCEPTION(st, ip, th);
      return SUCCESS;
   }
   GPR(rd) = (Reg_s)contents_rt << (SHAMT(ip->instr)+32);
   return SUCCESS;
}

ID(dsra)
{
   ip->dst[0].flags = REG_INT; ip->dst[0].reg = RDRI(ip->instr);
   ip->src[0].flags = REG_INT; ip->src[0].reg = RTRI(ip->instr); 
}
EX(dsra)
{
   uint  rd          = ip->dst[0].preg;
   Reg_s contents_rt = GPR(ip->src[0].preg);

   if (!ENABLE_64BIT(st)) {
      int inum = ip - st->iwin;
      EX_RECORD_EXCEPTION(st, inum, EXC_II, 0);
      EX_HANDLE_EXCEPTION(st, ip, th);
      return SUCCESS;
   }
   GPR(rd) = (Reg64_s)contents_rt >> SHAMT(ip->instr);
   return SUCCESS;
}

ID(dsrav)
{
   ip->dst[0].flags = REG_INT; ip->dst[0].reg = RDRI(ip->instr);
   ip->src[0].flags = REG_INT; ip->src[0].reg = RTRI(ip->instr); 
   ip->src[1].flags = REG_INT; ip->src[1].reg = RSRI(ip->instr);
}
EX(dsrav)
{
   uint  rd          = ip->dst[0].preg;
   Reg_s contents_rs = GPR(ip->src[1].preg);
   Reg_s contents_rt = GPR(ip->src[0].preg);

   if (!ENABLE_64BIT(st)) {
      int inum = ip - st->iwin;
      EX_RECORD_EXCEPTION(st, inum, EXC_II, 0);
      EX_HANDLE_EXCEPTION(st, ip, th);
      return SUCCESS;
   }
   GPR(rd) = (Reg64_s)contents_rt >> (contents_rs & 0x3f);
   return SUCCESS;
}

ID(dsra32)
{
   ip->dst[0].flags = REG_INT; ip->dst[0].reg = RDRI(ip->instr);
   ip->src[0].flags = REG_INT; ip->src[0].reg = RTRI(ip->instr); 
}
EX(dsra32)
{
   uint  rd          = ip->dst[0].preg;
   Reg_s contents_rt = GPR(ip->src[0].preg);

   if (!ENABLE_64BIT(st)) {
      int inum = ip - st->iwin;
      EX_RECORD_EXCEPTION(st, inum, EXC_II, 0);
      EX_HANDLE_EXCEPTION(st, ip, th);
      return SUCCESS;
   }
   GPR(rd) = (Reg64_s)contents_rt >> (SHAMT(ip->instr)+32);
   return SUCCESS;
}

ID(dsrl)
{
   ip->dst[0].flags = REG_INT; ip->dst[0].reg = RDRI(ip->instr);
   ip->src[0].flags = REG_INT; ip->src[0].reg = RTRI(ip->instr); 
}
EX(dsrl)
{
   uint  rd          = ip->dst[0].preg;
   Reg_s contents_rt = GPR(ip->src[0].preg);

   if (!ENABLE_64BIT(st)) {
      int inum = ip - st->iwin;
      EX_RECORD_EXCEPTION(st, inum, EXC_II, 0);
      EX_HANDLE_EXCEPTION(st, ip, th);
      return SUCCESS;
   }
   GPR(rd) = ((Reg64) contents_rt) >> SHAMT(ip->instr);
   return SUCCESS;
}

ID(dsrlv)
{
   ip->dst[0].flags = REG_INT; ip->dst[0].reg = RDRI(ip->instr);
   ip->src[0].flags = REG_INT; ip->src[0].reg = RTRI(ip->instr); 
   ip->src[1].flags = REG_INT; ip->src[1].reg = RSRI(ip->instr);
}
EX(dsrlv)
{
   uint  rd          = ip->dst[0].preg;
   Reg_s contents_rs = GPR(ip->src[1].preg);
   Reg_s contents_rt = GPR(ip->src[0].preg);

   if (!ENABLE_64BIT(st)) {
      int inum = ip - st->iwin;
      EX_RECORD_EXCEPTION(st, inum, EXC_II, 0);
      EX_HANDLE_EXCEPTION(st, ip, th);
      return SUCCESS;
   }
   GPR(rd) = ((Reg64) contents_rt) >> (contents_rs & 0x3f);
   return SUCCESS;
}

ID(dsrl32)
{
   ip->dst[0].flags = REG_INT; ip->dst[0].reg = RDRI(ip->instr);
   ip->src[0].flags = REG_INT; ip->src[0].reg = RTRI(ip->instr); 
}
EX(dsrl32)
{
   uint  rd          = ip->dst[0].preg;
   Reg_s contents_rt = GPR(ip->src[0].preg);

   if (!ENABLE_64BIT(st)) {
      int inum = ip - st->iwin;
      EX_RECORD_EXCEPTION(st, inum, EXC_II, 0);
      EX_HANDLE_EXCEPTION(st, ip, th);
      return SUCCESS;
   }
   GPR(rd) = ((Reg64) contents_rt) >> (SHAMT(ip->instr)+32);
   return SUCCESS;
}

ID(dsub)
{
   ip->dst[0].flags = REG_INT; ip->dst[0].reg = RDRI(ip->instr);
   ip->src[0].flags = REG_INT; ip->src[0].reg = RSRI(ip->instr); 
   ip->src[1].flags = REG_INT; ip->src[1].reg = RTRI(ip->instr);
}
EX(dsub)
{
   uint  rd          = ip->dst[0].preg;
   Reg_s contents_rs = GPR(ip->src[0].preg);
   Reg_s contents_rt = GPR(ip->src[1].preg);
   Reg64_s diff = contents_rs - contents_rt;
   int inum = ip - st->iwin;

   if (!ENABLE_64BIT(st)) {
      EX_RECORD_EXCEPTION(st, inum, EXC_II, 0);
   } else if (SIGN_BIT64(contents_rs) != SIGN_BIT64(contents_rt)
              && SIGN_BIT64(contents_rs) != SIGN_BIT64(diff)) {
      EX_RECORD_EXCEPTION(st, inum, EXC_OV, 0);
   }
   if (st->iwin_flags[inum] & IWIN_TLBFAULT) {
      EX_HANDLE_EXCEPTION(st, ip, th);
      return SUCCESS;
   }
   GPR(rd) = diff;
   return SUCCESS;
}

ID(dsubu)
{
   ip->dst[0].flags = REG_INT; ip->dst[0].reg = RDRI(ip->instr);
   ip->src[0].flags = REG_INT; ip->src[0].reg = RSRI(ip->instr); 
   ip->src[1].flags = REG_INT; ip->src[1].reg = RTRI(ip->instr);
}
EX(dsubu)
{
   uint  rd          = ip->dst[0].preg;
   Reg_s contents_rs = GPR(ip->src[0].preg);
   Reg_s contents_rt = GPR(ip->src[1].preg);

   if (!ENABLE_64BIT(st)) {
      int inum = ip - st->iwin;
      EX_RECORD_EXCEPTION(st, inum, EXC_II, 0);
      EX_HANDLE_EXCEPTION(st, ip, th);
      return SUCCESS;
   }
   GPR(rd) = ((Reg64)contents_rs - (Reg64)contents_rt);
   return SUCCESS;
}

ID(jalr)
{
   ip->dst[0].flags = REG_INT; ip->dst[0].reg = RDRI(ip->instr);
   ip->src[0].flags = REG_INT; ip->src[0].reg = RSRI(ip->instr); 
   ip->flags = INST_CALL | INST_INDIRECT;
}
EX(jalr)
{
   uint  rd          = ip->dst[0].preg;
   Reg_s contents_rs = GPR(ip->src[0].preg);

   int inum = ip - st->iwin;
   BrTREE *br = &st->branch_tree [st->iwin_br_node[inum]];
   BrTREE *right = &st->branch_tree [br->rchild];

   int ix = bp_pc_to_index(st->iwin_pc [inum]);
   GPR(rd) = th->returnpc;
   th->pc = contents_rs;
   st->bp_targets[ix] = th->pc;
   br->resolution =
      (right->condition == th->pc ? PRUNE_LEFT : PRUNE_RIGHT);
   st->iwin_branch_pc [inum] = contents_rs;
#if BRANCH_LATENCY > 1
   Add_to_worklist (st, BRANCH_LATENCY, finish_call, (MXSClientData)ip);
#else
   th->stall_branch = 0;
   UpdateStallFetch (th);
   prune_branch (st, st->iwin_br_node[inum]);
#endif
   return SUCCESS;
}

ID(jr)
{
   ip->src[0].flags = REG_INT; ip->src[0].reg = RSRI(ip->instr); 
   ip->flags = INST_BRANCH | INST_INDIRECT;
   if (ip->src[0].reg == REG_RA) {
      ip->flags |= INST_RETURN;
   }
}
EX(jr)
{
   Reg_s contents_rs = GPR(ip->src[0].preg);

   int inum = ip - st->iwin;
   BrTREE *br = &st->branch_tree [st->iwin_br_node[inum]];
   BrTREE *right = &st->branch_tree [br->rchild];

   if (ip->flags & INST_RETURN) {
      th->pc = contents_rs;
   } else {
      int ix = bp_pc_to_index(st->iwin_pc [inum]);
      th->pc = contents_rs;
      st->bp_targets[ix] = th->pc;
   }
   br->resolution =
      (right->condition == th->pc ? PRUNE_LEFT : PRUNE_RIGHT);
   st->iwin_branch_pc [inum] = contents_rs;
#if BRANCH_LATENCY > 1
   Add_to_worklist (st, BRANCH_LATENCY, finish_branch, (MXSClientData)ip);
#else
   th->stall_branch = 0;
   UpdateStallFetch (th);
   prune_branch (st, st->iwin_br_node[inum]);
#endif
   return SUCCESS;
}

ID(mfhi)
{
   ip->dst[0].flags = REG_INT; ip->dst[0].reg = RDRI(ip->instr);
   ip->src[0].flags = REG_INT; ip->src[0].reg = MXS_REG_HI;
}
EX(mfhi)
{
   uint  rd          = ip->dst[0].preg;
   uint  hi          = ip->src[0].preg;
   GPR(rd) = GPR(hi);
   return SUCCESS;
}

ID(mflo)
{
   ip->dst[0].flags = REG_INT; ip->dst[0].reg = RDRI(ip->instr);
   ip->src[0].flags = REG_INT; ip->src[0].reg = MXS_REG_LO; 
}
EX(mflo)
{
   uint  rd          = ip->dst[0].preg;
   uint  lo          = ip->src[0].preg;
   GPR(rd) = GPR(lo);
   return SUCCESS;
}

ID(movc)
{
   MXSDecodeC1Instruction(ip, pc);
}
EX(movc)
{
   if (!ENABLE_MIPS4(st)) {
      int inum = ip - st->iwin;
      EX_RECORD_EXCEPTION(st, inum,  EXC_II, 0);
      EX_HANDLE_EXCEPTION(st, ip, th);
      return SUCCESS;
   }
   MXSExecuteC1Instruction(st, ip, th);
   return SUCCESS;
}

ID(movn)
{
   ip->dst[0].flags = REG_INT; ip->dst[0].reg = RDRI(ip->instr);
   ip->src[0].flags = REG_INT; ip->src[0].reg = RSRI(ip->instr); 
   ip->src[1].flags = REG_INT; ip->src[1].reg = RTRI(ip->instr); 
   ip->src[2].flags = REG_INT; ip->src[2].reg = RDRI(ip->instr); 
}
EX(movn)
{
   uint  rd          = ip->dst[0].preg;
   Reg_s contents_rs = GPR(ip->src[0].preg);
   Reg_s contents_rt = GPR(ip->src[1].preg);
   Reg_s contents_rd = GPR(ip->src[2].preg);

   if (!ENABLE_MIPS4(st)) {
      int inum = ip - st->iwin;
      EX_RECORD_EXCEPTION(st, inum,  EXC_II, 0);
      EX_HANDLE_EXCEPTION(st, ip, th);
      return SUCCESS;
   }

   if (contents_rt != 0) {
      GPR(rd) = contents_rs;
   } else {
      GPR(rd) = contents_rd;
   }
   return SUCCESS;
}

ID(movz)
{
   ip->dst[0].flags = REG_INT; ip->dst[0].reg = RDRI(ip->instr);
   ip->src[0].flags = REG_INT; ip->src[0].reg = RSRI(ip->instr); 
   ip->src[1].flags = REG_INT; ip->src[1].reg = RTRI(ip->instr); 
   ip->src[2].flags = REG_INT; ip->src[2].reg = RDRI(ip->instr); 
}
EX(movz)
{
   uint  rd          = ip->dst[0].preg;
   Reg_s contents_rs = GPR(ip->src[0].preg);
   Reg_s contents_rt = GPR(ip->src[1].preg);
   Reg_s contents_rd = GPR(ip->src[2].preg);

   if (!ENABLE_MIPS4(st)) {
      int inum = ip - st->iwin;
      EX_RECORD_EXCEPTION(st, inum,  EXC_II, 0);
      EX_HANDLE_EXCEPTION(st, ip, th);
      return SUCCESS;
   }
   
   if (contents_rt == 0) {
      GPR(rd) = contents_rs;
   } else {
      GPR(rd) = contents_rd;
   }
   return SUCCESS;
}

ID(mthi)
{
   ip->dst[0].flags = REG_INT; ip->dst[0].reg = MXS_REG_HI;
   ip->src[0].flags = REG_INT; ip->src[0].reg = RSRI(ip->instr); 
}
EX(mthi)
{
   uint  hi          = ip->dst[0].preg;
   Reg_s contents_rs = GPR(ip->src[0].preg);
   GPR(hi) = contents_rs;
   return SUCCESS;
}

ID(mtlo)
{
   ip->dst[0].flags = REG_INT; ip->dst[0].reg = MXS_REG_LO;
   ip->src[0].flags = REG_INT; ip->src[0].reg = RSRI(ip->instr); 
}
EX(mtlo)
{
   uint  lo          = ip->dst[0].preg;
   Reg_s contents_rs = GPR(ip->src[0].preg);
   GPR(lo) = contents_rs;
   return SUCCESS;
}

ID(mult)
{
   ip->dst[0].flags = REG_INT; ip->dst[0].reg = MXS_REG_HI;
   ip->dst[1].flags = REG_INT; ip->dst[1].reg = MXS_REG_LO;
   ip->src[0].flags = REG_INT; ip->src[0].reg = RTRI(ip->instr); 
   ip->src[1].flags = REG_INT; ip->src[1].reg = RSRI(ip->instr);
}
EX(mult)
{
   uint  hi          = ip->dst[0].preg;
   uint  lo          = ip->dst[1].preg;
   Reg_s contents_rs = GPR(ip->src[1].preg);
   Reg_s contents_rt = GPR(ip->src[0].preg);
   bool negative = FALSE;
            
   if (contents_rs < 0) {
      negative = TRUE;
      contents_rs = -contents_rs;
   }
   if (contents_rt < 0) {
      negative = !negative;
      contents_rt = -contents_rt;
   }
   LongMultiply32(st, (Reg32)contents_rs, (Reg32)contents_rt, lo, hi);
   if (negative) {
      GPR(lo) = ~GPR(lo);
      GPR(hi) = ~GPR(hi);
      GPR(lo) += 1;
      if (GPR(lo) == 0)
         GPR(hi) += 1;
   }
   return SUCCESS;
}

ID(multu)
{
   ip->dst[0].flags = REG_INT; ip->dst[0].reg = MXS_REG_HI;
   ip->dst[1].flags = REG_INT; ip->dst[1].reg = MXS_REG_LO;
   ip->src[0].flags = REG_INT; ip->src[0].reg = RTRI(ip->instr); 
   ip->src[1].flags = REG_INT; ip->src[1].reg = RSRI(ip->instr);
}
EX(multu)
{
   uint  hi          = ip->dst[0].preg;
   uint  lo          = ip->dst[1].preg;
   Reg_s contents_rs = GPR(ip->src[1].preg);
   Reg_s contents_rt = GPR(ip->src[0].preg);

   LongMultiply32(st, (Reg32)contents_rs, (Reg32)contents_rt, lo, hi);
   return SUCCESS;
}

ID(nor)
{
   ip->dst[0].flags = REG_INT; ip->dst[0].reg = RDRI(ip->instr);
   ip->src[0].flags = REG_INT; ip->src[0].reg = RSRI(ip->instr); 
   ip->src[1].flags = REG_INT; ip->src[1].reg = RTRI(ip->instr);
}
EX(nor)
{
   uint  rd          = ip->dst[0].preg;
   Reg_s contents_rs = GPR(ip->src[0].preg);
   Reg_s contents_rt = GPR(ip->src[1].preg);

   GPR(rd) =  ~ (contents_rs | contents_rt);
   return SUCCESS;
}

ID(or)
{
   ip->dst[0].flags = REG_INT; ip->dst[0].reg = RDRI(ip->instr);
   ip->src[0].flags = REG_INT; ip->src[0].reg = RSRI(ip->instr); 
   ip->src[1].flags = REG_INT; ip->src[1].reg = RTRI(ip->instr);
}
EX(or)
{
   uint  rd          = ip->dst[0].preg;
   Reg_s contents_rs = GPR(ip->src[0].preg);
   Reg_s contents_rt = GPR(ip->src[1].preg);

   GPR(rd) = contents_rs | contents_rt;
   return SUCCESS;
}

ID(sll)
{
   ip->dst[0].flags = REG_INT; ip->dst[0].reg = RDRI(ip->instr);
   ip->src[0].flags = REG_INT; ip->src[0].reg = RTRI(ip->instr); 
}
EX(sll)
{
   uint  rd          = ip->dst[0].preg;
   Reg_s contents_rt = GPR(ip->src[0].preg);

   GPR(rd) = (Reg32_s)contents_rt << SHAMT(ip->instr);
   return SUCCESS;
}

ID(sllv)
{
   ip->dst[0].flags = REG_INT; ip->dst[0].reg = RDRI(ip->instr);
   ip->src[0].flags = REG_INT; ip->src[0].reg = RTRI(ip->instr); 
   ip->src[1].flags = REG_INT; ip->src[1].reg = RSRI(ip->instr);
}
EX(sllv)
{
   uint  rd          = ip->dst[0].preg;
   Reg_s contents_rs = GPR(ip->src[1].preg);
   Reg_s contents_rt = GPR(ip->src[0].preg);

   GPR(rd) = (Reg32_s)contents_rt << (contents_rs & 0x1f);
   return SUCCESS;
}

ID(slt)
{
   ip->dst[0].flags = REG_INT; ip->dst[0].reg = RDRI(ip->instr);
   ip->src[0].flags = REG_INT; ip->src[0].reg = RSRI(ip->instr); 
   ip->src[1].flags = REG_INT; ip->src[1].reg = RTRI(ip->instr);
}
EX(slt)
{
   uint  rd          = ip->dst[0].preg;
   Reg_s contents_rs = GPR(ip->src[0].preg);
   Reg_s contents_rt = GPR(ip->src[1].preg);

   if (contents_rs < contents_rt)
      GPR(rd) = 1;
   else
      GPR(rd) = 0;
   return SUCCESS;
}

ID(sltu)
{
   ip->dst[0].flags = REG_INT; ip->dst[0].reg = RDRI(ip->instr);
   ip->src[0].flags = REG_INT; ip->src[0].reg = RSRI(ip->instr); 
   ip->src[1].flags = REG_INT; ip->src[1].reg = RTRI(ip->instr);
}
EX(sltu)
{
   uint  rd          = ip->dst[0].preg;
   Reg_s contents_rs = GPR(ip->src[0].preg);
   Reg_s contents_rt = GPR(ip->src[1].preg);

   if ((Reg) contents_rs < (Reg) contents_rt)
      GPR(rd) = 1;
   else
      GPR(rd) = 0;
   return SUCCESS;
}

ID(sra)
{
   ip->dst[0].flags = REG_INT; ip->dst[0].reg = RDRI(ip->instr);
   ip->src[0].flags = REG_INT; ip->src[0].reg = RTRI(ip->instr); 
}
EX(sra)
{
   uint  rd          = ip->dst[0].preg;
   Reg_s contents_rt = GPR(ip->src[0].preg);

   GPR(rd) = (Reg32_s)contents_rt >> SHAMT(ip->instr);
   return SUCCESS;
}

ID(srav)
{
   ip->dst[0].flags = REG_INT; ip->dst[0].reg = RDRI(ip->instr);
   ip->src[0].flags = REG_INT; ip->src[0].reg = RTRI(ip->instr); 
   ip->src[1].flags = REG_INT; ip->src[1].reg = RSRI(ip->instr);
}
EX(srav)
{
   uint  rd          = ip->dst[0].preg;
   Reg_s contents_rs = GPR(ip->src[1].preg);
   Reg_s contents_rt = GPR(ip->src[0].preg);

   GPR(rd) = (Reg32_s)contents_rt >> (contents_rs & 0x1f);
   return SUCCESS;
}

ID(srl)
{
   ip->dst[0].flags = REG_INT; ip->dst[0].reg = RDRI(ip->instr);
   ip->src[0].flags = REG_INT; ip->src[0].reg = RTRI(ip->instr); 
}
EX(srl)
{
   uint  rd          = ip->dst[0].preg;
   Reg_s contents_rt = GPR(ip->src[0].preg);

   GPR(rd) = (Reg32_s)(((Reg32) contents_rt) >> SHAMT(ip->instr));
   return SUCCESS;
}

ID(srlv)
{
   ip->dst[0].flags = REG_INT; ip->dst[0].reg = RDRI(ip->instr);
   ip->src[0].flags = REG_INT; ip->src[0].reg = RTRI(ip->instr); 
   ip->src[1].flags = REG_INT; ip->src[1].reg = RSRI(ip->instr);
}
EX(srlv)
{
   uint  rd          = ip->dst[0].preg;
   Reg_s contents_rs = GPR(ip->src[1].preg);
   Reg_s contents_rt = GPR(ip->src[0].preg);

   GPR(rd) = (Reg32_s)((Reg32) contents_rt >> (contents_rs & 0x1f));
   return SUCCESS;
}

ID(sub)
{
   ip->dst[0].flags = REG_INT; ip->dst[0].reg = RDRI(ip->instr);
   ip->src[0].flags = REG_INT; ip->src[0].reg = RSRI(ip->instr); 
   ip->src[1].flags = REG_INT; ip->src[1].reg = RTRI(ip->instr);
}
EX(sub)
{
   uint  rd          = ip->dst[0].preg;
   Reg_s contents_rs = GPR(ip->src[0].preg);
   Reg_s contents_rt = GPR(ip->src[1].preg);
   Reg32_s diff;
   
   diff = (Reg32_s)contents_rs - (Reg32_s)contents_rt;
   if (SIGN_BIT32(contents_rs) != SIGN_BIT32(contents_rt)
       && SIGN_BIT32(contents_rs) != SIGN_BIT32(diff)) {
      int inum = ip - st->iwin;
      EX_RECORD_EXCEPTION(st, inum, EXC_OV, 0);
      EX_HANDLE_EXCEPTION(st, ip, th);
      return SUCCESS;
   }
   GPR(rd) = diff;
   return SUCCESS;
}

ID(subu)
{
   ip->dst[0].flags = REG_INT; ip->dst[0].reg = RDRI(ip->instr);
   ip->src[0].flags = REG_INT; ip->src[0].reg = RSRI(ip->instr); 
   ip->src[1].flags = REG_INT; ip->src[1].reg = RTRI(ip->instr);
}
EX(subu)
{
   uint  rd          = ip->dst[0].preg;
   Reg_s contents_rs = GPR(ip->src[0].preg);
   Reg_s contents_rt = GPR(ip->src[1].preg);

   GPR(rd) = (Reg32_s)((Reg32) contents_rs - (Reg32) contents_rt);
   return SUCCESS;
}

ID(sync)
{
   ip->flags = INST_SYNC;
}
EX(sync)
{
   /* sync is currently implemented in a conservative way:
    * it stalls the fetch unit until it graduates, ensuring
    * that subsequent memory ops are not issued until any
    * preceding loads/stores have completed.  So don't do
    * anything here, unstall the fetch unit at graduation */
   return SUCCESS;
}

ID(syscall)
{
   ip->flags = INST_SYSCALL;
}
EX(syscall)
{
   int inum = ip - st->iwin;

   EX_RECORD_EXCEPTION(st, inum, EXC_SYSCALL, 0);
   EX_HANDLE_EXCEPTION(st, ip, th);
   return SUCCESS;
}


ID(teq)
{
   ip->src[0].flags = REG_INT; ip->src[0].reg = RSRI(ip->instr); 
   ip->src[1].flags = REG_INT; ip->src[1].reg = RTRI(ip->instr);
   ip->flags = INST_TRAP;
}
EX(teq)
{
   Reg_s contents_rs = GPR(ip->src[0].preg);
   Reg_s contents_rt = GPR(ip->src[1].preg);
   int cond = (contents_rs == contents_rt);
   TrapOp(st, ip, th, cond);
   return SUCCESS;
}

ID(tge)
{
   ip->src[0].flags = REG_INT; ip->src[0].reg = RSRI(ip->instr); 
   ip->src[1].flags = REG_INT; ip->src[1].reg = RTRI(ip->instr);
   ip->flags = INST_TRAP;
}
EX(tge)
{
   Reg_s contents_rs = GPR(ip->src[0].preg);
   Reg_s contents_rt = GPR(ip->src[1].preg);
   int cond = (contents_rs >= contents_rt);
   TrapOp(st, ip, th, cond);
   return SUCCESS;
}

ID(tgeu)
{
   ip->src[0].flags = REG_INT; ip->src[0].reg = RSRI(ip->instr); 
   ip->src[1].flags = REG_INT; ip->src[1].reg = RTRI(ip->instr);
   ip->flags = INST_TRAP;
}
EX(tgeu)
{
   Reg_s contents_rs = GPR(ip->src[0].preg);
   Reg_s contents_rt = GPR(ip->src[1].preg);
   int cond = ((Reg)contents_rs >= (Reg)contents_rt);
   TrapOp(st, ip, th, cond);
   return SUCCESS;
}

ID(tlt)
{
   ip->src[0].flags = REG_INT; ip->src[0].reg = RSRI(ip->instr); 
   ip->src[1].flags = REG_INT; ip->src[1].reg = RTRI(ip->instr);
   ip->flags = INST_TRAP;
}
EX(tlt)
{
   Reg_s contents_rs = GPR(ip->src[0].preg);
   Reg_s contents_rt = GPR(ip->src[1].preg);
   int cond = (contents_rs < contents_rt);
   TrapOp(st, ip, th, cond);
   return SUCCESS;
}

ID(tltu)
{
   ip->src[0].flags = REG_INT; ip->src[0].reg = RSRI(ip->instr); 
   ip->src[1].flags = REG_INT; ip->src[1].reg = RTRI(ip->instr);
   ip->flags = INST_TRAP;
}
EX(tltu)
{
   Reg_s contents_rs = GPR(ip->src[0].preg);
   Reg_s contents_rt = GPR(ip->src[1].preg);
   int cond = ((Reg)contents_rs < (Reg)contents_rt);
   TrapOp(st, ip, th, cond);
   return SUCCESS;
}

ID(tne)
{
   ip->src[0].flags = REG_INT; ip->src[0].reg = RSRI(ip->instr); 
   ip->src[1].flags = REG_INT; ip->src[1].reg = RTRI(ip->instr);
   ip->flags = INST_TRAP;
}
EX(tne)
{
   Reg_s contents_rs = GPR(ip->src[0].preg);
   Reg_s contents_rt = GPR(ip->src[1].preg);
   int cond = (contents_rs != contents_rt);
   TrapOp(st, ip, th, cond);
   return SUCCESS;
}

ID(xor)
{
   ip->dst[0].flags = REG_INT; ip->dst[0].reg = RDRI(ip->instr);
   ip->src[0].flags = REG_INT; ip->src[0].reg = RSRI(ip->instr); 
   ip->src[1].flags = REG_INT; ip->src[1].reg = RTRI(ip->instr);
}
EX(xor)
{
   uint  rd          = ip->dst[0].preg;
   Reg_s contents_rs = GPR(ip->src[0].preg);
   Reg_s contents_rt = GPR(ip->src[1].preg);

   GPR(rd) = contents_rs ^ contents_rt;
   return SUCCESS;
}

/*****************************************************************
 * B C O N D    O P C O D E S
 *****************************************************************/

ID(bgez)
{
   int immediate     = IMMED(ip->instr);
   ip->src[0].flags = REG_INT; ip->src[0].reg = RSRI(ip->instr); 
   ip->target = (pc + 4) + SIGN_EXTEND32(18, immediate << 2);
   if (ip->src[0].reg == 0) {
      ip->flags = INST_BRANCH;
   } else {
      ip->flags = INST_BRANCH | INST_CONDITIONAL;
   }
}
EX(bgez)
{
   Reg_s contents_rs = GPR(ip->src[0].preg);
   int cond = (!(SIGN_BIT(contents_rs)));
   BranchOp(st, ip, th, cond, 0);
   return SUCCESS;
}
      
ID(bgezl)
{
   int immediate     = IMMED(ip->instr);
   ip->src[0].flags = REG_INT; ip->src[0].reg = RSRI(ip->instr); 
   ip->target = (pc + 4) + SIGN_EXTEND32(18, immediate << 2);
   ip->flags = INST_BRANCH | INST_CONDITIONAL | INST_LIKELY;
}
EX(bgezl)
{
   Reg_s contents_rs = GPR(ip->src[0].preg);
   int cond = (!(SIGN_BIT(contents_rs)));
   BranchOp(st, ip, th, cond, 1);
   return SUCCESS;
}
      
ID(bgezal)
{
   int immediate     = IMMED(ip->instr);
   ip->dst[0].flags = REG_INT; ip->dst[0].reg = REG_RA;
   ip->src[0].flags = REG_INT; ip->src[0].reg = RSRI(ip->instr); 
   ip->target = (pc + 4) + SIGN_EXTEND32(18, immediate << 2);
   if (ip->src[0].reg == 0) {
      ip->flags = INST_CALL;
   } else {
      ip->flags = INST_CALL | INST_CONDITIONAL;
   }
}
EX(bgezal)
{
   Reg_s contents_rs = GPR(ip->src[0].preg);
   int cond = (!(SIGN_BIT(contents_rs)));
   CallOp(st, ip, th, cond, 0);
   return SUCCESS;
}
      
ID(bgezall)
{
   int immediate     = IMMED(ip->instr);
   ip->dst[0].flags = REG_INT; ip->dst[0].reg = REG_RA;
   ip->src[0].flags = REG_INT; ip->src[0].reg = RSRI(ip->instr); 
   ip->target = (pc + 4) + SIGN_EXTEND32(18, immediate << 2);
   ip->flags = INST_CALL | INST_CONDITIONAL | INST_LIKELY;
}
EX(bgezall)
{
   Reg_s contents_rs = GPR(ip->src[0].preg);
   int cond = (!(SIGN_BIT(contents_rs)));
   CallOp(st, ip, th, cond, 1);
   return SUCCESS;
}
      
ID(bltz)
{
   int immediate     = IMMED(ip->instr);
   ip->src[0].flags = REG_INT; ip->src[0].reg = RSRI(ip->instr); 
   ip->target = (pc + 4) + SIGN_EXTEND32(18, immediate << 2);
   ip->flags = INST_BRANCH | INST_CONDITIONAL;
}
EX(bltz)
{
   Reg_s contents_rs = GPR(ip->src[0].preg);
   int cond = (SIGN_BIT(contents_rs) != 0);
   BranchOp(st, ip, th, cond, 0);
   return SUCCESS;
}
      
ID(bltzl)
{
   int immediate     = IMMED(ip->instr);
   ip->src[0].flags = REG_INT; ip->src[0].reg = RSRI(ip->instr); 
   ip->target = (pc + 4) + SIGN_EXTEND32(18, immediate << 2);
   ip->flags = INST_BRANCH | INST_CONDITIONAL | INST_LIKELY;
}
EX(bltzl)
{
   Reg_s contents_rs = GPR(ip->src[0].preg);
   int cond = (SIGN_BIT(contents_rs) != 0);
   BranchOp(st, ip, th, cond, 1);
   return SUCCESS;
}
      
ID(bltzal)
{
   int immediate     = IMMED(ip->instr);
   ip->dst[0].flags = REG_INT; ip->dst[0].reg = REG_RA;
   ip->src[0].flags = REG_INT; ip->src[0].reg = RSRI(ip->instr); 
   ip->target = (pc + 4) + SIGN_EXTEND32(18, immediate << 2);
   ip->flags = INST_CALL | INST_CONDITIONAL;
}
EX(bltzal)
{
   Reg_s contents_rs = GPR(ip->src[0].preg);
   int cond = (SIGN_BIT(contents_rs) != 0);
   CallOp(st, ip, th, cond, 0);
   return SUCCESS;
}
      
ID(bltzall)
{
   int immediate     = IMMED(ip->instr);
   ip->dst[0].flags = REG_INT; ip->dst[0].reg = REG_RA;
   ip->src[0].flags = REG_INT; ip->src[0].reg = RSRI(ip->instr); 
   ip->target = (pc + 4) + SIGN_EXTEND32(18, immediate << 2);
   ip->flags = INST_CALL | INST_CONDITIONAL | INST_LIKELY;
}
EX(bltzall)
{
   Reg_s contents_rs = GPR(ip->src[0].preg);
   int cond = (SIGN_BIT(contents_rs) != 0);
   CallOp(st, ip, th, cond, 1);
   return SUCCESS;
}

ID(teqi)
{
   ip->src[0].flags = REG_INT; ip->src[0].reg = RSRI(ip->instr); 
   ip->flags = INST_TRAP;
}
EX(teqi)
{
   int immediate     = IMMED(ip->instr);
   Reg_s contents_rs = GPR(ip->src[0].preg);
   int cond = (contents_rs == SIGN_EXTEND(16,immediate));
   TrapOp(st, ip, th, cond);
   return SUCCESS;
}
      
ID(tgei)
{
   ip->src[0].flags = REG_INT; ip->src[0].reg = RSRI(ip->instr); 
   ip->flags = INST_TRAP;
}
EX(tgei)
{
   int immediate     = IMMED(ip->instr);
   Reg_s contents_rs = GPR(ip->src[0].preg);
   int cond = (contents_rs >= SIGN_EXTEND(16,immediate));
   TrapOp(st, ip, th, cond);
   return SUCCESS;
}
      
ID(tgeiu)
{
   ip->src[0].flags = REG_INT; ip->src[0].reg = RSRI(ip->instr); 
   ip->flags = INST_TRAP;
}
EX(tgeiu)
{
   int immediate     = IMMED(ip->instr);
   Reg_s contents_rs = GPR(ip->src[0].preg);
   int cond = ((Reg)contents_rs >= (Reg)SIGN_EXTEND(16,immediate));
   TrapOp(st, ip, th, cond);
   return SUCCESS;
}
      
ID(tlti)
{
   ip->src[0].flags = REG_INT; ip->src[0].reg = RSRI(ip->instr); 
   ip->flags = INST_TRAP;
}
EX(tlti)
{
   int immediate     = IMMED(ip->instr);
   Reg_s contents_rs = GPR(ip->src[0].preg);
   int cond = (contents_rs < SIGN_EXTEND(16,immediate));
   TrapOp(st, ip, th, cond);
   return SUCCESS;
}
      
ID(tltiu)
{
   ip->src[0].flags = REG_INT; ip->src[0].reg = RSRI(ip->instr); 
   ip->flags = INST_TRAP;
}
EX(tltiu)
{
   int immediate     = IMMED(ip->instr);
   Reg_s contents_rs = GPR(ip->src[0].preg);
   int cond = ((Reg)contents_rs < (Reg)SIGN_EXTEND(16,immediate));
   TrapOp(st, ip, th, cond);
   return SUCCESS;
}
      
ID(tnei)
{
   ip->src[0].flags = REG_INT; ip->src[0].reg = RSRI(ip->instr); 
   ip->flags = INST_TRAP;
}
EX(tnei)
{
   int immediate     = IMMED(ip->instr);
   Reg_s contents_rs = GPR(ip->src[0].preg);
   int cond = (contents_rs != SIGN_EXTEND(16,immediate));
   TrapOp(st, ip, th, cond);
   return SUCCESS;
}

EX(bdoor)
{
   CPUState *P = (CPUState *)st->mipsyPtr;
   int64 bdoorRetval; 
   PA pAddr;
   Result ret;
   int inum = ip - st->iwin;
   
   if (!AllsQuiet (st, ip, ST_SYSCALL))
      return SUCCESS;

   /* Assume this is a call. Execute it on the real CPU. This */
   /* assumes that all backdoor calls take less than 4 */
   /* arguments. The prom stuff is a way for the simulated cpu to */
   /* actually run the code in simbootprom.c */
   ret = TranslateVirtualNoSideeffect(P, (VA) th->pc-4, &pAddr);
   ASSERT(ret==SUCCESS);
/* BASEM watch this! */
   bdoorRetval = ((int64 (*)(int,int,int,int))(pAddr))
      (GPR(th->regnames[4]), GPR(th->regnames[5]), 
       GPR(th->regnames[6]), GPR(th->regnames[7]));
   GPR(th->regnames[2]) = bdoorRetval >> 32;
   GPR(th->regnames[3]) = bdoorRetval & 0xffffffff;
   th->pc = GPR(th->regnames[31]);
   th->stall_sys = 0;
   UpdateStallFetch(th);
   return SUCCESS;
}

/*****************************************************************
 * BranchOp -
 * Handle all branches/jumps (except branch/jump-and-link)
 *****************************************************************/
void   
BranchOp(struct s_cpu_state *st, INST *ip, THREAD *th, int cond, int likely)
{
   int inum = ip - st->iwin;
   BrTREE *br = &st->branch_tree [st->iwin_br_node[inum]];
   BrTREE *right = &st->branch_tree [br->rchild];
   int ix = bp_pc_to_index(st->iwin_pc [inum]);
   Reg target = ip->target;
   if (cond) {
      st->bp_bits[ix]++;
      if (st->bp_bits[ix] > BP_MAX_VAL)
         st->bp_bits[ix] = BP_MAX_VAL;
      if (PredictTaken (right->condition, likely))
         br->resolution = PRUNE_LEFT;
      else
         br->resolution = PRUNE_RIGHT;
      st->iwin_branch_pc[inum] = ip->target;
      st->iwin_flags[inum] |= IWIN_TAKENBR;
      if (br->lchild == -1) {
         th->pc = ip->target;
      }
   } else {
      st->bp_bits[ix]--;
      if (st->bp_bits[ix] < 0)
         st->bp_bits[ix] = 0;
      if (PredictTaken (right->condition, likely))
         br->resolution = PRUNE_RIGHT;
      else
         br->resolution = PRUNE_LEFT;
      st->iwin_branch_pc[inum] = st->iwin_pc[inum] + (PC_INC*(1+BRANCH_SLOTS));
   }
   
#if BRANCH_LATENCY > 1
   Add_to_worklist (st, BRANCH_LATENCY, finish_branch, (MXSClientData)ip);
#else
   th->stall_branch = 0;
   UpdateStallFetch (th);
   prune_branch (st, st->iwin_br_node[inum]);
#endif
}

/*****************************************************************
 * CallOp -
 * Handle all branch/jump-and-link operations
 *****************************************************************/
static void   
CallOp(struct s_cpu_state *st, INST *ip, THREAD *th, int cond, int likely)
{
   int inum = ip - st->iwin;
   BrTREE *br = &st->branch_tree [st->iwin_br_node[inum]];
   BrTREE *right = &st->branch_tree [br->rchild];
   int ix = bp_pc_to_index(st->iwin_pc [inum]);
   Reg target = ip->target;
   int ra = ip->dst[0].preg;
   if (cond) {
      GPR(ra) = th->returnpc;
      st->bp_bits[ix]++;
      if (st->bp_bits[ix] > BP_MAX_VAL)
         st->bp_bits[ix] = BP_MAX_VAL;
      if (PredictTaken (right->condition, likely))
         br->resolution = PRUNE_LEFT;
      else
         br->resolution = PRUNE_RIGHT;
      st->iwin_branch_pc [inum] = ip->target;
      st->iwin_flags[inum] |= IWIN_TAKENBR;
      if (br->lchild == -1) {
         th->pc = ip->target;
      }
   } else {
      st->bp_bits[ix]--;
      if (st->bp_bits[ix] < 0)
         st->bp_bits[ix] = 0;
      if (PredictTaken (right->condition, likely))
         br->resolution = PRUNE_RIGHT;
      else
         br->resolution = PRUNE_LEFT;
      st->iwin_branch_pc [inum] = st->iwin_pc [inum] +
         (PC_INC*(1+BRANCH_SLOTS));
   }
#if BRANCH_LATENCY > 1
   Add_to_worklist (st, BRANCH_LATENCY, finish_call, (MXSClientData)ip);
#else
   th->stall_branch = 0;
   UpdateStallFetch (th);
   prune_branch (st, st->iwin_br_node[inum]);
#endif
}

/*****************************************************************
 * TrapOp -
 * Handle all conditional trap operations
 *****************************************************************/
static void   
TrapOp(struct s_cpu_state *st, INST *ip, THREAD *th, int cond)
{
   if (cond) {
      int inum = ip - st->iwin;
      EX_RECORD_EXCEPTION(st, inum, EXC_TRAP, 0);
      EX_HANDLE_EXCEPTION(st, ip, th);
   } else {
      th->stall_sys = 0;
      UpdateStallFetch(th);
   }
}


/*****************************************************************
 * LongMultiply64 -
 * Multiply two 64 bit numbers into the HI and LO registers.
 *****************************************************************/
static void 
LongMultiply64(struct s_cpu_state *st, Reg64 v_1, Reg64 v_2, uint lo, uint hi)
{
   Reg64 a, b, c, d;
   Reg64 bd, ad, cb, ac;
   Reg64 mid, mid2, carry_mid = 0;

#define _HI32(_x) (((_x)>>32) & 0xffffffff)
#define _LO32(_x) ((_x) & 0xffffffff)
  
   a = _HI32(v_1);
   b = _LO32(v_1);
   c = _HI32(v_2);
   d = _LO32(v_2);
  
   bd = b * d;
   ad = a * d;
   cb = c * b;
   ac = a * c;
  
   mid = ad + cb;
   if (mid < ad || mid < cb)
      /* Arithmetic overflow or carry-out */
      carry_mid = 1;
  
   mid2 = mid + _HI32(bd);
   if (mid2 < mid || mid2 < _HI32(bd))
      /* Arithmetic overflow or carry-out */
      carry_mid += 1;
   
   GPR(lo) =  _LO32(bd) | (_LO32(mid2)<<32);
   GPR(hi) =  ac + (carry_mid << 32) + _HI32(mid2);
}

/*****************************************************************
 * LongMultiply32 -
 * Multiply two 32 bit numbers into the HI and LO registers.
 *****************************************************************/
static void 
LongMultiply32(struct s_cpu_state *st, Reg32 v_1, Reg32 v_2, uint lo, uint hi)
{
   Reg64 product64;

   product64 = (Reg64)v_1 * (Reg64)v_2;   
#ifdef SIM_MIPS64
   /* Compiler bug workaround */
   GPR(lo) = SIGN_EXTEND(32, product64);
   GPR(hi) = SIGN_EXTEND(32, (product64>>32));
#else
   GPR(lo) = (Reg32_s) product64;
   GPR(hi) = (Reg32_s) (product64>>32);
#endif
}

/*
 *  AllsQuiet  -  A utility routine for instructions that
 *			need to wait until all outstanding
 *			instructions are completed.
 */

int
AllsQuiet(struct s_cpu_state *st, INST *ip, int excuse)
{
   int	i, r;

   /* Force in order issue around this instruction	*/
   /* by stalling until all work has been done		*/
   /* and all registers have been written.		*/
   
   if (st->work_head) goto return_false;
   if ((st->ldst_head) || (st->ldst_nextReserved)) goto return_false;
   
   for (i=0; i < NUM_PHYS_REGS; i++) {
      if (st->reg_rstat[i].reg_status & REG_IN_WIN) {
         /* OK if we're writing this register, otherwise not quiet */
         int our_dst = 0;
         for (r = 0; r < ip->ndst; r++) {
            RegInfo *ri = &ip->dst[r];
            if ((ri->flags != 0) && (i == ri->preg)) {
              our_dst = 1;
              break;
            }
         }
         if (!our_dst) goto return_false;
      }
   }
   return 1;

return_false:
   st->stall_type = excuse;
   st->stall_issue = 1;
   return 0;
}

void 
ClearFPCStall(struct s_cpu_state *st,THREAD *th, int inum) {
   BrTREE	*br;
   /* Now can start fetching new instructions	*/
   th->stall_fpc = 0;
   UpdateStallFetch (th);
   
   /* Unstall children, too			*/
   br = &st->branch_tree [st->iwin_br_node[inum]];
   if (br->rchild >= 0) {
      br = &st->branch_tree [br->rchild];
      th = &st->threads [br->thread];
      th->stall_fpc = 0;
      UpdateStallFetch (th);
   }
}



/*****************************************************************
 * O P C O D E   T A B L E S
 *****************************************************************/ 

MXSOpcodeDesc mxsOpcodes[] = {
   { "spec",     specID,    specEX,    0x00, MXS_NO_UNIT,       0,  0 }, 
   { "bcond",    bcondID,   bcondEX,   0x01, MXS_BRANCH_UNIT,   1,  1 },
   { "j",        jID,       jEX,       0x02, MXS_BRANCH_UNIT,   1,  1 },
   { "jal",      jalID,     jalEX,     0x03, MXS_BRANCH_UNIT,   1,  1 },
   { "beq",      beqID,     beqEX,     0x04, MXS_BRANCH_UNIT,   1,  1 },
   { "bne",      bneID,     bneEX,     0x05, MXS_BRANCH_UNIT,   1,  1 },
   { "blez",     blezID,    blezEX,    0x06, MXS_BRANCH_UNIT,   1,  1 },
   { "bgtz",     bgtzID,    bgtzEX,    0x07, MXS_BRANCH_UNIT,   1,  1 },
   { "addi",     addiID,    addiEX,    0x08, MXS_INT_UNIT,      1,  1 },
   { "addiu",    addiuID,   addiuEX,   0x09, MXS_INT_UNIT,      1,  1 },
   { "slti",     sltiID,    sltiEX,    0x0A, MXS_INT_UNIT,      1,  1 },
   { "sltiu",    sltiuID,   sltiuEX,   0x0B, MXS_INT_UNIT,      1,  1 },
   { "andi",     andiID,    andiEX,    0x0C, MXS_INT_UNIT,      1,  1 },
   { "ori",      oriID,     oriEX,     0x0D, MXS_INT_UNIT,      1,  1 },
   { "xori",     xoriID,    xoriEX,    0x0E, MXS_INT_UNIT,      1,  1 },
   { "lui",      luiID,     luiEX,     0x0F, MXS_LDST_UNIT,     1,  1 },  
   { "cop0",     cop0ID,    cop0EX,    0x10, MXS_INT_UNIT,      1,  1 },
   { "cop1",     cop1ID,    cop1EX,    0x11, MXS_INT_UNIT,      1,  1 },
   { "cop2",     cop2ID,    cop2EX,    0x12, MXS_INT_UNIT,      1,  1 },
   { "cop1x",    cop1xID,   cop1xEX,   0x13, MXS_INT_UNIT,      1,  1 },
   { "beql",     beqlID,    beqlEX,    0x14, MXS_BRANCH_UNIT,   1,  1 },
   { "bnel",     bnelID,    bnelEX,    0x15, MXS_BRANCH_UNIT,   1,  1 },
   { "blezl",    blezlID,   blezlEX,   0x16, MXS_BRANCH_UNIT,   1,  1 },
   { "bgtzl",    bgtzlID,   bgtzlEX,   0x17, MXS_BRANCH_UNIT,   1,  1 },
   { "daddi",    daddiID,   daddiEX,   0x18, MXS_INT_UNIT,      1,  1 },
   { "daddiu",   daddiuID,  daddiuEX,  0x19, MXS_INT_UNIT,      1,  1 },
   { "ldl",      ldlID,     ldlEX,     0x1A, MXS_LDST_UNIT,     1,  1 },
   { "ldr",      ldrID,     ldrEX,     0x1B, MXS_LDST_UNIT,     1,  1 },
   { "undef",    undefID,   undefEX,   0x1C, MXS_NO_UNIT,       0,  0 },
   { "undef",    undefID,   undefEX,   0x1D, MXS_NO_UNIT,       0,  0 },
   { "undef",    undefID,   undefEX,   0x1E, MXS_NO_UNIT,       0,  0 },
   { "undef",    undefID,   undefEX,   0x1F, MXS_NO_UNIT,       0,  0 },
   { "lb",       lbID,      lbEX,      0x20, MXS_LDST_UNIT,     1,  1 },
   { "lh",       lhID,      lhEX,      0x21, MXS_LDST_UNIT,     1,  1 },
   { "lwl",      lwlID,     lwlEX,     0x22, MXS_LDST_UNIT,     1,  1 },
   { "lw",       lwID,      lwEX,      0x23, MXS_LDST_UNIT,     1,  1 },
   { "lbu",      lbuID,     lbuEX,     0x24, MXS_LDST_UNIT,     1,  1 },
   { "lhu",      lhuID,     lhuEX,     0x25, MXS_LDST_UNIT,     1,  1 },
   { "lwr",      lwrID,     lwrEX,     0x26, MXS_LDST_UNIT,     1,  1 },
   { "lwu",      lwuID,     lwuEX,     0x27, MXS_LDST_UNIT,     1,  1 },
   { "sb",       sbID,      sbEX,      0x28, MXS_LDST_UNIT,     1,  1 },
   { "sh",       shID,      shEX,      0x29, MXS_LDST_UNIT,     1,  1 },
   { "swl",      swlID,     swlEX,     0x2A, MXS_LDST_UNIT,     1,  1 },
   { "sw",       swID,      swEX,      0x2B, MXS_LDST_UNIT,     1,  1 },
   { "sdl",      sdlID,     sdlEX,     0x2C, MXS_LDST_UNIT,     1,  1 },
   { "sdr",      sdrID,     sdrEX,     0x2D, MXS_LDST_UNIT,     1,  1 },
   { "swr",      swrID,     swrEX,     0x2E, MXS_LDST_UNIT,     1,  1 },
   { "cache",    cacheID,   cacheEX,   0x2F, MXS_LDST_UNIT,     1,  1 },
   { "ll",       llID,      llEX,      0x30, MXS_LDST_UNIT,     1,  1 }, 
   { "lwc1",     lwc1ID,    lwc1EX,    0x31, MXS_LDST_UNIT,     1,  1 },
   { "lwc2",     lwc2ID,    lwc2EX,    0x32, MXS_LDST_UNIT,     1,  1 },
   { "pref",     prefID,    prefEX,    0x33, MXS_LDST_UNIT,     1,  1 },
   { "lld",      lldID,     lldEX,     0x34, MXS_LDST_UNIT,     1,  1 },
   { "ldc1",     ldc1ID,    ldc1EX,    0x35, MXS_LDST_UNIT,     1,  1 },
   { "ldc2",     ldc2ID,    ldc2EX,    0x36, MXS_LDST_UNIT,     1,  1 },
   { "ld",       ldID,      ldEX,      0x37, MXS_LDST_UNIT,     1,  1 },
   { "sc",       scID,      scEX,      0x38, MXS_LDST_UNIT,     1,  1 },
   { "swc1",     swc1ID,    swc1EX,    0x39, MXS_LDST_UNIT,     1,  1 },
   { "swc2",     swc2ID,    swc2EX,    0x3A, MXS_LDST_UNIT,     1,  1 },
   { "undef",    undefID,   undefEX,   0x3B, MXS_NO_UNIT,       0,  0 },
   { "scd",      scdID,     scdEX,     0x3C, MXS_LDST_UNIT,     1,  1 },
   { "sdc1",     sdc1ID,    sdc1EX,    0x3D, MXS_LDST_UNIT,     1,  1 },
   { "sdc2",     sdc2ID,    sdc2EX,    0x3E, MXS_LDST_UNIT,     1,  1 },
   { "sd",       sdID,      sdEX,      0x3F, MXS_LDST_UNIT,     1,  1 },
};

/* Based on the func field of spec opcode */
MXSOpcodeDesc mxsSpecOpcodes[] = {
   { "sll",      sllID,     sllEX,     0x00, MXS_INT_UNIT,      1,  1 },
   { "movc",     movcID,    movcEX,    0x01, MXS_INT_UNIT,      1,  1 },
   { "srl",      srlID,     srlEX,     0x02, MXS_INT_UNIT,      1,  1 },
   { "sra",      sraID,     sraEX,     0x03, MXS_INT_UNIT,      1,  1 },
   { "sllv",     sllvID,    sllvEX,    0x04, MXS_INT_UNIT,      1,  1 },
   { "undef",    undefID,   undefEX,   0x05, MXS_NO_UNIT,       0,  0 },
   { "srlv",     srlvID,    srlvEX,    0x06, MXS_INT_UNIT,      1,  1 },
   { "srav",     sravID,    sravEX,    0x07, MXS_INT_UNIT,      1,  1 },
   { "jr",       jrID,      jrEX,      0x08, MXS_BRANCH_UNIT,   1,  1 },
   { "jalr",     jalrID,    jalrEX,    0x09, MXS_BRANCH_UNIT,   1,  1 },
   { "movz",     movzID,    movzEX,    0x0A, MXS_INT_UNIT,      1,  1 },
   { "movn",     movnID,    movnEX,    0x0B, MXS_INT_UNIT,      1,  1 },
   { "syscall",  syscallID, syscallEX, 0x0C, MXS_INT_UNIT,      1,  1 },
   { "break",    breakID,   breakEX,   0x0D, MXS_INT_UNIT,      1,  1 },
   { "spim",     undefID,   undefEX,   0x0E, MXS_NO_UNIT,       0,  0 },
   { "sync",     syncID,    syncEX,    0x0F, MXS_LDST_UNIT,     1,  1 },
   { "mfhi",     mfhiID,    mfhiEX,    0x10, MXS_INT_UNIT,      1,  1 },
   { "mthi",     mthiID,    mthiEX,    0x11, MXS_INT_UNIT,      1,  1 },
   { "mflo",     mfloID,    mfloEX,    0x12, MXS_INT_UNIT,      1,  1 },
   { "mtlo",     mtloID,    mtloEX,    0x13, MXS_INT_UNIT,      1,  1 },
   { "dsllv",    dsllvID,   dsllvEX,   0x14, MXS_INT_UNIT,      1,  1 },
   { "undef",    undefID,   undefEX,   0x15, MXS_NO_UNIT,       0,  0 },
   { "dsrlv",    dsrlvID,   dsrlvEX,   0x16, MXS_INT_UNIT,      1,  1 },
   { "dsrav",    dsravID,   dsravEX,   0x17, MXS_INT_UNIT,      1,  1 },
   { "mult",     multID,    multEX,    0x18, MXS_INT_UNIT,      5,  1 },
   { "multu",    multuID,   multuEX,   0x19, MXS_INT_UNIT,      6,  1 },
   { "div",      divID,     divEX,     0x1A, MXS_INT_UNIT,     34,  1 },
   { "divu",     divuID,    divuEX,    0x1B, MXS_INT_UNIT,     34,  1 },
   { "dmult",    dmultID,   dmultEX,   0x1C, MXS_INT_UNIT,      9,  1 },
   { "dmultu",   dmultuID,  dmultuEX,  0x1D, MXS_INT_UNIT,     10,  1 },
   { "ddiv",     ddivID,    ddivEX,    0x1E, MXS_INT_UNIT,     66,  1 },
   { "ddivu",    ddivuID,   ddivuEX,   0x1F, MXS_INT_UNIT,     66,  1 },
   { "add",      addID,     addEX,     0x20, MXS_INT_UNIT,      1,  1 },
   { "addu",     adduID,    adduEX,    0x21, MXS_INT_UNIT,      1,  1 },
   { "sub",      subID,     subEX,     0x22, MXS_INT_UNIT,      1,  1 },
   { "subu",     subuID,    subuEX,    0x23, MXS_INT_UNIT,      1,  1 },
   { "and",      andID,     andEX,     0x24, MXS_INT_UNIT,      1,  1 },
   { "or",       orID,      orEX,      0x25, MXS_INT_UNIT,      1,  1 },
   { "xor",      xorID,     xorEX,     0x26, MXS_INT_UNIT,      1,  1 },
   { "nor",      norID,     norEX,     0x27, MXS_INT_UNIT,      1,  1 },
   { "undef",    undefID,   undefEX,   0x28, MXS_NO_UNIT,       0,  0 },
   { "undef",    undefID,   undefEX,   0x29, MXS_NO_UNIT,       0,  0 },
   { "slt",      sltID,     sltEX,     0x2A, MXS_INT_UNIT,      1,  1 },
   { "sltu",     sltuID,    sltuEX,    0x2B, MXS_INT_UNIT,      1,  1 },
   { "dadd",     daddID,    daddEX,    0x2C, MXS_INT_UNIT,      1,  1 },
   { "daddu",    dadduID,   dadduEX,   0x2D, MXS_INT_UNIT,      1,  1 },
   { "dsub",     dsubID,    dsubEX,    0x2E, MXS_INT_UNIT,      1,  1 },
   { "dsubu",    dsubuID,   dsubuEX,   0x2F, MXS_INT_UNIT,      1,  1 },
   { "tge",      tgeID,     tgeEX,     0x30, MXS_INT_UNIT,      1,  1 },
   { "tgeu",     tgeuID,    tgeuEX,    0x31, MXS_INT_UNIT,      1,  1 },
   { "tlt",      tltID,     tltEX,     0x32, MXS_INT_UNIT,      1,  1 },
   { "tltu",     tltuID,    tltuEX,    0x33, MXS_INT_UNIT,      1,  1 },
   { "teq",      teqID,     teqEX,     0x34, MXS_INT_UNIT,      1,  1 },
   { "undef",    undefID,   undefEX,   0x35, MXS_NO_UNIT,       0,  0 },
   { "tne",      tneID,     tneEX,     0x36, MXS_INT_UNIT,      1,  1 },
   { "undef",    undefID,   undefEX,   0x37, MXS_INT_UNIT,      1,  1 },
   { "dsll",     dsllID,    dsllEX,    0x38, MXS_INT_UNIT,      1,  1 },
   { "undef",    undefID,   undefEX,   0x39, MXS_NO_UNIT,       0,  0 },
   { "dsrl",     dsrlID,    dsrlEX,    0x3A, MXS_INT_UNIT,      1,  1 },
   { "dsra",     dsraID,    dsraEX,    0x3B, MXS_INT_UNIT,      1,  1 },
   { "dsll32",   dsll32ID,  dsll32EX,  0x3C, MXS_INT_UNIT,      1,  1 },
   { "undef",    undefID,   undefEX,   0x3D, MXS_NO_UNIT,       0,  0 },
   { "dsrl32",   dsrl32ID,  dsrl32EX,  0x3E, MXS_INT_UNIT,      1,  1 },
   { "dsra32",   dsra32ID,  dsra32EX,  0x3F, MXS_INT_UNIT,      1,  1 }
};


/* Based on the rt field of bcond opcodes */
MXSOpcodeDesc mxsBcondOpcodes[] = {
   { "bltz",     bltzID,    bltzEX,    0x00, MXS_BRANCH_UNIT,   1,  1 },
   { "bgez",     bgezID,    bgezEX,    0x01, MXS_BRANCH_UNIT,   1,  1 },
   { "bltzl",    bltzlID,   bltzlEX,   0x02, MXS_BRANCH_UNIT,   1,  1 },
   { "bgezl",    bgezlID,   bgezlEX,   0x03, MXS_BRANCH_UNIT,   1,  1 },
   { "spimi",    undefID,   undefEX,   0x04, MXS_NO_UNIT,       0,  0 },
   { "undef",    undefID,   undefEX,   0x05, MXS_NO_UNIT,       0,  0 },
   { "undef",    undefID,   undefEX,   0x06, MXS_NO_UNIT,       0,  0 },
   { "undef",    undefID,   undefEX,   0x07, MXS_NO_UNIT,       0,  0 },
   { "tgei",     tgeiID,    tgeiEX,    0x08, MXS_INT_UNIT,      1,  1 },
   { "tgeiu",    tgeiuID,   tgeiuEX,   0x09, MXS_INT_UNIT,      1,  1 },
   { "tlti",     tltiID,    tltiEX,    0x0A, MXS_INT_UNIT,      1,  1 },
   { "tltiu",    tltiuID,   tltiuEX,   0x0B, MXS_INT_UNIT,      1,  1 },
   { "teqi",     teqiID,    teqiEX,    0x0C, MXS_INT_UNIT,      1,  1 },
   { "undef",    undefID,   undefEX,   0x0D, MXS_NO_UNIT,       0,  0 },
   { "tnei",     tneiID,    tneiEX,    0x0E, MXS_INT_UNIT,      1,  1 },
   { "undef",    undefID,   undefEX,   0x0F, MXS_NO_UNIT,       0,  0 },
   { "bltzal",   bltzalID,  bltzalEX,  0x10, MXS_BRANCH_UNIT,   1,  1 },
   { "bgezal",   bgezalID,  bgezalEX,  0x11, MXS_BRANCH_UNIT,   1,  1 },
   { "bltzall",  bltzallID, bltzallEX, 0x12, MXS_BRANCH_UNIT,   1,  1 },
   { "bgezall",  bgezallID, bgezallEX, 0x13, MXS_BRANCH_UNIT,   1,  1 },
   { "undef",    undefID,   undefEX,   0x14, MXS_NO_UNIT,       0,  0 },
   { "undef",    undefID,   undefEX,   0x15, MXS_NO_UNIT,       0,  0 },
   { "undef",    undefID,   undefEX,   0x16, MXS_NO_UNIT,       0,  0 },
   { "undef",    undefID,   undefEX,   0x17, MXS_NO_UNIT,       0,  0 },
   { "undef",    undefID,   undefEX,   0x18, MXS_NO_UNIT,       0,  0 },
   { "undef",    undefID,   undefEX,   0x19, MXS_NO_UNIT,       0,  0 },
   { "undef",    undefID,   undefEX,   0x1A, MXS_NO_UNIT,       0,  0 },
   { "undef",    undefID,   undefEX,   0x1B, MXS_NO_UNIT,       0,  0 },
   { "undef",    undefID,   undefEX,   0x1C, MXS_NO_UNIT,       0,  0 },
   { "undef",    undefID,   undefEX,   0x1D, MXS_NO_UNIT,       0,  0 },
   { "undef",    undefID,   undefEX,   0x1E, MXS_NO_UNIT,       0,  0 },
   { "undef",    undefID,   undefEX,   0x1F, MXS_NO_UNIT,       0,  0 }
};
