/*
 * Copyright (C) 1996-1998 by the Board of Trustees
 *    of Leland Stanford Junior University.
 * 
 * This file is part of the SimOS distribution. 
 * See LICENSE file for terms of the license. 
 *
 */

#include "ms.h"
#include "memref.h"
#include "cpu_state.h"
#include "cpu.h"

#include "simmisc.h"
#include "remap.h"
#include "machine_params.h"

extern int FlashliteWQFull(int cpu);

/*******************************************************************
 * Interface routines to R4000 simos
 * Basem A. Nayfeh
 ******************************************************************/

bool
InterruptIsPending(struct s_cpu_state *st)
{
  CPUState *P = (CPUState *) (st->mipsyPtr);
  P->CP0[C0_CAUSE] = ((P->CP0[C0_CAUSE] & ~CAUSE_EXTINTBITS) |
                      ((P->intrBitsPtr[0] << CAUSE_IPSHIFT) & CAUSE_EXTINTBITS));
  
  if (((P->CP0[C0_CAUSE] & P->CP0[C0_SR]) & SR_IMASK)
      && (P->CP0[C0_SR] & SR_IEC)
      && !(P->CP0[C0_SR] & (SR_EXL|SR_ERL))) {
    return TRUE;
  }
  return FALSE;
}

Result
ReadICache(int cpuNum, VA vAddr, PA pAddr, Inst *pInst)
{
  return MemRefReadInst(cpuNum,vAddr,pAddr,pInst);
}

/* RPB -- this won't work with SKIP_CACHES defined */

Result
DCacheFetchShared(int cpuNum, VA vAddr, PA pAddr, int type, char **dataPtr) {
   RefSize size;
   RefFlavor flavor = NO_FLAVOR;
   switch (type) {
   case LD_UNSIGNED_B:
   case LD_INTEGER_B:
      size = BYTE_SZ;
      break;
   case LD_UNSIGNED_H:
   case LD_INTEGER_H:
      size = HALF_SZ;
      break;
   case LD_UNSIGNED_L:
   case LD_UNSIGNED_R:
   case LD_UNSIGNED_W:
   case LD_INTEGER_W:
   case LD_FLOAT:
      size = WORD_SZ;
      break;
   case LD_UNSIGNED_DL:
   case LD_UNSIGNED_DR:
   case LD_INTEGER_D:
   case LD_DOUBLE:
      size = DOUBLE_SZ;
      break;
   case LD_INTEGER_LL:
      size = WORD_SZ;
      flavor = LL_FLAVOR;
      break;
   case LD_INTEGER_LLD:
      size = DOUBLE_SZ;
      flavor = LL_FLAVOR;
      break;
   }

   return MemRefReadData(cpuNum, vAddr, pAddr, dataPtr, size, flavor);
}

Result
DCacheFetchExclusive(int cpuNum, VA vAddr, PA pAddr, int type, char **dataPtr) {
   RefSize size;
   RefFlavor flavor = GETX_FLAVOR;
   switch (type) {
   case ST_INTEGER_B:
      size = BYTE_SZ;
      break;
   case ST_INTEGER_H:
      size = HALF_SZ;
      break;
   case ST_INTEGER_L:
   case ST_INTEGER_R:
   case ST_INTEGER_W:
   case ST_FLOAT:
      size = WORD_SZ;
      break;
   case ST_INTEGER_D:
   case ST_DOUBLE:
      size = DOUBLE_SZ;
      break;
   case ST_INTEGER_SC:
      size = WORD_SZ;
      flavor = SC_FLAVOR;
      break;
   case ST_INTEGER_SCD:
      size = DOUBLE_SZ;
      flavor = SC_FLAVOR;
      break;
   }

   return MemRefWriteData(cpuNum, vAddr, pAddr, (uint64) dataPtr, size, flavor);
}

Result
MXSCacheOp(int cpuNum, INST *ip, VA vAddr, PA pAddr, Reg pc)
{
   Result res = SUCCESS;
   uint targetCache = TARGET_CACHE(ip->instr);
   int cacheOp = CACHE_OP(ip->instr);

   switch(targetCache) {
   case CACH_PI:
      if (cacheOp == C_IINV) { 
      } else if (cacheOp == C_HINV) { 
      } else if (cacheOp == C_BARRIER) { 
         /* Cache barrier - mipsy doesn't do anything */
      } else {
         CPUWarning("Bad op 0x%x for the PICache at %#llx\n", cacheOp, (Reg64)pc);
      }
      break;
      
   case CACH_PD:
      if (cacheOp == C_IWBINV) { 
      } else if (cacheOp == C_HWBINV) { 
      } else if (cacheOp == C_HWB) { 

      } else {
         CPUWarning("Bad op 0x%x for the PDCache at %#llx\n", cacheOp, (Reg64)pc);
      }
      break;
      
   case CACH_SI:
       CPUWarning("Bad SI target for cache_op at %#llx\n", (Reg64)pc);
      break;
      
   case CACH_SD:
      if (FlashliteWQFull(cpuNum)) {
         return  STALL;
      }
      switch (cacheOp) {
      case C_IWBINV:
         res = MemRefIndexHitWBInval(cpuNum, vAddr, pAddr);
         break;
         
      case C_HWBINV:
         res = MemRefHitWBInval(cpuNum, vAddr, pAddr);
         break;

      case C_ILT:
      case C_IST:
         break;

      default:
         CPUWarning("BAD CACHE OP at (0x%x) %#llx\n", cacheOp, (Reg64)pc);
         break;
      }
      break;

   default:
      CPUWarning("BAD TARGET CACHE at %#llx\n", (Reg64)pc);
      break;
   }
   return res;
}
