/*
 * Copyright (C) 1996-1998 by the Board of Trustees
 *    of Leland Stanford Junior University.
 * 
 * This file is part of the SimOS distribution. 
 * See LICENSE file for terms of the license. 
 *
 */


/*
 *  ms_stats  -  Collect statistics on a per procedure basis during
 *		the running of the simulator.
 *
 *	Jim Bennett
 *	1994
 */

#include <stdlib.h>
#include "ms.h"
#include "ms_stats.h"
#include "ms_opcodes.h"
#include "cpu.h"
#include "sim_error.h"
#include "cpu_state.h"
#include "cpu_stats.h"
#include "mipsy.h"
#include "hw_events.h"
#include "statrecord.h"

static	double	convert_stat (struct s_cpu_state *st, int st_index);

static void PRINT(char *format, ...)
{
   va_list ap;
   CPUPrint("%MXS:");
   va_start(ap,format);
   vCPUPrint(format,ap);
   va_end(ap);
}

/*****************************************************************
 * STAT RECORD ROUTINES
 *
 * All hardware events that should be tracked by statrecord go here.
 *****************************************************************/
struct MSEventsBucket {
   StatRecordFieldDesc dMXSL1Stall;
   StatRecordFieldDesc dMXSL2Stall;
   StatRecordFieldDesc dMXSUpgradeStall;
   StatRecordFieldDesc iMXSL1Stall;
   StatRecordFieldDesc iMXSL2Stall;
   StatRecordFieldDesc pipelineMXSStall;
   StatRecordFieldDesc MXSGrad;
   struct {
      StatRecordFieldDesc fallthru;
      StatRecordFieldDesc wfallthru;
      StatRecordFieldDesc taken;
      StatRecordFieldDesc wtaken;
      StatRecordFieldDesc indbr;
      StatRecordFieldDesc jret;
      StatRecordFieldDesc likely;
   } branchC, branchI;
} msEventBucket;

#define DEF_FIELD(_field,_fl) (msEventBucket._field = StatRecordDefineField(# _field,_fl))

#define STAT_RECORD(_f, _inc) \
StatRecordEntry(cpuNum, pc, vAddr, msEventBucket._f, _inc);
#define STAT_RECORD_I(_f, _inc) \
StatRecordEntry(cpuNum, pc, 0, msEventBucket._f, _inc);
#define STAT_RECORD_D(_f, _inc) \
StatRecordEntry(cpuNum, 0, vAddr, msEventBucket._f, _inc);

void 
ms_events_init(void)
{
   /* MXS-related counters */
   DEF_FIELD(dMXSL1Stall, STATRECORD_INSTRUCTION|STATRECORD_DATA);
   DEF_FIELD(dMXSL2Stall, STATRECORD_INSTRUCTION|STATRECORD_DATA);
   DEF_FIELD(iMXSL1Stall, STATRECORD_INSTRUCTION);
   DEF_FIELD(iMXSL2Stall, STATRECORD_INSTRUCTION);
   DEF_FIELD(dMXSUpgradeStall, STATRECORD_INSTRUCTION|STATRECORD_DATA);
   DEF_FIELD(pipelineMXSStall, STATRECORD_INSTRUCTION);
   DEF_FIELD(MXSGrad, STATRECORD_INSTRUCTION);

   DEF_FIELD(branchC.fallthru, STATRECORD_INSTRUCTION);
   DEF_FIELD(branchC.wfallthru, STATRECORD_INSTRUCTION);
   DEF_FIELD(branchC.taken, STATRECORD_INSTRUCTION);
   DEF_FIELD(branchC.wtaken, STATRECORD_INSTRUCTION);
   DEF_FIELD(branchC.indbr, STATRECORD_INSTRUCTION);
   DEF_FIELD(branchC.jret, STATRECORD_INSTRUCTION);
   DEF_FIELD(branchC.likely, STATRECORD_INSTRUCTION);

   DEF_FIELD(branchI.fallthru, STATRECORD_INSTRUCTION);
   DEF_FIELD(branchI.wfallthru, STATRECORD_INSTRUCTION);
   DEF_FIELD(branchI.taken, STATRECORD_INSTRUCTION);
   DEF_FIELD(branchI.wtaken, STATRECORD_INSTRUCTION);
   DEF_FIELD(branchI.indbr, STATRECORD_INSTRUCTION);
   DEF_FIELD(branchI.jret, STATRECORD_INSTRUCTION);
   DEF_FIELD(branchI.likely, STATRECORD_INSTRUCTION);
}

void
ms_attribute_grad(int cpuNum, int pc, int stall)
{
   STAT_RECORD_I(MXSGrad, stall);
}

void
ms_attribute_branch(int cpuNum, int pc, int type)
{
   switch(type) {
   case ST_CORRECT_FALLTHRU:
      STAT_RECORD_I(branchC.fallthru, 1);
      break;
   case ST_CORRECT_W_FALLTHRU:
      STAT_RECORD_I(branchC.wfallthru, 1);
      break;
   case ST_CORRECT_W_TAKEN:
      STAT_RECORD_I(branchC.wtaken, 1);
      break;
   case ST_CORRECT_TAKEN:
      STAT_RECORD_I(branchC.taken, 1);
      break;
   case ST_CORRECT_IND_BR:
      STAT_RECORD_I(branchC.indbr, 1);
      break;
   case ST_CORRECT_JRET:
      STAT_RECORD_I(branchC.jret, 1);
      break;
   case ST_CORRECT_LIKELY:
      STAT_RECORD_I(branchC.likely, 1);
      break;

   case ST_INCORRECT_FALLTHRU:
      STAT_RECORD_I(branchI.fallthru, 1);
      break;
   case ST_INCORRECT_W_FALLTHRU:
      STAT_RECORD_I(branchI.wfallthru, 1);
      break;
   case ST_INCORRECT_TAKEN:
      STAT_RECORD_I(branchI.taken, 1);
      break;
   case ST_INCORRECT_W_TAKEN:
      STAT_RECORD_I(branchI.wtaken, 1);
      break;
   case ST_INCORRECT_IND_BR:
      STAT_RECORD_I(branchI.indbr, 1);
      break;
   case ST_INCORRECT_JRET:
      STAT_RECORD_I(branchI.jret, 1);
      break;
   case ST_INCORRECT_LIKELY:
      STAT_RECORD_I(branchI.likely, 1);
      break;
   default: 
      ASSERT(0);
   }
   return;
}

void 
ms_attribute_stall(int cpuNum, int vAddr, int pc, int stall, int stallType)
{
   if (stallType & E_PIPELINE) {
      STAT_RECORD_I(pipelineMXSStall,stall);
      return;
   }
   if (stallType & E_D) {
      if (stallType & E_L1) { 
         STAT_RECORD(dMXSL1Stall,stall);
         return;
      }
      if (((stallType & E_L1) && (stallType & E_UPGRADE)) 
          || ((stallType & E_L2) && (stallType & E_UPGRADE))) {
         STAT_RECORD(dMXSUpgradeStall,stall);
         return;
      }
      ASSERT(stallType & E_L2);
      STAT_RECORD(dMXSL2Stall,stall);
   } else if (vAddr == pc) {
      ASSERT(stallType & E_I);
      if (stallType & E_L1) {
         STAT_RECORD(iMXSL1Stall,stall);
      } else { 
         ASSERT(stallType & E_L2);
         STAT_RECORD(iMXSL2Stall,stall);
      }
   } else {
      CPUWarning("MemStat: no stall reason found for miss\n");
      CPUWarning("         should be a transient mipsy->mxs problem\n");
   }
}

/*----------------------------------------------------------------------*/
/*									*/
/*	ms_exec never returns, but exits when the app calls exit.	*/
/*	But just before exiting, the following routine gets called	*/
/*	to print out the collected statistics.				*/
/*									*/
/*----------------------------------------------------------------------*/

void print_stats (struct s_cpu_state *st)
{
}


/*
 *  convert_stat  -  Convert statistic to double precision
 *		     floating point number.
 */

static double convert_stat (struct s_cpu_state *st, int st_index)
{
   double	t;

   t = (double) (st->stats[st_index]);
   return (t);
}

void dump_stats (struct s_cpu_state *st)
{
   CPUState *P = (CPUState *) (st->mipsyPtr);
   int	i;

   if ((st->work_ticks == 0) && (st->work_cycle == 0)) {
      /* Must not of run yet. */
      return;
   }

   PRINT( "DUMPSTATS: CPU %d @ %lld %9d%0.5d\n", P->myNum,
          (uint64)MipsyReadTime(P->myNum), st->work_ticks, st->work_cycle);

/* Then output gathered statistics		*/

#define PRINT_IF_NOT_ZERO(_format, _name, _value) \
   if ((_value) != 0) PRINT(_format, _name, (double)(_value))

   for (i = 0; i < ST_NTYPES; i++) {
      if (sh_stat_names[i][0] == 0) continue;
      PRINT_IF_NOT_ZERO( "%s: %.0f\n", sh_stat_names[i], convert_stat(st,i));
   }

#undef PRINT_IF_NOT_ZERO
#define PRINT_IF_NOT_ZERO(_format, _value) \
   if ((_value) != 0) PRINT(_format, (double)(_value))
                          
   PRINT_IF_NOT_ZERO("SAMPLE: %.0f\n",st->sample_stats.samples);

#undef PRINT_IF_NOT_ZERO
#define PRINT_IF_NOT_ZERO(_format, _index, _value) \
   if ((_value) != 0) PRINT(_format, _index, (double)(_value))

   for (i = 0; i < (sizeof(st->sample_stats.nthreads_hist) / 
                    sizeof(st->sample_stats.nthreads_hist[0])); i++) {                
      PRINT_IF_NOT_ZERO("NTHREADS_HIST: %d %.0f\n",i,st->sample_stats.nthreads_hist[i]);
   }
   for (i = 0; i < (sizeof(st->sample_stats.reg_hist) / 
                    sizeof(st->sample_stats.reg_hist[0])); i++) {                
      PRINT_IF_NOT_ZERO("REGS_HIST: %d %.0f\n",i,st->sample_stats.reg_hist[i]);
   }
   for (i = 0; i < (sizeof(st->sample_stats.iwin_hist) / 
                    sizeof(st->sample_stats.iwin_hist[0])); i++) {                
      PRINT_IF_NOT_ZERO("IWIN_INST_HIST: %d %.0f\n",i,
                        st->sample_stats.iwin_hist[i].inst);
      PRINT_IF_NOT_ZERO("IWIN_SPECINST_HIST: %d %.0f\n",i,
                        st->sample_stats.iwin_hist[i].specInst);
      PRINT_IF_NOT_ZERO("IWIN_LDST_HIST: %d %.0f\n",i,
                        st->sample_stats.iwin_hist[i].ldstInst);
      PRINT_IF_NOT_ZERO("IWIN_LDSTDEP_HIST: %d %.0f\n",i,
                        st->sample_stats.iwin_hist[i].ldstdepInst);
      PRINT_IF_NOT_ZERO("IWIN_REGDEP_HIST: %d %.0f\n",i,
                        st->sample_stats.iwin_hist[i].regdepInst);
      PRINT_IF_NOT_ZERO("IWIN_SQUASH_HIST: %d %.0f\n",i,
                        st->sample_stats.iwin_hist[i].squashInst);
   }
   for (i = 0; i < (sizeof(st->sample_stats.ldst_hist) / 
                    sizeof(st->sample_stats.ldst_hist[0])); i++) {                
      PRINT_IF_NOT_ZERO("LDST_INST_HIST: %d %.0f\n",i,
                        st->sample_stats.ldst_hist[i].inst);
      PRINT_IF_NOT_ZERO("LDST_DONE_INST_HIST: %d %.0f\n",i,
                        st->sample_stats.ldst_hist[i].doneInst);
      PRINT_IF_NOT_ZERO("LDST_PEND_INST_HIST: %d %.0f\n",i,
                        st->sample_stats.ldst_hist[i].pendInst);
      PRINT_IF_NOT_ZERO("LDST_CONFLICT_INST_HIST: %d %.0f\n",i,
                        st->sample_stats.ldst_hist[i].conflictInst);
      PRINT_IF_NOT_ZERO("LDST_FAILED_INST_HIST: %d %.0f\n",i,
                        st->sample_stats.ldst_hist[i].failedInst);
      PRINT_IF_NOT_ZERO("LDST_STALL_INST_HIST: %d %.0f\n",i,
                        st->sample_stats.ldst_hist[i].stallInst);
   }
}

void MxsPrintStatus(struct s_cpu_state *st) 
{
   double grad_insts, grad_squashed, cond_br_correct, cond_br_incorrect, 
      ind_br_correct, ind_br_incorrect;
   double igrad_insts, igrad_squashed, icond_br_correct, icond_br_incorrect, 
      iind_br_correct, iind_br_incorrect;

   CPUState *P = (CPUState *) (st->mipsyPtr);
   static struct { 
      double grad_insts, grad_squashed, cond_br_correct, cond_br_incorrect, 
         ind_br_correct, ind_br_incorrect;
       
   } old[MIPSY_MAX_CPUS];
 
   if ((st->work_ticks == 0) && (st->work_cycle == 0)) {
      /* Must not of run yet. */
      return;
   }


   cond_br_correct = convert_stat (st, ST_CORRECT_FALLTHRU) + 
      convert_stat (st, ST_CORRECT_W_FALLTHRU) + 
      convert_stat (st, ST_CORRECT_W_TAKEN) +
      convert_stat (st, ST_CORRECT_TAKEN);

   cond_br_incorrect = convert_stat (st, ST_INCORRECT_FALLTHRU) + 
      convert_stat (st, ST_INCORRECT_W_FALLTHRU) + 
      convert_stat (st, ST_INCORRECT_W_TAKEN) +
      convert_stat (st, ST_INCORRECT_TAKEN);

   ind_br_correct = convert_stat (st, ST_CORRECT_IND_BR);
   ind_br_incorrect = convert_stat (st, ST_INCORRECT_IND_BR);

   grad_insts = (double) STATS_VALUE(P->myNum, numInstructions);
   grad_squashed = convert_stat (st, ST_GRAD_SQUASHED);
   icond_br_correct = cond_br_correct - old[P->myNum].cond_br_correct;
   icond_br_incorrect = cond_br_incorrect - old[P->myNum].cond_br_incorrect;
   iind_br_correct = ind_br_correct - old[P->myNum].ind_br_correct;
   iind_br_incorrect = ind_br_incorrect - old[P->myNum].ind_br_incorrect;
   igrad_insts = grad_insts - old[P->myNum].grad_insts;
   igrad_squashed = grad_squashed - old[P->myNum].grad_squashed;

   PRINT ("CPU%d BR (%3.1f%%) JR (%3.1f%%) Squash (%3.1f%%)\n",P->myNum,
          100.0*icond_br_correct/(icond_br_correct + icond_br_incorrect + .0000001),
          100.0*iind_br_correct/(iind_br_correct + iind_br_incorrect + .0000001),
          100.0*igrad_squashed/(igrad_squashed + igrad_insts + .0000001));

   old[P->myNum].grad_insts = grad_insts;
   old[P->myNum].grad_squashed = grad_squashed;
   old[P->myNum].cond_br_correct = cond_br_correct;
   old[P->myNum].cond_br_incorrect = cond_br_incorrect;
   old[P->myNum].ind_br_correct = ind_br_correct;
   old[P->myNum].ind_br_incorrect = ind_br_incorrect;
}


void 
MxsSampleStats(struct s_cpu_state *st) 
{
   int inum, r;
   struct s_ldst_buffer *entry;
   int  instInWindow , specInstInWindow , ldstInstInWindow ,
      ldstdepInstInWindow , regdepInstInWindow , squashInstInWindow;
   int  ldstBufferInst, ldstDoneInst , ldstConflictInst , ldstFailedInst, 
      ldstStallInst,  ldstPendInst;
   int regBusyCount;


   st->sample_stats.samples++;

   st->sample_stats.nthreads_hist[st->nthreads]++;

   regBusyCount = 0;
   for (r = 0; r < NUM_PHYS_REGS; r++) {
      if (st->reg_rstat[r].reg_status & REG_BUSY)
         regBusyCount++;
   }
   st->sample_stats.reg_hist[regBusyCount]++;

 
   instInWindow = specInstInWindow =  ldstInstInWindow =
      ldstdepInstInWindow = regdepInstInWindow = squashInstInWindow = 0;


   inum = st->iwin_headgrad; 
   while (inum >= 0) {
      int flags = st->iwin_flags[inum];
      instInWindow++;
      if (flags & IWIN_SPEC) 
         specInstInWindow++;
      if (flags & IWIN_LDST) 
         ldstInstInWindow++;
      if (flags & IWIN_LDST_DEP) 
         ldstdepInstInWindow++;
      if (flags & (IWIN_DEPS))
         regdepInstInWindow++;
      if (flags & IWIN_SQUASH) 
         squashInstInWindow++;

      inum = st->iwin_grad[inum];
   }

   st->sample_stats.iwin_hist[instInWindow].inst++;
   st->sample_stats.iwin_hist[specInstInWindow].specInst++;
   st->sample_stats.iwin_hist[ldstInstInWindow].ldstInst++;
   st->sample_stats.iwin_hist[ldstdepInstInWindow].ldstdepInst++;
   st->sample_stats.iwin_hist[regdepInstInWindow].regdepInst++;
   st->sample_stats.iwin_hist[squashInstInWindow].squashInst++;


   ldstBufferInst = ldstDoneInst = ldstConflictInst = ldstFailedInst = 
      ldstStallInst = ldstPendInst = 0;
   for (entry = st->ldst_head; entry; entry = entry->next) {
      int status = entry->ls->status;
      ldstBufferInst++;
      if (status & LS_ST_DONE) {
         ldstDoneInst++;
         continue;
      }
      if (status & LS_ST_PEND) 
         ldstPendInst++;
      if (status & LS_ST_CONFLICT) 
         ldstConflictInst++;
      if (status & LS_ST_FAILED) 
         ldstFailedInst++;
      if (status & LS_ST_STALL) 
         ldstStallInst++;
   }

   st->sample_stats.ldst_hist[ldstBufferInst].inst++;
   st->sample_stats.ldst_hist[ldstDoneInst].doneInst++;
   st->sample_stats.ldst_hist[ldstPendInst].pendInst++;
   st->sample_stats.ldst_hist[ldstConflictInst].conflictInst++;
   st->sample_stats.ldst_hist[ldstFailedInst].failedInst++;
   st->sample_stats.ldst_hist[ldstStallInst].stallInst++;
}

char printIwinMask[SIM_MAXCPUS];

void
dump_iwin_entry(struct s_cpu_state *st, int inum)
{
   static int spec = 0;
   static int branch = 0;
   static char branch_path[6] = "-----";
   
   int flags = st->iwin_flags[inum];
   int r;

   if (inum == st->iwin_headgrad) {
      spec = 0;
      branch = 0;
      strcpy(branch_path, "-----");
   }
   if (branch != st->iwin_br_node[inum]) {
      if (st->iwin_br_node[inum] == -1) {
         branch_path[spec] = 'X';
      } else if (branch == -1) {
         /* RPB hack -- resuming after an incorrect speculation */
         branch_path[spec] = '-';
      } else if (st->branch_tree[branch].lchild == st->iwin_br_node[inum]) {
         ASSERT(spec < 6);
         branch_path[spec++] = 'N';
      } else if (st->branch_tree[branch].rchild == st->iwin_br_node[inum]) {
         ASSERT(spec < 6);
         branch_path[spec++] = 'T';
      } else {
         ASSERT(0);
      }
      branch = st->iwin_br_node[inum];
   }

#ifdef PRINT_INSTRUCTIONS
   CPUPrint("<IWIN> entry %2d %#llx 0x%08x %-8s %-5s", 
            inum, (uint64)st->iwin_pc[inum], 
            st->iwin[inum].instr,
            mxsOpcodes[MAJOR_OP(st->iwin[inum].instr)].opname,
            branch_path);
#else
   CPUPrint("<IWIN> entry %2d %#llx %-8s %-5s", 
            inum, (uint64)st->iwin_pc[inum], 
            opcode_mnemonics[st->iwin[inum].op],
            branch_path);
#endif

   for (r = 0; r < MAX_DST_REGS; r++) {
      RegInfo *ri = &st->iwin[inum].dst[r];
      if (ri->flags == 0) {
         CPUPrint("  --- ");
      } else {
         CPUPrint("  %3d ", ri->preg);
      }
   }
   for (r = 0; r < MAX_SRC_REGS; r++) {
      RegInfo *ri = &st->iwin[inum].src[r];
      if (ri->flags == 0) {
         CPUPrint("  --- ");
      } else if (flags & IWIN_DEP(r)) {
         CPUPrint(" <%3d>", ri->preg);
      } else {
         CPUPrint("  %3d ", ri->preg);
      }
   }

   /* Pipeline stage information */
   if (flags & IWIN_SQUASH) {
      CPUPrint(" S");
   } else if (flags & IWIN_FREED) {
      CPUPrint(" F");
   } else if (flags & IWIN_ISSUED) {
      CPUPrint(" I");
      if (flags & IWIN_FAULT)
         CPUPrint(" FAULT");
      else if (flags & IWIN_TLBFAULT)
         CPUPrint(" TLBFAULT");
   } else if (flags & IWIN_AVAIL) {
      CPUPrint(" A");
   } else {
      CPUPrint(" -");
      if (flags & (IWIN_DEPS))
         CPUPrint(" DEP");
      else if (flags & IWIN_FLUSH)
         CPUPrint(" FLUSH");
      else if (flags & IWIN_CTL)
         CPUPrint(" CTL");
      else if (flags & IWIN_BRDLY)
         CPUPrint(" BRDLY");
      else if (flags & IWIN_LDST_DEP)
         CPUPrint(" LDST_DEP");
      else
         ASSERT(0);
   }

#ifdef rpb_notdef
   /* Flags that we don't care about for the time being */
   if (flags & IWIN_LDST)
      CPUPrint(" L");
   if (flags & IWIN_LDSTBUF)
      CPUPrint("B");

   if (flags & IWIN_BRANCH)
      CPUPrint(" B");
   if (flags & IWIN_TAKENBR)
      CPUPrint("T");

   if (flags & IWIN_UNCACHED)
      CPUPrint(" UNCACHED");
   if (flags & IWIN_ADDR_VALID)
      CPUPrint(" ADDR_VALID");
   if (flags & IWIN_STORE)
      CPUPrint(" STORE");
#endif
 
#ifdef rpb_notdef
   /* Flags that are redundant or irrelevant */
   if (flags & IWIN_BUSY)
      CPUPrint(" BUSY");
   if (flags & IWIN_SPEC)
      CPUPrint(" SPEC");
#endif
   
   CPUPrint("\n");
}
